#!/usr/bin/bash

#
# Copyright (c) 2014, 2021, Oracle and/or its affiliates.
#

# This script builds the opensource components in the opensource source
# archive whose license requires that they be buildable.

MAKE="/usr/bin/make"

Error() {
    echo "ERROR: $@" >&2
    exit 1
}

RunMake() {
    echo "*** Building $1..."
    echo "Building $1 in $(pwd):" >>$LOG
    shift
    $MAKE "$@" >>$LOG 2>&1 || Error "Build failed, see $LOG for details."
}

REQ_PKGS="
    pkg:/developer/assembler
    pkg:/developer/build/make
    pkg:/developer/build/onbld
    pkg:/developer/developerstudio-126/cc
    pkg:/developer/gcc/gcc-c-10
    pkg:/developer/gcc/gcc-c++-10
    pkg:/developer/lexer/flex
    pkg:/developer/parser/bison
    pkg:/file/gnu-coreutils
    pkg:/system/header
    pkg:/text/gnu-gettext
"

if [ "$(uname -p)" = "i386" ] ; then
    # Additional packages required to build GRUB
    REQ_PKGS+="pkg:/developer/build/pkg-config
	pkg:/system/library/freetype-2
    	"
fi

SKIP_CHECK="no"

while getopts o:s c; do
    case $c in
    o)	export LOG="$OPTARG"
	;;
    s)	SKIP_CHECK="yes"
	;;
    ?)	print "Usage: ${0##*/} [ -o output ] [ -s ]" 1>&2
	exit 2
	;;
    esac
done
shift $(expr $OPTIND - 1)

if [ "${0%/*}" = "$0" ] ; then
    FOSS_TOP=$(pwd)
else
    FOSS_TOP=$(cd ${0%/*} ; pwd)
fi

if [ -z "$LOG" ]; then
    export LOG=$FOSS_TOP/foss_build.$(date +%Y.%m.%d-%H:%M:%S).out
fi

export WS=$FOSS_TOP
export SRC=$FOSS_TOP/usr/src
export MACH=$(uname -p)
export ROOT=$FOSS_TOP/data/proto/root_$MACH
export STUBROOT=$FOSS_TOP/data/proto/root_$MACH
export ENVLDLIBS1="-L$ROOT/lib -L$ROOT/usr/lib"
export STUBENVLDLIBS1="-L$STUBROOT/lib -L$STUBROOT/usr/lib"
export BIP="yes"
export SPRO_ROOT=${SPRO_ROOT:-/opt}
case $MACH in
    i386) export MACH64="amd64" ;;
    sparc) export MACH64="sparcv9" ;;
esac
export ELFSIGN_O=\#
export ZASSERTDEFLIB=
export MAKEFLAGS=-e
export BUILD_TOOLS=/opt
# Did not export CC here as this will cause an error in the grub build
CC=$BUILD_TOOLS/developerstudio12.6/bin/cc
export CW_CC=${CC}
export NOT_FOSS_BUILD=\#
export POUND_SIGN=\#
export FISH_BUILD=\#

echo "Source root: $SRC"
rm -f $LOG ; touch $LOG
echo "Logging to:  $LOG"

if [ "$SKIP_CHECK" = "no" ]; then
    echo "Checking for required packages..."
    for pkg in $REQ_PKGS ; do
        pkg info -l -q $pkg
        if [ $? != 0 ] ; then
            Error "Required package $pkg is not installed!"
        fi
    done
fi

if [ ! -x $CC ] ; then
    Error "Could not find compiler: $CC!"
fi

if [ "$MACH" = "i386" ] ; then
    cd $SRC/grub
    RunMake Grub CCYFLAG="-I/usr/include"

    cd $SRC/lib/libparted
    RunMake libparted install
    cd $SRC/cmd/parted
    RunMake parted install

    cd $SRC/lib/libntfs
    RunMake libntfs install
    cd $SRC/cmd/ntfsprogs
    RunMake ntfsprogs install

fi

UTSDIR="$MACH"
if [ "$UTSDIR" = "i386" ]; then
    UTSDIR="intel"
fi

cd $SRC/uts/$UTSDIR/ecc
mkdir -p ../../common/crypto/io ../../../common/crypto/fips/
for i in ../../common/crypto/io/ecc.c \
    ../../../common/crypto/fips/fips_ecc_util.c
do
	echo "static int errno; /* dummy $i src */" >$i
done
RunMake ecc-module CCYFLAG=-I install

cd $SRC/../../
mkdir -p $ROOT/usr/include/hal $ROOT/usr/include/libpolkit
cp $SRC/lib/hal/libhal/common/libhal.h \
    $SRC/lib/hal/libhal-storage/common/libhal-storage.h \
    $ROOT/usr/include/hal
cp $SRC/lib/policykit/libpolkit/common/libpolkit.h \
    $ROOT/usr/include/libpolkit
cd $SRC/cmd/hal/tools
RunMake hal-storage install

cd $SRC/cmd/policykit
RunMake policykit install

cd $SRC/cmd/powertop
RunMake powertop install

cd $SRC/cmd/mail.local
RunMake mail.local install

cd $SRC/lib/hbaapi
RunMake hbaapi INS=$FOSS_TOP/stash install_h install

cd $SRC/lib/libfuse
rm -f libavl.so; ln -s /lib/libavl.so.1 libavl.so
rm -f $MACH64/libavl.so; ln -s /lib/$MACH64/libavl.so.1 $MACH64/libavl.so
RunMake libfuse ENVLDLIBS2="-L$SRC/lib/libfuse" install

cd $SRC/lib/libima
RunMake libima INS=$FOSS_TOP/stash install

cd $SRC/lib/librstp
RunMake librstp install

cd $SRC/lib/mpapi/libmpapi
RunMake libmpapi INS=$FOSS_TOP/stash install

cd $SRC/lib/pkcs11/pkcs11_tpm
mkdir -p $ROOT/lib/svc/manifest/system
RunMake pkcs11_tpm install

cd $SRC/lib/smhba
RunMake smhba INS=$FOSS_TOP/stash INCS="-I $ROOT/usr/include" install

echo "Build complete."
