/***************************************************************************
 *
 * devinfo_cpu : cpu devices
 *
 * Copyright (c) 2008, 2016, Oracle and/or its affiliates. All rights reserved.
 *
 * Licensed under the Academic Free License version 2.1
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <kstat2.h>
#include <sys/utsname.h>
#include <libdevinfo.h>
#include <sys/systeminfo.h>

#include "../osspec.h"
#include "../logger.h"
#include "../hald.h"
#include "../hald_dbus.h"
#include "../device_info.h"
#include "../util.h"
#include "devinfo_cpu.h"

static HalDevice *devinfo_cpu_add(HalDevice *, di_node_t, char *, char *);

DevinfoDevHandler devinfo_cpu_handler = {
	devinfo_cpu_add,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/*
 * During probing CPUs, hald only opens
 * kstat once and keeps it all the way.
 */
static kstat2_handle_t kc = NULL;
static kstat2_matcher_list_t mlist = NULL;

static di_prom_handle_t phdl = DI_PROM_HANDLE_NIL;

#define	HALD_PROM_CFGHDL_TO_CPUID(x)	(x & ~(0xful << 28))

/*
 * This variable is used to check if current platform is sun4v.
 *
 * -1 means this variable has not been initialized.
 *  1 means current platform is sun4v.
 *  0 means current platform is not sun4v.
 */
static int		is_sun4v = -1;

#define	CPU_INFO_KSTAT_GLOB	"/system/cpu/*/info"
#define	CPU_INFO_KSTAT_PATH_FMT	"/system/cpu/%d/info"

static HalDevice *
devinfo_cpu_add(HalDevice *parent, di_node_t node, char *devfs_path, char *device_type)
{

	HalDevice	*d;
	char		*prom_device_type = NULL;
	int		*int_cpu_id;
	static int	cpu_id = -1;
	uint64_t	clock_mhz;
	kstat2_status_t	status;
	kstat2_map_t	map;
	kstat2_nv_t	val;
	char		kspath[KSTAT2_PATHLEN];
	dbus_bool_t	is_supp_freqs;
	char		udi[HAL_PATH_MAX];
	char		*driver_name, *s;
	char		cpu_devfs_path[HAL_PATH_MAX];
	char		*node_name;
	char		machine[MAXNAMELEN];

	/*
	 * If it is x86, the software device tree node will have the
	 * device_type information which is the one passed above. If it is
	 * NULL, check if the node has a PROM entry, and check the device_type
	 * in case of sparc. Else return NULL
	 */
	if (device_type == NULL) {
		if (is_sun4v == -1) {
			if (sysinfo (SI_MACHINE, machine, sizeof (machine)) == -1) {
				HAL_ERROR (("Error in getting system info "
				    "handle to add cpu device: %s",
				    strerror (errno)));
				return (NULL);
			} else {
				if (strncmp (machine, "sun4v", MAXNAMELEN) == 0)
					is_sun4v = 1;
				else
					return (NULL);
			}
		}

		/*
		 * Check the device type if it has a PROM entry. Because
		 * in sparc, the property entry will in the PROM node
		 */
		if (di_nodeid (node) != DI_PROM_NODEID) {
			return (NULL);
		}

		/*
		 * On a sun4v architecture machine, the index to lookup
		 * CPU id in PROM entry is 'reg'.
		 */
		/*
		 * On some sun4v platforms, e.g. T3-4/T4-4, the
		 * PROM lookup is often slow, especially for string
		 * type data. So we use the device node name to
		 * check a CPU device instead of device_type in PROM
		 * entry to reduce the time cost.
		 */
		node_name = di_node_name (node);
		if (strcmp (node_name, "cpu") != 0) {
			return (NULL);
		}

		if (phdl == DI_PROM_HANDLE_NIL) {
			phdl = di_prom_init ();
			if (phdl == DI_PROM_HANDLE_NIL) {
				HAL_ERROR (("Error in Initializing the PROM "
				    "handle to find cpu device: %s",
				    strerror (errno)));
				return (NULL);
			}
		}

		/*
		 * Get cpuid if available
		 */
		if (di_prom_prop_lookup_ints (phdl, node,
		    "reg", &int_cpu_id) > 0) {
			cpu_id = HALD_PROM_CFGHDL_TO_CPUID((*int_cpu_id));
		} else {
			/*
			 * There is no cpuid entry in this arch.Just
			 * increment the cpuid which will be the
			 * current instance
			 */
			++cpu_id;
		}

	} else if (strcmp (device_type, "cpu") == 0) {
		/*
		 * This is a x86 arch, because software device tree node
		 * has the device_type entry for cpu. The "reg" property
		 * will have the cpuid. If not just increment the cpuid
		 * which will be the current cpu instance in the kstat
		 */
		if (di_prop_lookup_ints (DDI_DEV_T_ANY, node,
		    "reg", &int_cpu_id) > 0) {
			cpu_id = *int_cpu_id;
		} else {
			/*
			 * There is no cpuid entry in this arch. Just
			 * increment the cpuid which will be the
			 * current instance
			 */
			++cpu_id;
		}

	} else {
		return (NULL);
	}

	HAL_DEBUG (("CPUID=> %x", cpu_id));

	d = hal_device_new ();

	/*
	 * devinfo_set_default_properties () uses di_instance() as part of
	 * the udi. For some solaris devices like cpu di_instance() is not
	 * present and it returns -1. For the udi to be unique can use the
	 * cpu_id.
	 */
	hal_device_property_set_string (d, "info.parent",
	    "/org/freedesktop/Hal/devices/local");

	/*
	 * If cpu driver is not installed, then devfs_path returned by
	 * libdevinfo will be same for all cpu's.
	 * Since HAL stores the devices in its tree based on the devfs_path,
	 * To make it unique, will be concatenating devfs_path with cpu_id
	 */
	if (di_driver_name (node) == NULL) {
		snprintf (cpu_devfs_path, HAL_PATH_MAX, "%s_%d",
		    devfs_path, cpu_id);
	} else {
		snprintf (cpu_devfs_path, HAL_PATH_MAX, "%s", devfs_path);
	}

	HAL_DEBUG(("DevfsPath=> %s, CPUID=> %d", cpu_devfs_path, cpu_id));

	hal_util_compute_udi (hald_get_gdl (), udi, sizeof (udi),
	    "/org/freedesktop/Hal/devices%s_%d", cpu_devfs_path, cpu_id);
	hal_device_set_udi (d, udi);
	hal_device_property_set_string (d, "info.udi", udi);
	if (di_prop_lookup_strings (DDI_DEV_T_ANY, node, "model", &s) > 0) {
		hal_device_property_set_string (d, "info.product", s);
	} else {
		hal_device_property_set_string (d, "info.product",
		    di_node_name (node));
	}
	hal_device_property_set_string (d, "solaris.devfs_path",
	    cpu_devfs_path);
	if ((driver_name = di_driver_name (node)) != NULL) {
		hal_device_property_set_string (d, "info.solaris.driver",
		    driver_name);
	}

	hal_device_add_capability (d, "processor");

	hal_device_property_set_int (d, "processor.number", cpu_id);

	/*
	 * Get the cpu related info from the kstat
	 */
	if (kc != NULL) {
		/*
		 * The kstat connection is already open but we don't
		 * know when the tree was last updated, so call for an
		 * update. If this fails, close the kstat connection and
		 * allow the code below to attempt to reestablish it.
		 */
		if ((status = kstat2_update(kc)) != KSTAT2_S_OK) {
			(void) kstat2_close(&kc);
			(void) kstat2_free_matcher_list(&mlist);
		}
	}

	if (kc == NULL) {
		/*
		 * To reduce the load on the kernel, try to restrict the
		 * kstat tree to just the cpu kstats.
		 */
		status = kstat2_alloc_matcher_list(&mlist);
		if (status == KSTAT2_S_OK) {
			(void) kstat2_add_matcher(KSTAT2_M_GLOB,
			    CPU_INFO_KSTAT_GLOB, mlist);
			if ((status = kstat2_open(&kc, mlist)) != KSTAT2_S_OK) {
				(void) kstat2_free_matcher_list(&mlist);
				HAL_ERROR(("Could not open kstat to get "
				    "cpu info: %s",
				    kstat2_status_string(status)));
			}
		} else {
			HAL_ERROR(("Failed to alloc kstat matcher list: %s",
			    kstat2_status_string(status)));
		}

		if (status != KSTAT2_S_OK) {
			goto next;
		}
	}

	/* Try reading the cpu info kstat */
	(void) snprintf(kspath, sizeof (kspath),
	    CPU_INFO_KSTAT_PATH_FMT, cpu_id);

	status = kstat2_lookup_map(kc, kspath, &map);
	if (status != KSTAT2_S_OK) {
		HAL_ERROR (("Could not find cpu info kstat \"%s\": %s",
		    kspath, kstat2_status_string(status)));
		(void) kstat2_close(&kc);
		(void) kstat2_free_matcher_list(&mlist);
		return (NULL);
	}

	status = kstat2_map_get(map, "clock_MHz", &val);
	if (status != KSTAT2_S_OK) {
		HAL_ERROR (("Could not get kstat clock_MHz data for cpu: %s",
		    kstat2_status_string(status)));
		goto next;
	}
	clock_mhz = val->kstat2_integer;

	if (hal_device_property_set_uint64 (d, "processor.maximum_speed",
	    clock_mhz) == FALSE) {
		HAL_INFO (("Could not set the processor speed device prop"));
	}


	status = kstat2_map_get(map, "supported_frequencies_Hz", &val);
	if (status != KSTAT2_S_OK) {
		HAL_INFO (("Could not get kstat supported_frequencies_Hz data"
		    " for cpu: %s", kstat2_status_string(status)));
		is_supp_freqs = FALSE;
	} else {
		/*
		 * Frequencies are in an array.
		 * If there are 0 or 1 elements in the array
		 * there is only a single frequency.
		 */
		if (val->kstat2_integers.length <= 1) {
			is_supp_freqs = FALSE;
		} else {
			is_supp_freqs = TRUE;
		}
	}

	if (hal_device_property_set_bool (d, "processor.can_throttle",
	    is_supp_freqs) == FALSE) {
		HAL_INFO (("Could not set the processor.can_throttle"
		    " device prop"));
	}

next:
	devinfo_add_enqueue (d, cpu_devfs_path, &devinfo_cpu_handler);
	return (d);
}
