/*
 * Copyright 2009, Intel Corporation
 * Copyright (c) 2009, 2016, Oracle and/or its affiliates. All rights reserved.
 *
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 *	Eric C Saxe <eric.saxe@sun.com>
 *	Aubrey Li <aubrey.li@intel.com>
 */


#include <unistd.h>
#include <stropts.h>
#include <string.h>
#include <kstat.h>
#include <errno.h>
#include <sys/acpi_drv.h>
#include "powertop.h"

#define	mW2W(value)	((value) / 1000)

typedef struct battery_state {
	uint32_t exist;
	uint32_t power_unit;
	uint32_t bst_state;
	double present_rate;
	double remain_cap;
	double last_cap;
} battery_state_t;

static battery_state_t	battery_state;

static int		pt_battery_stat_snapshot(int fd);

void
pt_battery_print(int fd)
{

	int err;

	/*
	 * This idicates that the open of /dev/acpi_drv failed because
	 * battery information is not available on this system
	 */
	if (fd < 0)
		return;

	(void) memset(&battery_state, 0, sizeof (battery_state_t));

	/*
	 * The return value of pt_battery_stat_snapshot() can be used for
	 * debug or to show/hide the acpi power line. We currently don't
	 * make the distinction of a system that runs only on AC and one
	 * that runs on battery but has no kstat battery info.
	 *
	 * We still display the estimate power usage for systems
	 * running on AC with a fully charged battery because some
	 * batteries may still consume power.
	 *
	 * If pt_battery_mod_lookup() didn't find a kstat battery module, don't
	 * bother trying to take the snapshot
	 */
	if ((err = pt_battery_stat_snapshot(fd)) < 0)
		pt_error("battery ioctl failed (%d)\n", err);

	pt_display_acpi_power(battery_state.exist, battery_state.present_rate,
	    battery_state.remain_cap, battery_state.last_cap,
	    battery_state.bst_state);
}

static int
pt_battery_stat_snapshot(int fd)
{
	acpi_bif_t bif;
	acpi_bst_t bst;

	(void) memset(&bif, 0, sizeof (acpi_bif_t));
	if (ioctl(fd, ACPI_DRV_IOC_INFO, &bif) < 0) {
		return (-1);
	}

	(void) memset(&bst, 0, sizeof (acpi_bst_t));
	if (ioctl(fd, ACPI_DRV_IOC_STATUS, &bst) < 0) {
		return (-1);
	}

	battery_state.power_unit = bif.bif_unit;
	battery_state.last_cap = mW2W((double)(bif.bif_last_cap));

	if (bst.bst_rate == 0xFFFFFFFF)
		battery_state.present_rate = 0;
	else {
		battery_state.exist = 1;
		battery_state.present_rate = mW2W((double)(bst.bst_rate));
	}

	battery_state.remain_cap = mW2W((double)(bst.bst_rem_cap));
	battery_state.bst_state = bst.bst_state;
	return (0);
}
