
/*
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 *	Eric C Saxe <eric.saxe@sun.com>
 *	Aubrey Li <aubrey.li@intel.com>
 */

/*
 * Copyright (c) 2009, Intel Corporation.
 * All Rights Reserved.
 */

/*
 * Copyright (c) 2009, 2016, Oracle and/or its affiliates. All rights reserved.
 */


#include <stdlib.h>
#include <string.h>
#include <dtrace.h>
#include <kstat.h>
#include <errno.h>
#include "powertop.h"

#include <stdio.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/pm.h>

static boolean_t pt_check_cpupm_enabled(void);
static void pt_cpufreq_null_enable(void);

#define	DTP_ARG_LENGTH	5

static int	pt_cpufreq_check_pm(void);

/*
 * Perform setup necessary to enumerate and track CPU speed changes
 */

int
pt_cpufreq_stat_prepare(cpu_table_t *cpu_table)
{
	uint64_t	supported_max_pstates;
	int	err;

	err = pt_get_kstat_integer(cpu_table->pstate_map,
	    "supported_max_pstates", &supported_max_pstates);

	if (err != 0) {
		g_max_pstates = 0;
		return (err);
	}

	g_max_pstates = supported_max_pstates;
	return (0);
}

/*
 * The administrative interface offered by libpower is not especially
 * suited to determine with certainty that any given specific PM
 * feature is enabled, nor for being sure that
 * it *has* been enabled if we try to set it using those
 * responsiveness abstractions.
 *
 *  For the time-being, therefore, we do not do this using the
 * user-level administrative abstractions offered by libpower,
 * instead, we will use a private interface offered directly by the pm
 * device driver (see corresponding functions below)
 */


/*
 * pt_cpufreq_check_pm() -
 *
 *	This routine is called repeatedly (via pt_cpufreq_suggest())
 * from the main loop of pmconfig.c to see whether PM is active or not.
 *
 * This implementation now checks directly (using /dev/pm), if CPUPM
 * is enabled (thus allowing the use of P-states)
 *
 * returns:
 *
 *     0 if PM is disabled or if CPUPM is not enabled
 *
 *     1 if there's nothing for us to do because:
 *         (a) the system does not support frequency scaling
 *         (b) we cannot open /dev/pm (need check for proper
 *		 authorization above)
 *
 *     2 if CPUPM is enabled
 *
 * Notice the ordering of the return values, they will be picked up and
 * switched upon ascendingly.
 */
static int
pt_cpufreq_check_pm(void)
{
	if (g_max_pstates < 2) {
		return (1);
	}

	/*
	 * Now, let's see if CPUPM is enabled
	 */
	if (pt_check_cpupm_enabled() == B_FALSE) {
		return (0);
	}
	return (2);
}

/*
 * We define a g_cpufreq_msg here, to say that CPUPM is disabled
 * and a null function, which we need to associate with the
 * suggestion that we will post when CPUPM is disabled.
 */
static char *g_cpufreq_msg = "(CPU PM is disabled)";

static void
pt_cpufreq_null_enable(void)
{
}

/*
 * pt_check_cpupm_enabled() -
 *	access the kernel directly (using /dev/pm) to determine
 * whether PM is active or not.
 */
static boolean_t pt_check_cpupm_enabled(void)
{
	int		ioc_ret = 0;

	/*
	 *  Check that PM is active
	 */

	ioc_ret = ioctl(g_pm_fd, PM_IS_ACTIVE, NULL);
	if (ioc_ret < 0) {
		return (B_FALSE); /* Should be a sep. error condition really */
	}
	return (ioc_ret);
}

/*
 * pt_cpufreq_suggest() -
 *
 * 	This is an extern that is called by powertop.c
 * It installs a suggestion if CPUPM is disabled.
 *
 * If there is a null function, then nothing gets called when the
 * associated key ('X' here) is struck.
 */

void
pt_cpufreq_suggest(void)
{
	int result = pt_cpufreq_check_pm();

	switch (result) {
	/*
	 * 0 -  means that CPUPM (P-states) is not enabled.
	 * Put up a suggestion to let the user know that they
	 * can enable it using poweradm(8).
	 */
	case 0:
		pt_sugg_add("Suggestion: enable CPU power management "
		    "using poweradm(8)", 40, 'X', (char *)g_cpufreq_msg,
		    pt_cpufreq_null_enable);
		break;
	/*
	 * 2 -  means that CPUPM (P-states) is enabled.
	 * Take down the suggestion.
	 */
	case 2:
		(void) pt_sugg_remove(pt_cpufreq_null_enable);
		break;
	}
}
