/*
 * Copyright 2009, Intel Corporation
 * Copyright (c) 2012, 2019, Oracle and/or its affiliates. All rights reserved.
 *
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 *	Eric C Saxe <eric.saxe@sun.com>
 *	Aubrey Li <aubrey.li@intel.com>
 */


#include <getopt.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <poll.h>
#include <kstat.h>
#include <fcntl.h>

#include "powertop.h"

/*
 * Global variables, see powertop.h for comments and extern declarations.
 * These are ordered by type, grouped by usage.
 */
int			g_bit_depth;
int 			g_total_events, g_top_events;
uint64_t		g_max_pstates, g_max_cstates;
uint_t			g_features;
int			g_ncpus;
uint_t			g_ncpus_observed = 0;
kstat2_handle_t		g_kh;
char			cstates_name[CPU_MAX_CSTATES][CSTATES_MAX_LEN];
int			g_pm_fd;
int			g_acpi_drv_fd;
cpu_table_t		*g_cpu_table;
int			*g_cpuid_map;

double			g_interval_length;
hrtime_t		g_interval_start;
hrtime_t		g_total_c_time;
hrtime_t		g_total_p_time;

struct timespec		g_rqtp;
uchar_t			g_op_mode = PT_MODE_DEFAULT;
boolean_t		g_gui;
int			g_observed_cpu = -1;

event_info_t		g_event_info[EVENT_NUM_MAX];
cstate_info_t		g_cstate_info[CPU_MAX_CSTATES];
pstate_info_t		g_pstate_info[CPU_MAX_PSTATES];

boolean_t		g_sig_resize;

uint_t			g_argc;
char			**g_argv;

/* kstat globals */


cpu_snapshot_t *g_new_ssp = NULL;
cpu_snapshot_t *g_old_ssp = NULL;

static const int	true = 1;

void
pt_sig_handler(int sig)
{
	switch (sig) {
	case SIGWINCH:
		g_sig_resize = B_TRUE;
		break;
	}
}

int
main(int argc, char **argv)
{
	double		interval;
	hrtime_t 	interval_start;
	int		index2 = 0, c, dump_count = 0;
	char		*endptr, key;
	boolean_t	root_user = B_FALSE;
	struct pollfd	pollset;
	int		optval;
	int		ret;
	kstat2_status_t	stat;

#define	LONGOPS "d:t:hvc:"

	static struct option opts[] = {
		{ "dump", 1, NULL, 'd' },
		{ "time", 1, NULL, 't' },
		{ "help", 0, NULL, 'h' },
		{ "cpu", 1, NULL, 'c' },
		{ "verbose", 0, NULL, 'v' },
		{ 0, 0, NULL, 0 }
	};

	pt_set_progname(argv[0]);
	/*
	 * Setup our nanosleep timer info for 200 msecs
	 */
	g_rqtp.tv_sec = 0;
	g_rqtp.tv_nsec =  RETRY_DELAY;

	if ((g_bit_depth = pt_get_bit_depth()) < 0)
		exit(EXIT_FAILURE);

	g_features = 0;
	interval = INTERVAL_DEFAULT;
	g_max_cstates = 0;
	g_argv = NULL;
	g_argc = 0;
	g_observed_cpu = -1;
	g_sig_resize = B_FALSE;
	g_curr_sugg = NULL;

	while ((c = getopt_long(argc, argv, LONGOPS, opts, &index2)) != EOF) {
		if (c == -1)
			break;

		switch (c) {

		case 'd':
			if (PT_ON_DUMP) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			g_op_mode |= PT_MODE_DUMP;
			g_gui = B_FALSE;

			optval = (int)strtol(optarg, &endptr, 10);
			if (optval <= 0) {
				pt_usage();
				exit(EXIT_USAGE);
			}
			dump_count = (uint_t)optval;

			if (*endptr != NULL) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			break;
		case 't':
			if (PT_ON_TIME) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			g_op_mode |= PT_MODE_TIME;
			interval = (double)strtod(optarg,
			    &endptr);

			if (*endptr != NULL || interval < 1 ||
			    interval > INTERVAL_MAX) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			break;
		case 'v':
			if (PT_ON_VERBOSE) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			g_op_mode |= PT_MODE_VERBOSE;
			break;

		case 'c':
			if (PT_ON_CPU) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			g_op_mode |= PT_MODE_CPU;
			optval = (int)strtol(optarg, &endptr, 10);
			if (optval < 0) {
				pt_usage();
				exit(EXIT_USAGE);
			}
			g_observed_cpu = (uint_t)optval;

			if (*endptr != NULL) {
				pt_usage();
				exit(EXIT_USAGE);
			}

			g_argc = 1;
			g_ncpus_observed = 1;

			if ((g_argv = malloc(sizeof (char *))) == NULL)
				return (EXIT_FAILURE);

			if ((*g_argv = malloc(sizeof (char) * 5)) == NULL)
				return (EXIT_FAILURE);

			(void) snprintf(*g_argv, 5, "%d\0", g_observed_cpu);
			break;
		case 'h':
			pt_usage();
			exit(EXIT_SUCCESS);
		default:
			pt_usage();
			exit(EXIT_USAGE);
		}
	}

	if (optind < argc) {
		pt_usage();
		exit(EXIT_USAGE);
	}

	/*
	 * Obtain a kstat2 handle and exit on failure.
	 */
	if (pt_open_kstat(&g_kh) != KSTAT2_S_OK) {
		pt_error("open_kstat  failed");
		exit(EXIT_FAILURE);
	}

	if ((g_pm_fd = open("/dev/pm", O_RDWR)) < 0) {
		pt_error("open(/dev/pm)  failed");
		exit(EXIT_FAILURE);
	}

	/*
	 * if the open fails, then battery information is not available
	 * on this system, but everything else can be displayed, so do not exit
	 */
	g_acpi_drv_fd = open("/dev/acpi_drv", O_RDWR);

	/*
	 * Enumerate the system's CPUs, populate cpu_table, g_ncpus
	 */
	if ((g_ncpus = pt_table_init(g_observed_cpu, g_kh, &g_cpu_table,
	    &g_cpuid_map)) == -1)
		exit(EXIT_FAILURE);

	if (g_ncpus_observed == 1) {
		if (!pt_verify_cpuid((int)g_observed_cpu, NULL)) {
			pt_notify_offline((int)g_observed_cpu);
			pt_usage();
			exit(EXIT_USAGE);
		}
	}

	(void) printf("%s   %s\n\n", TITLE, COPYRIGHT_INTEL);

	(void) printf("Collecting data for %.2f second(s) \n",
	    (float)interval);


	/* Prepare P-state statistics */

	if (pt_cpufreq_stat_prepare(g_cpu_table) == 0)
		g_features |= FEATURE_PSTATE;

	/* Prepare C-state statistics */
	if (pt_cpuidle_stat_prepare() == 0)
		g_features |= FEATURE_CSTATE;

	if (!PT_ON_NOEVENTS) {
		/* Prepare event statistics */
		if (pt_events_stat_prepare() == 0)
			g_features |= FEATURE_EVENTS;
	}

	/* Get the initial snapshot */
	g_new_ssp = pt_acquire_kstat_snapshot(g_ncpus, g_cpu_table);

	/* Get the c-state name info */
	get_cstate_name_info(g_cpu_table);

	/*
	 * Initialize the display.
	 */

	if (!PT_ON_DUMP) {
		pt_display_init_curses();
		pt_display_setup(B_FALSE);
		(void) signal(SIGWINCH, pt_sig_handler);
		pt_display_title_bar();
		pt_display_status_bar();

		g_gui = B_TRUE;
		pollset.fd = STDIN_FILENO;
		pollset.events = POLLIN;
	}
	/*
	 * Installs the initial suggestions, running as root and turning CPU
	 * power management ON.
	 */

	if (geteuid() != 0) {
		pt_sugg_as_root();
	} else {
		root_user = B_TRUE;
		pt_cpufreq_suggest();
	}

	while (true) {
		interval_start = gethrtime();
		g_interval_start = interval_start;

		key = 0;

		if (g_sig_resize)
			pt_display_resize(g_acpi_drv_fd);

		if (!PT_ON_DUMP) {
			if (poll(&pollset, (nfds_t)1,
			    (int)(interval * MILLISEC)) > 0)
				(void) read(STDIN_FILENO, &key, 1);
		} else {
			(void) sleep((int)interval);
		}

		g_interval_length = (double)(gethrtime() - interval_start) /
		    NANOSEC;

		g_top_events = 0;
		g_total_events = 0;

		(void) memset(g_event_info, 0, EVENT_NUM_MAX *
		    sizeof (event_info_t));
		(void) memset(g_cstate_info, 0, CPU_MAX_CSTATES *
		    sizeof (cstate_info_t));

		stat = kstat2_update(g_kh);
		if (stat != KSTAT2_S_OK) {
			pt_error("kstat_chain_update failed");
			exit(EXIT_FAILURE);
		}

		ret = pt_cpu_table_changed(g_observed_cpu, g_ncpus,
		    g_cpuid_map);

		if (ret == 1) {
			(void) printf("List of available CPUs changed, "
			    "updating list\n");

			if ((g_ncpus = pt_table_init(g_observed_cpu, g_kh,
			    &g_cpu_table, &g_cpuid_map)) == 0) {
				exit(EXIT_FAILURE);
			}
			continue;
		} else if (ret == -1) {
			(void) printf("Observered CPU %d, has gone offline\n",
			    g_observed_cpu);
				exit(EXIT_FAILURE);
		}

		g_old_ssp = g_new_ssp;
		g_new_ssp = pt_acquire_kstat_snapshot(g_ncpus, g_cpu_table);

		if (g_new_ssp != NULL) {
			pt_kstat_cpu_walk(g_ncpus, g_old_ssp, g_new_ssp);
		}
		pt_free_snapshot(g_ncpus, g_old_ssp);
		g_old_ssp = NULL;

		if (!PT_ON_NOEVENTS) {
			/* Collect event statistics */
			if (g_features & FEATURE_EVENTS &&
			    pt_events_stat_collect() < 0) {
				/* Reinitialize event statistics */
				(void) pt_events_stat_prepare();
				continue;
			}
		}

		/* Show CPU power states */

		pt_display_states();

		/* Show wakeups events affecting PM */
		if (g_features & FEATURE_EVENTS) {
			pt_display_wakeups(g_interval_length);
			pt_display_events(g_interval_length);
		}

		pt_battery_print(g_acpi_drv_fd);

		if (key && !PT_ON_DUMP) {
			switch (toupper(key)) {
			case 'Q':
				(void) kstat2_close(&g_kh);
				exit(EXIT_SUCCESS);

			case 'R':
				interval = 3;
				break;
			}

			/*
			 * Check if the user has activated the current
			 * suggestion.
			 */
			if (g_curr_sugg != NULL &&
			    toupper(key) == g_curr_sugg->key &&
			    g_curr_sugg->func)
				g_curr_sugg->func();
		}

		if (dump_count)
			dump_count--;

		/* Exits if user requested a dump */
		if (PT_ON_DUMP && !dump_count) {
			(void) kstat2_close(&g_kh);
			exit(EXIT_SUCCESS);
		}

		/* No key pressed, will suggest something */
		if (!key && !dump_count)
			pt_sugg_pick();

		/* Refresh display */
		if (!PT_ON_DUMP)
			pt_display_update();

		if (root_user)
			pt_cpufreq_suggest();
	}
	(void) kstat2_close(&g_kh);
	return (EXIT_SUCCESS);
}
