/*
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 *	Eric C Saxe <eric.saxe@sun.com>
 *	Aubrey Li <aubrey.li@intel.com>
 */

/*
 * Copyright (c) 2009, Intel Corporation.
 * All Rights Reserved.
 */

/*
 * Copyright (c) 2009, 2016, Oracle and/or its affiliates. All rights reserved.
 */

#ifndef __INCLUDE_GUARD_POWERTOP_H_
#define	__INCLUDE_GUARD_POWERTOP_H_

#include <sys/types.h>
#include <sys/processor.h>
#include <sys/pset.h>
#include <sys/time.h>
#include <kstat2.h>
#define	DBGPRINT
#define	DBGPRINT1
#define	DBGPRINT2
#define	DBGPRINT3

#define	HZ2MHZ(speed) (((speed) + ((MICROSEC)/2)) / (MICROSEC))
#define	MIN(A, B) (((A) < (B)) ? (A) : (B))
#define	max(A, B) (((A) < (B)) ? (B) : (A))


#define	TITLE			"Solaris PowerTOP version 1.3"
#define	COPYRIGHT_INTEL		"(C) 2009 Intel Corporation"

/*
 * Exit values. stdlib.h defines EXIT_SUCCESS as 0 and
 * EXIT_FAILURE as 1
 */
#define	EXIT_USAGE		2

/*
 * PowerTOP Features
 * These may not be available everywhere
 */
#define	FEATURE_CSTATE		0x01
#define	FEATURE_PSTATE		0x02
#define	FEATURE_EVENTS		0x04
#define	FEATURE_TURBO		0x08

#define	BIT_DEPTH_BUF		10

#define	INTERVAL_DEFAULT	5.0
#define	INTERVAL_MAX		30.0
#define	INTERVAL_UPDATE(l)						\
	((l/INTERVAL_DEFAULT) * INTERVAL_DEFAULT + INTERVAL_DEFAULT)
#define	RETRY_DELAY		200*MICROSEC
#define	PT_KSTAT_CHAIN_RETRYS	3

#define	STATE_NAME_MAX		16
#define	EVENT_NAME_MAX 		64
#define	EVENT_NUM_MAX 		100
#define	CPU_MAX_PSTATES		64
#define	CPU_MAX_CSTATES		16
#define	CSTATES_MAX_LEN		4

/*
 * Available op modes. The PT_ON_* macros allow for a simple way of checking
 * under which mode PowerTOP is operating.
 */
#define	PT_MODE_DEFAULT		0x01
#define	PT_MODE_DUMP		0x02
#define	PT_MODE_VERBOSE		0x04
#define	PT_MODE_CPU		0x08
#define	PT_MODE_TIME		0x10
#define	PT_MODE_NOEVENTS	0x20	/* can be patched in by debugger */
#define	PT_MODE_CPU_VERBOSE	0x0c

#define	PT_ON_DEFAULT	(g_op_mode & PT_MODE_DEFAULT)
#define	PT_ON_DUMP	(g_op_mode & PT_MODE_DUMP)
#define	PT_ON_VERBOSE	(g_op_mode & PT_MODE_VERBOSE)
#define	PT_ON_CPU	(g_op_mode & PT_MODE_CPU)
#define	PT_ON_TIME	(g_op_mode & PT_MODE_TIME)
#define	PT_ON_NOEVENTS	(g_op_mode & PT_MODE_NOEVENTS)
#define	PT_ON_CPU_VERBOSE	(g_op_mode & PT_MODE_CPU_VERBOSE)

#define	ID_NO_CPU	-1
#define	CPU_ONLINE(s)	((s) == P_ONLINE || (s) == P_NOINTR)
#define	CPU_ACTIVE(c)	(CPU_ONLINE((c)->cs_state) && (c)->cs_id != ID_NO_CPU)

#define	CPUID(idx)	g_cpu_table[idx].cpuid
#define	PT_SUCCESS	0
#define	PT_FAIL		-1


#define	CSTATE_PATH_FMT	"kstat:/pm/cpu/%d/cstate"
#define	PSTATE_PATH_FMT	"kstat:/pm/cpu/%d/pstate"

/*
 * Structures and typedefs
 */

typedef struct int_array {
	int		array_len;
	uint64_t	*array;
} int_array_t;

typedef struct cpu_snapshot {
	processorid_t		cs_id;		/* cpu_id */
	int			cs_state;	/* as in p_online(2) */
	uint64_t		actual_cycles;
	uint64_t		base_cycles;
	int_array_t		cstates_nsec;
	int_array_t		cstates_count;
	int_array_t		pstates_nsec;
	int_array_t		supp_freq;
} cpu_snapshot_t;



typedef struct event_info {
	char		offender_name[EVENT_NAME_MAX];
	char		offense_name[EVENT_NAME_MAX];
	uint64_t	total_count;
} event_info_t;


/*
 * P/C state information
 */
typedef struct cstate_info {
	char		name[STATE_NAME_MAX];
	hrtime_t	total_time;
	hrtime_t	last_time;
	uint64_t	events;
} cstate_info_t;

typedef struct pstate_info {
	uint64_t	speed;
	hrtime_t	total_time;
} pstate_info_t;


/*
 * Suggestions
 */
typedef	void		(sugg_func_t)(void);

typedef struct suggestion {
	char *text;
	char key;
	char *sb_msg;
	int weight;
	int slice;
	sugg_func_t *func;
	struct suggestion *prev;
	struct suggestion *next;
} sugg_t;

typedef struct _cpu_table {
	int		status;
	int		cpuid;
	kstat2_map_t	cstate_map;	/* kstat map for cstates */
	kstat2_map_t	pstate_map;	/* kstat map for pstates */
	int_array_t	cstates_nsec;
	int_array_t	cstates_count;
	int_array_t	pstates_nsec;
	int_array_t	supp_freq;
	int_array_t	eff_freq;
} cpu_table_t;

extern int			g_bit_depth;
extern int			g_pm_fd;

extern kstat2_handle_t		g_kh;

extern char			cstates_name[CPU_MAX_CSTATES][CSTATES_MAX_LEN];
/*
 * Event accounting
 */
extern int 			g_total_events;
extern int 			g_top_events;

/*
 * Command line arguments
 */
extern uchar_t			g_op_mode;
extern int			g_observed_cpu;
extern boolean_t		g_gui;
/*
 * Event info array
 */
extern event_info_t		g_event_info[EVENT_NUM_MAX];

/*
 * Lookup table, sequential CPU id to Solaris CPU id
 */
extern cpu_table_t 		*g_cpu_table;
extern int			*g_cpuid_map;

/*
 * Number of idle/frequency states
 */
extern uint64_t			g_max_pstates;
extern uint64_t			g_max_cstates;

/*
 * Total time, used to display different idle states
 */
extern hrtime_t			g_total_c_time;
/*
 * Total time, used to display different active states
 */
extern hrtime_t			g_total_p_time;

/*
 * Current interval length
 */
extern double			g_interval_length;
/*
 * nanosleep timespec struct for 200 msec delay
 */
extern struct timespec		g_rqtp;

/*
 * P/C state info arrays
 */
extern cstate_info_t		g_cstate_info[CPU_MAX_CSTATES];
extern pstate_info_t		g_pstate_info[CPU_MAX_PSTATES];

extern uint_t			g_features;
extern int			g_ncpus;
extern uint_t			g_ncpus_observed;

extern sugg_t			*g_curr_sugg;

/*
 * DTrace scripts for the events report
 */
extern const char		*g_dtp_events;
extern const char		*g_dtp_events_v;
extern const char		*g_dtp_events_c;
extern const char		*g_dtp_events_c_v;
/*
 * Arguments for dtrace_program_strcompile(). Contents vary according to
 * the specified operation mode.
 */
extern uint_t			g_argc;
extern char			**g_argv;

/*
 * Platform specific messages
 */
extern const char		*g_msg_cstate;
extern const char		*g_msg_pstate;

/*
 * Flags for signal handling
 */
extern boolean_t		g_sig_resize;

/*
 * Exit message buffer
 */
#define	EXIT_MSG_BUFSIZE	1024
extern char		*pt_exit_message;
extern char		pt_exit_message_buf[];


extern void		pt_sig_handler(int);

/*
 * Suggestions related
 */
extern void 		pt_cpufreq_suggest(void);
extern void		pt_sugg_as_root(void);

/*
 * See util.c
 */
extern void 		pt_error(char *, ...);
extern void 		pt_set_progname(char *);
extern void		pt_usage(void);
extern int		pt_get_bit_depth(void);
extern int		pt_event_compare(const void *, const void *);
extern int		pt_verify_cpuid(processorid_t, int *);
extern void		pt_notify_offline(processorid_t);

/*
 * Display/curses related
 */
extern void		pt_display_setup(boolean_t);
extern void 		pt_display_init_curses(void);
extern void		pt_display_update(void);
extern void 		pt_display_title_bar(void);
extern void		pt_display_status_bar(void);
extern void		pt_display_mod_status_bar(char *);
extern void 		pt_display_states(void);
extern void		pt_display_acpi_power(uint32_t, double, double, double,
	uint32_t);
extern void 		pt_display_wakeups(double);
extern void 		pt_display_events(double);
extern void 		pt_display_suggestions(char *);
extern void		pt_display_resize(int fd);

/*
 * Suggestions
 */
extern void 		pt_sugg_add(char *, int, char, char *, sugg_func_t *);
extern int		pt_sugg_remove(sugg_func_t *);
extern void 		pt_sugg_pick(void);
/*
 * Battery
 */
extern void 		pt_battery_print(int fd);

/*
 * DTrace stats
 */

extern int 		pt_cpufreq_stat_prepare(cpu_table_t *cpu_table);
extern int 		pt_cpuidle_stat_prepare(void);
extern int 		pt_events_stat_prepare(void);
/*
 * Kstat
 */
int 			pt_events_stat_collect(void);

extern cpu_snapshot_t *g_new_ssp;
extern cpu_snapshot_t *g_old_ssp;


/*
 * Kstat routines
 */

extern void pt_kstat_cpu_walk(int ncpus, cpu_snapshot_t *old,
    cpu_snapshot_t *new);
extern void pt_update_cpu_kstats(int count, struct cpu_snapshot *c1,
    struct cpu_snapshot *c2);
extern cpu_snapshot_t *pt_acquire_kstat_snapshot(int count,
    cpu_table_t *cpu_table);
extern void pt_free_snapshot(int count, cpu_snapshot_t *ss);
extern kstat2_status_t	pt_open_kstat(kstat2_handle_t *handle);

extern int pt_get_kstat_integer(kstat2_map_t map, char *name,
    uint64_t *interger);
extern int pt_get_kstat_integer_array(kstat2_map_t map, char *name,
    int_array_t *array);
extern void pt_free_int_array(int_array_t *array);
extern uint_t pt_table_init(int cpu, kstat2_handle_t handle,
    cpu_table_t **cpu_table, int **cpuid_map);
extern int pt_cpu_table_changed(int cpu, int map_count, int *cpuid_map);

extern void get_cstate_name_info(cpu_table_t *cpu_table);

#endif /* __INCLUDE_GUARD_POWERTOP_H_ */
