/*
 * Copyright 2009, Intel Corporation
 * Copyright (c) 2009, 2020, Oracle and/or its affiliates. All rights reserved.
 *
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 *	Eric C Saxe <eric.saxe@sun.com>
 *	Aubrey Li <aubrey.li@intel.com>
 */


#include <stdarg.h>
#include <stdlib.h>
#include <libgen.h>
#include <unistd.h>
#include <strings.h>
#include <sys/systeminfo.h>
#include <kstat.h>
#include <time.h>
#include <errno.h>
#include "powertop.h"

static char 	PROG_FMT[] = "%s: ";
static char 	ERR_FMT[] = ": %s";
static char 	*progname;
static int pt_obtain_cpu_kstats(int i, cpu_snapshot_t *cpu_ss,
    cpu_table_t *cpu_table);
static	void pt_set_pstate_speed(int i, int len, uint64_t eff_freq,
    struct cpu_snapshot *c2);


void
pt_set_progname(char *name)
{
	progname = basename(name);
}

/*PRINTFLIKE1*/
void
pt_error(char *format, ...)
{
	int 	err = errno;
	va_list alist;

	if (g_gui)
		return;

	if (progname != NULL)
		(void) fprintf(stderr, PROG_FMT, progname);

	va_start(alist, format);
	(void) vfprintf(stderr, format, alist);
	va_end(alist);

	if (strchr(format, '\n') == NULL)
		(void) fprintf(stderr, ERR_FMT, strerror(err));
}

static int
pt_get_index_from_cpuid(processorid_t cpuid)
{
	int	index;
	int	max_cpuid	= sysconf(_SC_CPUID_SEQ_MAX);

	if (cpuid <= max_cpuid) {
		index = g_cpuid_map[cpuid];
		if (index >= 0 && index < g_ncpus)
			return (index);
	}
	return (-1);
}

int
pt_verify_cpuid(processorid_t cpuid, int *idxp)
{
	int	index;
	int	*p;

	p = (idxp != NULL) ? idxp : &index;
	if ((*p = pt_get_index_from_cpuid(cpuid)) != -1)
		return (1);
	else
		return (0);
}

void
pt_notify_offline(processorid_t cpuid)
{
	(void) snprintf(pt_exit_message_buf, EXIT_MSG_BUFSIZE,
	    "Error: The CPU#%d may be off-line.\n", cpuid);
	pt_exit_message = pt_exit_message_buf;

	(void) fprintf(stderr, "%s", pt_exit_message);
}

/*
 * Returns the number of online CPUs.
 */
uint_t
pt_table_init(int cpu, kstat2_handle_t handle, cpu_table_t **cpu_table,
    int **cpuid_map)
{
	int	i;
	uint_t	ncpus = 0;
	kstat2_status_t	stat;
	int	cpuid;
	int	max_cpuid	= sysconf(_SC_CPUID_SEQ_MAX);
	int	cpus_conf	= sysconf(_SC_NPROCESSORS_CONF);
	char	path[80];
	int	status;
	cpu_table_t	*ptable;
	int		*pmap;

	/* Fall back to one CPU if any of the sysconf calls above failed */
	if (max_cpuid == -1 || cpus_conf == -1) {
		max_cpuid = cpus_conf = 1;
	}

	if (cpu != -1)
		cpus_conf = 1;

	/* Free any previous cpu table */
	if (*cpu_table != NULL) {
		free(*cpu_table);
		*cpu_table = NULL;
	}
	if (*cpuid_map != NULL) {
		free(*cpuid_map);
		*cpuid_map = NULL;
	}

	if ((ptable = calloc(cpus_conf, sizeof (cpu_table_t))) == NULL)
		return (0);
	if ((pmap = malloc((max_cpuid + 1) * sizeof (int))) == NULL) {
		free(ptable);
		return (0);
	}
	for (i = 0; i <= max_cpuid; i++) {
		pmap[i] = -1;
	}

	for (i = 0, ncpus = 0; i <= max_cpuid && ncpus < cpus_conf; i++) {

		if (cpu == -1)
			cpuid = i;
		else
			cpuid = cpu;

		status = p_online(cpuid, P_STATUS);

		if (status == P_ONLINE) {
			(void) snprintf(path, sizeof (path), CSTATE_PATH_FMT,
			    cpuid);
			stat = kstat2_lookup_map(handle, path,
			    &ptable[ncpus].cstate_map);
			if (stat != KSTAT2_S_OK) {
				(void) printf("can't lookup kstat cstate: %s",
				    kstat2_status_string(stat));
				free(ptable);
				free(pmap);
				return (-1);
			}

			(void) snprintf(path, sizeof (path), PSTATE_PATH_FMT,
			    cpuid);
			stat = kstat2_lookup_map(handle, path,
			    &ptable[ncpus].pstate_map);
			if (stat != KSTAT2_S_OK) {
				(void) printf("can't lookup kstat pstate: %s",
				    kstat2_status_string(stat));
				free(ptable);
				free(pmap);
				return (-1);
			}
			ptable[ncpus].status = status;
			ptable[ncpus].cpuid = cpuid;
			pmap[cpuid] = ncpus; /* reverse reference */
			ncpus++;
		}
	}

	*cpu_table = ptable;
	*cpuid_map = pmap;

	return (ncpus);
}

/*
 * Has the CPU talbe changed?
 * 1 means yes
 * 0 mean no
 * -1 means the one CPU that was being observed has gone offline
 * Genaerate a new CPU map, if the number of eneries in the map has
 * changed, then the CPU table has changed, if the contents of the
 * map has chnaged, then the table has changed.
 * If NO on line CPUs can be found then, the observed CPU must have
 * gone offline
 * NOTE: we use the CPU map and not the CPU table becasue the CPU map
 * is an array of int, while the CPU table os an array of cpu_table_t,
 * the memcmp() is faster.
 */
int
pt_cpu_table_changed(int cpu, int map_count, int *cpuid_map)
{
	int	i;
	uint_t	ncpus = 0;
	int	cpuid;
	int	max_cpuid	= sysconf(_SC_CPUID_SEQ_MAX);
	int	cpus_conf	= sysconf(_SC_NPROCESSORS_CONF);
	int	status;
	int	*cpuid_map_new;

	/* Fall back to one CPU if any of the sysconf calls above failed */
	if (max_cpuid == -1 || cpus_conf == -1) {
		max_cpuid = cpus_conf = 1;
	}

	if (cpu != -1) {
		cpus_conf = 1;
	}


	if ((cpuid_map_new = malloc((max_cpuid + 1) * sizeof (int))) == NULL) {
		return (0);
	}
	for (i = 0; i <= max_cpuid; i++) {
		cpuid_map_new[i] = -1;
	}

	for (i = 0; i <= max_cpuid && ncpus < cpus_conf; i++) {
		if (cpu != -1)
			cpuid = cpu;
		else
			cpuid = i;

		status = p_online(cpuid, P_STATUS);

		if (status == P_ONLINE) {
			cpuid_map_new[cpuid] = ncpus; /* reverse reference */
			ncpus++;
		}
	}

	if (ncpus == 0) {
		free(cpuid_map_new);
		return (-1);
	}

	if (ncpus != map_count) {
		free(cpuid_map_new);
		return (1);
	}

	if (memcmp(cpuid_map, cpuid_map_new, max_cpuid + 1) != 0) {
		free(cpuid_map_new);
		return (1);
	}

	free(cpuid_map_new);
	return (0);
}

/*
 * Simple integer comparison routine for the event report qsort(3C).
 */
int
pt_event_compare(const void *p1, const void *p2)
{
	event_info_t *ip = ((event_info_t *)p1);
	event_info_t *jp = ((event_info_t *)p2);

	if (ip->total_count > jp->total_count)
		return (-1);

	if (ip->total_count < jp->total_count)
		return (1);

	return (0);
}

void
pt_usage(void)
{
	(void) fprintf(stderr, "%s   %s\n\n", TITLE, COPYRIGHT_INTEL);
	(void) fprintf(stderr, "usage: powertop [option]\n");
	(void) fprintf(stderr, "  -d, --dump [count]	Read wakeups count "
	    "times and print list of top offenders\n");
	(void) fprintf(stderr, "  -t, --time [interval]	Default time to gather "
	    "data in seconds [1-30s]\n");
	(void) fprintf(stderr, "  -v, --verbose		Verbose mode, reports "
	    "kernel cyclic activity\n");
	(void) fprintf(stderr, "  -c, --cpu [CPU]	Only observe a specific"
	    " CPU\n");
	(void) fprintf(stderr, "  -h, --help		Show this help "
	    "message\n");
}

int
pt_get_bit_depth(void)
{
	/*
	 * This little routine was derived from isainfo.c to look up
	 * the system's bit depth. It feeds a 10 byte long buffer to
	 * sysinfo (we only need the first word, sysinfo truncates and
	 * \0 terminates the rest) from which we figure out which isa
	 * we're running on.
	 */
	char	buf[BIT_DEPTH_BUF];

	if (sysinfo(SI_ARCHITECTURE_64, buf, BIT_DEPTH_BUF) == -1)
		if (sysinfo(SI_ARCHITECTURE_32, buf, BIT_DEPTH_BUF) == -1)
			return (-2);

	if (strcmp(buf, "sparc") == 0 || strcmp(buf, "i386") == 0)
		return (32);

	if (strcmp(buf, "sparcv9") == 0 || strcmp(buf, "amd64") == 0)
		return (64);

	return (-3);
}

void
pt_kstat_cpu_walk(int ncpus, cpu_snapshot_t *old, cpu_snapshot_t *new)
{
	int i;

	if (new == NULL || old == NULL)
		return;

	for (i = 0; i < CPU_MAX_PSTATES; i++) {
		g_pstate_info[i].total_time = 0;
	}
	for (i = 0; i < CPU_MAX_CSTATES; i++) {
		g_cstate_info[i].total_time = 0;
		g_cstate_info[i].last_time = 0;
		g_cstate_info[i].events = 0;
	}
	g_total_c_time = 0;
	g_total_p_time = 0;
	g_total_events = 0;

	pt_update_cpu_kstats(ncpus, old, new);
}


void
pt_update_cpu_kstats(int ncpus, struct cpu_snapshot *c1,
    struct cpu_snapshot *c2)
{
	int i, j, k, len;
	uint64_t	count;
	uint64_t	eff_freq;
	uint64_t	supp_freq;

	g_total_c_time = 0;

	for (k = 0; k < ncpus; k++, c1++, c2++) {

		if (g_features & FEATURE_CSTATE) {
			if (c1->cstates_nsec.array_len !=
			    c2->cstates_nsec.array_len)
				return;
			if (c1->cstates_count.array_len !=
			    c2->cstates_count.array_len)
				return;

			/*
			 * C-State Kstats are in C(n) ... C(0) Order
			 * Powertop displays them in C(0) to C(n) order.
			 */
			len = c1->cstates_count.array_len;

			for (i = 0; i < len; i++) {
				j = len - i;
				count = c2->cstates_count.array[j] -
				    c1->cstates_count.array[j];
				g_cstate_info[i].events += count;
			}

			len = c1->cstates_nsec.array_len;
			for (i = 0; i < len; i++) {
				j = (len -1) - i;
				count = c2->cstates_nsec.array[j] -
				    c1->cstates_nsec.array[j];
				g_cstate_info[i].last_time = count;
				g_cstate_info[i].total_time += count;
				g_total_c_time += count;
			}
		}

		/*
		 * If only P0 is present on the machine, use the first
		 * (and only) supported frequency, which corresponds to P0
		 */
		if (g_max_pstates == 0) {
			g_pstate_info[0].speed =
			    HZ2MHZ(c2->supp_freq.array[0]);
		} else {
			if (g_features & FEATURE_PSTATE) {
				if (c1->pstates_nsec.array_len !=
				    c2->pstates_nsec.array_len)
					return;

				len = c1->pstates_nsec.array_len;
				/*
				 * compute effective frequency for this cpu
				 */
				eff_freq = 0;
				supp_freq = c2->supp_freq.array[len-1];
				if (c2->actual_cycles != 0 &&
				    c2->base_cycles != 0 &&
				    c1->actual_cycles != 0 &&
				    c1->base_cycles != 0) {
					eff_freq = supp_freq *
					    (c2->actual_cycles -
					    c1->actual_cycles) /
					    (c2->base_cycles - c1->base_cycles);

				}

				/*
				 * P-State Kstats are in P(n) .. P(0) order
				 * powertop displays them is this order.
				 */
				for (i = 0; i < len; i++) {
					count = c2->pstates_nsec.array[i] -
					    c1->pstates_nsec.array[i];
					g_pstate_info[i].total_time += count;
					g_total_p_time += count;
					pt_set_pstate_speed(i, len, eff_freq,
					    c2);
				}
			}
		}
	}
}

static	void
pt_set_pstate_speed(int i, int len, uint64_t eff_freq, struct cpu_snapshot *c2)
{
	uint64_t supp_freq = HZ2MHZ(c2->supp_freq.array[i]);

	/*
	 * if this is P0 (i == len -1)
	 * and effective frequency is supported (c2->eff_freq.array != NULL)
	 * and the effective frequency is > the supported frequency
	 * then display the effective frequency instead of the
	 * supported frequency
	 */
	if (i == len -1 && eff_freq > c2->supp_freq.array[i]) {
		eff_freq = HZ2MHZ(eff_freq);
		if (eff_freq > g_pstate_info[i].speed && eff_freq > supp_freq) {
			g_pstate_info[i].speed = eff_freq;
		} else {
			g_pstate_info[i].speed = supp_freq;
		}
	} else {
		g_pstate_info[i].speed = supp_freq;
	}
}

/*
 * Note: the following helpers do not clean up on the failure case,
 * because pt_free_snapshot() is called in main each time through the loop
 */
int
pt_acquire_cpus(int count, cpu_snapshot_t *cpu_ss, cpu_table_t *cpu_table)
{
	int	err = 0;
	int	ret = 0;
	int	i;

	for (i = 0; i < count; i++) {
		err = pt_obtain_cpu_kstats(i, cpu_ss, cpu_table);
		if (ret == 0 && err != 0)
			ret = err;
	}
	return (ret);

}

static	int
pt_obtain_cpu_kstats(int i, cpu_snapshot_t *cpu_ss, cpu_table_t *cpu_table)
{
	int	ret = 0;
	int	err = 0;

	if (g_features & FEATURE_CSTATE) {
		err = pt_get_kstat_integer_array(cpu_table[i].cstate_map,
		    "cstates_nsec", &cpu_ss[i].cstates_nsec);

		if (ret == 0 && err != 0)
			ret = err;

		err = pt_get_kstat_integer_array(cpu_table[i].cstate_map,
		    "cstates_count", &cpu_ss[i].cstates_count);

		if (ret == 0 && err != 0)
			ret = err;
	}

	if (g_features & FEATURE_PSTATE) {
		err = pt_get_kstat_integer(cpu_table[i].pstate_map,
		    "actual_cycles", &cpu_ss[i].actual_cycles);

		if (ret == 0 && err != 0)
			ret = err;

		err = pt_get_kstat_integer(cpu_table[i].pstate_map,
		    "base_cycles", &cpu_ss[i].base_cycles);

		if (ret == 0 && err != 0)
			ret = err;

		err = pt_get_kstat_integer_array(cpu_table[i].pstate_map,
		    "pstates_nsec", &cpu_ss[i].pstates_nsec);

		if (ret == 0 && err != 0)
			ret = err;

		err = pt_get_kstat_integer_array(cpu_table[i].pstate_map,
		    "supported_frequencies", &cpu_ss[i].supp_freq);

		if (ret == 0 && err != 0)
			ret = err;

	}

	return (ret);
}

int
pt_get_kstat_integer(kstat2_map_t map, char *name, uint64_t *integer)
{
	int		err = 0;
	kstat2_nv_t	val;
	kstat2_status_t stat;
	kstat2_mapref_t	ref = NULL;

	stat = kstat2_map_get(map, name, &val);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer() kstat2_map_get() "
		    "failed with %s", kstat2_status_string(stat));
		return (err);
	}

	stat = kstat2_mapref_alloc(map, &ref);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer() kstat2_mapref_alloc() "
		    "failed with %s", kstat2_status_string(stat));
		goto done;
	}

	stat = kstat2_update(g_kh);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer() kstat2_update() failed "
		    "with %s", kstat2_status_string(stat));
		goto done;
	}

	stat = kstat2_mapref_deref(ref, &map);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer() kstat2_mapref_deref() "
		    "failed with %s", kstat2_status_string(stat));
		goto done;
	}

	*integer = val->kstat2_integer;
	err = 0;
done:
	(void) kstat2_mapref_free(&ref);

	return (err);
}

int
pt_get_kstat_integer_array(kstat2_map_t map, char *name, int_array_t *array)
{
	int		err;
	kstat2_nv_t	val;
	kstat2_status_t stat;
	kstat2_mapref_t	ref = NULL;

	stat = kstat2_map_get(map, name, &val);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer_array() kstat2_map_get() "
		    "failed with %s\n", kstat2_status_string(stat));
		pt_free_int_array(array);
		return (err);
	}

	stat = kstat2_mapref_alloc(map, &ref);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer_array() "
		    "kstat2_mapref_alloc() failed with %s\n",
		    kstat2_status_string(stat));
		pt_free_int_array(array);
		return (err);
	}

	stat = kstat2_update(g_kh);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer_array() kstat2_update() "
		    "failed with %s\n", kstat2_status_string(stat));
		pt_free_int_array(array);
		goto done;
	}

	stat = kstat2_mapref_deref(ref, &map);
	if (stat != KSTAT2_S_OK) {
		err = errno;
		(void) printf("pt_get_kstat_integer_array() "
		    "kstat2_mapref_deref() failed with %s\n",
		    kstat2_status_string(stat));
		pt_free_int_array(array);
		goto done;
	}

	err = 0;
	array->array_len = val->kstat2_integers.length;

	if (val->kstat2_integers.length == 0) {
		pt_free_int_array(array);
		goto done;
	}

	if (array->array == NULL) {
		array->array = calloc(val->kstat2_integers.length,
		    sizeof (uint64_t));
		if (array->array == NULL) {
			pt_free_int_array(array);
			goto done;
		}
	}

	(void) memcpy(array->array, val->kstat2_integers.array,
	    sizeof (uint64_t) * val->kstat2_integers.length);

done:
	(void) kstat2_mapref_free(&ref);

	return (err);
}

cpu_snapshot_t *
pt_acquire_kstat_snapshot(int count, cpu_table_t *cpu_table)
{
	int err = 0;
	cpu_snapshot_t *cpu_ss;

	if ((cpu_ss = calloc(count, sizeof (cpu_snapshot_t))) == NULL)
		return (0);

	err = pt_acquire_cpus(count, cpu_ss, cpu_table);

	if (err) {
		free(cpu_ss);
		return (NULL);
	}

	return (cpu_ss);
}

void
pt_free_snapshot(int count, cpu_snapshot_t *ss)
{
	int i;

	if (ss == NULL)
		return;
	for (i = 0; i < count; i++) {
		pt_free_int_array(&(ss[i].cstates_nsec));
		pt_free_int_array(&(ss[i].cstates_count));
		pt_free_int_array(&(ss[i].pstates_nsec));
		pt_free_int_array(&(ss[i].supp_freq));
	}
	free(ss);
}

void
pt_free_int_array(int_array_t *array)
{
	if (array->array != NULL) {
		free(array->array);
	}
	array->array = NULL;
	array->array_len = 0;
}

kstat2_status_t
pt_open_kstat(kstat2_handle_t *handle)
{
	kstat2_status_t stat;

	stat = kstat2_open(handle, NULL);
	if (stat != KSTAT2_S_OK) {
		pt_error("kstat2_open failed");
	}

	return (stat);
}

void
get_cstate_name_info(cpu_table_t *cpu_table)
{
	int	err;
	uint64_t	i;
	uint64_t	supported_max_cstates;

	err = pt_get_kstat_integer(cpu_table->cstate_map,
	    "supported_max_cstates", &supported_max_cstates);

	if (err != 0) {
		g_max_cstates = 0;
		return;
	}

	g_max_cstates = supported_max_cstates;

	/*
	 * assume c-state names to be C1,C2...Cmax_cstates.
	 */
	for (i = 1; i <= g_max_cstates; i++) {
		(void) snprintf(cstates_name[i], CSTATES_MAX_LEN, "c%ld", i);
	}


}
