/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/* LINTLIBRARY */

#include	<sys/machelf.h>
#include	<sys/elf.h>
#include	<sys/types.h>

#if	!defined(_KERNEL) && !defined(_BOOT)
#include	<stddef.h>
#include	<_elfsignrange.h>
#else
#include	<sys/sysmacros.h>
#include	<elfsignrange/_elfsignrange.h>
#endif

#if	!defined(_KERNEL) && !defined(_BOOT) && defined(ELF_SIGN_RANGE_DEBUG)
#include	<stdio.h>
#define		DBG_CALLBACK(a, o, s)	_Dbg_callback(a, o, s)
#else
#define		DBG_CALLBACK(a, o, s)
#endif


#if	defined(_ELF64)
#if	!defined(_KERNEL) && !defined(_BOOT) && defined(ELF_SIGN_RANGE_DEBUG)
void
_Dbg_callback(char *addr, size_t offset, size_t size)
{
	(void) printf("  callback: addr 0x%llx: offset 0x%llx:%llu - "
	    "0x%llx:%llu, cnt 0x%llx:%llu\n", EC_NATPTR(addr),
	    EC_XWORD(offset), EC_XWORD(offset),
	    EC_XWORD(offset + size), EC_XWORD(offset + size),
	    EC_XWORD(size), EC_XWORD(size));
}
#endif

/* Define a simple checksum capability (stolen from libcmd) */
void
_elf_checksum_init(void *cookie, char *base, uint_t flags)
{
	_Elf_Checksum_desc_t	*ecsp = (_Elf_Checksum_desc_t *)cookie;

	ecsp->ecs_base = base;
	ecsp->ecs_sum = 0;
	ecsp->ecs_flags = flags;
}

void
_elf_checksum_callback(size_t offset, size_t size, void *cookie)
{
	_Elf_Checksum_desc_t	*ecsp = (_Elf_Checksum_desc_t *)cookie;
	char			*addr = ecsp->ecs_base + offset;

	if (size == 0)
		return;

	DBG_CALLBACK(addr, offset, size);

	while (size--)
		ecsp->ecs_sum += *addr++ & 0xff;
}

void
_elf_checksum_fini(void *cookie)
{
	_Elf_Checksum_desc_t	*ecsp = (_Elf_Checksum_desc_t *)cookie;
	long			_sum, sum = ecsp->ecs_sum;

	_sum = _ELF_LSW(sum) + _ELF_MSW(sum);
	ecsp->ecs_sum = ((ushort_t)(_ELF_LSW(_sum) + _ELF_MSW(_sum)));
}
#endif

/*
 * Determine whether a section should be skipped.
 *
 * When processing ELF_SR_RELOBJ, any SHT_SUNW_SIGNATURE section data is always
 * omitted from the address range, as the data can differ even when the rest of
 * the file contents remain unchanged.  Note, objects can contain more than one
 * SHT_SUNW_SIGNATURE section.  Sections named .SUNW_signature and .SUNW_fips
 * have this same type.  Both of these sections are skipped.
 *
 * SHT_DYNAMIC and SHT_SUNW_ancillary sections are omitted when calculating
 * an ELF checksum, as these sections are updated with the checksums when the
 * file is being created by ld(1).
 */
static int
skip_section(Shdr *shdr, boolean_t skip_checksum)
{
	if (shdr->sh_type == SHT_SUNW_SIGNATURE)
		return (1);

	if (skip_checksum &&
	    ((shdr->sh_flags & SHF_SUNW_ABSENT) == 0) &&
	    ((shdr->sh_type == SHT_DYNAMIC) ||
	    (shdr->sh_type == SHT_SUNW_ancillary)))
		return (1);

	return (0);
}

/*
 * Low level, machine and class specific function to return the important
 * memory ranges of an ELF file.  This core functionality can be compiled
 * twice to create both _ELF64 and non_ELF64 variants, and thus provides the
 * interfaces:
 *
 *	_elf32_sign_range()
 *	_elf64_sign_range()
 *
 * Ehdr points to the ELF files ELF header
 * Phdr points to the ELF files array of Program headers
 * Shdr points to the ELF files array of Section headers
 *
 * There are a some assumptions made in regards ELF layout that relate to
 * existing practice, and the use of libelf.  The Program Header table (Phdr[])
 * is assumed to follow the Elf Header (Ehdr[]).  The Sections (Scn) follow the
 * Program Header table.  The Section Header Table (Shdr[]) is at the end of
 * the file, and follows the sections.  Although libelf creates this layout,
 * ELF does not dictate it.  For instance, files exist where the Section Header
 * table can be located between sections.  The GNU objcopy utility has been
 * known to create files in this manner.
 *
 * The most optimal use of elf_sign_range (where the number of memory ranges
 * conveyed to the caller is kept to a minimum) is achieved when any
 * SHT_SUNW_signature section is the last section of the file, and the Section
 * Header table follows the Sections.  The creation of this optimal layout is
 * the responsibility of the tools used to add signature information.
 */
int
_elf_sign_range(elf_sign_range_cb_func cb, Ehdr *ehdr, Phdr *phdr, Shdr *shdr,
    size_t msize, Elf_Signrange type, void *cookie)
{
	Phdr		*nphdr, *fphdr = NULL, *lphdr = NULL;
	Shdr		*nshdr;
	size_t		offset, size;
	uint_t		num, shnum;
	int		sig;
	boolean_t	process_segments, skip_checksum, skip_signature;
	boolean_t	extended_sections = _B_FALSE;

	/* If there's no Ehdr, bail */
	if (ehdr == NULL)
		return (0);

	/*
	 * Determine whether to drive the range processing using segments or
	 * sections from the ELF file type.
	 */
	switch (ehdr->e_type) {
	case ET_EXEC:
	case ET_DYN:
		process_segments = _B_TRUE;
		break;
	case ET_REL:
	case ET_SUNW_ANCILLARY:
		process_segments = _B_FALSE;
		break;
	default:
		return (0);
	}

	/* Catch a whole file request first */
	if (type == ELF_SR_FILE) {
		if (msize)
			(*cb)(0, msize, cookie);
		return (1);
	}

	/*
	 * Adjust the default processing if necessary from the sign range type.
	 */
	switch (type) {
	case ELF_SR_SIGNED_INTERPRET:
		skip_checksum = _B_FALSE;
		skip_signature = _B_FALSE;
		break;
	case ELF_SR_INTERPRET:
		skip_checksum = _B_FALSE;
		skip_signature = _B_TRUE;
		break;
	case ELF_SR_RELOBJ:
		process_segments = _B_FALSE;
		skip_checksum = _B_FALSE;
		skip_signature = _B_TRUE;
		break;
	case ELF_SR_CHECKSUM:
		skip_checksum = _B_TRUE;
		skip_signature = _B_TRUE;
		break;
	default:
		return (0);
	}

	/*
	 * Determine the information within an ELF header that can be verified.
	 *
	 * Dynamic objects only have their allocatable sections mapped into
	 * memory, and these segments are all that need to be verified.  Thus,
	 * there is some flexibility in that non-allocatable sections can be
	 * added or removed, and need never be part of a files verification.
	 * Relocatable objects have all their sections, besides any signature
	 * section verified.  However, signature sections can be added after
	 * the original file is built.
	 *
	 * Omit the e_shoff element to allow for section modification.
	 * Capture the range ehdr -> ehdr->e_shoff.
	 */
	size = offsetof(Ehdr, e_shoff);
	if (size > msize)
		return (0);
	(*cb)(0, size, cookie);

	/* Reset the offset to the next element ehdr->e_flags */
	offset = offsetof(Ehdr, e_flags);

	/*
	 * Omit the e_shum and e_shstrndx elements to allow for section addition
	 * or deletion, and the possibility that either action could modify the
	 * Section Header String Table offset.  These two elements are the last
	 * elements of the Ehdr.
	 * Capture the range ehdr->e_flags -> ehdr->e_shnum.
	 */
	size = offsetof(Ehdr, e_shnum) - offset;
	if ((offset + size) > msize)
		return (0);
	(*cb)(offset, size, cookie);

	/* Reset the offset to the end of the Ehdr */
	offset = sizeof (Ehdr);

	if (process_segments) {
		uint_t	phnum = ehdr->e_phnum;

		/*
		 * Process a dynamic executable or shared object.  If there is
		 * no Phdr[], something is odd.
		 */
		if (phdr == NULL)
			return (0);

		/*
		 * If the program header number is PN_XNUM, then the actual
		 * number must be obtained from Shdr[0].
		 */
		if (phnum == PN_XNUM) {
			if ((shdr == NULL) ||
			    ((phnum = shdr[0].sh_info) == 0))
				return (0);
		}

		/*
		 * Scan the Phdr[]'s to determine the address range of the
		 * loadable segments.
		 */
		for (num = 0, nphdr = phdr; num < phnum; num++, nphdr++) {
			/* Ignore any segment that has no backing data */
			if (nphdr->p_filesz == 0)
				continue;

			if (nphdr->p_type == PT_DYNAMIC) {
				if (skip_checksum) {
					/*
					 * Under ELF_SR_CHECKSUM omit this
					 * Phdr[] data range, and return the
					 * range collected up until the start
					 * of this Phdr[].
					 */
					size = (size_t)nphdr->p_offset - offset;
					if ((offset + size) > msize)
						return (0);
					if (size)
						(*cb)(offset, size, cookie);

					/*
					 * Reset the offset to the end of the
					 * PT_DYNAMIC data.
					 */
					offset = (size_t)(nphdr->p_offset) +
					    (size_t)(nphdr->p_filesz);
				}
				continue;
			}

			if (nphdr->p_type != PT_LOAD)
				continue;

			/* Keep track of the first and last PT_LOAD segments */
			if ((fphdr == NULL) ||
			    (nphdr->p_offset < fphdr->p_offset))
				fphdr = nphdr;
			if ((lphdr == NULL) ||
			    (nphdr->p_offset >= lphdr->p_offset))
				lphdr = nphdr;
		}

		/* If there are no loadable segments, something is odd */
		if (fphdr == NULL)
			return (0);

		/*
		 * In theory, the order of Ehdr, Phdr[]'s and the sections the
		 * Phdr[]'s capture, could provide a number of combinations.
		 * But, existing practice has Phdr[]'s following the Ehdr.
		 * For userland dynamic objects, the first PT_LOAD segment
		 * captures the Ehdr and Phdr[]'s.  For a kernel a.out, the
		 * Ehdr and Phdr[]'s can be before the sections that are
		 * captured by the Phdr[] PT_LOAD range, and are not included
		 * in the Phdr[] PT_LOAD range.
		 *
		 * To keep life really simple, we continue the range from the
		 * end of the Ehdr's processing, and continue to the end of
		 * the last PT_LOAD.
		 */
		if (offset) {
			size = (size_t)lphdr->p_offset +
			    (size_t)lphdr->p_filesz - offset;
			if ((offset + size) > msize)
				return (0);
			if (size)
				(*cb)(offset, size, cookie);
		}
	}

	/*
	 * If this is a dynamic object, and this is not a
	 * ELF_SR_SIGNED_INTERPRET request, we're done.
	 */
	if (process_segments && (skip_signature == _B_TRUE))
		return (1);

	/*
	 * If this is relocatable object or kernel module, we're going to
	 * essentially capture the whole file.  If this is a dynamic object
	 * then we're only looking for any SHT_SUNW_SIGNATURE section.
	 *
	 * If there is no Shdr[], something is odd.
	 */
	if (shdr == NULL)
		return (0);

	/*
	 * If the section header number is 0, then the actual number must be
	 * obtained from Shdr[0].
	 */
	if ((shnum = ehdr->e_shnum) == 0) {
		if ((ehdr->e_shoff == 0) ||
		    ((shnum = shdr[0].sh_size) == 0))
			return (0);

		extended_sections = _B_TRUE;
	}

	/*
	 * Scan the section headers to find the end range.  Note, if this is a
	 * kernel module that contains a Phdr[], this range will be captured
	 * after the Ehdr and before the first Shdr[].
	 */
	for (num = 0, nshdr = shdr; num < shnum; num++, nshdr++) {
		/*
		 * If we're processing a dynamic object, we're only interested
		 * in collecting the data for any SHT_SUNW_SIGNATURE section.
		 * Note, if the section has already have been captured during
		 * Phdr[] processing, ignore it.
		 */
		if (process_segments) {
			offset = (size_t)nshdr->sh_offset;

			if ((nshdr->sh_type == SHT_SUNW_SIGNATURE) &&
			    ((fphdr && (offset < (size_t)fphdr->p_offset)) ||
			    (lphdr && (offset > (size_t)lphdr->p_offset)))) {
				size = (size_t)nshdr->sh_size;

				if ((offset + size) > msize)
					return (0);
				if (size)
					(*cb)(offset, size, cookie);
			}
			continue;
		}

		/*
		 * Reset the offset if it has been cleared from processing a
		 * previously skipped section.
		 */
		if (offset == 0)
			offset = (size_t)nshdr->sh_offset;

		if (skip_section(nshdr, skip_checksum)) {
			/*
			 * Omit this section, and return the range collected up
			 * until the start of this section.
			 */
			size = (size_t)nshdr->sh_offset - offset;
			if ((offset + size) > msize)
				return (0);
			if (size)
				(*cb)(offset, size, cookie);

			/*
			 * Clear the offset, so that it can be assigned to the
			 * next section.
			 */
			offset = 0;
		}
	}

	/* If we're processing a dynamic object, we're done */
	if (process_segments)
		return (1);

	/*
	 * Scan the Shdr[] entries themselves, rather than the data the section
	 * headers point to.
	 */
	if (offset == 0)
		offset = (size_t)ehdr->e_shoff;

	/*
	 * Collect all elements of the Shdr[] until a SHT_SUNW_SIGNATURE section
	 * is found.  This section header should be skipped completely, as it
	 * may not have existed in the file when the original hash was created.
	 * Any sections that follow a SHT_SUNW_SIGNATURE section have their
	 * sh_offset skipped.  This allows the SHT_SUNW_SIGNATURE section to be
	 * added or to have changed size.  The most optimal layout for a signed
	 * file is for the SHT_SUNW_SIGNATURE to be the last section.
	 */
	for (sig = 0, num = 0, nshdr = shdr; num < shnum; num++, nshdr++) {

		if (offset == 0) {
			offset = (size_t)(ehdr->e_shoff) +
			    (size_t)(num * sizeof (Shdr));
		}

		/*
		 * If extended sections exist, skip the sh_size (e_shnum) and
		 * the sh_link (e_shstrndx) of Shdr[0].
		 */
		if (extended_sections == _B_TRUE) {
			size = (size_t)(ehdr->e_shoff) +
			    offsetof(Shdr, sh_size) - offset;
			if ((offset + size) > msize)
				return (0);
			if (size)
				(*cb)(offset, size, cookie);

			/* Reset the offset to follow sh_link */
			offset = (size_t)(ehdr->e_shoff) +
			    offsetof(Shdr, sh_link);
			extended_sections = _B_FALSE;

		} else if (nshdr->sh_type == SHT_SUNW_SIGNATURE) {
			size = (size_t)(ehdr->e_shoff) +
			    (size_t)(num * sizeof (Shdr)) - offset;
			if ((offset + size) > msize)
				return (0);
			if (size)
				(*cb)(offset, size, cookie);

			/* Reset the offset to capture next the next section */
			offset = 0;
			sig++;

		} else if (sig) {
			/*
			 * If a SHT_SUNW_SIGNATURE Shdr[] has been found, then
			 * any sections that follow may be at a different
			 * offset, should the SHT_SUNW_SIGNATURE section get
			 * rewritten and change size.  If SHT_SUNW_SIGNATURE
			 * sections are fixed length, none of this is necessary.
			 *
			 * Optimally, a SHT_SUNW_SIGNATURE section should be the
			 * last section of a file, which would result in this
			 * code not being exercised.
			 */
			size = (size_t)(ehdr->e_shoff) +
			    (size_t)(num * sizeof (Shdr)) +
			    offsetof(Shdr, sh_offset) - offset;
			if ((offset + size) > msize)
				return (0);
			if (size)
				(*cb)(offset, size, cookie);

			/* Reset the offset to the following element */
			offset = (size_t)(ehdr->e_shoff) +
			    (size_t)(num * sizeof (Shdr)) +
			    offsetof(Shdr, sh_size);
		}
	}

	/* Flush out the range remaining to the end of that last Shdr */
	if (offset) {
		size = (size_t)(ehdr->e_shoff) +
		    (size_t)(num * sizeof (Shdr)) - offset;
		if ((offset + size) > msize)
			return (0);
		if (size)
			(*cb)(offset, size, cookie);
	}
	return (1);
}
