/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Verify the elfsign(1) signature for the initial Solaris unix module.
 * Extract policy settings and keys.
 *
 * No ELF- or crypto-specific code should be in this file.
 *
 * ZVBOOT
 *   Define this for use by GRUB2 Solaris kernel zones zvboot loader to
 *   verify the unix module signature.
 *
 * EFI_VBOOT
 *   Define this for use by the GRUB2 multiboot loader to
 *   verify the unix module signature.
 */

#include <sys/types.h>
#include <sys/verified_boot.h>	/* Verified Boot types */
#include "verify.h"

#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/mm.h>		/* grub_*alloc* and other memory management */

#include <grub/dl.h>
GRUB_MOD_LICENSE("GPLv3+");

#if defined(ZVBOOT)
#include "zvcons.h"		/* panic(), console I/O */
#elif defined(EFI_VBOOT)
/* Invoke a panic by returning a non-0 value: */
#define	panic(...)	grub_printf("PANIC: ");  grub_printf(__VA_ARGS__); \
			return (1)
#endif	/* ZVBOOT, EFI_VBOOT */


/*
 * Default, hard-coded keys
 * default_modulus	is the modulus from
 *			/etc/certs/elfsign/Oracle_Solaris_2017.pem.
 *			This is for the current Solaris version.
 * s11_modulus		is the modulus from /etc/certs/elfsign/ORCLS11SE.
 *			This is for Solaris 11. Some non-ON Oracle
 *			software is signed with this key.
 * s12_modulus		is the modulus from /etc/certs/elfsign/ORCLS12SE.
 *			This is for Solaris 12.
 * internal_modulus	is the modulus from /etc/certs/elfsign/INTERNALSE.
 *			This is for developer builds only.
 */
#include <sys/verify_key.h>
#ifdef USE_INTERNAL_CERT
#include <sys/verify_key2.h>
#endif

/* Internal key storage values (not for use in firmware or virtual APIs) */
#define	ADDITIONAL_KEYS		1	/* ORCLS12SE */
#define	MAX_KEYS_INTERNAL	(MAX_VERIFY_PUBLIC_KEYS_V1 + ADDITIONAL_KEYS)

/*
 * Return the policies and elfsign public key(s).
 * On failure, only add the default keys.
 * The default policy is 'warning' set by the caller.
 */
static void
verify_get_settings(verified_boot_cert_t *certs, uint_t max_key_count,
    uint_t *key_count, verify_policy_t *boot_policy)
{
	/* Copy default, hard-coded keys */
	grub_memcpy(certs[*key_count].key, default_modulus,
	    VERIFY_MODULUS_LEN);
	grub_memcpy(certs[*key_count].dn_hash, default_dn_hash,
	    VERIFY_DN_HASH_LEN);
#ifdef	VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL,
	    "verify_get_settings: Oracle_Solaris_2017.pem key:\n");
	vboot_hex_print(&(*certs)[*key_count].key, VERIFY_MODULUS_LEN);
#endif
	(*key_count)++;

	grub_memcpy(certs[*key_count].key, s11_modulus,
	    VERIFY_MODULUS_LEN);
	grub_memcpy(certs[*key_count].dn_hash, s11_dn_hash,
	    VERIFY_DN_HASH_LEN);
#ifdef	VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL, "verify_get_settings: ORCLS11SE key:\n");
	vboot_hex_print(&(*certs)[*key_count].key, VERIFY_MODULUS_LEN);
#endif
	(*key_count)++;

	grub_memcpy(certs[*key_count].key, s12_modulus,
	    VERIFY_MODULUS_LEN);
	grub_memcpy(certs[*key_count].dn_hash, s12_dn_hash,
	    VERIFY_DN_HASH_LEN);
#ifdef	VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL, "verify_get_settings: ORCLS12SE key:\n");
	vboot_hex_print(&(*certs)[*key_count].key, VERIFY_MODULUS_LEN);
#endif
	(*key_count)++;

#ifdef USE_INTERNAL_CERT
	/* INTERNALSE key for in-house Solaris development builds only */
	grub_memcpy(certs[*key_count].key, internal_modulus,
	    VERIFY_MODULUS_LEN);
	grub_memcpy(certs[*key_count].dn_hash, internal_dn_hash,
	    VERIFY_DN_HASH_LEN);
#ifdef	VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL, "verify_get_settings: INTERNALSE key:\n");
	vboot_hex_print(&(*certs)[*key_count].key, VERIFY_MODULUS_LEN);
#endif
	(*key_count)++;
#endif /* USE_INTERNAL_CERT */

	(void) verify_get_settings_hv(certs, max_key_count, key_count,
	    boot_policy);

	grub_dprintf(DBGLABEL, "verify_get_settings: "
	    "have %d total keys with policy %d\n",
	    *key_count, *boot_policy);
}


/*
 * Verify an elfsign signature for a file.
 * For signature failures, ignore, print a warning message, or panic
 * depending on the boot_policy setting.
 * Return 0 on success, non-zero on failure.
 */
int
verify_module(const unsigned char *file)
{
	verify_policy_t		boot_policy = VERIFY_EFI_DEFAULT;
	uint_t			key_count = 0;
	verified_boot_cert_t	certs[MAX_KEYS_INTERNAL];
	int			rc = 0;

	grub_memset(certs, 0, sizeof (verify_public_certs_t));
	verify_get_settings(&certs[0], MAX_KEYS_INTERNAL,
	    &key_count, &boot_policy);

	if (boot_policy == VERIFY_NONE)
		return (0);

	/* Verify elfsign signature in ELF file module */
	if ((rc = verify_module_signature(file, certs, key_count))
	    != 0) {
		grub_dprintf(DBGLABEL, "verify_module: UNIX kernel image "
		    "verification failed; rc = %d.\n", rc);
	} else {
		grub_dprintf(DBGLABEL, "verify_module: UNIX kernel image "
		    "verification successful\n");
		return (0);
	}

	switch (boot_policy) {
	case VERIFY_ENFORCE:
		/* Fatal-level messages must have NO prefix: */
		grub_printf("Signature verification of UNIX kernel image "
		    "failed; boot-policy is set to \"enforce\".\n");
		panic("Unable to boot: boot-policy is set to \"enforce\".\n");
	case VERIFY_WARNING:
		/* Non-fatal messages must have a WARNING: or NOTICE: prefix: */
		grub_printf("WARNING: Signature verification of "
		    "UNIX kernel image failed\n");
		break;
	default:
		panic("Invalid Verified Boot policy value %d\n", boot_policy);
	}
	return (0);
}
