/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Internal header file for use in GRUB zvboot or multiboot.
 *
 * ZVBOOT
 *   Define this for use by GRUB2 Solaris kernel zones zvboot loader to
 *   verify the unix module signature.
 *
 * EFI_VBOOT
 *   Define this for use by the GRUB2 multiboot loader to
 *   verify the unix module signature.
 */

#ifndef _VERIFY_H
#define	_VERIFY_H

#include <sys/types.h>
#include <sys/verified_boot.h>	/* Verified Boot types */

#include <grub/mm.h>		/* grub_*alloc* and other memory management */
#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/time.h>		/* grub_sleep(), for debugging */

#if defined(EFI_VBOOT)		/* multiboot2 for GRUB UEFI mode */
#define	DBGLABEL		"vboot"
#include <grub/file.h>		/* grub_file_t defintions and functions */
#include <ctype.h>

#ifdef _BIG_ENDIAN
#define	htonl(x) (x)
#else
#define	htonl(x) (((uint32_t)(x) << 24) | (((uint32_t)(x) << 8) & 0xff0000) | \
	(((uint32_t)(x) >> 8) & 0xff00) | ((uint32_t)(x)  >> 24))
#endif

#elif defined(ZVBOOT)		/* zvboot for Kernel Zones */
#define	DBGLABEL		"zvboot"
#else
#error	"ZVBOOT or EFI_VBOOT needs to be defined."
#endif	/* ZVBOOT, EFI_VBOOT */

#ifdef	__cplusplus
extern "C" {
#endif

#define	vboot_alloc(size)	grub_malloc(size)
#define	vboot_free(buf, size)	grub_free(buf)
#define	bcopy(src, dst, n)	grub_memcpy((dst), (src), (n))
#define	bcmp(s1, s2, n)		grub_memcmp((s1), (s2), (n))
#define	bzero(s, n)		grub_memset((s), 0, (n))
#define	strchr(s, c)		grub_strchr((s), (c))
#define	strncpy(dst, src, n)	grub_strncpy((dst), (src), (n))

#define	VBOOT_DEBUG(...)	grub_dprintf(DBGLABEL, __VA_ARGS__)
#define	VBOOT_ERROR(...)	grub_printf(__VA_ARGS__)
#define	VBOOT_NOTICE(...)	grub_printf(__VA_ARGS__)

#define	VERIFY_EFI_DEFAULT	VERIFY_WARNING

#define	BOOT_POLICY_STRING(n)	((n) == 1 ? "none" : ((n) == 2 ? "warning" : \
	((n) == 3 ? "enforce" : "unknown")))

/* verify_zvboot.c or verify_efi.c */
extern int verify_get_settings_hv(verified_boot_cert_t *certs,
    uint_t max_cert_count, uint_t *cert_count, verify_policy_t *boot_policy);

/* verify_elf.c */
extern void vboot_hex_print(const void *data, size_t length);
extern void vboot_char_print(const void *data, size_t length);
extern int verify_module_signature(const uchar_t *file,
    verified_boot_cert_t *certs, uint_t num_keys);

/* verify_rsa.c */
extern int verify_signature_against_keys(verified_boot_cert_t *certs,
    unsigned int num_keys, const unsigned char *signature,
    size_t sig_len, const unsigned char *dn_hash,
    const unsigned char *data, size_t data_len);

/* verify.c */
extern int verify_module(const unsigned char *file);

#if defined(EFI_VBOOT)
/* verify_efi.c */
extern int multiboot2_verify_module(grub_file_t file);

/* verify_cert.c */
extern int get_decoded_keys_from_image(uchar_t *sbuf, uint64_t buffer_size,
    const char *cert_pathname, uchar_t **key);

/* verify_base64.c */
extern ssize_t b64_decode(void *outbuf, size_t outbufsz, const char *inbuf,
    const char *alpha, uint64_t flags);
#endif

#ifdef	__cplusplus
}
#endif

#endif	/* _VERIFY_H */
