/*
 *  GRUB  --  GRand Unified Bootloader
 * Copyright (c) 2007, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Base64 decoding function using 65 printable US-ASCII characters.
 * Defined in https://www.ietf.org/rfc/rfc4648.txt
 * For use only with Solaris Verified Boot verification.
 */

#include <sys/types.h>
#include <string.h>
#ifdef EFI_VBOOT
#include "verify.h"
#endif

static inline boolean_t
is_base64(unsigned char c)
{
	return ((c >= '0' && c <= '9') ||
	    (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ||
	    (c == '+') || (c == '/'));
}


/*
 * Decode printable characters, using the base64 algorithm.
 * Encoded string must be NUL-terminated.
 */
ssize_t
b64_decode(void *outbuf, size_t outbufsz, const char *inbuf, const char *alpha,
    uint64_t flags)
{
	static const char *b64_digits =
	    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	uchar_t		*decoded_str = (uchar_t *)outbuf;
	uchar_t		*encoded_str = (uchar_t *)inbuf;
	size_t		i = 0;
	size_t		j = 0;
	size_t		encoded_index = 0;
	ssize_t		decoded_index = 0;
	uchar_t		u, array4[4], array3[3];
	char		*p;

	if ((alpha != NULL) || (flags != 0)) {
		return (-1);
	}

	while ((u = encoded_str[encoded_index]) != '\0') {
		if (!is_base64(u)) {
			encoded_index++;
			continue;
		}
		if (decoded_index > outbufsz) {
			return (-1);
		}

		array4[i++] = encoded_str[encoded_index++];
		if (i == 4) {
			for (i = 0; i < 4; i++) {
				p = strchr(b64_digits, (char)array4[i]);
				if (p == NULL) {
					return (-1);
				}

				/* save base64 index */
				array4[i] = p - b64_digits;
			}

			array3[0] = (array4[0] << 2) +
			    ((array4[1] & 0x30) >> 4);
			array3[1] = ((array4[1] & 0xf) << 4) +
			    ((array4[2] & 0x3c) >> 2);
			array3[2] = ((array4[2] & 0x3) << 6) +
			    array4[3];

			for (i = 0; i < 3; i++) {
				decoded_str[decoded_index++] = array3[i];
			}

			i = 0;
		}
	}

	if (i != 0) {
		/* fill in the rest with zero */
		for (j = i; j < 4; j++)
			array4[j] = 0;

		for (j = 0; j < 4; j++) {
			p = strchr(b64_digits, (char)array4[j]);
			if (p == NULL) {
				return (-1);
			}

			array4[j] = p - b64_digits;
		}

		array3[0] = (array4[0] << 2) +
		    ((array4[1] & 0x30) >> 4);
		array3[1] = ((array4[1] & 0xf) << 4) +
		    ((array4[2] & 0x3c) >> 2);
		array3[2] = ((array4[2] & 0x3) << 6) +
		    array4[3];

		for (j = 0; j < (i - 1); j++) {
			decoded_str[decoded_index++] = array3[j];
		}
	}

	return (decoded_index);
}
