/*
 *  GRUB  --  GRand Unified Bootloader
 * Copyright (c) 2003, 2016, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Read and parse RSA X.509 certificates.
 * For use only with Solaris Verified Boot verification.
 *
 * Modified from several libkmf and other files.
 */

#include <sys/types.h>
#include <sys/stdbool.h>
#include <sys/note.h>
#include <sys/verified_boot.h>

#ifdef EFI_VBOOT
#include "verify.h"

#else
#include <sys/byteorder.h>	/* htonl() */
#include <sys/systm.h>		/* kernel bcopy() definition, strchr() */
#include <sys/kobj_impl.h>	/* internal bcopy() implementation */
#include "vboot/vboot.h"
#endif	/* EFI_VBOOT */

/* BER-encoded ASN.1 object tag identifiers */
#define	BER_INTEGER			0x02
#define	BER_BIT_STRING			0x03
#define	BER_CONSTRUCTED_SEQUENCE	0x30
#define	BER_CONSTRUCTED_SET		0x31

#define	BEGIN_CERT			"-----BEGIN"

typedef int32_t ber_tag_t;


/* Skip over a BER tag (type, length, value). */
static ber_tag_t
ber_skip_tag(uchar_t **sbuf, size_t *len, size_t max_len)
{
	ber_tag_t	tag;
	uchar_t		lc;
	size_t		noctects, diff;
	uint32_t	netlen;

	/* Assume no big tag */
	tag = lc = **sbuf;
	(*sbuf)++;

	*len = 0;
	netlen = 0;
	lc = **sbuf;
	(*sbuf)++;

	if ((lc & 0x80) != 0) {	/* long-form length */
		noctects = (lc & 0x7f);
		if (noctects > sizeof (ber_tag_t)) {
			return (-1);
		}
		diff = sizeof (ber_tag_t) - noctects;
		bcopy(*sbuf, (uchar_t *)&netlen + diff, noctects);
		*sbuf += noctects;
		*len = htonl(netlen);
	} else {		/* short-form length */
		*len = lc;
	}

	if (*len > max_len) {
		VBOOT_DEBUG("BER len %ld for tag 0x%x > max_len %ld\n",
		    *len, tag, max_len);
		*len = 0;
		return (-1);
	}

	return (tag);
}


/* Get the length for a tag. */
static size_t
calc_taglen(ber_tag_t tag)
{
	int		i;
	uint32_t	mask;

	/* Find first non-zero byte in tag. */
	for (i = sizeof (ber_tag_t) - 1; i > 0; i--) {
		mask = (0xffffff << (i * 8));
		if ((tag & mask) != 0) {
			break;
		}
	}

	return (i + 1);
}


/* Get the length of the length field. */
static size_t
calc_lenlen(size_t len)
{
	if (len <= 0x7f) {
		return (1);
	} else if (len <= 0xff) {
		return (2);
	} else if (len <= 0xffff) {
		return (3);
	} else if (len <= 0xffffff) {
		return (4);
	} else {
		return (5);
	}
}


/* Get the cert's valid date range (not before, not after). */
static void
get_validity(uchar_t **sbuf, size_t max_len)
{
	uchar_t		*buf, *tmpbuf;
	ber_tag_t	ber_tag = 0;
	size_t		ber_len = 0;

	tmpbuf = buf  = *sbuf;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf = tmpbuf;

	if (ber_tag != BER_CONSTRUCTED_SEQUENCE) { /* skip */
		return;
	}

	/* BER '{tata}' */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf += ber_len;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf += ber_len;

	*sbuf = buf;
}


/* Get the Relative Distinguished Name, RDN. */
static void
get_rdn(uchar_t **sbuf, size_t max_len)
{
	uchar_t		*buf, *tmpbuf;
	ber_tag_t	ber_tag = 0;
	size_t		ber_len = 0;

	buf = *sbuf;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);

	tmpbuf = buf;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf = tmpbuf;

	while (ber_tag == BER_CONSTRUCTED_SET) {
		ber_tag = ber_skip_tag(&buf, &ber_len, max_len);

		/* skip over the set */
		buf += ber_len;
		tmpbuf = buf;
		ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
		buf = tmpbuf;
	}

	*sbuf = buf;
}


/* Get the signing algorithm's OID. */
static void
get_algoid(uchar_t **sbuf, size_t max_len)
{
	uchar_t		*buf, *tmpbuf;
	ber_tag_t	ber_tag = 0;
	size_t		ber_len = 0;
	size_t		size;

	tmpbuf = buf = *sbuf;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf = tmpbuf;

	size = ber_len + calc_taglen(ber_tag) +
	    calc_lenlen(ber_len);
	buf += size;
	*sbuf = buf;
}


/* Extract the SPKI (Simple Public Key Infrastructure) data from the cert. */
static int
extract_spki_from_decoded_cert(uchar_t *decoded, uchar_t **ret,
    size_t *ret_len, size_t max_len)
{
	uchar_t		*tmpbuf, *spki;
	uchar_t		*buf = decoded;
	ber_tag_t	ber_tag = 0;
	size_t		ber_len = 0;

	/* Sequence header - skip tag */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);

	/* Extract to-be-signed (TBS) data. */
	/* BER '{' - skip tag */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);

	/* BER 't' - peek tag */
	tmpbuf = buf;
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	buf = tmpbuf;

	/* Get version number, if present. */
	if (ber_tag == 0xa0) {
		/* BER 'T' - skip tag */
		ber_tag = ber_skip_tag(&buf, &ber_len, max_len);

		/* BER 'i' - read int */
		ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
		buf += ber_len;
	}

	/* Get serial number. */
	/* BER 'I' - read bigint */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	if (ber_tag != BER_INTEGER) {
		VBOOT_DEBUG("tag is not BER_INTEGER\n");
		return (1);
	}
	buf += ber_len;

	/* Get signature algorithm, CA DN, validity dates, DN. */
	get_algoid(&buf, max_len);
	get_rdn(&buf, max_len);
	get_validity(&buf, max_len);
	get_rdn(&buf, max_len);

	/* BER '{' - skip tag */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	if (ber_len > max_len) {
		VBOOT_DEBUG("ber_len %ld > max_len %ld\n", ber_len, max_len);
		return (1);
	}

	get_algoid(&buf, max_len);

	/* BER 'B}' - get string */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	if (ber_tag == BER_BIT_STRING) {
		spki = vboot_alloc(ber_len);
		bcopy(buf, spki, ber_len);
	} else {
		VBOOT_DEBUG("tag is not BER_BIT_STRING\n");
		return (3);
	}

	*ret = spki;
	*ret_len = ber_len;

	return (0);
}


/*
 * Extract the RSA public key modulus from the SPKI data.
 * KeyParts has modulus at index 0 and pub exp at index 1.
 * Allocated memory at ret must be freed by caller.
 * Return 0 on success, otherwise non-zero.
 */
static int
extract_rsa_public_key_modulus_from_spki(uchar_t *spki, size_t spki_size,
    uchar_t **ret, size_t *ret_len)
{
	uchar_t		*modulus;
	uchar_t		*buf = spki;
	ber_tag_t	ber_tag;
	size_t		ber_len;
	size_t		max_len = spki_size;

	/* Strip leading 0's */
	while ((*buf == 0x00) && (spki_size > 0)) {
		buf++;
		spki_size--;
	}

	/* The first sequence is '{' */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	if (ber_len > max_len) {
		VBOOT_DEBUG("ber_len %ld > max_len %ld\n", ber_len, max_len);
		return (1);
	}

	/* The second sequence is 'I' */
	ber_tag = ber_skip_tag(&buf, &ber_len, max_len);
	if (ber_len > max_len) {
		VBOOT_DEBUG("ber_len %ld > max_len %ld\n", ber_len, max_len);
		return (1);
	}
	if (ber_tag != BER_INTEGER) {
		VBOOT_DEBUG("tag is not BER_INTEGER\n");
		return (1);
	}

	/* Strip leading 0's if the high-order bit is set. */
	while ((*buf == 0x00) && (ber_len > 0) && ((buf[1] & 0x80) != 0)) {
		buf++;
		ber_len--;
	}
	modulus = vboot_alloc(ber_len);
	(void) bcopy(buf, modulus, ber_len);

	*ret = modulus;
	*ret_len = ber_len;

	return (0);
}


/*
 * Determine if this is a binary DER format cert.
 * Otherwise, assume it's an ASCII PEM format file.
 */
static bool
is_der_format(const uchar_t *cert, size_t size)
{
	size_t	value_len, cert_len, i;

	if ((size < 2) || (cert[0] != 0x30)) {
		return (false);
	}

	if ((cert[1] & 0x80) == 0) {		/* short-form length */
		cert_len = cert[1] & 0x7f;
		size -= 2;
	} else {				/* long-form length */
		value_len = cert[1] & 0x7f;
		cert_len = 0;

		/* Ensure range doesn't overflow */
		if (size < 2 + value_len) {
			VBOOT_DEBUG("DER value length check failed "
			    "(value_len=%d, size=%lu)\n", value_len, size);
			return (false);
		}

		for (i = 2; i < (2 + value_len); i++) {
			cert_len = (cert_len << 8) + cert[i];
		}
		size -= (2 + value_len);
	}

	if (cert_len != size) {
		VBOOT_DEBUG("DER length check failed (len=%lu, size=%lu)\n",
		    cert_len, size);
		return (false);
	}

	return (true);
}


/*
 * Given an in-memory image of a elfsign X.509 certificate file,
 * extract the RSA-2048 public key modulus value in binary.
 * Allocated memory at *key must be freed by caller.
 */
int
get_decoded_keys_from_image(uchar_t *sbuf, size_t buffer_size,
    const char *cert_pathname, uchar_t **key)
{
	uchar_t		*encoded;
	uchar_t		*decoded = NULL;
	uchar_t		*spki = NULL;
	uchar_t		*modulus = NULL;
	size_t		i, j;
	ssize_t		decoded_size = 0;
	size_t		spki_size = 0;
	size_t		modulus_size = 0;
	int		rv;

	/* Sanity check */
	if (buffer_size == 0) {
		VBOOT_DEBUG("Empty cert file: %s\n", cert_pathname);
		return (3);
	}

	if (!is_der_format(sbuf, buffer_size)) {
		/*
		 * PEM format file. Convert to binary.
		 * Locate the BEGIN and END lines and copy lines in between.
		 */
		uchar_t		*tmpbuf;
		size_t		tmp_size;

		encoded = vboot_alloc(buffer_size);

		/* Flush everything up to -----BEGIN */
		for (i = 0; (i < buffer_size - sizeof (BEGIN_CERT)) &&
		    (bcmp(&sbuf[i], BEGIN_CERT, sizeof (BEGIN_CERT) - 1) != 0);
		    ++i) {
			;
		}
		if (i == buffer_size - sizeof (BEGIN_CERT)) { /* no BEGIN */
			VBOOT_DEBUG("No X.509 certificate found in file %s\n",
			    cert_pathname);
			vboot_free(encoded, buffer_size);
			return (9);
		}

		while ((sbuf[i] != '\n') && (i < buffer_size)) {
			++i;
		}
		++i;
		if (i == buffer_size) { /* no BEGIN line */
			VBOOT_DEBUG("No X.509 certificate found in file %s\n",
			    cert_pathname);
			vboot_free(encoded, buffer_size);
			return (9);
		}

		/* Copy encoded data between BEGIN and END CERTIFICATE lines */
		j = 0;
		while ((sbuf[i] != '-') && (i < buffer_size)) {
			encoded[j++] = sbuf[i++];
		}
		if ((j == 0) || (i == buffer_size)) { /* no END CERTIFICATE */
			VBOOT_DEBUG("No X.509 certificate found in file %s\n",
			    cert_pathname);
			vboot_free(encoded, buffer_size);
			return (10);
		}

		/* No need to decode the new line at the end */
		tmp_size = (((j + 3) / 4) * 3);
		if (tmp_size > buffer_size) {
			VBOOT_DEBUG("Cert strlen %ld > buffer size %ld\n",
			    tmp_size, buffer_size);
			vboot_free(encoded, buffer_size);
			return (11);
		}
		tmpbuf = vboot_alloc(tmp_size);

		encoded[j] = '\0';
		decoded_size = b64_decode(tmpbuf, tmp_size, (char *)encoded,
		    NULL, 0);
		vboot_free(encoded, buffer_size);
		if (decoded_size <= 0) {
			VBOOT_DEBUG("b64_decode failed for file %s; rc %ld\n",
			    cert_pathname, decoded_size);
			vboot_free(tmpbuf, tmp_size);
			return (12);
		} else if (decoded_size > buffer_size) {
			VBOOT_DEBUG("Decoded size %ld > buffer size %ld\n",
			    decoded_size, buffer_size);
			vboot_free(tmpbuf, tmp_size);
			return (18);
		}

		/* Check the decoded format's cert len */
		if (!is_der_format(tmpbuf, decoded_size)) {
			VBOOT_DEBUG("DER format check failed for file %s\n",
			    cert_pathname);
			vboot_free(tmpbuf, tmp_size);
			return (13);
		}

		/* Allocate the exact decoded size and copy to it */
		decoded = vboot_alloc(decoded_size);
		bcopy(tmpbuf, decoded, decoded_size);
		vboot_free(tmpbuf, tmp_size);

		sbuf = decoded;
		buffer_size = decoded_size;
	}

#ifdef	VERBOSE_VBOOT_DEBUG
	VBOOT_DEBUG("Decoded output, %lu bytes:\n", buffer_size);
	vboot_hex_print((unsigned char *)sbuf, buffer_size);
#endif

	/* Convert decoded data to SPKI */
	rv = extract_spki_from_decoded_cert(sbuf, &spki, &spki_size,
	    buffer_size);
	vboot_free(decoded, decoded_size);
	if (rv != 0) {
		VBOOT_DEBUG("DecodeSignedCertificate failed, cert file %s\n",
		    cert_pathname);
		return (15);
	}

#ifdef	VERBOSE_VBOOT_DEBUG
	VBOOT_DEBUG("SPKI output, %d bytes:\n", spki_size);
	vboot_hex_print((unsigned char *)spki, spki_size);
#endif

	rv = extract_rsa_public_key_modulus_from_spki(spki, spki_size,
	    &modulus, &modulus_size);
	vboot_free(spki, spki_size);
	if (rv != 0) {
		VBOOT_DEBUG("ExtractSPKI_RSA failed, cert file %s\n",
		    cert_pathname);
		return (16);
	}

	/* Check modulus length */
	if (modulus_size != VERIFY_MODULUS_LEN) {
		VBOOT_DEBUG("Key length is not 256 bytes, cert file %s\n",
		    cert_pathname);
		vboot_free(modulus, modulus_size);
		return (17);
	}

#ifdef	VERBOSE_VBOOT_DEBUG
	VBOOT_DEBUG("Modulus output, %d bytes:\n", modulus_size);
	vboot_hex_print((unsigned char *)modulus, modulus_size);
#endif

	*key = modulus;

	VBOOT_DEBUG("get_decoded_keys_from_image(%s) OK\n", cert_pathname);
	return (0);
}
