/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2017, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * X86 GRUB UEFI-specific functions for Solaris Verified Boot.
 *
 * No ELF- or crypto-specific code should be in this file.
 */

#ifndef EFI_VBOOT
#error	"EFI_VBOOT needs to be defined."
#endif
#ifndef GRUB_MACHINE_EFI
#error	"GRUB_MACHINE_EFI needs to be defined."
#endif

#include <sys/types.h>
#include <sys/verified_boot.h>	/* Verified Boot types */
#include "verify.h"

#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/mm.h>		/* grub_*alloc* and other memory management */
#include <grub/file.h>		/* grub_file_t defintions and functions */
#include <grub/env.h>		/* grub_env_get() */
#include <grub/time.h>		/* grub_sleep(), for debugging */

#include <grub/efi/efi.h>	/* grub_efi_* types and functions for UEFI */
#include <grub/efi/api.h>	/* GRUB_EFI_* constants */

#define	SIGNATURE_DATABASE	"db"
#define	MIN_DER_HEADER_LEN	2

/* GUID definitions */
#define	GUIDS_EQUAL(s1, s2)	(bcmp((s1), (s2), sizeof (*(s1))) == 0)
#define	EFI_IMAGE_SECURITY_DATABASE_GUID {0xd719b2cb, 0x3d3a, 0x4596, \
	{0xa3, 0xbc, 0xda, 0xd0, 0x0e, 0x67, 0x65, 0x6f}}
#define	EFI_X509_GUID {0xa5c059a1, 0x94e4, 0x4aa7, \
	{0x87, 0xb5, 0xab, 0x15, 0x5c, 0x2b, 0xf0, 0x72}}
#define	EFI_RSA2048_GUID {0x3c5766e8, 0x269c, 0x4e34, \
	{0xaa, 0x14, 0xed, 0x77, 0x6e, 0x85, 0xb3, 0xb6}}
#define	EFI_PKCS7_GUID {0x4aafd29d, 0x68df, 0x49ee, \
	{0x8a, 0xa9, 0x34, 0x7d, 0x37, 0x56, 0x65, 0xa7}}
#define	EFI_CERT_SHA256_GUID {0xc1c41626, 0x504c, 0x4092, \
	{0xac, 0xa9, 0x41, 0xf9, 0x36, 0x93, 0x43, 0x28}}
#define	EFI_MICROSOFT_KEK_CA_2011_GUID {0x77fa9abd, 0x0359, 0x4d32, \
	{0xbd, 0x60, 0x28, 0xf4, 0xe7, 0x8f, 0x78, 0x4b}}

/* UEFI certificate and signature definitions */
#pragma pack(1) /* disable alignment */
typedef struct _efi_signature_data {
	grub_efi_guid_t	SignatureOwner;
	uchar_t		SignatureData[];
} efi_signature_data;

typedef struct _efi_signature_list {
	grub_efi_guid_t	SignatureType;
	uint32_t	SignatureListSize;
	uint32_t	SignatureHeaderSize; /* 0 */
	uint32_t	SignatureSize; /* includes GUID size */
} efi_signature_list;
#pragma pack()


/* Access the boot_policy EFI SecureBoot variable and determine boot_policy. */
static verify_policy_t
verify_get_boot_policy(void)
{
	uchar_t		*secure_boot;
	grub_size_t	secure_boot_size = 0;
	int		secure_boot_enabled = 0;
	grub_efi_guid_t	global_guid = GRUB_EFI_GLOBAL_VARIABLE_GUID;
	const char	*boot_policy_str;
	verify_policy_t	boot_policy = VERIFY_EFI_DEFAULT;

	/* See if Secure Boot is enabled by reading EFI variable SecureBoot. */
	secure_boot = grub_efi_get_variable("SecureBoot", &global_guid,
	    &secure_boot_size);
	if ((secure_boot != NULL) && (secure_boot_size != 0) &&
	    (*secure_boot != 0)) {
		secure_boot_enabled = 1;
	}
	grub_dprintf(DBGLABEL, "SecureBoot %sabled (%u, len %lu)\n",
	    secure_boot_enabled ? "en" : "dis", secure_boot ? *secure_boot : 0,
	    secure_boot_size);
	grub_free(secure_boot);

	/* Read and validate boot_policy variable, if present. */
	boot_policy_str = grub_env_get("boot_policy");
	if (boot_policy_str != NULL) {
		if (grub_strcmp(boot_policy_str, "enforce") == 0) {
			boot_policy = VERIFY_ENFORCE;
		} else if (grub_strcmp(boot_policy_str, "warning") == 0) {
			boot_policy = VERIFY_WARNING;
		} else if (grub_strcmp(boot_policy_str, "none") == 0) {
			boot_policy = VERIFY_NONE;
		} else {
			grub_printf("Unrecognized boot_policy=%s "
			    "ignored.\n", boot_policy_str);
		}
	} else if (secure_boot_enabled) { /* Secure Boot default */
		boot_policy = VERIFY_ENFORCE;
		grub_dprintf(DBGLABEL,
		    "Default SecureBoot boot_policy=enforce.\n");
	} else { /* non-Secure Boot default */
		boot_policy = VERIFY_EFI_DEFAULT;
		grub_dprintf(DBGLABEL,
		    "Non-SecureBoot default boot_policy=%d\n", boot_policy);
	}

	/* In Secure Boot mode, the boot_policy must be enforce. */
	if (secure_boot_enabled) {
		if (boot_policy != VERIFY_ENFORCE) {
			boot_policy = VERIFY_ENFORCE;
			grub_dprintf(DBGLABEL, "boot_policy overridden in "
			    "Secure Boot mode to boot_policy=enforce.\n");
		}
	}

	grub_dprintf(DBGLABEL, "SecureBoot %sabled, boot_policy=%s\n",
	    secure_boot_enabled ? "en" : "dis",
	    BOOT_POLICY_STRING(boot_policy));

	return (boot_policy);
}


/* Convert a 128-bit GUID to a formatted hex string. */
static const char *
guid_str(const grub_efi_guid_t *guid)
{
	static char	buf[40];

	(void) grub_snprintf(buf, sizeof (buf),
	    "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
	    guid->data1, guid->data2, guid->data3,
	    guid->data4[0], guid->data4[1], guid->data4[2], guid->data4[3],
	    guid->data4[4], guid->data4[5], guid->data4[6], guid->data4[7]);
	return (buf);
}


/*
 * Access the X.509 certificate in the db database.
 * Return 0 if OK.
 * Otherwise return non-zero for other errors.
 */
static int
verify_get_certs(verified_boot_cert_t *certs, uint_t max_key_count,
    uint_t *key_count)
{
	static const grub_efi_guid_t	efi_security_db_guid =
	    EFI_IMAGE_SECURITY_DATABASE_GUID;
	static const grub_efi_guid_t	x509_guid = EFI_X509_GUID;
	static const grub_efi_guid_t	rsa2048_guid = EFI_RSA2048_GUID;
	static const grub_efi_guid_t	pkcs7_guid = EFI_PKCS7_GUID;
	static const grub_efi_guid_t	cert_sha256_guid = EFI_CERT_SHA256_GUID;
	static const grub_efi_guid_t	microsoft_kek_ca_2011_guid =
	    EFI_MICROSOFT_KEK_CA_2011_GUID;
	grub_size_t		data_size = 0;
	grub_ssize_t		remainder, cert_size;
	uchar_t			*data, *key, *key_buf;
	const char		*cert_type, *guid_strp;
	efi_signature_list	*cert_list;
	efi_signature_data	*cert;
	grub_efi_guid_t		guid;
	int			rv, cert_index;

	grub_dprintf(DBGLABEL, "verify_get_certs() called\n");

	data = grub_efi_get_variable(SIGNATURE_DATABASE, &efi_security_db_guid,
	    &data_size);
	if (data != NULL) {
		cert_list = (efi_signature_list *)data;
	} else {
		grub_dprintf(DBGLABEL, "no data in the db database\n");
		return (1);
	}

	/* A UEFI signature db consists of zero or more signature lists. */
	for (cert_list = (efi_signature_list *)data, remainder = data_size;
	    remainder > 0;
	    remainder -= cert_list->SignatureListSize,
	    cert_list = (efi_signature_list *)((uchar_t *)cert_list +
	    cert_list->SignatureListSize)) {

		/* For Verified Boot, we're only interested in X.509 certs */
		if (!GUIDS_EQUAL(&cert_list->SignatureType, &x509_guid)) {
			if (GUIDS_EQUAL(&cert_list->SignatureType,
			    &rsa2048_guid)) {
				cert_type = "RSA-2048";
			} else if (GUIDS_EQUAL(&cert_list->SignatureType,
			    &pkcs7_guid)) {
				cert_type = "PKCS-7";
			} else if (GUIDS_EQUAL(&cert_list->SignatureType,
			    &cert_sha256_guid)) {
				cert_type = "SHA-256";
			} else { /* unknown */
				cert_type = guid_str(&cert_list->SignatureType);
			}
			grub_dprintf(DBGLABEL, "Skipping unsupported "
			    "SignatureType %s.\n", cert_type);
			continue;
		}

		/* Look at each cert in this signature list */
		for (cert = (efi_signature_data *)((uchar_t *)cert_list +
		    sizeof (efi_signature_list) +
		    cert_list->SignatureHeaderSize);
		    (uchar_t *)cert < ((uchar_t *)cert_list +
		    cert_list->SignatureListSize);
		    cert = (efi_signature_data *)((uchar_t *)cert +
		    cert_list->SignatureSize)) {

			guid = cert->SignatureOwner;
			guid_strp = guid_str(&guid);
			cert_size = cert_list->SignatureSize -
			    sizeof (grub_efi_guid_t);

			if (cert_size <= MIN_DER_HEADER_LEN) { /* skip */
				grub_dprintf(DBGLABEL,
				    "Skipping undersized cert %s, size %ld\n",
				    guid_strp, cert_size);
				continue;
			}

			/* Filter-out certs irrelevant to Verified Boot */
			if (GUIDS_EQUAL(&guid, &microsoft_kek_ca_2011_guid)) {
				continue;
			}

			grub_dprintf(DBGLABEL, "cert %s, size %lu\n",
			    guid_strp, cert_size);

			/* Extract the public key from the X.509 cert */
			rv = get_decoded_keys_from_image(
			    &cert->SignatureData[0], cert_size,
			    guid_strp, &key);
			if (rv == 0) {
				cert_index = *key_count;
				grub_dprintf(DBGLABEL,
				    "get_decoded_keys_from_image(%s), "
				    "key %d, OK\n", cert_index, guid_strp);
				if (cert_index < max_key_count) {
					key_buf = certs[cert_index].key;
					bcopy(key, key_buf,
					    VERIFY_MODULUS_LEN);
					vboot_free(key, VERIFY_MODULUS_LEN);
					bzero(certs[cert_index].dn_hash,
					    VERIFY_DN_HASH_LEN);
#ifdef	VERBOSE_VBOOT_DEBUG
					grub_dprintf(DBGLABEL,
					    "key %d from file %s:\n",
					    cert_index, cert_pathname);
					vboot_hex_print(key_buf,
					    VERIFY_MODULUS_LEN);
#endif
					++*key_count;
				} else {
					grub_printf("Key %s ignored; "
					    "exceeded maximum of %u keys\n",
					    cert_index);
					goto ret;
				}
			} else {
				grub_dprintf(DBGLABEL,
				    "get_decoded_keys_from_image(%s) "
				    "FAILED; rv=%d\n", guid_strp, rv);
			}
		}
	}

ret:
	grub_dprintf(DBGLABEL, "verify_get_certs() returning; key_count=%u\n",
	    *key_count);
	grub_free(data);
	return (0);
}


/*
 * Get keys and policy information from X86 UEFI.
 */
int
verify_get_settings_hv(verified_boot_cert_t *certs, uint_t max_key_count,
    uint_t *key_count, verify_policy_t *policy)
{
	*policy = verify_get_boot_policy();

	return (verify_get_certs(certs, max_key_count, key_count));
}


/*
 * Verify an elfsign signature in an ELF file:
 * Allocate memory for an ELF file and read the entire file into memory.
 * Verify the elfsign signature in the file, if any.
 * Free the file memory before returning.
 * Called from multiboot_mbi2.c#grub_multiboot_load().
 * Return 0 on success, non-zero on failure.
 * Returning non-zero causes GRUB to go back to the GRUB boot menu.
 */
int
efi_verify_module(grub_file_t file)
{
	grub_off_t		file_size;
	grub_properly_aligned_t	*file_buffer;
	grub_ssize_t		len_read;
	grub_err_t		err;

	file_size = grub_file_size(file);
	if (file_size == GRUB_FILE_SIZE_UNKNOWN) {
		grub_dprintf(DBGLABEL, "efi_verify_module: "
		    "UNIX file size unknown\n");
		return (GRUB_ERR_FILE_READ_ERROR);
	}
	grub_dprintf(DBGLABEL, "efi_verify_module: "
	    "UNIX file size = 0x%lx\n", file_size);

	file_buffer = grub_malloc(file_size);
	if (file_buffer == NULL) {
		grub_dprintf(DBGLABEL, "efi_verify_module: "
		    "cannot allocate buffer for UNIX\n");
		return (GRUB_ERR_OUT_OF_MEMORY);
	}

	if ((grub_file_seek(file, 0)) == (grub_off_t)-1) {
		grub_dprintf(DBGLABEL, "efi_verify_module: "
		    "cannot seek UNIX file to 0\n");
		err = GRUB_ERR_FILE_READ_ERROR;
		goto err_ret;
	}

	len_read = grub_file_read(file, file_buffer, file_size);
	if (len_read < (grub_ssize_t)file_size) {
		grub_dprintf(DBGLABEL, "efi_verify_module: "
		    "short length 0x%lx read from file UNIX\n", len_read);
		err = GRUB_ERR_FILE_READ_ERROR;
		goto err_ret;
	}

	err = verify_module((uchar_t *)file_buffer);

err_ret:
	grub_free(file_buffer);
	return (err);
}
