/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Verify the elfsign(1) signature for the initial Solaris unix module.
 * Calculate the ELF digest and extract the ELF signature.
 *
 * No Kernel Zone-specific code should be in this file.
 */

#include <sys/types.h>
#include <sys/stdbool.h>
#define	ELF_TARGET_ALL		/* include both X86 and SPARC ELF specifics */
#include <sys/elf.h>
#include <sys/sha2.h>		/* SHA256_DIGEST_LENGTH, SHA512_DIGEST_LENGTH */
#include <sys/note.h>

#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/mm.h>		/* grub_*alloc* and other memory management */
#include <grub/crypto.h>	/* crypto and digest functions */

#include <sys/verified_boot.h>		/* Verified Boot types */
#include "verify.h"

#ifndef	_ELF64
#define	_ELF64			/* Required to define _elf64_sign_range() */
#endif
#include "elfsignrange/_elfsignrange.h"

#define	EMPTY_SIGNATURE_SECTION_SIZE	8

/* Get number of sections. For extended sections, get from shdr[0].sh_size. */
#define	GET_SHNUM(ehdrp, shdrp) \
	((ehdrp)->e_shstrndx != SHN_XINDEX ? (ehdrp)->e_shnum : \
	(shdrp)[0].sh_size)
/* Get string table section. For extended sections, get from shdr[0].sh_link. */
#define	GET_SHSTRNDX(ehdrp, shdrp) \
	((ehdrp)->e_shstrndx != SHN_XINDEX ? (ehdrp)->e_shstrndx : \
	(shdrp)[0].sh_link)

/* Return a pointer to a byte offset into the ELF file */
#define	ELF_GET_PTR(base, off)		(void *)((uchar_t *)(base) + (off))

/* Unknown size, so use a generous maximum */
#define	MAX_KERNEL_MODULE	0x10000000

/*
 * Use this to pass parameters to elfsignrange callback functions.
 * In the future when multiple digest algorithms are supported, that
 * will have to be added to this callback function too.
 */
typedef struct {
	uchar_t		*elf_file;
	size_t		cur_size;
	size_t		sig_offset;
	size_t		sig_size;
	SHA2_CTX	ctx;
} elf_callback_data_t;

/*
 * Print the filename and elfsign hash string for a module (for debugging).
 * Called by verify_hash_legacy().
 */
#ifdef	VERBOSE_VBOOT_DEBUG
static void
verify_print_es_hash(const char *function, const uchar_t *hash, size_t hash_len)
{
	char	digest_string[SHA256_DIGEST_LENGTH * 2 + 1];
	int	i;

	if (hash_len > SHA256_DIGEST_LENGTH) {
		hash_len = SHA256_DIGEST_LENGTH;
	}

	/* Convert hex digits to ASCII characters */
	for (i = 0; i < hash_len; ++i) {
		grub_snprintf(digest_string + i * 2, 3, "%02x",
		    hash[i]);
	}
	digest_string[i * 2] = '\0';

	/* Print message */
	grub_dprintf(DBGLABEL, "%s: hash = %s\n", function, digest_string);
}
#endif


/* Print binary data in hexadecimal (for debugging) */
void
vboot_hex_print(const void *data, size_t length)
{
#define	BYTES_PER_LINE	32
	int		i, j;
	char		linebuf[81];
	int		line_limit, remaining;

	for (i = 0; i < length; i += line_limit) {
		remaining = length - i;
		line_limit = (remaining > BYTES_PER_LINE) ?
		    BYTES_PER_LINE : remaining;

		for (j = 0; j < line_limit; ++j) {
			(void) grub_snprintf(linebuf + j * 2, 3, "%02x",
			    ((unsigned char *)data)[i + j]);
		}
		linebuf[j * 2] = '\0';
		grub_dprintf(DBGLABEL, "  %s\n", linebuf);
	}
}

/* Print ASCII characters in binary data (otherwise ".") (for debugging) */
void
vboot_char_print(const void *data, size_t length)
{
#define	CHARS_PER_LINE	64
	int	i, j;
	char	linebuf[81];
	int	line_limit, remaining;

	for (i = 0; i < length; i += line_limit) {
		remaining = length - i;
		line_limit = (remaining > CHARS_PER_LINE) ?
		    CHARS_PER_LINE : remaining;

		for (j = 0; j < line_limit; ++j) {
			char	dbyte = ((char *)data)[i + j];
			linebuf[j] = grub_isprint(dbyte) ? dbyte : '.';
		}
		linebuf[j] = '\0';
		grub_dprintf(DBGLABEL, "  %s\n", linebuf);
	}
}

/*
 * Sanity self-test for SHA-256.
 * Return 0 for pass, non-zero for fail.
 *
 * SHA Test data from NIST Cryptographic Algorithm Validation Program (CAVP).
 * Test vector "SHA256ShortMsg.rsp Len = 512" downloaded from
 * http://csrc.nist.gov/groups/STM/cavp/documents/shs/shabytetestvectors.zip
 */
static int
verify_signature_self_test_sha256(void)
{
	/* Divide data arbitrarily to test two SHA256Update operations: */
	static const uchar_t sha256_testdata1[53] = {0x5a, 0x86, 0xb7, 0x37,
	    0xea, 0xea, 0x8e, 0xe9, 0x76, 0xa0, 0xa2, 0x4d, 0xa6, 0x3e, 0x7e,
	    0xd7, 0xee, 0xfa, 0xd1, 0x8a, 0x10, 0x1c, 0x12, 0x11, 0xe2, 0xb3,
	    0x65, 0x0c, 0x51, 0x87, 0xc2, 0xa8, 0xa6, 0x50, 0x54, 0x72, 0x08,
	    0x25, 0x1f, 0x6d, 0x42, 0x37, 0xe6, 0x61, 0xc7, 0xbf, 0x4c, 0x77,
	    0xf3, 0x35, 0x39, 0x03, 0x94};
	static const uchar_t sha256_testdata2[11] = {
	    0xc3, 0x7f, 0xa1, 0xa9, 0xf9, 0xbe, 0x83, 0x6a, 0xc2, 0x85, 0x09};
	static const uchar_t sha256_digest[SHA256_DIGEST_LENGTH] = {
	    0x42, 0xe6, 0x1e, 0x17, 0x4f, 0xbb, 0x38, 0x97, 0xd6, 0xdd, 0x6c,
	    0xef, 0x3d, 0xd2, 0x80, 0x2f, 0xe6, 0x7b, 0x33, 0x19, 0x53, 0xb0,
	    0x61, 0x14, 0xa6, 0x5c, 0x77, 0x28, 0x59, 0xdf, 0xc1, 0xaa};

	grub_uint8_t	sha256_ctx[GRUB_MD_SHA256->contextsize];
	uchar_t		digest[SHA256_DIGEST_LENGTH];
	int		r;

	/* Test SHA-256 */
	GRUB_MD_SHA256->init(sha256_ctx);
	GRUB_MD_SHA256->write(sha256_ctx, sha256_testdata1,
	    sizeof (sha256_testdata1));
	GRUB_MD_SHA256->write(sha256_ctx, sha256_testdata2,
	    sizeof (sha256_testdata2));
	GRUB_MD_SHA256->final(sha256_ctx);
	grub_memcpy(digest, GRUB_MD_SHA256->read(sha256_ctx),
	    GRUB_MD_SHA256->mdlen);

#ifdef	VERBOSE_VBOOT_DEBUG
	verify_print_es_hash("verify_signature_self_test_sha256", digest,
	    GRUB_MD_SHA256->mdlen);
#endif

	r = grub_memcmp(digest, sha256_digest, SHA256_DIGEST_LENGTH);
	if (r != 0) {
		grub_dprintf(DBGLABEL,
		    "verify_signature_self_test_sha256 failed\n");
	}

	return (r);
}


/*
 * Detect overlap between data offset and length and the signature section
 * offset and length.
 * Return 0 if OK (no overlap).
 * Return 1 if there's overlap, along with information about the
 * non-overlapping parts, if any.
 */
static int
verify_detect_overlap(size_t data_off, size_t data_len, size_t sig_off,
    size_t sig_len, size_t *prefix_off, size_t *prefix_len,
    size_t *postfix_off, size_t *postfix_len)
{
	size_t	data_end = data_off + data_len - 1;
	size_t	sig_end = sig_off + sig_len - 1;

	*prefix_off = *prefix_len = *postfix_off = *postfix_len = 0;

	if (data_off == 0 || data_len == 0 || sig_off == 0 ||
	    sig_len <= EMPTY_SIGNATURE_SECTION_SIZE ||
	    sig_len > SIG_MAX_LENGTH) {
		return (0); /* no valid signature or not enough information */
	} else if ((data_end < sig_off) || (data_off > sig_end)) {
		return (0); /* no overlap! */
	}

	/* Now that we know we have overlap, specify the overlap regions */
	grub_dprintf(DBGLABEL, "verify_detect_overlap: detected overlap with "
	    "signature section at offset=%#zx, length=%#zx, end=%#zx",
	    sig_off, sig_len, sig_end);
	if ((data_off == sig_off) && (data_len == sig_len)) {
		grub_dprintf(DBGLABEL, "verify_detect_overlap: "
		    "exact overlap with signature section detected");
	} else {
		if (data_end >= sig_off) {
			grub_dprintf(DBGLABEL, "verify_detect_overlap: "
			    "data end >= sign begin");
		}
		if (data_off <= sig_end) {
			grub_dprintf(DBGLABEL, "verify_detect_overlap: "
			    "data begin <= sign end");
		}
	}

	/* Calculate non-overlapping regions, if any */
	if (data_off < sig_off) {
		*prefix_off = data_off;
		*prefix_len = sig_off - data_off;
		grub_dprintf(DBGLABEL, "verify_detect_overlap: "
		    "prefix %#zx, len %#zx, end %#zx",
		    *prefix_off,  *prefix_len,
		    *prefix_off + *prefix_len - 1);
	}
	if (data_end > sig_end) {
		*postfix_off = sig_end + 1;
		*postfix_len = data_end - sig_end;
		grub_dprintf(DBGLABEL, "verify_detect_overlap: "
		    "postfix %#zx, len %#zx, end %#zx",
		    *postfix_off, *postfix_len,
		    *postfix_off + *postfix_len - 1);
	}

	return (1);
}


/*
 * Callback function for elfsignrange() for accumulating the digest.
 * Pass parameters via the opaque elf_callback_data_t pointer, udata.
 *
 * To handle corrupt/incorrect ELF files, we exclude data that overlaps the
 * signature section region of the file, if any.
 */
static void
verify_digest_callback(size_t src_off, size_t size, void *udata)
{
	elf_callback_data_t	*cb_data = (elf_callback_data_t *)udata;
	size_t			prefix_off = 0;
	size_t			prefix_len = 0;
	size_t			postfix_off = 0;
	size_t			postfix_len = 0;

	grub_dprintf(DBGLABEL, "vboot_digest_callback"
	    "(src_off 0x%lx, size 0x%lx, end 0x%lx, udata %p)\n",
	    (unsigned long)src_off, (unsigned long)size,
	    (unsigned long)src_off + size - 1, udata);
	if (size == 0) {
		return; /* nothing to do */
	}

	if (verify_detect_overlap(src_off, size, cb_data->sig_offset,
	    cb_data->sig_size, &prefix_off, &prefix_len,
	    &postfix_off, &postfix_len) != 0) {
		/* Handle parts that are not overlapping, if any */
		if (prefix_len != 0) {
			grub_dprintf(DBGLABEL, "verify_digest_callback: "
			    "processing non-overlapping part before signature");
			verify_digest_callback(prefix_off, prefix_len, udata);
		}
		if (postfix_len != 0) {
			grub_dprintf(DBGLABEL, "verify_digest_callback: "
			    "processing non-overlapping part after signature");
			verify_digest_callback(postfix_off, postfix_len,
			    udata);
		}

	} else {
		GRUB_MD_SHA256->write(&cb_data->ctx,
		    ELF_GET_PTR(cb_data->elf_file, src_off), size);
		cb_data->cur_size += size;
	}
}


/*
 * Calculate the elfsign hash of the ELF sections using elfsignrange.
 * Return 0 on success, else non-zero.
 */
static int
verify_hash_elfsignrange(filesig_vers_t filesig_version, const Elf64_Ehdr *hdr,
    const Elf64_Shdr *shdrs, const Elf64_Phdr *phdrs, const uchar_t *elf_file,
    uchar_t *hash, size_t sig_offset, size_t sig_len)
{
	Elf_Signrange		range_type;
	elf_callback_data_t	cb_data = {NULL, 0, 0, 0, {0}};
	int			ret = 0;

	/* Convert elfsign version to elfsignrange range type */
	switch (filesig_version) {
	case FILESIG_VERSION7_INTERPRET:
		range_type = ELF_SR_INTERPRET;
		break;
	case FILESIG_VERSION9_RELOBJ:
		range_type = ELF_SR_RELOBJ;
		break;
	default:
		grub_dprintf(DBGLABEL, "verify_hash_elfsignrange: "
		    "unsupported version %u\n", filesig_version);
		return (-1);
	}

	/*
	 * Calculate hash using elfsignrange callbacks that determine
	 * what regions are hashed.
	 */
	grub_memset(&cb_data.ctx, 0, sizeof (&cb_data.ctx));
	GRUB_MD_SHA256->init(&cb_data.ctx);
	cb_data.elf_file = (uchar_t *)elf_file;
	cb_data.sig_offset = sig_offset;
	cb_data.sig_size = sig_len;

	grub_dprintf(DBGLABEL, "verify_hash_elfsignrange: "
	    "calling _elf64_sign_range(%srange %u)\n",
	    ((phdrs != NULL) ? "phdrs, " : ""), range_type);

	if (_elf64_sign_range(verify_digest_callback, (Elf64_Ehdr *)hdr,
	    (Elf64_Phdr *)phdrs, (Elf64_Shdr *)shdrs, MAX_KERNEL_MODULE,
	    range_type, &cb_data) == 0) {
		grub_dprintf(DBGLABEL, "verify_hash_elfsignrange: "
		    "_elf64_sign_range(range %u) failed\n", range_type);
		ret = -1;
	}

	GRUB_MD_SHA256->final(&cb_data.ctx);
	grub_memcpy(hash, GRUB_MD_SHA256->read(&cb_data.ctx),
	    GRUB_MD_SHA256->mdlen);

	/* Cleanup */
	grub_dprintf(DBGLABEL, "verify_hash_elfsignrange:  hash value is\n");
	vboot_hex_print(hash, SHA256_DIGEST_LENGTH);

	grub_memset(&cb_data.ctx, 0, sizeof (&cb_data.ctx));

	/* Sanity check */
	if (cb_data.cur_size == 0) {
		grub_dprintf(DBGLABEL, "verify_hash_elfsign_range: "
		    "No data parsed by gelf_sign_range()\n");
	}
	grub_dprintf(DBGLABEL, "verify_hash_elfsignrange: "
	    "processed %lu bytes\n", cb_data.cur_size);
	return (ret);
}


/*
 * Calculate the elfsign hash of ELF sections using the legacy range algorithm.
 * Only elfsign version 5 (SHA-256) is supported here.
 * Called by verify_module_signature().
 */
static void
verify_hash_legacy(filesig_vers_t filesig_version, const Elf64_Ehdr *hdr,
    uchar_t *hash, const char *shstrtab, const uchar_t *elf_file)
{
#ifndef VERBOSE_VBOOT_DEBUG
	_NOTE(ARGUNUSED(shstrtab));
#endif
	size_t		shn, shnum;
	Elf64_Shdr	*shdrs = (Elf64_Shdr *)(ELF_GET_PTR(elf_file,
	    hdr->e_shoff));
	Elf64_Shdr	*shp;
	grub_uint8_t	sha256_ctx[GRUB_MD_SHA256->contextsize];

	if (filesig_version != FILESIG_VERSION5) {
		grub_dprintf(DBGLABEL, "elfsign signature version %u "
		    "is not supported.\n", filesig_version);
		return;
	}

	/* Initialize digest computation */
	grub_memset(sha256_ctx, 0, sizeof (sha256_ctx));
	GRUB_MD_SHA256->init(sha256_ctx);

	/*
	 * Examine each ELF section header.
	 * Skip section 0 which is required to be empty.
	 */
	shnum = GET_SHNUM(hdr, shdrs);
	for (shn = 1; shn < shnum; ++shn) {
		shp = (Elf64_Shdr *)((char *)shdrs + (shn * hdr->e_shentsize));

		/* Skip empty, SHT_NOBITS, and SHT_SUNW_SIGNATURE sections */
		if ((shp->sh_size == 0) || (shp->sh_type == SHT_NOBITS) ||
		    (shp->sh_type == SHT_SUNW_SIGNATURE)) {
#ifdef	VERBOSE_VBOOT_DEBUG
			grub_dprintf(DBGLABEL,
			    "verify_hash_legacy: skipping section %s\n",
			    shstrtab + shp->sh_name);
#endif
			continue;
		}

#ifdef	VERBOSE_VBOOT_DEBUG
		grub_dprintf(DBGLABEL, "verify_hash_legacy: Section %d: ", shn);
		grub_dprintf(DBGLABEL, "type %u, ", shp->sh_type);
		grub_dprintf(DBGLABEL, "addr %lx, ", shp->sh_addr);
		grub_dprintf(DBGLABEL, "size 0x%lx, ", shp->sh_size);
		grub_dprintf(DBGLABEL, "name %s\n", shstrtab + shp->sh_name);
#endif

		GRUB_MD_SHA256->write(sha256_ctx,
		    ELF_GET_PTR(elf_file, shp->sh_offset), shp->sh_size);
	}

	GRUB_MD_SHA256->final(sha256_ctx);
	grub_memcpy(hash, GRUB_MD_SHA256->read(sha256_ctx),
	    GRUB_MD_SHA256->mdlen);

#ifdef	VERBOSE_VBOOT_DEBUG
	verify_print_es_hash("verify_hash_legacy", hash,
	    GRUB_MD_SHA256->mdlen);
#endif
}


/* Calculate the SHA-256 hash of some data, specifically the DN */
static void
verify_hash_dn(uchar_t *dn_hash, uchar_t *dn, size_t dn_len)
{
	grub_uint8_t    sha256_ctx[GRUB_MD_SHA256->contextsize];

	GRUB_MD_SHA256->init(sha256_ctx);
	GRUB_MD_SHA256->write(sha256_ctx, dn, dn_len);
	GRUB_MD_SHA256->final(sha256_ctx);
	grub_memcpy(dn_hash, GRUB_MD_SHA256->read(sha256_ctx),
	    VERIFY_DN_HASH_LEN);
}

/*
 * Verify the that the module matches the signature in the
 * .SUNW_SIGNATURE section for the module, if it exists.
 *
 * Return 0 if verified, non-0 if not verified.
 *
 * Called by verify_module().
 */
int
verify_module_signature(const uchar_t *elf_file,
    verified_boot_cert_t *certs, uint_t num_keys)
{
	Elf64_Ehdr		hdr;
	Elf64_Shdr		*shdrs, *shp;
	Elf64_Phdr		*phdrs = NULL;
	union {
		char			cdata[SIG_MAX_LENGTH];
		filesignatures_t	filesig;
	} *buf = NULL;
	size_t			shstrlen, phdrslen;
	char			*shstrtab = NULL;
	uint_t			signature_count = 0; /* should end up == 1 */
	size_t			sig_offset, sig_size_only;
	filesig_vers_t		filesig_version;
	size_t			shn, shnum, shstrndx;
	uchar_t			*dn = NULL;
	uchar_t			*sig = NULL;
	char			*sig_oid = NULL;
	size_t			dn_len, sig_len, sig_oid_len;
	uchar_t			dn_hash[VERIFY_DN_HASH_LEN];
	uchar_t			elf_hash[SHA256_DIGEST_LENGTH] = {0};
	int			r, ret;
	static bool		first_time = true;

	if (first_time) {
		/* Perform one-time self tests */
		if (verify_signature_self_test_sha256() != 0) {
			grub_dprintf(DBGLABEL, "Error: "
			    "Failed verify_signature_self_test_sha256.\n");
			return (1);
		}
		first_time = false;
	}

	/* Get ELF header */
	grub_memcpy(&hdr, ELF_GET_PTR(elf_file, 0), sizeof (hdr));

	/* ELF file format sanity checks */
	VERIFY_SANITY_CHECK_ELF_HDR(VBOOT_DEBUG, &hdr, MAX_KERNEL_MODULE, r);
	if (r != 0) {
		VERIFY_SANITY_UNIX_WANBOOT_ELF_HDR(VBOOT_DEBUG, &hdr, r);
	}
	if (r != 0) {
		return (2);
	}

	/* Get ELF section header */
	shdrs = (Elf64_Shdr *)ELF_GET_PTR(elf_file, hdr.e_shoff);
	shnum = GET_SHNUM(&hdr, shdrs);
	shstrndx = GET_SHSTRNDX(&hdr, shdrs);

	/* Get ELF section table length and pointer */
	shp = (Elf64_Shdr *)((char *)shdrs + shstrndx * hdr.e_shentsize);
	shstrlen = shp->sh_size;

	shstrtab = (char *)ELF_GET_PTR(elf_file, shp->sh_offset);

	/* Get ELF Program header (if any) */
	phdrslen = (size_t)hdr.e_phnum * hdr.e_phentsize;
	if (phdrslen > 0) {
		phdrs = (Elf64_Phdr *)ELF_GET_PTR(elf_file, hdr.e_phoff);
		grub_dprintf(DBGLABEL, "Phdr found; length %lu",
		    (long unsigned)phdrslen);
	}

	/*
	 * Search the ELF section table for the SUNW_SIGNATURE section.
	 * Skip section 0 which is required to be empty.
	 */
	for (shn = 1; shn < shnum; ++shn) {
		shp = (Elf64_Shdr *)((char *)shdrs + (shn * hdr.e_shentsize));

		if (shp->sh_type != SHT_SUNW_SIGNATURE)
			continue;

		if ((shp->sh_flags & SHF_ALLOC) != 0) {
			grub_dprintf(DBGLABEL, "verify_module_signature: "
			    "Error: memory-resident signature section %u\n",
			    shn);
			ret = 3;
			goto cleanup;
		}

		if ((shp->sh_size != 0) &&
		    (shp->sh_name < shstrlen) &&
		    (grub_strncmp(shstrtab + shp->sh_name,
		    ELF_SIGNATURE_SECTION,
		    sizeof (ELF_SIGNATURE_SECTION)) == 0) &&
		    ((shp->sh_flags & SHF_SUNW_ABSENT) == 0)) {
			/* hit!--signature found */
			++signature_count;
			break;
		}
	}

	if (signature_count == 0) {
		grub_dprintf(DBGLABEL, "verify_module_signature: Error: "
		    "kernel module is not signed with elfsign(1)\n");
		ret = 4;
		goto cleanup;
	} else if (signature_count != 1) {
		grub_dprintf(DBGLABEL, "verify_module_signature: Error: "
		    "kernel module has multiple ELF signature sections\n");
		ret = 5;
		goto cleanup;
	}

	/* Sanity check signature section size and allocate memory for it. */
	sig_size_only = shp->sh_size;
	sig_offset = shp->sh_offset;
	if ((sig_size_only > SIG_MAX_LENGTH) || (sig_size_only == 0)) {
		grub_dprintf(DBGLABEL, "verify_module_signature: Error: "
		    ".SUNW_SIGNATURE length error %lu\n",
		    (long unsigned)sig_size_only);
		ret = 6;
		goto cleanup;
	}

	buf = grub_malloc(sig_size_only);

	/* Copy elfsign signature section to an aligned buffer */
	grub_memcpy(buf->cdata, ELF_GET_PTR(elf_file, sig_offset),
	    sig_size_only);

	/*
	 * Extract information from .SUNW_SIGNATURE section.
	 */
	filesig_version = buf->filesig.filesig_sig.filesig_version;

#ifdef VERBOSE_VBOOT_DEBUG
	/* Print information from .SUNW_SIGNATURE section. */
	grub_dprintf(DBGLABEL, "verify_module_signature: "
	    "filesig_version is 0x%x\n", filesig_version);
	grub_dprintf(DBGLABEL, "verify_module_signature: Dumping "
	    "SHT_SUNW_SIGNATURE section (length %ld):\n", sig_size_only);
	vboot_hex_print(buf->cdata, sig_size_only);
	vboot_char_print(buf->cdata, sig_size_only);
#endif

	/*
	 * Check elfsign version.
	 * Elfsign versions 1-4, 6 are not supported for Solaris Verified Boot.
	 * Version 5 works with Solaris Verified Boot back to S11.2.
	 */
	switch (filesig_version) {
	case FILESIG_VERSION5:
	case FILESIG_VERSION7_INTERPRET:
	case FILESIG_VERSION9_RELOBJ:
		break;
	default:
		grub_dprintf(DBGLABEL, "verify_module_signature: Error: "
		    "elfsign signature version %u not supported;",
		    filesig_version);
		ret = 7;
		goto cleanup;
	}

	/*
	 * Extract DN, signature, OID, & lengths from .SUNW_SIGNATURE section.
	 */
	dn_len = buf->filesig.filesig_sig.filesig_v3_dnsize;
	dn = (uchar_t *)&buf->filesig.filesig_sig.filesig_v3_data[0];
	sig_len = buf->filesig.filesig_sig.filesig_v3_sigsize;
	sig = (uchar_t *)&buf->filesig.filesig_sig.filesig_v3_data[dn_len];
	sig_oid_len = buf->filesig.filesig_sig.filesig_v3_oidsize;
	sig_oid = (char *)&buf->filesig.filesig_sig.filesig_v3_data[
	    dn_len + sig_len];

	/* Sanity check signature block */
	VERIFY_SANITY_CHECK_SIGNATURE_SECTION(VBOOT_DEBUG, sig_offset, sig,
	    sig_len, dn, dn_len, sig_oid, sig_oid_len, MAX_KERNEL_MODULE, r);
	if (r != 0) {
		ret = 8;
		goto cleanup;
	}

	/* Check hash algorithm OID */
	if (grub_strncmp(sig_oid, OID_SHA256WITHRSAENCRYPTION, sig_oid_len)
	    != 0) {
		grub_dprintf(DBGLABEL, "verify_module_signature: unexpected "
		    "OID %s; expected %s for elfsign signature version %u\n",
		    sig_oid, OID_SHA256WITHRSAENCRYPTION, filesig_version);
		ret = 11;
		goto cleanup;
	}

	verify_hash_dn(dn_hash, dn, dn_len);

	/*
	 * Calculate elfsign hash.
	 */
	if (filesig_version >= ELFSIGN_RANGE_BASE) {
		if (verify_hash_elfsignrange(filesig_version, &hdr,
		    shdrs, phdrs, elf_file, elf_hash,
		    sig_offset, sig_size_only) != 0) {
			ret = 12;
			goto cleanup;
		}
	} else {
		verify_hash_legacy(filesig_version, &hdr, elf_hash, shstrtab,
		    elf_file);
	}

	/*
	 * Now that we have the elfsign hash, signature, and keys,
	 * verify the hash against the signature using one or more public keys.
	 */
	if (verify_signature_against_keys(certs, num_keys, sig, sig_len,
	    dn_hash, elf_hash, sizeof (elf_hash)) != 0) {
		ret = 13;
		goto cleanup;
	}

	ret = 0; /* success */

cleanup:
	if (buf != NULL)
		grub_free(buf);

	return (ret);
}
