/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Verify elfsign RSA-2048 signature using libgcrypt MPI
 * (arbitrary precision or "bignum") functions.
 * The actual SHA-256 digest used here is computed beforehand.
 *
 * Libgcrypt MPI documentation:
 * http://www.gnupg.org/documentation/manuals/gcrypt/MPI-library.html
 */

#include <sys/types.h>
#include <sys/stdbool.h>
#include <visibility.h>		/* _gcry_mpi_* */
#include <gcrypt.h>		/* gcry_mpi_* */
#include <sys/verified_boot.h>	/* VERIFY_MODULUS_LEN */
#include "verify.h"
#include <grub/misc.h>		/* grub_str*() and other useful utilities */

#define	MIN_PKCS1_PADLEN	11

/*
 * RSA-2048 public key modulus for testing.
 * The modulus "n" must match the "n" in the private key.
 */
static const unsigned char raw_test_modulus[] = {
	0xc3, 0xa1, 0xab, 0x2c, 0x88, 0x01, 0x48, 0x4f,
	0x62, 0x4a, 0x35, 0x8e, 0x3d, 0x95, 0x32, 0xe4,
	0x7a, 0x48, 0x68, 0x66, 0xb8, 0x92, 0x01, 0x9d,
	0xe7, 0x25, 0xd8, 0x75, 0xbb, 0xe3, 0x60, 0xa0,
	0xd0, 0x06, 0x2d, 0xbd, 0x98, 0xc9, 0x8b, 0xd2,
	0x69, 0x04, 0xc0, 0x28, 0x1f, 0x2f, 0xd0, 0xd6,
	0x47, 0x71, 0x82, 0x78, 0x62, 0x35, 0x1e, 0xa8,
	0x9d, 0xb4, 0x7c, 0x66, 0x05, 0x25, 0x77, 0xf2,
	0x9b, 0x7b, 0xaf, 0x8d, 0xfa, 0x5d, 0x34, 0xee,
	0xec, 0xb4, 0xc4, 0xb2, 0xb1, 0x76, 0x35, 0xf9,
	0x47, 0xdd, 0xd1, 0xc0, 0xb9, 0xb3, 0xd6, 0x70,
	0xaf, 0x0c, 0x48, 0x9f, 0x7c, 0xce, 0xb5, 0x04,
	0xf3, 0x73, 0xdc, 0x2c, 0x26, 0x47, 0xd9, 0x46,
	0x28, 0x52, 0xff, 0x70, 0x08, 0x03, 0x4e, 0xa6,
	0x72, 0x62, 0x53, 0xfe, 0x87, 0x8e, 0x44, 0x86,
	0x73, 0xb6, 0x5d, 0xe7, 0xf8, 0xea, 0xf8, 0x1f,
	0x46, 0xb6, 0x5b, 0xcb, 0x90, 0x74, 0x62, 0xf4,
	0xad, 0xf6, 0x24, 0xe8, 0xa9, 0x2f, 0x53, 0xd0,
	0x09, 0x43, 0x41, 0x9a, 0x5b, 0xf1, 0xcf, 0xb4,
	0x0c, 0x54, 0x16, 0xda, 0x66, 0x53, 0xa9, 0x48,
	0x69, 0xa8, 0xe5, 0x15, 0x08, 0x89, 0xc9, 0xbc,
	0x2e, 0x82, 0xdb, 0x87, 0x29, 0x98, 0x60, 0x34,
	0x59, 0xda, 0x34, 0xf8, 0x26, 0x1e, 0xa8, 0x4a,
	0x49, 0xa0, 0xe3, 0xa2, 0xeb, 0x37, 0x63, 0xfe,
	0x2c, 0x21, 0x9d, 0x9c, 0x1f, 0xac, 0x45, 0xba,
	0x0e, 0x4c, 0x59, 0xb3, 0xb5, 0xed, 0xc4, 0xcf,
	0xd6, 0xf7, 0xa8, 0x93, 0xd2, 0x8c, 0xac, 0x8f,
	0xbd, 0xb9, 0x80, 0xef, 0x17, 0x26, 0x0a, 0xe0,
	0x68, 0xb0, 0xd4, 0x18, 0x2b, 0x86, 0xf2, 0x8b,
	0xa6, 0xf7, 0x1b, 0xa7, 0xa0, 0x4f, 0xb5, 0xad,
	0x1b, 0xcc, 0x3a, 0x4b, 0x5d, 0xf7, 0x82, 0xf4,
	0xf1, 0xb2, 0xa5, 0x18, 0x03, 0xdf, 0x4b, 0x0d};

static const unsigned char hash_unpadded_test_data[] = {
	0xd1, 0x2d, 0xf9, 0x27, 0xc9, 0x6b, 0xdd, 0x47,
	0xb8, 0xc0, 0x30, 0x78, 0xfd, 0xe6, 0x6f, 0x8b,
	0x6c, 0xfa, 0x5b, 0x62, 0xd7, 0xb2, 0xc9, 0x51,
	0x2c, 0x48, 0x94, 0xa3, 0x93, 0x36, 0xcb, 0x79};

/* Test signature for the above key and data */
static const unsigned char raw_test_signature[] = {
	0xae, 0x42, 0xc8, 0xf7, 0x57, 0xb4, 0x73, 0xe6,
	0xef, 0x11, 0x32, 0x0e, 0xc3, 0xb5, 0x36, 0x9a,
	0xb0, 0x25, 0x7e, 0xfd, 0x71, 0xa0, 0x49, 0x4b,
	0x09, 0x5d, 0xee, 0x95, 0x49, 0xd7, 0xc0, 0x48,
	0x5f, 0x9a, 0x52, 0x70, 0x30, 0xa9, 0x48, 0xbf,
	0xae, 0x3c, 0x07, 0xc4, 0xdd, 0x7e, 0xa0, 0x43,
	0xac, 0x54, 0xf7, 0x5b, 0x21, 0x5d, 0xc4, 0xe2,
	0x5d, 0xf6, 0xca, 0xd3, 0xbe, 0x20, 0xac, 0xda,
	0xb4, 0xec, 0xd6, 0x04, 0x9f, 0x02, 0x37, 0xcc,
	0x37, 0x29, 0x34, 0xb9, 0x37, 0x39, 0x4a, 0xa9,
	0xe1, 0xfd, 0xfe, 0xfc, 0x60, 0x84, 0x1e, 0x17,
	0x5d, 0x54, 0xf4, 0x24, 0xb7, 0xcf, 0x6f, 0x99,
	0xff, 0xfa, 0xfb, 0xb3, 0x0a, 0x43, 0xab, 0x12,
	0x24, 0x90, 0x3e, 0xde, 0xfb, 0xed, 0x37, 0xa1,
	0x03, 0x1a, 0x41, 0xf9, 0x2f, 0xc4, 0x85, 0x28,
	0xb2, 0x62, 0x0e, 0x94, 0xb6, 0xf6, 0xda, 0x22,
	0x32, 0xce, 0xbc, 0x4d, 0x76, 0xc7, 0x0c, 0x68,
	0x11, 0xf0, 0xd9, 0x91, 0x14, 0x02, 0x76, 0xde,
	0x9f, 0x6f, 0x9e, 0xb8, 0x04, 0x9e, 0x40, 0x4b,
	0xfa, 0xed, 0x05, 0x0b, 0x1e, 0x94, 0x27, 0xe9,
	0xd3, 0x9e, 0x1c, 0x72, 0x22, 0x8c, 0x36, 0x9e,
	0x70, 0xe5, 0x13, 0x28, 0x17, 0x4a, 0xb5, 0x83,
	0xfc, 0x50, 0x58, 0xb4, 0x8c, 0xba, 0x67, 0x97,
	0xd2, 0x2f, 0x51, 0x0e, 0x5a, 0xbb, 0xf0, 0xea,
	0x6e, 0xc0, 0xee, 0x15, 0xb6, 0x4a, 0x28, 0x5f,
	0x75, 0x6f, 0xaa, 0xec, 0x80, 0xba, 0x08, 0xf8,
	0xde, 0x2c, 0x8a, 0xff, 0x51, 0x28, 0xe3, 0xac,
	0x27, 0xaa, 0xa6, 0xce, 0x47, 0xe0, 0x66, 0x26,
	0xc7, 0x3d, 0x80, 0xc8, 0x60, 0x28, 0xbf, 0x4b,
	0xd3, 0xe4, 0x56, 0x1f, 0x4b, 0x94, 0x16, 0xc0,
	0x64, 0xd4, 0x8a, 0x2d, 0x09, 0x9e, 0x93, 0xf3,
	0xd9, 0xcf, 0xba, 0xd2, 0xe2, 0xb7, 0x89, 0x48};


/*
 * pad_data_to_signature_length()
 * Prepend padding to data in PKCS#1 format prior to verifying a signature.
 * The PKCS#1 EME-PKCS1-v1_5 padding format for signatures is:
 *	0x00, 0x01, 0xff, . . ., 0xff, 0x00, (original unpadded data)
 */
static int
pad_data_to_signature_length(unsigned char *padded_buffer, size_t buff_len,
    const unsigned char *data, size_t data_len)
{
	ssize_t		start = buff_len - data_len;

	/* Sanity checks */
	if ((buff_len > VERIFY_MODULUS_LEN) || (start < MIN_PKCS1_PADLEN)) {
		return (1);
	}

	/* PKCS#1 padding for signatures begins with 0x00, 0x01 */
	padded_buffer[0] = 0x00;
	padded_buffer[1] = 0x01;

	/* Pad remaining bytes, except for last byte, with 0xff */
	grub_memset(&padded_buffer[2], 0xff, start - (2 + 1));

	/* Last byte of padding ends with 0x00 */
	padded_buffer[start -1] = 0x00;

	/* Copy data at end of buffer */
	grub_memcpy(&padded_buffer[start], data, data_len);

	return (0);
}


/*
 * verify_signature()
 * Verify an elfsign(1) RSA signature, given data and a RSA elfsign public key.
 * Return 0 if a signature and a key matches against data, otherwise non-zero.
 *
 * A signature is verified using RSA encryption as follows:
 *		hash data = (signature^exponent) % modulus
 * where:
 * hash data	the expected hash (SHA-256) of selected ELF sections.
 *		This is compared against the actual hash
 * signature	the 256-byte RSA-2048 signature from the ELF
 *		.SUNW_SIGNATURE section
 * exponent	65537 (constant). The exponent part of the elfsign public key
 * modulus	the 256-byte RSA-2048 modulus part of the elfsign public key
 *
 * Parameters:
 * modulus		The modulus from the elfsign public key
 *			(RSA-2048 for versions 5-6)
 * signature		RSA signature from ELF SUNW_signature section
 * data			Hash of selected ELF sections (SHA-256 for version 5-6)
 * *_len		Byte lengths of the key, signature, and data above
 */
static int
verify_signature(const unsigned char *modulus,
    const unsigned char *signature, size_t signature_len,
    const unsigned char *data, size_t data_len)
{
	static const unsigned char	exponent65537[] = {0x01, 0x00, 0x01};
	int				r;
	gcry_mpi_t			pkey_mpi = NULL;
	gcry_mpi_t			padded_data_mpi = NULL;
	gcry_mpi_t			sig_mpi = NULL;
	gcry_mpi_t			exponent_mpi = NULL;
	gcry_mpi_t			result_mpi = NULL;
	unsigned char			padded_buffer[VERIFY_MODULUS_LEN];

	/* sanity checks */
	if ((modulus == NULL) || (signature == NULL) || (data == NULL)) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "Error: NULL pointer passed to verify_signature()\n");
		return (2);
	}

#ifdef	VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL,  "verify_signature: Unpadded hash data:\n");
	vboot_hex_print(data, data_len);
#endif

	/* Pad data-to-be-verified to signature length using PKCS#1 padding */
	r = pad_data_to_signature_length(padded_buffer, sizeof (padded_buffer),
	    data, data_len);
	if (r != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "pad_data_to_signature_length() length error; rc %d\n", r);
		goto ret;
	}

	/* Convert data, key, signature, and exponent to internal MPI format */
	if ((r = gcry_mpi_scan(&padded_data_mpi, GCRYMPI_FMT_USG,
	    padded_buffer, sizeof (padded_buffer), NULL)) != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "gcry_mpi_scan(data) returned %d\n", r);
		goto ret;
	}

	if ((r = gcry_mpi_scan(&pkey_mpi, GCRYMPI_FMT_USG,
	    modulus, VERIFY_MODULUS_LEN, NULL)) != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "gcry_mpi_scan(key) returned %d\n", r);
		goto ret;
	}

	if ((r = gcry_mpi_scan(&sig_mpi, GCRYMPI_FMT_USG,
	    signature, signature_len, NULL)) != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "gcry_mpi_scan(sig) returned %d\n", r);
		goto ret;
	}

	if ((r = gcry_mpi_scan(&exponent_mpi, GCRYMPI_FMT_USG,
	    exponent65537, sizeof (exponent65537), NULL)) != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: "
		    "gcry_mpi_scan(sig) returned %d\n", r);
		goto ret;
	}

	/* Allocate memory to hold result */
	result_mpi = gcry_mpi_new(VERIFY_MODULUS_LEN * 8);
	if (result_mpi == NULL) {
		r = 4;
		goto ret;
	}

	/*
	 * Verify RSA signature by computing the following:
	 *   expectedSHAChecksumPadded = (signature^65537) % publicKeyModulus
	 * Then comparing the result with the actual SHA checksum + padding.
	 */
	gcry_mpi_powm(result_mpi, sig_mpi, exponent_mpi, pkey_mpi);

	r = gcry_mpi_cmp(result_mpi, padded_data_mpi);
	if (r != 0) {
		grub_dprintf(DBGLABEL, "verify_signature: signature "
		    "does NOT match; gcry_mpi_cmp() returned: %d\n", r);
	}

ret:
	gcry_mpi_release(pkey_mpi);
	gcry_mpi_release(padded_data_mpi);
	gcry_mpi_release(sig_mpi);
	gcry_mpi_release(exponent_mpi);
	gcry_mpi_release(result_mpi);

	return (r);
}


/*
 * Sanity test for signature verification.
 * Return 0 for pass, non-zero for fail.
 */
static int
verify_signature_self_test_rsa(void)
{
	/* Verify data with a known signature and RSA key */
	return (verify_signature(raw_test_modulus,
	    raw_test_signature, sizeof (raw_test_signature),
	    hash_unpadded_test_data, sizeof (hash_unpadded_test_data)));
}


/*
 * verify_signature_against_keys()
 * Verify an elfsign(1) RSA signature, given data and 1 or more
 * elfsign raw RSA public keys.
 * Return 0 if a signature and a key matches against data, otherwise non-zero.
 *
 * Parameters:
 * keys			One or more RSA-2048 elfsign public key(s).
 * dn_hash		SHA-256 hash of the DN
 * signature		RSA signature from ELF SUNW_signature section
 * data			SHA-256 Hash of selected ELF sections
 * *_len		Byte lengths of the key, signature, and data
 */
int
verify_signature_against_keys(
    verified_boot_cert_t *certs, unsigned int num_keys,
    const unsigned char *signature, size_t sig_len,
    const unsigned char *dn_hash, const unsigned char *data, size_t data_len)
{
	static const unsigned char	zero_key[VERIFY_MODULUS_LEN] = {'\0'};
	static const unsigned char	zero_dn_hash[VERIFY_DN_HASH_LEN] = {0};
	const unsigned char		*cur_key, *prev_key;
	const unsigned char		*cur_dn_hash;
	int				i;
	int				r = -1;
	static bool			first_time = true;

	if (first_time) {
		/* Perform one-time self tests */
		if (verify_signature_self_test_rsa() != 0) {
			grub_dprintf(DBGLABEL, "verify_signature_against_keys: "
			    "Failed verify_signature_self_test_rsa.\n");
			return (-2);
		}
		first_time = false;
	}

#ifdef VERBOSE_VBOOT_DEBUG
	grub_dprintf(DBGLABEL, "verify_signature_against_keys: dn_hash:\n");
	vboot_hex_print(dn_hash, VERIFY_DN_HASH_LEN);
#endif

	if (num_keys == 0) {
		grub_printf("Internal error: no verified boot keys\n");
		return (-3);
	}

	prev_key = zero_key;

	/* Verify signature against all keys until successful */
	for (i = 0; i < num_keys; i++, prev_key = cur_key) {

		cur_key = certs[i].key;
		cur_dn_hash = certs[i].dn_hash;

		/* skip key if it's all 0s */
		if (grub_memcmp(cur_key, zero_key,
		    VERIFY_MODULUS_LEN) == 0) {
			grub_dprintf(DBGLABEL, "verify_signature_against_keys: "
			    "skipping zero key #%u\n", i);
			r = 1;
			continue;
		}

		/* skip key if it's the same as the previous key */
		if (grub_memcmp(cur_key, prev_key,
		    VERIFY_MODULUS_LEN) == 0) {
			grub_dprintf(DBGLABEL, "verify_signature_against_keys: "
			    "skipping dup key #%u\n", i);
			r = 2;
			continue;
		}

		/* skip if current DN does not match the desired DN */
		if ((grub_memcmp(cur_dn_hash, zero_dn_hash,
		    VERIFY_DN_HASH_LEN) != 0) && (grub_memcmp(cur_dn_hash,
		    dn_hash, VERIFY_DN_HASH_LEN) != 0)) {
			grub_dprintf(DBGLABEL, "verify_signature_against_keys: "
			    "skipping key #%u (DN does not match)\n", i);
			r = 3;
			continue;
		}

		if ((r = verify_signature(cur_key, signature, sig_len,
		    data, data_len)) == 0) {
			return (0); /* success */
		}
	}

	grub_dprintf(DBGLABEL,
	    "verify_signature_against_keys: FAILED, ret=%d\n", r);
	return (r); /* fail */
}
