/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Solaris kernel zones zvboot-specific functions for Solaris Verified Boot.
 *
 * No ELF- or crypto-specific code should be in this file.
 */

#include <sys/types.h>
#include <sys/verified_boot.h>	/* Verified Boot types */
#include "verify.h"

#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/mm.h>		/* grub_*alloc* and other memory management */

#include "zv_asm.h"			/* do_hypercall() */
#include "zvcons.h"			/* panic(), console I/O */
#include <sys/zvmm/zvmm_hcall.h>	/* Hypercalls */
#include <sys/zvmm/zvmm_mach_hcall.h>	/* X86 Hypercalls */
#include <sys/zone.h>

#ifndef ZVBOOT
#error	"ZVBOOT needs to be defined."
#endif

#define	ALIGN_SIZE 4096


/*
 * Return non-0 if a key is already present, else 0.
 */
static int
verify_dup_key_check(const uchar_t *key, verify_public_certs_t *certs,
    uint_t key_count)
{
	int i;

	for (i = 0; i < key_count; ++i) {
		if (grub_memcmp(key, (*certs)[i].key, VERIFY_MODULUS_LEN)
		    == 0) {
			return (1);
		}
	}
	return (0);
}


/*
 * Get cert key/dn pairs and other information from the Hypervisor.
 * Return 0 if OK.
 * Return -1 if the HV call is not implemented.
 * Otherwise return non-zero for other errors.
 */
int
verify_get_settings_hv(verified_boot_cert_t *certs, uint_t max_key_count,
    uint_t *key_count, verify_policy_t *boot_policy)
{
	static uint64_t buffer[2 * ALIGN_SIZE / sizeof (uint64_t)];
	union {
		verified_boot_info_v1_t	*p;
		intptr_t		n;
	} vboot_align = { .n = ((intptr_t)buffer + ALIGN_SIZE - 1) &
	    ~(ALIGN_SIZE - 1UL) };

	uint64_t vboot_size = sizeof (verified_boot_info_v1_t);
	int i, rc;

	rc = do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_GET_ZONEATTR,
	    ZONE_ATTR_VBOOT_INFO, vboot_align.p, &vboot_size);
	if (rc != 0) {
		grub_dprintf("zvboot", "verify_get_settings_hv: "
		    "do_hypercall() failed; error %d\n", rc);
		return (-1);
	}

	grub_dprintf("zvboot", "verify_get_settings_hv: "
	    "do_hypercall() succeeded with policy %d, key_count %d\n",
	    (int)vboot_align.p->policy, (int)vboot_align.p->key_count);

	*boot_policy = vboot_align.p->policy;

	for (i = 0; (i < vboot_align.p->key_count) &&
	    (*key_count < max_key_count); i++) {
		if (verify_dup_key_check(vboot_align.p->cert[i].key,
		    certs, *key_count) != 0) {
			grub_dprintf("zvboot", "Skipping dup key %d.\n", i);
#ifdef	VERBOSE_VBOOT_DEBUG
			vboot_hex_print(vboot_align.p->cert[i].key,
			    VERIFY_MODULUS_LEN);
#endif
			continue;
		}

		grub_memcpy(&certs[*key_count], (void *)&vboot_align.p->cert[i],
		    sizeof (verified_boot_cert_t));

		grub_dprintf("zvboot", "verify_get_settings_hv: HV key %d.\n",
		    i);
#ifdef	VERBOSE_VBOOT_DEBUG
		vboot_hex_print(certs[*key_count].key,
		    VERIFY_MODULUS_LEN);
#endif

		++(*key_count);
	}

	return (0);
}
