/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2013, 2014, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Gnasher: a randomising hypercall stress test.
 *
 * Gnasher attempts to test the input validation and error handling of
 * host extender drivers by performing pseudo-random hypercalls. It
 * will determine the set of extenders configured for the guest domain
 * and then attempt to exercise them at random.
 *
 * For each test hypercall, gnasher will generate random arguments:
 *
 * - argument 1 is the extender number.
 * - argument 2 is usually the hypercall function number. A value
 *   between 0 and 63 is generated.
 * - argument 3 is often a guest physical address. A value between 0
 *   and 4G is generated.
 * - arguments 4 though 6 vary in use. 64 bit random values are
 *   generated.
 *
 * The simplest way to use gnasher is to boot with the "-B
 * zvboot_gnasher=true" flag:
 *
 *   # zoneadm -z colony boot -- -B zvboot_gnasher=true
 *
 * Standard host behaviour is to kill a guest that issues an invalid
 * hypercall. Most of the hypercalls generated by gnasher will be
 * invalid, which means that testing can be slow. The behaviour of the
 * host can be modified by setting the kernel variable
 * "zvmm_allow_bad_hypercall" to 1 before beginning any testing:
 *
 *   # echo "zvmm_allow_bad_hypercall/w 1" | mdb -kw
 *
 * Similarly, the host will object if the guest causes a host extender
 * to write to memory that should not be writable by the guest. This
 * can be modified using:
 *
 *   # echo "zvmm_allow_write_to_readonly/w 1" | mdb -kw
 *
 * It is a good idea to monitor the output of gnasher (using "zlogin
 * -C colony") to see the hypercall sequence that leads up to a host
 * panic.
 *
 * Please tag any bugs found using gnasher with "gnasher". 
 */

#include <stdarg.h>
#include <sys/types.h>
#include <sys/errno.h>

#include <sys/zvmm/zvmm_hcall.h>
#include <sys/zvmm/zvmm_mach_hcall.h>
#include <sys/zvmm/zvmm_ext_common.h>
#include <sys/zvmm/zvterm_common.h>
#include <sys/zvmm/zvblk_common.h>
#include <sys/zvmm/zvsdir_common.h>
#include <sys/zvmm/zvnet_common.h>

#include <grub/mm.h>
#include <grub/misc.h>

#include "zv_asm.h"
#include "zv_mem.h"
#include "zvcons.h"
#include "gnasher.h"

typedef struct blacklist_entry {
	struct blacklist_entry *next;
	char *type;
	uint64_t extender;
	uint64_t function;
	char *reason;
} blacklist_entry_t;

static blacklist_entry_t *blacklist = NULL;

static zv_extinfo_v2_t *ext_p;
static int ext_n;

static void
blacklist_add(char *t, uint64_t e, uint64_t f, char *r)
{
	blacklist_entry_t *new = grub_malloc(sizeof (*new));
	if (new == NULL)
		panic("Cannot allocate blacklist entry.");

	new->type = NULL;
	if (t != NULL)
		new->type = grub_strdup(t);
	new->extender = e;
	new->function = f;
	new->reason = r;

	new->next = blacklist;
	blacklist = new;
}

static blacklist_entry_t *
blacklist_find(uint64_t e, uint64_t f, uint64_t arg3, uint64_t arg4,
    uint64_t arg5, uint64_t arg6)
{
	char *type = ext_p[e].ze_name;

	blacklist_entry_t *l;
	for (l = blacklist; l != NULL; l = l->next) {
		/*
		 * If this blacklist entry specifies an extender
		 * number and it matches, this hypercall is
		 * blacklisted (i.e. an extender with a blacklisted
		 * extender number is never tested).
		 */
		if ((l->extender >= 0) && (l->extender == e))
			return (l);

		/*
		 * If the blacklist entry specifies an extender type
		 * and it doesn't match, this hypercall can't be
		 * blacklisted.
		 */
		if ((l->type != NULL) &&
		    (grub_strcmp(type, l->type) != 0))
			continue;

		/*
		 * If the blacklist entry specifies a function number
		 * and it matches, this hypercall is blacklisted.
		 */
		if ((l->function >= 0) && (l->function == f))
			return (l);
	}

	return (NULL);
}

static void
find_extenders(void)
{
	/* Allows only for 4k of extenders. */
	static void *enum_page = NULL;
	boolean_t support_new_enum = (do_hypercall(ZVMM_CNTRL_EXTID,
	    ZVCNTRL_FEATURE, ZVCNTRL_ENUM_V2) == 0);

	if (enum_page == NULL)
		enum_page = page_alloc(PGSIZE);
	ext_p = (zv_extinfo_v2_t *)enum_page;

	if (support_new_enum) {
		zv_extinfo_v2_t *ei;

		if (do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_ENUM_V2,
		    enum_page, 0) == 0)
			panic("No extenders?");
		for (ei = (zv_extinfo_v2_t *)enum_page, ext_n = 0;
		     ei->ze_name[0] != '\0';
		     ei++, ext_n++)
			;
	} else {
		zv_extinfo_v1_t *ei;
		zv_extinfo_v2_t *ein;
		void *page = page_alloc(PGSIZE);

		if (do_hypercall(ZVMM_CNTRL_EXTID,
		    ZVCNTRL_ENUM_V1, page, 0) == 0)
			panic("No extenders?");
		for (ei = (zv_extinfo_v1_t *)page, ein = ext_p, ext_n = 0;
		     ei->ze_name[0] != '\0';
		     ei++, ein++, ext_n++) {
			grub_memcpy(ein->ze_name, ei->ze_name,
			    sizeof (ein->ze_name));
			ein->ze_id = ei->ze_id;
			(void) grub_snprintf(ein->ze_dev_id,
			    ZVMM_ID_MAX_LEN, "%d", ei->ze_addr);
		}
	}

	grub_printf("Found %d extenders at 0x%p.\n", ext_n, ext_p);
}

static uint64_t
rand(void)
{
	return (do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_GET_RANDOM_BYTES));
}

static void
gnash(uint64_t ext, int iter)
{
	uint64_t args[6];

	args[0] = ext;
	/* Function numbers are typically small integers. */
	args[1] = rand() % 64;
	args[2] = rand() & 0xffffffff;
	args[3] = rand();
	args[4] = rand();
	args[5] = rand();

	grub_printf("iter %d: "
	    "hypercall(0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx)...\n", iter,
	    args[0], args[1], args[2], args[3], args[4], args[5]);

	blacklist_entry_t *be = blacklist_find(args[0], args[1], args[2],
	    args[3], args[4], args[5]);
	if (be != NULL) {
		grub_printf("iter %d: "
		    "                                                 "
		    "...blacklisted (%s)\n", iter, be->reason);
	} else {
		uint64_t rc = do_hypercall(args[0], args[1], args[2], args[3],
		    args[4], args[5]);

		grub_printf("iter %d: "
		    "                                                 "
		    "...result = 0x%lx\n", iter, rc);
	}
}

void
gnasher(void)
{
	int n;

	grub_printf("\nGnash, gnash!\n\n");

	find_extenders();

	/*
	 * Describe the extenders discovered.
	 */
	zv_extinfo_v2_t *ei;
	for (n = 0, ei = ext_p; n < ext_n; n++, ei++) {
		grub_printf("extender %d is type %s\n",
		    n, ei->ze_name);
	}

	blacklist_add("zvcntrl", -1, ZVCNTRL_REBOOT, "stops testing");
	blacklist_add("zvcntrl", -1, ZVCNTRL_HALT, "stops testing");
	blacklist_add("zvcntrl", -1, ZVCNTRL_PAUSE, "stops testing");
	blacklist_add("zvcntrl", -1, ZVCNTRL_SUSPEND, "stops testing");
	blacklist_add("zvcntrl", -1, ZVCNTRL_CRASHDUMP, "stops testing");
	blacklist_add("zvcntrl", -1, ZVCNTRL_STOP_CPU, "stops testing");

	blacklist_add(NULL, 1, -1, "not yet ready to test the console");

	/*
	 * Describe the configured blacklist.
	 */
	blacklist_entry_t *be = blacklist;
	for (be = blacklist; be != NULL; be = be->next) {
		grub_printf("blacklist: "
		    "type %s, extender %ld, function %ld, reason %s\n",
		    be->type, be->extender, be->function, be->reason);
	}

	for (n = 0; ; n++)
		gnash(rand() % ext_n, n);
}
