/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2011, 2012, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/x86_archext.h>
#include <sys/archsystm.h>
#include <sys/controlregs.h>
#include "zv_mem.h"
#include "zv_asm.h"

/*
 * This program loads into memory at 3Meg virt/phys.
 *
 * The memory from 1Meg up to 3Meg will be available for page tables.
 * Which allows us to build mappings for up to 1Gig. That should be plenty.
 */

/*
 * Standard bits used in PTE (page level) and PTP (internal levels)
 */
static x86pte_t ptp_bits = PT_VALID | PT_REF | PT_WRITABLE | PT_USER;
static x86pte_t pte_bits = PT_VALID | PT_REF | PT_WRITABLE | PT_MOD |
    PT_NOCONSIST;

uintptr_t last_phys = 0;
static uintptr_t next_ptable = MBYTE(1);

static uint_t shift_amt[] = {12, 21, 30, 39};
static uint_t ptes_per_table = 512;
static uint_t top_level = 3;

/*
 * get the index corresponding to a virt address at a given page table level.
 */
static uint_t
vatoindex(uintptr_t va, uint_t level)
{
	return ((va >> shift_amt[level]) & (ptes_per_table - 1));
}

/*
 * Return the page table entry that maps a virtual address.
 */
x86pte_t *
find_pte(uintptr_t va, uint_t level)
{
	uint_t l;
	uint_t index;
	x86pte_t *table;
	x86pte_t *new_table;
	x86pte_t pteval;

	/*
	 * Walk down the page tables creating any needed intermediate tables.
	 */
	table = (x86pte_t *)(getcr3() & PGMASK);
	for (l = top_level; l != level; --l) {

		index = vatoindex(va, l);
		pteval = table[index];

		if (pteval & PT_VALID) {
			table = (x86pte_t *)(pteval & PGMASK);
		} else {
			new_table = (x86pte_t *)next_ptable;
			next_ptable += PGSIZE;
			pteval = (x86pte_t)new_table | ptp_bits;
			table[index] = pteval;
			table = new_table;
		}
	}

	/*
	 * Return a pointer into the current pagetable.
	 */
	index = vatoindex(va, l);
	return (table + index);
}

/*
 * Add a mapping for the machine page at the given virtual address.
 */
void
map_pa_at_va(uint64_t pa, void *vaddr, uint_t level)
{
	uintptr_t va = (uintptr_t)vaddr;
	x86pte_t *pte_ptr;
	x86pte_t pteval;

	pteval = pa | pte_bits;
	if (level > 0)
		pteval |= PT_PAGESIZE;
	if (va != pa)
		pteval |= PT_GLOBAL;

	/*
	 * Find the pte that will map this address. This creates any
	 * missing intermediate level page tables
	 */
	pte_ptr = find_pte(va, level);
	*pte_ptr = pteval;
}

/*
 * Allocate guest physical from low memory towards high
 */
void *
page_alloc(size_t size)
{
	uintptr_t val;
	uintptr_t v;

	val = last_phys;
	last_phys = val + ROUND_UP(size, PGSIZE);
	for (v = val; v < last_phys; v += PGSIZE)
		map_pa_at_va(v, (void *)v, 0);
	return ((void *)val);
}
