/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2011, 2015, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>

#include "zv_asm.h"
#include "zv_mem.h"
#include "zvcons.h"

#include <sys/zvmm/zvmm_hcall.h>
#include <sys/zvmm/zvmm_ext_common.h>
#include <sys/zvmm/zvblk_common.h>

#include <grub/disk.h>
#include <grub/misc.h>
#include <grub/mm.h>

#define	ZVBLK_NAME	"zvblk"

typedef struct zvblk_dev {
	zv_extinfo_v2_t ei;
	struct zvblk_dev *next;
} zvblk_dev_t;

static zvblk_dev_t *zvblk_dev = NULL;
static zvblk_dev_t *zvblk_last = NULL;

static zvblk_dev_t *
find_dev(int addr)
{
	zvblk_dev_t *d;
	zv_idstr_t idstr;

	(void) grub_snprintf(idstr, ZVMM_ID_MAX_LEN, "%d", addr);
	for (d = zvblk_dev; d != NULL; d = d->next)
		if (grub_strcmp(d->ei.ze_dev_id, idstr) == 0)
			return (d);

	return (NULL);
}

uint64_t
disk_get_info(int drive, uint64_t flag)
{
	static zvblk_info_t *zi = NULL;

	if (zi == NULL)
		zi = page_alloc(PGSIZE);

	if (do_hypercall(drive, ZVBLK_GET_INFO_V1, flag, zi) != 0)
		return (0);

	if (flag == ZVBLK_DEV_SIZE)
		return (zi->zdi_size);

	/* Grub should only care about logical block size. */
	if (flag == ZVBLK_DEV_BSIZE)
		return (zi->zdi_lblksize);

	if (flag == ZVBLK_DEV_TYPE)
		return (zi->zdi_removable);

	return (0);
}

uint64_t
zvblk_disk_get_info(const char *name, uint64_t flag)
{
	const unsigned int l = grub_strlen(ZVBLK_NAME);

	if (grub_strncmp(name, ZVBLK_NAME, l) != 0) {
		grub_dprintf("zvblk", "%s: you're not one of us!\n",
			name);
		return (-1);
	}

	unsigned int addr = grub_strtoul(name + l, 0, 0);
	zvblk_dev_t *d = find_dev(addr);
	if (d == NULL)
		return (-1);

	return (disk_get_info(d->ei.ze_id, flag));
}

static int
zvblk_iterate(int (*hook)(const char *name), grub_disk_pull_t pull)
{
	zvblk_dev_t *d;

	if (pull != GRUB_DISK_PULL_NONE)
		return (0);

	for (d = zvblk_dev; d != NULL; d = d->next) {
		char devname[40];

		grub_snprintf(devname, sizeof (devname),
		    "%s%s", ZVBLK_NAME, d->ei.ze_dev_id);
		if (hook(devname))
			return (1);
	}
	return (0);
}

static grub_err_t
zvblk_open(const char *name, grub_disk_t disk)
{
	const unsigned int l = grub_strlen(ZVBLK_NAME);
	uint64_t bsize;

	grub_dprintf("zvblk", "attempt to open %s:\n", name);

	if (grub_strncmp(name, ZVBLK_NAME, l) != 0) {
		grub_dprintf("zvblk", "%s: you're not one of us!\n",
			name);
		return (grub_error(GRUB_ERR_UNKNOWN_DEVICE,
			"not a zvblk device"));
	}

	unsigned int addr = grub_strtoul(name + l, 0, 0);
	zvblk_dev_t *d = find_dev(addr);
	if (d == NULL) {
		grub_dprintf("zvblk", "open of %s fails.\n", name);
		return (grub_error(GRUB_ERR_UNKNOWN_DEVICE,
			"unknown extender"));
	}

	bsize = disk_get_info(d->ei.ze_id, ZVBLK_DEV_BSIZE);
	disk->total_sectors =
	    disk_get_info(d->ei.ze_id, ZVBLK_DEV_SIZE) / bsize;
	for (disk->log_sector_size = 0;
	    (1U << disk->log_sector_size) < bsize;
	    disk->log_sector_size++)
		;

	disk->id = addr;
	disk->data = d;

	grub_errno = GRUB_ERR_NONE;
	grub_dprintf("zvblk", "open of %s succeeds.\n", name);

	return (GRUB_ERR_NONE);
}

static void
zvblk_close(grub_disk_t disk)
{
	/* Nothing. */
}

static grub_err_t
zvblk_read(grub_disk_t disk, grub_disk_addr_t sector,
    grub_size_t size, char *buf)
{
	static zvblk_first_desc_t *iodesc = NULL;
	zvblk_dev_t *d = disk->data;
	zvblk_vec_desc_t *vec;
	uint64_t *pa;
	uint64_t seg;
	uint64_t max_segs;
	uint64_t rc;
	uint64_t real_size = size << disk->log_sector_size;

	grub_dprintf(ZVBLK_NAME, "zvblk_read: "
	    "disk: %s, sector 0x%lx, size 0x%lx, real 0x%lx, buf %p\n",
	    disk->name, sector, size, real_size, buf);

	/*
	 * Page-aligned iodesc.  Both the iodesc and disk_buffer are guest
	 * physical addresses, as required by the host side driver. In
	 * zvboot, VA == PA, so no conversion is required.
	 */
	if (iodesc == NULL)
		iodesc = page_alloc(PGSIZE);
	iodesc->zid_hdr.zih_ver = 0;
	iodesc->zid_hdr.zih_desc_type = ZVBLK_IODESC_TYPE_FIRST;
	iodesc->zid_hdr.zih_next_desc_gpa = 0;
	iodesc->zid_id = 0;
	iodesc->zid_op = ZVBLK_OP_READ;
	iodesc->zid_op |= ZVBLK_OP_FLAG_SYNC;
	/* We don't need more than one descriptor page for max 1MB I/O size. */
	iodesc->zid_desc_num = 1;
	iodesc->zid_block = sector;
	iodesc->zid_size = real_size;
	vec = (zvblk_vec_desc_t *)iodesc->zid_data;
	/* We only need one vector descriptor. */
	pa = iodesc->zid_data + 1;
	iodesc->zid_vec_num = 1;

	max_segs = (PGSIZE - sizeof (zvblk_first_desc_t)) / sizeof (*pa);

	for (seg = 0; real_size > 0; seg++) {
		size_t avail = PGSIZE - ((uint64_t)buf & PGOFFSET);
		size_t this_len = (real_size > avail) ? avail : real_size;

		if (seg > max_segs)
			panic("Too many segments");

		pa[seg] = (uint64_t)buf;

		buf += this_len;
		real_size -= this_len;
	}

	iodesc->zid_page_num = seg;
	vec->zvd_desc_index = 0;
	vec->zvd_desc_offset = (uintptr_t)pa & PGOFFSET;
	vec->zvd_size = iodesc->zid_size;

	/* start the disk blk read then wait for it to complete */
	rc = do_hypercall(d->ei.ze_id, ZVBLK_IO, iodesc);
	if (rc != ZVBLK_IO_DONE) {
		return (grub_error(GRUB_ERR_READ_ERROR,
			"IO operation failed"));
	}
	if ((iodesc->zid_result & ZVBLK_RES_OKAY) == 0) {
		return (grub_error(GRUB_ERR_READ_ERROR,
			"failed to read block"));
	}

	return (GRUB_ERR_NONE);
}

static grub_err_t
zvblk_write(grub_disk_t disk, grub_disk_addr_t sector,
    grub_size_t size, const char *buf)
{
	return (GRUB_ERR_NOT_IMPLEMENTED_YET);
}

static struct grub_disk_dev grub_zvblk_dev = {
	.name = ZVBLK_NAME,
	/*
	 * There is no GRUB_DISK_DEVICE_*_ID identifier assigned for
	 * zvblk, so we re-use that for ATA devices. Given that we
	 * will never see anything other than zvblk disk devices, this
	 * is safe.
	 */
	.id = GRUB_DISK_DEVICE_ATA_ID,
	.iterate = zvblk_iterate,
	.open = zvblk_open,
	.close = zvblk_close,
	.read = zvblk_read,
	.write = zvblk_write,
	.next = NULL
};

static void
zvblk_add(zv_extinfo_v2_t *eip)
{
	zvblk_dev_t *new;

	grub_dprintf("zvblk", "add zvblk extender addr %s.\n", eip->ze_dev_id);

	new = grub_malloc(sizeof (*new));
	if (new == NULL)
		panic("out of memory");

	new->ei = *eip;
	new->next = NULL;
	if (zvblk_last != NULL)
		zvblk_last->next = new;
	else
		zvblk_dev = new;
	zvblk_last = new;
}

void
zvblk_init(void)
{
	boolean_t support_new_enum = (do_hypercall(ZVMM_CNTRL_EXTID,
	    ZVCNTRL_FEATURE, ZVCNTRL_ENUM_V2) == 0);

	/* Declare the zvblk disk type. */
	grub_disk_dev_register(&grub_zvblk_dev);

	/* Enumerate the extenders, adding disk devices. */
	void *enum_page = page_alloc(PGSIZE);
	uint64_t rc;

	rc = 0;
	for (;;) {
		if (support_new_enum) {
			zv_extinfo_v2_t *eip;

			rc = do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_ENUM_V2,
			    enum_page, rc);
			if (rc == 0)
				break;
			for (eip = enum_page; eip->ze_name[0] != '\0';
			    eip++) {
				if (grub_strcmp(eip->ze_name, "zvblk") == 0)
					zvblk_add(eip);
			}
		} else {
			zv_extinfo_v1_t *ei;
			zv_extinfo_v2_t ein;

			rc = do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_ENUM_V1,
			    enum_page, rc);
			if (rc == 0)
				break;
			for (ei = enum_page; ei->ze_name[0] != '\0'; ei++) {
				if (grub_strcmp(ei->ze_name, "zvblk") == 0) {
					grub_memcpy(ein.ze_name, ei->ze_name,
					    sizeof (ein.ze_name));
					ein.ze_id = ei->ze_id;
					(void) grub_snprintf(ein.ze_dev_id,
					    ZVMM_ID_MAX_LEN, "%d", ei->ze_addr);
					zvblk_add(&ein);
				}
			}
		}
	}
}
