/*
 *  GRUB  --  GRand Unified Bootloader
 * Copyright (c) 2011, 2016, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/varargs.h>

#include <grub/device.h>
#include <grub/disk.h>
#include <grub/file.h>
#include <grub/misc.h>
#include <grub/mm.h>
/* Avoid clashing definitions with kernel headers. */
#define	GRUB_DL_H 1
#include <grub/partition.h>
#include <grub/zfs/zfs.h>

#include "zvload.h"
#include "zvcons.h"
#include "zv_mem.h"
#include "zv_asm.h"
#include "gnasher.h"

#include <sys/zvmm/zvmm_hcall.h>
#include <sys/zvmm/zvmm_ext_common.h>
#include <sys/zvmm/zvmm_mach_hcall.h>
#include <sys/zvmm/zvblk_common.h>

#include <sys/zone.h>

#define	KEY_CTRL_F	(0x06)
#define	KEY_CTRL_B	(0x02)
#define	KEY_ESC		(0x1b)
#define	KEY_ENTER	(0x0d)
#define	KEY_DEL		(0x7f)
#define	KEY_BACK	(0x08)

#define	LINELEN		(BOOTARGS_MAX)

#define	STR(s) (((s) != NULL ? (s) : ""))

/* The grub2 name for our boot device. */
static char *boot_device_name = NULL;
/* The type of the filesystem on the boot device. */
static char *boot_fstype = NULL;

static char *config_file_path = "/boot/menu.conf";

/* Default boot device. */
#define	DEFAULT_BOOT_DEVICE	"disk0"

/* grub2 device name for a ramdisk. */
#define	GRUB_RAMDISK_DEVICE	"memdisk"
/* grub2 device prefix for a zvmm block device. */
#define	GRUB_ZVBLK_DEVICE	"zvblk"

extern grub_device_t memdisk_set(char *, grub_off_t);

static char *config;

typedef struct config_entry {
	char *ce_title;
	char *ce_rootdev;
	char *ce_bootfs;
	char *ce_bootargs;
	char *ce_kernel;
	char *ce_module;
	int ce_default;
	int ce_index;
} config_entry_t;

typedef struct boot_args {
	char *ba_dev;
	char *ba_be;
	char *ba_kern;
	char *ba_kernargs;
	char *ba_debugargs;
	boolean_t ba_list;
	boolean_t ba_gnasher;
} boot_args_t;

static config_entry_t chosen_entry;

static grub_file_t
file_open(const char *dev, const char *dataset, const char *path)
{
	char fullpath[2 * LINELEN];

	fullpath[0] = '\0';

	/*
	 * Generate a grub2 style pathname, including the device
	 * specifier.
	 *
	 * We assume that:
	 * - a `path' that includes the @ character already has the
         *   ZFS filesystem name prepended to it,
	 * - `dataset' includes a ZFS pool name (as well as filesystem)
         *   if it doesn't being with a slash (/),
	 * - only the boot device might be ZFS.
	 *
	 * XXX: This is a complicated mess, as it tries to deal with
	 * ZFS and non-ZFS, the presence/absence of a dataset, etc.
	 * It, and the callers, should be cleaned up.
	 */

	grub_strcat(fullpath, "(");
	grub_strcat(fullpath, dev);
	grub_strcat(fullpath, ")");

	if ((grub_strcmp(dev, boot_device_name) == 0) &&
	    (grub_strcmp(boot_fstype, "zfs") == 0) &&
	    (dataset != NULL) &&
	    (dataset[0] != '\0') &&
	    (grub_strchr(path, '@') == NULL)) {
		char *slash;

		/* Remove the dataset name, if present. */
		if ((dataset[0] != '/') &&
		    ((slash = grub_strchr(dataset, '/')) != NULL))
			grub_strcat(fullpath, slash);
		else
			grub_strcat(fullpath, dataset);
		grub_strcat(fullpath, "/@/");
	}
	grub_strcat(fullpath, path);

	grub_dprintf("zvboot", "file_open(<%s>, %s, %s) -> %s\n",
	    dev, dataset, path, fullpath);

	return (grub_file_open(fullpath));
}

static grub_ssize_t
file_read(grub_file_t file, void *buf, grub_size_t len)
{
	char *p = buf;
	grub_size_t remain = len;
	grub_ssize_t r = 0;

	/* Read in 1MB chunks to help out zvblk. */
	while (remain > 0) {
		grub_size_t this = remain;
		grub_size_t good;

		if (this > MBYTE(1))
			this = MBYTE(1);

		if ((good = grub_file_read(file, p, this)) != this) {
			r += good;
			return (r);
		}

		remain -= this;
		r += good;
		p += this;
	}

	return (r);
}

static void
delete_char(char *buf, size_t pos)
{
	char *cp;

	if (buf[pos] == '\0')
		return;

	cp = &buf[pos + 1];

	do {
		*(cp - 1) = *cp;
		cp++;
	} while (*(cp - 1) != '\0');
}

static void
insert_char(char *buf, size_t pos, char ch)
{
	char *cp = buf + grub_strlen(buf);

	*(cp + 1) = '\0';

	while (cp != &buf[pos]) {
		*cp = *(cp - 1);
		cp--;
	}

	buf[pos] = ch;
}

static char *
eat_white(char *cp)
{
	while (grub_isspace(*cp))
		++cp;
	return (cp);
}

static char *
eat_word(char *cp, char *buf, size_t size)
{
	if (buf != NULL)
		*buf = '\0';

	while (!grub_isspace(*cp) && *cp != '\n' && *cp != '\0') {
		if (buf != NULL && size > 1) {
			*buf = *cp;
			size--;
			buf++;
		}

		cp++;
	}

	if (buf != NULL)
		*buf = '\0';

	return (cp);
}

static int
get_num(const char *cp, int *valid)
{
	int x = 0;
	int sign = 1;

	if (*cp == '-') {
		sign = -1;
		++cp;
	}

	while ('0' <= *cp && *cp <= '9') {
		if (valid != NULL)
			*valid = 1;
		x = x * 10 + *cp - '0';
		++cp;
	}

	if (*cp != '\0' && *cp != '\t' && *cp != '\n') {
		if (valid != NULL)
			*valid = 0;
	}

	return (x * sign);
}

static void
clear_line(void)
{
	zcons_putchar(0x1b);
	zcons_putchar('[');
	zcons_putchar('2');
	zcons_putchar('K');
	zcons_putchar('\r');
}

/*
 * primitive interactive line editing
 */
static void
getline(char *buf, size_t size)
{
	size_t pos = 0;

	buf[0] = '\0';

	for (;;) {
		size_t len;
		char ch;

		len = grub_strlen(buf);

		clear_line();
		grub_printf("%s", buf);

		switch ((ch = zcons_getchar())) {
		case KEY_CTRL_F:
			if (pos < len)
				pos++;
			break;

		case KEY_CTRL_B:
			if (pos > 0)
				pos--;
			break;

		case KEY_ENTER:
			zcons_putchar('\n');
			return;

		/*
		 * Rather than attempt to support delete, we're better off
		 * treating them the same due to the world's insanity over
		 * backspace/delete.
		 */
		case KEY_DEL:
		case KEY_BACK:
			if (pos == 0)
				break;

			delete_char(buf, --pos);
			break;

		default:
			if (len + 1 == size)
				break;

			insert_char(buf, pos++, ch);
			break;
		}
	}
}

static char *
get_poolname(const char *device)
{
	grub_device_t dev;
	dev = grub_device_open(device);
	if (dev == NULL)
		return (NULL);

	grub_err_t err;
	char *nvlist;
	err = grub_zfs_fetch_nvlist(dev, &nvlist);
	grub_device_close(dev);
	if (err != 0)
		return (NULL);

	char *poolname;
	poolname = grub_zfs_nvlist_lookup_string(nvlist,
	    ZPOOL_CONFIG_POOL_NAME);
	if (poolname == NULL) {
		grub_free(nvlist);
		return (NULL);
	}
	grub_free(nvlist);

	return (poolname);
}

static char *
get_default_bootfs(const char *device)
{
	grub_device_t dev;
	dev = grub_device_open(device);
	if (dev == NULL) {
		/* Any failure should not carry forward. */
		grub_errno = GRUB_ERR_NONE;
		return (NULL);
	}

	/* Find the default boot filesystem. */
	char *default_bootfs;
	grub_err_t err;
	err = grub_zfs_defaultbootfsname(dev, &default_bootfs);
	if (err != 0) {
		grub_device_close(dev);

		/*
		 * Any error should not carry forward - this function
		 * can be called with a device which has a non-ZFS
		 * filesystem.
		 */
		grub_errno = GRUB_ERR_NONE;
		return (NULL);
	}

	grub_device_close(dev);

	return (default_bootfs);
}

static char *
get_default_dataset(const char *device)
{
	/* Find the pool name. */
	char *poolname = get_poolname(device);
	if (poolname == NULL) {
		/*
		 * Any error should not carry forward - this function
		 * can be called with a device which has a non-ZFS
		 * filesystem.
		 */
		grub_errno = GRUB_ERR_NONE;
		return (NULL);
	}

	/* Find the default boot filesystem. */
	char *default_bootfs = get_default_bootfs(device);;
	if (default_bootfs == NULL)
		return (NULL);

	/* Combine the two. */
	grub_size_t sz = grub_strlen(poolname) + grub_strlen(default_bootfs)
	    + 2; /* / and \0. */
	char *combined = grub_malloc(sz);
	if (combined == NULL) {
		grub_dprintf("zvboot", "Could not allocate %ld bytes for "
		    "default bootfs name.\n", sz);

		grub_free(default_bootfs);
		grub_free(poolname);
		return (NULL);
	}

	grub_snprintf(combined, sz, "%s/%s", poolname, default_bootfs);

	grub_free(default_bootfs);
	grub_free(poolname);

	grub_dprintf("zvboot", "default bootfs for %s is %s.\n",
	    device, combined);

	return (combined);
}

/*
 * We have various things to fix up before we can use the chosen config entry.
 */
static void
complete_entry(config_entry_t *entry, char *default_device, boot_args_t *ba)
{
#if defined(amd64)
#define	DEFAULT_KERNEL	"/platform/i86pc/kernel/amd64/unix"
#define	DEFAULT_ARCHIVE	"/platform/i86pc/amd64/boot_archive"
#else
#error "Unknown platform"
#endif

	if (entry->ce_title == NULL)
		entry->ce_title = "default";
	/* If no root device is specified, use the default. */
	if (entry->ce_rootdev == NULL)
		entry->ce_rootdev = default_device;
	/* If no bootfs is specified, used the default for the pool. */
	if (entry->ce_bootfs == NULL)
		entry->ce_bootfs = get_default_dataset(entry->ce_rootdev);

	if (ba != NULL && ba->ba_kern != NULL)
		entry->ce_kernel = ba->ba_kern;
	else if (entry->ce_kernel == NULL)
		entry->ce_kernel = DEFAULT_KERNEL;

	if (ba != NULL && ba->ba_kernargs != NULL)
		entry->ce_bootargs = ba->ba_kernargs;

	if (entry->ce_module == NULL)
		entry->ce_module = DEFAULT_ARCHIVE;
}

static void
print_entry(const char *title, config_entry_t *c)
{
	grub_dprintf("zvboot", "%s:\n", title);
	grub_dprintf("zvboot", "   title: %s\n", c->ce_title);
	grub_dprintf("zvboot", " rootdev: %s\n", c->ce_rootdev);
	grub_dprintf("zvboot", "  bootfs: %s\n", c->ce_bootfs);
	grub_dprintf("zvboot", "bootargs: %s\n", c->ce_bootargs);
	grub_dprintf("zvboot", "  kernel: %s\n", c->ce_kernel);
	grub_dprintf("zvboot", "  module: %s\n", c->ce_module);
	grub_dprintf("zvboot", " default: %d\n", c->ce_default);
	grub_dprintf("zvboot", "   index: %d\n", c->ce_index);
}

static void
print_ba(const char *title, boot_args_t *ba)
{
	grub_dprintf("zvboot", "%s:\n", title);
	grub_dprintf("zvboot", "      ba_dev: %s\n", STR(ba->ba_dev));
	grub_dprintf("zvboot", "       ba_be: %s\n", STR(ba->ba_be));
	grub_dprintf("zvboot", "     ba_kern: %s\n", STR(ba->ba_kern));
	grub_dprintf("zvboot", " ba_kernargs: %s\n", STR(ba->ba_kernargs));
	grub_dprintf("zvboot", "     ba_list: %d\n", ba->ba_list);
	grub_dprintf("zvboot", "ba_debugargs: %s\n", STR(ba->ba_debugargs));
	grub_dprintf("zvboot", "  ba_gnasher: %d\n", ba->ba_gnasher);
}

static char *
trim_space(char *s)
{
	char *new = s;

	/* Trim the beginning. */
	while ((new != '\0') && (*new == ' '))
		new++;

	/* Trim the end. */
	char *e = new + grub_strlen(new) - 1;
	while ((e > new) && (*e == ' ')) {
		*e = '\0';
		e--;
	}

	return (new);
}

static void
parse_kv(char *s, char **k, char **v)
{
	*k = *v = NULL;

	char *eq = grub_strchr(s, '=');
	if (eq == NULL)
		return;

	*eq = '\0';
	*k = trim_space(s);
	*v = trim_space(eq + 1);
}

/*
 * Walk a pybootmgmt menu.conf file.
 */
static void
walk_config(int (*cb)(config_entry_t *, void *), void *arg)
{
	config_entry_t entry;
	grub_memset(&entry, 0, sizeof (entry));

	/*
	 * This routine is destructive on the `config' buffer, so work
	 * with a copy.
	 */
	char *copy = grub_strdup(config);
	boolean_t valid_entry = B_FALSE;
	char *p = copy;
	int index = 0;
	while ((p != NULL) && (*p != '\0')) {
		char *next = grub_strchr(p, '\n');
		if (next != NULL)
			*next = '\0';

		switch (p[0]) {
		case '\0':
			/* Blank. */
			break;

		case '#':
			/* Comment. */
			break;

		case '[':
			/* Start of section. */
			p[grub_strlen(p) - 1] = '\0';

			/* Callback for the previous entry. */
			if (valid_entry) {
				print_entry("Found BE", &entry);

				if (cb(&entry, arg)) {
					grub_free(copy);
					return;
				}
			}

			/* We ignore "meta" and "global" sections. */
			valid_entry = !((grub_strcmp(p + 1, "meta") == 0)
			    || (grub_strcmp(p + 1, "global") == 0));

			/* Reset for the next entry. */
			grub_free(entry.ce_title);
			grub_free(entry.ce_rootdev);
			grub_free(entry.ce_bootfs);
			grub_free(entry.ce_bootargs);
			grub_free(entry.ce_kernel);
			grub_free(entry.ce_module);
			grub_memset(&entry, 0, sizeof (entry));

			if (valid_entry)
				entry.ce_index = index++;
			break;

		default:;
			char *k, *v;
			parse_kv(p, &k, &v);

#define	SET_FIELD(key, field) \
	if (grub_strcmp(k, key) == 0) entry.field = grub_strdup(v);

			SET_FIELD("title", ce_title);
			SET_FIELD("bootfs", ce_bootfs);
			SET_FIELD("kargs", ce_bootargs);
			SET_FIELD("kernel", ce_kernel);
			SET_FIELD("boot_archive", ce_module);
#undef SET_FIELD

			if ((grub_strcmp(k, "default") == 0) &&
			    (grub_strcmp(v, "True") == 0)) {
				entry.ce_default = 1;
			}
		}

		if (next != NULL)
			p = next + 1;
	}

	/* Callback for the final entry. */
	if (valid_entry) {
		print_entry("Found BE", &entry);

		(void) cb(&entry, arg);
	}

	grub_free(copy);
}

static int search_address;
static int search_removable;
static char *found_device;

static int
select_boot_hook(const char *name)
{
	/* Only zvblk devices can match. */
	if (grub_strncmp(name, GRUB_ZVBLK_DEVICE,
		grub_strlen(GRUB_ZVBLK_DEVICE)) != 0) {
		grub_dprintf("zvboot",
		    "%s is not a %s device -> not acceptable.\n",
		    name, GRUB_ZVBLK_DEVICE);
		return (0);
	}

	/* If specified, the unit number must match. */
	if (search_address >= 0) {
		char proposal[16];
		int plen;
		char *comma;

		plen = grub_snprintf(proposal, sizeof (proposal),
		    "%s%d", GRUB_ZVBLK_DEVICE, search_address);
		/*
		 * The name can be in a form zvblkN or zvblkN,gptM. In the
		 * second case we want to match just the first part.
		 *
		 * If we find a comma in the provided name, we calculate
		 * the length of the first part, otherwise we just use strlen()
		 * for comparison against the length of our proposed device
		 * string. This is necessary to prevent matching e.g zvblk1
		 * with zvblk100.
		 */
		comma = grub_strchr(name, ',');
		if ((plen != (comma ? (comma - name) : grub_strlen(name))) ||
		    (grub_strncmp(name, proposal, plen) != 0)) {
			grub_dprintf("zvboot",
			    "wrong unit number (%s != %s) -> not acceptable.\n",
			    name, proposal);
			return (0);
		}
	}

	/* From zvblk.c. */
	extern uint64_t zvblk_disk_get_info(const char *, uint64_t);
	uint64_t removable = zvblk_disk_get_info(name, ZVBLK_DEV_TYPE);
	if (search_removable != removable) {
		grub_dprintf("zvboot", "%s is %sremovable -> not acceptable.\n",
		    name, removable ? "" : "not ");
		return (0);
	}

	/* The device details match. */

	/* If no filesystem, no point in trying to use this device. */
	grub_device_t dev = grub_device_open(name);
	if (dev == NULL) {
		grub_dprintf("zvboot", "cannot open %s -> not acceptable.\n", name);
		/* Error should not carry forward */
		grub_errno = GRUB_ERR_NONE;
		return (0);
	}
	grub_fs_t fs = grub_fs_probe(dev);
	grub_device_close(dev);
	if (fs == NULL) {
		grub_dprintf("zvboot",
		    "cannot determine filesystem of %s -> not acceptable.\n",
		    name);
		/* Error should not carry forward */
		grub_errno = GRUB_ERR_NONE;
		return (0);
	}

	/* Remember the filesystem type of the chosen boot device. */
	boot_fstype = grub_strdup(fs->name);

	found_device = grub_strdup(name);

	grub_dprintf("zvboot", "boot device: %s, fs: %s.\n",
	    found_device, boot_fstype);

	return (1);
}

static char *
select_boot_device(const char *boot_dev)
{
	int removable;
	int address;
	int valid = 0;
	char *space;
	const char *dev = boot_dev;

	/*
	 * Matching rules:
	 * - "cdrom" means a CD with any address,
	 * - "cdromN" means a CD with address N,
	 * - "disk" means disk0,
	 * - "diskN" means a disk with address N.
	 */

	/*
	 * Iterate the ba_dev list (space separated) looking for a bootable
	 * device.
	 */
	for (;;) {
		space = grub_strchr(dev, ' ');
		if (space != NULL)
			*space = '\0';

		if (grub_strncmp(dev, "cdrom", grub_strlen("cdrom")) == 0) {
			removable = 1;
			address = get_num(dev + grub_strlen("cdrom"),
			    &valid);
			if (!valid)
				address = -1;
		} else if (grub_strncmp(dev, "disk",
		    grub_strlen("disk")) == 0) {
			removable = 0;
			address = get_num(dev + grub_strlen("disk"), &valid);
			if (!valid)
				address = 0;
		}

		if (space != NULL)
			*space = ' ';

		search_address = address;
		search_removable = removable;
		found_device = NULL;

		grub_dprintf("zvboot", "Looking for boot device "
		    "- address %d, removable %d.\n",
		    search_address, search_removable);

		grub_device_iterate(select_boot_hook);
		if (found_device != NULL || space == NULL)
			break;

		for (dev = space; *dev == ' '; dev++)
			;
		if (*dev == '\0')
			break;
	}

	if (found_device == NULL)
		panic("Cannot determine boot device from '%s'.\n", boot_dev);

	return (found_device);
}

static char *
get_boot_device(boot_args_t *ba)
{
	uint64_t bootdevlen = 1024;
	char *bootdev;
	int err = 0;

	if (ba->ba_dev != NULL)
		return (grub_strdup(ba->ba_dev));

	/* If no boot device specified, get the boot device based on bootpri. */

	bootdev = grub_malloc(bootdevlen);

	if (bootdev == NULL || (err = do_hypercall(ZVMM_CNTRL_EXTID,
	    ZVCNTRL_GET_ZONEATTR, ZONE_ATTR_BOOT_DEVICE,
	    bootdev, &bootdevlen)) != 0 || bootdevlen < 2) {
		grub_free(bootdev);
		bootdev = grub_strdup(DEFAULT_BOOT_DEVICE);
		grub_dprintf("zvboot", "Using default boot device %s "
		    "bootdev: err %d\n", DEFAULT_BOOT_DEVICE, err);
	} else {
		grub_dprintf("zvboot", "Boot device based on bootpri:\n");
		grub_dprintf("zvboot", "     bootdev = %s\n", bootdev);
	}

	return (bootdev);
}

/*
 * Options -L and -Z <dataset> are not checked as they are not considered as
 * real boot arguments. They are rather meta arguments allowing a selection of
 * desired boot entry.
 */
static boolean_t
ba_is_empty(boot_args_t *ba)
{
	if (ba->ba_dev != NULL || ba->ba_kern != NULL ||
	    ba->ba_kernargs != NULL || ba->ba_debugargs != NULL ||
	    ba->ba_gnasher)
		return (B_FALSE);

	return (B_TRUE);
}

static void
parse_bootargs(char *buf, boot_args_t *ba)
{
	char debugstr[BOOTARGS_MAX] = { 0 };
	char *cp = buf;

	grub_memset(ba, 0, sizeof (boot_args_t));

	if (cp == NULL)
		return;

	cp = eat_white(cp);

	/*
	 * Skip kz= arguments. These are for kzhost only. See kzhost's argument
	 * parsing.
	 */
	if (grub_strncmp(cp, "kz=", grub_strlen("kz=")) == 0) {
		cp = eat_word(cp, NULL, 0);
		cp = eat_white(cp);
	}
	
	/*
	 * Handle various invocations:
	 *
	 * boot -L
	 * boot -Z rpool/ROOT/be
	 * boot cdrom
	 * boot disk0 unix
	 * boot unix
	 *
	 * All potentially with trailing kernel bootargs. In summary:
	 *
	 * boot [cdrom|cdrom[0-9]|disk|disk[0-9]] \
	 *      [-L|-Z rpool/ROOT/be] [unix] [args]...
	 */

	/*
	 * Including "zvboot_debug" anywhere in the boot string
	 * enables zvboot debugging output. Typically this is done
	 * using "zoneadm -z colony boot -- -B zvboot_debug=true".
	 */
	if (grub_strstr(cp, "zvboot_debug") != NULL) {
		(void) grub_strncat(debugstr, " zvboot",
		    sizeof (debugstr) - grub_strlen(debugstr) - 1);
	}
	/*
	 * Including "zvblk_debug" anywhere in the boot string
	 * enables zvblk debugging output. Typically this is done
	 * using "zoneadm -z colony boot -- -B zvblk_debug=true".
	 */
	if (grub_strstr(cp, "zvblk_debug") != NULL) {
		(void) grub_strncat(debugstr, " zvblk",
		    sizeof (debugstr) - grub_strlen(debugstr) - 1);
	}

	/*
	 * Including "zvboot_gnasher" anywhere in the boot string
	 * enables the gnasher hypercall test tool. Typically this is
	 * done using "zoneadm -z colony boot -- -B
	 * zvboot_gnasher=true".
	 */
	if (grub_strstr(cp, "zvboot_gnasher") != NULL)
		ba->ba_gnasher = B_TRUE;

	do {
		char word[LINELEN];
		char *ocp = cp;

		cp = eat_word(cp, word, sizeof (word));
		cp = eat_white(cp);

		if (word[0] == '\0')
			break;

		if (word[0] == '-') {
			switch (word[1]) {
			case 'L':
				if (ba->ba_be != NULL) {
					grub_printf("zvboot: "
					    "both -Z and -L specified "
					    "- assuming -L.\n");
					ba->ba_be = NULL;
				}
				ba->ba_list = B_TRUE;
				break;

			case 'Z':
				cp = eat_word(cp, word, sizeof (word));
				cp = eat_white(cp);
				ba->ba_be = grub_strdup(word);
				if (ba->ba_list) {
					grub_printf("zvboot: "
					    "both -Z and -L specified "
					    "- assuming -L.\n");
					ba->ba_be = NULL;
				}
				break;

			default:
				/*
				 * Anything other than -L or -Z must
				 * be a kernel argument. That means
				 * that we're done - the rest are all
				 * kernel arguments.
				 */
				while ((cp != NULL) && (cp[0] != '\0')) {
					word[grub_strlen(word)] = ' ';
					cp = eat_word(cp, word, sizeof (word));
					cp = eat_white(cp);
				}
				cp = NULL;
				ba->ba_kernargs = grub_strdup(ocp);
				break;
			}
		} else if ((ba->ba_dev == NULL) &&
		    ((grub_strncmp(word, "cdrom",
			    grub_strlen("cdrom")) == 0) ||
			(grub_strncmp(word, "disk",
			    grub_strlen("disk")) == 0))) {
			/* This is a device name. */
			ba->ba_dev = grub_strdup(word);
		} else if (ba->ba_kern == NULL) {
			/* This is the name of a kernel. */
			ba->ba_kern = grub_strdup(word);
		} else {
			/*
			 * Anything else is a standard kernel
			 * argument.
			 */
			while ((cp != NULL) && (cp[0] != '\0')) {
				word[grub_strlen(word)] = ' ';
				cp = eat_word(cp, word, sizeof (word));
				cp = eat_white(cp);
			}
			cp = NULL;
			ba->ba_kernargs = grub_strdup(ocp);
		}
	} while (cp != NULL);

	if (debugstr[0] != '\0')
		ba->ba_debugargs = grub_strdup(debugstr);
}

static void
process_debug_bootargs(boot_args_t *ba)
{
	extern grub_err_t grub_env_set(const char *, const char *);

	if (ba->ba_debugargs != NULL)
		(void) grub_env_set("debug", ba->ba_debugargs);
	
	if (ba->ba_gnasher)
		gnasher();
}

static int
get_host_bootargs_attr(int type, boot_args_t *ba)
{
	char buf[BOOTARGS_MAX];
	size_t buf_sz = sizeof (buf);
	int rc;

	rc = do_hypercall(ZVMM_CNTRL_EXTID, ZVCNTRL_GET_ZONEATTR, type,
	    buf, &buf_sz);
	if (rc != 0)
		return (rc);
	
	parse_bootargs(buf, ba);

	return (0);
}

/*
 * Retrieve transient and default host boot arguments. Default ones come from
 * the zonecfg(8) "bootargs" property, transient were passed by a user/client
 * to zoneadmd(8).
 */
static void
get_host_bootargs(boot_args_t *ba, boot_args_t *ba_def)
{
	int rc;

	if ((rc = get_host_bootargs_attr(ZONE_ATTR_BOOTARGS_TRAN, ba)) != 0 ||
	    (rc = get_host_bootargs_attr(ZONE_ATTR_BOOTARGS_CFG, ba_def)) != 0)
	{
		/*
		 * If the host does not know requested zone bootargs attributes
		 * zvcntrl_get_zoneattr() returns EPERM. Do not panic as the
		 * host may just be old.
		 */
		if (rc != EPERM) {
			panic("Failed to read default/transient host bootargs "
			    "with error %d", rc);
		}
		/*
		 * The host is old and does not provide transient/default host
		 * bootargs zone attributes. Fail back to the legacy behavior.
		 * Retrieve zone boot arguments and mark them as transient, no
		 * matter where they came from - this is for a compatibility
		 * reason.
		 */
		if ((rc = get_host_bootargs_attr(ZONE_ATTR_BOOTARGS, ba)) != 0)
			panic("Failed to read host bootargs with error %d", rc);
	}

	if (!ba_is_empty(ba) || ba->ba_list || ba->ba_be != NULL) {
		process_debug_bootargs(ba);
		print_ba("Transient host boot arguments", ba);
	}
}

static int
show_be(config_entry_t *entry, void *arg)
{
	size_t *nep = arg;

	grub_printf("%ld %s", *nep + 1, entry->ce_title);
	if (entry->ce_bootfs != NULL)
		grub_printf(" (%s)", entry->ce_bootfs);
	grub_printf("\n");
	*nep += 1;
	return (0);
}

static int
match_be_index(config_entry_t *entry, void *arg)
{
	int index = *((int *)arg);

	if (entry->ce_index == index) {
		grub_memcpy(&chosen_entry, entry, sizeof (chosen_entry));
		return (1);
	}

	return (0);
}

static void
list_bes(void)
{
	int num;

	for (;;) {
		size_t nr_entries = 0;
		int valid = 0;
		char buf[10];

		walk_config(show_be, &nr_entries);

		if (nr_entries == 0)
			panic("No entries to list.");

		grub_printf("Select environment to boot: [ 1 - %ld ]:\n",
		    nr_entries);

		getline(buf, sizeof (buf));

		num = get_num(buf, &valid);

		if (valid && num >= 1 && num <= nr_entries)
			break;

		grub_printf("Invalid entry.\n");
	}

	grub_dprintf("zvboot", "entry %d chosen (index %d).\n", num, num-1);

	num--;

	walk_config(match_be_index, &num);
}

/*
 * A matching BE is either one explicitly chosen via bootfs, or the default one.
 */
/*ARGSUSED*/
static int
match_be(config_entry_t *entry, void *arg)
{
	char *ba_be = arg;

	/* Administrator selected BE. */
	if (ba_be != NULL) {
		if (grub_strcmp(entry->ce_bootfs, ba_be) == 0) {
			grub_memcpy(&chosen_entry, entry, sizeof (chosen_entry));
			return (1);
		}
		return (0);
	}

	if (entry->ce_default) {
		grub_memcpy(&chosen_entry, entry, sizeof (chosen_entry));
		return (1);
	}

	return (0);
}

static void
load_module(const char *device, const char *pool, const char *module_path)
{
	/*
	 * Load module if present.
	 */
	if (module_path[0] == '\0')
		return;

	grub_file_t module;
	if ((module = file_open(device, pool, module_path)) == NULL) {
		/*
		 * If we couldn't find a boot archive and this is a ramdisk,
		 * assume that the ramdisk _is_ the boot archive and continue on
		 * with it.
		 */
		if (grub_strcmp(device, GRUB_RAMDISK_DEVICE) == 0)
			return;

		panic("Failed to open boot_archive: '%s' (%d)",
		    module_path, grub_errno);
	}

	grub_printf("reading module %s...", module_path);

	module_size = grub_file_size(module);
	if (module_size == 0)
		panic("Zero sized module %s", module_path);

	if ((module_bytes = page_alloc(module_size)) == NULL)
		panic("Cannot allocate %ld bytes for module.", module_size);
	grub_dprintf("zvboot", "module is %ld (0x%lx) bytes at %p.\n",
	    module_size, module_size, module_bytes);

	if (file_read(module, module_bytes, module_size) != module_size)
		panic("Read of module file %s failed", module_path);

	grub_file_close(module);

	grub_printf("done.\n");

	/* Add the boot_archive as a ramdisk. */
	memdisk_set(module_bytes, module_size);
}

/*
 * Load kernel from disk.
 */
static void
load_kernel(const char *device, const char *pool, const char *kernel_path)
{
	grub_printf("reading kernel file %s...", kernel_path);

	grub_file_t kernel;
	if ((kernel = file_open(device, pool, kernel_path)) == NULL)
		panic("Cannot open kernel file: %s (%d)",
		    kernel_path, grub_errno);

	grub_ssize_t kernel_size = grub_file_size(kernel);
	if (kernel_size == 0)
		panic("Zero sized kernel %s", kernel_path);

	if ((kernel_bytes = page_alloc(kernel_size)) == NULL)
		panic("Cannot allocate %ld bytes for kernel.", kernel_size);
	grub_dprintf("zvboot", "kernel is %ld (0x%lx) bytes at %p.\n",
	    kernel_size, kernel_size, kernel_bytes);

	if (file_read(kernel, kernel_bytes, kernel_size) != kernel_size)
		panic("Read of kernel file %s failed", kernel_path);
	grub_file_close(kernel);

	grub_printf("done.\n");

	path_to_kernel = (char *)kernel_path;
}

/*
 * Lifted from grub_cmd_zfs_bootfs().
 */
static char *
set_zfs_bootargs(const char *device, const char *dataset)
{
	static char zfs_args[BOOTARGS_MAX];
	char work[BOOTARGS_MAX];

	zfs_args[0] = '\0';

	grub_device_t dev;
	dev = grub_device_open(device);
	if (dev == NULL) {
		grub_dprintf("zvboot", "Cannot open device %s (%d).\n",
		    device, grub_errno);
		return (NULL);
	}

	/* Strip the pool name, if present. */
	const char *slash, *fs;
	if ((dataset[0] != '/') &&
	    ((slash = grub_strchr(dataset, '/')) != NULL))
		fs = slash;
	else
		fs = dataset;

	grub_err_t err;
	grub_uint64_t mdnobj;
	err = grub_zfs_getmdnobj(dev, fs, &mdnobj);
	if (err != 0) {
		grub_dprintf("zvboot", "Cannot find mdnobj for "
		    "filesystem %s on device %s (%d).\n",
		    fs, device, err);
		grub_device_close(dev);
		return (NULL);
	}

	char *poolname = get_poolname(device);
	if (poolname == NULL) {
		grub_dprintf("zvboot", "Cannot find poolname for %s.\n",
		    device);
		grub_device_close(dev);
		return (NULL);
	}
	
	grub_snprintf(work, sizeof (work), " -B zfs-bootfs=%s/%ld",
	    poolname, mdnobj);
	grub_strncat(zfs_args, work, BOOTARGS_MAX - 1);
	grub_free(poolname);

	/*
	 * If the pool was previously imported somewhere else, the bootpath
	 * in the vdev label is useless.  Kernel zones are a very controlled
	 * environment - we know what the physical path is.  That is, we need to
	 * translate "zvblk0,gpt2" to "/zvnex/zvblk@0:b".
	 */
	const char letters[] = "abcdefgh";
	const int zvblklen = grub_strlen("zvblk");
	char bootpath[MAXPATHLEN];
	int inst, partnum;
	char *p;

	if (grub_strncmp(device, "zvblk", zvblklen) != 0) {
		grub_dprintf("zvboot", "Invalid device %s.\n", device);
		grub_device_close(dev);
		return (NULL);
	}

	inst = get_num(device + zvblklen, NULL);

	if ((p = grub_strchr(device + zvblklen, ',')) == NULL) {
		grub_dprintf("zvboot", "No partition in %s.\n", device);
		grub_device_close(dev);
		return (NULL);
	}
	p++;
	if (grub_strncmp(p, "gpt", grub_strlen("gpt")) != 0) {
		grub_dprintf("zvboot", "Invalid partition in %s.\n", device);
		grub_device_close(dev);
		return (NULL);
	}

	partnum = get_num(p + grub_strlen("gpt"), NULL);
	if (partnum < 1 || partnum > grub_strlen(letters)) {
		grub_dprintf("zvboot", "Invalid gpt partition %s in %s.\n", p,
		    device);
		grub_device_close(dev);
		return (NULL);
	}

	grub_snprintf(bootpath, sizeof (bootpath), "/zvnex/zvblk@%d:%c", inst,
	    letters[partnum - 1]);
	grub_snprintf(work, sizeof (work), " -B bootpath=\"%s\"", bootpath);
	grub_strncat(zfs_args, work, BOOTARGS_MAX - 1);

	char *nvlist;
	err = grub_zfs_fetch_nvlist(dev, &nvlist);
	if (err != 0) {
		grub_dprintf("zvboot", "Cannot fetch nvlist for %s.\n",
		    device);
		grub_device_close(dev);
		return (NULL);
	}
	grub_device_close(dev);

	grub_uint64_t devguid;
	if (grub_zfs_nvlist_lookup_uint64(nvlist, ZPOOL_CONFIG_GUID,
		&devguid) == 0) {
		grub_dprintf("zvboot", "Cannot find GUID for %s.\n",
		    device);
		grub_free(nvlist);
		return (NULL);
	}

	char *nv = grub_zfs_nvlist_lookup_nvlist(nvlist,
	    ZPOOL_CONFIG_VDEV_TREE);
	if (nv == NULL) {
		grub_dprintf("zvboot", "Cannot find vdev in nvlist for %s.\n",
		    device);
		grub_free(nvlist);
		return (NULL);
	}

	/* From zfsinfo.c. */
	extern grub_err_t get_bootpath(char *, grub_uint64_t, char **, char **);
	char *cfg_bootpath, *devid;
	if (get_bootpath(nv, devguid, &cfg_bootpath, &devid) == GRUB_ERR_NONE) {
		if (grub_strcmp(bootpath, cfg_bootpath) != 0) {
			grub_dprintf("zvboot",
			    "Preferring calculated bootpath (%s) to that from "
			    "the pool (%s).\n", bootpath, cfg_bootpath);
		}

		if (devid != NULL) {
			grub_snprintf(work, sizeof (work),
			    " -B diskdevid=\"%s\"", devid);
			grub_strncat(zfs_args, work, BOOTARGS_MAX - 1);
		}
	}

	grub_free(nv);
	grub_free(nvlist);

	return (zfs_args);
}

static void
set_bootargs(const char *device, const char *dataset, const char *inargs,
    zv_boot_info_t *binfo)
{
	if ((boot_args = grub_malloc(BOOTARGS_MAX)) == NULL)
		panic("Cannot allocate %d bytes for boot args.", BOOTARGS_MAX);
		   
	boot_args[0] = '\0';

	/*
	 * Boot arguments specified at boot time override anything in
	 * the grub configuration file.
	 */
	if (inargs != NULL)
		grub_strncat(boot_args, inargs, BOOTARGS_MAX - 1);

	if (grub_strcmp(boot_fstype, "zfs") == 0) {
		char *zfs_args = set_zfs_bootargs(device, dataset);

		grub_strncat(boot_args, zfs_args, BOOTARGS_MAX - 1);
	}

	if (grub_strstr(boot_args, "zvmm_platform=true") == NULL) {
		(void) grub_strncat(boot_args, " -B zvmm_platform=true",
		    BOOTARGS_MAX - 1);
	}

	if (grub_strstr(boot_args, "boot-max-ncpus=") == NULL) {
		char ncpu[5];

		(void) grub_snprintf(ncpu, sizeof (ncpu) - 1, "%d",
		    binfo->zv_ncpu);
		(void) grub_strncat(boot_args, " -B boot-max-ncpus=",
		    BOOTARGS_MAX - 1);
		(void) grub_strncat(boot_args, ncpu, BOOTARGS_MAX - 1);
	}

#if defined(amd64)
	if (grub_strstr(boot_args, "console=ttya") == NULL)
		(void) grub_strncat(boot_args, " -B console=ttya", BOOTARGS_MAX - 1);
#endif /* amd64 */

	grub_dprintf("zvboot", "final boot_args: %s\n", boot_args);
}

static void
open_grub_configuration(const char *boot_dev)
{
	grub_file_t menu_lst;
	grub_ssize_t sz;

	menu_lst = file_open(boot_dev, "/", config_file_path);
	if (menu_lst == NULL) {
		grub_printf("zvboot: cannot open %s, using default "
		    "configuration.\n", config_file_path);
		config = "";
		return;
	}
	sz = grub_file_size(menu_lst);

	grub_dprintf("zvboot", "Config file is %ld bytes.\n", sz);

	config = grub_malloc(sz + 1);

	if (file_read(menu_lst, config, sz) != sz)
		panic("Read of %s failed", config_file_path);
	grub_file_close(menu_lst);

	config[sz] = '\0';
}

static void
select_be(boot_args_t *ba)
{
	grub_memset(&chosen_entry, 0, sizeof (chosen_entry));

	if (ba->ba_list)
		list_bes();
	else
		walk_config(match_be, (void *)ba->ba_be);
}

/*
 * Select arguments to boot the kernel with. The priority of boot arguments is
 * following: Transient host arguments, transient guest arguments (stored in the
 * "Solaris_reboot_transient" entry), common entry arguments, default host
 * arguments (zonecfg property "bootargs").
 */
static boot_args_t *
select_bootargs(boot_args_t *ba_htrans, boot_args_t *ba_guest,
    boot_args_t *ba_hdef)
{
	if (!ba_is_empty(ba_htrans) || ba_htrans->ba_list ||
	    ba_htrans->ba_be != NULL) {

		select_be(ba_htrans);
		if (chosen_entry.ce_title == NULL)
			return (NULL);
		if (!ba_is_empty(ba_htrans))
			return (ba_htrans);
		/*
		 * If a host user called zoneadm reboot just with -L or -Z
		 * <dataset> then we proceed with arguments from the grub entry
		 * and then with the default host arguments.
		 */
		parse_bootargs(chosen_entry.ce_bootargs, ba_guest);

		if (!ba_is_empty(ba_guest))
			return (ba_guest);

		return (ba_hdef);
	}

	/*
	 * The host invoked (re)boot without any arguments OR 'reboot' was
	 * called within the ba_guest without or with arguments (transient
	 * ba_guest arguments).
	 */

	walk_config(match_be, NULL);
	if (chosen_entry.ce_title == NULL)
		return (NULL);

	/*
	 * In order to prevent a chaining of -L and -Z <dataset> options we
	 * permit them only in the transient entry. All other occurences are
	 * ignored.
	 */

	if (grub_strcmp(chosen_entry.ce_title,
	    "Solaris_reboot_transient") == 0) {
		parse_bootargs(chosen_entry.ce_bootargs, ba_guest);
		select_be(ba_guest);
		if (chosen_entry.ce_title == NULL)
			return (NULL);
		if (!ba_is_empty(ba_guest))
			return (ba_guest);
		grub_free(ba_guest->ba_be);
	}

	parse_bootargs(chosen_entry.ce_bootargs, ba_guest);

	if (!ba_is_empty(ba_guest))
		return (ba_guest);

	return (ba_hdef);
}

void
do_grubby_stuff(zv_boot_info_t *binfo)
{
	/*
	 * Provide some heap for grub_malloc().
	 *
	 * Be careful using any grub_* functions before this point, as
	 * they will not be able to allocate any memory.
	 */
#define	HEAP_SIZE	(MBYTE(2))
	void *heap = page_alloc(HEAP_SIZE);
	grub_mm_init_region(heap, HEAP_SIZE);

	boot_args_t *ba;
	boot_args_t ba_htrans = { 0 };
	boot_args_t ba_hdef = { 0 };
	boot_args_t ba_guest = { 0 };

	get_host_bootargs(&ba_htrans, &ba_hdef);

	char *boot_device = get_boot_device(&ba_htrans);

	/* After `get_bootargs' as that might enable debug output. */
	grub_dprintf("zvboot", "last_phys = %p\n", (void *)last_phys);
	grub_dprintf("zvboot", "heap is 0x%lx bytes at %p\n", HEAP_SIZE, heap);
	
	/* If there is a ramdisk loaded, declare it. */
	if ((binfo->zv_ramdisk_start != 0) &&
	    (binfo->zv_ramdisk_size > 0)) {
		memdisk_set(binfo->zv_ramdisk_start,
		    binfo->zv_ramdisk_size);
	}

	/* Discover the disks available. */
	extern void zvblk_init(void);
	zvblk_init();

	/* Declare the partition schemes. */
#if defined(amd64)
	extern struct grub_partition_map *_grub_msdos_partition_map;
	grub_partition_map_register(_grub_msdos_partition_map);
	extern struct grub_partition_map *_grub_sun_pc_partition_map;
	grub_partition_map_register(_grub_sun_pc_partition_map);
#endif /* amd64 */
	extern struct grub_partition_map *_grub_sun_partition_map;
	grub_partition_map_register(_grub_sun_partition_map);
	extern struct grub_partition_map *_grub_gpt_partition_map;
	grub_partition_map_register(_grub_gpt_partition_map);

	/* Declare the filesystems. */
	extern struct grub_fs *_grub_iso9660_fs;
	grub_fs_register(_grub_iso9660_fs);
	extern struct grub_fs *_grub_zfs_fs;
	grub_fs_register(_grub_zfs_fs);
	extern struct grub_fs *_grub_ufs_fs;
	grub_fs_register(_grub_ufs_fs);

	/* Declare the GZIP file reader. */
	extern grub_file_t grub_gzio_open (grub_file_t);
	grub_file_filter_register (GRUB_FILE_FILTER_GZIO, grub_gzio_open);

	/*
	 * If a ramdisk was specified, load the configuration file,
	 * kernel, etc. from it. Otherwise choose a boot device,
	 * presumably a disk or cdrom.
	 */
	grub_device_t dev;
	if ((dev = grub_device_open(GRUB_RAMDISK_DEVICE)) != 0) {
		boot_device_name = GRUB_RAMDISK_DEVICE;
		grub_device_close(dev);
	} else {
		/*
		 * Failure to open the ramdisk should not carry
		 * forward.
		 */
		grub_errno = GRUB_ERR_NONE;

		boot_device_name = select_boot_device(boot_device);
	}

	open_grub_configuration(boot_device_name);

	ba = select_bootargs(&ba_htrans, &ba_guest, &ba_hdef);
	if (ba != NULL) {
		process_debug_bootargs(ba);
		print_ba("Chosen boot arguments", ba);
	}

	if (chosen_entry.ce_title != NULL)
		print_entry("Chosen BE", &chosen_entry);
	else 
		grub_printf("zvboot: No 'default' entry, using default "
		    "configuration.\n");

	/*
	 * Complete the entry by merging grub entry arguments with provided or
	 * found arguments. Provided arguments have a higher priority. If we
	 * failed to find the entry in the grub menu (ba is NULL) provided
	 * arguments cannot be trusted and so system default values will be
	 * used.
	 */

	complete_entry(&chosen_entry, boot_device_name, ba);
	print_entry("Completed BE", &chosen_entry);

	grub_printf("Boot device: %s  File and args: %s%s%s\n",
	    boot_device, STR(chosen_entry.ce_kernel),
	    chosen_entry.ce_kernel != NULL ? " " : "",
	    STR(chosen_entry.ce_bootargs));

	load_module(chosen_entry.ce_rootdev, chosen_entry.ce_bootfs,
	    chosen_entry.ce_module);

	/*
	 * If we can mount the boot_archive, load unix from
	 * it. Otherwise, load it from the disk.
	 */
	grub_device_t boot_archive = grub_device_open(GRUB_RAMDISK_DEVICE);
	char *unixdev = chosen_entry.ce_rootdev;
	if (boot_archive != NULL) {
		if (grub_fs_probe(boot_archive) != NULL)
			unixdev = GRUB_RAMDISK_DEVICE;
		grub_device_close(boot_archive);
	}

	load_kernel(unixdev, chosen_entry.ce_bootfs, chosen_entry.ce_kernel);

	set_bootargs(chosen_entry.ce_rootdev, chosen_entry.ce_bootfs,
	    chosen_entry.ce_bootargs, binfo);
}
