/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2011, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/varargs.h>
#include <sys/elf.h>
#include <sys/elf_notes.h>
#ifdef amd64
#include <sys/archsystm.h>
#include <sys/x86_archext.h>
#include <sys/controlregs.h>
#endif /* amd64 */

#include <grub/misc.h>
#include <grub/mm.h>

#include "zv_mem.h"
#include "zv_asm.h"
#include "zvcons.h"
#include "zvload.h"
#include "verify.h"

#ifdef amd64
/*
 * from Solaris kernel source for now
 */
#define	_BOOT_TARGET_amd64
#include <sys/bootinfo.h>
#endif /* amd64 */

#include <sys/zvmm/zvmm_hcall.h>
#include <sys/zvmm/zvmm_mach_hcall.h>

/*
 * values of command line arguments
 */
char *boot_args;	/* arguments passed to the 64 bit kernel */
char *path_to_kernel;

char *kernel_bytes;	/* 64 bit ELF kernel that is loaded */
char *module_bytes;	/* module file that is loaded */
size_t module_size;

/*
 * from ELF - virt addr for kernel nucleus 8Meg for amd64
 */
uint64_t kernel_virt;

#ifdef amd64
static struct boot_modules boot_module;
static struct xboot_info xboot;
#endif /* amd64 */

#define	PGETBYTES(off) (void *)(kernel_bytes + (off))

/*
 * Parse the kernel ELF file for program information and copy the kernel
 * text and data segments to their physical memory locations.
 */
static void
elfload64(void)
{
	Elf64_Ehdr *eh;
	Elf64_Phdr *phdr;
	caddr_t allphdrs;
	int i;
	char *src;
	char *dst;
	uchar_t *ident;

	ident = PGETBYTES(0);
	if (ident == NULL)
		panic("Cannot read kernel ELF header");

	if (ident[EI_MAG0] != ELFMAG0 || ident[EI_MAG1] != ELFMAG1 ||
	    ident[EI_MAG2] != ELFMAG2 || ident[EI_MAG3] != ELFMAG3)
		panic("malformed kernel ELF file");

	if (ident[EI_CLASS] == ELFCLASS64)
		eh = PGETBYTES(0);
	else
		panic("kernel not ELF64");

	if (eh->e_type != ET_EXEC)
		panic("not ET_EXEC, e_type = 0x%x", eh->e_type);

	if (eh->e_phnum == 0 || eh->e_phoff == 0)
		panic("no program headers");

	/*
	 * Get the program headers.
	 */
	allphdrs = PGETBYTES(eh->e_phoff);

	/*
	 * Next look for interesting program headers.
	 */
	for (i = 0; i < eh->e_phnum; i++) {
		/*LINTED [ELF program header alignment]*/
		phdr = (Elf64_Phdr *)(allphdrs + eh->e_phentsize * i);

		/*
		 * Dynamically-linked executables are not expected nor
		 * usable.
		 */
		if (phdr->p_type == PT_INTERP) {
			panic("Illegal PT_INTERP section.");
		}

		/*
		 * At this point we only care about PT_LOAD segments.
		 */
		if (phdr->p_type != PT_LOAD)
			continue;

		if (phdr->p_flags == (PF_R | PF_W) && phdr->p_vaddr == 0)
			panic("warning: krtld reloc info?");

		/*
		 * If memory size is zero just ignore this header.
		 */
		if (phdr->p_memsz == 0)
			continue;

#ifdef amd64
		grub_dprintf("zvboot", "p: %p, v: %p\n",
		    (void *)phdr->p_paddr, (void *)phdr->p_vaddr);
		/*
		 * If load address 1:1 then ignore this header, it's for dboot.
		 */
		if (phdr->p_paddr == phdr->p_vaddr)
			continue;

		/*
		 * Copy the data to kernel area.
		 */
		switch (phdr->p_paddr) {
		case MBYTE(4): /* KTEXT_PHYS */
			/* Kernel text loads at PA 4M. */
			kernel_virt = phdr->p_vaddr;
			break;
		case MBYTE(12): /* KDATA_PHYS */
			/* Kernel data loads at PA 12M. */
			break;
		default:
			panic("Bad p_paddr for nucleus segment");
		}
		dst = (char *)phdr->p_paddr;
#endif /* amd64 */

		src = (char *)PGETBYTES(phdr->p_offset);
		grub_dprintf("zvboot", "elf: move 0x%lx from %p to %p\n",
		    phdr->p_filesz, src, dst);
		grub_memcpy(dst, src, phdr->p_filesz);
	}
}

/*
 * Boot a kernel zone
 */
void
boot(zv_boot_info_t *zv_boot_info)
{
	char *cp;

	/*
	 * Make sure memory allocator is aligned.
	 */
	last_phys = ROUND_UP((uintptr_t)zv_boot_info->free_mem_start, PGSIZE);

	/*
	 * Read in the kernel and module files.
	 */
	extern void do_grubby_stuff(zv_boot_info_t *zv_boot_info);
	do_grubby_stuff(zv_boot_info);

	/*
	 * Verify the kernel module before loading.
	 */
	(void) verify_module((uchar_t *)kernel_bytes);

	/*
	 * Copy the kernel text/data from the ELF file to its desired
	 * RAM.
	 */
	elfload64();

#ifdef amd64
	/*
	 * Fill in xboot info.
	 */
	grub_memset(&xboot, 0, sizeof (xboot));
	xboot.bi_phys_install = (uint64_t)zv_boot_info->zv_memlist;

	/*
	 * Record where the boot module is in memory.
	 */
	if (module_bytes != NULL) {
		boot_module.bm_addr = (uint64_t)module_bytes;
		boot_module.bm_size = module_size;
	} else {
		boot_module.bm_addr =
		    (uint64_t)zv_boot_info->zv_ramdisk_start;
		boot_module.bm_size = zv_boot_info->zv_ramdisk_size;
	}
	xboot.bi_modules = (uint64_t)&boot_module;
	xboot.bi_module_cnt = 1;

	/*
	 * Build the kernel command line.
	 */
	cp = page_alloc(PGSIZE);
	grub_strcpy(cp, path_to_kernel);
	if (boot_args != NULL) {
		grub_strcat(cp, " ");
		grub_strcat(cp, boot_args);
	}
	xboot.bi_cmdline = (uint64_t)cp;

	grub_dprintf("zvboot", "kernel command line: %s\n", cp);

	/*
	 * Add high memory page table entries.
	 */
	uint64_t off;
	xboot.bi_top_page_table = getcr3() & ~0xffful;
	xboot.bi_pt_window = (uint64_t)page_alloc(PGSIZE);
	/* size is 2 * KERN_NUCLEUS_SIZE */
	for (off = 0; off < 4 * MBYTE(4); off += MBYTE(2)) {
		grub_dprintf("zvboot", "map %p at %p\n",
		    (void *)(MBYTE(4) + off), (void *)(kernel_virt + off));
		map_pa_at_va(MBYTE(4) + off, (void *)(kernel_virt + off), 1);
	}
	xboot.bi_pte_to_pt_window =
	    (uint64_t)find_pte((uintptr_t)xboot.bi_pt_window, 0);

	/*
	 * Finish off the xboot data.
	 */
	xboot.bi_rsvdmem = NULL;	/* no BIOS memory */
	xboot.bi_pcimem = NULL;		/* no device memory */
	xboot.bi_uefi_systab = NULL;
	xboot.bi_smbios_entry = NULL;
	xboot.bi_smbios_entry_len = 0;
	xboot.bi_framebuffer_info = 0;
	xboot.bi_mb_info = NULL;	/* since we didn't use grub */

	xboot.bi_use_largepage = 1;
	xboot.bi_use_pae = 1;
	xboot.bi_use_nx = 1;
	xboot.bi_use_pge = 1;
	xboot.bi_kseg_size = (uint64_t)MBYTE(8);
	xboot.bi_next_paddr = (uint64_t)last_phys;
	xboot.bi_next_vaddr = (uint64_t)last_phys;

	grub_dprintf("zvboot", "xboot.bi_cmdline = %p\n",
	    (void *)xboot.bi_cmdline);
	grub_dprintf("zvboot", "xboot.bi_next_paddr = %p\n",
	    (void *)xboot.bi_next_paddr);
	grub_dprintf("zvboot", "xboot.bi_pt_window = %p\n",
	    (void *)xboot.bi_pt_window);
	grub_dprintf("zvboot", "xboot.bi_pte_to_pt_window = %p\n",
	    (void *)xboot.bi_pte_to_pt_window);
	grub_dprintf("zvboot", "boot_module.bm_addr = %p\n",
	    (void *)boot_module.bm_addr);
	grub_dprintf("zvboot", "boot_module.bm_size = 0x%x\n",
	    boot_module.bm_size);

	/* That's it, just enter the kernel. */
	grub_dprintf("zvboot", "Entering kernel at VA = %p\n",
	    (void *)kernel_virt);

	void (*entry_addr)(struct xboot_info *);
	entry_addr = (void (*)(struct xboot_info *))kernel_virt;
	entry_addr(&xboot);
#endif /* amd64 */
}
