/* ask_passwd.c - test module for dynamic loading */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
 *  Copyright (C) 2003,2007  Free Software Foundation, Inc.
 *  Copyright (C) 2003  NIIBE Yutaka <gniibe@m17n.org>
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/types.h>
#include <grub/misc.h>
#include <grub/time.h>
#include <grub/err.h>
#include <grub/dl.h>
#include <grub/extcmd.h>
#include <grub/i18n.h>
#include <grub/auth.h>

GRUB_MOD_LICENSE ("GPLv3+");

struct pbkdf2_password
{
  grub_uint8_t *salt;
  grub_size_t saltlen;
  unsigned int c;
  grub_uint8_t *expected;
  grub_size_t buflen;
};

static inline int
hex2val (char hex)
{
  if ('0' <= hex && hex <= '9')
    return hex - '0';
  if ('a' <= hex && hex <= 'f')
    return hex - 'a' + 10;
  if ('A' <= hex && hex <= 'F')
    return hex - 'A' + 10;
  return -1;
}

static grub_err_t
check_menulock_password_pbkdf2 (const char *entered, void *pin)
{
  grub_uint8_t *buf;
  struct pbkdf2_password *pass = pin;
  gcry_err_code_t err;

  buf = grub_malloc (pass->buflen);
  if (!buf)
    return grub_crypto_gcry_error (GPG_ERR_OUT_OF_MEMORY);

  err = grub_crypto_pbkdf2 (GRUB_MD_SHA512, (grub_uint8_t *) entered,
                            grub_strlen (entered),
                            pass->salt, pass->saltlen, pass->c,
                            buf, pass->buflen);
  if (err)
    {
      grub_free (buf);
      return grub_crypto_gcry_error (err);
    }

  if (grub_crypto_memcmp (buf, pass->expected, pass->buflen) != 0)
    return GRUB_ACCESS_DENIED;

  return GRUB_ERR_NONE;
}

static grub_err_t
grub_ask_passwd (grub_extcmd_context_t ctxt __attribute__ ((unused)),
                                  int argc, char **args)
{
  char *ptr, *ptr2;
  grub_uint8_t *ptro;
  struct pbkdf2_password *pass;

  int max_num_attempt;
  char entered[GRUB_AUTH_MAX_PASSLEN];
  int num_attempt = 0;
  unsigned long punishment_delay = 1;

  if (argc != 2)
    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("two argument expected"));

  if (grub_memcmp (args[0], "grub.pbkdf2.sha512.",
                   sizeof ("grub.pbkdf2.sha512.") - 1) != 0)
    return grub_error (GRUB_ERR_BAD_ARGUMENT, "incorrect PBKDF2 password");

  ptr = args[0] + sizeof ("grub.pbkdf2.sha512.") - 1;

  pass = grub_malloc (sizeof (*pass));
  if (!pass)
    return grub_errno;

  pass->c = grub_strtoul (ptr, &ptr, 0);
  if (*ptr != '.')
    {
      grub_free (pass);
      return grub_error (GRUB_ERR_BAD_ARGUMENT, "incorrect PBKDF2 password");
    }
  ptr++;

  ptr2 = grub_strchr (ptr, '.');
  if (!ptr2 || ((ptr2 - ptr) & 1) || grub_strlen (ptr2 + 1) & 1)
    {
      grub_free (pass);
      return grub_error (GRUB_ERR_BAD_ARGUMENT, "incorrect PBKDF2 password");
    }

  pass->saltlen = (ptr2 - ptr) >> 1;
  pass->buflen = grub_strlen (ptr2 + 1) >> 1;
  ptro = pass->salt = grub_malloc (pass->saltlen);
  if (!ptro)
    {
      grub_free (pass);
      return grub_errno;
    }
  while (ptr < ptr2)
    {
      int hex1, hex2;
      hex1 = hex2val (*ptr);
      ptr++;
      hex2 = hex2val (*ptr);
      ptr++;
      if (hex1 < 0 || hex2 < 0)
        {
          grub_free (pass->salt);
          grub_free (pass);
          return grub_error (GRUB_ERR_BAD_ARGUMENT,
                             "incorrect PBKDF2 password");
        }

      *ptro = (hex1 << 4) | hex2;
      ptro++;
    }

  ptro = pass->expected = grub_malloc (pass->buflen);
  if (!ptro)
    {
      grub_free (pass->salt);
      grub_free (pass);
      return grub_errno;
    }
  ptr = ptr2 + 1;
  ptr2 += grub_strlen (ptr2);
  while (ptr < ptr2)
    {
      int hex1, hex2;
      hex1 = hex2val (*ptr);
      ptr++;
      hex2 = hex2val (*ptr);
      ptr++;
      if (hex1 < 0 || hex2 < 0)
        {
          grub_free (pass->expected);
          grub_free (pass->salt);
          grub_free (pass);
          return grub_error (GRUB_ERR_BAD_ARGUMENT,
                             "incorrect PBKDF2 password");
        }

      *ptro = (hex1 << 4) | hex2;
      ptro++;
    }

  max_num_attempt = grub_strtoul (args[1], 0, 10);
  /*User gets N number of attempts before system reboots*/
  while (num_attempt < max_num_attempt)
    {
      grub_puts_ (N_("Enter password: "));

      if (!grub_password_get (entered, GRUB_AUTH_MAX_PASSLEN))
        {
          num_attempt++;
          grub_sleep (punishment_delay);
          continue;
        }

      if (check_menulock_password_pbkdf2(entered, pass))
        {
          num_attempt++;
          grub_printf ("Access denied\n");
          grub_sleep (punishment_delay);
          continue;
        }
      return GRUB_ERR_NONE;
    }
  /*reboot the system if incorrect password is entered*/
  grub_printf ("Rebooting system");
  grub_sleep (punishment_delay);
  grub_reboot ();
}

static grub_extcmd_t ask_passwd;

GRUB_MOD_INIT(ask_passwd)
{
  ask_passwd = grub_register_extcmd ("ask_passwd", grub_ask_passwd, 0,
                              N_("PASSWORD <N>"),
                              N_("Ask password <N> times."),
                              0);
}

GRUB_MOD_FINI(ask_passwd)
{
  grub_unregister_extcmd (ask_passwd);
}
