/* read.c - Command to read variables from user.  */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2006,2007,2008  Free Software Foundation, Inc.
 *  Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/dl.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/env.h>
#include <grub/term.h>
#include <grub/types.h>
#include <grub/extcmd.h>
#include <grub/i18n.h>

GRUB_MOD_LICENSE ("GPLv3+");

static char *
grub_getline (int noblock, int noecho)
{
  int i;
  char *line;
  char *tmp;
  char c;

  i = 0;
  line = grub_malloc (1 + i + sizeof('\0'));
  if (! line)
    return NULL;

  while (1)
    {
      if (noblock && grub_checkkey() <= 0)
        break;
      c = grub_getkey ();
      if ((c == '\n') || (c == '\r'))
	break;

      line[i] = c;
      if (!noecho && grub_isprint (c))
	grub_printf ("%c", c);
      i++;
      tmp = grub_realloc (line, 1 + i + sizeof('\0'));
      if (! tmp)
	{
	  grub_free (line);
	  return NULL;
	}
      line = tmp;
    }
  line[i] = '\0';

  return line;
}

static grub_err_t
grub_cmd_read (grub_extcmd_context_t ctxt, int argc, char **args)
{
  int noblock = 0, noecho = 0;
  char *line;
  if (ctxt->state[0].set)
     noblock = 1;
  if (ctxt->state[1].set)
     noecho = 1;
  line = grub_getline (noblock, noecho);
  if (! line)
    return grub_errno;
  if (argc > 0 && *line != 0)
    grub_env_set (args[0], line);

  grub_free (line);
  return 0;
}

static const struct grub_arg_option options[] =
  {
    {"noblock", 'd', 0, "don't block on the read; use whatever "
	"input is available", 0, 0},
    {"noecho", 'n', 0, "don't echo input as it's read", 0, 0},
    {0, 0, 0, 0, 0, 0}
  };

static grub_extcmd_t cmd;

GRUB_MOD_INIT(read)
{
  cmd = grub_register_extcmd ("read", grub_cmd_read, 0,
			       N_("[-dn] [ENVVAR]"),
			       N_("Set variable with user input."),
                              options);
}

GRUB_MOD_FINI(read)
{
  grub_unregister_extcmd (cmd);
}
