/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1999,2003,2007,2008,2009 Free Software Foundation, Inc.
 *  Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 *  IN NO EVENT SHALL ANY DEVELOPER OR DISTRIBUTOR BE LIABLE FOR ANY CLAIM,
 *  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 *  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 *  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *  GRUB is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Emulate required libgcrypt functions for use by the MPI library in GRUB.
 * This includes string handling, memory management, error handling,
 * and random functions.  The emulation in most cases is performed by
 * simply calling the equivalent grub_*() function.
 *
 * Random functions are dummy functions for linking only, as they are only
 * needed for key generation and sessions, not RSA signature verification.
 */

#include <sys/types.h>		/* size_t */
#include <string.h>		/* str* */
#include <strings.h>		/* bcopy */

#include <gcrypt.h>		/* gcry_* */
#include <random.h>		/* _gcry_*random* */

#include <grub/misc.h>		/* grub_str*() and other useful utilities */
#include <grub/mm.h>		/* grub_*alloc* and other memory management */
#include <grub/err.h>		/* grub_errno */

#include <gpg-error.h>		/* GPG_ERR_* codes */

/*
 * String handling
 */

size_t
strlen(const char *s)
{
	return (grub_strlen(s));
}

void
bcopy(const void *src, void *dest, size_t n)
{
	(void) grub_memcpy(dest, src, n);
}

/*
 * Memory Management
 */

void *
_gcry_malloc(size_t n)
{
	return (grub_malloc(n));
}

void *
_gcry_calloc(size_t n, size_t m)
{
	size_t	nm = n * m;
	void	*p;

	p = grub_malloc(nm);
	if (p != NULL)
		grub_memset(p, 0, nm);
	return (p);
}

void
_gcry_free(void *p)
{
	if (p != NULL)
		grub_free(p);
}

void *
_gcry_xrealloc(void *p, size_t n)
{
	return (grub_realloc(p, n));
}

void
_gcry_check_heap(const void *p)
{
	(void) p; /* Not used */
}

void *
_gcry_xmalloc(size_t n)
{
	return (grub_malloc(n));
}

/*
 * Other processes are prevented from accessing the data and you must be
 * privileged to use this. It doesn't apply to GRUB or other boot-type
 * environment.
 */
void *
_gcry_malloc_secure(size_t n)
{
	return (grub_malloc(n));
}

void *
_gcry_xmalloc_secure(size_t n)
{
	return (grub_malloc(n));
}

void *
_gcry_xcalloc(size_t n, size_t m)
{
	return (_gcry_calloc(n, m));
}

void *
_gcry_xcalloc_secure(size_t n, size_t m)
{
	return (_gcry_calloc(n, m));
}

int
_gcry_is_secure(const void *a)
{
	(void) a; /* Not used */
	return (0);
}


/*
 * Logging functions
 */
void
_gcry_log_bug(const char *fmt, ...)
{
	va_list	pvar;

	va_start(pvar, fmt);
	(void) grub_vprintf(fmt, pvar);
	va_end(pvar);
}

void
_gcry_log_printf(const char *fmt, ...)
{
	va_list	pvar;

	if (fmt != NULL) {
		va_start(pvar, fmt);
		(void) grub_vprintf(fmt, pvar);
		va_end(pvar);
	}
}

void
_gcry_assert_failed(const char *expr, const char *file, int line,
    const char *func)
{
	grub_fatal("Assertion %s failed at %s:%s():%d\n",
	    expr, file, func, line);
}

/*
 * Retrieve the error code directly from the ERRNO variable. This
 * returns GPG_ERR_UNKNOWN_ERRNO if the system error is not mapped
 * (report this) and GPG_ERR_MISSING_ERRNO if ERRNO has the value 0.
 */
gpg_err_code_t
gpg_err_code_from_syserror(void)
{
	if (grub_errno == 0) {
		return (GPG_ERR_MISSING_ERRNO);
	} else if (grub_errno < (grub_err_t)GPG_ERR_CODE_DIM) {
		return (GPG_ERR_SYSTEM_ERROR | grub_errno);
	} else {
		return (GPG_ERR_UNKNOWN_ERRNO);
	}
}

/*
 * Dummy random functions
 */
void *
_gcry_random_bytes(size_t nbytes, enum gcry_random_level level)
{
	(void) nbytes;
	(void) level;
	return (NULL);
}

void *
_gcry_random_bytes_secure(size_t nbytes, enum gcry_random_level level)
{
	return (_gcry_random_bytes(nbytes, level));
}

void
_gcry_create_nonce(void *buffer, size_t length)
{
	if ((buffer != NULL) && (length > 0))
		grub_memset(buffer, 0xff, length);
}
