/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2009, Free Software Foundation, Inc.
 *  Copyright (c) 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Emulate required libgcrypt functions for use by the MPI library in GRUB.
 * This for Solaris Verified boot's use of libgcrypt functions in multiboot.
 */

#ifndef _GCRYPT_H
#define	_GCRYPT_H

#include <sys/types.h>	/* size_t */
/* gpg_err_code_t, gpg_error_t, gcry_error_t, gcry_err_code_t:  */
#include <gpg-error.h>
#include <string.h>	/* strcasecmp() */

#ifdef	__cplusplus
extern "C" {
#endif

/* Different formats of external big integer representation. */
enum gcry_mpi_format {
	GCRYMPI_FMT_NONE = 0,
	GCRYMPI_FMT_STD = 1, /* Twos complement stored without length. */
	GCRYMPI_FMT_PGP = 2, /* As used by OpenPGP (unsigned only). */
	GCRYMPI_FMT_SSH = 3, /* As used by SSH (like STD but with length). */
	GCRYMPI_FMT_HEX = 4, /* Hex format. */
	GCRYMPI_FMT_USG = 5  /* Like STD but unsigned. */
};

/* Flags used for creating big integers. */
enum gcry_mpi_flag {
	GCRYMPI_FLAG_SECURE = 1, /* Allocate the number in "secure" memory. */
	/*
	 * The number is not a real one but just a way to store some bytes.
	 * This is useful for encrypted big integers.
	 */
	GCRYMPI_FLAG_OPAQUE = 2
};

typedef enum gcry_random_level {
	GCRY_WEAK_RANDOM = 0,
	GCRY_STRONG_RANDOM = 1,
	GCRY_VERY_STRONG_RANDOM = 2
} gcry_random_level_t;

enum gcry_ctl_cmds {
	GCRYCTL_DUMMY_UNUSED
};

typedef struct gcry_md_handle {
	void *dummy_unused;
} *gcry_md_hd_t;

typedef struct gcry_sexp {
	void *dummy_unused;
} *gcry_sexp_t;


struct gcry_mpi;	/* defined in mpi.h */
typedef struct gcry_mpi *gcry_mpi_t;

/* gcrypt_util.c */
void *_gcry_malloc(size_t n);
void *_gcry_calloc(size_t n, size_t m);
void _gcry_free(void *p);
void *_gcry_xrealloc(void *p, size_t n);
void _gcry_check_heap(const void *p);
void *_gcry_xmalloc(size_t n);
void *_gcry_malloc_secure(size_t n);
void *_gcry_xmalloc_secure(size_t n);
void *_gcry_xcalloc(size_t n, size_t m);
void *_gcry_xcalloc_secure(size_t n, size_t m);
int _gcry_is_secure(const void *a);
void _gcry_log_bug(const char *fmt, ...);
void _gcry_log_printf(const char *fmt, ...);
void _gcry_assert_failed(const char *expr, const char *file, int line,
    const char *func);
void *_gcry_random_bytes(size_t nbytes, enum gcry_random_level level);
void *_gcry_random_bytes_secure(size_t nbytes, enum gcry_random_level level);
void _gcry_create_nonce(void *buffer, size_t length);

/* Gcrypt MPI functions used by Verified Boot */
gcry_mpi_t gcry_mpi_new(unsigned int nbits);
void gcry_mpi_release(gcry_mpi_t a);
int gcry_mpi_cmp(const gcry_mpi_t u, const gcry_mpi_t v);
void gcry_mpi_powm(gcry_mpi_t w, const gcry_mpi_t b, const gcry_mpi_t e,
	const gcry_mpi_t m);
gcry_error_t gcry_mpi_scan(gcry_mpi_t *ret_mpi, enum gcry_mpi_format format,
	const void *buffer, size_t buflen, size_t *nscanned);

#ifdef	__cplusplus
}
#endif

#endif	/* _GCRYPT_H */
