/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2010,2011  Free Software Foundation, Inc.
 *  Copyright (c) 2012, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/net/netbuff.h>
#include <grub/dl.h>
#include <grub/net.h>
#include <grub/time.h>
#include <grub/term.h>
#include <grub/efi/api.h>
#include <grub/efi/efi.h>
#include <grub/i18n.h>

GRUB_MOD_LICENSE ("GPLv3+");

/* GUID.  */
static grub_efi_guid_t mnp_guid = GRUB_EFI_MANAGED_NETWORK_GUID;
static grub_efi_guid_t mnsbp_guid = GRUB_EFI_MANAGED_NETWORK_SB_GUID;
static grub_efi_guid_t pxe_io_guid = GRUB_EFI_PXE_GUID;

#define	EFINET_FILTER_UNICAST	1
#define	EFINET_FILTER_BROADCAST	1
#define	EFINET_FILTER_MULTICAST	0
#define	EFINET_FILTER_PROMISC	0
#define	EFINET_FILTER_ETHPROTO	0

#define	GRUB_EFI_RX_RING_SIZE	40
#define	GRUB_EFI_TX_RING_SIZE	20

struct grub_efi_mnp_rxtx {
  volatile int ready;
  volatile grub_efi_managed_network_completion_token_t token;
  void * event_data;
};

struct grub_efinet_dev_data
{
  struct grub_efi_managed_network *efi_mnet;
  grub_efi_handle_t efi_handle;
  grub_efi_managed_network_binding_t *efi_mnet_binding;
  grub_efi_handle_t efi_mnet_handle;

  struct grub_efi_mnp_rxtx efi_rx_ring[GRUB_EFI_RX_RING_SIZE];
  unsigned efi_rx_next;
  unsigned efi_rx_count;

  struct grub_efi_mnp_rxtx efi_tx_ring[GRUB_EFI_TX_RING_SIZE];
  unsigned efi_tx_next;
  unsigned efi_tx_count;
};

static void
efinet_signal_completion(grub_efi_event_t __attribute__((unused)) event,
  void *ctx)
{
  /*
   * In this case, ctx is just a pointer to an int, so we just
   * need to set to to 1 to indicate that this packet is ready.
   */
  *(int *)ctx = 1;
}

static void
efinet_recycle_tx_token(struct grub_efinet_dev_data *dev,
  struct grub_efi_mnp_rxtx *next_tx)
{
  grub_efi_status_t st;

  /* All done; prepare the tx token for reuse */
  grub_memset(next_tx->token.packet.tx_data, 0,
    sizeof (struct grub_efi_managed_network_xmit_data));
  next_tx->ready = 0;
  dev->efi_tx_next = (dev->efi_tx_next + 1) % dev->efi_tx_count;
  /*
   * Before the token can be reused, we need to deallocate the event
   * associated with it, then allocate a new event.
   */
  grub_efi_close_event(next_tx->token.event, &next_tx->event_data);
  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, efinet_signal_completion, (void **)&next_tx->ready,
    &next_tx->event_data, (grub_efi_event_t *)&next_tx->token.event);
  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("efinet", "Could not create tx token event: 0x%lx.\n",
	st);
      /*
       * Not a fatal error at this point, but if it keeps happening, it
       * will be, since we won't have any tx tokens for the tx queue.
       */
    }
}

static int
setup_filters_and_configure(grub_efi_managed_network_t *mnet)
{
   grub_efi_status_t st;
   grub_efi_managed_net_config_data_t mnet_cfg;
   int mod = 0;

   st = efi_call_3(mnet->get_mode_data, mnet, &mnet_cfg, 0);
   switch (st)
     {
     case GRUB_EFI_NOT_STARTED:
       /*FALLTHRU*/
     case GRUB_EFI_SUCCESS:
       break;
     default:
       grub_dprintf("efinet", "mnp->get_mode_data failed: 0x%lx\n",
	st);
       return -1;
     }

#define	CMP_AND_SET(x, y, z) \
	if ((x) != (y)) { \
	   (x) = (y); \
	   (z) = 1; \
	}

   CMP_AND_SET(mnet_cfg.protocol_type_filter, EFINET_FILTER_ETHPROTO, mod);
   CMP_AND_SET(mnet_cfg.enable_ucast_rcv, EFINET_FILTER_UNICAST, mod);
   CMP_AND_SET(mnet_cfg.enable_mcast_rcv, EFINET_FILTER_MULTICAST, mod);
   CMP_AND_SET(mnet_cfg.enable_bcast_rcv, EFINET_FILTER_BROADCAST, mod);
   CMP_AND_SET(mnet_cfg.enable_promisc_rcv, EFINET_FILTER_PROMISC, mod);
   CMP_AND_SET(mnet_cfg.flushqs_onreset, 0, mod);
   /* Disable background polling-- GRUB will poll manually */
   CMP_AND_SET(mnet_cfg.disable_bg_polling, 1, mod);

   if (mod)
     {
       grub_dprintf("efinet", "Configuring MNP instance...\n");
       st = efi_call_2(mnet->configure, mnet, &mnet_cfg);
       if (st != GRUB_EFI_SUCCESS)
         {
	   grub_dprintf("efinet", "mnp->configure failed: 0x%lx\n", st);
	   return -1;
         }
     }

   return 0;
}

static int
efinet_mnp_poll(grub_efi_managed_network_t *mnet)
{
  grub_efi_status_t st;

  st = efi_call_1(mnet->poll, mnet);
  if (st == GRUB_EFI_NOT_READY) /* No data */
    return 0;
  else if (st == GRUB_EFI_TIMEOUT)
    {
      grub_dprintf("efinet", "MNP->Poll indicates packet drops.\n");
      /* We still may have packets that were received, so fall through */
    }
  else if (st == GRUB_EFI_DEVICE_ERROR)
    {
      grub_dprintf("efinet", "MNP->Poll device error.\n");
      /* Reconfigure the interface since it was reset to defaults */
      (void) setup_filters_and_configure(mnet);
      return -1;
    }
  else if (st == GRUB_EFI_NOT_STARTED)
    {
      grub_dprintf("efinet", "MNP->Poll => instance not started.\n");
      return -1;
    }

  return 1;
}

static grub_err_t
send_card_buffer (struct grub_net_card *netcard,
		  struct grub_net_buff *pack)
{
  struct grub_efinet_dev_data *dev = (struct grub_efinet_dev_data *)netcard->data;
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_uint64_t limit_time = grub_get_time_ms () + 4000;
  grub_efi_managed_network_xmit_data_t *txdatap;
  struct grub_efi_mnp_rxtx *next_tx;

  if (! mnet)
    return grub_error (GRUB_ERR_IO, "EFI network not initialized");

  next_tx = &dev->efi_tx_ring[dev->efi_tx_next];

  txdatap = next_tx->token.packet.tx_data;

  txdatap->hdr_len = 14; /* Ethernet header length (XXX VLANs?) */
  txdatap->data_len = (pack->tail - pack->data) - txdatap->hdr_len;
  txdatap->frag_cnt = 1;
  txdatap->frag_table[0].frag_len = txdatap->data_len + txdatap->hdr_len;
  txdatap->frag_table[0].frag_buf = pack->data;

  st = efi_call_2(mnet->transmit, mnet, &next_tx->token);
  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("efinet", "Error transmitting: 0x%lx.\n", st);
      return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));
    }

  while (1)
    {
      if (efinet_mnp_poll(mnet) < 0)
	return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));

      if (next_tx->ready)
        {
	  if (next_tx->token.status == GRUB_EFI_SUCCESS)
	    {
	      efinet_recycle_tx_token(dev, next_tx);
	      return GRUB_ERR_NONE;
	    }
	  else
	    {
	      grub_dprintf("efinet", "Problem with tx: 0x%lx.\n",
		next_tx->token.status);
	      efinet_recycle_tx_token(dev, next_tx);
	      return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));
	    }
	}
      
      if (limit_time < grub_get_time_ms ())
	return grub_error (GRUB_ERR_TIMEOUT, N_("couldn't send network packet"));
    }

  return 0;
}

static void
efinet_recycle_rx_token(struct grub_efinet_dev_data *dev,
  struct grub_efi_mnp_rxtx *next_rx)
{
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_efi_managed_network_receive_data_t *rxdatap;

  rxdatap = next_rx->token.packet.rx_data;

  /* All done; release the rx token and stick it back on the rx queue */
  if (rxdatap != NULL)
    grub_efi_signal_event(rxdatap->recycle_event);
  next_rx->token.packet.rx_data = 0;
  next_rx->ready = 0;
  dev->efi_rx_next = (dev->efi_rx_next + 1) % dev->efi_rx_count;
  /*
   * Before the token can be reused, we need to deallocate the event
   * associated with it, then allocate a new event.
   */
  grub_efi_close_event(next_rx->token.event, &next_rx->event_data);
  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, efinet_signal_completion, (void **)&next_rx->ready,
    &next_rx->event_data, (grub_efi_event_t *)&next_rx->token.event);
  if (st == GRUB_EFI_SUCCESS)
    {
      st = efi_call_2(mnet->receive, mnet, &next_rx->token);
      if (st != GRUB_EFI_SUCCESS)
        {
          grub_dprintf("efinet", "Failed to recycle rx token: 0x%lx.\n",
	    st);
          /*
           * Not a fatal error at this point, but if it keeps happening, it
           * will be, since we won't have any rx tokens in the rx queue.
           */
        }
    }
  else
    {
      grub_dprintf("efinet", "Could not create rx token event: 0x%lx.\n",
	st);
      /*
       * Not a fatal error at this point, but if it keeps happening, it
       * will be, since we won't have any rx tokens in the rx queue.
       */
    }

    /* Zero out the received data pointer */
    next_rx->token.packet.rx_data = NULL;
}

static struct grub_net_buff *
get_card_packet (struct grub_net_card *netcard)
{
  struct grub_efinet_dev_data *dev = (struct grub_efinet_dev_data *)netcard->data;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_err_t err;
  struct grub_net_buff *nb;
  grub_efi_managed_network_receive_data_t *rxdatap;
  struct grub_efi_mnp_rxtx *next_rx;

  if (! mnet)
    return NULL;

  if (efinet_mnp_poll(mnet) < 0)
    return NULL;

  next_rx = &dev->efi_rx_ring[dev->efi_rx_next];
  if (! next_rx->ready)
    return NULL;

  rxdatap = next_rx->token.packet.rx_data;

  if (next_rx->token.status != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("efinet", "Error with signalled rx token: 0x%lx.\n",
        next_rx->token.status);
      efinet_recycle_rx_token(dev, next_rx);
      return NULL;
    }
  else if (rxdatap == NULL)
    {
      grub_dprintf("efinet", "Rx data pointer is NULL!\n");
      efinet_recycle_rx_token(dev, next_rx);
      return NULL;
    }

  /* Copy packet from rx token */
  nb = grub_netbuff_alloc(rxdatap->packet_len + 4 /* for alignment */);
  if (!nb)
    {
      grub_dprintf("efinet", "Couldn't allocate 0x%x bytes for packet.\n",
	rxdatap->packet_len);
      return NULL;
    }

  /* Reserve 2 bytes so that 2 + 14/18 bytes of ethernet header is divisible
     by 4. So that IP header is aligned on 4 bytes. */
  grub_netbuff_reserve (nb, 2);
  if (!nb)
    {
      grub_netbuff_free (nb);
      return NULL;
    }

  grub_memcpy(nb->data, rxdatap->media_hdr, rxdatap->packet_len);

  err = grub_netbuff_put (nb, rxdatap->packet_len);
  if (err)
    {
      grub_netbuff_free (nb);
      return NULL;
    }

  efinet_recycle_rx_token(dev, next_rx);

  return nb;
}

static void
efinet_close (struct grub_net_card *netcard)
{
  struct grub_efinet_dev_data *dev = (struct grub_efinet_dev_data *)netcard->data;
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet;
  grub_efi_managed_network_binding_t *mnetb;

  if (dev == NULL)
    return;

  mnet = dev->efi_mnet;
  mnetb = dev->efi_mnet_binding;

  if (mnet && mnetb)
    {
      st = efi_call_2(mnetb->destroy_child, mnetb, dev->efi_mnet_handle);
      if (st != GRUB_EFI_SUCCESS)
        grub_dprintf("net", "Error destroying MNP child %s: 0x%lx\n",
	  netcard->name, (long)st);
      else
       {
	 dev->efi_mnet = 0;
	 dev->efi_mnet_binding = 0;
	 dev->efi_mnet_handle = 0;
	 dev->efi_handle = 0;
       }

    }
  grub_free(dev);
  netcard->data = 0;
}

static struct grub_net_card_driver efidriver =
  {
    .name = "efinet",
    .close = efinet_close,
    .send = send_card_buffer,
    .recv = get_card_packet
  };

grub_efi_handle_t
grub_efinet_get_device_handle (struct grub_net_card *card)
{
  struct grub_efinet_dev_data *dev;
  if (!card || !card->data || card->driver != &efidriver)
    return 0;
  dev = (struct grub_efinet_dev_data *)card->data;
  return dev->efi_handle;
}

static int
efinet_derive_mac_from_handle(grub_efi_handle_t handle,
  grub_efi_mac_address_t *macp)
{
  grub_efi_device_path_t *dp;

  dp = grub_efi_get_device_path (handle);
  if (! dp)
    {
      grub_dprintf("efinet", "derive_mac: Couldn't get device path.\n");
      return -1;
    }

  /* Parse the device path. Currently, we only support Ethernet NICs */
  while (1)
    {
      grub_efi_uint8_t type = GRUB_EFI_DEVICE_PATH_TYPE (dp);
      grub_efi_uint8_t subtype = GRUB_EFI_DEVICE_PATH_SUBTYPE (dp);
      grub_efi_uint16_t len = GRUB_EFI_DEVICE_PATH_LENGTH (dp);

      if (type == GRUB_EFI_MESSAGING_DEVICE_PATH_TYPE &&
          subtype == GRUB_EFI_MAC_ADDRESS_DEVICE_PATH_SUBTYPE)
        {
	  grub_efi_mac_address_device_path_t mac;
	  grub_memcpy (&mac, dp, len);
	  grub_dprintf ("efinet",
	    "/MacAddr(%02x:%02x:%02x:%02x:%02x:%02x,%x)\n",
		(unsigned) mac.mac_address[0],
		(unsigned) mac.mac_address[1],
		(unsigned) mac.mac_address[2],
		(unsigned) mac.mac_address[3],
		(unsigned) mac.mac_address[4],
		(unsigned) mac.mac_address[5],
		(unsigned) mac.if_type);

	  /* XXX- Assumption is this is Ethernet */

	  grub_memcpy(macp, &mac.mac_address[0],
	    sizeof(grub_efi_mac_address_t));
	  return 0;
	}

      if (GRUB_EFI_END_ENTIRE_DEVICE_PATH (dp))
	break;

      dp = (grub_efi_device_path_t *) ((char *) dp + len);
    }

  /* MAC address Not found */
  return -1;  
}

static int
efinet_init_rxtx_entry(struct grub_efi_mnp_rxtx *xp)
{
  grub_efi_status_t st;

  grub_memset(xp, 0, sizeof (struct grub_efi_mnp_rxtx));

  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, efinet_signal_completion, (void **)&xp->ready,
    &xp->event_data, (grub_efi_event_t *)&xp->token.event);

  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("efinet", "error initting rxtx token: 0x%lx.\n",
        st);
      return -1;
    }

  return 0;
}

static void
efinet_init_rx_ring(struct grub_efinet_dev_data *dev)
{
  grub_efi_status_t st;
  int i;

  dev->efi_rx_next = 0;
  dev->efi_rx_count = 0;

  for (i = 0; i < GRUB_EFI_RX_RING_SIZE; i++)
    {
      if (efinet_init_rxtx_entry(&dev->efi_rx_ring[dev->efi_rx_count]) != 0)
        {
          grub_dprintf("efinet", "error initting rx item; %sproceeding "
	    "with %d rx tokens.\n", dev->efi_rx_count == 0 ? "not " : "",
	    dev->efi_rx_count);
          break;
        }

      /* Add the token to the receive queue */
      st = efi_call_2(dev->efi_mnet->receive, dev->efi_mnet,
	&dev->efi_rx_ring[dev->efi_rx_count].token);
      if (st == GRUB_EFI_SUCCESS)
        dev->efi_rx_count++;
      else
        grub_dprintf("efinet", "failed to add token to rx: 0x%lx.\n", st);
    }
}

static void
efinet_init_tx_ring(struct grub_efinet_dev_data *dev)
{
  int i;

  dev->efi_tx_next = 0;
  dev->efi_tx_count = 0;

  for (i = 0; i < GRUB_EFI_TX_RING_SIZE; i++)
    {
      void *txdata;

      if (efinet_init_rxtx_entry(&dev->efi_tx_ring[dev->efi_tx_count]) != 0)
        {
          grub_dprintf("efinet", "error initting tx item; %sproceeding "
	    "with %d tx tokens.\n", dev->efi_tx_count == 0 ? "not " : "",
	    dev->efi_tx_count);
          break;
        }

      /*
       * Actual allocation size of the MNP transmit data includes a single
       * fragment ONLY.
       */
      txdata = grub_zalloc(sizeof (struct grub_efi_managed_network_xmit_data));
      if (txdata == NULL)
	{
	  grub_dprintf("efinet", "Failed to allocate txdata structure.\n");
	  continue;
	}

      dev->efi_tx_ring[dev->efi_tx_count].token.packet.tx_data = txdata;

      dev->efi_tx_count++;
    }
}

static void
grub_efinet_findcards (void)
{
  grub_efi_uintn_t num_handles;
  grub_efi_handle_t *handles;
  grub_efi_handle_t *handle;
  grub_efi_handle_t mnet_handle;
  grub_efi_managed_network_binding_t *mnetb;
  int i = 0;
  grub_efi_status_t status;

  /* Find handles which support the managed net binding interface.  */
  handles = grub_efi_locate_handle (GRUB_EFI_BY_PROTOCOL, &mnsbp_guid,
				    0, &num_handles);
  if (! handles)
    return;
  for (handle = handles; num_handles--; handle++)
    {
      struct grub_efinet_dev_data *dev;
      struct grub_efi_simple_network_mode snp_mode;
      struct grub_net_card *card;
      grub_efi_uint32_t mtu;
      grub_efi_mac_address_t mac;
      grub_efi_managed_network_t *mnet;

      mnetb = grub_efi_open_protocol (*handle, &mnsbp_guid,
	GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
      if (mnetb == 0)
	{
	  grub_dprintf("efinet", "Could not open the MNPBP protocol.\n");
	  continue;
	}

      /* Must be a NULL pointer for create_child to create a handle */
      mnet_handle = 0;

      status = efi_call_2(mnetb->create_child, mnetb, &mnet_handle);

      if (status != GRUB_EFI_SUCCESS || mnet_handle == 0)
	{
	  grub_dprintf("efinet", "Could not create a new MNP child: st=0x%lx "
	    "mnet_handle=%p\n", status, mnet_handle);
	  continue;
	}

      mnet = grub_efi_open_protocol (mnet_handle, &mnp_guid,
	GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
      if (mnet == 0)
	{
	  grub_dprintf("efinet", "Could not get MNP protocol.\n");
	  status = efi_call_2(mnetb->destroy_child, mnetb, mnet_handle);
	  grub_dprintf("efinet", "MNP->destroy_child => 0x%lx\n",
	    status);
	  continue;
	}

      /* Bring up the interface */
      setup_filters_and_configure(mnet);

      /*
       * Try to get the SNP mode data for the MTU and MAC.  If they're
       * not available (since the spec says it's optional), we'll have to
       * punt on the MTU size, but we can derive the MAC from the device
       * path.
       */
      status = efi_call_3(mnet->get_mode_data, mnet, 0, &snp_mode);
      if (status == GRUB_EFI_SUCCESS)
	{
	  mtu = snp_mode.max_packet_size;
	  grub_memcpy(mac, snp_mode.current_address,
	    sizeof (grub_efi_mac_address_t));
	}
      else
	{
	  grub_dprintf("efinet", "MNP->get_mode_data failed: 0x%lx\n",
	    status);
	  mtu = 1500; /* PUNT */
	  if (efinet_derive_mac_from_handle(*handle, &mac) != 0)
	    {
		  /*
		   * Something went wrong getting the MAC; abort configuring
		   * this instance.
		   */
	      	  status = efi_call_2(mnetb->destroy_child, mnetb, mnet_handle);
		  grub_dprintf("efinet", "MNP->destroy_child => 0x%lx\n",
		    status);
		  continue;
	    }
	}

      card = grub_zalloc (sizeof (struct grub_net_card));
      if (!card)
	{
	  grub_print_error ();
	  grub_free (handles);
	  return;
	}

      dev = grub_zalloc(sizeof (struct grub_efinet_dev_data)); 
      if (!dev)
	{
	  grub_print_error ();
	  grub_free(card);
	  grub_free (handles);
	  return;
	}

      card->name = grub_xasprintf ("efinet%d", i++);
      card->driver = &efidriver;
      card->flags = 0;
      card->default_address.type = GRUB_NET_LINK_LEVEL_PROTOCOL_ETHERNET;

      card->mtu = mtu;
      grub_memcpy (card->default_address.mac, mac,
		   sizeof (card->default_address.mac));

      dev->efi_mnet = mnet;
      dev->efi_mnet_binding = mnetb;
      dev->efi_mnet_handle = mnet_handle;
      dev->efi_handle = *handle;

      efinet_init_rx_ring(dev);
      efinet_init_tx_ring(dev);

      if (dev->efi_rx_count == 0 || dev->efi_tx_count == 0)
        {
	  efinet_close(card);
	  grub_free(dev);
          grub_free((char *)card->name);
          grub_free(card);
        }
      else
        {
	  card->data = dev;
	  grub_net_card_register (card);
        }
    }
  grub_free (handles);
}

/*
 * This function assumes that card->efi_mnet was already initialized to
 * a protocol handle for MNP.
 */
static void
grub_efi_net_config_real (grub_efi_handle_t hnd, char **device,
			  char **path)
{
  struct grub_net_card *card;
  grub_efi_mac_address_t mac;

  /*
   * Extract the MAC address from the device path and use it to match the
   * efinet card below.
   */
  if (efinet_derive_mac_from_handle(hnd, &mac) != 0)
    return;

  FOR_NET_CARDS (card)
  {
    struct grub_efi_pxe *pxe;
    volatile struct grub_efi_pxe_mode *pxe_mode;
    grub_efi_status_t st;

    if (card->driver != &efidriver)
      continue;

    grub_dprintf("efinet", "Matching %x:%x:%x:%x:%x:%x with %x:%x:%x:%x:%x:%x\n",
	 mac[0], mac[1], mac[2], mac[3], mac[4], mac[5],
	 card->default_address.mac[0],
	 card->default_address.mac[1],
	 card->default_address.mac[2],
	 card->default_address.mac[3],
	 card->default_address.mac[4],
	 card->default_address.mac[5]);

    if (grub_memcmp(&mac, &card->default_address.mac,
	sizeof(card->default_address.mac)) != 0)
      continue;

    pxe = grub_efi_open_protocol (hnd, &pxe_io_guid,
				  GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
    if (! pxe)
      return;

    pxe_mode = pxe->mode;
    grub_net_configure_by_dhcp_ack (card->name, card, 0,
				    (struct grub_net_bootp_packet *)
				    &pxe_mode->dhcp_ack,
				    sizeof (pxe_mode->dhcp_ack),
				    1, device, path);

    /*
     * Stop the PXEBC protocol for this handle.  This prevents performance
     * issues and protocol-level duplication (e.g. ARP replies, ICMP errors)
     * [[Note this only done on the boot interface; if there are PXEBC
     * instances associated with other active efinet instances, they may
     * experience these negative effects, and this code may need revisiting.
     * As of now, though, there are no implementations that kick off PXEBC
     * on instances of NICs other than the one chosen for boot.]]
     */
    st = efi_call_1(pxe->stop, pxe);
    if (st != GRUB_EFI_SUCCESS)
      grub_dprintf("efinet", "Warning: Failed to stop UEFI PXE Base Code Protocol for MAC "
	"%x:%x:%x:%x:%x:%x.\n\tNetwork performance may be affected.\n",
	mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

    return;
  }
}

static grub_err_t
grub_efinet_fini_hw (int noreturn __attribute__ ((unused)))
{
  struct grub_net_card *card;
  FOR_NET_CARDS (card)
    if (card->driver == &efidriver)
      efinet_close (card);
  return GRUB_ERR_NONE;
}

static grub_err_t
grub_efinet_restore_hw (void)
{
  return GRUB_ERR_NONE;
}

static struct grub_preboot *efinet_fini_hnd;

GRUB_MOD_INIT(efinet)
{
  grub_efinet_findcards ();

  grub_efi_net_config = grub_efi_net_config_real;

  efinet_fini_hnd = grub_loader_register_preboot_hook (grub_efinet_fini_hw,
                                                       grub_efinet_restore_hw,
                                                       GRUB_LOADER_PREBOOT_HOOK_PRIO_DISK);
}

GRUB_MOD_FINI(efinet)
{
  struct grub_net_card *card, *next;

  FOR_NET_CARDS_SAFE (card, next)
    if (card->driver == &efidriver)
      grub_net_card_unregister (card);

  grub_loader_unregister_preboot_hook (efinet_fini_hnd);
}
