/* wbefinet - WANBoot-specific version of efinet.c */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2010,2011  Free Software Foundation, Inc.
 *  Copyright (c) 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/net/netbuff.h>
#include <grub/dl.h>
#include <grub/net.h>
#include <grub/time.h>
#include <grub/term.h>
#include <grub/efi/api.h>
#include <grub/efi/efi.h>
#include <grub/efi/wanboot.h>
#include <grub/i18n.h>

GRUB_MOD_LICENSE ("GPLv3+");

/* GUID.  */
static grub_efi_guid_t mnp_guid = GRUB_EFI_MANAGED_NETWORK_GUID;
static grub_efi_guid_t mnsbp_guid = GRUB_EFI_MANAGED_NETWORK_SB_GUID;
static grub_efi_guid_t wanboot_guid = WANBOOT_GUID;

#define	EFINET_FILTER_UNICAST	1
#define	EFINET_FILTER_BROADCAST	1
#define	EFINET_FILTER_MULTICAST	0
#define	EFINET_FILTER_PROMISC	0
#define	EFINET_FILTER_ETHPROTO	0

#define	GRUB_EFI_RX_RING_SIZE	40
#define	GRUB_EFI_TX_RING_SIZE	20

struct grub_efi_mnp_rxtx {
  volatile int ready;
  volatile grub_efi_managed_network_completion_token_t token;
  void * event_data;
};

struct grub_wbefinet_dev_data
{
  struct grub_efi_managed_network *efi_mnet;
  grub_efi_handle_t efi_handle;
  grub_efi_managed_network_binding_t *efi_mnet_binding;
  grub_efi_handle_t efi_mnet_handle;

  struct grub_efi_mnp_rxtx efi_rx_ring[GRUB_EFI_RX_RING_SIZE];
  unsigned efi_rx_next;
  unsigned efi_rx_count;

  struct grub_efi_mnp_rxtx efi_tx_ring[GRUB_EFI_TX_RING_SIZE];
  unsigned efi_tx_next;
  unsigned efi_tx_count;
};

static void
wbefinet_signal_completion(grub_efi_event_t __attribute__((unused)) event,
  void *ctx)
{
  /*
   * In this case, ctx is just a pointer to an int, so we just
   * need to set to to 1 to indicate that this packet is ready.
   */
  *(int *)ctx = 1;
}

static void
wbefinet_recycle_tx_token(struct grub_wbefinet_dev_data *dev,
  struct grub_efi_mnp_rxtx *next_tx)
{
  grub_efi_status_t st;

  /* All done; prepare the tx token for reuse */
  grub_memset(next_tx->token.packet.tx_data, 0,
    sizeof (struct grub_efi_managed_network_xmit_data));
  next_tx->ready = 0;
  dev->efi_tx_next = (dev->efi_tx_next + 1) % dev->efi_tx_count;
  /*
   * Before the token can be reused, we need to deallocate the event
   * associated with it, then allocate a new event.
   */
  grub_efi_close_event(next_tx->token.event, &next_tx->event_data);
  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, wbefinet_signal_completion, (void **)&next_tx->ready,
    &next_tx->event_data, (grub_efi_event_t *)&next_tx->token.event);
  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("wbefinet", "Could not create tx token event: 0x%lx.\n",
	st);
      /*
       * Not a fatal error at this point, but if it keeps happening, it
       * will be, since we won't have any tx tokens for the tx queue.
       */
    }
}

static int
setup_filters_and_configure(grub_efi_managed_network_t *mnet)
{
   grub_efi_status_t st;
   grub_efi_managed_net_config_data_t mnet_cfg;
   int mod = 0;

   st = efi_call_3(mnet->get_mode_data, mnet, &mnet_cfg, 0);
   switch (st)
     {
     case GRUB_EFI_NOT_STARTED:
       /*FALLTHRU*/
     case GRUB_EFI_SUCCESS:
       break;
     default:
       grub_dprintf("wbefinet", "mnp->get_mode_data failed: 0x%lx\n",
	st);
       return -1;
     }

#define	CMP_AND_SET(x, y, z) \
	if ((x) != (y)) { \
	   (x) = (y); \
	   (z) = 1; \
	}

   CMP_AND_SET(mnet_cfg.protocol_type_filter, EFINET_FILTER_ETHPROTO, mod);
   CMP_AND_SET(mnet_cfg.enable_ucast_rcv, EFINET_FILTER_UNICAST, mod);
   CMP_AND_SET(mnet_cfg.enable_mcast_rcv, EFINET_FILTER_MULTICAST, mod);
   CMP_AND_SET(mnet_cfg.enable_bcast_rcv, EFINET_FILTER_BROADCAST, mod);
   CMP_AND_SET(mnet_cfg.enable_promisc_rcv, EFINET_FILTER_PROMISC, mod);
   CMP_AND_SET(mnet_cfg.flushqs_onreset, 0, mod);
   /* Disable background polling-- GRUB will poll manually */
   CMP_AND_SET(mnet_cfg.disable_bg_polling, 1, mod);

   if (mod)
     {
       grub_dprintf("wbefinet", "Configuring MNP instance...\n");
       st = efi_call_2(mnet->configure, mnet, &mnet_cfg);
       if (st != GRUB_EFI_SUCCESS)
         {
	   grub_dprintf("wbefinet", "mnp->configure failed: 0x%lx\n", st);
	   return -1;
         }
     }

   return 0;
}

static int
wbefinet_mnp_poll(grub_efi_managed_network_t *mnet)
{
  grub_efi_status_t st;

  st = efi_call_1(mnet->poll, mnet);
  if (st == GRUB_EFI_NOT_READY) /* No data */
    return 0;
  else if (st == GRUB_EFI_TIMEOUT)
    {
      grub_dprintf("wbefinet", "MNP->Poll indicates packet drops.\n");
      /* We still may have packets that were received, so fall through */
    }
  else if (st == GRUB_EFI_DEVICE_ERROR)
    {
      grub_dprintf("wbefinet", "MNP->Poll device error.\n");
      /* Reconfigure the interface since it was reset to defaults */
      (void) setup_filters_and_configure(mnet);
      return -1;
    }
  else if (st == GRUB_EFI_NOT_STARTED)
    {
      grub_dprintf("wbefinet", "MNP->Poll => instance not started.\n");
      return -1;
    }

  return 1;
}

static grub_err_t
send_card_buffer (struct grub_net_card *netcard,
		  struct grub_net_buff *pack)
{
  struct grub_wbefinet_dev_data *dev = (struct grub_wbefinet_dev_data *)netcard->data;
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_uint64_t limit_time = grub_get_time_ms () + 4000;
  grub_efi_managed_network_xmit_data_t *txdatap;
  struct grub_efi_mnp_rxtx *next_tx;

  if (! mnet)
    return grub_error (GRUB_ERR_IO, "EFI network not initialized");

  next_tx = &dev->efi_tx_ring[dev->efi_tx_next];

  txdatap = next_tx->token.packet.tx_data;

  txdatap->hdr_len = 14; /* Ethernet header length (XXX VLANs?) */
  txdatap->data_len = (pack->tail - pack->data) - txdatap->hdr_len;
  txdatap->frag_cnt = 1;
  txdatap->frag_table[0].frag_len = txdatap->data_len + txdatap->hdr_len;
  txdatap->frag_table[0].frag_buf = pack->data;

  st = efi_call_2(mnet->transmit, mnet, &next_tx->token);
  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("wbefinet", "Error transmitting: 0x%lx.\n", st);
      return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));
    }

  while (1)
    {
      if (wbefinet_mnp_poll(mnet) < 0)
	return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));

      if (next_tx->ready)
        {
	  if (next_tx->token.status == GRUB_EFI_SUCCESS)
	    {
	      wbefinet_recycle_tx_token(dev, next_tx);
	      return GRUB_ERR_NONE;
	    }
	  else
	    {
	      grub_dprintf("wbefinet", "Problem with tx: 0x%lx.\n",
		next_tx->token.status);
	      wbefinet_recycle_tx_token(dev, next_tx);
	      return grub_error (GRUB_ERR_IO, N_("couldn't send network packet"));
	    }
	}
      
      if (limit_time < grub_get_time_ms ())
	return grub_error (GRUB_ERR_TIMEOUT, N_("couldn't send network packet"));
    }

  return 0;
}

static void
wbefinet_recycle_rx_token(struct grub_wbefinet_dev_data *dev,
  struct grub_efi_mnp_rxtx *next_rx)
{
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_efi_managed_network_receive_data_t *rxdatap;

  rxdatap = next_rx->token.packet.rx_data;

  /* All done; release the rx token and stick it back on the rx queue */
  if (rxdatap != NULL)
    grub_efi_signal_event(rxdatap->recycle_event);
  next_rx->token.packet.rx_data = 0;
  next_rx->ready = 0;
  dev->efi_rx_next = (dev->efi_rx_next + 1) % dev->efi_rx_count;
  /*
   * Before the token can be reused, we need to deallocate the event
   * associated with it, then allocate a new event.
   */
  grub_efi_close_event(next_rx->token.event, &next_rx->event_data);
  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, wbefinet_signal_completion, (void **)&next_rx->ready,
    &next_rx->event_data, (grub_efi_event_t *)&next_rx->token.event);
  if (st == GRUB_EFI_SUCCESS)
    {
      st = efi_call_2(mnet->receive, mnet, &next_rx->token);
      if (st != GRUB_EFI_SUCCESS)
        {
          grub_dprintf("wbefinet", "Failed to recycle rx token: 0x%lx.\n",
	    st);
          /*
           * Not a fatal error at this point, but if it keeps happening, it
           * will be, since we won't have any rx tokens in the rx queue.
           */
        }
    }
  else
    {
      grub_dprintf("wbefinet", "Could not create rx token event: 0x%lx.\n",
	st);
      /*
       * Not a fatal error at this point, but if it keeps happening, it
       * will be, since we won't have any rx tokens in the rx queue.
       */
    }

    /* Zero out the received data pointer */
    next_rx->token.packet.rx_data = NULL;
}

static struct grub_net_buff *
get_card_packet (struct grub_net_card *netcard)
{
  struct grub_wbefinet_dev_data *dev = (struct grub_wbefinet_dev_data *)netcard->data;
  grub_efi_managed_network_t *mnet = dev->efi_mnet;
  grub_err_t err;
  struct grub_net_buff *nb;
  grub_efi_managed_network_receive_data_t *rxdatap;
  struct grub_efi_mnp_rxtx *next_rx;

  if (! mnet)
    return NULL;

  if (wbefinet_mnp_poll(mnet) < 0)
    return NULL;

  next_rx = &dev->efi_rx_ring[dev->efi_rx_next];
  if (! next_rx->ready)
    return NULL;

  rxdatap = next_rx->token.packet.rx_data;

  if (next_rx->token.status != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("wbefinet", "Error with signalled rx token: 0x%lx.\n",
        next_rx->token.status);
      wbefinet_recycle_rx_token(dev, next_rx);
      return NULL;
    }
  else if (rxdatap == NULL)
    {
      grub_dprintf("wbefinet", "Rx data pointer is NULL!\n");
      wbefinet_recycle_rx_token(dev, next_rx);
      return NULL;
    }

  /* Copy packet from rx token */
  nb = grub_netbuff_alloc(rxdatap->packet_len + 4 /* for alignment */);
  if (!nb)
    {
      grub_dprintf("wbefinet", "Couldn't allocate 0x%x bytes for packet.\n",
	rxdatap->packet_len);
      return NULL;
    }

  /* Reserve 2 bytes so that 2 + 14/18 bytes of ethernet header is divisible
     by 4. So that IP header is aligned on 4 bytes. */
  grub_netbuff_reserve (nb, 2);
  if (!nb)
    {
      grub_netbuff_free (nb);
      return NULL;
    }

  grub_memcpy(nb->data, rxdatap->media_hdr, rxdatap->packet_len);

  err = grub_netbuff_put (nb, rxdatap->packet_len);
  if (err)
    {
      grub_netbuff_free (nb);
      return NULL;
    }

  wbefinet_recycle_rx_token(dev, next_rx);

  return nb;
}

static void
wbefinet_close (struct grub_net_card *netcard)
{
  struct grub_wbefinet_dev_data *dev = (struct grub_wbefinet_dev_data *)netcard->data;
  grub_efi_status_t st;
  grub_efi_managed_network_t *mnet;
  grub_efi_managed_network_binding_t *mnetb;

  if (dev == NULL)
    return;

  mnet = dev->efi_mnet;
  mnetb = dev->efi_mnet_binding;

  if (mnet && mnetb)
    {
      st = efi_call_2(mnetb->destroy_child, mnetb, dev->efi_mnet_handle);
      if (st != GRUB_EFI_SUCCESS)
        grub_dprintf("net", "Error destroying MNP child %s: 0x%lx\n",
	  netcard->name, (long)st);
      else
       {
	 dev->efi_mnet = 0;
	 dev->efi_mnet_binding = 0;
	 dev->efi_mnet_handle = 0;
	 dev->efi_handle = 0;
       }

    }
  grub_free(dev);
  netcard->data = 0;
}

static struct grub_net_card_driver wbefidriver =
  {
    .name = "wbefinet",
    .close = wbefinet_close,
    .send = send_card_buffer,
    .recv = get_card_packet
  };

grub_efi_handle_t
grub_wbefinet_get_device_handle (struct grub_net_card *card)
{
  struct grub_wbefinet_dev_data *dev;
  if (!card || !card->data || card->driver != &wbefidriver)
    return 0;
  dev = (struct grub_wbefinet_dev_data *)card->data;
  return dev->efi_handle;
}

static int
wbefinet_init_rxtx_entry(struct grub_efi_mnp_rxtx *xp)
{
  grub_efi_status_t st;

  grub_memset(xp, 0, sizeof (struct grub_efi_mnp_rxtx));

  st = grub_efi_create_event(GRUB_EFI_EVT_NOTIFY_SIGNAL,
    GRUB_EFI_TPL_CALLBACK, wbefinet_signal_completion, (void **)&xp->ready,
    &xp->event_data, (grub_efi_event_t *)&xp->token.event);

  if (st != GRUB_EFI_SUCCESS)
    {
      grub_dprintf("wbefinet", "error initting rxtx token: 0x%lx.\n",
        st);
      return -1;
    }

  return 0;
}

static void
wbefinet_init_rx_ring(struct grub_wbefinet_dev_data *dev)
{
  grub_efi_status_t st;
  int i;

  dev->efi_rx_next = 0;
  dev->efi_rx_count = 0;

  for (i = 0; i < GRUB_EFI_RX_RING_SIZE; i++)
    {
      if (wbefinet_init_rxtx_entry(&dev->efi_rx_ring[dev->efi_rx_count]) != 0)
        {
          grub_dprintf("wbefinet", "error initting rx item; %sproceeding "
	    "with %d rx tokens.\n", dev->efi_rx_count == 0 ? "not " : "",
	    dev->efi_rx_count);
          break;
        }

      /* Add the token to the receive queue */
      st = efi_call_2(dev->efi_mnet->receive, dev->efi_mnet,
	&dev->efi_rx_ring[dev->efi_rx_count].token);
      if (st == GRUB_EFI_SUCCESS)
        dev->efi_rx_count++;
      else
        grub_dprintf("wbefinet", "failed to add token to rx: 0x%lx.\n", st);
    }
}

static void
wbefinet_init_tx_ring(struct grub_wbefinet_dev_data *dev)
{
  int i;

  dev->efi_tx_next = 0;
  dev->efi_tx_count = 0;

  for (i = 0; i < GRUB_EFI_TX_RING_SIZE; i++)
    {
      void *txdata;

      if (wbefinet_init_rxtx_entry(&dev->efi_tx_ring[dev->efi_tx_count]) != 0)
        {
          grub_dprintf("wbefinet", "error initting tx item; %sproceeding "
	    "with %d tx tokens.\n", dev->efi_tx_count == 0 ? "not " : "",
	    dev->efi_tx_count);
          break;
        }

      /*
       * Actual allocation size of the MNP transmit data includes a single
       * fragment ONLY.
       */
      txdata = grub_zalloc(sizeof (struct grub_efi_managed_network_xmit_data));
      if (txdata == NULL)
	{
	  grub_dprintf("wbefinet", "Failed to allocate txdata structure.\n");
	  continue;
	}

      dev->efi_tx_ring[dev->efi_tx_count].token.packet.tx_data = txdata;

      dev->efi_tx_count++;
    }
}

/*
 * get_wb_macaddr
 */
static int
get_wb_macaddr (grub_net_link_level_address_t *mac_addr)
{
  grub_size_t datasize;
  grub_efi_mac_address_t *data;

  data = grub_efi_get_variable (WB_BI_MACADDR, &wanboot_guid, &datasize);
  if (data == NULL || datasize < WB_MACADDR_LEN) {
    grub_memset(&mac_addr->mac, 0, sizeof (mac_addr->mac));
    return 0;
  }

  mac_addr->type = GRUB_NET_LINK_LEVEL_PROTOCOL_ETHERNET;
  grub_memcpy(&mac_addr->mac, data, sizeof (mac_addr->mac));

  return 1;
}

/*
 * get_wb_host_ip
 */
static int
get_wb_host_ip (grub_net_network_level_address_t *host_addr)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_HOST_IP, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  /*
   * At this time, IPv4 support only
   */
  host_addr->type = GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV4;
  grub_memcpy (&host_addr->ipv4, data, sizeof (host_addr->ipv4));

  return 1;
}

/*
 * get_wb_netmask
 */
static int
get_wb_netmask (int *netmask)
{
  grub_size_t datasize;
  unsigned char *data;
  grub_uint32_t invmask;
  int i;

  data = grub_efi_get_variable (WB_BI_SUBNET_MASK, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  if (datasize < 4)
    {
      *netmask = 0;
      return 0;
    }

  grub_memcpy (&invmask, data, sizeof (invmask));
  invmask = grub_be_to_cpu32 (invmask);
  invmask = ~invmask;

  i=0;
  while (invmask)
    {
      invmask >>= 1;
      i++;
    }

  *netmask = 32 - i;

  return 1;
}

/*
 * get_wb_router_ip
 */
static int
get_wb_router_ip (grub_net_network_level_address_t *router_addr)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_ROUTER_IP, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  /*
   * At this time, IPv4 support only
   */
  router_addr->type = GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV4;
  grub_memcpy (&router_addr->ipv4, data, sizeof (router_addr->ipv4));

  return 1;
}

/*
 * get_wb_server_ip
 */
static int
get_wb_server_ip (grub_net_network_level_address_t *server_addr)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_SERVER_IP, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  /*
   * At this time, IPv4 support only
   */
  server_addr->type = GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV4;
  grub_memcpy (&server_addr->ipv4, data, sizeof (server_addr->ipv4));

  return 1;
}

/*
 * get_wb_server_port
 */
static int
get_wb_server_port (grub_uint16_t *port)
{
  grub_size_t datasize;
  grub_uint16_t *data;

  data = grub_efi_get_variable (WB_BI_SERVER_PORT, &wanboot_guid, &datasize);
  if (data == NULL || datasize < sizeof (grub_uint16_t)) {
    *port = WB_DEF_SERVER_PORT;
    return 0;
  }

  *port = *data;

  return 1;
}

/*
 * get_wb_bootp_response
 * This is alternately called a DHCP ACK.
 */
static int
get_wb_bootp_response ( grub_uint8_t *bootp_resp,
			grub_size_t *bootp_resp_len )
{
  grub_size_t datasize;
  grub_uint16_t *data;

  data = grub_efi_get_variable (WB_BI_BOOTP_RESPONSE, &wanboot_guid, &datasize);
  if (data == NULL || datasize == 0 || datasize > *bootp_resp_len)
    {
      grub_memset(bootp_resp, 0, *bootp_resp_len);
      *bootp_resp_len = 0;
      return 0;
    }

  /*
   * The bootp response consists of the base information described in struct
   * grub_net_bootp_packet plus additional information, much of which are
   * "options" not included in the struct.
   *
   * The bootp response is never looked at in this module; it is just copied
   * into a Multiboot2 NETWORK tag at some point.
   */

  grub_memcpy (bootp_resp, data, datasize);
  *bootp_resp_len = datasize;

  return 1;
}

static void
grub_wbefinet_findcards (void)
{
  grub_efi_uintn_t num_handles;
  grub_efi_handle_t *handles;
  grub_efi_handle_t *handle;
  grub_net_link_level_address_t mac_addr;
  grub_net_network_level_address_t server_ip_addr;
  grub_net_network_level_netaddress_t host_ip_addr;
  grub_net_network_level_address_t router_ip_addr;
  grub_uint8_t *bootp_resp;
  grub_size_t bootp_resp_len;
  grub_efi_mac_address_t mac;
  grub_efi_uint32_t mtu = 0;
  struct grub_net_card *card;
  struct grub_wbefinet_dev_data *dev;
  grub_efi_status_t status;

  /* Get the WANBoot MAC address, used to find the correct MNP */
  if (! get_wb_macaddr (&mac_addr))
    {
      grub_printf ("Could not get the WANBoot MAC address\n");
      return;
    }

  /* Get the WANBoot host IP address */
  if (! get_wb_host_ip (&(host_ip_addr.addr)))
    {
      grub_printf ("Could not get the WANBoot host IP address\n");
      return;
    }

  /* Get the WANBoot subnet mask */
  if (! get_wb_netmask (&(host_ip_addr.masksize)))
    {
      grub_printf ("Could not get the WANBoot subnet mask\n");
      return;
    }

  /* Get the WANBoot router IP address */
  if (! get_wb_router_ip (&router_ip_addr))
    {
      grub_printf ("Could not get the WANBoot router IP address\n");
      return;
    }

  /* Get the WANBoot server IP address */
  if (! get_wb_server_ip (&server_ip_addr))
    {
      grub_printf ("Could not get the WANBoot server IP address\n");
      return;
    }

  /* Get the client bootp response (DHCP ACK) */

  bootp_resp_len = 576;		/* See RFC 791 for explanation of size */
  bootp_resp = grub_malloc (bootp_resp_len);
  if (! bootp_resp )
    {
      grub_printf ("Could not allocate the bootp response buffer\n");
      return;
    }

  if (! get_wb_bootp_response ( bootp_resp, &bootp_resp_len ))
    {
      /*
       * If there is no bootp_response is in EFI, make sure that the buffer
       * info is zeroed so that the Multiboot2 network info is not set
       */
      grub_free (bootp_resp);
      bootp_resp = NULL;
      bootp_resp_len = 0;
    }

  /*
   * The network configuration has been retreived from UEFI NVRAM
   * variables set by the WANBoot application.
   *
   * Next, walk through all of the Managed Network Handles and reset
   * them all.
   */

  /* Get the Managed Network handles */
  handles = grub_efi_locate_handle (GRUB_EFI_BY_PROTOCOL, &mnp_guid,
                                    0, &num_handles);
  if (! handles) {
    grub_printf("No managed network handles\n");
    if (bootp_resp)
      grub_free (bootp_resp);
    return;
  }

  for (handle = handles; num_handles--; handle++)
    {
      grub_efi_managed_network_t *mnet;

      mnet = grub_efi_open_protocol (*handle, &mnp_guid,
                                     GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
      if (mnet == 0)
        {
          grub_printf("Could not open the MN protocol.\n");
          continue;
        }

      status = efi_call_2(mnet->configure, mnet, 0);
      if (status != GRUB_EFI_SUCCESS)
        {
          grub_printf("Could not reset the MN protocol.\n");
          continue;
        }
    }

  grub_free(handles);

  /*
   * Find a Managed Network Service Binding and create Managed Network
   * handles until there is one with the WANBoot MAC address. Destroy any
   * other handles created along the way.
   */

  handles = grub_efi_locate_handle (GRUB_EFI_BY_PROTOCOL, &mnsbp_guid,
                                    0, &num_handles);
  if (! handles)
    {
      grub_printf("No managed network service bindings found\n");
      if (bootp_resp)
	grub_free (bootp_resp);
      return;
    }

  for (handle = handles; num_handles--; handle++)
    {
      grub_efi_managed_network_binding_t *mnetb;
      grub_efi_managed_network_t *mnet;
      grub_efi_handle_t mnet_handle;
      struct grub_efi_simple_network_mode snp_mode;

      mnetb = grub_efi_open_protocol (*handle, &mnsbp_guid,
        GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
      if (mnetb == 0)
        {
          grub_dprintf("wbefinet", "Could not open the MNP SB protocol.\n");
          continue;
        }

      /* Must be a NULL pointer for create_child to create a handle */
      mnet_handle = 0;

      status = efi_call_2(mnetb->create_child, mnetb, &mnet_handle);

      if (status != GRUB_EFI_SUCCESS || mnet_handle == 0)
        {
          grub_dprintf("wbefinet", "Could not create a new MNP child: st=0x%lx "
            "mnet_handle=%p\n", status, mnet_handle);
          continue;
        }

      /* Open the new child handle to find its MAC address */

      mnet = grub_efi_open_protocol (mnet_handle, &mnp_guid,
                                     GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL);
      if (mnet == 0)
        {
          grub_dprintf("wbefinet", "Could not get MNP protocol.\n");
          status = efi_call_2(mnetb->destroy_child, mnetb, mnet_handle);
          grub_dprintf("wbefinet", "MNP->destroy_child => 0x%lx\n",
            status);
          continue;
        }

      setup_filters_and_configure(mnet);

      status = efi_call_3(mnet->get_mode_data, mnet, 0, &snp_mode);
      if (status == GRUB_EFI_SUCCESS)
        {
          /*
           * If the MAC Address for the handle created on this Service
           * Binding doesn't match the WANBoot MAC Address, destroy this
           * handle and move on.
           */

          if (grub_memcmp(&(mac_addr.mac), snp_mode.current_address,
                          sizeof(mac_addr.mac)) != 0)
            {
              status = efi_call_2(mnetb->destroy_child, mnetb, mnet_handle);
              grub_dprintf ("wbefinet", "MNP->destroy_child => 0x%lx\n",
                            status);

              continue;
            }

          mtu = snp_mode.max_packet_size;
          grub_memcpy (mac, snp_mode.current_address, sizeof (mac));
        }
      else
        {
          /*
           * No Simple Network data, don't try to find MAC address elsewhere
           */

          status = efi_call_2(mnetb->destroy_child, mnetb, mnet_handle);
          grub_dprintf("wbefinet", "MNP->destroy_child => 0x%lx\n",
                       status);
          continue;
        }

      /* Allocate and initialize card and dev structs */

      card = grub_zalloc (sizeof (struct grub_net_card));
      if (!card)
        {
          grub_print_error ();
          grub_free (handles);
	  if (bootp_resp)
	    grub_free (bootp_resp);
          return;
        }

      dev = grub_zalloc (sizeof (struct grub_wbefinet_dev_data));
      if (!dev)
        {
          grub_print_error ();
          grub_free(card);
          grub_free (handles);
	  if (bootp_resp)
	    grub_free (bootp_resp);
          return;
        }

      card->name = grub_xasprintf ("wbefinet0");
      card->driver = &wbefidriver;
      card->flags = 0;
      card->default_address.type = GRUB_NET_LINK_LEVEL_PROTOCOL_ETHERNET;

      card->mtu = mtu;
      grub_memcpy (card->default_address.mac, mac,
                   sizeof (card->default_address.mac));

      dev->efi_mnet = mnet;
      dev->efi_mnet_binding = mnetb;
      dev->efi_mnet_handle = mnet_handle;
      dev->efi_handle = *handle;

      wbefinet_init_rx_ring(dev);
      wbefinet_init_tx_ring(dev);

      if (dev->efi_rx_count == 0 || dev->efi_tx_count == 0)
        {
          wbefinet_close(card);
          grub_free(dev);
          grub_free((char *)card->name);
          grub_free(card);
        }
      else
        {
          card->data = dev;
          grub_net_card_register (card);
        }

      /* initialize card with that network configuration */
      grub_net_configure_by_ip4_addrs (card->name, card, 0, &host_ip_addr,
                                       &router_ip_addr, &mac_addr,
				       bootp_resp, bootp_resp_len);

      if (bootp_resp)
	grub_free (bootp_resp);

      grub_printf("Network interface configured\n");

      break;
    }

  grub_free (handles);
}

/*
 * This function assumes that card->efi_mnet was already initialized to
 * a protocol handle for MNP.
 */
static void
grub_efi_net_config_real (grub_efi_handle_t hnd __attribute__((unused)),
                          char **device, char **path)
{
  grub_net_network_level_address_t server_ip;
  grub_uint32_t ip_addr;                       /* XXX IPv4 only */

  if (get_wb_server_ip (&server_ip))
    {
      grub_uint16_t port;

      ip_addr = grub_be_to_cpu32 (server_ip.ipv4);

      if (get_wb_server_port(&port))
        {
          *device = grub_xasprintf ("http,%u.%u.%u.%u,%d",
                                (ip_addr >> 24) & 0xff, (ip_addr >> 16) & 0xff,
                                (ip_addr >> 8) & 0xff, ip_addr & 0xff, port);

        }
      else
        {
          *device = grub_xasprintf ("http,%u.%u.%u.%u",
                                (ip_addr >> 24) & 0xff, (ip_addr >> 16) & 0xff,
                                (ip_addr >> 8) & 0xff, ip_addr & 0xff);

        }

      *path = grub_xasprintf ("/");
    }
  else
    {
      grub_dprintf("wbefinet", "Failed to get WANBoot server IP address\n");
    }
}

static grub_err_t
grub_wbefinet_fini_hw (int noreturn __attribute__ ((unused)))
{
  struct grub_net_card *card;
  FOR_NET_CARDS (card)
    if (card->driver == &wbefidriver)
      wbefinet_close (card);
  return GRUB_ERR_NONE;
}

static grub_err_t
grub_wbefinet_restore_hw (void)
{
  return GRUB_ERR_NONE;
}

static struct grub_preboot *wbefinet_fini_hnd;


GRUB_MOD_INIT(wbefinet)
{
  grub_wbefinet_findcards ();

  grub_efi_net_config = grub_efi_net_config_real;

  wbefinet_fini_hnd = grub_loader_register_preboot_hook (grub_wbefinet_fini_hw,
                                                         grub_wbefinet_restore_hw,
							 GRUB_LOADER_PREBOOT_HOOK_PRIO_DISK);

}

GRUB_MOD_FINI(wbefinet)
{
  struct grub_net_card *card, *next;

  FOR_NET_CARDS_SAFE (card, next)
    if (card->driver == &wbefidriver)
      grub_net_card_unregister (card);

  grub_loader_unregister_preboot_hook (wbefinet_fini_hnd);

}
