/* wanbootfs.c - download and mount wanbootfs */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/types.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/err.h>
#include <grub/crypto.h>
#include <grub/dl.h>
#include <grub/disk.h>
#include <grub/env.h>
#include <grub/extcmd.h>
#include <grub/efi/api.h>
#include <grub/efi/efi.h>
#include <grub/efi/wanboot.h>
#include <grub/file.h>
#include <grub/fs.h>
#include <grub/time.h>
#include <grub/net/tcp.h>
#include <grub/net/ip.h>
#include <grub/net/ethernet.h>
#include <grub/net/netbuff.h>
#include <grub/net.h>
#include <grub/i18n.h>
#include <grub/normal.h>

GRUB_MOD_LICENSE ("GPLv3+");

static grub_efi_guid_t wanboot_guid = WANBOOT_GUID;

/*
 * wanbootfs data
 */
typedef struct wbfs_info
{
  /* efi variables */
  char			                server_rsrc[WB_SERVER_RSRC_LEN+1];
  char                                  client_id[WB_CLIENT_ID_STR_LEN+1];
  grub_uint16_t                         server_port;
  grub_net_link_level_address_t         mac_addr;
  grub_net_network_level_netaddress_t   host_addr;
  grub_net_network_level_address_t      server_addr;
  grub_net_network_level_address_t      gw_addr;

  /* crypto */
  char					nonce[WB_NONCE_STR_LEN+1];

  /* misc */
  char			                network_id[WB_IPADDR_STR_LEN+1];
  grub_net_tcp_socket_t                 sock;
  int                                   http_resp_off;
  int					content_length;
  int					response_done;
  char                                  *mime_boundary;

  /* buffers and pointers into it */
  int					http_resp_len;
  char					*http_resp;	/* used as wbfs buff */
  int					fs_buffer_len;
  unsigned char				*fs_buffer;	/* ptr into http resp */
  int					fs_digest_len;
  unsigned char				*fs_digest;	/* ptr into http resp */

} *wbfs_info_t;

/*
 * This will point into a buffer allocated to read the HTTP response from the
 * server. There is no method to release the WANBoot fs device, so the pointer
 * to the start of that response buffer is saved right now.
 */
unsigned char *wbfs_addr = 0;
int wbfs_size = 0;

/*
 * get_hex_digit_char
 * Convert a nibble to an upper-case hexadecimal digit character.
 */
static char
get_hex_digit_char (unsigned char digit)
{
  if (digit > 0xf)
    return '0';

  if (digit <= 0x9)
    return ('0' + digit);
  else
    return ('A' + (digit - 0xa));
}

/*
 * build_cid_string
 * This was originally implemented used grub_snprintf(), but it does not
 * support %x and the client ID query string value needs uppercase hexadecimal
 * digits.
 */
static void
build_cid_string (char *string, unsigned char *value, grub_size_t length, int is_mac)
{
  grub_size_t i;
  char *cp = string;

  /* Prepend the starting portion */
  if (is_mac)
    {
      grub_strcpy (cp, "01");
      cp += grub_strlen ("01");
    }

  /* Fill in with client_id or macaddr value */
  for (i=0; i<length; i++)
    {
      *cp++ = get_hex_digit_char ((value[i] >> 4) & 0xf);
      *cp++ = get_hex_digit_char (value[i] & 0xf);
    }
  *cp = '\0';
}

/*
 * Get WANBoot UEFI variables
 */

/*
 * wb_get_server_ip
 */
static int
wb_get_server_ip (grub_net_network_level_address_t *server_addr)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_SERVER_IP, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  /*
   * At this time, IPv4 support only
   */
  server_addr->type = GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV4;
  grub_memcpy (&server_addr->ipv4, data, sizeof (server_addr->ipv4));

  return 1;
}

/*
 * wb_get_server_port
 * wanboot-cgi server port, valid port number
 */
static int
wb_get_server_port (grub_uint16_t *port)
{
  grub_size_t datasize;
  grub_uint16_t *data;

  data = grub_efi_get_variable (WB_BI_SERVER_PORT, &wanboot_guid, &datasize);
  if (data == NULL || datasize < sizeof (grub_uint16_t))
    {
      *port = WB_DEF_SERVER_PORT;
      return 0;
    }

  *port = *data;

  return 1;
}

/*
 * wb_get_host_ip
 */
static int
wb_get_host_ip (grub_uint32_t *host_addr)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_HOST_IP, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  grub_memcpy (host_addr, data, sizeof (grub_uint32_t));

  return 1;
}

/*
 * wb_get_subnet_mask
 */
static int
wb_get_subnet_mask (grub_uint32_t *subnet_mask)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_SUBNET_MASK, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  /*
   * Look for a EFI_IP_ADDRESS (16 bytes) or a EFI_IPv4_ADDRESS (4 bytes).
   * The IPv4 address is in the same place either way.
   */

  if (datasize != 4 && datasize != 16)
    return 0;

  grub_memcpy (subnet_mask, data, sizeof (grub_uint32_t));

  return 1;
}

/*
 * wb_get_macaddr
 * get MAC address
 */
static int
wb_get_macaddr (grub_net_link_level_address_t *mac_addr)
{
  grub_size_t datasize;
  grub_efi_mac_address_t *data;

  data = grub_efi_get_variable (WB_BI_MACADDR, &wanboot_guid, &datasize);
  if (data == NULL || datasize < sizeof (mac_addr->mac))
    {
      grub_memset (&mac_addr->mac, 0, sizeof (mac_addr->mac));
      return 0;
    }

  mac_addr->type = GRUB_NET_LINK_LEVEL_PROTOCOL_ETHERNET;
  grub_memcpy (&mac_addr->mac, data, sizeof (mac_addr->mac));

  return 1;
}

/*
 * wb_get_server_resource
 * wanboot-cgi server resource name
 */
static int
wb_get_server_resource (char *resource)
{
  grub_size_t datasize;
  char *data;

  grub_memset (resource, 0, WB_SERVER_RSRC_LEN+1);

  data = grub_efi_get_variable (WB_BI_SERVER_RSRC, &wanboot_guid, &datasize);

  if (datasize > WB_SERVER_RSRC_LEN)
    return 0;

  if (data == NULL || datasize == 0)
    grub_strcpy (resource, WB_DEF_SERVER_RSRC);
  else
    grub_strncpy (resource, data, datasize);

  return 1;
}

/*
 * wb_get_client_id
 * Return the client ID as a string.
 * If a client-id is defined, check to see if it is longer than 14 hex digits
 * (7 bytes). If it is truncate. Then use the client-id. Otherwise, prepend
 * '01' to the MAC address.
 */
static int
wb_get_client_id (char *client_id)
{
  grub_size_t datasize;
  unsigned char *data;

  grub_memset (client_id, 0, WB_CLIENT_ID_STR_LEN+1);

  data = grub_efi_get_variable (WB_BI_CID, &wanboot_guid, &datasize);

  /* If too long, truncate to what the AI Server can use */
  if (datasize > WB_CLIENT_ID_LEN)
    {
      grub_printf ("Client ID longer than can be used by the AI Server.\n");
      grub_printf ("Truncating to %d bytes\n", WB_CLIENT_ID_LEN);
      datasize = WB_CLIENT_ID_LEN;
    }

  if (data == NULL || datasize == 0)
    {
      data = grub_efi_get_variable (WB_BI_MACADDR, &wanboot_guid, &datasize);
      if (data == NULL || datasize < WB_MACADDR_LEN)
        return 0;

      build_cid_string (client_id, data, WB_MACADDR_LEN, 1);
    }
  else
    {
      build_cid_string (client_id, data, datasize, 0);
    }

  return 1;
}

/*
 * wb_use_client_keys
 * Returns true if the variable can be read and is non-zero, otherwise false.
 */
static int
wb_use_client_keys (void)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_USE_CLIENT_KEYS, &wanboot_guid,
				&datasize);

  if (data == NULL)
    return 0;

  if (*data != 0)
    return 1;

  return 0;
}

/*
 * wb_get_hmac_key
 * Get SHA-256 hashing key, used to verify WANBoot filesystem digest
 */
static int
wb_get_hmac_key (unsigned char *hmac_key)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_HMAC_KEY, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  if (datasize != SHA256_DIGEST_LEN)
    return 0;

  grub_memcpy (hmac_key, data, SHA256_DIGEST_LEN);
  
  return 1;
}

/*
 * wb_get_aes_key
 */
static int
wb_get_aes_key (unsigned char *aes_key, int *keylen)
{
  grub_size_t datasize;
  unsigned char *data;

  data = grub_efi_get_variable (WB_BI_AES_KEY, &wanboot_guid, &datasize);
  if (data == NULL)
    return 0;

  switch (datasize)
    {
    case AES128_KEY_LEN:
    case AES192_KEY_LEN:
    case AES256_KEY_LEN:
      if (aes_key != 0)
        grub_memcpy (aes_key, data, datasize);
      *keylen = datasize;
      return 1;
    default:
      return 0;
    }
}

/*
 * wb_get_network_id
 * Network ID for HTTP request
 */
static int
wb_get_network_id (char *network_id)
{
  int rv;
  grub_uint32_t host_ip;
  grub_uint32_t subnet_mask;

  rv = wb_get_host_ip (&host_ip);
  if (rv)
    {
      rv = wb_get_subnet_mask (&subnet_mask);
      if (rv)
	{
	  grub_uint32_t nid = host_ip & subnet_mask;
	  grub_uint8_t *b_nid = (grub_uint8_t *)&nid;

	  grub_snprintf(network_id, WB_IPADDR_STR_LEN, "%u.%u.%u.%u",
			b_nid[0], b_nid[1], b_nid[2], b_nid[3]);

	  return 1;
	}
    }

  /* Can't calculate the network id? Assign a default */
  grub_strcpy (network_id, "192.168.100.0");

  return 1;
}

/*
 * Crypto Support
 */

/*
 * generate_nonce
 * In the secure case, hash the ms time. Since a SHA-256 key is available,
 * use SHA-256, then truncate to 160 bits.
 * In the insecure case, it does not matter what is used.
 */
static int
generate_nonce (char *nonce_buf)
{
  if (wb_use_client_keys ())
    {
      char *cp;
      int i;
      gcry_err_code_t err;
      unsigned char hmac_key[SHA256_DIGEST_LEN];
      unsigned char computed[SHA256_DIGEST_LEN];
      grub_uint64_t seed = grub_get_time_ms();

      if (!wb_get_hmac_key (hmac_key))
	return 0;

      err = grub_crypto_hmac_buffer (GRUB_MD_SHA256, hmac_key,
				     SHA256_DIGEST_LEN,
				     (char *)&seed, sizeof (grub_uint64_t),
				     computed);

      if (err)
	return 0;

      /* Truncating the resulting hash */

      cp = nonce_buf;
      for (i=0; i<(WB_NONCE_STR_LEN/2); i++)
	{
	  *cp++ = get_hex_digit_char ((computed[i] >> 4) & 0xf);
	  *cp++ = get_hex_digit_char (computed[i] & 0xf);
	}
      *cp = '\0';
    }
  else
    {
      grub_strcpy (nonce_buf, "F909D552F25B86A0BF42AEC62441EBFA8D5E327C");
    }

  return 1;
}

static int
check_nonce(char *dl_nonce)
{
  int rv = 0;
  grub_file_t file;
  grub_ssize_t bytes_read;
  grub_size_t nonce_buf_len = WB_NONCE_LEN;
  char read_nonce[WB_NONCE_LEN];

  file = grub_file_open ("(wbfs)/nonce");

  /* if there is no nonce file, return success */
  if (! file)
    {
      grub_dprintf("wanboot", "Nonce file missing on WANBoot fs\n");
      return 1;
    }

  bytes_read = grub_file_read (file, read_nonce, nonce_buf_len);
  if (bytes_read >= WB_NONCE_LEN)
    {
      if (! grub_strncmp (read_nonce, dl_nonce, WB_NONCE_LEN))
	rv = 1;
    }

  grub_file_close (file);

  return rv;
}

/*
 * verify_digest
 * Verify the supplied digest against the message using SHA-256.
 */
static int
verify_digest (unsigned char *msg, grub_size_t msglen, unsigned char *digest)
{
  int rv = 0;
  gcry_err_code_t err;
  unsigned char computed[SHA256_DIGEST_LEN];
  unsigned char hmac_key[SHA256_DIGEST_LEN];

  if (!wb_get_hmac_key (hmac_key))
    return 0;

  err = grub_crypto_hmac_buffer (GRUB_MD_SHA256, hmac_key, SHA256_DIGEST_LEN,
				 msg, msglen, computed);

  if (! err && ! grub_crypto_memcmp (digest, computed, SHA256_DIGEST_LEN))
    rv = 1;

  return rv;
}

static int
decrypt_payload (unsigned char *in_out_buffer, grub_size_t buffer_len)
{
  grub_crypto_cipher_handle_t hdl = NULL;
  gcry_err_code_t err;
  unsigned char *aes_key;
  unsigned char *crypt_buffer;
  unsigned char *iv;
  int key_len = 0;
  int rv = 0;

  /* Allocate the key buffer for the maximum supported length */
  aes_key = grub_malloc (AES256_KEY_LEN);
  if (! aes_key)
    return 0;

  /* Get the key and the key length */
  rv = wb_get_aes_key (aes_key, &key_len);
  if (rv == 0)
    return 0;

  /* Decrypt the payload */

  hdl = grub_crypto_cipher_open (GRUB_CIPHER_AES);
  if (! hdl)
    return 0;

  err = grub_crypto_cipher_set_key (hdl, aes_key, key_len);
  if (err)
    {
      grub_crypto_cipher_close (hdl);
      grub_free (aes_key);
      return 0;
    }

  /*
   * The IV is at the start of the downloaded buffer, followed by the
   * encrypted payload
   */

  iv = in_out_buffer;
  crypt_buffer = in_out_buffer + key_len;

  err = grub_crypto_cbc_decrypt (hdl, crypt_buffer, crypt_buffer,
				 buffer_len - key_len, iv);
  if (! err)
    rv = 1;

  grub_crypto_cipher_close (hdl);
  grub_free (aes_key);

  return rv;
}

/*
 * Memory-based disk WANBoot FS support
 */

/*
 * create_wanbootfs
 * The WANBoot fs is supposed to be an ISO-9660 filesystem, so check for a
 * Volume Descriptor to validate it.
 */
static int
create_wanbootfs (unsigned char *addr, int size)
{
  char *vold_id;

  grub_dprintf ("wanbootfs", "wanbootfs at %p, %d\n", addr, size);
  
  /*
   * The first Volume Descriptor starts after the System Area (32768 byte).
   * It is a one-byte type, followed by "CD001", followed by a one-byte
   * version (0x01), followed by data.
   *
   * Verify by checking for the Volume Descriptor identifier, which should be
   * at offset 0x8001.
   */

  vold_id = (char *)(addr + 0x8001);

  if (grub_strncmp (vold_id, "CD001", 5))
    {
      int j;

      grub_printf ("WANBoot fs image missing Volume Descriptor identifier\n");

      grub_dprintf ("wanbootfs", "\nStart of fs buffer:\n");
      for (j=0; j<0x20; j++)
	grub_dprintf ("wanbootfs", "0x%x ", addr[j]);
      grub_dprintf ("wanbootfs", "Start of ISOFS Volume Descriptor:\n");
      for (j=0x8000; j<0x8020; j++)
	grub_dprintf ("wanbootfs", "0x%x ", addr[j]);
      grub_dprintf ("wanbootfs", "\n");

      return 0;
    }

  wbfs_size = size;
  wbfs_addr = addr;

  return 1;
}

static grub_err_t
destroy_wanbootfs (void)
{
  grub_free (wbfs_addr);
  wbfs_addr = 0;
  wbfs_size = 0;

  return 0;
}

static int
grub_wbfs_iterate (int (*hook) (const char *name),
                      grub_disk_pull_t pull)
{
  if (pull != GRUB_DISK_PULL_NONE)
    return 0;

  return hook ("wbfs");
}

static grub_err_t
grub_wbfs_open (const char *name, grub_disk_t disk)
{
  if (grub_strcmp (name, "wbfs"))
      return grub_error (GRUB_ERR_UNKNOWN_DEVICE, "not a wanbootfs disk");

  disk->total_sectors = wbfs_size / GRUB_DISK_SECTOR_SIZE;
  disk->id = (unsigned long) "wbfs";

  return GRUB_ERR_NONE;
}

static void
grub_wbfs_close (grub_disk_t disk __attribute((unused)))
{
}

static grub_err_t
grub_wbfs_read (grub_disk_t disk __attribute((unused)),
                     grub_disk_addr_t sector,
                     grub_size_t size, char *buf)
{
  unsigned int offset = sector << GRUB_DISK_SECTOR_BITS;
  unsigned int length = size << GRUB_DISK_SECTOR_BITS;

  if (wbfs_addr == 0)
    return GRUB_ERR_READ_ERROR;

  grub_memcpy (buf, wbfs_addr + offset, length);
  return 0;
}

static grub_err_t
grub_wbfs_write (grub_disk_t disk __attribute ((unused)),
                      grub_disk_addr_t sector __attribute ((unused)),
                      grub_size_t size __attribute ((unused)),
                      const char *buf __attribute ((unused)))
{
  return GRUB_ERR_NOT_IMPLEMENTED_YET;
}

/*
 * HTTP response parsing
 */

/*
 * parse_response_code
 * Response should be "HTTP/1.1 <3-digit response code>".
 */
static int
parse_response_code (char *linep, int *resp)
{
  int i;
  char *lptr = linep;

  *resp = -1;

  if (grub_strncmp (lptr, "HTTP/1.1", grub_strlen ("HTTP/1.1")))
    return 0;

  lptr += grub_strlen ("HTTP/1.1");
  while (grub_isspace (*lptr))
    lptr++;

  i = 0;
  while (grub_isdigit (lptr[i]))
    i++;

  if (i != 3)		/* must be 3 digits */
    return 0;

  *resp = grub_strtol (lptr, NULL, 10);

  return 1;
}

/*
 * parse_http_status
 * Parse the response header and check the response code. Only code 200 (OK)
 * is supported.
 */
static char *
parse_http_status (char *buffer, int length)
{
  char *bptr;
  int resp, rv;

  /* Find the end of the first line */
  bptr = grub_memchr (buffer, '\n', length);
  if (bptr)
    {
      /* Advance past the CRLF to the start of the next line */
      bptr++;
    }
  else
    {
      grub_dprintf ("wanboot", "Cannot find end of HTTP Status line\n");
      return NULL;
    }

  /* Parse the line to extract the HTTP Status Code */
  rv = parse_response_code (buffer, &resp);
  if (rv == -1)
    {
      grub_dprintf ("wanboot", "Bad HTTP status or response header\n");
      return NULL;
    }

  /* Status Code 200 is "OK" Response */
  if (resp != 200)
    {
      grub_dprintf ("wanboot", "Bad HTTP response type - %d\n", resp);
      return NULL;
    }

  return bptr;
}

/*
 * parse_mime_header
 * Parses the MIME header fields to find the Content-Length and MIME
 * boundary as well as confirm the Content Type.
 */
static char *
parse_mime_header (char *buffer, int length, wbfs_info_t wbinfo)
{
  char *bptr = buffer;
  int blen = length;
  char *nptr = bptr;
  int blank_cnt = 0;

  while (nptr < (buffer + length))
    {
      /* Get the next line */
      nptr = grub_memchr (bptr, '\n', blen);
      if (nptr)
        {
          /* Advance past the end of the line (CRLF) */
          nptr++;
        }
      else
        {
          grub_printf ("Cannot find end of HTTP Response\n");
          return NULL;
        }

      if (grub_strncmp (bptr, "\r\n", 2) == 0)
        {
          /* Blank line */
          blank_cnt++;
          if (blank_cnt >= 2)
            {
              /* Two consecutive blank lines mark the end of the header */
              return nptr;
            }
        }
      else if (grub_strncmp (bptr, "Content-Type: multipart/mixed; boundary=",
               grub_strlen ("Content-Type: multipart/mixed; boundary=")) == 0)
        {
          /* Multi-part boundary separator */
          char *mptr, *eptr;
          int mlen;

          mptr = bptr + grub_strlen("Content-Type: multipart/mixed; boundary=");
          eptr = grub_memchr (mptr, '\r', blen - (mptr - bptr));
          if (! eptr)
            {
              grub_dprintf ("wanboot", "Error parsing boundary content type\n");
              return NULL;
            }
          mlen = eptr - mptr;

          wbinfo->mime_boundary = grub_malloc (mlen + 1);
          if (! wbinfo->mime_boundary)
            {
              grub_dprintf ("wanboot", "Could not allocate MIME boundary\n");
              return NULL;
            }

          grub_memcpy (wbinfo->mime_boundary, mptr, mlen);
          wbinfo->mime_boundary[mlen] = 0;

          blank_cnt = 0;
        }
      else if (grub_strncmp (bptr, "Content-Length: ",
               grub_strlen ("Content-Length: ")) == 0)
        {
          /* Entire response length */
          char *lptr, *eptr;

          lptr = bptr + grub_strlen ("Content-Length: ");
          eptr = grub_memchr (lptr, '\r', blen - (lptr - bptr));
          if (! eptr)
            {
              grub_dprintf ("wanboot", "Error parsing content length\n");
              return NULL;
            }

	  wbinfo->content_length = grub_strtol (lptr, NULL, 10);

          blank_cnt = 0;
        }
      else
        {
          blank_cnt = 0;
        }

      /* On to the next line */
      blen -= (nptr - bptr);
      bptr = nptr;
    }

  grub_printf ("Missing end of WANBoot filesystem MIME header\n");
  return nptr;
}

/*
 * find_mime_part
 * Find the start and length of a MIME part, based on its header
 */
static char *
find_mime_part (char *buffer, int length, int *part_length)
{
  char *bptr = buffer;
  int blen = length;
  char *nptr = bptr;

  if (part_length == NULL)
    return NULL;

  while (bptr < (buffer + length))
    {
      /* Get the next line */
      nptr = grub_memchr (bptr, '\n', length - (nptr - buffer));
      if (nptr)
        {
          /* Advance past the end of the line (CRLF) */
          nptr++;
        }
      else
        {
          grub_dprintf ("wanboot", "Cannot find end of HTTP Response\n");
          return NULL;
        }

      if (grub_strncmp (bptr, "\r\n", 2) == 0)
        {
          /* The contents begin after a blank line */

          if (*part_length == 0)
            {
              grub_dprintf ("wanboot", "Start of content with no length\n");
              return NULL;
            }

          return nptr;
        }
      else if (grub_strncmp (bptr, "Content-Type: ",
                             grub_strlen ("Content-Type: ")) == 0)
        {
          /* Verify that the content type is correct */
          char *lptr, *eptr;

          lptr = bptr + grub_strlen ("Content-Type: ");
          eptr = grub_memchr (lptr, '\r', blen - (eptr - lptr));
          if (grub_strncmp (lptr, "application/octet-stream",
                       grub_strlen("application/octet-stream")) != 0)
            {
              grub_dprintf ("wanboot", "Incorrect content type for WBFS\n");
              return NULL;
            }
        }
      else if (grub_strncmp (bptr, "Content-Length: ",
                             grub_strlen ("Content-Length: ")) == 0)
        {
          /* Parse out the length and put it in the WANBoot data */
          char *lptr, *eptr;

          lptr = bptr + grub_strlen ("Content-Length: ");
          eptr = grub_memchr (lptr, '\r', blen - (bptr - lptr));
          if (! eptr)
            {
              grub_dprintf ("wanboot", "Error parsing content length\n");
              return NULL;
            }

          /*
           * Get the length from the Content-Length, then parse ahead in the
           * response buffer to find the start of the content.
           */

          *part_length = grub_strtol (lptr, NULL, 10);
        }

      blen -= (nptr - bptr);
      bptr = nptr;
    }

  return nptr;
}

/*
 * parse_http_response
 * Parse the MIME response in the HTTP response buffer in the wbfs info
 * and set up the wbfs info to point into the buffer (wbfs image or hmac
 * digest) and otherwise be initialized for later use.
 */
static grub_err_t
parse_http_response (wbfs_info_t wbinfo)
{
  char *rptr = wbinfo->http_resp;
  int rlen = wbinfo->http_resp_len;
  char *nptr;
  int plen;
  grub_err_t err = GRUB_ERR_NONE;

  /* Look for the boundary for the first MIME part */

  nptr = grub_memchr (rptr, '\n', rlen);
  if (nptr)
    {
      /* Move past the end of the line */
      nptr++;
    }
  else
    {
      grub_printf ("Malformed HTTP Response (cannot find end)\n");

      return GRUB_ERR_BAD_FS;
    }

  if ((grub_strncmp (rptr, "--", grub_strlen ("--")) != 0) &&
      (grub_strncmp (rptr + grub_strlen ("--"), wbinfo->mime_boundary,
                         grub_strlen (wbinfo->mime_boundary)) != 0))
    {
      grub_printf ("Malformed HTTP Response (missing MIME boundary)\n");

      return GRUB_ERR_BAD_FS;
    }

  rlen -= (nptr - rptr);
  rptr = nptr;

  /* The first MIME part is expected to be the WANBoot filesystem */
  nptr = find_mime_part (rptr, rlen, &plen);

  if (nptr == NULL)
    {
      /* find_mime_part prints the error message */
      return GRUB_ERR_BAD_FS;
    }

  /* Consumes the MIME part fields */

  rlen -= (nptr - rptr);
  rptr = nptr;

  /* point to the start of the wanboot fs and jump to the end */

  wbinfo->fs_buffer = (unsigned char *)rptr;
  wbinfo->fs_buffer_len = plen;

  rptr += wbinfo->fs_buffer_len;
  rlen -= wbinfo->fs_buffer_len;

  nptr = grub_memchr (rptr, '\n', rlen);
  if (nptr)
    {
      /* Move past the end of the line */
      nptr++;
    }
  else
    {
      grub_printf ("Malformed HTTP Response (cannot find end)\n");

      return GRUB_ERR_BAD_FS;
    }

  /* A blank line must follow the MIME part (WANBoot filesystem) */

  if (grub_strncmp(rptr, "\r\n", grub_strlen ("\r\n")) != 0)
    {
      grub_printf ("Malformed HTTP Response (missing blank line)\n");

      return GRUB_ERR_BAD_FS;
    }

  rlen -= (nptr - rptr);
  rptr = nptr;

  /* A MIME boundary should be next */

  nptr = grub_memchr (rptr, '\n', rlen);
  if (nptr != NULL)
    {
      /* Move past the end of the line */
      nptr++;
    }
  else
    {
      grub_printf ("Malformed HTTP Response (cannot find end)\n");

      return GRUB_ERR_BAD_FS;
    }

  if ((grub_strncmp (rptr, "--", grub_strlen ("--")) != 0) ||
      (grub_strncmp (rptr + grub_strlen ("--"),
                     wbinfo->mime_boundary,
                     grub_strlen (wbinfo->mime_boundary)) != 0))
    {
      grub_printf ("Malformed HTTP Response (missing MIME boundary)\n");

      return GRUB_ERR_BAD_FS;
    }

  rlen -= (nptr - rptr);
  rptr = nptr;

  /* The second MIME part is expected to be the digest */

  nptr = find_mime_part (rptr, rlen, &plen);

  if (nptr == NULL)
    {
      grub_printf ("Malformed HTTP Response (missing digest)\n");

      return GRUB_ERR_BAD_FS;
    }
  else
    {
      wbinfo->fs_digest_len = plen;
      wbinfo->fs_digest = (unsigned char *)nptr;

      return GRUB_ERR_NONE;
    }

  return err;
}

/*
 * Networking support
 */

static void
wbhttp_finish (grub_net_tcp_socket_t sock __attribute__ ((unused)),
            void *d)
{
  wbfs_info_t wbinfo = d;

  /* Mark the buffer as done so polling stops */
  wbinfo->response_done = 1;
}

static void
wbhttp_err (grub_net_tcp_socket_t sock __attribute__ ((unused)),
            void *d)
{
  wbfs_info_t wbinfo = d;

  grub_printf ("Network error while reading WANBoot filesystem\n");

  /* Mark the buffer as done so polling stops */
  wbinfo->response_done = 1;

  if (wbinfo->sock)
    grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);
  wbinfo->sock = 0;
}

/*
 * wbhttp_receive
 * The first buffer will include the HTTP status and the general header and
 * MIME header fields. The MIME header will include the MIME boundary and the
 * overall content length. The overall content length will be used to determine
 * when all of the HTTP response has been read in.
 *
 * The first buffer should correspond to the first TCP packet and the minimum
 * MTU is large enough that the entire header should fit in the first buffer.
 *
 * Once the content length has been determined, the buffer is saved off and
 * subsequent buffers are appended to it until it has the content length.
 */
static grub_err_t
wbhttp_receive (grub_net_tcp_socket_t sock __attribute__ ((unused)),
                struct grub_net_buff *nb,
                void *d)
{
  wbfs_info_t wbinfo = d;
  char *bptr, *nptr;		/* buffer position and next position in buff */
  int blen;			/* buffer length */

  if (wbinfo->http_resp_len == 0)
    {
      char *tmp;

      /* First buffer case */

      bptr = (char *) nb->data;
      blen = (char *) nb->tail - (char *) nb->data;
      nptr = bptr;

      nptr = parse_http_status (bptr, blen);
      if (nptr == NULL)
	{
	  grub_printf ("Invalid HTTP status getting WANBoot filesystem\n");

	  wbinfo->response_done = 1;

	  grub_netbuff_free (nb);
	  grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);

	  return GRUB_ERR_NONE;
	}

      blen -= (nptr - bptr);
      bptr = nptr;

      nptr = parse_mime_header (bptr, blen, wbinfo);
      if (nptr == NULL)
	{
	  grub_printf ("Failed to extract WANBoot filesystem\n");

	  wbinfo->response_done = 1;

	  grub_netbuff_free (nb);
	  grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);

	  return GRUB_ERR_NONE;
	}

      blen -= (nptr - bptr);
      bptr = nptr;

      if (! wbinfo->mime_boundary || wbinfo->content_length == 0)
	{
	  grub_printf ("Error while downloading WANBoot filesystem\n");

	  wbinfo->response_done = 1;

	  grub_netbuff_free (nb);
	  grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);

	  return GRUB_ERR_NONE;
	}

      /*
       * Allocate and initialize a buffer for the remainder of the HTTP
       * response based on the content length specified in the MIME header
       */

      tmp = grub_malloc (wbinfo->content_length);

      if (!tmp)
        {
	  grub_printf ("Error allocating WANBoot filesystem image\n");

	  wbinfo->response_done = 1;

	  grub_netbuff_free (nb);
	  grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);

	  return GRUB_ERR_NONE;
        }

      wbinfo->http_resp = (char *)tmp;
      wbinfo->http_resp_len = wbinfo->content_length;

      /*
       * Copy the remainder of the buffer into the http response and note where
       * the next buffer should be appended
       */
      
      grub_memcpy (wbinfo->http_resp, bptr, blen);
      wbinfo->http_resp_off = blen;
    }
  else
    {
      /*
       * Append the buffer to the end of the http response until we have the
       * content length
       */
      bptr = (char *) nb->data;
      blen = (char *) nb->tail - (char *) nb->data;

      if (wbinfo->http_resp_off + blen > wbinfo->http_resp_len)
	{
	  grub_printf ("Truncating last buffer\n");
	  blen = wbinfo->http_resp_len - wbinfo->http_resp_off;
	}

      grub_memcpy (wbinfo->http_resp + wbinfo->http_resp_off, bptr, blen);
      wbinfo->http_resp_off += blen;

      if (wbinfo->http_resp_off >= wbinfo->content_length)
	{
	  wbinfo->response_done = 1;

	  grub_netbuff_free (nb);
	  grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_DISCARD);

	  return GRUB_ERR_NONE;
	}
    }

  return GRUB_ERR_NONE;
}

/*
 * wbhttp_establish
 */
static grub_err_t
wbhttp_establish (wbfs_info_t wbinfo)
{
  grub_err_t err;
  grub_uint8_t *ptr;
  struct grub_net_buff *nb;
  int key_len = 0;
  int i;

  /*
   * Based on the netbuff allocation from http.c
   */

  /* Allocate network buffer */
  nb = grub_netbuff_alloc (GRUB_NET_TCP_RESERVE_SIZE
                           + sizeof ("GET ") - 1
                           + grub_strlen (wbinfo->server_rsrc)
                           + sizeof ("/?CONTENT=bootfs&IP=") - 1
                           + grub_strlen (wbinfo->network_id)
                           + sizeof ("&CID=") - 1
                           + grub_strlen (wbinfo->client_id)
                           + sizeof ("&NONCE=") -1
                           + grub_strlen (wbinfo->nonce)
                           + sizeof ("&ARCH=i386 HTTP/1.1\r\nHost:\r\n\r\n"));

  if (!nb)
    {
      grub_printf ("Failed to allocate network buffer\n");
      return grub_errno;
    }

  /* Reserve TCP space */
  grub_netbuff_reserve (nb, GRUB_NET_TCP_RESERVE_SIZE);
  ptr = nb->tail;

  /* "GET" */
  err = grub_netbuff_put (nb, sizeof ("GET ") - 1);
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, "GET ", sizeof ("GET ") - 1);
  ptr = nb->tail;

  /* URI - server resource */
  err = grub_netbuff_put (nb, grub_strlen (wbinfo->server_rsrc));
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, wbinfo->server_rsrc, grub_strlen (wbinfo->server_rsrc));
  ptr = nb->tail;

  /* Argument - CONTENT and network ID */
  err = grub_netbuff_put (nb, sizeof ("/?CONTENT=bootfs&IP=") - 1);
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, "/?CONTENT=bootfs&IP=",
               sizeof ("/?CONTENT=bootfs&IP=") - 1);
  ptr = nb->tail;

  /* Value - network ID */
  err = grub_netbuff_put (nb, grub_strlen (wbinfo->network_id));
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, wbinfo->network_id, grub_strlen (wbinfo->network_id));
  ptr = nb->tail;

  /* Argument - client ID */
  err = grub_netbuff_put (nb, sizeof ("&CID=") - 1);
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, "&CID=", sizeof ("&CID=") - 1);
  ptr = nb->tail;

  /* Value - client ID */
  err = grub_netbuff_put (nb, grub_strlen (wbinfo->client_id));
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, wbinfo->client_id, grub_strlen (wbinfo->client_id));
  ptr = nb->tail;

  /* Argument - nonce */
  err = grub_netbuff_put (nb, sizeof ("&NONCE=") - 1);
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, "&NONCE=", sizeof ("&NONCE=") - 1);
  ptr = nb->tail;

  /* Value - nonce */
  err = grub_netbuff_put (nb, grub_strlen (wbinfo->nonce));
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, wbinfo->nonce, grub_strlen (wbinfo->nonce));
  ptr = nb->tail;

  /* End of HTTP request */
  err = grub_netbuff_put (nb, sizeof ("&ARCH=i386 HTTP/1.1\r\nHost:\r\n\r\n"));
  if (err)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to initialize network buffer\n");
      return err;
    }
  grub_memcpy (ptr, "&ARCH=i386 HTTP/1.1\r\nHost:\r\n\r\n",
               sizeof ("&ARCH=i386 HTTP/1.1\r\nHost:\r\n\r\n"));

  /*
   * Open the socket
   */

  wbinfo->sock = grub_net_tcp_open_byaddr (wbinfo->server_addr,
					   wbinfo->server_port,
					   wbhttp_receive,
					   wbhttp_err,
					   wbhttp_finish,
					   wbinfo);

  if (!wbinfo->sock)
    {
      grub_netbuff_free (nb);

      grub_printf ("Failed to open connection to server\n");
      return grub_errno;
    }

  wbinfo->fs_buffer = NULL;
  wbinfo->fs_buffer_len = 0;
  wbinfo->fs_digest = NULL;
  wbinfo->fs_digest_len = 0;
  wbinfo->response_done = 0;

  err = grub_net_send_tcp_packet (wbinfo->sock, nb, 1);
  if (err)
    {
      grub_net_tcp_close (wbinfo->sock, GRUB_NET_TCP_ABORT);

      grub_printf ("Failed to send HTTP request, error - %d\n", err);
      return err;
    }

  grub_printf ("WANBoot filesystem download started\n");

  for (i = 0; !wbinfo->response_done && i < 300; i++)
    {
      grub_net_tcp_retransmit ();
      grub_net_poll_cards (300, &wbinfo->response_done);
    }

  if (! wbinfo->response_done)
    {
      grub_printf ("Polling timed out without receiving full buffer\n");
      return GRUB_ERR_BAD_FS;
    }

  err = parse_http_response (wbinfo);
  if (err != GRUB_ERR_NONE)
    {
      /* parse_http_response should emit an appropriate error message */
      return err;
    }

  grub_printf ("WANBoot filesystem downloaded\n");

  if (wb_use_client_keys ())
    {
      if (! verify_digest (wbinfo->fs_buffer, wbinfo->fs_buffer_len,
                           wbinfo->fs_digest))
        {
          grub_printf ("Digest unverified\n");
          return GRUB_ERR_BAD_FS;
        }

      grub_printf ("Decrypting WANBoot filesystem\n");

      if (! decrypt_payload (wbinfo->fs_buffer, wbinfo->fs_buffer_len))
        {
          grub_printf ("Failed to decrypt WANBoot filesystem\n");
          return GRUB_ERR_BAD_FS;
        }

      (void) wb_get_aes_key (NULL, &key_len);
      if (! create_wanbootfs (wbinfo->fs_buffer + key_len,
                              wbinfo->fs_buffer_len - key_len))
        {
          grub_printf ("Failed to mount WANBoot filesystem\n");
          return GRUB_ERR_BAD_FS;
        }
    }
  else
    {
      int j;

      /*
       * Check for non-zero wanboot fs SHA digest for indication that
       * the wanboot fs is encrypted.
       */
      for (j=0; j<wbinfo->fs_digest_len &&
        wbinfo->fs_digest[j] == '\0'; j++)
        ;

      if (j < wbinfo->fs_digest_len)
        {
          grub_printf("Mismatch between server and client security\n");
          return GRUB_ERR_BAD_FS;
        }
      if (! create_wanbootfs (wbinfo->fs_buffer, wbinfo->fs_buffer_len))
        {
          grub_printf ("Failed to mount WANBoot filesystem\n");
          return GRUB_ERR_BAD_FS;
        }
    }

  if (! check_nonce (wbinfo->nonce))
    {
      grub_printf ("Invalid Nonce value in WANBoot filesystem\n");
      return GRUB_ERR_BAD_FS;
    }

  return GRUB_ERR_NONE;
}

static struct grub_disk_dev grub_wanbootfs_dev =
  {
    .name = "wbfs",
    .id = GRUB_DISK_DEVICE_WBFS_ID,
    .iterate = grub_wbfs_iterate,
    .open = grub_wbfs_open,
    .close = grub_wbfs_close,
    .read = grub_wbfs_read,
    .write = grub_wbfs_write,
    .next = 0
  };


/*
 * Command support
 */

static grub_err_t
grub_cmd_wanbootfs (grub_extcmd_context_t ctxt __attribute__ ((unused)),
                    int argc __attribute__ ((unused)),
                    char **args __attribute__ ((unused)))
{
  wbfs_info_t	wbinfo;
  int rv;

  /* Allocate and zero the structure for the data associated with the command */
  wbinfo = grub_zalloc (sizeof (struct wbfs_info));
  if (wbinfo == NULL)
    {
      grub_printf ("Cannot allocate command data\n");
      return GRUB_ERR_OUT_OF_MEMORY;
    }
  
  /* Get wanboot variables from efi */

  /* macaddr */
  rv = wb_get_macaddr (&wbinfo->mac_addr);
  if (!rv)
    {
      grub_printf ("Failed to get UEFI variable '%s'\n", WB_BI_MACADDR);
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* server address */
  rv = wb_get_server_ip (&wbinfo->server_addr);
  if (!rv)
    {
      grub_printf ("Failed to get UEFI variable '%s'\n", WB_BI_SERVER_IP);
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* server port */
  rv = wb_get_server_port (&wbinfo->server_port);
  if (!rv)
    {
      grub_printf ("Failed to get UEFI variable '%s'\n", WB_BI_SERVER_PORT);
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* server resource */
  rv = wb_get_server_resource (wbinfo->server_rsrc);

  if (!rv)
    {
      grub_printf ("Failed to get UEFI variable '%s'\n", WB_BI_SERVER_RSRC);
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* client id */
  rv = wb_get_client_id (wbinfo->client_id);

  if (!rv)
    {
      grub_printf ("Failed to get client id\n");
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* Crypto */

  /* nonce */
  rv = generate_nonce (wbinfo->nonce);
  if (! rv)
    {
      grub_printf ("Failed to generate nonce\n");
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  /* network id */
  rv = wb_get_network_id (wbinfo->network_id);

  if (! rv)
    {
      grub_printf ("Failed to generate network id\n");
      grub_free (wbinfo);

      return GRUB_ERR_BAD_ARGUMENT;
    }

  rv = wbhttp_establish (wbinfo);
  if (rv != GRUB_ERR_NONE)
    {
      if (wbinfo->mime_boundary)
	grub_free (wbinfo->mime_boundary);

      grub_free (wbinfo);

      return rv;
    }

  /* Success! Let the world know if they ask. */
  grub_env_set (WANBOOT_ENV_ENABLED, "1");
  grub_env_export (WANBOOT_ENV_ENABLED);

  return GRUB_ERR_NONE;
}


/*
 * WANBoot environment initialization support
 */

static int
set_wb_root (char *scheme, char *port)
{
  char *server_ip;
  grub_size_t server_ip_len;
  char *env_buf;

  if (! scheme)
    return 0;

  server_ip = grub_efi_get_variable (WB_BI_SERVER_IP, &wanboot_guid,
                                    &server_ip_len);

  if (! server_ip || server_ip_len <= 4)
    return 0;

  env_buf = grub_malloc (80);
  if (! env_buf)
    return 0;

  /* XXX The server IP address should come from the wanboot.conf file */
  /* either as an IP address or a hostname looked up in the /etc/hosts */
  /* file on the WANBoot fs. Instead, we are using the value in UEFI */

  if (port != 0)
    grub_snprintf (env_buf, 80, "(%s,%u.%u.%u.%u,%s)", scheme,
                   server_ip[0] & 0xff, server_ip[1] & 0xff,
                   server_ip[2] & 0xff, server_ip[3] & 0xff, port);
  else
    grub_snprintf (env_buf, 80, "(%s,%u.%u.%u.%u)", scheme,
                   server_ip[0] & 0xff, server_ip[1] & 0xff,
                   server_ip[2] & 0xff, server_ip[3] & 0xff);

  grub_env_set ("root", env_buf);

  grub_free (env_buf);

  return 1;
}

/*
 * check_set_store_vars
 * Check for the existence of the various store files used by https
 * and set the appropriate environment variables if they exist. If the
 * scheme is specified, they will be used, otherwise they won't.
 */
static void
check_set_store_vars (void)
{
  grub_file_t file;

  /* certstore */
  file = grub_file_open (WBFS_FILE_CERTSTORE);
  if (file)
    {
      grub_file_close (file);
      grub_env_set (HTTPS_ENV_CERTSTORE, WBFS_FILE_CERTSTORE);
      grub_env_export (HTTPS_ENV_CERTSTORE);
    }
  grub_errno = 0;

  /* keystore */
  file = grub_file_open (WBFS_FILE_KEYSTORE);
  if (file)
    {
      grub_file_close (file);
      grub_env_set (HTTPS_ENV_KEYSTORE, WBFS_FILE_KEYSTORE);
      grub_env_export (HTTPS_ENV_KEYSTORE);
    }
  grub_errno = 0;

  /* trusttstore */
  file = grub_file_open (WBFS_FILE_TRUSTSTORE);
  if (file)
    {
      grub_file_close (file);
      grub_env_set (HTTPS_ENV_TRUSTSTORE, WBFS_FILE_TRUSTSTORE);
      grub_env_export (HTTPS_ENV_TRUSTSTORE);
    }
    grub_errno = 0;
}

static grub_err_t
grub_cmd_wanboot_init (grub_extcmd_context_t ctxt __attribute__ ((unused)),
                       int argc, char **args)
{
  grub_file_t file;
  char *linebuf = NULL, *cp;
  grub_err_t err = GRUB_ERR_NONE;

  if (argc != 1)
    return grub_error (GRUB_ERR_BAD_ARGUMENT, "only one parameter allowed");

  file = grub_file_open (args[0]);
  if (! file)
    return grub_error (GRUB_ERR_FILE_NOT_FOUND, "Cannot open '%s'", args[0]);

  for (; err == GRUB_ERR_NONE; grub_free(linebuf))
    {
      linebuf = grub_file_getline (file);

      if (! linebuf)
	break;

      /*
       * the only entries of importance here are 'root_server',
       * 'server_authentication' and 'client_authentication'
       */

      if ((cp = grub_strstr (linebuf, "server_authentication=")) != NULL)
	{
	  cp += grub_strlen ("server_authentication=");
	  if (!grub_strcmp (cp, "yes"))
	    {
	      grub_env_set (HTTPS_ENV_SERVERAUTH, "yes");
	      grub_env_export (HTTPS_ENV_SERVERAUTH);
	    }
	  else
	    {
	      grub_env_set (HTTPS_ENV_SERVERAUTH, "no");
	      grub_env_export (HTTPS_ENV_SERVERAUTH);
	    }
	}
      else if ((cp = grub_strstr (linebuf, "client_authentication=")) != NULL)
	{
	  cp += grub_strlen ("client_authentication=");
	  if (!grub_strcmp (cp, "yes"))
	    {
	      grub_env_set (HTTPS_ENV_CLIENTAUTH, "yes");
	      grub_env_export (HTTPS_ENV_CLIENTAUTH);
	    }
	  else
	    {
	      grub_env_set (HTTPS_ENV_CLIENTAUTH, "no");
	      grub_env_export (HTTPS_ENV_CLIENTAUTH);
	    }
	}
      else if ((cp = grub_strstr (linebuf, "root_server=")) != NULL)
	{
	  char *scheme, *port = NULL;

	  cp += grub_strlen ("root_server=");

	  /* cp should be at the start of a http/https URL */
	  scheme = cp;

	  /* look for ':' following the scheme */
	  if ((cp = grub_strchr (cp, ':')) == NULL)
	    continue;

	  *cp = '\0';
	  cp++;

	  /* jump out if the scheme isn't http or https */
	  if (grub_strcmp (scheme, "http") && grub_strcmp (scheme, "https"))
	    {
	      err = GRUB_ERR_BAD_FILE_TYPE;
	      continue;
	    }

	  /* looking for a port number, which is preceded by ':' */
	  if ((cp = grub_strchr (cp, ':')) == NULL)
	    continue;

	  cp++;
	  port = cp;

	  /* look for '/'  following the port number */
	  if ((cp = grub_strchr (cp, '/')) == NULL)
	    {
	      port = 0;
	      continue;
	    }

	  *cp = '\0';

	  if (!set_wb_root (scheme, port))
	    err = GRUB_ERR_OUT_OF_MEMORY;	/* only way this may fail now */
	}
    }

  if (file)
    grub_file_close (file);

  if (err == GRUB_ERR_NONE)
    {
      check_set_store_vars ();
      return err;
    }

  return grub_error (err, "Failed to parse '%s'", args[0]);
}


static grub_extcmd_t wbfs_cmd, wbi_cmd;

GRUB_MOD_INIT(wanbootfs)
{
  wbfs_cmd = grub_register_extcmd ("wanbootfs",
                                  grub_cmd_wanbootfs,
                                  0,
                                  0,
                                  N_("Download and mount WANBoot fs"),
                                  0);                          /* options */

  wbi_cmd = grub_register_extcmd ("wanbootinit",
                                 grub_cmd_wanboot_init,
                                 0,
                                 0,
                                 N_("Parse wanboot.conf and init WANBoot"),
                                 0);                           /* options */
  grub_disk_dev_register (&grub_wanbootfs_dev);
}

GRUB_MOD_FINI(wanbootfs)
{
  (void) destroy_wanbootfs ();

  grub_unregister_extcmd (wbfs_cmd);
  grub_unregister_extcmd (wbi_cmd);

  grub_disk_dev_unregister (&grub_wanbootfs_dev);
}
