/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/misc.h>
#include <grub/net/tcp.h>
#include <grub/net/ip.h>
#include <grub/net/ethernet.h>
#include <grub/net/netbuff.h>
#include <grub/net.h>
#include <grub/mm.h>
#include <grub/dl.h>
#include <grub/file.h>
#include <grub/i18n.h>
#include <grub/time.h>
#include <grub/datetime.h>
#include <grub/env.h>
#include <stddef.h>
#include <stdint.h>
#include <time.h>
#include <gnutls/gnutls.h>
#include <gnutls/x509.h>
#include <gnutls/pkcs12.h>
#include "/usr/include/sys/errno.h"

GRUB_MOD_LICENSE ("GPLv3+");

typedef void (*cofunction_t) (void *const arg);

typedef struct coroutine_struct
{
  cofunction_t cr_function;	/* 0x00, 0x00 */
  void *cr_arg;			/* 0x04, 0x08 */
  void *cr_stackstart;		/* 0x08, 0x10 */
  size_t cr_stacksize;		/* 0x0c, 0x18 */
  void *cr_stackptr;		/* 0x10, 0x20 */
} coroutine_t;			/* 0x14, 0x28 */

#define COROUTINE_STACKSIZE     8192

#define HTTPS_PORT              443

#define HTTP_LINELEN            256
#define HTTPS_BUFSIZE           8192

#define HTTPS_ENV_CERTSTORE     "https_certstore"
#define HTTPS_ENV_KEYSTORE      "https_keystore"
#define HTTPS_ENV_TRUSTSTORE    "https_truststore"
#define HTTPS_ENV_CLIENTAUTH    "https_client_auth"
#define HTTPS_ENV_SERVERAUTH    "https_server_auth"

#define WANBOOT_PASSPHRASE      "boy with goldfish"

typedef struct https_data
{
  grub_net_packets_t pkts;
  int rcv_data;
  grub_size_t linelength;
  int headers_recv;
  int size_recv;
  grub_net_tcp_socket_t sock;
  gnutls_session_t session;
  char *filename;
  grub_off_t offset;
  int initial;
  grub_err_t err;
  char *errmsg;
  int state;
  int chunked;
  int finished;
  grub_size_t contentlength;
  int have_contentlength;
  grub_size_t chunklength;
  int have_chunklength;
  grub_size_t bodysize;
  grub_size_t filepos;
  coroutine_t coroutine;
  unsigned long long stack[COROUTINE_STACKSIZE];
  char line[HTTP_LINELEN];
  char buffer[HTTPS_BUFSIZE];
} *https_data_t;


#define STATE_BAD        (-1)
#define STATE_START         0
#define STATE_HEADERS       1
#define STATE_CHUNK_START   2
#define STATE_BODY          3
#define STATE_CHUNK_END     4
#define STATE_DONE          5

#ifndef GRUB_TARGET_CPU_X86_64

struct __FILE_TAG
{
  int _cnt;			/* number of available characters in buffer */
  unsigned char *_ptr;		/* next character from/to here in buffer */
  unsigned char *_base;		/* the buffer */
  unsigned char _flag;		/* the state of the stream */
  unsigned char _magic;		/* Old home of the file descriptor */
  unsigned __orientation:2;	/* the orientation of the stream */
  unsigned __ionolock:1;	/* turn off implicit locking */
  unsigned __seekable:1;	/* is file seekable? */
  unsigned __extendedfd:1;	/* enable extended FILE */
  unsigned __xf_nocheck:1;	/* no extended FILE runtime check */
  unsigned __filler:10;
};

#else /* GRUB_TARGET_CPU_X86_64 */

struct __FILE_TAG
{
  long __pad[16];
};

#endif /* GRUB_TARGET_CPU_X86_64 */

typedef struct __FILE_TAG FILE;
#define stdin   (&__iob[0])
#define stdout  (&__iob[1])
#define stderr  (&__iob[2])

#define SZ_TOTAL (257 + 257 + 7)
#define _U      0x00000001	/* Upper case */
#define _L      0x00000002	/* Lower case */
#define _N      0x00000004	/* Numeral (digit) */
#define _S      0x00000008	/* Spacing character */
#define _P      0x00000010	/* Punctuation */
#define _C      0x00000020	/* Control character */
#define _B      0x00000040	/* Blank */
#define _X      0x00000080	/* heXadecimal digit */

typedef struct error_convert_struct
{
  grub_err_t ec_grub_error;
  int ec_unix_error;
} error_convert_t;

typedef unsigned int uint_t;
typedef unsigned long ulong_t;
typedef unsigned long uintptr_t;

#define ENABLE_ECDH   0
#define ENABLE_ANON   0
#define ENABLE_DHE    0
#define ENABLE_PSK    0
#define ENABLE_SRP    1

static void
https_set_error (grub_file_t const file, int const err,
		 const char *format, ...)
		 __attribute__ ((format (printf, 3, 4)));

static void https_warning (const char *format, ...)
			   __attribute__ ((format (printf, 1, 2)));

static int parse_http_status (grub_file_t const file);

static int parse_http_header (grub_file_t const file);

static int parse_chunk_start (grub_file_t const file);

static int parse_chunk_end (grub_file_t const file);

static void https_err (grub_net_tcp_socket_t const sock, void *const f);

static grub_err_t https_receive (grub_net_tcp_socket_t const sock
				 __attribute__ ((unused)),
				 struct grub_net_buff *const nb,
				 void *const f);

static void https_establish (void *const arg);

static grub_err_t https_seek (grub_file_t const file, grub_off_t const off);

static grub_err_t https_open (grub_file_t const file,
			      const char *const filename);

static int
https_read_file (grub_file_t const file,
		 gnutls_datum_t * const blob,
		 const char *const name, const char *const label);

static int
https_read_pkcs12_file (grub_file_t const file,
			gnutls_datum_t * blob,
			gnutls_pkcs12_t * pkcs12,
			const char *filename,
			const char *password, const char *label);

static int
https_process_truststore (gnutls_x509_crt_t ** chain,
			  unsigned int *chain_len,
			  grub_file_t const file,
			  const char *filename, const char *password);

static int
https_process_keystore (gnutls_x509_privkey_t * key,
			unsigned char *key_id,
			grub_file_t const file,
			const char *filename, const char *password);

static int
https_process_certstore (gnutls_x509_crt_t ** chain,
			 unsigned int *chain_len,
			 gnutls_x509_crl_t * crl,
			 grub_file_t const file,
			 const char *filename, const char *password,
			 const unsigned char *key_id);

static int https_print_x509_certificate_info (gnutls_session_t const session);

static void https_session_info (gnutls_session_t const session);

static int https_print_x509cred (gnutls_certificate_credentials_t const
				 x509cred);

static const char *gnutls_errormsg (int const errnum);

static grub_err_t https_close (grub_file_t const file);

static grub_err_t https_packets_pulled (grub_file_t const file);

static grub_err_t https_finish (grub_file_t const file);

extern ssize_t https_recv (gnutls_transport_ptr_t f, void *buf, size_t len);

extern int https_recv_timeout (gnutls_transport_ptr_t f, unsigned int ms);

extern ssize_t https_writev (gnutls_transport_ptr_t f,
			     giovec_t const *iovec, int iovec_cnt);

extern ssize_t https_write (gnutls_transport_ptr_t f,
			    const void *buffer, size_t size);

extern int https_set_errno (gnutls_transport_ptr_t f
			    __attribute__ ((unused)));

extern void coroutine_create (coroutine_t * const new_coroutine,
			      cofunction_t const function,
			      void *const arg,
			      void *const stackstart, size_t const stacksize);

extern void coroutine_resume_stub (void);

extern void coroutine_resume (coroutine_t * const current_coroutine,
			      coroutine_t const *const new_coroutine);

int https_verbose = 0;
int https_debug = 0;
int https_errno = ENOTSUP;
coroutine_t main_routine;
coroutine_t *current_coroutine = &main_routine;

static error_convert_t error_convert[] = {
  {GRUB_ERR_NONE, 0},
  {GRUB_ERR_TEST_FAILURE, ENOTACTIVE},
  {GRUB_ERR_BAD_MODULE, ENOEXEC},
  {GRUB_ERR_OUT_OF_MEMORY, ENOMEM},
  {GRUB_ERR_BAD_FILE_TYPE, EBADFD},
  {GRUB_ERR_FILE_NOT_FOUND, ENOENT},
  {GRUB_ERR_FILE_READ_ERROR, EIO},
  {GRUB_ERR_BAD_FILENAME, ENAMETOOLONG},
  {GRUB_ERR_UNKNOWN_FS, ENOTBLK},
  {GRUB_ERR_BAD_FS, ENOTBLK},
  {GRUB_ERR_BAD_NUMBER, EINVAL},
  {GRUB_ERR_OUT_OF_RANGE, ERANGE},
  {GRUB_ERR_UNKNOWN_DEVICE, ENODEV},
  {GRUB_ERR_BAD_DEVICE, ENXIO},
  {GRUB_ERR_READ_ERROR, EIO},
  {GRUB_ERR_WRITE_ERROR, EIO},
  {GRUB_ERR_UNKNOWN_COMMAND, ENOTSUP},
  {GRUB_ERR_INVALID_COMMAND, ENOTSUP},
  {GRUB_ERR_BAD_ARGUMENT, EINVAL},
  {GRUB_ERR_BAD_PART_TABLE, EBADRQC},
  {GRUB_ERR_UNKNOWN_OS, ENOTSUP},
  {GRUB_ERR_BAD_OS, ENOTSUP},
  {GRUB_ERR_NO_KERNEL, ENOTSUP},
  {GRUB_ERR_BAD_FONT, EBFONT},
  {GRUB_ERR_NOT_IMPLEMENTED_YET, ENOTSUP},
  {GRUB_ERR_SYMLINK_LOOP, ELOOP},
  {GRUB_ERR_BAD_COMPRESSED_DATA, EILSEQ},
  {GRUB_ERR_MENU, EBADRQC},
  {GRUB_ERR_TIMEOUT, ETIME},
  {GRUB_ERR_IO, EIO},
  {GRUB_ERR_ACCESS_DENIED, EPERM},
  {GRUB_ERR_EXTRACTOR, EBADR},
  {GRUB_ERR_NET_BAD_ADDRESS, EADDRNOTAVAIL},
  {GRUB_ERR_NET_ROUTE_LOOP, ELOOP},
  {GRUB_ERR_NET_NO_ROUTE, EHOSTUNREACH},
  {GRUB_ERR_NET_NO_ANSWER, ETIMEDOUT},
  {GRUB_ERR_WAIT, ESRCH},
  {GRUB_ERR_BUG, EADV},
  {GRUB_ERR_NET_PORT_CLOSED, ESHUTDOWN},
  {GRUB_ERR_NET_INVALID_RESPONSE, EBADMSG},
  {GRUB_ERR_NET_UNKNOWN_ERROR, EPROTONOSUPPORT},
  {GRUB_ERR_NET_PACKET_TOO_BIG, EMSGSIZE},
  {GRUB_ERR_NET_NO_DOMAIN, EADDRNOTAVAIL},
};


static void
https_set_error (grub_file_t const file, int const err,
		 const char *format, ...)
{
  https_data_t const data = file->data;
  char *errmsg;
  va_list ap;

  va_start (ap, format);
  errmsg = grub_xvasprintf (format, ap);
  va_end (ap);

  if (https_debug)
    grub_printf ("Error %d, %s\n", err, errmsg);

  if (data->err == 0)
    {
      data->err = err;
      data->errmsg = errmsg;
    }
  else
    grub_free (errmsg);
  data->state = STATE_BAD;
}


static void
https_warning (const char *format, ...)
{
  char *errmsg;
  va_list ap;

  va_start (ap, format);
  errmsg = grub_xvasprintf (format, ap);
  va_end (ap);

  if (https_debug)
    grub_printf ("Warning, %s\n", errmsg);

  grub_free (errmsg);
}


static int
parse_http_status (grub_file_t const file)
{
  https_data_t const data = file->data;
  const char *line = data->line;
  const char *sptr;
  char *eptr;
  int code;

  if (grub_memcmp (line, "HTTP/1.1 ", sizeof ("HTTP/1.1 ") - 1) != 0)
    {
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', invalid http status: `%s'"),
		       data->filename, line);
      return (-1);
    }

  sptr = line + sizeof ("HTTP/1.1 ") - 1;
  code = grub_strtoul (sptr, &eptr, 10);

  if (eptr == sptr || *eptr != ' ')
    {
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', invalid http status: `%s'"),
		       data->filename, line);
      return (-1);
    }

  switch (code)
    {
    case 200:
    case 206:
      data->state = STATE_HEADERS;
      return (1);

    case 404:
      https_set_error (file, GRUB_ERR_FILE_NOT_FOUND,
		       _("file `%s', not found: `%s'"), data->filename, line);
      return (-1);

    default:
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', bad http status: `%s'"),
		       data->filename, line);
      return (-1);
    }
}


static int
parse_http_header (grub_file_t const file)
{
  https_data_t const data = file->data;
  const char *line = data->line;
  const char *sptr;
  char *eptr;
  grub_size_t contentlength;

  if (*line == 0)
    {
      if (data->chunked)
	data->state = STATE_CHUNK_START;
      else
	{
	  if (data->have_contentlength)
	    data->bodysize = data->contentlength;
	  else
	    data->bodysize = (grub_size_t) - 1;
	  data->state = STATE_BODY;
	  data->filepos = data->offset;
	}
      data->headers_recv = 1;
      return (1);
    }

  if (grub_memcmp (line, "Content-Length: ",
		   sizeof ("Content-Length: ") - 1) == 0 && !data->size_recv)
    {
      sptr = line + sizeof ("Content-Length: ") - 1;
      contentlength = grub_strtoull (sptr, &eptr, 10);
      if (eptr == sptr || *eptr != 0)
	{
	  https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
			   _("file `%s', bad content length `%s'"),
			   data->filename, line);
	  return (-1);
	}
      data->contentlength = contentlength;
      file->size = contentlength;
      data->have_contentlength = 1;
    }

  if (grub_memcmp (line, "Transfer-Encoding: chunked",
		   sizeof ("Transfer-Encoding: chunked") - 1) == 0)
    {
      data->chunked = 1;
    }

  return (0);
}

static int
parse_chunk_start (grub_file_t const file)
{
  https_data_t const data = file->data;
  const char *line = data->line;
  char *eptr;
  grub_size_t chunklength;

  chunklength = grub_strtoull (line, &eptr, 16);
  if (eptr == line || *eptr != 0)
    {
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', bad chunk length `%s'"),
		       data->filename, line);
      return (-1);
    }

  if (chunklength == 0)
    {
      data->state = STATE_DONE;
      return (1);
    }

  data->chunklength = chunklength;
  data->have_chunklength = 1;
  data->bodysize = chunklength;
  data->state = STATE_BODY;
  return (1);
}


static int
parse_chunk_end (grub_file_t const file)
{
  https_data_t const data = file->data;
  const char *line = data->line;

  if (*line != 0)
    {
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', bad chunk end `%s'"),
		       data->filename, line);
      return (-1);
    }

  data->state = STATE_CHUNK_START;
  data->have_chunklength = 0;

  return (1);
}


static void
https_err (grub_net_tcp_socket_t const sock __attribute__ ((unused)),
	   void *const f)
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t const data = file->data;

  while (data->pkts.first)
    grub_net_remove_packet (data->pkts.first);

  data->rcv_data++;
  if (data->sock)
    {
      grub_net_tcp_close (data->sock, GRUB_NET_TCP_ABORT);
      data->sock = NULL;
    }
  file->device->net->eof = 1;
  file->device->net->stall = 1;

  coroutine_resume (current_coroutine, &data->coroutine);

  https_finish (file);
}


static grub_err_t
https_receive (grub_net_tcp_socket_t const sock __attribute__ ((unused)),
	       struct grub_net_buff *const nb, void *const f)
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t const data = file->data;
  grub_err_t status;

  status = grub_net_put_packet (&data->pkts, nb);
  if (status == GRUB_ERR_NONE)
    data->rcv_data++;

  coroutine_resume (current_coroutine, &data->coroutine);

  if (data->finished)
    {
      https_finish (file);
    }
  else
    {
      if (file->device->net->packs.count >= 20)
	file->device->net->stall = 1;
      if (file->device->net->packs.count >= 100)
	grub_net_tcp_stall (data->sock);

    }
  return (status);
}


static void
https_establish (void *const arg)
{
  grub_file_t const file = (grub_file_t) arg;
  https_data_t const data = file->data;
  gnutls_certificate_credentials_t xcred;
  size_t cmdlength;
  const char *writeptr;
  char *bufptr;
  int status;
  int have_trust;
  int have_key;
  int have_cert;
  int ok = 1;
  ssize_t readstatus;
  ssize_t writestatus;
  size_t writecount;
  size_t bufleft;
  size_t curlength;
  const char *nlptr;
  struct grub_net_buff *nb;
  const char *serverauth;
  const char *clientauth;
  const char *certstore;
  const char *keystore;
  const char *truststore;
  gnutls_x509_crt_t *trust_chain;
  unsigned int trust_len;
  gnutls_x509_privkey_t key = NULL;
  unsigned char key_id[20];
  gnutls_x509_crt_t *cert_chain;
  unsigned int cert_len;
  gnutls_x509_crl_t crl = NULL;

  serverauth = grub_env_get (HTTPS_ENV_SERVERAUTH);
  clientauth = grub_env_get (HTTPS_ENV_CLIENTAUTH);
  certstore = grub_env_get (HTTPS_ENV_CERTSTORE);
  keystore = grub_env_get (HTTPS_ENV_KEYSTORE);
  truststore = grub_env_get (HTTPS_ENV_TRUSTSTORE);

  if (https_verbose && data->initial)
    {
      if (serverauth == NULL)
	grub_printf ("https serverauth NONE\n");
      else
	grub_printf ("https serverauth `%s'\n", serverauth);

      if (truststore == NULL)
	grub_printf ("https truststore NONE\n");
      else
	grub_printf ("https truststore `%s'\n", truststore);

      if (clientauth == NULL)
	grub_printf ("https clientauth NONE\n");
      else
	grub_printf ("https clientauth `%s'\n", clientauth);

      if (keystore == NULL)
	grub_printf ("https keystore NONE\n");
      else
	grub_printf ("https keystore `%s'\n", keystore);

      if (certstore == NULL)
	grub_printf ("https certstore NONE\n");
      else
	grub_printf ("https certstore `%s'\n", certstore);
    }

  /* for backwards compatibility with gnutls < 3.3.0 */
  gnutls_global_init ();

  status = gnutls_certificate_allocate_credentials (&xcred);
  if (status < 0)
    {
      https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
		       _("file `%s', unable to allocate credentials: %s"),
		       data->filename, gnutls_errormsg (status));
      goto invalid;
    }

  if (grub_strcmp (serverauth, "yes") == 0)
    {
      if (truststore == NULL)
	{
	  https_set_error (file, GRUB_ERR_FILE_READ_ERROR,
			   _("file `%s', no value set for truststore"),
			   data->filename);
	  have_trust = 0;
	}

      else
	have_trust = https_process_truststore (&trust_chain, &trust_len,
					       file, truststore,
					       WANBOOT_PASSPHRASE);

      if (have_trust)
	{
	  status =
	    gnutls_certificate_set_x509_trust (xcred, trust_chain, trust_len);
	  if (status < 0)
	    {
	      https_set_error (file, GRUB_ERR_ACCESS_DENIED,
			       _("file `%s', unable to set trust certificates: %s"),
			       data->filename, gnutls_errormsg (status));
	      have_trust = 0;
	    }
	}

      if (!have_trust)
	ok = 0;
    }

  if (grub_strcmp (clientauth, "yes") == 0)
    {
      if (keystore == NULL)
	{
	  https_set_error (file, GRUB_ERR_FILE_READ_ERROR,
			   _("file `%s', no value set for keystore"),
			   data->filename);
	  have_key = 0;
	}

      else
	have_key =
	  https_process_keystore (&key, key_id, file, keystore,
				  WANBOOT_PASSPHRASE);

      if (certstore == NULL)
	{
	  https_set_error (file, GRUB_ERR_FILE_READ_ERROR,
			   _("file `%s', no value set for certstore"),
			   data->filename);
	  have_cert = 0;
	}

      else
	{
	  if (!have_key)
	    have_cert = https_process_certstore (&cert_chain, &cert_len, &crl,
						 file, certstore,
						 WANBOOT_PASSPHRASE, NULL);
	  else
	    have_cert = https_process_certstore (&cert_chain, &cert_len, &crl,
						 file, certstore,
						 WANBOOT_PASSPHRASE, key_id);
	}

      if (have_key && have_cert)
	{
	  status =
	    gnutls_certificate_set_x509_key (xcred, cert_chain, cert_len,
					     key);
	  if (status < 0)
	    {
	      https_set_error (file, GRUB_ERR_ACCESS_DENIED,
			       _("file `%s', unable to set key: %s"),
			       data->filename, gnutls_errormsg (status));
	      ok = 0;
	    }

	  if (crl)
	    {
	      status = gnutls_certificate_set_x509_crl (xcred, &crl, 1);
	      if (status < 0)
		{
		  https_warning (_("file `%s', unable to set crl: %s"),
				 data->filename, gnutls_errormsg (status));
		}
	    }
	}
      else
	ok = 0;
    }

  if (!ok)
    goto invalid;

  if (https_verbose && data->initial)
    https_print_x509cred (xcred);

  /* Initialize TLS session */
  gnutls_init (&data->session, GNUTLS_CLIENT);

  gnutls_server_name_set (data->session, GNUTLS_NAME_DNS,
			  file->device->net->server,
			  grub_strlen (file->device->net->server));

  /* use default priorities */
  gnutls_set_default_priority (data->session);

  gnutls_credentials_set (data->session, GNUTLS_CRD_CERTIFICATE, xcred);

  gnutls_transport_set_push_function (data->session, https_write);

  gnutls_transport_set_vec_push_function (data->session, https_writev);

  gnutls_transport_set_pull_function (data->session, https_recv);

  gnutls_transport_set_pull_timeout_function (data->session,
					      https_recv_timeout);

  gnutls_transport_set_errno_function (data->session, https_set_errno);

  gnutls_transport_set_ptr (data->session, (void *) file);

  /* Perform the TLS handshake */

  do
    {
      status = gnutls_handshake (data->session);
    }
  while (status < 0 && gnutls_error_is_fatal (status) == 0);

  if (status < 0)
    {
      https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
		       _("file `%s', gnutls handshake failed: %s"),
		       data->filename, gnutls_errormsg (status));
      goto invalid;
    }

  if (https_verbose && data->initial)
    https_session_info (data->session);

  if (!data->initial)
    grub_snprintf (data->line, HTTP_LINELEN,
		   "GET %s HTTP/1.1\r\n"
		   "Host: %s\r\n"
		   "User-Agent: " PACKAGE_STRING "\r\n"
		   "Range: bytes=%" PRIuGRUB_UINT64_T "-\r\n"
		   "\r\n",
		   data->filename, file->device->net->server, data->offset);
  else
    grub_snprintf (data->line, HTTP_LINELEN,
		   "GET %s HTTP/1.1\r\n"
		   "Host: %s\r\n"
		   "User-Agent: " PACKAGE_STRING "\r\n"
		   "\r\n", data->filename, file->device->net->server);
  cmdlength = grub_strlen (data->line);

  for (writecount = cmdlength, writeptr = data->line;
       writecount > 0; writecount -= writestatus, writeptr += writestatus)
    {

      writestatus = gnutls_record_send (data->session, writeptr, writecount);
      if (writestatus <= 0)
	{
	  https_set_error (file, GRUB_ERR_NET_UNKNOWN_ERROR,
			   _("file `%s', unable to send record: %s"),
			   data->filename, gnutls_errormsg (writestatus));
	  goto invalid;
	}
    }

  for (bufleft = 0, bufptr = data->buffer, readstatus = 0;
       data->state >= STATE_START && data->state < STATE_DONE;
       bufleft -= curlength, bufptr += curlength)
    {
      if (bufleft == 0)
	{
	  readstatus = gnutls_record_recv (data->session,
					   data->buffer, HTTPS_BUFSIZE);

	  if (readstatus <= 0)
	    {
	      if (data->state == STATE_BODY)
		{
		  data->state = STATE_DONE;
		  break;
		}

	      https_set_error (file, GRUB_ERR_FILE_READ_ERROR,
			       _("file `%s', unable to receive record: %s"),
			       data->filename, gnutls_errormsg (readstatus));
	      break;
	    }

	  bufleft = readstatus;
	  bufptr = data->buffer;
	}

      if (data->state == STATE_BODY)
	{
	  curlength = bufleft;
	  if (curlength > data->bodysize)
	    curlength = data->bodysize;

	  nb = grub_netbuff_alloc (curlength);
	  if (!nb)
	    {
	      https_set_error (file, grub_errno,
			       _("file `%s', out of memory"), data->filename);
	      goto invalid;
	    }

	  grub_netbuff_put (nb, curlength);
	  grub_memcpy (nb->data, bufptr, curlength);
	  grub_net_put_packet (&file->device->net->packs, nb);

	  data->bodysize -= curlength;
	  data->filepos += curlength;
	  if (data->bodysize == 0)
	    {
	      if (data->chunked)
		data->state = STATE_CHUNK_END;
	      else
		data->state = STATE_DONE;
	    }

	}
      else
	{
	  nlptr = grub_memchr (bufptr, '\n', bufleft);
	  if (nlptr == NULL)
	    curlength = bufleft;
	  else
	    curlength = nlptr - bufptr + 1;

	  if (data->linelength + curlength >= HTTP_LINELEN)
	    {
	      https_set_error (file, GRUB_ERR_FILE_READ_ERROR,
			       _("file `%s', line too long"), data->filename);
	      break;
	    }

	  grub_memcpy (data->line + data->linelength, bufptr, curlength);
	  data->linelength += curlength;

	  if (nlptr == NULL)
	    continue;

	  data->linelength--;
	  if (data->linelength > 0 &&
	      data->line[data->linelength - 1] == '\r')
	    data->linelength--;
	  data->line[data->linelength] = 0;

	  switch (data->state)
	    {
	    case STATE_START:
	      status = parse_http_status (file);
	      break;

	    case STATE_HEADERS:
	      status = parse_http_header (file);
	      break;

	    case STATE_CHUNK_START:
	      status = parse_chunk_start (file);
	      break;

	    case STATE_CHUNK_END:
	      status = parse_chunk_end (file);
	      break;
	    }

	  if (status < 0)
	    break;

	  data->linelength = 0;
	}
    }

  if (data->state == STATE_DONE)
    {
      gnutls_bye (data->session, GNUTLS_SHUT_RDWR);
      data->finished = 1;
    }

invalid:
  if (!data->headers_recv)
    {
      if (data->sock != NULL)
	{
	  grub_net_tcp_close (data->sock, GRUB_NET_TCP_ABORT);
	  data->sock = NULL;
	}
    }

  for (;;)
    coroutine_resume (current_coroutine, &main_routine);
}


static int
https_read_file (grub_file_t const file,
		 gnutls_datum_t * const blob,
		 const char *const filename, const char *const label)
{
  grub_file_t datafile;
  char *p = NULL;
  char *np;
  size_t size = 0;
  ssize_t status;

  blob->data = NULL;
  blob->size = 0;

  datafile = grub_file_open (filename);
  if (datafile == NULL)
    {
      https_set_error (file, grub_errno, _("unable to open %s file `%s': %s"),
		       label, filename, grub_errmsg);
      return (0);
    }

  for (;;)
    {
      np = grub_realloc (p, size + 8192);
      if (np == NULL)
	{
	  https_set_error (file, grub_errno,
			   _("unable to realloc memory for %s file `%s': %s"),
			   label, filename, grub_errmsg);
	  grub_file_close (datafile);
	  if (p != NULL)
	    grub_free (p);
	  return (0);
	}
      p = np;

      status = grub_file_read (datafile, &p[size], 8192);
      if (status < 0)
	{
	  https_set_error (file, grub_errno,
			   _("unable to read %s file `%s': %s"), label,
			   filename, grub_errmsg);
	  grub_file_close (datafile);
	  if (p != NULL)
	    grub_free (p);
	  return (0);
	}
      else if (status == 0)
	break;

      size += status;
    }

  grub_file_close (datafile);

  np = grub_realloc (p, size + 1);
  if (np == NULL)
    {
      https_set_error (file, grub_errno,
		       _("unable to realloc memory for %s file `%s': %s"),
		       label, filename, grub_errmsg);
      if (p != NULL)
	grub_free (p);
      return (0);
    }

  p = np;
  p[size] = 0;

  blob->data = (unsigned char *) p;
  blob->size = (unsigned int) size;

  return (1);
}


static int
https_read_pkcs12_file (grub_file_t const file,
			gnutls_datum_t * blob,
			gnutls_pkcs12_t * pkcs12,
			const char *filename,
			const char *password, const char *label)
{
  int status;

  if (!https_read_file (file, blob, filename, label))
    return (0);

  status = gnutls_pkcs12_init (pkcs12);
  if (status < 0)
    {
      https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
		       _("%s file `%s', unable to allocate pkcs12 structure: %s"),
		       label, filename, gnutls_errormsg (status));
      grub_free (blob->data);
      return (0);
    }

  status = gnutls_pkcs12_import (*pkcs12, blob, GNUTLS_X509_FMT_DER, 0);
  if (status < 0)
    {
      https_set_error (file, GRUB_ERR_BAD_COMPRESSED_DATA,
		       _("%s file `%s', unable to import pkcs12 data: %s"),
		       label, filename, gnutls_errormsg (status));
      gnutls_pkcs12_deinit (*pkcs12);
      grub_free (blob->data);
      return (0);
    }

  if (password != NULL)
    {
      status = gnutls_pkcs12_verify_mac (*pkcs12, password);
      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_ACCESS_DENIED,
			   _("%s file `%s', unable to verify mac: %s"),
			   label, filename, gnutls_errormsg (status));
	  gnutls_pkcs12_deinit (*pkcs12);
	  grub_free (blob->data);
	  return (0);
	}
    }

  return (1);
}

static int
https_process_truststore (gnutls_x509_crt_t ** chain,
			  unsigned int *chain_len,
			  grub_file_t file,
			  const char *filename, const char *password)
{
  int status;
  gnutls_datum_t blob;
  gnutls_pkcs12_t pkcs12;
  gnutls_pkcs12_bag_t bag = NULL;
  int bag_on;
  int bag_type;
  int element_count;
  int element_on;
  int element_type;
  gnutls_datum_t element_data;
  gnutls_x509_crt_t cert;
  gnutls_x509_crt_t *chain_ptr = NULL;
  unsigned int chain_on = 0;

  *chain = NULL;
  *chain_len = 0;

  if (!https_read_pkcs12_file
      (file, &blob, &pkcs12, filename, password, "truststore"))
    return (0);

  for (bag_on = 0;; bag_on++)
    {
      status = gnutls_pkcs12_bag_init (&bag);
      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
			   _("truststore file `%s', unable to allocate bag structure: %s"),
			   filename, gnutls_errormsg (status));
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      status = gnutls_pkcs12_get_bag (pkcs12, bag_on, bag);
      if (status == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
	break;

      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_BAD_PART_TABLE,
			   _("truststore file `%s', unable to get bag structure %d: %s"),
			   filename, bag_on, gnutls_errormsg (status));
	  gnutls_pkcs12_bag_deinit (bag);
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      bag_type = gnutls_pkcs12_bag_get_type (bag, 0);
      if (bag_type < 0)
	{
	  https_warning (_("truststore file `%s', unable to get bag %d type: %s"),
			 filename, bag_on, gnutls_errormsg (bag_type));
	  continue;
	}

      if (bag_type == GNUTLS_BAG_ENCRYPTED)
	{
	  status = gnutls_pkcs12_bag_decrypt (bag, password);
	  if (status < 0)
	    {
	      https_warning (_("truststore file `%s', unable to decrypt bag %d: %s"),
			     filename, bag_on, gnutls_errormsg (status));
	      continue;
	    }
	}

      element_count = gnutls_pkcs12_bag_get_count (bag);
      if (element_count < 0)
	{
	  https_warning (_("truststore file `%s', unable to get bag %d element count: %s"),
			 filename, bag_on, gnutls_errormsg (element_count));
	  continue;
	}

      for (element_on = 0; element_on < element_count; element_on++)
	{
	  element_type = gnutls_pkcs12_bag_get_type (bag, element_on);
	  if (element_type < 0)
	    {
	      https_warning (_("truststore file `%s', unable to get bag %d element %d type: %s"),
			     filename, bag_on, element_on,
			     gnutls_errormsg (element_type));
	      continue;
	    }

	  switch (element_type)
	    {

	    case GNUTLS_BAG_CERTIFICATE:

	      status =
		gnutls_pkcs12_bag_get_data (bag, element_on, &element_data);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to get bag %d element %d data: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      status = gnutls_x509_crt_init (&cert);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("truststore file `%s', unable to allocate crt structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      status =
		gnutls_x509_crt_import (cert, &element_data,
					GNUTLS_X509_FMT_DER);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_BAD_FILE_TYPE,
				   _("truststore file `%s', unable to import crt structure from bag %d, element %d: %s"),
				   filename, bag_on, element_on,
				   gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      chain_ptr =
		(gnutls_x509_crt_t *) gnutls_realloc (chain_ptr,
						      sizeof (*chain_ptr)
						      * (chain_on + 1));
	      if (chain_ptr == NULL)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("truststore file `%s', unable to allocate crt chain structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      chain_ptr[chain_on] = cert;
	      chain_on++;
	      break;
	    }
	}

      gnutls_pkcs12_bag_deinit (bag);
    }

  if (chain_on == 0)
    {
      https_set_error (file, GRUB_ERR_FILE_NOT_FOUND,
		       _("truststore file `%s', no certificates found"),
		       filename);
      gnutls_pkcs12_deinit (pkcs12);
      grub_free (blob.data);
      return (0);
    }


  *chain = chain_ptr;
  *chain_len = chain_on;

  gnutls_pkcs12_deinit (pkcs12);
  grub_free (blob.data);
  return (1);
}


static int
https_process_keystore (gnutls_x509_privkey_t * key,
			unsigned char *key_id,
			grub_file_t const file,
			const char *filename, const char *password)
{
  int status;
  gnutls_datum_t blob;
  gnutls_pkcs12_t pkcs12;
  gnutls_pkcs12_bag_t bag = NULL;
  int bag_on;
  int bag_type;
  int element_count;
  int element_on;
  int element_type;
  gnutls_datum_t element_data;
  gnutls_x509_privkey_t key_ptr = NULL;
  size_t key_id_size;

  *key = NULL;

  if (!https_read_pkcs12_file
      (file, &blob, &pkcs12, filename, password, "keystore"))
    return (0);

  for (bag_on = 0;; bag_on++)
    {
      status = gnutls_pkcs12_bag_init (&bag);
      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
			   _("keystore file `%s', unable to allocate bag structure: %s"),
			   filename, gnutls_errormsg (status));
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      status = gnutls_pkcs12_get_bag (pkcs12, bag_on, bag);
      if (status == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
	break;

      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_BAD_PART_TABLE,
			   _("keystore file `%s', unable to get bag structure %d: %s"),
			   filename, bag_on, gnutls_errormsg (status));
	  gnutls_pkcs12_bag_deinit (bag);
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      bag_type = gnutls_pkcs12_bag_get_type (bag, 0);
      if (bag_type < 0)
	{
	  https_warning (_("keystore file `%s', unable to get bag %d type: %s"),
			 filename, bag_on, gnutls_errormsg (bag_type));
	  continue;
	}

      if (bag_type == GNUTLS_BAG_ENCRYPTED)
	{
	  status = gnutls_pkcs12_bag_decrypt (bag, password);
	  if (status < 0)
	    {
	      https_warning (_("keystore file `%s', unable to decrypt bag %d: %s"),
			     filename, bag_on, gnutls_errormsg (status));
	      continue;
	    }
	}

      element_count = gnutls_pkcs12_bag_get_count (bag);
      if (element_count < 0)
	{
	  https_warning (_("keystore file `%s', unable to get bag %d element count: %s"), filename, bag_on, gnutls_errormsg (element_count));
	  continue;
	}

      for (element_on = 0; element_on < element_count; element_on++)
	{
	  element_type = gnutls_pkcs12_bag_get_type (bag, element_on);
	  if (element_type < 0)
	    {
	      https_warning (_("keystore file `%s', unable to get bag %d element %d type: %s"),
			     filename, bag_on, element_on,
			     gnutls_errormsg (element_type));
	      continue;
	    }

	  switch (element_type)
	    {
	    case GNUTLS_BAG_PKCS8_ENCRYPTED_KEY:
	      status =
		gnutls_pkcs12_bag_get_data (bag, element_on, &element_data);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to get bag %d element %d data: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      status = gnutls_x509_privkey_init (&key_ptr);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("keystore file `%s', unable to allocate privkey structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      status =
		gnutls_x509_privkey_import_pkcs8 (key_ptr, &element_data,
						  GNUTLS_X509_FMT_DER,
						  password, 0);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to import pkcs8 structure from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      key_id_size = 20;
	      status =
		gnutls_x509_privkey_get_key_id (key_ptr, 0, key_id,
						&key_id_size);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to get key id from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      gnutls_pkcs12_bag_deinit (bag);
	      gnutls_pkcs12_deinit (pkcs12);
	      grub_free (blob.data);
	      *key = key_ptr;
	      return (1);

	    case GNUTLS_BAG_PKCS8_KEY:
	      status =
		gnutls_pkcs12_bag_get_data (bag, element_on, &element_data);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to get bag %d element %d data: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      status = gnutls_x509_privkey_init (&key_ptr);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("keystore file `%s', unable to allocate privkey structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      status =
		gnutls_x509_privkey_import_pkcs8 (key_ptr, &element_data,
						  GNUTLS_X509_FMT_DER,
						  password,
						  GNUTLS_PKCS_PLAIN);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to import pkcs8 structure from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      key_id_size = 20;
	      status =
		gnutls_x509_privkey_get_key_id (key_ptr, 0, key_id,
						&key_id_size);
	      if (status < 0)
		{
		  https_warning (_("keystore file `%s', unable to get key id from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      gnutls_pkcs12_bag_deinit (bag);
	      gnutls_pkcs12_deinit (pkcs12);
	      grub_free (blob.data);
	      *key = key_ptr;
	      return (1);
	    }
	}

      gnutls_pkcs12_bag_deinit (bag);
    }

  https_set_error (file, GRUB_ERR_FILE_NOT_FOUND,
		   _("keystore file `%s', no key found"), filename);
  gnutls_pkcs12_deinit (pkcs12);
  grub_free (blob.data);
  return (0);
}


static int
https_process_certstore (gnutls_x509_crt_t ** chain,
			 unsigned int *chain_len,
			 gnutls_x509_crl_t * crl,
			 grub_file_t const file,
			 const char *filename, const char *password,
			 const unsigned char *key_id)
{
  int status;
  gnutls_datum_t blob;
  gnutls_pkcs12_t pkcs12;
  gnutls_pkcs12_bag_t bag = NULL;
  int bag_on;
  int bag_type;
  int element_count;
  int element_on;
  int element_type;
  gnutls_datum_t element_data;
  unsigned char cert_id[20];
  gnutls_x509_crt_t cert;
  size_t cert_id_size;
  gnutls_x509_crt_t *chain_ptr = NULL;
  unsigned int chain_on = 0;
  gnutls_x509_crl_t crl_ptr = NULL;

  if (!https_read_pkcs12_file
      (file, &blob, &pkcs12, filename, password, "truststore"))
    return (0);

  for (bag_on = 0;; bag_on++)
    {
      status = gnutls_pkcs12_bag_init (&bag);
      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
			   _("truststore file `%s', unable to allocate bag structure: %s"),
			   filename, gnutls_errormsg (status));
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      status = gnutls_pkcs12_get_bag (pkcs12, bag_on, bag);
      if (status == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
	break;

      if (status < 0)
	{
	  https_set_error (file, GRUB_ERR_BAD_PART_TABLE,
			   _("truststore file `%s', unable to get bag structure %d: %s"),
			   filename, bag_on, gnutls_errormsg (status));
	  gnutls_pkcs12_bag_deinit (bag);
	  gnutls_pkcs12_deinit (pkcs12);
	  grub_free (blob.data);
	  return (0);
	}

      bag_type = gnutls_pkcs12_bag_get_type (bag, 0);
      if (bag_type < 0)
	{
	  https_warning (_("truststore file `%s', unable to get bag %d type: %s"),
			 filename, bag_on, gnutls_errormsg (bag_type));
	  continue;
	}

      if (bag_type == GNUTLS_BAG_ENCRYPTED)
	{
	  status = gnutls_pkcs12_bag_decrypt (bag, password);
	  if (status < 0)
	    {
	      https_warning (_("truststore file `%s', unable to decrypt bag %d: %s"),
			     filename, bag_on, gnutls_errormsg (status));
	      continue;
	    }
	}

      element_count = gnutls_pkcs12_bag_get_count (bag);
      if (element_count < 0)
	{
	  https_warning (_("truststore file `%s', unable to get bag %d element count: %s"),
			 filename, bag_on, gnutls_errormsg (element_count));
	  continue;
	}

      for (element_on = 0; element_on < element_count; element_on++)
	{
	  element_type = gnutls_pkcs12_bag_get_type (bag, element_on);
	  if (element_type < 0)
	    {
	      https_warning (_("truststore file `%s', unable to get bag %d element %d type: %s"),
			     filename, bag_on, element_on,
			     gnutls_errormsg (element_type));
	      continue;
	    }

	  switch (element_type)
	    {
	    case GNUTLS_BAG_CERTIFICATE:
	      status =
		gnutls_pkcs12_bag_get_data (bag, element_on, &element_data);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to get bag %d element %d data: %s"), filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      status = gnutls_x509_crt_init (&cert);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("truststore file `%s', unable to allocate crt structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      status =
		gnutls_x509_crt_import (cert, &element_data,
					GNUTLS_X509_FMT_DER);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to import crt structure from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      cert_id_size = 20;
	      status =
		gnutls_x509_crt_get_key_id (cert, 0, cert_id, &cert_id_size);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to get cert id from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  gnutls_x509_crt_deinit (cert);
		  continue;
		}

	      if (key_id != NULL && memcmp (cert_id, key_id, 20) == 0)
		{
		  if (chain_on == 0)
		    {
		      chain_ptr =
			(gnutls_x509_crt_t *) gnutls_malloc (sizeof (*chain));
		      if (chain_ptr == NULL)
			{
			  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
					   _("truststore file `%s', unable to allocate crt structure: %s"),
					   filename,
					   gnutls_errormsg (status));
			  gnutls_pkcs12_bag_deinit (bag);
			  gnutls_pkcs12_deinit (pkcs12);
			  grub_free (blob.data);
			  return (0);
			}

		      *chain_ptr = cert;
		      chain_on++;
		    }
		  else
		    gnutls_x509_crt_deinit (cert);
		}
	      else
		gnutls_x509_crt_deinit (cert);
	      break;

	    case GNUTLS_BAG_CRL:
	      if (crl_ptr != NULL)
		break;

	      status =
		gnutls_pkcs12_bag_get_data (bag, element_on, &element_data);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to get bag %d element %d data: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}

	      status = gnutls_x509_crl_init (&crl_ptr);
	      if (status < 0)
		{
		  https_set_error (file, GRUB_ERR_OUT_OF_MEMORY,
				   _("truststore file `%s', unable to allocate crl structure: %s"),
				   filename, gnutls_errormsg (status));
		  gnutls_pkcs12_bag_deinit (bag);
		  gnutls_pkcs12_deinit (pkcs12);
		  grub_free (blob.data);
		  return (0);
		}

	      status =
		gnutls_x509_crl_import (crl_ptr, &element_data,
					GNUTLS_X509_FMT_DER);
	      if (status < 0)
		{
		  https_warning (_("truststore file `%s', unable to import crl structure from bag %d, element %d: %s"),
				 filename, bag_on, element_on,
				 gnutls_errormsg (status));
		  continue;
		}
	      break;
	    }
	}

      gnutls_pkcs12_bag_deinit (bag);
    }

  if (chain_on != 1)
    {
      https_set_error (file, GRUB_ERR_FILE_NOT_FOUND,
		       _("truststore file `%s', no matching certificate found"),
		       filename);
      gnutls_pkcs12_deinit (pkcs12);
      grub_free (blob.data);
      return (0);
    }

  *crl = crl_ptr;
  *chain = chain_ptr;
  *chain_len = chain_on;
  gnutls_pkcs12_deinit (pkcs12);
  grub_free (blob.data);
  return (1);
}

static int
https_print_x509_certificate_info (gnutls_session_t const session)
{

  gnutls_datum_t const *cert_list;
  int status;
  uint_t cert_list_size;
  gnutls_x509_crt_t cert;
  gnutls_datum_t cinfo;

  /* This function only works for X.509 certificates. */

  if (gnutls_certificate_type_get (session) != GNUTLS_CRT_X509)
    return (-1);

  cert_list = gnutls_certificate_get_peers (session, &cert_list_size);

  grub_printf ("- Peer provided %d certificates\n", cert_list_size);

  if (cert_list_size == 0)
    return (0);

  /* we only print information about the first certificate. */

  gnutls_x509_crt_init (&cert);

  gnutls_x509_crt_import (cert, &cert_list[0], GNUTLS_X509_FMT_DER);

  grub_printf ("- Certificate info:\n");

  /* This is the preferred way of printing short information   */
  /* about a certificate                                       */

  status = gnutls_x509_crt_print (cert, GNUTLS_CRT_PRINT_ONELINE, &cinfo);
  if (status == 0)
    {
      grub_printf ("- %s\n", cinfo.data);
      gnutls_free (cinfo.data);
    }

  gnutls_x509_crt_deinit (cert);

  return (cert_list_size);
}


void
https_session_info (gnutls_session_t session)
{
#if ENABLE_PSK
  const char *hint;
  const char *username;
#endif /* ENABLE_PSK */
  gnutls_kx_algorithm_t kx;
#if ENABLE_DHE || ENABLE_ANON
  int dhe = 0;
#endif /* ENABLE_DHE || ENABLE_ANON */
#if ENABLE_ECDH
  int ecdh = 0;
#endif /* ENABLE_ECDH */

  /* print the key exchange's algorithm name */

  kx = gnutls_kx_get (session);
  grub_printf ("- Key Exchange: %s\n", gnutls_kx_get_name (kx));

  /* check the authentication type used, and switch to the     */
  /* appropriate.                                              */

  switch (gnutls_auth_get_type (session))
    {
    case GNUTLS_CRD_IA:
      grub_printf ("- TLS/IA session\n");
      break;

#if ENABLE_SRP

    case GNUTLS_CRD_SRP:
#if 0
      grub_printf ("- SRP session with username %s\n",
		   gnutls_srp_server_get_username (session));
#else
      grub_printf ("- SRP session with username\n");
#endif
      break;

#endif /* ENABLE_SRP */

    case GNUTLS_CRD_PSK:

#if ENABLE_PSK
      hint = gnutls_psk_client_get_hint (session);
      if (hint != NULL)
	grub_printf ("- PSK authentication. PSK hint \"%s\"\n");

      /* This returns NULL in client side. */

      username = gnutls_psk_server_get_username (session);
      if (username != NULL)
	grub_printf ("- PSK authentication.  Connected as \"%s\"\n",
		     username);

#endif /* ENABLE_PSK */

#if ENABLE_ECDH
      if (kx == GNUTLS_KX_ECDHE_PSK)
	ecdh = 1;
#endif /* ENABLE_ECDH */
#if ENABLE_DHE || ENABLE_ANON
      if (kx == GNUTLS_KX_DHE_PSK)
	dhe = 1;
#endif /* ENABLE_DHE || ENABLE_ANON */
      break;
    case GNUTLS_CRD_ANON:	/* anonymous authentication */
      grub_printf ("- Anonymous authentication.\n");

#if ENABLE_ECDH
      if (kx == GNUTLS_KX_ANON_ECDH)
	ecdh = 1;
#endif /* ENABLE_ECDH */
#if ENABLE_DHE || ENABLE_ANON
      if (kx == GNUTLS_KX_ANON_DH)
	dhe = 1;
#endif /* ENABLE_DHE || ENABLE_ANON */
      break;

    case GNUTLS_CRD_CERTIFICATE:	/* certificate authentication */

      /* Check if we have been using ephemeral Diffie-Hellman. */

#if ENABLE_DHE || ENABLE_ANON
      if (kx == GNUTLS_KX_DHE_RSA || kx == GNUTLS_KX_DHE_DSS)
	dhe = 1;
#endif /* ENABLE_DHE || ENABLE_ANON */
#if ENABLE_ECDH
      if (kx == GNUTLS_KX_ECDHE_RSA || kx == GNUTLS_KX_ECDHE_ECDSA)
	ecdh = 1;
#endif /* ENABLE_ECDH */

      /* if the certificate list is available, then print  */
      /* some information about it.                        */

      https_print_x509_certificate_info (session);
      break;
    }

#if ENABLE_ECDH

  if (ecdh != 0)
    grub_printf ("- Ephemeral ECDH using curve %s\n",
		 gnutls_ecc_curve_get_name (gnutls_ecc_curve_get (session)));

#endif /* ENABLE_ECDH */

#if ENABLE_DHE || ENABLE_ANON

  if (dhe != 0)
    grub_printf ("- Ephemeral DH using prime of %d bits\n",
		 gnutls_dh_get_prime_bits (session));

#endif /* ENABLE_DHE || ENABLE_ANON */

  /* print the protocol's name (ie TLS 1.0) */

  grub_printf ("- Protocol: %s\n",
	       gnutls_protocol_get_name (gnutls_protocol_get_version
					 (session)));

  grub_printf ("- Certificate Type: %s\n",
	       gnutls_certificate_type_get_name (gnutls_certificate_type_get
						 (session)));

  /* print the name of the cipher used.  ie 3DES. */

  grub_printf ("- Cipher: %s\n",
	       gnutls_cipher_get_name (gnutls_cipher_get (session)));

  grub_printf ("- MAC: %s\n", gnutls_mac_get_name (gnutls_mac_get (session)));

  grub_printf ("- Max size: %lu\n",
	       (ulong_t) gnutls_record_get_max_size (session));
}


static int
https_print_x509cred (gnutls_certificate_credentials_t const x509cred)
{
  uint_t crton;
  int status = 1;
  int ret;
  size_t dataon;
  unsigned const char *dataptr;
  gnutls_x509_privkey_t get_key;
  gnutls_x509_crt_t *get_crts;
  unsigned n_get_crts;
  gnutls_x509_trust_list_t trust_list;
  gnutls_x509_trust_list_iter_t trust_iter;
  gnutls_x509_crt_t get_ca_crt;
  unsigned n_get_ca_crts;
  gnutls_datum_t get_datum;

  ret = gnutls_certificate_get_x509_key (x509cred, 0, &get_key);
  if (ret < 0)
    {
      grub_printf ("unable to get x509 key\n"
		   "System message: %s\n", gnutls_errormsg (status));
      status = 0;
    }
  else
    {
      ret = gnutls_x509_privkey_export2 (get_key,
					 GNUTLS_X509_FMT_PEM, &get_datum);
      if (ret < 0)
	{
	  grub_printf ("unable to get key data\n"
		       "System message: %s\n", gnutls_errormsg (status));
	  status = 0;
	}

      else
	{
	  grub_printf ("x509 key:\n");
	  for (dataon = 0, dataptr = get_datum.data;
	       dataon < get_datum.size; dataon++, dataptr++)
	    grub_printf ("%c", *dataptr);
	  grub_printf ("\n");
	  gnutls_free (get_datum.data);
	}

      gnutls_x509_privkey_deinit (get_key);
    }

  ret = gnutls_certificate_get_x509_crt (x509cred, 0, &get_crts, &n_get_crts);
  if (ret < 0)
    {
      grub_printf ("unable to get x509 crt\n"
		   "System message: %s\n", gnutls_errormsg (status));
      status = 0;
    }

  else if (n_get_crts == 0)
    {
      grub_printf ("no x509 crts found\n");
      status = 0;
    }

  else
    {
      for (crton = 0; crton < n_get_crts; crton++)
	{
	  ret = gnutls_x509_crt_export2 (get_crts[0],
					 GNUTLS_X509_FMT_PEM, &get_datum);
	  if (ret < 0)
	    {
	      grub_printf ("unable to get crt[%u] data\n"
			   "System message: %s\n",
			   crton, gnutls_errormsg (status));
	      status = 0;
	    }

	  else
	    {
	      grub_printf ("x509 crt[%u]:\n", crton);
	      for (dataon = 0, dataptr = get_datum.data;
		   dataon < get_datum.size; dataon++, dataptr++)
		grub_printf ("%c", *dataptr);
	      grub_printf ("\n");
	      gnutls_free (get_datum.data);
	    }

	  gnutls_x509_crt_deinit (get_crts[crton]);
	}

      gnutls_free (get_crts);
    }

  gnutls_certificate_get_trust_list (x509cred, &trust_list);

  n_get_ca_crts = 0;
  trust_iter = NULL;
  while (gnutls_x509_trust_list_iter_get_ca (trust_list,
					     &trust_iter,
					     &get_ca_crt) !=
	 GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
    {
      ret = gnutls_x509_crt_export2 (get_ca_crt,
				     GNUTLS_X509_FMT_PEM, &get_datum);
      if (ret < 0)
	{
	  grub_printf ("unable to get trust[%u] data\n"
		       "System message: %s\n",
		       n_get_ca_crts, gnutls_errormsg (status));
	  status = 0;
	}

      else
	{
	  grub_printf ("trust[%u]:\n", n_get_ca_crts);
	  for (dataon = 0, dataptr = get_datum.data;
	       dataon < get_datum.size; dataon++, dataptr++)
	    grub_printf ("%c", *dataptr);
	  grub_printf ("\n");
	  gnutls_free (get_datum.data);
	}

      gnutls_x509_crt_deinit (get_ca_crt);
      ++n_get_ca_crts;
    }

  return (status);
}


static const char *
gnutls_errormsg (int const errnum)
{
  const char *msg;
  static char buffer[64];

  msg = gnutls_strerror (errnum);
  if (msg != NULL)
    return (msg);
  grub_snprintf (buffer, sizeof (buffer), "Error %d", errnum);
  return (buffer);
}


static grub_err_t
https_seek (grub_file_t const file, grub_off_t const off)
{
  struct https_data *const old_data = file->data;
  struct https_data *data;
  grub_uint16_t port;
  grub_err_t err;
  int i;

  if (old_data->filename != NULL && old_data->sock != NULL &&
      old_data->state == STATE_BODY && old_data->filepos == off)
    return (GRUB_ERR_NONE);

  if (old_data->sock != NULL)
    {
      grub_net_tcp_close (old_data->sock, GRUB_NET_TCP_ABORT);
      old_data->sock = NULL;
    }

  while (file->device->net->packs.first)
    {
      grub_netbuff_free (file->device->net->packs.first->nb);
      grub_net_remove_packet (file->device->net->packs.first);
    }

  file->device->net->stall = 0;
  file->device->net->offset = off;

  data = grub_zalloc (sizeof (*data));
  if (!data)
    return grub_errno;

  data->size_recv = 1;
  data->filename = old_data->filename;
  if (!data->filename)
    {
      grub_free (data);
      file->data = NULL;
      return grub_errno;
    }
  grub_free (old_data);

  file->data = data;

  coroutine_create (&data->coroutine, https_establish, (void *) file,
		    (void *) &data->stack, sizeof (data->stack));

  if (file->device->net->port != 0)
    port = file->device->net->port;
  else
    port = HTTPS_PORT;

  data->sock = grub_net_tcp_open (file->device->net->server,
				  port, https_receive, https_err, https_err,
				  file);

  if (!data->sock)
    {
      return grub_errno;
    }

  data->offset = off;
  data->initial = 0;

  coroutine_resume (current_coroutine, &data->coroutine);

  for (i = 0; i < 100; i++)
    {
      if (data->headers_recv)
	break;

      grub_net_tcp_retransmit ();

      if (data->headers_recv)
	break;

      grub_net_poll_cards (300, &data->headers_recv);
    }

  err = GRUB_ERR_NONE;

  if (!data->headers_recv)
    {
      err = data->err;
      if (err)
	{
	  grub_error (err, "%s", data->errmsg);
	}
      else
	err = grub_error (GRUB_ERR_TIMEOUT,
			  _("file `%s', timed out opening"), data->filename);

      https_finish (file);
    }

  return (err);
}

static grub_err_t
https_open (grub_file_t const file, const char *const filename)
{
  grub_err_t err;
  grub_uint16_t port;
  struct https_data *data;
  int i;

  data = grub_zalloc (sizeof (*data));
  if (!data)
    return (grub_errno);

  file->size = (grub_off_t) GRUB_FILE_SIZE_UNKNOWN;

  data->filename = grub_strdup (filename);
  if (!data->filename)
    {
      grub_free (data);
      return grub_errno;
    }

  file->not_easily_seekable = 0;
  file->data = data;

  coroutine_create (&data->coroutine, https_establish, (void *) file,
		    (void *) &data->stack, sizeof (data->stack));

  if (file->device->net->port != 0)
    port = file->device->net->port;
  else
    port = HTTPS_PORT;

  data->sock = grub_net_tcp_open (file->device->net->server,
				  port, https_receive, https_err, https_err,
				  file);

  if (!data->sock)
    {
      return grub_errno;
    }

  data->offset = 0;
  data->initial = 1;

  coroutine_resume (current_coroutine, &data->coroutine);

  for (i = 0; i < 100; i++)
    {
      if (data->headers_recv)
	break;

      grub_net_tcp_retransmit ();

      if (data->headers_recv)
	break;

      grub_net_poll_cards (300, &data->headers_recv);
    }

  err = GRUB_ERR_NONE;

  if (!data->headers_recv)
    {
      err = data->err;
      if (err)
	{
	  grub_error (err, "%s", data->errmsg);
	}
      else
	err = grub_error (GRUB_ERR_TIMEOUT,
			  _("file `%s', timed out opening"), data->filename);

      https_finish (file);
    }

  return (err);
}


static grub_err_t
https_close (grub_file_t const file)
{
  struct https_data *const data = file->data;
  grub_err_t err;

  err = https_finish (file);

  if (data)
    {
      if (data->filename)
	grub_free (data->filename);

      if (data->errmsg != NULL)
	grub_free (data->errmsg);

      grub_free (data);
    }

  return (err);
}

static grub_err_t
https_packets_pulled (grub_file_t const file)
{
  https_data_t const data = file->data;

  if (file->device->net->packs.count >= 20)
    return (GRUB_ERR_NONE);

  if (!file->device->net->eof)
    file->device->net->stall = 0;
  if (data && data->sock)
    grub_net_tcp_unstall (data->sock);

  return (GRUB_ERR_NONE);
}

static grub_err_t
https_finish (grub_file_t const file)
{
  https_data_t const data = file->data;
  grub_net_t net;
  struct grub_net_packet *pack;
  grub_off_t size;

  if (file->size == (grub_off_t) GRUB_FILE_SIZE_UNKNOWN)
    {
      net = file->device->net;
      size = net->offset;

      for (pack = net->packs.first; pack != NULL; pack = pack->next)
	size += pack->nb->tail - pack->nb->data;

      file->size = size;
    }

  if (data == NULL)
    return (GRUB_ERR_NONE);

  while (data->pkts.first != NULL)
    grub_net_remove_packet (data->pkts.first);

  if (data->sock)
    {
      grub_net_tcp_close (data->sock, GRUB_NET_TCP_ABORT);
      data->sock = NULL;
    }

  return (data->err);
}


ssize_t
https_recv (gnutls_transport_ptr_t f, void *buf, size_t len)
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t const data = file->data;
  grub_net_packet_t *pkt;
  struct grub_net_buff *nb;
  size_t left = len;
  char *bufptr = (char *) buf;
  size_t size;

  if (data == NULL)
    {
      grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
      return (-1);
    }

  for (;;)
    {
      if (data->err != GRUB_ERR_NONE)
	{
	  grub_error (data->err, "%s", data->errmsg);
	  return (-1);
	}

      if (data->sock == NULL)
	{
	  grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
	  return (-1);
	}

      if (data->rcv_data)
	break;

      coroutine_resume (current_coroutine, &main_routine);
    }

  while (left > 0 && (pkt = data->pkts.first) != NULL)
    {
      nb = pkt->nb;
      size = nb->tail - nb->data;
      if (size > left)
	size = left;
      grub_memcpy (bufptr, nb->data, size);
      left -= size;
      bufptr += size;
      grub_netbuff_pull (nb, size);
      if (nb->tail == nb->data)
	{
	  grub_netbuff_free (nb);
	  grub_net_remove_packet (pkt);
	  if (data->pkts.first == NULL)
	    data->rcv_data = 0;
	}
      else
	break;
    }

  return (len - left);
}

int
https_recv_timeout (gnutls_transport_ptr_t const f,
		    unsigned int ms __attribute__ ((unused)))
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t data = file->data;

  if (data == NULL)
    {
      grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
      return (-1);
    }

  for (;;)
    {
      if (data->err != GRUB_ERR_NONE)
	{
	  grub_error (data->err, "%s", data->errmsg);
	  return (-1);
	}

      if (data->sock == NULL)
	{
	  grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
	  return (-1);
	}

      if (data->rcv_data)
	break;

      coroutine_resume (current_coroutine, &main_routine);
    }

  return (1);
}


ssize_t
https_writev (gnutls_transport_ptr_t f,
	      giovec_t const *const iovec, int iovec_cnt)
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t const data = file->data;
  struct grub_net_buff *nb;
  giovec_t const *iovptr;
  int iovon;
  grub_uint8_t *ptr;
  size_t size = 0;
  int err;

  if (data == NULL || data->sock == NULL)
    {
      grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
      return (-1);
    }

  for (iovon = 0, iovptr = iovec; iovon < iovec_cnt; iovon++, iovptr++)
    size += iovptr->iov_len;

  nb = grub_netbuff_alloc (GRUB_NET_TCP_RESERVE_SIZE + size);
  if (nb == NULL)
    return (-1);

  grub_netbuff_reserve (nb, GRUB_NET_TCP_RESERVE_SIZE);

  for (iovon = 0, iovptr = iovec; iovon < iovec_cnt; iovon++, iovptr++)
    {
      ptr = nb->tail;
      grub_netbuff_put (nb, iovptr->iov_len);
      grub_memcpy (ptr, iovptr->iov_base, iovptr->iov_len);
    }

  err = grub_net_send_tcp_packet (data->sock, nb, 1);
  if (err)
    {
      grub_net_tcp_close (data->sock, GRUB_NET_TCP_ABORT);
      data->sock = NULL;
      grub_errno = err;
      return (-1);
    }

  return (size);
}


ssize_t
https_write (gnutls_transport_ptr_t f,
	     const void *buffer, size_t size)
{
  grub_file_t const file = (grub_file_t) f;
  https_data_t const data = file->data;
  struct grub_net_buff *nb;
  grub_uint8_t *ptr;
  grub_err_t err;

  if (data == NULL || data->sock == NULL)
    {
      grub_error (GRUB_ERR_NET_PORT_CLOSED, _("file closed"));
      return (-1);
    }

  nb = grub_netbuff_alloc (GRUB_NET_TCP_RESERVE_SIZE + size);
  if (nb == NULL)
    return (-1);

  grub_netbuff_reserve (nb, GRUB_NET_TCP_RESERVE_SIZE);

  ptr = nb->tail;
  grub_netbuff_put (nb, size);
  grub_memcpy (ptr, buffer, size);

  err = grub_net_send_tcp_packet (data->sock, nb, 1);
  if (err)
    {
      grub_net_tcp_close (data->sock, GRUB_NET_TCP_ABORT);
      data->sock = NULL;
      grub_errno = err;
      return -1;
    }

  return (size);
}


int
https_set_errno (gnutls_transport_ptr_t f __attribute__ ((unused)))
{
  size_t erron;
  const error_convert_t *errptr;

  for (erron = 0, errptr = error_convert;
       erron < sizeof (error_convert) / sizeof (error_convert[0]);
       erron++, errptr++)
    {
      if (grub_errno == errptr->ec_grub_error)
	return (errptr->ec_unix_error);
    }
  return (EIO);
}


static struct grub_net_app_protocol grub_https_protocol = {
  .name = "https",
  .open = https_open,
  .close = https_close,
  .seek = https_seek,
  .packets_pulled = https_packets_pulled
};

GRUB_MOD_INIT (https)
{
  grub_net_app_level_register (&grub_https_protocol);
}

GRUB_MOD_FINI (https)
{
  grub_net_app_level_unregister (&grub_https_protocol);
}


#ifndef lint

#ifndef GRUB_TARGET_CPU_X86_64

typedef struct coroutine_stack_struct
{
  uintptr_t cs_eax;		/* 0x00 */
  uintptr_t cs_ebx;		/* 0x04 */
  uintptr_t cs_ecx;		/* 0x08 */
  uintptr_t cs_edx;		/* 0x0c */
  uintptr_t cs_esi;		/* 0x10 */
  uintptr_t cs_edi;		/* 0x18 */
  uintptr_t cs_ebp;		/* 0x1c */
  uintptr_t cs_retpc;		/* 0x20 */
  uintptr_t cs_newpc;		/* 0x24 */
  uintptr_t cs_arg;		/* 0x28 */
} coroutine_stack_t;		/* 0x2c */

void
coroutine_create (coroutine_t * const new_coroutine,
		  cofunction_t const function,
		  void *const arg,
		  void *const stackstart, size_t const stacksize)
{
  coroutine_stack_t *stackptr = (coroutine_stack_t *)
    ((char *) stackstart + stacksize - sizeof (coroutine_stack_t));
  unsigned long align = (unsigned long) stackptr & 0xf;

  new_coroutine->cr_function = function;
  new_coroutine->cr_arg = arg;
  new_coroutine->cr_stackstart = stackstart;
  new_coroutine->cr_stacksize = stacksize;
  new_coroutine->cr_stackptr = (void *) ((char *) stackptr - align);

  stackptr->cs_eax = 0;
  stackptr->cs_ebx = 0;
  stackptr->cs_ecx = 0;
  stackptr->cs_edx = 0;
  stackptr->cs_esi = 0;
  stackptr->cs_edi = 0;
  stackptr->cs_ebp = 0;
  stackptr->cs_retpc = (uintptr_t) function;
  stackptr->cs_newpc = 0;
  stackptr->cs_arg = (uintptr_t) arg;
}


void
coroutine_resume_stub (void)
{
  asm ("  .globl  coroutine_resume           ");
  asm ("coroutine_resume:                    ");
  asm ("  .type   coroutine_resume, @function");
  asm ("  pushl   %ebp                       ");  /* save frame */
  asm ("  movl    %esp, %ebp                 ");  /* set up a stack frame */
  asm ("  pushl   %edi			     ");  /* save edi */
  asm ("  pushl   %esi			     ");  /* save esi */
  asm ("  pushl   %edx			     ");  /* save edx */
  asm ("  pushl   %ecx			     ");  /* save ecx */
  asm ("  pushl   %ebx                       ");  /* save ebx */
  asm ("  pushl   %eax                       ");  /* save eax */
  asm ("  movl    8(%ebp), %esi              ");  /* source coroutine */
  asm ("  movl    12(%ebp), %edi             ");  /* new coroutine */
  asm ("  movl    %esp, 0x10(%esi)           ");  /* save our stack */
  asm ("  movl    %edi, current_coroutine    ");  /* point at new coroutine */
  asm ("  movl    0x10(%edi), %esp           ");  /* on new stack */
  asm ("  popl    %eax                       ");  /* restore eax */
  asm ("  popl    %ebx                       ");  /* restore ebx */
  asm ("  popl    %ecx                       ");  /* restore ecx */
  asm ("  popl    %edx                       ");  /* restore edx */
  asm ("  popl    %esi                       ");  /* restore esi */
  asm ("  popl    %edi                       ");  /* restore edi */
  asm ("  popl    %ebp			     ");  /* restore ebp */
  asm ("  ret				     ");  /* return */
  asm ("  .size   coroutine_resume, . - coroutine_resume");
}

#else /* GRUB_TARGET_CPU_X86_64 */

typedef struct coroutine_stack_struct
{
  uintptr_t cs_rax;		/* 0x00 */
  uintptr_t cs_rbx;		/* 0x08 */
  uintptr_t cs_rcx;		/* 0x10 */
  uintptr_t cs_rdx;		/* 0x18 */
  uintptr_t cs_rsi;		/* 0x20 */
  uintptr_t cs_rdi;		/* 0x28 */
  uintptr_t cs_r8;		/* 0x30 */
  uintptr_t cs_r9;		/* 0x38 */
  uintptr_t cs_r10;		/* 0x40 */
  uintptr_t cs_r11;		/* 0x48 */
  uintptr_t cs_r12;		/* 0x50 */
  uintptr_t cs_r13;		/* 0x58 */
  uintptr_t cs_r14;		/* 0x60 */
  uintptr_t cs_r15;		/* 0x68 */
  uintptr_t cs_rbp;		/* 0x70 */
  uintptr_t cs_retpc;		/* 0x78 */
  uintptr_t cs_newpc;		/* 0x80 */
} coroutine_stack_t;		/* 0x88 */

void
coroutine_create (coroutine_t * const new_coroutine,
		  cofunction_t const function,
		  void *const arg,
		  void *const stackstart, size_t const stacksize)
{
  coroutine_stack_t *stackptr = (coroutine_stack_t *)
    ((char *) stackstart + stacksize - sizeof (coroutine_stack_t));
  unsigned long align = (unsigned long) stackptr & 0xf;

  stackptr = (coroutine_stack_t *) ((char *) stackptr - align -
				    sizeof (uintptr_t));

  new_coroutine->cr_function = function;
  new_coroutine->cr_arg = arg;
  new_coroutine->cr_stackstart = stackstart;
  new_coroutine->cr_stacksize = stacksize;
  new_coroutine->cr_stackptr = (void *) stackptr;

  stackptr->cs_rax = 0;
  stackptr->cs_rbx = 0;
  stackptr->cs_rcx = 0;
  stackptr->cs_rdx = 0;
  stackptr->cs_rsi = 0;
  stackptr->cs_rdi = (uintptr_t) arg;
  stackptr->cs_r8 = 0;
  stackptr->cs_r9 = 0;
  stackptr->cs_r10 = 0;
  stackptr->cs_r11 = 0;
  stackptr->cs_r12 = 0;
  stackptr->cs_r13 = 0;
  stackptr->cs_r14 = 0;
  stackptr->cs_r15 = 0;
  stackptr->cs_rbp = 0;
  stackptr->cs_retpc = (uintptr_t) function;
  stackptr->cs_newpc = 0;
}


void
coroutine_resume_stub (void)
{
  asm ("  .globl  coroutine_resume           ");
  asm ("coroutine_resume:                    ");
  asm ("  .type   coroutine_resume, @function");
  asm ("  pushq   %rbp                       ");  /* save frame */
  asm ("  movq    %rsp, %rbp                 ");  /* set up a stack frame */
  asm ("  pushq   %r15			     ");  /* save r15 */
  asm ("  pushq   %r14			     ");  /* save r14 */
  asm ("  pushq   %r13			     ");  /* save r13 */
  asm ("  pushq   %r12			     ");  /* save r12 */
  asm ("  pushq   %r11			     ");  /* save r11 */
  asm ("  pushq   %r10			     ");  /* save r10 */
  asm ("  pushq   %r9			     ");  /* save r9 */
  asm ("  pushq   %r8			     ");  /* save r8 */
  asm ("  pushq   %rdi			     ");  /* save rdi */
  asm ("  pushq   %rsi			     ");  /* save rsi */
  asm ("  pushq   %rdx			     ");  /* save rdx */
  asm ("  pushq   %rcx			     ");  /* save rcx */
  asm ("  pushq   %rbx                       ");  /* save rbx */
  asm ("  pushq   %rax                       ");  /* save rax */
  asm ("  movq    %rsp, 0x20(%rdi)           ");  /* save our stack */
  asm ("  movabsq $current_coroutine, %rax   ");  /* point at new coroutine */
  asm ("  movq    %rsi, (%rax)               ");
  asm ("  movq    0x20(%rsi), %rsp           ");  /* on new stack */
  asm ("  popq    %rax                       ");  /* restore eax */
  asm ("  popq    %rbx                       ");  /* restore ebx */
  asm ("  popq    %rcx                       ");  /* restore ecx */
  asm ("  popq    %rdx                       ");  /* restore edx */
  asm ("  popq    %rsi                       ");  /* restore esi */
  asm ("  popq    %rdi                       ");  /* restore edi */
  asm ("  popq    %r8                        ");  /* restore r8 */
  asm ("  popq    %r9                        ");  /* restore r9 */
  asm ("  popq    %r10                       ");  /* restore r10 */
  asm ("  popq    %r11                       ");  /* restore r11 */
  asm ("  popq    %r12                       ");  /* restore r12 */
  asm ("  popq    %r13                       ");  /* restore r13 */
  asm ("  popq    %r14                       ");  /* restore r14 */
  asm ("  popq    %r15                       ");  /* restore r15 */
  asm ("  popq    %rbp			     ");  /* restore rbp */
  asm ("  ret				     ");  /* return */
  asm ("  .size   coroutine_resume, . - coroutine_resume");
}


#endif /* GRUB_TARGET_CPU_X86_64 */

#else /* lint */

/* ARGSUSED */
void
coroutine_create (coroutine_t * const new_coroutine,
		  cofunction_t const function,
		  void *const arg,
		  void *const stackstart, size_t const stacksize)
{
}

void
corroutine_resume_stub (void)
{
}

/* ARGSUSED */
void
coroutine_resume (coroutine_t * const current_coroutine,
		  const coroutine_t * const new_coroutine)
{
}

#endif /* lint */

extern void abort (void);

extern void __assert (const char *assertion,
		      const char *filename, int line_num);

extern void __assert_c99 (const char *assertion,
			  const char *filename, int line_num,
			  char const *function);

extern int abs (int x);

extern int atoi (const char *str);

extern long atol (const char *str);

extern char *basename (char *path);

extern void *calloc (size_t nelem, size_t elsize);

extern int close (int f);

extern int closedir (void *dirp);

extern void exit (int status);

extern int *___errno (int unused);

extern int fclose (FILE * stream);

extern int ferror (FILE * stream);

extern int fflush (FILE * stream);

extern char *fgets (char *s, int n, FILE * stream);

extern int fileno (FILE * stream);

extern FILE *fopen (const char *filename, const char *mode);

extern FILE *fopen64 (const char *filename, const char *mode);

extern int setvbuf (FILE *stream, char *buf, int mode, size_t size);

extern int fprintf (FILE * stream, const char *format, ...)
__attribute__ ((format (printf, 2, 3)));

extern int fputc (int c, FILE * stream);

extern int fputs (const char *s, FILE * stream);

extern size_t fread (void *ptr, size_t size, size_t nitems, FILE * stream);

extern void free (void *p);

extern int fscanf (FILE * stream, const char *format
		   __attribute__ ((unused)), ...)
		   __attribute__ ((format (scanf, 2, 3)));

extern int fstat (int file, void *statbuf);

extern int fstat64 (int file, void *statbuf);

extern long ftello (FILE * stream);

extern long long ftello64 (FILE * stream);

size_t fwrite (const void *ptr, size_t size, size_t nitems, FILE * p);

extern int getc (FILE * stream);

extern int getentropy (void *buf, size_t buflen);

extern char *getenv (const char *name);

extern ssize_t
getline (char **lineptr, size_t * n, FILE * stream);

extern int gettimeofday (struct timeval *tp, void *tzp);

extern long getpid (void);

extern struct tm *gmtime_r (const time_t * clock, struct tm *res);

extern int inet_aton (const char *cp, void *addr);

extern int inet_pton (int af, const char *cp, void *addr);

extern int kill (long pid, int sig);

extern void *malloc (size_t nbytes);

extern void *memchr (const void *s, int c_in, size_t n);

extern void *_memcpy (void *s1, const void *s2, size_t n);

extern void *memmem (const void *haystack, size_t haystacklen,
		     const void *needle, size_t needlelen);

extern void *_memset (void *s, int c, size_t n);

extern int mkdir (const char *path, unsigned int mode);

extern int nanosleep (const struct timespec *rqtp, struct timespec *rmtp);

extern int open (const char *path, int oflag, ...);

extern void *opendir (const char *dirname);

extern int poll (void *fds, unsigned long nfds, int timeout);

extern int printf (const char *format, ...)
__attribute__ ((format (printf, 1, 2)));

extern int puts (const char *s);

extern void qsort (void *base,
		   size_t nel,
		   size_t width, int (*compar) (const void *, const void *));

extern int raise (int sig);

extern ssize_t read (int f, void *buffer, size_t size);

extern void *readdir (void *dirp);

extern void *readdir_r (void *dirp, void *entry, void *result);

extern void *readdir64_r (void *dirp, void *entry, void *result);

extern void *realloc (void *p, size_t n);

extern ssize_t recv (int f, void *buf, size_t len, int flags);

extern int scanf (const char *format __attribute__ ((unused)), ...)
		   __attribute__ ((format (scanf, 1, 2)));

extern int select (int nfds,
		   void *readfds,
		   void *writefds, void *errorfds, void *timeout);

extern ssize_t send (int s, const void *msg, size_t len, int flags);

extern int snprintf (char *s, size_t n, const char *format, ...)
__attribute__ ((format (printf, 3, 4)));

extern int socket (int domain, int type, int protocol);

extern int sprintf (char *s, const char *format, ...)
__attribute__ ((format (printf, 2, 3)));

extern int sscanf (const char *s, const char *format, ...);

extern char *stpcpy (char *s1, const char *s2);

extern int strcasecmp (const char *s1, const char *s2);

extern char *strcat (char *s1, const char *s2);

extern char *strchr (const char *s1, int c);

extern int strcmp (const char *s1, const char *s2);

extern char *strcpy (char *s1, const char *s2);

extern size_t strcspn (const char *s1, const char *s2);

extern char *strdup (const char *s);

extern char *strerror (int errnum);

extern size_t strftime (char *s,
			size_t maxsize, const char *format, struct tm *tmptr);

static int strftime_process (char *s, size_t maxsize, char defflag,
			     const char *format, struct tm *tmptr);

static int strftime_value (char *s, size_t maxsize, char flag,
			   int width, int value);

static int strftime_entry (char *s, size_t maxsize, char flag,
			   const char *const *table, size_t tabsize,
			   int width, int value);

static int strftime_string (char *s, size_t maxsize, int width,
			    const char *entry);

extern size_t strlen (const char *s);

extern int strncasecmp (const char *s1, const char *s2, size_t n);

extern int strncmp (const char *s1, const char *s2, size_t n);

extern char *strncat (char *s1, const char *s2, size_t n);

extern char *strncpy (char *s1, const char *s2, size_t n);

extern char *strndup (const char *s, size_t n);

extern char *strpbrk (const char *s1, const char *s2);

extern char *strrchr (const char *s1, int c);

extern size_t strspn (const char *s1, const char *s2);

extern char *strstr (const char *s1, const char *s2);

extern long strtol (const char *str, char **endptr, int base);

extern unsigned long strtoul (const char *str, char **endptr, int base);

extern unsigned long long strtoull (const char *str, char **endptr, int base);

extern time_t time (time_t * tloc);

extern int vasprintf (char **ret, const char *format, va_list ap);

extern int vfprintf (FILE * stream, const char *format, va_list ap);

extern int vprintf (const char *format, va_list ap);

extern int vsnprintf (char *s, size_t n, const char *format, va_list ap);

extern int vsprintf (char *s, const char *format, va_list ap);

extern int vsscanf (const char *s, const char *format, va_list arg);

extern unsigned long long __umoddi3 (unsigned long long s,
				     unsigned long long d);

extern ssize_t write (int f, const void *buffer, size_t size);

extern ssize_t writev (int f, const giovec_t * iovec, int iovec_cnt);

extern ssize_t __xnet_sendmsg(int s, const void *msg, int flags);

extern unsigned char __ctype[SZ_TOTAL];

extern FILE __iob[3];

extern char **environ;

void
abort ()
{
  grub_abort ();
}

void
__assert (const char *assertion, const char *filename, int line_num)
{
  grub_printf ("Assertion failed: %s, file %s, line %d\n",
	       assertion, filename, line_num);
  grub_abort ();
}

void
__assert_c99 (const char *assertion, const char *filename, int line_num,
	      char const *function)
{
  grub_printf ("Assertion failed: %s, file %s, line %d, function %s\n",
	       assertion, filename, line_num, function);
  grub_abort ();
}

int
abs (int x)
{
  return (grub_abs (x));
}

int
atoi (const char *str)
{
  return ((int) grub_strtol (str, NULL, 10));
}

long
atol (const char *str)
{
  return (grub_strtol (str, NULL, 10));
}

char *
basename (char *path)
{
  char *s;

  for (s = path; *s != 0; s++);
  for (s--; s >= path && *s != '/'; s--);
  return (s + 1);
}

void *
calloc (size_t nelem, size_t elsize)
{
  return (grub_zalloc (nelem * elsize));
}

int
close (int f __attribute__ ((unused)))
{
  return (0);
}

int
closedir (void *dirp __attribute__ ((unused)))
{
  return (-1);
}

void
exit(int status __attribute__ ((unused)))
{
}

int *
___errno (int unused __attribute__ ((unused)))
{
  return (&https_errno);
}

int
fclose (FILE * stream __attribute__ ((unused)))
{
  return (-1);
}

int
ferror (FILE * stream)
{
  grub_error (GRUB_ERR_BUG, "unsupported function ferror (0x%lx) called",
    (unsigned long) stream);
  return (-1);
}

int
fflush(FILE *stream __attribute__ ((unused)))
{
  return 0;
}

char *
fgets(char *s, int n, FILE *stream)
{
  grub_error (GRUB_ERR_BUG, "unsupported function fgets(0x%lx, %d, 0x%lx) called",
    (unsigned long) s, n, (unsigned long) stream);
  return (NULL);
}

int
fileno (FILE * stream)
{
  grub_error (GRUB_ERR_BUG, "unsupported function fileno (0x%lx) called",
    (unsigned long) stream);
  return (-1);
}

FILE *
fopen (const char *filename __attribute__ ((unused)),
       const char *mode __attribute__ ((unused)))
{
  return (NULL);
}

FILE *
fopen64 (const char *filename __attribute__ ((unused)),
	 const char *mode __attribute__ ((unused)))
{
  return (NULL);
}

int
setvbuf (FILE *stream, char *buf, int mode, size_t size)
{
  grub_error (GRUB_ERR_BUG, "unsupported function setvbuf(0x%lx, 0x%lx, %d, %zu) called",
    (unsigned long) stream, (unsigned long) buf, mode, size);
  return (-1);
}

int
fprintf (FILE * stream, const char *format, ...)
{
  int count;
  va_list ap;

  if (stream == stdin || stream == stdout)
    {
      va_start (ap, format);
      count = grub_vprintf (format, ap);
      va_end (ap);
      return (count);
    }
  grub_error (GRUB_ERR_BUG,
	      "unsupported function fprintf (0x%lx, \"%s\", ...) called",
	      (unsigned long) stream, format);
  return (-1);
}

int
fscanf (FILE * stream __attribute__ ((unused)), const char *format
	__attribute__ ((unused)), ...)
{
  return (-1);
}

int
fputc (int c, FILE * stream)
{
  char buffer[2];

  if (stream == stdin || stream == stdout)
    {
      buffer[0] = (char) c;
      buffer[1] = 0;
      grub_xputs (buffer);
      return (1);
    }
  grub_error (GRUB_ERR_BUG, "unsupported function fputc (0x%x, 0x%lx) called",
	      c, (unsigned long) stream);
  return (-1);
}

int
fputs (const char *s, FILE * stream)
{
  if (stream == stdout || stream == stderr)
    {
      grub_xputs (s);
      return (grub_strlen (s));
    }

  grub_error (GRUB_ERR_BUG,
	      "unsupported function fputs (\"%s\", 0x%lx) called", s,
	      (unsigned long) stream);
  return (-1);
}

size_t
fread (void *ptr, size_t size, size_t nitems, FILE * stream)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function fread (0x%lx, %lu, %lu, 0x%lx) called",
	      (unsigned long) ptr, (unsigned long) size,
	      (unsigned long) nitems, (unsigned long) stream);
  return (0);
}

void
free (void *p)
{
  grub_free (p);
}

int
fstat (int file, void *statbuf)
{
  grub_error (GRUB_ERR_BUG, "unsupported function fstat (%d, 0x%lx) called",
	      file, (unsigned long) statbuf);
  return (-1);
}

int
fstat64 (int file, void *statbuf)
{
  grub_error (GRUB_ERR_BUG, "unsupported function fstat64 (%d, 0x%lx) called",
	      file, (unsigned long) statbuf);
  return (-1);
}

long
ftello (FILE * stream)
{
  grub_error (GRUB_ERR_BUG, "unsupported function ftello (0x%lx) called",
	      (unsigned long) stream);
  return (-1);
}

long long
ftello64 (FILE * stream __attribute__ ((unused)))
{
  grub_error (GRUB_ERR_BUG, "unsupported function ftello64 (0x%lx) called",
	      (unsigned long) stream);
  return (-1);
}

size_t
fwrite (const void *ptr, size_t size, size_t nitems, FILE * stream)
{
  char *bufptr;
  size_t length;

  if (stream == stdout || stream == stdin)
    {
      length = size * nitems;
      bufptr = grub_malloc (length + 1);
      if (bufptr == NULL)
	return (0);
      grub_memcpy (bufptr, ptr, length);
      bufptr[length] = 0;
      grub_xputs (bufptr);
      grub_free (bufptr);
      return (length);
    }

  return (0);
}

int
getc (FILE * stream)
{
  grub_error (GRUB_ERR_BUG, "unsupported function getc (0x%lx) called",
	      (unsigned long) stream);
  return (-1);
}

char const *entropy_filename = "(wbfs)/dev/urandom";
grub_file_t entropy_file;

int
getentropy (void *buf, size_t buflen)
{
  char *bufptr = (char *) buf;
  grub_size_t bufleft = buflen;
  grub_ssize_t readsize;

  if (entropy_file == NULL)
    {
      entropy_file = grub_file_open (entropy_filename);
      if (entropy_file == NULL)
	{
	  grub_fatal ("%s %d: unable to open entropy file `%s'\n",
		      __FILE__, __LINE__, entropy_filename);
	  return (-1);
	}
    }

  while (bufleft > 0)
    {
      readsize = grub_file_read (entropy_file, bufptr, bufleft);

      if (readsize == 0)
	{
	  if (grub_file_seek (entropy_file, 0) == (grub_off_t) -1)
	    {
	      grub_fatal ("%s %d: unable to seek entropy file `%s'\n",
			  __FILE__, __LINE__, entropy_filename);
	      return (-1);
	    }

	  readsize = grub_file_read (entropy_file, bufptr, bufleft);
	}

      if (readsize <= 0)
	{
	  grub_fatal ("%s %d: unable to read entropy file `%s'\n",
		      __FILE__, __LINE__, entropy_filename);
	  return (-1);
	}

      bufptr += readsize;
      bufleft -= readsize;
    }

  return (0);
}

char *
getenv (const char *name __attribute__ ((unused)))
{
  return (NULL);
}

ssize_t
getline (char **lineptr, size_t * n, FILE * stream)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function getline (0x%lx, 0x%lx, 0x%lx) called",
	      (unsigned long) lineptr, (unsigned long) n,
	      (unsigned long) stream);
  return (-1);
}

int
gettimeofday (struct timeval *tp, void *tzp __attribute__ ((unused)))
{
  time_t oldms;
  time_t newms;
  time_t curtime;

  oldms = grub_get_time_ms ();
  curtime = time (NULL);
  newms = grub_get_time_ms ();
  if (newms < oldms)
     curtime++;

  tp->tv_sec = curtime;
  tp->tv_usec = (newms % 1000) * 1000;

  return (0);
}

long
getpid (void)
{
  return (-1);
}

struct tm *
gmtime_r (const time_t * clock, struct tm *res)
{
  time_t curtime = *clock;
  time_t days = curtime / 60 / 60 / 24;
  time_t ydays;
  int years = 70;
  int isleap = 0;

  res->tm_sec = curtime % 60;
  res->tm_min = curtime / 60 % 60;
  res->tm_hour = curtime / 60 / 60 % 24;

  years += (days / (365 * 4 + 1) * 4);
  ydays = days % (365 * 4 + 1);
  if (ydays >= 365)
    {
      ydays -= 365;
      years++;
      if (ydays >= 365)
	{
	  ydays -= 365;
	  years++;
	  if (ydays >= 366)
	    {
	      ydays -= 366;
	      years++;
	    }
	  else if (ydays >= 59)
	    isleap = 1;
	}
    }

  if (ydays < 31)
    {
      res->tm_mday = ydays + 1;
      res->tm_mon = 0;
    }
  else if (ydays < 59 + isleap)
    {
      res->tm_mday = ydays - 31 + 1;
      res->tm_mon = 1;
    }
  else if (ydays < 90 + isleap)
    {
      res->tm_mday = ydays - (59 + isleap) + 1;
      res->tm_mon = 2;
    }
  else if (ydays < 120 + isleap)
    {
      res->tm_mday = ydays - (90 + isleap) + 1;
      res->tm_mon = 3;
    }
  else if (ydays < 151 + isleap)
    {
      res->tm_mday = ydays - (120 + isleap) + 1;
      res->tm_mon = 4;
    }
  else if (ydays < 181 + isleap)
    {
      res->tm_mday = ydays - (151 + isleap) + 1;
      res->tm_mon = 5;
    }
  else if (ydays < 212 + isleap)
    {
      res->tm_mday = ydays - (181 + isleap) + 1;
      res->tm_mon = 6;
    }
  else if (ydays < 243 + isleap)
    {
      res->tm_mday = ydays - (212 + isleap) + 1;
      res->tm_mon = 7;
    }
  else if (ydays < 273 + isleap)
    {
      res->tm_mday = ydays - (243 + isleap) + 1;
      res->tm_mon = 8;
    }
  else if (ydays < 304 + isleap)
    {
      res->tm_mday = ydays - (273 + isleap) + 1;
      res->tm_mon = 9;
    }
  else if (ydays < 334 + isleap)
    {
      res->tm_mday = ydays - (304 + isleap) + 1;
      res->tm_mon = 10;
    }
  else
    {
      res->tm_mday = ydays - (334 + isleap) + 1;
      res->tm_mon = 11;
    }
  res->tm_year = years;
  res->tm_wday = (days + 4) % 7;
  res->tm_yday = (int) ydays;
  res->tm_isdst = 0;

  return (res);
}

int
inet_aton (const char *cp __attribute__ ((unused)),
	   void *addr __attribute__ ((unused)))
{
  return (-1);
}

int
inet_pton (int af __attribute__ ((unused)),
	   const char *cp __attribute__ ((unused)),
	   void *addr __attribute__ ((unused)))
{
  return (-1);
}

int
kill (long pid __attribute__ ((unused)), int sig)
{
  grub_printf ("Killed with signal %d\n", sig);
  grub_abort ();
  return (0);
}

void *
malloc (size_t nbytes)
{
  return (grub_malloc (nbytes));
}

void *
memchr (const void *s, int c_in, size_t n)
{
  return (grub_memchr (s, c_in, n));
}

void *
_memcpy (void *s1, const void *s2, size_t n)
{
  return (grub_memcpy (s1, s2, n));
}

void *
memmem (const void *haystack, size_t haystacklen,
	const void *needle, size_t needlelen)
{
  size_t haystackon;
  const char *haystackptr;
  size_t needleon;
  const char *needleptr = (const char *) needle;

  for (haystackon = 0, haystackptr = (const char *) haystack;
       haystackon + needlelen < haystacklen; haystackon++, haystackptr++)
    {
      for (needleon = 0; needleon < needlelen; needleon++)
	{
	  if (haystackptr[needleon] != needleptr[needleon])
	    break;
	}
      if (needleon >= needlelen)
	return ((void *) haystackptr);
    }
  return (NULL);
}

void *
_memset (void *s, int c, size_t n)
{
  return (grub_memset (s, c, n));
}

int
mkdir (const char *path __attribute__ ((unused)), unsigned int mode
       __attribute__ ((unused)))
{
  return (-1);
}

int
nanosleep (const struct timespec *rqtp, struct timespec *rmtp)
{
  grub_uint64_t start;
  grub_uint64_t end;
  time_t diff;
  grub_uint32_t ms;

  start = grub_get_time_ms ();

  ms = rqtp->tv_sec * 1000 + rqtp->tv_nsec / 1000000;

  grub_millisleep (ms);

  end = grub_get_time_ms ();
  diff = end - start;

  rmtp->tv_sec = diff / 1000;
  rmtp->tv_nsec = (diff % 1000) * 1000000;

  return (0);
}

int
open (const char *path __attribute__ ((unused)), int oflag
      __attribute__ ((unused)), ...)
{
  return (-1);
}

void *
opendir (const char *dirname __attribute__ ((unused)))
{
  return (NULL);
}

int
poll (void *fds, unsigned long nfds, int timeout)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function poll (0x%lx, %lu, %d) called",
	      (unsigned long) fds, nfds, timeout);
  return (-1);
}

int
printf (const char *format, ...)
{
  int count;
  va_list ap;

  va_start (ap, format);
  count = grub_vprintf (format, ap);
  va_end (ap);
  return (count);
}

int
puts (const char *s)
{
  grub_puts (s);
  return ((int) grub_strlen (s));
}

void
qsort (void *base,
       size_t nel, size_t width, int (*compar) (const void *, const void *))
{
  size_t el1;
  char *ptr1;
  char *cptr1;
  size_t el2;
  char *ptr2;
  char *cptr2;
  size_t charon;
  char ch;

  for (el1 = 0; el1 < nel; el1++)
    {
      for (el2 = 0, ptr1 = (char *) base, ptr2 = (char *) base + width;
	   el2 + 1 < nel; el2++, ptr1 += width, ptr2 += width)
	{
	  if (compar (ptr1, ptr2) > 0)
	    {
	      for (charon = 0, cptr1 = ptr1, cptr2 = ptr2;
		   charon < width; charon++, cptr1++, cptr2++)
		{
		  ch = *cptr1;
		  *cptr1 = *cptr2;
		  *cptr2 = ch;
		}
	    }
	}
    }
}

ssize_t
read (int f, void *buf, size_t nbyte)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function read (%d, 0x%lx, %lu) called", f,
	      (unsigned long) buf, (unsigned long) nbyte);
  return (-1);
}

int
raise (int sig)
{
  grub_printf ("Raised signal %d\n", sig);
  grub_abort ();
  return (0);
}

void *
readdir (void *dirp __attribute__ ((unused)))
{
  return (NULL);
}

void *
readdir_r (void *dirp __attribute__ ((unused)),
	   void *entry __attribute__ ((unused)),
	   void *result __attribute__ ((unused)))
{
  return (NULL);
}

void *
readdir64_r (void *dirp __attribute__ ((unused)),
	     void *entry __attribute__ ((unused)),
	     void *result __attribute__ ((unused)))
{
  return (NULL);
}

void *
realloc (void *p, size_t n)
{
  return (grub_realloc (p, n));
}

ssize_t
recv (int f, void *buf, size_t len, int flags)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function recv (%d, 0x%lx, %lu, 0x%x) called", f,
	      (unsigned long) buf, (unsigned long) len, flags);
  return (-1);
}

int
select (int nfds __attribute__ ((unused)),
	void *readfds, void *writefds, void *errorfds, void *timeout)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function select (%d, 0x%lx, 0x%lx, 0x%lx, 0x%lx) called",
	      nfds, (unsigned long) readfds, (unsigned long) writefds,
	      (unsigned long) errorfds, (unsigned long) timeout);
  return (-1);
}

int
scanf (const char *format __attribute__ ((unused)), ...)
{
  return (-1);
}

ssize_t
send (int s, const void *msg, size_t len, int flags)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function send (%d, 0x%lx, %lu, 0x%x) called", s,
	      (unsigned long) msg, (unsigned long) len, flags);
  return (-1);
}

int
snprintf (char *s, size_t n, const char *format, ...)
{
  int count;
  va_list ap;

  va_start (ap, format);
  count = grub_vsnprintf (s, n, format, ap);
  va_end (ap);

  return (count);
}

int
socket (int domain, int type, int protocol)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function socket (%d, %d, %d) called",
	      domain, type, protocol);
  return (-1);
}

int
sprintf (char *s, const char *format, ...)
{
  int count;
  va_list ap;

  va_start (ap, format);
  count = grub_vsnprintf (s, 0x1000, format, ap);
  va_end (ap);

  return (count);
}

int
sscanf (
  char const *s,
  char const *format,
  ...)
{
    int matches;

    va_list args;

    va_start (args, format);
    matches = vsscanf (s, format, args);
    va_end (args);

    return (matches);
    }

char *
stpcpy (char *s1, const char *s2)
{
  grub_strcpy (s1, s2);
  return (s1 + strlen (s1));
}

int
strcasecmp (const char *s1, const char *s2)
{
  return (grub_strcasecmp (s1, s2));
}

char *
strcat (char *s1, const char *s2)
{
  return (grub_strcat (s1, s2));
}

char *
strchr (const char *s1, int c)
{
  return (grub_strchr (s1, c));
}

int
strcmp (const char *s1, const char *s2)
{
  return (grub_strcmp (s1, s2));
}

char *
strcpy (char *s1, const char *s2)
{
  return (grub_strcpy (s1, s2));
}

size_t
strcspn (const char *s1, const char *s2)
{
  const char *s1ptr;
  const char *s2ptr;
  char ch;

  for (s1ptr = s1; *s1ptr != 0; s1ptr++)
    {
      ch = *s1ptr;
      for (s2ptr = s2; *s2ptr != 0; s2ptr++)
	if (ch == *s2ptr)
	  break;

      if (*s2ptr != 0)
	break;
    }

  return (s1ptr - s1);
}

char *
strdup (const char *s)
{
  return (grub_strdup (s));
}

char *
strerror (int errnum)
{
  static char buffer[32];

  grub_snprintf (buffer, sizeof (buffer), "Error %d", errnum);
  return (buffer);
}

size_t
strftime (char *s, size_t maxsize, const char *format, struct tm *tmptr)
{
  if (format == NULL)
    format = "%c";
  return (strftime_process (s, maxsize, 0, format, tmptr));
}

static const char *strftime_full_month_table[] = {
  "January",
  "February",
  "March",
  "April",
  "May",
  "June",
  "July",
  "August",
  "September",
  "October",
  "November",
  "December",
};

static const char *strftime_abbrev_month_table[] = {
  "Jan",
  "Feb",
  "Mar",
  "Apr",
  "May",
  "Jun",
  "Jul",
  "Aug",
  "Sep",
  "Oct",
  "Nov",
  "Dec",
};

static const char *strftime_full_weekday_table[] = {
  "Sunday",
  "Monday",
  "Tuesday",
  "Wednesday",
  "Thursday",
  "Friday",
  "Saturday",
};

static const char *strftime_abbrev_weekday_table[] = {
  "Sun",
  "Mon",
  "Tue",
  "Wed",
  "Thu",
  "Fri",
  "Sat",
};

static const char *strftime_ampm_table[] = {
  "am",
  "pm",
};


static int
strftime_process (char *s, size_t maxsize, char defflag,
		  const char *format, struct tm *tmptr)
{
  int count = 0;
  int flag;
  int width;
  int size;
  int ch;

  while (maxsize > 1 && (ch = *format++) != 0)
    {
      if (ch != '%')
	{
	  *s++ = (char) ch;
	  maxsize--;
	  count++;
	  continue;
	}

      ch = *format++;

      flag = defflag;
      if (ch == '#' || ch == '-' || ch == '0' || ch == '^' || ch == '_')
	{
	  flag = ch;
	  ch = *format++;
	}

      width = 0;
      while (grub_isdigit (ch))
	{
	  width = width * 10 + ch - '0';
	  ch = *format++;
	}

      if (ch == 'E' || ch == 'O')
	ch = *format++;

      switch (ch)
	{
	case '+':		/* date and time as produced by date */
	  size = strftime_process (s, maxsize, flag,
				   "%a %b %_d %0H:%0M:%0S %Z %0Y", tmptr);
	  break;

	case '%':		/* percent sign */
	  size = strftime_string (s, maxsize, width, "%");
	  break;

	case 'a':		/* abbreviated weekday name */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_abbrev_weekday_table,
				 ARRAY_SIZE (strftime_abbrev_weekday_table),
				 width, tmptr->tm_wday);
	  break;

	case 'A':		/* full weekday name */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_full_weekday_table,
				 ARRAY_SIZE (strftime_full_weekday_table),
				 width, tmptr->tm_wday);
	  break;

	case 'b':		/* abbreviated month name */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_abbrev_month_table,
				 ARRAY_SIZE (strftime_abbrev_month_table),
				 width, tmptr->tm_mon);
	  break;

	case 'B':		/* full month name */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_full_month_table,
				 ARRAY_SIZE (strftime_full_month_table),
				 width, tmptr->tm_mon);
	  break;

	case 'c':		/* date and time */
	  size = strftime_process (s, maxsize, flag,
				   "%a %b %0d %0H:%0M:%0S %0Y", tmptr);
	  break;

	case 'C':		/* date and time */
	  size = strftime_process (s, maxsize, flag,
				   "%a %b %_d %0H:%0M:%0S %Z %0Y", tmptr);
	  break;

	case 'd':		/* day of month */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_mday);
	  break;

	case 'D':		/* date as %m/%d/%y */
	  size = strftime_process (s, maxsize, flag, "%m/%0d/%0y", tmptr);
	  break;

	case 'e':		/* Day of month */
	  size = strftime_value (s, maxsize, flag ? flag : '_',
				 width ? width : 2, tmptr->tm_mday);
	  break;

	case 'F':		/* ISO 8601:2000 standard date */
	  size = strftime_process (s, maxsize, flag, "%0Y-%0m-%0d", tmptr);
	  break;

	case 'h':		/* abbreviated month name */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_abbrev_month_table,
				 ARRAY_SIZE (strftime_abbrev_month_table),
				 width, tmptr->tm_mon);
	  break;

	case 'H':		/* hour (24-hour clock) */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_hour);
	  break;

	case 'I':		/* hour (12-hour clock) */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2,
				 (tmptr->tm_hour + 11) % 12 + 1);
	  break;

	case 'j':		/* day number */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 3, tmptr->tm_yday + 1);
	  break;

	case 'k':		/* hour (24-hour clock) */
	  size = strftime_value (s, maxsize, flag ? flag : '_',
				 width ? width : 2, tmptr->tm_hour);
	  break;

	case 'l':		/* hour (12-hour clock) */
	  size = strftime_value (s, maxsize, flag ? flag : '_',
				 width ? width : 2,
				 (tmptr->tm_hour + 11) % 12 + 1);
	  break;

	case 'm':		/* month number */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_mon + 1);
	  break;

	case 'M':		/* minute */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_min);
	  break;

	case 'n':		/* newline */
	  size = strftime_string (s, maxsize, width, "\n");
	  break;

	case 'p':		/* am/pm */
	  size = strftime_entry (s, maxsize, flag == '#' ? 0 : '^',
				 strftime_ampm_table,
				 ARRAY_SIZE (strftime_ampm_table),
				 width, tmptr->tm_hour >= 12);
	  break;

	case 'P':		/* am/pm lower case */
	  size = strftime_entry (s, maxsize, flag,
				 strftime_ampm_table,
				 ARRAY_SIZE (strftime_ampm_table),
				 width, tmptr->tm_hour >= 12);
	  break;

	case 'r':		/* time in 12-hour clock */
	  size = strftime_process (s, maxsize, flag, "%I:%0M:%0S %p", tmptr);
	  break;

	case 'R':		/* hour:minute */
	  size = strftime_process (s, maxsize, flag, "%H:%0M", tmptr);
	  break;

	case 'S':		/* seconds */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_sec);
	  break;

	case 't':		/* tab */
	  size = strftime_string (s, maxsize, width, "\t");
	  break;

	case 'T':		/* time */
	  size = strftime_process (s, maxsize, flag, "%H:%0M:%0S", tmptr);
	  break;

	case 'u':		/* weekday */
	  size = strftime_value (s, maxsize, flag ? flag : '_',
				 width ? width : 1,
				 (tmptr->tm_wday + 6) % 7 + 1);
	  break;

	case 'w':		/* weekday */
	  size = strftime_value (s, maxsize, flag ? flag : '_',
				 width ? width : 1, tmptr->tm_wday);
	  break;

	case 'x':		/* date */
	  size = strftime_process (s, maxsize, flag, "%m/%0d/%0y", tmptr);
	  break;

	case 'X':		/* time */
	  size = strftime_process (s, maxsize, flag, "%H:%0M:%0S", tmptr);
	  break;

	case 'y':		/* year */
	  size = strftime_value (s, maxsize, flag == '-' || flag == '0'
				 || flag == '_' ? flag : '0',
				 width ? width : 2, tmptr->tm_year % 100);
	  break;

	case 'Y':		/* year */
	  size = strftime_value (s, maxsize, flag ? flag : '0',
				 width ? width : 4, tmptr->tm_year + 1900);
	  break;

	case 'z':		/* timezone offset from UTC */
	  size = 0;
	  break;

	case 'Z':		/* timezone name */
	  size = 0;
	  break;

	default:
	  size = strftime_string (s, maxsize, width, "?");
	  break;
	}

      s += size;
      maxsize -= size;
      count += size;
    }

  *s = 0;
  return (count);
}


static int
strftime_value (char *s, size_t maxsize, char flag, int width, int value)
{
  switch (flag)
    {
    case '-':			/* Do not pad anything for numeric values */
      return (grub_snprintf (s, maxsize, "%d", value));

    case '0':			/* Pad left with zeros */
      return (grub_snprintf (s, maxsize, "%0*d", width, value));

    case '_':			/* Pad left with spaces */
    default:
      return (grub_snprintf (s, maxsize, "%*d", width, value));
    }
}


static int
strftime_entry (char *s, size_t maxsize, char flag, const char *const *table,
		size_t tabsize, int width, int value)
{
  const char *entry = NULL;
  int length;
  int count = 0;
  char ch;

  if (value >= 0 && (size_t) value < tabsize)
    entry = table[value];
  if (entry == NULL)
    entry = "?";

  length = (int) grub_strlen (entry);
  for (; maxsize > 1 && length < width; maxsize--, s++, length++, count++)
    *s = ' ';

  switch (flag)
    {
    case '#':			/* Convert to other case */
    case '^':			/* Convert to upper case */
      for (; maxsize > 1 && (ch = *entry) != 0;
	   maxsize--, s++, entry++, count++)
	{
	  if (grub_islower (ch))
	    ch = grub_toupper (ch);
	  *s = ch;
	}
      break;

    default:
      for (; maxsize > 1 && (ch = *entry) != 0;
	   maxsize--, s++, entry++, count++)
	{
	  *s = ch;
	}
      break;
    }

  return (count);
}


static int
strftime_string (char *s, size_t maxsize, int width, const char *entry)
{
  char ch;
  int length;
  int count = 0;

  length = (int) grub_strlen (entry);
  for (; maxsize > 1 && length < width; maxsize--, s++, length++, count++)
    *s = ' ';

  for (; maxsize > 1 && (ch = *entry) != 0; maxsize--, s++, entry++, count++)
    {
      *s = ch;
    }

  return (count);
}

size_t
strlen (const char *s1)
{
  return (grub_strlen (s1));
}
int
strncasecmp (const char *s1, const char *s2, size_t n)
{
  return (grub_strncasecmp (s1, s2, n));
}

int
strncmp (const char *s1, const char *s2, size_t n)
{
  return (grub_strncmp (s1, s2, n));
}

char *
strncat (char *s1, const char *s2, size_t n)
{
  return (grub_strncat (s1, s2, n));
}

char *
strncpy (char *s1, const char *s2, size_t n)
{
  return (grub_strncpy (s1, s2, n));
}

char *
strndup (const char *s, size_t n)
{
  return (grub_strndup (s, n));
}

char *
strpbrk (const char *s1, const char *s2)
{
  const char *s1ptr;
  const char *s2ptr;
  char ch;

  for (s1ptr = s1; *s1ptr != 0; s1ptr++)
    {
      ch = *s1ptr;
      for (s2ptr = s2; *s2ptr != 0; s2ptr++)
	if (ch == *s2ptr)
	  return ((char *) s1ptr);
    }
  return (NULL);
}

char *
strrchr (const char *s1, int c)
{
  return (grub_strrchr (s1, c));
}

size_t
strspn (const char *s1, const char *s2)
{
  const char *s1ptr;
  const char *s2ptr;
  char ch;

  for (s1ptr = s1; *s1ptr != 0; s1ptr++)
    {
      ch = *s1ptr;
      for (s2ptr = s2; *s2ptr != 0; s2ptr++)
	if (ch == *s2ptr)
	  break;

      if (*s2ptr == 0)
	break;
    }

  return (s1ptr - s1);
}

char *
strstr (const char *s1, const char *s2)
{
  return (grub_strstr (s1, s2));
}

long
strtol (const char *str, char **endptr, int base)
{
  return (grub_strtol (str, endptr, base));
}

unsigned long
strtoul (const char *str, char **endptr, int base)
{
  return (grub_strtoul (str, endptr, base));
}

unsigned long long
strtoull (const char *str, char **endptr, int base)
{
  return (grub_strtoull (str, endptr, base));
}

time_t
time (time_t * tloc)
{
  struct grub_datetime datetime;
  int unixtime = 0;

  grub_get_datetime (&datetime);
  grub_datetime2unixtime (&datetime, &unixtime);

  if (tloc != NULL)
    *tloc = unixtime;
  return (unixtime);
}

int
vasprintf (char **ret, const char *format, va_list ap)
{
  char *s;

  s = grub_xvasprintf (format, ap);
  if (ret != NULL)
    *ret = s;
  if (s == NULL)
    return (0);
  return (grub_strlen (s));
}

int
vfprintf (FILE * stream, const char *format, va_list ap)
{
  if (stream == stdout || stream == stderr)
    return (grub_vprintf (format, ap));
  return (-1);
}

int
vprintf (const char *format, va_list ap)
{
  return (grub_vprintf (format, ap));
}

int
vsnprintf (char *s, size_t n, const char *format, va_list ap)
{
  return (grub_vsnprintf (s, n, format, ap));
}

int
vsprintf (char *s, const char *format, va_list ap)
{
  return (grub_vsnprintf (s, 0x1000, format, ap));
}

int
vsscanf (
  char const *s,
  char const *format,
  va_list arg)
{
	char const *sourceptr = s;
	char const *formatptr = format;
	int suppress;
	int length;
	size_t size;
	char matchvalue;
	int minus;
	int base;
	long long value;
	int ch;
	int valid;
	char *argptr;
	int matchcount = 0;
	char match[256];

	while (*formatptr != 0) {
	    
	    if (grub_isspace (*formatptr)) {
		while (grub_isspace (*sourceptr))
		    sourceptr++;
		formatptr++;
		continue;
		}

	    if (*formatptr != '%') {
		if (*sourceptr == 0 && matchcount == 0)
		    return (-1);
		if (*sourceptr != *formatptr)
		    return (matchcount);
		sourceptr++;
		formatptr++;
		continue;
		}

	    formatptr++;
	    if (*formatptr == '%') {
		while (grub_isspace (*sourceptr))
		    sourceptr++;
		if (*sourceptr == 0 && matchcount == 0)
		    return (-1);
		if (*sourceptr != '%')
		    return (matchcount);
		sourceptr++;
		formatptr++;
		continue;
		}

	    suppress = 0;
	    if (*formatptr == '*') {
		suppress = 1;
		formatptr++;
		}

	    length = 0;
	    while (*formatptr >= '0' && *formatptr <= '9') {
		length = length * 10 + *formatptr - '0';
		formatptr++;
		}

	    if (length == 0)
		length = GRUB_USHRT_MAX;

	    size = 0;
	    switch (*formatptr) {
		case 'h':
		    if (formatptr[1] == 'h') {
			size = sizeof (char);
		        formatptr += 2;
			}
		    else {
			size = sizeof (short);
			formatptr++;
			}
		    break;

		case 'l':
		    if (formatptr[1] == 'l') {
			size = sizeof (long long);
		        formatptr += 2;
			}
		    else {
			size = sizeof (long);
			formatptr++;
			}
		    break;

		case 'j':
		    size = sizeof (long long);
		    formatptr++;
		    break;

		case 'z': case 't':
		    size = sizeof (size_t);
		    formatptr++;
		    break;
		}
		
	    switch (*formatptr++) {
		case '[':
		    if (*formatptr == '^') {
			match[0] = 0;
			memset (&match[1], 1, sizeof (match) - 1);
			matchvalue = 0;
			formatptr++;
			}
		    else {
			memset (match, 0, sizeof (match));
			matchvalue = 1;
			}

		    if (*formatptr == ']' || *formatptr == '-') {
			match[(unsigned char) *formatptr] = matchvalue;
			formatptr++;
			}

		    while (*formatptr != ']' && *formatptr != 0) {
			if (*formatptr == '-' && formatptr[1] != ']' &&
			  (unsigned char) formatptr[-1] < (unsigned char) formatptr[1]) {
			    memset (&match[(unsigned char) formatptr[-1]], matchvalue,
			      (unsigned char) formatptr[1] - (unsigned char) formatptr[-1] + 1);
			    formatptr += 2;
			    }
			else {
			    match[(unsigned char) *formatptr] = matchvalue;
			    formatptr++;
			    }
			}

		    if (*formatptr != ']')
			return (-1);
		    formatptr++;

		    if (!suppress) {
			argptr = va_arg (arg, char *);

			while (match[(unsigned char) *sourceptr]) {
			    *argptr++ = *sourceptr;
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }

			*argptr = 0;
			matchcount++;
			}

		    else {
			while (match[(unsigned char) *sourceptr]) {
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }
			}
		    break;

		case 'c': case 'C':
		    if (length == GRUB_USHRT_MAX)
			length = 1;

		    if (!suppress) {
			argptr = va_arg (arg, char *);

			while (*sourceptr != 0) {
			    *argptr++ = *sourceptr;
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }

			matchcount++;
			}

		    else {
			while (*sourceptr != 0) {
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }
			}
		    break;

		case 's': case 'S':
		    while (grub_isspace (*sourceptr)) 
			sourceptr++;

		    if (!suppress) {
			argptr = va_arg (arg, char *);

			while (*sourceptr != 0 && !grub_isspace (*sourceptr)) {
			    *argptr++ = *sourceptr;
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }

			*argptr = 0;
			matchcount++;
			}

		    else {
			while (*sourceptr != 0 && !grub_isspace (*sourceptr)) {
			    sourceptr++;
			    length--;
			    if (length <= 0)
				break;
			    }
			}
		    break;

		case 'i':
		    base = 0;
		    goto donumber;

		case 'd': case 'u':
		    base = 10;
		    goto donumber;

		case 'o':
		    base = 8;
		    goto donumber;

		case 'x': case 'X':
		    base = 16;
		    goto donumber;

		case 'p':
		    if (size == 0)
			size = sizeof (void *);
		    base = 16;
		    goto donumber;

		donumber:
		    if (size == 0)
			size = sizeof (int);

		    while (grub_isspace (*sourceptr)) 
			sourceptr++;

		    valid = 0;
		    value = 0;
		    minus = 0;
		    if (length >= 1 && *sourceptr == '-') {
			minus = 1;
			sourceptr++;
			length--;
			}

		    else if (length >= 1 && *sourceptr == '+') {
			sourceptr++;
			length--;
			}

		    if ((base == 0 || base == 16) && length >= 2 &&
		      sourceptr[0] == '0' && (sourceptr[1] == 'x' || sourceptr[1] == 'X')) {
			base = 16;
			sourceptr += 2;
			length -= 2;
			}

		    else if (base == 0 && length >= 2 && 
		      sourceptr[0] == '0' && sourceptr[1] >= '0' && sourceptr[1] <= '7') {
			base = 8;
			sourceptr++;
			length--;
			}

		    else if (base == 0)
		       base = 10;

		    while (length > 0 && *sourceptr != 0) {
			ch = *sourceptr;
			if (ch >= '0' && ch <= '9')
			    ch = ch - '0';
			else if (ch >= 'a' && ch <= 'f')
			    ch = ch - 'a' + 10;
			else if (ch >= 'A' && ch <= 'F')
			    ch = ch - 'A' + 10;
			else
			    break;

			if (ch >= base)
			    break;

			value = value * base + ch;
			valid = 1;
			sourceptr++;
			length--;
			}

		    if (minus)
			value = -value;

		    if (!valid) {
			if (*sourceptr == 0 && matchcount == 0)
			    return (-1);
			return (matchcount);
			}

		    if (!suppress) {
			argptr = va_arg (arg, char *);

			if (size == sizeof (char))
			    *argptr = (char) value;

			else if (size == sizeof (short))
			    *(short *) argptr = (short) value;

			else if (size == sizeof (int))
			    *(int *) argptr = (int) value;

			else if (size == sizeof (long))
			    *(long *) argptr = (long) value;

			else if (size == sizeof (long long))
			    *(long long *) argptr = value;
			matchcount++;
			}
		    break;

		case 'n':
		    if (size == 0)
			size = sizeof (int);

		    value = sourceptr - s;

		    if (!suppress) {
			argptr = va_arg (arg, char *);

			if (size == sizeof (char))
			    *argptr = (char) value;

			else if (size == sizeof (short))
			    *(short *) argptr = (short) value;

			else if (size == sizeof (int))
			    *(int *) argptr = (int) value;

			else if (size == sizeof (long))
			    *(long *) argptr = (long) value;

			else if (size == sizeof (long long))
			    *(long long *) argptr = value;
			}
		    break;

		default:
		    return (matchcount);
	    	}
	    }

	return (matchcount);
	}



unsigned long long
__umoddi3 (unsigned long long n, unsigned long long d)
{
  grub_uint64_t r;

  grub_divmod64 (n, d, &r);
  return (r);
}

ssize_t
write (int f, const void *buffer, size_t size)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function write (%d, 0x%lx, %lu) called", f,
	      (unsigned long) buffer, (unsigned long) size);
  return (-1);
}

ssize_t
writev (int f, const giovec_t * iovec, int iovec_cnt)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function writev (%d, 0x%lx, %d) called", f,
	      (unsigned long) iovec, (unsigned long) iovec_cnt);
  return (-1);
}

ssize_t
__xnet_sendmsg (int s, const void *msg, int flags)
{
  grub_error (GRUB_ERR_BUG,
	      "unsupported function __xnet_sendmsg (%d, 0x%lx, 0x%x) called", s,
	      (unsigned long) msg, flags);
  return (-1);
}



unsigned char __ctype[SZ_TOTAL] = {
  0,				/* EOF */
  _C,      _C,      _C,      _C,      _C,      _C,      _C,      _C,
  _C,      _S | _C, _S | _C, _S | _C, _S | _C, _S | _C, _C,      _C,
  _C,      _C,      _C,      _C,      _C,      _C,      _C,      _C,
  _C,      _C,      _C,      _C,      _C,      _C,      _C,      _C,
  _S | _B, _P,      _P,      _P,      _P,      _P,      _P,      _P,
  _P,      _P,      _P,      _P,      _P,      _P,      _P,      _P,
  _N | _X, _N | _X, _N | _X, _N | _X, _N | _X, _N | _X, _N | _X, _N | _X,
  _N | _X, _N | _X, _P,      _P,      _P,      _P,      _P,      _P,
  _P,      _U | _X, _U | _X, _U | _X, _U | _X, _U | _X, _U | _X, _U,
  _U,      _U,      _U,      _U,      _U,      _U,      _U,      _U,
  _U,      _U,      _U,      _U,      _U,      _U,      _U,      _U,
  _U,      _U,      _U,      _P,      _P,      _P,      _P,      _P,
  _P,      _L | _X, _L | _X, _L | _X, _L | _X, _L | _X, _L | _X, _L,
  _L,      _L,      _L,      _L,      _L,      _L,      _L,      _L,
  _L,      _L,      _L,      _L,      _L,      _L,      _L,      _L,
  _L,      _L,      _L,      _P,      _P,      _P,      _P,      _C,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,
  0,       0,       0,       0,       0,       0,       0,       0,

/* tolower()  and toupper() conversion table */ 0,
    0,   1,   2,   3,   4,   5,   6,   7,
    8,   9,  10,  11,  12,  13,  14,  15,
   16,  17,  18,  19,  20,  21,  22,  23,
   24,  25,  26,  27,  28,  29,  30,  31,
   32,  33,  34,  35,  36,  37,  38,  39,
   40,  41,  42,  43,  44,  45,  46,  47,
   48,  49,  50,  51,  52,  53,  54,  55,
   56,  57,  58,  59,  60,  61,  62,  63,
   64,  97,  98,  99, 100, 101, 102, 103,
  104, 105, 106, 107, 108, 109, 110, 111,
  112, 113, 114, 115, 116, 117, 118, 119,
  120, 121, 122,  91,  92,  93,  94,  95,
   96,  65,  66,  67,  68,  69,  70,  71,
   72,  73,  74,  75,  76,  77,  78,  79,
   80,  81,  82,  83,  84,  85,  86,  87,
   88,  89,  90, 123, 124, 125, 126, 127,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
/* CSWIDTH information */
    1,   0,   0,   1,   0,   0,   1,
};

FILE __iob[3];

char **environ;
