/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (c) 2010, 2012, Oracle and/or its affiliates. All rights reserved.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <grub/mm.h>
#include <grub/types.h>
#include <grub/misc.h>
#include <grub/util/misc.h>
#include <grub/util/version.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

static int compare_version (const char *, const char *);

struct grub_util_version_info grub_util_solaris_grub_version =
{
  "solaris_grub_version",
  compare_version,
  NULL,
  0
};

/*
 * compare_dotted_version()
 * Compares two strings with an arbitrary long number of dot-separated numbers.
 * Returns:	0  - if the version numbers are equal
 *             >0  - if str1 version number is more recent than str2
 *             <0  - if str2 version number is more recent than str1
 *
 * Comparison is done field by field, by retrieving an unsigned integer value,
 * (missing fields are assumed as 0, but explict zeroes take precedence) so:
 *   4.1.2.11 > 4.1.2.2 > 4.1.2.0 > 4.1.2
 *
 * where ">" means "more recent than".
 */
static int
compare_dotted_version(const char *str1, const char *str2)
{
  int  retval = 0;
  char *verstr1, *verstr2, *freeptr1, *freeptr2;
  char *parsep1, *parsep2;
  int val_str1, val_str2;

  freeptr1 = verstr1 = xstrdup (str1);
  freeptr2 = verstr2 = xstrdup (str2);

  while (verstr1 != NULL && verstr2 != NULL)
    {
      parsep1 = strsep (&verstr1, ".");
      parsep2 = strsep (&verstr2, ".");

      val_str1 = atoi (parsep1);
      val_str2 = atoi (parsep2);

      retval = val_str1 - val_str2;
      if (retval != 0)
        goto out; 
    }

  /* Common portion of the version string is equal. */
  if (verstr1 == NULL && verstr2 != NULL)
    retval = -1;
  if (verstr2 == NULL && verstr1 != NULL)
    retval = 1;

out:
  free (freeptr1);
  free (freeptr2);
  grub_util_info("comparing dotted version %s %s returning %d",
                 str1, str2, retval);
  return (retval);
}

/*
 * compare_timestamps()
 * Currently, timestamp is in %Y%m%dT%H%M%SZ format in UTC, which means that
 * we can simply do a lexicographic comparison to know which one is the most
 * recent.
 *
 * Returns:   0  - if timestamps coincide
 *            1  - if the timestamp in str1 is more recent
 *            2  - if the timestamp in str2 is more recent
 */
static int
compare_timestamps (const char *str1, const char *str2)
{
  int  retval;

  retval = strcmp (str1, str2);
  grub_util_info("comparing timestamps %s %s returning %d", str1, str2, retval);
  return (retval);
}

/*
 * compare_version()
 * Given two solaris versions, compare the two and returns which one is more
 * "recent". Comparison is based on dotted version number fields and a
 * timestamp.
 *
 * Returns:     0   - if the two versions coincide
 *             >0   - if the version in str1 is more recent
 *             <0   - if the version in str2 is more recent
 */
static int
compare_version (const char *str1, const char *str2)
{
  int retval = 0;
  char *verstr1, *verstr2, *freeptr1, *freeptr2;
  char *parsep1, *parsep2;

  freeptr1 = verstr1 = xstrdup(str1);
  freeptr2 = verstr2 = xstrdup(str2);

  parsep1 = verstr1;
  parsep2 = verstr2;

  while (parsep1 != NULL && parsep2 != NULL)
    {
      parsep1 = strsep (&verstr1, ",:-");
      parsep2 = strsep (&verstr2, ",:-");

     /* verstr1 or verstr2 will be NULL before parsep1 or parsep2. */
     if (verstr1 == NULL || verstr2 == NULL)
       {
         retval = compare_timestamps (parsep1, parsep2);
         goto out;
       }

     retval = compare_dotted_version (parsep1, parsep2);
     if (retval == 0)
       continue;
     else
       goto out;
    }

out:
  free (freeptr1);
  free (freeptr2);
  return (retval);
}
