/*
 * Description
 *	HBAAPILIB-sun.c - Implements the Sun Extension for Target mode
 *		FCHBA discovery
 *
 * License:
 *	The contents of this file are subject to the SNIA Public License
 *	Version 1.0 (the "License"); you may not use this file except in
 *	compliance with the License. You may obtain a copy of the License at
 *
 *	http://www.snia.org/English/Resources/Code/OpenSource.html
 *
 *	Software distributed under the License is distributed on an "AS IS"
 *	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 *	the License for the specific language governing rights and limitations
 *	under the License.
 */
/*
 * Copyright (c) 2008, 2020, Oracle and/or its affiliates.
 */

#include <dlfcn.h>
#include <strings.h>
#include <stdio.h>
#include <time.h>
#include "hbaapi.h"
#include "hbaapi-sun.h"
#include "vendorhbaapi.h"
#include <stdlib.h>
#include <assert.h>
#include <syslog.h>
#include <sys/varargs.h>

/*
 * LIBRARY_NUM is a shortcut to figure out which library we need to call.
 *  The top 16 bits of handle are the library index
 */
#define	LIBRARY_NUM(handle)	((handle)>>16)

/*
 * VENDOR_HANDLE turns a global library handle into a vendor specific handle,
 * with all upper 16 bits set to 0
 */
#define	VENDOR_HANDLE(handle)	((handle)&0xFFFF)

#define	HBA_HANDLE_FROM_LOCAL(library, vendor) \
	(((library)<<16) | ((vendor)&0x0000FFFF))

extern int _hbaapi_debuglevel;
extern int _hbaapi_sysloginit;

#include <pthread.h>
/*
 * When multiple mutex's are grabed, they must be always be grabbed in
 * the same order, or deadlock can result.  There are three levels
 * of mutex's involved in this API.  If LL_mutex is grabbed, always grap
 * it first.  If AL_mutex is grabbed, it may not be grabbed before
 * LL_mutex.  If grabbed in a multi grab sequence, the mutex's protecting
 * the callback lists must always be grabbed last and release before calling
 * a vendor specific library function that might invoke a callback function
 * on the same thread.
 */
#define	GRAB_MUTEX(M)			grab_mutex(M)
#define	RELEASE_MUTEX(M)		release_mutex(M)
#define	RELEASE_MUTEX_RETURN(M, RET)	release_mutex(M); return (RET)

static pthread_mutex_t _hbaapi_debug_fd_mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * HBA_LIBRARY_STATUS and HBA_LIBRARY_INFO are redefined here.
 * Avoid any change in the common code.
 */
typedef enum {
    HBA_LIBRARY_UNKNOWN,
    HBA_LIBRARY_LOADED,
    HBA_LIBRARY_NOT_LOADED
} HBA_LIBRARY_STATUS;

typedef struct hba_library_info {
    struct hba_library_info *next;
    char		*LibraryName;
    void*		hLibrary;		/* Handle to a loaded DLL */
    char		*LibraryPath;
    HBA_ENTRYPOINTSV2	functionTable;		/* Function pointers */
    HBA_LIBRARY_STATUS	status;			/* info on this library */
    HBA_UINT32		index;
} HBA_LIBRARY_INFO, *PHBA_LIBRARY_INFO;

#define	ARE_WE_INITED() \
	if (_hbaapi_librarylist == NULL) { \
		return (HBA_STATUS_ERROR); \
	}

extern HBA_LIBRARY_INFO *_hbaapi_librarylist;
extern HBA_UINT32 _hbaapi_total_library_count;
int _hbaapi_loaded_refcnt = 0;
extern pthread_mutex_t _hbaapi_LL_mutex;
extern pthread_mutex_t _hbaapi_AL_mutex;
extern pthread_mutex_t _hbaapi_APE_mutex;
pthread_mutex_t	_hbaapi_LL_loaded_mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * Function type def fop Sun extentions.
 */
typedef HBA_UINT32	(* Sun_HBAGetNumberOfTgtAdaptersFunc)();
typedef HBA_STATUS	(* Sun_HBAGetTgtAdapterNameFunc)(HBA_UINT32, char *);
typedef HBA_HANDLE	(* Sun_HBAOpenTgtAdapterFunc)(char *);
typedef HBA_STATUS	(* Sun_HBAOpenTgtAdapterByWWNFunc)
			    (HBA_HANDLE *, HBA_WWN);
typedef	HBA_STATUS	(* Sun_HBANPIVGetAdapterAttributesFunc)
			    (HBA_HANDLE, HBA_ADAPTERATTRIBUTES *);
typedef	HBA_STATUS	(* Sun_HBAGetNPIVPortInfoFunc)
			    (HBA_HANDLE, HBA_UINT32, HBA_UINT32,
			    HBA_NPIVATTRIBUTES *);
typedef HBA_STATUS	(* Sun_HBADeleteNPIVPortFunc)
			    (HBA_HANDLE, HBA_UINT32, HBA_WWN);
typedef HBA_STATUS	(* Sun_HBACreateNPIVPortFunc)
			    (HBA_HANDLE, HBA_UINT32, HBA_WWN, HBA_WWN,
			    HBA_INT8 *, HBA_UINT32 *);
typedef	HBA_STATUS	(* Sun_HBAAdapterReturnWWNFunc)
			    (HBA_HANDLE, HBA_UINT32, HBA_WWN *, HBA_WWN *);
typedef	HBA_STATUS	(* Sun_HBAAdapterCreateWWNFunc)
			    (HBA_WWN *, HBA_WWN *,
			    HBA_WWN *, HBA_INT32);
typedef	HBA_STATUS	(* Sun_HBAGetPortNPIVAttributesFunc)
			    (HBA_HANDLE, HBA_UINT32, HBA_PORTNPIVATTRIBUTES *);
typedef	HBA_STATUS	(* Sun_HBARegisterForAdapterDeviceEventsFunc)
			    (void (*)(void *, HBA_WWN, HBA_UINT32, HBA_UINT32),
			    void *, HBA_HANDLE, HBA_WWN, HBA_CALLBACKHANDLE *);
typedef	HBA_STATUS	(* Sun_HBADoForceLipFunc)(HBA_HANDLE, int *);

/*
 * Individual adapter (hba) information
 * Same as hbaadapter with different structure name.
 */
typedef struct hba_tgtadapter_info {
    struct hba_tgtadapter_info *next;
    HBA_STATUS		GNstatus; /* status from GetTgtAdapterNameFunc */
    char		*name;
    HBA_WWN		nodeWWN;
    HBA_LIBRARY_INFO	*library;
    HBA_UINT32		index;
} HBA_TGTADAPTER_INFO;

/*
 * Make the list as an array with max size 16
 */
HBA_TGTADAPTER_INFO *_hbaapi_tgtadapterlist;
HBA_UINT32 _hbaapi_total_tgtadapter_count = 0;
pthread_mutex_t _hbaapi_tgtAL_mutex = PTHREAD_MUTEX_INITIALIZER;

char *
WWN2str(char *buf, HBA_WWN *wwn) {
	int j;
	unsigned char *pc = (unsigned char *)&(wwn->wwn[0]);
	buf[0] = '\0';
	for (j = 0; j < 16; j += 2) {
		(void) sprintf(&buf[j], "%02X", (int)*pc++);
	}
	return (buf);
}

void
trace_debug(int level, const char *fmt, ...)
{
	va_list ap;

	if (level > _hbaapi_debuglevel)
		return;

	(void) pthread_mutex_lock(&_hbaapi_debug_fd_mutex);
	if (_hbaapi_sysloginit == 0) {
		openlog("HBAAPI", LOG_PID|LOG_ODELAY, LOG_USER);
		_hbaapi_sysloginit = 1;
	}
	va_start(ap, fmt);
	vsyslog(LOG_USER|LOG_DEBUG, fmt, ap);
	va_end(ap);
	(void) pthread_mutex_unlock(&_hbaapi_debug_fd_mutex);
}

/*
 * Common library internal. Mutex handling
 */
static void
grab_mutex(pthread_mutex_t *mp) {
    int ret;
    if ((ret = pthread_mutex_lock(mp)) != 0) {
		perror("pthread_mutex_lock - HBAAPI:");
		trace_debug(0, "pthread_mutex_lock returned %d", ret);
	}
}

static void
release_mutex(pthread_mutex_t *mp) {
    int ret;
    if ((ret = pthread_mutex_unlock(mp)) != 0) {
		perror("pthread_mutex_unlock - HBAAPI:");
		trace_debug(0, "pthread_mutex_unlock returned %d", ret);
	}
}

/*
 * The API used to use fixed size tables as its primary data structure.
 * Indexing from 1 to N identified each adapters.  Now the adapters are
 * on a linked list.  There is a unique "index" foreach each adapter.
 * Adapters always keep their index, even if they are removed from the
 * hardware.  The only time the indexing is reset is on HBA_FreeLibrary
 */
HBA_UINT32
Sun_HBA_GetNumberOfTgtAdapters()
{
	int j = 0;
	HBA_LIBRARY_INFO	*lib_infop;
	Sun_HBAGetNumberOfTgtAdaptersFunc
	    GetNumberOfTgtAdaptersFunc = NULL;
	Sun_HBAGetTgtAdapterNameFunc
	    GetTgtAdapterNameFunc = NULL;
	HBA_BOOLEAN		found_name;
	HBA_TGTADAPTER_INFO	*adapt_infop;
	HBA_STATUS		status;

	char adaptername[HBA_ADAPTER_NAME_MAX_LEN];
	int num_adapters; /* local */

	if (_hbaapi_librarylist == NULL) {
		return (0);
	}
	GRAB_MUTEX(&_hbaapi_LL_mutex); /* pay attention to order */
	GRAB_MUTEX(&_hbaapi_tgtAL_mutex);

	for (lib_infop = _hbaapi_librarylist;
	    lib_infop != NULL;
	    lib_infop = lib_infop->next) {

		if (lib_infop->status != HBA_LIBRARY_LOADED) {
			continue;
		}

		if (lib_infop->hLibrary != NULL) {
			GetNumberOfTgtAdaptersFunc =
			    (Sun_HBAGetNumberOfTgtAdaptersFunc)
			    dlsym(lib_infop->hLibrary,
			    "Sun_fcGetNumberOfTgtAdapters");
			GetTgtAdapterNameFunc = (Sun_HBAGetTgtAdapterNameFunc)
			    dlsym(lib_infop->hLibrary,
			    "Sun_fcGetTgtAdapterName");
			if (GetNumberOfTgtAdaptersFunc == NULL ||
			    GetTgtAdapterNameFunc == NULL)	{
				GetNumberOfTgtAdaptersFunc = NULL;
				GetTgtAdapterNameFunc = NULL;
				continue;
			}
		} else {
			continue;
		}

		num_adapters = ((GetNumberOfTgtAdaptersFunc)());
		trace_debug(1, "HBAAPI: number of target mode "
		    "adapters for %s = %d\n",
		    lib_infop->LibraryName, num_adapters);

		for (j = 0; j < num_adapters; j++) {
			found_name = 0;
			status = (GetTgtAdapterNameFunc)(j,
			    (char *)&adaptername);
			if (status == HBA_STATUS_OK) {
				for (adapt_infop = _hbaapi_tgtadapterlist;
				    adapt_infop != NULL;
				    adapt_infop = adapt_infop->next) {
					/*
					 * check for duplicates, really, this
					 * may just be a second call to this
					 * function
					 * ??? how do we know when a name
					 * becomes stale?
					 */
					if (strcmp(adaptername,
					    adapt_infop->name) == 0) {
						/* already got this one */
						found_name++;
						break;
					}
				}
				if (found_name != 0) {
					continue;
				}
			}

			adapt_infop = (HBA_TGTADAPTER_INFO *)
			    calloc(1, sizeof (HBA_TGTADAPTER_INFO));
			if (adapt_infop == NULL) {
				(void) fprintf(stderr,
				    "HBA_GetNumberOfAdapters: calloc failed "
				    "on sizeof:%ld\n",
				    (long)sizeof (HBA_TGTADAPTER_INFO));
				RELEASE_MUTEX(&_hbaapi_tgtAL_mutex);
				RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex,
				    _hbaapi_total_tgtadapter_count);
			}
			if ((adapt_infop->GNstatus = status) == HBA_STATUS_OK) {
				adapt_infop->name = strdup(adaptername);
			} else {
				char dummyname[256];
				(void) snprintf(dummyname, sizeof (dummyname),
				    "NULLADAPTER-%s-%03d",
				    lib_infop->LibraryPath,
				    _hbaapi_total_tgtadapter_count);
				dummyname[255] = '\0';
				adapt_infop->name = strdup(dummyname);
			}
			adapt_infop->library = lib_infop;
			adapt_infop->next = _hbaapi_tgtadapterlist;
			adapt_infop->index = _hbaapi_total_tgtadapter_count;
			_hbaapi_tgtadapterlist = adapt_infop;
			_hbaapi_total_tgtadapter_count++;
		}
		GetNumberOfTgtAdaptersFunc = GetTgtAdapterNameFunc = NULL;
	}
	RELEASE_MUTEX(&_hbaapi_tgtAL_mutex);
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, _hbaapi_total_tgtadapter_count);
}

HBA_STATUS
Sun_HBA_GetTgtAdapterName(
    HBA_UINT32 adapterindex,
    char *adaptername)
{
	HBA_TGTADAPTER_INFO	*adapt_infop;
	HBA_STATUS		ret = HBA_STATUS_ERROR_ILLEGAL_INDEX;

	if (adaptername == NULL) {
		return (HBA_STATUS_ERROR_ARG);
	}
	/*
	 * The adapter index is from old code, but we have
	 * to support it.  Go down the list looking for
	 * the adapter
	 */
	ARE_WE_INITED();
	GRAB_MUTEX(&_hbaapi_tgtAL_mutex);
	*adaptername = '\0';
	for (adapt_infop = _hbaapi_tgtadapterlist;
	    adapt_infop != NULL;
	    adapt_infop = adapt_infop->next) {

		if (adapt_infop->index == adapterindex) {
			if (adapt_infop->name != NULL &&
			    adapt_infop->GNstatus == HBA_STATUS_OK) {
				(void) strlcpy(adaptername, adapt_infop->name,
				    HBA_ADAPTER_NAME_MAX_LEN);
			} else {
				*adaptername = '\0';
			}
			ret = adapt_infop->GNstatus;
			break;
		}
	}
	trace_debug(2, "GetAdapterName for index:%d ->%s",
	    adapterindex, adaptername);
	RELEASE_MUTEX_RETURN(&_hbaapi_tgtAL_mutex, ret);
}

HBA_HANDLE
Sun_HBA_OpenTgtAdapter(char *adaptername)
{
	HBA_HANDLE		handle;
	Sun_HBAOpenTgtAdapterFunc	OpenTgtAdapterFunc;
	HBA_TGTADAPTER_INFO	*adapt_infop;
	HBA_LIBRARY_INFO	*lib_infop;

	trace_debug(2, "OpenAdapter: %s", adaptername);

	if (_hbaapi_librarylist == NULL) {
		return (HBA_HANDLE_INVALID);
	}
	if (adaptername == NULL) {
		return (HBA_STATUS_ERROR_ARG);
	}
	handle = HBA_HANDLE_INVALID;
	GRAB_MUTEX(&_hbaapi_AL_mutex);
	for (adapt_infop = _hbaapi_tgtadapterlist;
	    adapt_infop != NULL;
	    adapt_infop = adapt_infop->next) {
		if (strcmp(adaptername, adapt_infop->name) != 0) {
			continue;
		}
		lib_infop = adapt_infop->library;
		OpenTgtAdapterFunc = (Sun_HBAOpenTgtAdapterFunc)
		    dlsym(lib_infop->hLibrary, "Sun_fcOpenTgtAdapter");
		if (OpenTgtAdapterFunc != NULL) {
			/* retrieve the vendor handle */
			handle = (OpenTgtAdapterFunc)(adaptername);
			if (handle != 0) {
				/*
				 * or this with the library index to get
				 * the common handle
				 */
				handle = HBA_HANDLE_FROM_LOCAL(
				    lib_infop->index, handle);
			}
		}
		break;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_AL_mutex, handle);
}

/*
 * This function ignores the list of known adapters and instead tries
 * each vendors open function to see if one of them
 * can open an adapter when referenced with a particular WWN
 */
HBA_STATUS
Sun_HBA_OpenTgtAdapterByWWN(HBA_HANDLE *phandle, HBA_WWN nodeWWN)
{
	HBA_HANDLE		handle;
	HBA_LIBRARY_INFO	*lib_infop;
	Sun_HBAGetNumberOfTgtAdaptersFunc
	    GetNumberOfTgtAdaptersFunc;
	Sun_HBAOpenTgtAdapterByWWNFunc
	    OpenTgtAdapterByWWNFunc;
	HBA_STATUS		status;
	char	wwn_str[17];

	trace_debug(2, "OpenAdapterByWWN: %s", WWN2str(wwn_str, &nodeWWN));

	if (phandle == NULL) {
		return (HBA_STATUS_ERROR_ARG);
	}

	ARE_WE_INITED();

	*phandle = HBA_HANDLE_INVALID;

	GRAB_MUTEX(&_hbaapi_LL_mutex);
	for (lib_infop = _hbaapi_librarylist;
	    lib_infop != NULL;
	    lib_infop = lib_infop->next) {

		status = HBA_STATUS_ERROR_ILLEGAL_WWN;

		if (lib_infop->status != HBA_LIBRARY_LOADED) {
			continue;
		}

		GetNumberOfTgtAdaptersFunc = (Sun_HBAGetNumberOfTgtAdaptersFunc)
		    dlsym(lib_infop->hLibrary,
		    "Sun_fcGetNumberOfTgtAdapters");
		OpenTgtAdapterByWWNFunc = (Sun_HBAOpenTgtAdapterByWWNFunc)
		    dlsym(lib_infop->hLibrary,
		    "Sun_fcOpenTgtAdapterByWWN");
		if (GetNumberOfTgtAdaptersFunc == NULL ||
		    OpenTgtAdapterByWWNFunc == NULL) {
			GetNumberOfTgtAdaptersFunc = NULL;
			OpenTgtAdapterByWWNFunc = NULL;
			continue;
		}

		(void) ((GetNumberOfTgtAdaptersFunc)());

		if ((status = (OpenTgtAdapterByWWNFunc)(&handle, nodeWWN))
		    != HBA_STATUS_OK) {
			GetNumberOfTgtAdaptersFunc = NULL;
			OpenTgtAdapterByWWNFunc = NULL;
			continue;
		}
		/* OK, make a vendor non-specific handle */
		*phandle = HBA_HANDLE_FROM_LOCAL(lib_infop->index, handle);
		status = HBA_STATUS_OK;
		break;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

static HBA_STATUS
HBA_NPIV_CheckLibrary(HBA_HANDLE handle,
	HBA_LIBRARY_INFO **lib_infopp,
	HBA_HANDLE *vendorhandle) {
	HBA_UINT32		libraryIndex;
	HBA_LIBRARY_INFO	*lib_infop;

	if (vendorhandle == NULL || handle == HBA_HANDLE_INVALID) {
		return (HBA_STATUS_ERROR_ARG);
	}
	if (_hbaapi_librarylist == NULL) {
		return (HBA_STATUS_ERROR);
	}
	libraryIndex = LIBRARY_NUM(handle);

	GRAB_MUTEX(&_hbaapi_LL_mutex);
	for (lib_infop = _hbaapi_librarylist;
	    lib_infop != NULL;
	    lib_infop = lib_infop->next) {
		if (lib_infop->index == libraryIndex) {
			if (lib_infop->status != HBA_LIBRARY_LOADED) {
				RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex,
				    HBA_STATUS_ERROR_NOT_LOADED);
			}
			*lib_infopp = lib_infop;
			*vendorhandle = VENDOR_HANDLE(handle);
			/* caller will release the mutex */
			return (HBA_STATUS_OK);
		}
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex,
	    HBA_STATUS_ERROR_INVALID_HANDLE);
}
#define	NPIVCHECKLIBRARY() \
	status = HBA_NPIV_CheckLibrary(handle, &lib_infop, &vendorHandle); \
	if (status != HBA_STATUS_OK) { \
		return (status); \
	}

HBA_STATUS
Sun_HBA_NPIVGetAdapterAttributes(
    HBA_HANDLE		handle,
    HBA_ADAPTERATTRIBUTES *hbaattributes)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBANPIVGetAdapterAttributesFunc	NPIVGetAdapterAttributesFunc;

	trace_debug(2, "HBA_NPIVGetAdapterAttributes");

	NPIVCHECKLIBRARY();
	NPIVGetAdapterAttributesFunc = (Sun_HBANPIVGetAdapterAttributesFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcNPIVGetAdapterAttributes");
	if (NPIVGetAdapterAttributesFunc != NULL) {
		status = ((NPIVGetAdapterAttributesFunc)(vendorHandle,
		    hbaattributes));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

HBA_STATUS
Sun_HBA_GetNPIVPortInfo(
    HBA_HANDLE		handle,
    HBA_UINT32		portindex,
    HBA_UINT32		vportindex,
    HBA_NPIVATTRIBUTES	*attributes)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBAGetNPIVPortInfoFunc	GetNPIVPortInfoFunc;

	NPIVCHECKLIBRARY();
	GetNPIVPortInfoFunc = (Sun_HBAGetNPIVPortInfoFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcGetNPIVPortInfo");
	if (GetNPIVPortInfoFunc != NULL) {
		status = ((GetNPIVPortInfoFunc)(vendorHandle, portindex,
		    vportindex, attributes));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

HBA_STATUS
Sun_HBA_DeleteNPIVPort(
    HBA_HANDLE		handle,
    HBA_UINT32		portindex,
    HBA_WWN		vportWWN)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBADeleteNPIVPortFunc	DeleteNPIVPortFunc;

	NPIVCHECKLIBRARY();
	DeleteNPIVPortFunc = (Sun_HBADeleteNPIVPortFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcDeleteNPIVPort");
	if (DeleteNPIVPortFunc != NULL) {
		status = ((DeleteNPIVPortFunc)(vendorHandle,
		    portindex, vportWWN));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

HBA_STATUS
Sun_HBA_CreateNPIVPort(
    HBA_HANDLE		handle,
    HBA_UINT32		portindex,
    HBA_WWN		vnodeWWN,
    HBA_WWN		vportWWN,
    HBA_INT8		*zonename,
    HBA_UINT32		*vportindex)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBACreateNPIVPortFunc	CreateNPIVPortFunc;

	NPIVCHECKLIBRARY();
	CreateNPIVPortFunc = (Sun_HBACreateNPIVPortFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcCreateNPIVPort");
	if (CreateNPIVPortFunc != NULL) {
		status = ((CreateNPIVPortFunc)(vendorHandle,
		    portindex, vnodeWWN, vportWWN, zonename, vportindex));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

HBA_STATUS
Sun_HBA_GetPortNPIVAttributes(
    HBA_HANDLE		handle,
    HBA_UINT32		portindex,
    HBA_PORTNPIVATTRIBUTES	*portnpivattributes)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBAGetPortNPIVAttributesFunc	GetPortNPIVAttributesFunc;

	NPIVCHECKLIBRARY();
	GetPortNPIVAttributesFunc = (Sun_HBAGetPortNPIVAttributesFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcGetPortNPIVAttributes");
	if (GetPortNPIVAttributesFunc != NULL) {
		status = ((GetPortNPIVAttributesFunc)(
		    vendorHandle, portindex, portnpivattributes));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

HBA_STATUS
Sun_HBA_AdapterCreateWWN(
    HBA_WWN		*nwwn,
    HBA_WWN		*pwwn,
    HBA_WWN		*OUI,
    HBA_INT32		method)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	Sun_HBAAdapterCreateWWNFunc	AdapterCreateWWNFunc;

	if (_hbaapi_librarylist == NULL) {
		return (HBA_STATUS_ERROR);
	}

	GRAB_MUTEX(&_hbaapi_LL_mutex);
	for (lib_infop = _hbaapi_librarylist;
	    lib_infop != NULL;
	    lib_infop = lib_infop->next) {
		if (lib_infop->status != HBA_LIBRARY_LOADED) {
			continue;
		}
		AdapterCreateWWNFunc = (Sun_HBAAdapterCreateWWNFunc)
		    dlsym(lib_infop->hLibrary, "Sun_fcAdapterCreateWWN");
		if (AdapterCreateWWNFunc != NULL) {
			status = ((AdapterCreateWWNFunc)(
			    nwwn, pwwn, OUI, method));
		} else {
			status = HBA_STATUS_ERROR_NOT_SUPPORTED;
		}
		RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
	}

	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, HBA_STATUS_ERROR);
}

HBA_STATUS
Sun_HBA_AdapterReturnWWN(
    HBA_HANDLE		handle,
    HBA_UINT32		portindex,
    HBA_WWN		*nwwn,
    HBA_WWN		*pwwn)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;
	Sun_HBAAdapterReturnWWNFunc	AdapterReturnWWNFunc;

	NPIVCHECKLIBRARY();
	AdapterReturnWWNFunc = (Sun_HBAAdapterReturnWWNFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcAdapterReturnWWN");
	if (AdapterReturnWWNFunc != NULL) {
		status = ((AdapterReturnWWNFunc)(vendorHandle,
		    portindex, nwwn, pwwn));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

typedef struct hba_npivadaptercallback_elem {
    struct hba_npivadaptercallback_elem *next;
    HBA_LIBRARY_INFO	*lib_info;
    void		*userdata;
    HBA_CALLBACKHANDLE	vendorcbhandle;
    void		(*callback)();
} HBA_NPIVADAPTERCALLBACK_ELEM;
extern HBA_NPIVADAPTERCALLBACK_ELEM *_hbaapi_adapterdeviceevents_callback_list;

/* Adapter Device Events */
static void
adapterdeviceevents_callback(void *data,
    HBA_WWN	PortWWN,
    HBA_UINT32	eventType,
    HBA_UINT32	fabricPortID)
{
	HBA_NPIVADAPTERCALLBACK_ELEM	*acbp;
	char	wwn_str[17];

	trace_debug(3, "AdapterDeviceEvent, port:%s, "
	    "eventType:%d fabricPortID:0X%06x",
	    WWN2str(wwn_str, &PortWWN), eventType, fabricPortID);

	GRAB_MUTEX(&_hbaapi_APE_mutex);

	for (acbp = _hbaapi_adapterdeviceevents_callback_list;
	    acbp != NULL;
	    acbp = acbp->next) {
		if (data == (void *)acbp) {
			(*acbp->callback)(acbp->userdata, PortWWN,
			    eventType, fabricPortID);
			break;
		}
	}
	RELEASE_MUTEX(&_hbaapi_APE_mutex);
}

HBA_STATUS
Sun_HBA_RegisterForAdapterDeviceEvents(
    void		(*callback) (
		void		*data,
		HBA_WWN		PortWWN,
		HBA_UINT32	eventType,
		HBA_UINT32	fabricPortID),
    void		*userData,
    HBA_HANDLE		handle,
    HBA_WWN		PortWWN,
    HBA_CALLBACKHANDLE	*callbackHandle)
{
	HBA_NPIVADAPTERCALLBACK_ELEM	*acbp;
	HBA_STATUS			status;
	HBA_LIBRARY_INFO		*lib_infop;
	HBA_HANDLE			vendorHandle;
	Sun_HBARegisterForAdapterDeviceEventsFunc
	    registeredfunc;

	if (callbackHandle == NULL) {
		return (HBA_STATUS_ERROR_ARG);
	}

	NPIVCHECKLIBRARY();
	registeredfunc = (Sun_HBARegisterForAdapterDeviceEventsFunc)
	    dlsym(lib_infop->hLibrary,
	    "Sun_fcRegisterForAdapterDeviceEvents");
	if (registeredfunc == NULL) {
		RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex,
		    HBA_STATUS_ERROR_NOT_SUPPORTED);
	}

	acbp = (HBA_NPIVADAPTERCALLBACK_ELEM *)
	    calloc(1, sizeof (HBA_NPIVADAPTERCALLBACK_ELEM));

	if (acbp == NULL) {
		RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, HBA_STATUS_ERROR);
	}

	*callbackHandle = (HBA_CALLBACKHANDLE) acbp;
	acbp->callback = callback;
	acbp->userdata = userData;
	acbp->lib_info = lib_infop;

	status = (registeredfunc)(adapterdeviceevents_callback,
	    (void *)acbp,
	    vendorHandle,
	    PortWWN,
	    &acbp->vendorcbhandle);
	if (status != HBA_STATUS_OK) {
		free(acbp);
		RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
	}

	GRAB_MUTEX(&_hbaapi_APE_mutex);
	acbp->next = _hbaapi_adapterdeviceevents_callback_list;
	_hbaapi_adapterdeviceevents_callback_list = acbp;
	RELEASE_MUTEX(&_hbaapi_APE_mutex);

	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, HBA_STATUS_OK);
}

HBA_STATUS
Sun_HBA_ForceLip(HBA_HANDLE handle, int *rval)
{
	HBA_STATUS		status;
	HBA_LIBRARY_INFO	*lib_infop;
	HBA_HANDLE		vendorHandle;

	Sun_HBADoForceLipFunc	DoForceLipFunc;

	trace_debug(2, "Sun_HBA_DoForceLip");

	NPIVCHECKLIBRARY();
	DoForceLipFunc = (Sun_HBADoForceLipFunc)
	    dlsym(lib_infop->hLibrary, "Sun_fcDoForceLip");
	if (DoForceLipFunc != NULL) {
		status = ((DoForceLipFunc)(vendorHandle, rval));
	} else {
		status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_mutex, status);
}

static HBA_STATUS
Sun_HBA_GetNPIVCapability(HBA_PVOID capability, HBA_UINT32 *capcount)
{
	HBA_STATUS	status = HBA_STATUS_OK;
	HBA_HANDLE	handle;
	HBA_ADAPTERATTRIBUTES	attrs;
	HBA_PORTATTRIBUTES	portattrs;
	HBA_PORTNPIVATTRIBUTES	npivattrs;
	HBA_UINT32	portIndex;
	char		adapterName[HBA_ADAPTER_NAME_MAX_LEN];
	int		cntl_cnt = 0;
	int		numAdapters, i;
	HBA_VPORTCAPABILITY	*cap = (HBA_VPORTCAPABILITY *)capability;

	if ((status = Sun_HBA_LoadLibrary()) != HBA_STATUS_OK) {
		return (status);
	}

	*capcount = 0;
	numAdapters = HBA_GetNumberOfAdapters();
	for (i = 0; i < numAdapters; i++) {
		status = HBA_GetAdapterName(i, adapterName);
		if (status != HBA_STATUS_OK) {
			(void) HBA_FreeLibrary();
			return (status);
		}

		if ((handle = HBA_OpenAdapter(adapterName)) == 0) {
			HBA_CloseAdapter(handle);
			(void) HBA_FreeLibrary();
			return (HBA_STATUS_ERROR);
		}

		if ((status = Sun_HBA_NPIVGetAdapterAttributes(handle, &attrs))
		    != HBA_STATUS_OK)  {
			HBA_CloseAdapter(handle);
			(void) HBA_FreeLibrary();
			return (status);
		}

		for (portIndex = 0; portIndex < attrs.NumberOfPorts;
		    portIndex++) {
			status = HBA_GetAdapterPortAttributes(handle,
			    portIndex, &portattrs);
			if (status != HBA_STATUS_OK) {
				HBA_CloseAdapter(handle);
				(void) HBA_FreeLibrary();
				return (status);
			}
			status = Sun_HBA_GetPortNPIVAttributes(handle,
			    portIndex, &npivattrs);
			if (status != HBA_STATUS_OK) {
				HBA_CloseAdapter(handle);
				(void) HBA_FreeLibrary();
				return (status);
			}

			(void) strcpy(cap[cntl_cnt].Controller,
			    portattrs.OSDeviceName + strlen("/dev/cfg/"));
			bcopy(&portattrs.PortWWN,
			    &cap[cntl_cnt].PhysicalPortWWN,
			    sizeof (HBA_WWN));
			cap[cntl_cnt].PhysicalPortIndex = portIndex;
			cap[cntl_cnt].VPortFlags = 0;
			if (npivattrs.MaxNumberOfNPIVPorts == 0 ||
			    npivattrs.MaxNumberOfNPIVPorts ==
			    npivattrs.NumberOfNPIVPorts ||
			    portattrs.PortState != HBA_PORTSTATE_ONLINE) {
				cap[cntl_cnt].VPortFlags |=
				    HBA_VPORT_FLAG_UNSUPPORTED;
			}
			cap[cntl_cnt].MaxVPorts =
			    npivattrs.MaxNumberOfNPIVPorts;
			cap[cntl_cnt].CurVPorts = npivattrs.NumberOfNPIVPorts;
			cntl_cnt++;
		}
		HBA_CloseAdapter(handle);
	}

	*capcount = cntl_cnt;
	Sun_HBA_FreeLibrary();

	return (HBA_STATUS_OK);
}

HBA_STATUS
Sun_HBA_GetVPortCapability(HBA_VPORTTYPE type, HBA_PVOID capability,
    HBA_UINT32 *capcount)
{
	HBA_STATUS	status;

	assert(capability != NULL && capcount != NULL);

	switch (type) {
		case HBA_VPORT_TYPE_NPIV:
			status =
			    Sun_HBA_GetNPIVCapability(capability, capcount);
			break;
		case HBA_VPORT_TYPE_FCOEOVERVNIC:
			status = HBA_STATUS_ERROR_NOT_SUPPORTED;
			break;
		default:
			status = HBA_STATUS_ERROR_NOT_SUPPORTED;
	}
	return (status);
}

HBA_STATUS
Sun_HBA_LoadLibrary(void) {
	HBA_STATUS status = HBA_STATUS_OK;

	GRAB_MUTEX(&_hbaapi_LL_loaded_mutex);
	if (_hbaapi_loaded_refcnt >= 1) {
		_hbaapi_loaded_refcnt++;
		RELEASE_MUTEX_RETURN(&_hbaapi_LL_loaded_mutex, status);
	}
	status = HBA_LoadLibrary();
	if (status == HBA_STATUS_OK) {
		_hbaapi_loaded_refcnt++;
	}
	RELEASE_MUTEX_RETURN(&_hbaapi_LL_loaded_mutex, status);
}

void
Sun_HBA_FreeLibrary(void) {
	GRAB_MUTEX(&_hbaapi_LL_loaded_mutex);

	_hbaapi_loaded_refcnt--;
	if (_hbaapi_loaded_refcnt == 0) {
		(void) HBA_FreeLibrary();
	}

	RELEASE_MUTEX(&_hbaapi_LL_loaded_mutex);
}
