/*
 * Copyright (c) 2009, 2015, Oracle and/or its affiliates. All rights reserved.
 */

#include <dlfcn.h>
#include <pthread.h>
#include <string.h>
#include <sys/sem.h>

#include "libsun_ima.h"
#include "ima.h"
#include "ima-plugin.h"

extern int number_of_plugins;
extern pthread_mutex_t libMutex;
extern IMA_PLUGIN_INFO	 plugintable[IMA_MAX_NUM_PLUGINS];

static void os_obtainmutex(pthread_mutex_t *mutex);
static void os_releasemutex(pthread_mutex_t *mutex);

IMA_API IMA_STATUS SUN_IMA_SetTunableProperties(
		IMA_OID oid,
		ISCSI_TUNABLE_PARAM *param) {
	SUN_IMA_SetTunablePropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	PassFunc = NULL;
	if ((param == NULL) || (param->tunable_objectValue == NULL)) {
		return (IMA_ERROR_INVALID_PARAMETER);
	}

	if ((oid.objectType != IMA_OBJECT_TYPE_LHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_TARGET)) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (SUN_IMA_SetTunablePropertiesFn)
			    dlsym(plugintable[i].hPlugin,
			    "SUN_IMA_SetTunableProperties");
			if (PassFunc != NULL) {
				status = PassFunc(oid, param);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS SUN_IMA_GetTunableProperties(
		IMA_OID oid,
		ISCSI_TUNABLE_PARAM *param) {
	SUN_IMA_GetTunablePropertiesFn PassFunc = NULL;
	int i;
	IMA_STATUS status;

	if ((param == NULL) || (param->tunable_objectValue == NULL)) {
		return (IMA_ERROR_INVALID_PARAMETER);
	}

	if ((oid.objectType != IMA_OBJECT_TYPE_LHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_TARGET)) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (SUN_IMA_GetTunablePropertiesFn)
			    dlsym(plugintable[i].hPlugin,
			    "SUN_IMA_GetTunableProperties");
			if (PassFunc != NULL) {
				status = PassFunc(oid, param);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS SUN_IMA_GetSendTargetsList(
		IMA_OID oid,
		IMA_TARGET_ADDRESS address,
		SUN_IMA_DISCOVERY_ADDRESS_KEY_PROPERTIES **ppList)
{
	SUN_IMA_GetSendTargetsListFn PassFunc = NULL;
	int i;
	IMA_STATUS status;

	if (ppList == NULL) {
		return (IMA_ERROR_INVALID_PARAMETER);
	}

	if ((oid.objectType != IMA_OBJECT_TYPE_LHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_PHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS)) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (SUN_IMA_GetSendTargetsListFn)
			    dlsym(plugintable[i].hPlugin,
			    "SUN_IMA_GetSendTargetsList");
			if (PassFunc != NULL) {
				status = PassFunc(oid, address, ppList);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

static void os_obtainmutex(pthread_mutex_t *mutex) {
	(void) pthread_mutex_lock(mutex);
}

static void os_releasemutex(pthread_mutex_t *mutex) {
	(void) pthread_mutex_unlock(mutex);
}
