#!/usr/bin/python

#
# Copyright (c) 2015, 2020, Oracle and/or its affiliates.
#

#
# Check that Makefiles in the changeset
#     - can find all referenced include files
#     - have correct KEEP_STATE_FILE syntax
#

import os
import re
import sys


INVALIDSTATEMSG = "{0}: {1}: ERROR - invalid state-file syntax found\n"
INCLNFMSG = "{0}: {1}: ERROR - includes '{2}' not found in this repo\n"
NOTAMFMSG = "{0}: {1}: WARNING - includes '{2}' which is not a Makefile\n"
TOPLVLMKFILE = "{0}: {1}: WARNING - top-level Makefile '{2}' included via " \
    "relative path\n"
OLDSTATEMSG = "{0}: {1}: ERROR - Old state-file syntax found\n"
SRCRELMSG = "{0}: {1}: ERROR - $(SRC_REL) found. Use $(SRC) instead\n"
UNKPREFMSG = "{0}: {1}: ERROR - unknown prefix variable ({2}) found\n"
WARNMSG = "{0}: {1}: WARNING - state-file without prior include statement\n"
WAITMSG = "{0}: {1}: WARNING - .WAIT directive used\n"
WHITESPMSG = "{0}: {1}: ERROR - statement has trailing whitespace\n"
ERROFFMSG = "{0}: WARNING - lint -erroff used. Try removing these or fixing " \
    "the lint issues: {1}\n"
INSDIRMSG = "{0}: {1}: ERROR - Use of $(INS.dir) is deprecated.\n"
PLUSEQUALSMSG = "{0}: {1}: ERROR - '+=' must be surrounded by whitespace.\n"
USELESSADDMSG = "{0}: {1}: WARNING - Useless additive macro.\n"
CCVERBEXTRAMSG = "{0}: {1}: Redundant CCVERBOSE.\n"
CCVERBDISABLEMSG = "{0}: {1}: ERROR - CCVERBOSE should not be disabled.\n"
CCVERBLINTMSG = "{0}: {1}: ERROR - CCVERBOSE (-v) included in LINTFLAGS.\n"
CCVERBCPPMSG = "{0}: {1}: ERROR - CCVERBOSE (-v) included in CPPFLAGS.\n"
ATDEPMSG = "{0}: {1}: WARNING - '$@' detected, did you mean '$$@' ?\n"

CLOSED = "usr/closed/"
MAN = "usr/man/"

IGNORERE = re.compile(r"^[#\s*]")
OLDSTATEFILERE = re.compile(r".KEEP_STATE:")
STATEFILESTMTRE = re.compile(r"(.KEEP_STATE_FILE:)(\s*)(\S*)(.*)")
BDSRE = re.compile(r"(\$\(BUILDDIR\)/.make.state.\$\(MACH\))$")
INSDIRRE = re.compile(r"(.*INS.dir)")
INSDIRASSIGNRE = re.compile(r"INS.dir=")
WAITRE = re.compile(r"r,(.*\.WAIT)")
PLUSEQUALSRE = re.compile(r".*\+=[^\s].*|.*[^\s]\+=.*")
USELESSADDRE = re.compile(r".*\+=\s*$")
CCVERBEXTRARE = re.compile(r"CFLAGS.*\+=.*CCVERBOSE")
CCVERBDISABLERE = re.compile(r"CCVERBOSE.*=\s*$")
CCVERBLINTRE = re.compile(r"LINTFLAGS.*\+=.*CCVERBOSE|LINTFLAGS.*\+=.*-v")
CCVERBCPPRE = re.compile(r"CPPFLAGS.*\+=.*CCVERBOSE|CPPFLAGS.*\+=.*-v")
ATDEPRE = re.compile(r"^\S.*:\s*\$\(*@")

# warn when Makefiles use lint -erroff options
ERROFFRE = re.compile(".*-erroff")

TONICRE = re.compile(r"^(\$\(TONIC.*)")
__ok_masks = [
    "INTEL_BLD",
    "SPARC_BLD",
    "KSPLICE_\\S*",
    "BUILDTESTSPLICE",
    "ANCILLARY",
    "SKIP_\\S*",
    "USESTUB\\S*",
    "CLOSED\\S*",
    "CONS32",
    "CONS64",
    "EXPAND_NIGHTLYID",
    "FISH_BUILD"
]
__ok_masks_re = "|".join(__ok_masks)

OKINCLRE = re.compile(r"\$\((%s)" % __ok_masks_re)
INCLUDERE = re.compile(r"^(.*)(include)(\s*)(.*)")

__prefixstr = [
    # We set these first three ourselves, because we know
    # what they are ALL THE TIME.
    "UTSBASE",
    "UTSCLOSED",
    "UTSTREE",
    "ARCHDIR",
    "BOOTSRCDIR",
    "BRAND_SHARED",
    "CLOSED",
    "CONF_SRCDIR",
    "LIBCDIR",
    "NDMP_DIR",
    "PLATFORM",
    "PROMIF",
    "PSMBASE"
]
__prefixstr_re = "|".join(__prefixstr)
KNOWNPREFIXRE = re.compile(r"^(%s)\)/" % __prefixstr_re)
KNOWNPREFIXDEFRE = re.compile(r"^(%s)\s*=\s*(\S*)" % __prefixstr_re)

# This is the list of files which are exempt from checking
# for erroneous use of $(INS.dir) rules.
__insdir_blessed_files = [
    # These files are supposed to have $(INS.dir) rules.
    # Do NOT add to this list without getting agreement from gatekeeper
    # first.
    # Makefiles for certain Open Source components are listed here
    # since license conditions require that what we make available is
    # buildable outside the firewall.
    #
    # List ON-native files first, and Open Source files second.
    "usr/src/Makefile",
    "usr/src/Targetdirs",
    "usr/src/cmd/sgs/link_audit/Makefile.targ",
    "usr/src/cmd/sgs/packages/Makefile",
    "usr/src/tools/Makefile",
    "usr/src/uts/Makefile.targ",
    "usr/src/uts/common/sys/Makefile.syshdrs",
    # Open Source files
    "usr/src/cmd/dtrace_toolkit/Makefile",
    "usr/src/grub/Makefile.lgrub",
    "usr/src/lib/policykit/Makefile.com",
    "usr/src/lib/libparted/Makefile.com",
    "usr/src/lib/libsqlite/Makefile.com"
]

# This is the list of files which are exempt from checking for incorrect use
# of CCVERBOSE. Makefile.master and Makefile.uts must have it, and the others
# in the list are either legacy code which will never be modified to accept
# CCVERBOSE, or third party code which can't support it, but shouldn't be
# changed (to avoid unnecessary divergence from "upstream").
__ccverbose_blessed_files = [
    "usr/src/cmd/csh/Makefile.com",
    "usr/src/cmd/mail/Makefile",
    "usr/src/cmd/ntfsprogs/Makefile",
    "usr/src/cmd/sh/Makefile",
    "usr/src/cmd/svr4pkg/libinst/Makefile",
    "usr/src/lib/libresolv2/Makefile.com",
    "usr/src/lib/libsqlite/Makefile.com",
    "usr/src/lib/libxcurses/Makefile.com",
    "usr/src/Makefile.master",
    "usr/src/uts/Makefile.uts"
]


def makefilechk(fh, realname, reporoot, verbose=False, allmakefiles=None,
                output=sys.stderr, erroff_warnings=False):

    prefixes = {}
    errcount = 0

    # Firstly, a bit of a hack to see if we're operating in
    # usr/closed rather than usr/src:
    if reporoot.endswith("usr/closed") or reporoot.endswith("usr/man"):
        wsp = os.path.abspath(os.path.join(reporoot, "../..")) + "/"
    else:
        wsp = os.path.abspath(reporoot)
        wsp = wsp + "/"

    outfn = format(realname.replace(wsp, ""))
    basedir = os.path.dirname(realname)

    srcp = os.path.join(wsp, "usr/src")

    # This file contains lots of patterns which are picked up as violations
    # when checking. So we skip all the checks if we detect ourself (since we
    # are not a Makefile).
    if outfn == "usr/src/tools/onbld/Checks/CheckMakefile.py":
        return 0

    # Some Makefiles in uts depend on $(PLATFORM), but we don't
    # always get that specifically defined. Make a WAG...
    if "sun4v" in basedir:
        platform = "sun4v"
    else:
        platform = "i86pc"

    if "sparc" in basedir or "sun" in basedir or "usr/src/stand" in basedir:
        mach = "sparc"
    else:
        mach = "i386"

    # Set up the defaults that we need. Individual Makefiles may
    # override these, but sometimes they're used via an included
    # file's definitions so we set an initial value here.
    prefixes["PLATFORM"] = platform
    prefixes["MACH"] = mach
    prefixes["SRC"] = srcp
    prefixes["CLOSED"] = wsp + CLOSED
    prefixes["BRAND_SHARED"] = srcp + "/lib/brand/shared"
    prefixes["UTSCLOSED"] = "../.."
    prefixes["PROMIF"] = "ieee1275"
    prefixes["BOOTSRCDIR"] = "../.."
    prefixes["LIBCDIR"] = srcp + "/lib/libc"
    if reporoot.endswith("usr/closed"):
        prefixes["UTSBASE"] = "../../../../src/uts"
        prefixes["UTSTREE"] = prefixes["UTSCLOSED"]
    else:
        prefixes["UTSBASE"] = "../.."
        prefixes["UTSTREE"] = prefixes["UTSBASE"]
        prefixes["PSMBASE"] = prefixes["UTSBASE"] + "/../psm"

    includeSeen = False
    erroffs = set()

    for num, line in enumerate(fh, 1):
        nok = False
        if IGNORERE.match(line) or TONICRE.match(line):
            continue

        if PLUSEQUALSRE.match(line):
            # Check that any += directives are surrounded by whitespace,
            # as per the make(1S) man page, applicable to dmake too.
            output.write(PLUSEQUALSMSG.format(outfn, num))

        if USELESSADDRE.match(line):
            # Macro lines ending += with nothing after are pointless.
            # Warn about them.
            output.write(USELESSADDMSG.format(outfn, num))

        if ERROFFRE.match(line) and erroff_warnings:
            # split line into tokens, look for erroffs and try to extract
            # the flags we're disabling. -erroff on its own is also caught
            # (and very bad since that disables all lint errors!)
            toks = line.split()
            erroffs.update([
                t.strip(" \t\r\n").replace("-erroff=", "")
                for t in toks if "-erroff" in t])

        # check the statefile syntax - old-style first
        if OLDSTATEFILERE.match(line):
            output.write(OLDSTATEMSG.format(outfn, num))
            errcount = errcount + 1

        # Check for $(INS.dir) rules outside of those blessed files
        # and paths. Note that assignments are not reported.
        if INSDIRRE.search(line) and outfn not in __insdir_blessed_files:
            if not INSDIRASSIGNRE.match(line):
                output.write(INSDIRMSG.format(outfn, num))
                errcount = errcount + 1

        # Check for .WAIT directives
        if WAITRE.search(line):
            output.write(WAITMSG.format(outfn, num))
            errcount = errcount + 1

        # Check for CCVERBOSE additions or removals, outside of those blessed
        # files.
        if outfn not in __ccverbose_blessed_files:
            if CCVERBEXTRARE.match(line):
                output.write(CCVERBEXTRAMSG.format(outfn, num))
                errcount = errcount + 1
            if CCVERBDISABLERE.match(line):
                output.write(CCVERBDISABLEMSG.format(outfn, num))
                errcount = errcount + 1
            if CCVERBLINTRE.match(line):
                output.write(CCVERBLINTMSG.format(outfn, num))
                errcount = errcount + 1
            if CCVERBCPPRE.match(line):
                output.write(CCVERBCPPMSG.format(outfn, num))
                errcount = errcount + 1

        if ATDEPRE.search(line):
            output.write(ATDEPMSG.format(outfn, num))
            errcount = errcount + 1

        # Now we need to match any prefixes to the included
        # Makefile - such as $(UTSCLOSED)/intel, or $(SRC_REL)
        # We can have several of these validly defined in a Makefile,
        # (eg, UTSBASE and UTSCLOSED) in $CLOSED) so a list is
        # necessary.
        pf = KNOWNPREFIXDEFRE.search(line)
        if pf:
            prefixes[pf.group(1)] = os.path.relpath(
                os.path.join(os.getcwd(), pf.group(2)))
            continue

        # Now we need to match on the "include.*Makefile.*" pattern
        #
        # Now we check that each included Makefile.* is
        # (a) stat-able in the workspace, and
        # (b) part of the manifest (ie, it's been "hg add"ed)
        # (c) if a well-known top-level makefile, included via $(SRC) rather
        #     than relatively.
        #
        # Zerothly: identify and strip off any blessed $(..) variables
        # Any non-blessed $(..) variables are an error

        inclstr = INCLUDERE.match(line)
        if inclstr is not None:
            includeSeen = True
            # Do we have an allowable mask string?
            g1 = inclstr.group(1)
            if re.findall("[=:]", line):
                # This is most likely an assignment line
                # (either variable or target), skip it
                continue
            elif g1.startswith("$"):
                okp = OKINCLRE.match(inclstr.group(1))
                if okp is None:
                    # invalid mask applied
                    output.write(UNKPREFMSG.format(outfn, num, g1))
                    errcount = errcount + 1
            # end of allowable mask check

            # Now we need to do variable substitution
            k = inclstr.group(4)
            if k.endswith(" ") or k.endswith("\t"):
                # We abort checking the line at this point because we
                # don't know just how much there is on the end of the
                # line. The gateling should remove that trailer and
                # recheck.
                output.write(WHITESPMSG.format(outfn, num))
                errcount = errcount + 1
                continue

            unmunged = k   # keep a copy for later
            # We do this twice because some variables include
            # other variables in their definition.
            for n in [0, 1]:
                for i in list(prefixes.keys()):
                    k = re.sub(r"(%s\))" % i, prefixes[i], k)
            j = k.replace("$", "").replace("(", "")
            j = j.replace(")", "")
            # If we've got a $(SRC_REL) in the statement then
            # complain about it.
            if re.match(r"(SRC_REL)", j):
                output.write(SRCRELMSG.format(outfn, num))
                errcount = errcount + 1

            # Do some tidying of relative paths to cope
            # with $(MACH)/../Makefile.com etc
            j = os.path.join(basedir, j)

            # If we're a Makefile.$mumble, then we need to
            # account for being included by a lower-level
            # Makefile. Best-effort is ok - but if you're
            # more than 4 levels below *this* Makefile, you
            # really should consider re-orging your subtree.
            k = j
            if re.match(r"(..)", k):
                found = False
                nok = False
                for i in range(1, 4):
                    apn = os.path.abspath(os.path.join(
                        basedir, k.replace("../", "", i)))
                    if os.path.isfile(apn):
                        k = apn
                        found = True
                        break
                if found is False:
                    ap0 = os.path.normpath(os.path.join(basedir, k))
                    if os.path.isfile(ap0):
                        k = ap0
                    else:
                        nok = True
            # normalise the path before we munge it again
            j = os.path.normpath(k)
            j = j.replace(wsp + "usr", "usr")
            j = j.replace(CLOSED, "")
            j = j.replace(MAN, "")
            j = j.replace("/./", "/").lstrip("/")

            if "Targetdirs" not in j and "Makefile" not in j and \
               not nok:
                # we know it's a file, but it's not a Makefile
                # this is a warning. I'm looking at you, pmcs!
                output.write(NOTAMFMSG.format(outfn, num, j))
            elif j not in allmakefiles or nok is True:
                output.write(INCLNFMSG.format(outfn, num, unmunged))
                errcount = errcount + 1

            # Check if the Makefile being included is a top-level Makefile by
            # measuring the distance from the workspace root. For the purposes
            # of measurement, "top-level" is defined as being within 3
            # directories of $WS, e.g. "usr/src/cmd" or "usr/closed/lib". Files
            # at this level should not be included relatively, unless the file
            # doing the including is also a top-level Makefile (to avoid noise
            # from e.g. cmd/Makefile.cmd including ../Makefile.master).
            if unmunged.startswith("..") and j.count(os.path.sep) <= 3 \
                    and outfn.count(os.path.sep) > 3:
                output.write(TOPLVLMKFILE.format(outfn, num, j))
                errcount = errcount + 1

        # acceptable style matches
        # .KEEP_STATE_FILE:[whitespace]$(BUILDDIR)/.make.state.$(MACH)$
        #
        sfgrs = STATEFILESTMTRE.match(line) or None
        if sfgrs is not None:
            # We need to have $(BUILDDIR) set prior to a valid
            # state-file statement. It is not, however, an error
            # to not have an include statement prior to a state-
            # file statement, since this might be included from
            # a Makefile where we've already got $(BUILDDIR).
            # Warn in this case.
            if includeSeen is False:
                output.write(WARNMSG.format(outfn, num))
            bdsf = BDSRE.match(sfgrs.group(3))
            if bdsf is None or sfgrs.group(4) is not "":
                output.write(
                    INVALIDSTATEMSG.format(outfn, num))
                errcount = errcount + 1
            #  Comment is not required.

    if erroffs:
        output.write(ERROFFMSG.format(outfn, ", ".join(sorted(erroffs))))
    return errcount
