#!/usr/bin/python

#
# Copyright (c) 2008, 2021, Oracle and/or its affiliates.
#

#
# Check commit comments:
# 	- Have the correct form.
# 	- Have a synopsis matching that of the CR or ARC case.
# 	- Appear only once.
# 	- do not have the pipe character '|'  (bugdaemon data delimiter)
#

import re
import sys
import email.utils
from onbld.Checks.DbLookups import BugDB, ARC, Rti, RtiException, \
    defaultLookupFields
from onbld.Checks import get_user_email
from onbld.Checks.Rti import determine_rti_gate


arcre = re.compile(r"^([A-Z][A-Z]*ARC[/ \t][12]\d{3}/\d{3}) (.*)$")
bugre = re.compile(r"^(\d{8}) (.*)$")
followupre = re.compile(r".*\(fix .*\)$")

# These are the acceptable BugDB product_id codes. If your CR is not
# found with one of these, then you really shouldn't be using ON's
# comment checking functions! One restriction we enforce here is that
# all the CRs present in your changeset MUST be part of the same
# product_id except for the product_ids 10006 and 10026.
KNOWN_PRODUCTS = [
    9990,           # LDOMs
    9994,           # ErrorInjector
    10005,          # SolarisCluster
    10006,          # Solaris
    10007,          # SolarisPerfPIT
    10008,          # Solaris-preint
    10009,          # SolarisSystemTest
    10010,          # STC
    10023,          # NWS
    10026,          # ZFSSA
    10027,          # ZFS AIE
    10034,          # TrustedSolaris2
    10358,          # OSSETEST
    12598,          # MiniCluster Specific Software
    "unsupported",  # Other unsupported productids
]


def isARC(comment):
    return arcre.match(comment)


def isBug(comment):
    return bugre.match(comment)


def isFollowUp(comment):
    """Even if the comment appears to be a follow up fix, the caller should
    still check the bug state to make sure it's in state 60."""
    return followupre.match(comment)


#
# Translate any acceptable case number format into "<ARC> <YEAR>/<NUM>"
# format.
#
def normalize_arc(caseid):
    return re.sub(r"^([A-Z][A-Z]*ARC)[/ \t]", "\\1 ", caseid)


def comchk(
        comments, check_db=True, output=sys.stderr,
        bugdb_email=None, gatePath=None, gateCset0="", gateBranch=None,
        bugdb=None):
    """Validate checkin comments against ON standards.

    Comments must be a list of one-line comments, with no trailing
    newline.

    If check_db is True (the default), validate CR and ARC
    synopses against the databases.

    Error messages intended for the user are written to output,
    which defaults to stderr
    """

    # Construct a dictionary, mapping the items in KNOWN_PRODUCTS
    # to empty lists which we can populate with the CRs that correspond
    # to those products.
    okproducts = dict(
        list(zip(KNOWN_PRODUCTS, [[] for item in KNOWN_PRODUCTS])))

    bugnospcre = re.compile(r"^(\d{8})([^ ].*)")
    backoutre = re.compile(r"^backout (\d{8}(/\d{8})*)+: (.*)$")
    sunbtre = re.compile(r"^SUNBT\d{7}\D")
    btreleasere = re.compile(r"^(SUNBT\d{7})-([\w\.\-]+) (.*)$")

    errors = {
        "arc": {},  # we can have different varieties of errors here.
        "bugnospc": [],
        "bugextraspc": [],
        "mutant": [],
        "dup": [],
        "nomatch": [],
        "non-existent": [],
        "badproductid": [],
        "btrelease": [],
        "followup": [],
        "nobugs": False,
        "invalidbackout": [],
        "containspipe": [],
    }
    bugs = {}
    arcs = {}
    ret = 0
    blanks = False
    security = False
    followup_push = False
    backouts = False

    for com in comments:
        # Our input must be newline-free, comments are line-wise.
        if com.find("\n") != -1:
            raise ValueError("newline in comment '%s'" % com)

        # Check for pipe character.
        if com.find("|") != -1:
            errors["containspipe"].append(com)

        # Ignore valid backouts.
        if backoutre.search(com):
            backouts = True
            continue

        # Catch invalid backouts:
        if com.startswith("backout"):
            errors["invalidbackout"].append(com)
            continue

        if not com or com.isspace():
            blanks = True
            continue

        match = bugre.search(com)
        if match:
            if match.group(1) not in bugs:
                bugs[match.group(1)] = []
            bugs[match.group(1)].append(match.group(2))

            if match.group(2).strip() != match.group(2):
                errors["bugextraspc"].append(com)
            continue

        #
        # Bugs missing a space after the ID are still bugs
        # for the purposes of the duplicate ID and synopsis
        # checks.
        #
        match = bugnospcre.search(com)
        if match:
            if match.group(1) not in bugs:
                bugs[match.group(1)] = []
            bugs[match.group(1)].append(match.group(2))
            errors["bugnospc"].append(com)
            if match.group(2).strip() != match.group(2):
                errors["bugextraspc"].append(com)
            continue

        # ARC case
        match = arcre.search(com)
        if match:
            arc, case = re.split("[/ \t]", match.group(1), 1)
            arcs.setdefault((arc, case), []).append(match.group(2))
            continue

        # Anything else is bogus
        errors["mutant"].append(com)

    if len(bugs) == 0 and not backouts:
        # Most of the time, our 'mutant' check will handle this case, but
        # commit comments with only an ARC case and no bugids can escape it,
        # so check for that here.
        errors["nobugs"] = True

    if len(bugs) > 0 and backouts:
        errors["invalidbackout"].append(True)

    if len(bugs) > 0 and check_db:
        if not bugdb:
            if not bugdb_email:
                bugdb_email = get_user_email()
            # parseaddr() returns an empty string when passed None
            bugdb_email = email.utils.parseaddr(bugdb_email)[1]
            if not bugdb_email:
                raise ValueError("Invalid email address")
            bugdb = BugDB(bugdb_email)
        results = bugdb.lookup(
            list(bugs.keys()),
            fields=defaultLookupFields)

    for crid, insts in list(bugs.items()):
        if len(insts) > 1:
            errors["dup"].append(crid)

        if not check_db:
            continue

        if crid not in results:
            errors["non-existent"].append(crid)
            continue

        if results[crid]["security"] == "Y":
            security = True

        #
        # For each synopsis, compare the real synopsis with
        # that in the comments, allowing for possible '(fix
        # stuff)'-like trailing text. For migrated bugs,
        # ensure that release name after "SUNBTXXXXXXX"
        # has been removed. In addition, check that
        # the productid is in our approved list and that each
        # CR in the changeset has the same productid except
        # for the productids 10006 and 100026.
        #
        for entered in insts:
            synopsis = results[crid]["synopsis"]
            pid = results[crid]["product_id"]
            status = results[crid]["status"]

            if not re.search(
                    r"^" + re.escape(synopsis) + r"( \([^)]+\))?$", entered):
                errors["nomatch"].append(
                    [crid, synopsis, entered])

            if sunbtre.match(synopsis):
                match = btreleasere.search(synopsis)
                if match:
                    proper_synopsis = "%s %s" % (
                        match.group(1), match.group(3))
                    errors["btrelease"].append(
                        [crid, synopsis, proper_synopsis])

            if int(pid) not in list(okproducts.keys()):
                errors["badproductid"].append([crid, pid])
                okproducts["unsupported"].append([crid, pid])
            else:
                okproducts[int(pid)].append([crid, pid])

            # We expect that follow-up pushes have existing bugs in state 60.
            # We allow states 80 and 82 here *only* so that we can run
            # unit-tests against the current data in BugDB. RTI approval
            # requires bugs to be in state 25 for integration.
            if isFollowUp(entered) and status in ["60", "80", "82"]:
                followup_push = True

    if followup_push and len(comments) > 1:
        errors["followup"] = True

    if check_db and not followup_push:
        valid = ARC(list(arcs.keys()))
        rtiType, gateName, consolidation = determine_rti_gate(
            gatePath, gateCset0, gateBranch)

        # keep a cache of the Rti objects we've created to reduce the
        # impact of repeated <rti>.queryARC(..) calls, which lazy-loads
        # from an XML webservice.
        known_rtis = {}
        for crid, insts in list(bugs.items()):
            try:
                rti = Rti(crid, gateName, consolidation, rtiType)
                rti_number = rti.rtiNumber()[0]

                if rti_number in known_rtis:
                    rti = known_rtis[rti_number]
                else:
                    known_rtis[rti_number] = rti

                rti_arcs = rti.queryARC()

                # If the ARC case in the comments is not in the RTI, it
                # is an error.  However, it is NOT required for all ARC
                # cases in the RTI to be listed in the comments for a
                # single repo as the ARC case may only be applicable to
                # some repos but not others.
                for arc in set(arcs.keys()):
                    if arc not in set(rti_arcs):
                        errors["arc"]["rti"] = rti_arcs
                        errors["arc"]["changeset"] = list(arcs.keys())

            except RtiException as e:
                output.write(
                    "Unable to cross-check RTI with ARC entries in comments: "
                    "%s\n" % e)

    for case, insts in list(arcs.items()):
        if len(insts) > 1:
            errors["dup"].append(" ".join(case))

        if not check_db or followup_push:
            continue

        if case not in valid:
            errors["non-existent"].append(" ".join(case))
            continue

        #
        # We first try a direct match between the actual case name
        # and the entered comment.  If that fails we remove a possible
        # trailing (fix nit)-type comment, and re-try.
        #
        for entered in insts:
            if entered == valid[case]:
                break
            else:
                # Try again with trailing (fix ...) removed.
                dbcom = re.sub(r" \([^)]+\)$", "", entered)
                if dbcom != valid[case]:
                    errors["nomatch"].append(
                        [" ".join(case), valid[case], entered])

    if blanks:
        output.write("WARNING: Blank line(s) in comments\n")
        ret = 1

    if errors["dup"]:
        ret = 1
        output.write("These IDs appear more than once in your comments:\n")
        for err in errors["dup"]:
            output.write("  %s\n" % err)

    if errors["bugnospc"]:
        ret = 1
        output.write(
            "These bugs are missing a single space following the ID:\n")
        for com in errors["bugnospc"]:
            output.write("  %s\n" % com)

    if errors["bugextraspc"]:
        ret = 1
        output.write(
            "These bugs have leading or trailing spaces in the synopsis:\n")
        for com in errors["bugextraspc"]:
            output.write(" %s\n" % com)

    if errors["btrelease"]:
        ret = 1
        output.write(
            "These migrated bugs have superfluous release "
            "name in synopsis,\nplease update them both in "
            "BugDB and in changeset comment :\n")
        for err in errors["btrelease"]:
            output.write("Synopsis of %s is wrong:\n" % err[0])
            output.write("  should be: '%s'\n" % err[2])
            output.write("         is: '%s'\n" % err[1])

    if errors["mutant"]:
        ret = 1
        output.write("These comments are neither bug nor ARC case:\n")
        for com in errors["mutant"]:
            output.write("  %s\n" % com)
    elif errors["nobugs"]:
        ret = 1
        output.write("Comments must include at least one bugid\n")

    if errors["non-existent"]:
        ret = 1
        output.write("These bugs/ARC cases were not found in the databases:\n")
        for id in errors["non-existent"]:
            output.write("  %s\n" % id)

    if errors["nomatch"]:
        ret = 1
        output.write(
            "These bugs/ARC case synopsis/names don't match the "
            "database entries:\n")
        if security:
            output.write(
                "Security bugs found, please obscure "
                "the bug subject in your changeset comment:\n")
        for err in errors["nomatch"]:
            output.write("Synopsis/name of %s is wrong:\n" % err[0])
            output.write("  should be: '%s'\n" % err[1])
            output.write("         is: '%s'\n" % err[2])

    if len(errors["badproductid"]) > 0:
        ret = 1
        output.write(
            "Your changeset contains CRs with an unsupported product id\n")
        for bpid in errors["badproductid"]:
            output.write("CR %s has product id %s\n" % (bpid[0], bpid[1]))
        output.write("\n")

    if errors["arc"]:
        ret = 1
        output.write(
            "The ARC cases in the comments do not match those in the RTI.\n")
        output.write("ARC cases in the RTI:\n")
        if len(errors["arc"]["rti"]) == 0:
            output.write("    <None found>\n")
        else:
            for case in errors["arc"]["rti"]:
                output.write("    %s %s\n" % case)

        output.write("\nARC cases in the comments:\n")
        if len(errors["arc"]["changeset"]) == 0:
            output.write("    <None found>\n")
        else:
            for case in errors["arc"]["changeset"]:
                output.write("    %s %s\n" % case)

    if errors["followup"]:
        ret = 1
        output.write(
            "Follow up pushes should only include a single comment line "
            "with the most relevant bug, and should omit ARC cases.\n")

    if errors["invalidbackout"]:
        ret = 1
        if True in errors["invalidbackout"]:
            output.write(
                "Bugfixes cannot be pushed at the same time as backouts.")
        else:
            output.write(
                "The following comments are badly formatted for backouts: %s" %
                "    %s\n".join(errors["invalidbackout"]))

    if errors["containspipe"]:
        ret = 1
        output.write("The following comments contain a pipe symbol. This "
                     "should be removed, or seek gatekeeper assistance: %s" %
                     "    %s\n".join(errors["containspipe"]))

    seen = list()
    for n in list(okproducts.keys()):
        if (len(okproducts[n]) > 0) and (okproducts[n] not in seen):
            seen.append(n)

    # CRs in the changeset must have the same productid except for
    # the productids 10006 and 100026.
    if (len(seen) == 2 and (seen != [10006, 10026])) or len(seen) > 2:
        ret = 1
        output.write(
            "All the CRs in your changeset must have the same "
            "product ID assignment.\nPlease check and fix for "
            "these CRs:\n")
        for j in seen:
            output.write("CR %s has product id %s\n" % (
                okproducts[j][0][0],
                okproducts[j][0][1]))
        output.write("\n")
    return ret
