#!/usr/bin/python

#
# Copyright (c) 2008, 2021, Oracle and/or its affiliates.
#

#
# Make sure there is a correctly formed copyright message containing
# the current year.
#
# We treat extant but incorrect copyrights of known format as present
# for the purposes of the "no copyright found" messages to avoid
# treating every otherwise incorrect copyright as also not present
#

import time
import re
import sys


def err(stream, msg, fname, line=None):
    if line:
        stream.write("%s: %d: %s\n" % (fname, line, msg))
    else:
        stream.write("%s: %s\n" % (fname, msg))


# pre-2002 copyright with '(c)'
oldcopyright = re.compile(r"Copyright \(c\) .* Sun Microsystems, Inc\.")

# pre-2002 copyright with 'by'
oldcopyright1 = re.compile(r"Copyright .* by Sun Microsystems, Inc\.")

# last valid Sun copyright
suncopyright = re.compile(
    r"Copyright ([\d, -]+) Sun Microsystems, Inc\." +
    r"(\s+)(All rights reserved\.)?")

# old, check to make sure no longer present
licterms = "Use is subject to license terms."

# initial Oracle copyright, pre-2020
orcl_pre_2020_copyright = re.compile(
    r"Copyright \(c\) (\d\d\d\d, )?(\d\d\d\d)(,)? " +
    r"Oracle and/or its affiliates\.(\s+)" +
    r"All rights reserved\.")

# current Oracle copyright from 2020 onwards.
goodcopyright = re.compile(
    r"Copyright \(c\) (\d\d\d\d)?(,)?( )?(\d\d\d\d)(,)? " +
    r"Oracle and/or its affiliates\.")


def copyright(fh, filename=None, output=sys.stderr):
    ret = lineno = rights = 0
    check_license = False

    if not filename:
        filename = fh.name

    for line in fh:
        lineno += 1

        if check_license:
            check_license = False
            if licterms in line:
                err(
                    output, "old '%s' message found" % licterms, filename,
                    lineno)
                ret = 1
                continue

        if oldcopyright.search(line):
            err(output, "ancient Sun copyright", filename, lineno)
            rights += 1
            ret = 1
            check_license = True
            continue
        elif oldcopyright1.search(line):
            err(output, "pre-2002 Sun copyright", filename, lineno)
            rights += 1
            ret = 1
            check_license = True
            continue
        elif suncopyright.search(line):
            err(output, "old Sun copyright", filename, lineno)
            rights += 1
            ret = 1
            check_license = True
            continue
        elif orcl_pre_2020_copyright.search(line):
            err(output, "old Oracle copyright (remove \"All rights "
                "reserved\" phrase)", filename, lineno)
            rights += 1
            ret = 1
            check_license = True
            continue

        #
        # For the "good" copyright, we locate the following:
        # group 1 = optional initial year
        # group 2 = comma after initial year
        # group 3 = space after initial year
        # group 4 = current year
        # group 5 = comma after current year
        # group 6 = spacing between phrases
        #
        match = goodcopyright.search(line)

        if match:
            # Save them out as 'convenience' variables:
            match_initial_year = match.group(1)
            match_initial_comma = match.group(2)
            match_initial_space = match.group(3)
            match_current_year = match.group(4)
            match_current_comma = match.group(5)
            # match_spacing = match.group(6)

            # only check for the old license message on the line
            # following a copyright match
            check_license = True
            rights += 1

            year = time.strftime("%Y")
            if match_current_year != year:
                err(
                    output, "wrong copyright year %s, should be %s" %
                    (match_current_year, year), filename, lineno)
                ret = 1

            if match_current_comma != ",":
                err(output, "need comma after current year", filename, lineno)
                ret = 1

            if match_initial_year is not None:
                if match_initial_comma != ",":
                    err(output, "need comma after initial year",
                        filename, lineno)
                    ret = 1

                if match_initial_space != " ":
                    err(output, "need space after initial year",
                        filename, lineno)
                    ret = 1

                if match_initial_year >= match_current_year:
                    err(output,
                        "start year %s must be earlier than end year %s" %
                        (match_initial_year, match_current_year),
                        filename, lineno)
                    ret = 1

    if rights == 0:
        err(output, "no copyright message found", filename)
        ret = 1

    return ret
