#!/usr/bin/python

#
# Copyright (c) 2008, 2021, Oracle and/or its affiliates.
#

#
# Various database lookup classes/methods, i.e.:
#
#     * bug.oraclecorp.com (via BugDB proxy daemon)
#     * psarc.us.oracle.com project DB (for ARC on OWAN)
#     * onnv.us.oracle.com/cgi-bin/arc.cgi (for ARC on OWAN, legacy method)
#

# To assist 2->3 transition
from __future__ import print_function

import copy
import csv
import re
import os
import sys

import lxml.etree as etree
from socket import socket, AF_INET, SOCK_STREAM
from onbld.Checks import onOWAN

from six.moves import urllib

# field name mapping
fieldNames = {
    "cr_number": "p_rptno",
    "product_id": "p_product_id",
    "category": "p_component",
    "subcategory": "p_sub_component",
    "release": "p_release",
    "build": "p_fixed_in_label",
    "responsible_engineer": "p_programmer",
    "priority": "p_cs_priority",
    "status": "p_status",
    "synopsis": "p_abstract",
    "assignee_email": "email",
    "security": "security_flag",
    "tags": "tags",
    "bug_type": "CRtype",
}

# A default set of fields that all putback checks require. If you change
# these, make sure the values are sufficient for all putback checks.
defaultLookupFields = [
    "cr_number", "product_id", "security", "synopsis",
    "category", "subcategory", "status", "tags"]

# We prefer the official elpaso bug daemon, but allow an environment
# variable to override this in case of outages, or when testing changes to
# the bug daemon itself.
BUGDB_URL = os.environ.get(
    "ONBLD_BUGDB_URL", "http://elpaso.us.oracle.com:24601")


class NonExistentBug(Exception):
    def __str__(self):
        return "Bug %s does not exist" % (Exception.__str__(self))


class BugDBException(Exception):
    def __str__(self):
        return "Not operating inside Oracle: %s" % (Exception.__str__(self))


class BugDBEmailMissing(Exception):
    def __str__(self):
        s = """\
ERROR: Cannot determine email address for BugDB authentication.

This may be configured in $HOME/.hgrc with the "from" key in the [email]
section or with the "bugdb_email" key in the [cdm] section.  If neither is
present, then $LOGNAME must be set to your userid as a key to look up the
address in LDAP (for which the package python-ldap must be installed).
"""
        return s


class BugDB(object):
    """Lookup change requests.

    Object may ONLY be used inside OWAN, since it connects to the
    BugDB (bug.oraclecorp.com) proxy daemon run by the ON gatekeepers.

    Usage example:
    bdb = BugDB(email="firstname.lastname@oracle.com")
    r = bdb.lookup("15957065")
    print(r["15957065"]["synopsis"])
    r = bdb.lookup(["15957065", "15960625"])
    print(r["15960625"]["synopsis"])
    """

    def __init__(
            self, email=None,
            bugdb_url=BUGDB_URL):
        """Create a BugDB object.
        """
        if not email:
            raise BugDBEmailMissing
        else:
            self.__email = email.split("@")[0]

        self.__validBugDB = ["bugdb"]
        self.__onOWAN = onOWAN()

        # Cache both the fields used by DBlookup for pbchk
        # as well as the results
        self._cache_results = {}
        self._cache_fields = []

        if self.__onOWAN is False:
            raise BugDBException

    def __BugDB(
            self, crs, fields=None,
            bugdb_url=BUGDB_URL, use_cache=True):
        """Return all info for requested change reports.

        Arguments:
        crs: list of change request ids

        fields: optional list of fields to retrieve

        Returns:
        Dictionary, mapping CR=>dictionary, where the nested dictionary
        is a mapping of field=>value
        """

        #
        # See if "maxcrs" for maximal batch query size is defined
        # if not, default to 200.
        #
        try:
            maxcrs
        except NameError:
            maxcrs = 500

        # try to avoid doing the same lookup twice
        if use_cache and fields == self._cache_fields and \
                set(crs) == set(self._cache_results.keys()):
            return copy.deepcopy(self._cache_results)

        urlPattern = "%(bugdb_url)s/usrsrv/?" \
            "email=%%s&crs=%%s&operation=checkcr" % {"bugdb_url": bugdb_url}

        if fields is None:
            fields = defaultLookupFields

        # cr_number must always be present and always come first
        if "cr_number" not in fields:
            fields.insert(0, "cr_number")
        elif fields.index("cr_number") != 0:
            fields.remove("cr_number")
            fields.insert(0, "cr_number")
        i = 0
        urlPattern += "&fields=" + ','.join(fieldNames[f] for f in fields)

        i = 0
        results = {}
        data = []

        while i < len(crs):
            if len(crs) < (i + maxcrs):
                j = len(crs)
            else:
                j = i + maxcrs

            crstmp = crs[i:j]
            # now for some ugliness
            crstr = str()
            for n in range(0, len(crstmp)):
                crstr = crstmp[n] + "," + crstr
            #
            # We request synopsis last, and split on only
            # the number of separators that we expect to
            # see such that a | in the synopsis doesn't
            # throw us out of whack.
            #
            url = urlPattern % (self.__email, crstr[0:-1])
            try:
                data += urllib.request.urlopen(url).readlines()
            except urllib.error.HTTPError as e:
                print(
                    "ERROR: HTTP error at " + url +
                    " got error: " + str(e.code))
                raise e

            except urllib.error.URLError as e:
                print(
                    "ERROR: could not connect to " + url +
                    " got error: %s" % e.reason[1])
                raise e
            i += maxcrs

        for line in data:
            line = line.decode().rstrip("\n")
            values = line.split("|", len(fields) - 1)
            v = 0
            cr = values[0]
            results[cr] = {}
            for field in fields:
                results[cr][field] = values[v]
                v += 1

        self._cache_results = copy.deepcopy(results)
        self._cache_fields = copy.deepcopy(fields)

        return results

    def lookup(
            self, crs, fields=None,
            bugdb_url=BUGDB_URL):
        """
        Return all info for requested change reports.

        Arguments:
        crs: one change request id (may be integer, string, or list),
                 or multiple change request ids (must be a list)

        fields: optional list of fields to retrieve

        Returns:
        Dictionary, mapping CR=>dictionary, where the nested dictionary
        is a mapping of field=>value
        """
        results = {}
        if not isinstance(crs, list):
            crs = [str(crs)]

        results.update(self.__BugDB(crs, fields, bugdb_url=bugdb_url))
        # the CR has already been found by one bug database
        # so don't bother looking it up in the others
        for cr in crs:
            if cr in results:
                crs.remove(cr)

        return results


ARC_XML_HOST = "spe-tools-websvcs.us.oracle.com"
ARC_XML_PORT = "9105"
ARC_XML_URL = "http://%(host)s:%(port)s/arc-webservice/arc?case=%%s" % {
    "host": ARC_XML_HOST,
    "port": ARC_XML_PORT}


def ARC(arclist):
    #
    # connect to ARC web service
    #
    ret = {}
    for a, c in arclist:
        url = ARC_XML_URL % c
        try:
            result = urllib.request.urlopen(url).read()
            parser = etree.XMLParser(strip_cdata=False)
            tree = etree.XML(result, parser)
            if tree.find('.//state').text == "Approved":
                ret[(a, c)] = tree.find('.//name').text
        except Exception:
            print(
                "ERROR: failed to query case %s: %s" % (c, sys.exc_info()[1]))
    return ret


# Pointers to the webrti server hostname & port to use
# Using it directly is probably not *officially* supported, so we'll
# have a pointer to the official `webrticli` command line interface
# if using a direct socket connection fails for some reason, so we
# have a fallback
WEBRTI_HOST = "webrti.us.oracle.com"
WEBRTI_PORT = 9188
WEBRTICLI = "/net/tufnel.us.oracle.com/export/on12-gate/public/bin/webrticli"

WEBRTI_XML_HOST = "spe-tools-websvcs.us.oracle.com"
WEBRTI_SCHEMA = "http://spe-tools-websvcs.us.oracle.com/schema/webrti"
WEBRTI_XML_PORT = "9100"
WEBRTI_XML_URL = "http://%(host)s:%(port)s/WebRTI/rti?webrti=%%s&gate=%%s" % {
    "host": WEBRTI_XML_HOST,
    "port": WEBRTI_XML_PORT}


class RtiException(Exception):
    pass


class RtiCallFailed(RtiException):
    def __str__(self):
        return "Unable to call webrti: %s" % (RtiException.__str__(self))


class RtiSystemProblem(RtiException):
    def __str__(self):
        return "RTI status cannot be determined for CR: %s" % \
            RtiException.__str__(self)


class RtiIncorrectCR(RtiException):
    def __str__(self):
        return "Incorrect CR number specified: %s" % \
            RtiException.__str__(self)


class RtiNotFound(RtiException):
    def __str__(self):
        return "RTI not found for CR: %s" % RtiException.__str__(self)


class RtiNeedConsolidation(RtiException):
    def __str__(self):
        return "More than one consolidation has this CR: %s" % \
            RtiException.__str__(self)


class RtiBadGate(RtiException):
    def __str__(self):
        return "Incorrect gate name specified: %s" % \
            RtiException.__str__(self)


class RtiUnknownException(RtiException):
    def __str__(self):
        return "Unknown webrti return code: %s" % \
            RtiException.__str__(self)


class RtiOffOwan(RtiException):
    def __str__(self):
        return "RTI status checks need OWAN access: %s" % \
            RtiException.__str__(self)


WEBRTI_ERRORS = {
    "1": RtiSystemProblem,
    "2": RtiIncorrectCR,
    "3": RtiNotFound,
    "4": RtiNeedConsolidation,
    "5": RtiBadGate,
}

# Our Rti object which we'll use to represent an Rti query
# It's really just a wrapper around the Rti connection, and attempts
# to establish a direct socket connection and query the webrti server
# directly (thus avoiding a system/fork/exec call).  If it fails, it
# falls back to the webrticli command line client.

returnCodeRe = re.compile(r".*RETURN_CODE=(\d+)")


class Rti:
    """Lookup an RTI.

    Example usage:
      try:
        r = Rti("15555525")
        print(r.rtiNumber())
      except RtiException as e:
        print(e)
    """

    def __init__(self, cr, gate=None, consolidation=None, rtitype=None):
        """Create an Rti object for the specified change request.

        Argument:
        cr: change request id

        Keyword arguments, to limit scope of RTI search:
        gate: path to gate workspace (default=None)
        consolidation: consolidation name (default=None)
        """

        bufSz = 1024
        addr = (WEBRTI_HOST, WEBRTI_PORT)
        # If the passed 'cr' was given as an int, then wrap it
        # into a string to make our life easier
        if isinstance(cr, int):
            cr = str(cr)
        self.__queryCr = cr
        self.__queryGate = gate
        self.__queryConsolidation = consolidation

        self.__webRtiOutput = []
        self.__mainCR = []
        self.__rtiNumber = []
        self.__consolidation = []
        self.__project = []
        self.__status = []
        self.__rtiType = []
        self.__rti_xml = None
        try:
            # try to use a direct connection to the
            # webrti server first
            sock = socket(AF_INET, SOCK_STREAM)
            sock.connect(addr)
            command = "WEBRTICLI/1.0\nRTIstatus\n%s\n" % cr
            if consolidation:
                command += "-c\n%s\n" % consolidation
            if gate:
                command += "-g\n%s\n" % gate
            command += "\n"
            sock.send(command)
            dataList = []
            # keep receiving data from the socket until the
            # server closes the connection
            stillReceiving = True
            while stillReceiving:
                dataPiece = sock.recv(bufSz)
                if dataPiece:
                    dataList.append(dataPiece)
                else:
                    stillReceiving = False
            # create the lines, skipping the first
            # ("WEBRTCLI/1.0\n")
            data = "\n".join("".join(dataList).split("\n")[1:])
        except Exception:
            if not onOWAN():
                raise RtiOffOwan(cr)

            if not os.path.exists(WEBRTICLI):
                raise RtiCallFailed("not found")

            # fallback to the "supported" webrticli interface
            command = WEBRTICLI
            if consolidation:
                command += " -c " + consolidation
            if gate:
                command += " -g " + gate
            command += " RTIstatus " + cr

            try:
                cliPipe = os.popen(command)
            except Exception:
                # we couldn't call the webrticli for some
                # reason, so return a failure
                raise RtiCallFailed("unknown")

            data = cliPipe.readline()

        # parse the data to see if we got a return code
        # if we did, then that's bad.  if we didn't,
        # then our call was successful
        m = returnCodeRe.search(data)
        if m:
            rc = m.group(1)
            # we got a return code, set it in our
            # object, set the webRtiOutput for debugging
            # or logging, and return a failure
            if rc in WEBRTI_ERRORS:
                exc = WEBRTI_ERRORS[rc]
                if exc == RtiBadGate:
                    edata = gate
                else:
                    edata = cr
            else:
                exc = RtiUnknownException
                edata = rc
            raise exc(edata)

        data = data.splitlines()
        # At this point, we should have valid data
        for line in data:
            line = line.rstrip("\r\n")
            # We keep all output here, even if we discard some lines because
            # they don't match the rtitype criteria.
            self.__webRtiOutput.append(line)
            fields = line.split(":")
            if rtitype and fields[5] != rtitype:
                continue
            self.__mainCR.append(fields[0])
            self.__rtiNumber.append(fields[1])
            self.__consolidation.append(fields[2])
            self.__project.append(fields[3])
            self.__status.append(fields[4])
            self.__rtiType.append(fields[5])
        if not self.__mainCR:
            raise RtiNotFound(cr)

    def __load_xml(self):
        if self.__rti_xml:
            return
        # we don't support queries with more than one RTI associated.
        if len(self.rtiNumber()) > 1:
            raise RtiNeedConsolidation(
                ", ".join(set(self.mainCR())) + (" (missing gate path?)"))

        if len(self.rtiNumber()) != 1:
            raise RtiException("We don't have exactly 1 RTI number")
        if len(self.project()) != 1:
            raise RtiException("We don't have exactly 1 project ID")
        # e.g.
        # http://spe-tools-websvcs.us.oracle.com:9100/WebRTI/rti?webrti=402394&gate=5.12
        url = WEBRTI_XML_URL % (self.rtiNumber()[0], self.project()[0])
        try:
            op = urllib.request.urlopen(url)
            self.__rti_xml = op.read()
            op.close()
        except urllib.error.HTTPError as e:
            raise RtiException("HTTP error for {url} got: {e}".format(
                url=url, e=e))

    # accessors in case callers need the raw data
    def mainCR(self):
        return self.__mainCR

    def rtiNumber(self):
        return self.__rtiNumber

    def consolidation(self):
        return self.__consolidation

    def project(self):
        """Return an array of projects for this RTI. Note that "project"
        here means e.g. "5.11" or "5.12"."""
        return self.__project

    def status(self):
        return self.__status

    def rtiType(self):
        return self.__rtiType

    def queryCr(self):
        return self.__queryCr

    def queryGate(self):
        return self.__queryGate

    def queryConsolidation(self):
        return self.__queryConsolidation

    def queryARC(self):
        """Unlike the rest of this class, this uses the XML
        interface provided by spe-tools-websvcs. Eventually, we should
        convert the rest of the methods in this class to consuming
        that xml rather than rtitool output, since we get access to much
        more of the RTI content, which could be useful."""

        self.__load_xml()
        elroot = etree.fromstring(self.__rti_xml)
        schema = None
        try:
            schema = elroot.nsmap[None]
        except (ValueError, KeyError):
            # WebRTI sometimes returns xml for temporary or preliminary RTIs
            # without a corresponding namespace. In those cases, using an
            # empty schema is fine.
            pass
        # if we're using a default namespace schema, then Etree expects
        # to see that in xpath expressions to find tags.
        # See http://lxml.de/xpathxslt.html
        if schema:
            schema = "{%s}" % schema
        else:
            schema = ""
        cases = elroot.findall(".//%sarcCase" % schema)
        if not cases:
            return []
        arc_ids = []
        for case in cases:
            arc = None
            year = None
            number = None

            arc_node = case.find(".//%sarc" % schema)
            if arc_node is not None:
                arc = arc_node.text.strip()
            year_node = case.find(".//%syear" % schema)
            if year_node is not None:
                year = year_node.text.strip()
            number_node = case.find(".//%snumber" % schema)
            if number_node is not None:
                number = number_node.text.strip()

            if year and number and arc:
                arc_ids.append((arc, "%s/%s" % (year, number)))
        return arc_ids

    # in practice, most callers only care about the following
    def accepted(self):
        for status in self.__status:
            if status != "S_ACCEPTED":
                return False
        return True

    # for logging/debugging in case the caller wants the raw webrti output
    def webRtiOutput(self):
        return self.__webRtiOutput
