#!/usr/bin/python

#
# Copyright (c) 2014, 2018, Oracle and/or its affiliates. All rights reserved.
#

import os
import sys

# We set this as a global so we don't have to rebuild the dictionary
# each time we call the parfait_remind method.
fortify_db = {}
printed_issues_header = False


def fortify_remind(
        filename, repo_root, branch, output=sys.stderr, subrepo=""):
    """Remind users about source files that have unresolved Fortify issues.
    This putback check is purely informative, and never fails. We return
    a non-zero status if we have problems accessing the parfait-remind
    database so that the caller can choose to avoid subsequent calls to
    this method.
    If 'subrepo' is set to "fish" or "closed", we prepend paths to those
    repositories when searching for Fortify issues with the given filename.
    """

    global printed_issues_header

    lookup_filename = os.path.abspath(filename)
    lookup_filename = lookup_filename.replace(repo_root, "").lstrip("/")

    # The filename is relative to the root of the repository, however
    # the fortify-remind database contains a path relative to $WS.
    if subrepo:
        lookup_filename = os.path.join("usr", subrepo, lookup_filename)

    fortify_db_path = os.path.join(
        os.environ.get("PARENT_WS", "/ws/on-gate"), "public", branch,
        "fortify-remind-db.txt")
    if not fortify_db:
        try:
            if not os.path.exists(fortify_db_path):
                output.write(
                    "Warning: {} does not exist\n".format(fortify_db_path))
                output.write(
                    "         unable to check for existing Fortify issues.\n")
                return 1
            with open(fortify_db_path, "r") as fortify_db_file:
                for line in fortify_db_file:
                    if line.startswith("#"):
                        continue
                    path, verified, unverified = line.strip().split(" ")
                    fortify_db[path] = (
                        unverified, int(verified) + int(unverified))
        except (IOError, ValueError, KeyError) as e:
            output.write(
                "Warning: unable to load {}: {}\n".format(fortify_db_path, e))
            output.write(
                "         unable to check for existing fortify issues.\n")
            return 1

    if lookup_filename in fortify_db:
        if not printed_issues_header:
            output.write(
                "Existing issues were found for files in this putback. "
                "(unverified/total)\n")
            printed_issues_header = True
        output.write(
            "  {}: {}/{}\n".format(
                lookup_filename, fortify_db[lookup_filename][0],
                fortify_db[lookup_filename][1]))
    return 0
