#!/usr/bin/python

#
# Copyright (c) 2016, 2017, Oracle and/or its affiliates. All rights reserved.
#
import logging
import os
import sys
import textwrap

from onbld.build import load_tests


def notify(filelist, root, output=sys.stderr):
    """
    Issue developer notifications and advice about the presence of unit tests
    that match the files being modified.

    filelist is a list of 2-tuples. The first element is a path
    relative to the current directory, the second is an entry from the
    active  list, or None if an explicit file list was given. This
    corresponds to what is retured by buildfilelist(..) in the ON
    Cadmium mercurial extension, cdm.py.

    root is the root of the current hg repository.

    output is a file object where we write messages.
    """
    if not filelist:
        return

    ws = root
    if ws.endswith("usr/closed"):
        ws = ws.replace("usr/closed", "")
    elif ws.endswith("usr/fish"):
        ws = ws.replace("usr/fish", "")
    elif ws.endswith("usr/man"):
        ws = ws.replace("usr/man", "")

    logger = logging.getLogger("cdm_notify")
    logger.addHandler(logging.StreamHandler(output))
    logger.setLevel(logging.INFO)

    all_tests = load_tests(ws, os.environ, logger)

    matching_tests = {}
    notices = {}

    for f, e in filelist:
        if e and e.is_removed():
            continue
        lookup_filename = os.path.abspath(f)
        lookup_filename = lookup_filename.replace(root, "").lstrip("/")
        if root.endswith("usr/closed"):
            lookup_filename = "usr/closed/%s" % lookup_filename
        elif root.endswith("usr/fish"):
            lookup_filename = "usr/fish/%s" % lookup_filename
        elif root.endswith("usr/man"):
            lookup_filename = "usr/man/%s" % lookup_filename
        lookup_filename = lookup_filename.replace(ws, "").lstrip("/")

        for test in all_tests:
            test_file = os.path.basename(test.cfg_path)
            matching_pattern = None
            for cpattern, pattern in test.source_regexps:
                if cpattern.match(lookup_filename):
                    matching_pattern = pattern
                    break

            if matching_pattern:
                if test.script:
                    matching_tests.setdefault(test_file, set()).add(
                        matching_pattern)
                if test.notice:
                    notices.setdefault(test_file, set()).add(test.notice)

    if matching_tests:
        output.write(
            "Source files changed in this putback that match the following "
            "patterns:\n\n")
        for test_file in matching_tests:
            output.write("%s:\n" % test_file)
            for pattern in matching_tests[test_file]:
                output.write("\t%s\n" % pattern)
            output.write("\n")
        output.write(
            "Unit tests are available in the gate to test this code.\n"
            "Please ensure all tests are run before putback.\n"
            "See test configuration files in $SRC/tools/test.d/\n\n"
            "Use 'build test <file or section name> -v' to run tests\n"
            "and 'build test' to list all known tests.\n")

    if notices:
        output.write(
            "Developer advice exists for files in this integration:\n\n")
        for test_file in notices:
            output.write("%s:\n" % test_file)
            for notice in notices[test_file]:
                # ConfigParser doesn't allow embedded blank lines.
                # to work around that, we convert the literal "\n\n"
                # string to two newline chars.
                # textwrap doesn't respect paragraphs, so we wrap each
                # paragraph separately to preserve those in the output.
                paras = notice.replace("\\n\\n", "\n\n").split("\n\n")
                wrapped_notice = ""
                for para in paras:
                    wrapped_para = "\n".join(
                        textwrap.wrap(
                            para.strip(), initial_indent="\t",
                            subsequent_indent="\t"))
                    wrapped_notice += "%s\n\n" % wrapped_para
                output.write(wrapped_notice)
