#!/usr/bin/python

#
# Copyright (c) 2014, 2017, Oracle and/or its affiliates. All rights reserved.
#

import os
import sys

# We set this as a global so we don't have to rebuild the dictionary
# each time we call the parfait_remind method.
parfait_db = {}
printed_issues_header = False


def parfait_remind(
        filename, repo_root, branch, output=sys.stderr, subrepo=""):
    """Remind users about source files that have unresolved Parfait issues.
    This putback check is purely informative, and never fails. We return
    a non-zero status if we have problems accessing the parfait-remind
    database so that the caller can choose to avoid subsequent calls to
    this method.
    If 'subrepo' is set to "fish" or "closed", we prepend paths to those
    repositories when searching for parfait issues with the given filename.
    """

    global printed_issues_header

    lookup_filename = os.path.abspath(filename)
    lookup_filename = lookup_filename.replace(repo_root, "").lstrip("/")

    # The filename is relative to the root of the repository, however
    # the parfait-remind database contains a path relative to $WS.
    if subrepo:
        lookup_filename = "usr/%s/%s" % (subrepo, lookup_filename)

    parfait_db_path = os.path.join(
        os.environ.get("PARENT_WS", "/ws/on-gate"), "public", branch,
        "parfait-remind-db.txt")
    if not parfait_db:
        try:
            if not os.path.exists(parfait_db_path):
                output.write("Warning: %s does not exist\n" % parfait_db_path)
                output.write(
                    "         unable to check for existing parfait issues.\n")
                return 1
            with open(parfait_db_path, "r") as parfait_db_file:
                for line in parfait_db_file:
                    if line.startswith("#"):
                        continue
                    path, verified, unverified = line.strip().split(" ")
                    parfait_db[path] = (
                        unverified, int(verified) + int(unverified))
        except (IOError, ValueError, KeyError) as e:
            output.write(
                "Warning: unable to load %s: %s\n" % (parfait_db_path, e))
            output.write(
                "         unable to check for existing parfait issues.\n")
            return 1

    if lookup_filename in parfait_db:
        if not printed_issues_header:
            output.write(
                "Existing issues were found for files in this putback. "
                "(unverified/total)\n")
            printed_issues_header = True
        output.write(
            "  %s: %s/%s\n" %
            (lookup_filename, parfait_db[lookup_filename][0],
                parfait_db[lookup_filename][1]))
    return 0
