#!/usr/bin/python

#
# Copyright (c) 2008, 2018, Oracle and/or its affiliates. All rights reserved.
#

#
# Check on RTI status for bug IDs passed.
#
# How we obtain the bug IDs will vary per SCM.
#       - For Teamware, we want to check the active list comments.
#       - For Mercurial, we can check the incoming changegroup (via the
#         pretxnchangegroup hook) and abort if necessary
#
# This module is implemented as a generic checking module given a list of
# bug IDs.  It can then be wrapped or hooked into whatever SCM with an
# SCM-specific hook to parse and pass the requisite bug IDs
#

from collections import defaultdict
import os
import re
import sys
from onbld.Checks.DbLookups import Rti
from onbld.Checks.DbLookups import RtiException
from onbld.Checks.DbLookups import RtiOffOwan

opensolarisGateRe = re.compile(r".*osol[0-9]+-sust$")
patchGateRe = re.compile(r".*-patch.*")
updateGateRe = re.compile(r".*update.*")
s11uxGateRe = re.compile(r".*on11u.*")
oseGateRe = re.compile(r".*onose[0-9]+-sust.*")
sruGateRe = re.compile(r".*sru.*")
ldomSustainingGateRe = re.compile(r"ldom.*sust.*")
ldomMicroGateRe = re.compile(r"ldom.*micro.*")
testGateRe = re.compile(r"(stc.*-gate|on.*-(test|stc2))$")
akMinorGateRe = re.compile(r"ak-[2-9][0-9][0-9][0-9].*")

# Map the first changeset in each repo to its corresponding gate name.
cset0map = {
    "68f95e015346": "ON",  # usr/src
    "2be569732930": "ON",  # usr/closed
    "a4354b0c67d9": "ON",  # usr/man
    "442de104a6ad": "ON",  # usr/fish
    "a62e9e23639a": "userland",
}

# Map the consolidation name to a dict mapping branch name to the type of RTI.
branchmaps = defaultdict(
    lambda: {
        # At least ON and Userland follow this branch naming strategy.
        "default": ("MarketingRelease", "trunk"),
        "server": ("MarketingRelease", "trunk"),
        "s11u[123]": ("SolarisUpdate", "update"),
        ".*-update": ("SolarisUpdate", "update"),
        ".*-sru": ("SolarisUpdate", "sru"),
        "ak-.*": ("SolarisUpdate", "ak"),
        "s11u4": ("SolarisUpdate", "s11u4"),
    }
)


def determine_rti_gate(gatePath=None, gateCset0="", gateBranch=None):
    """Based on what we know about a repository, return the information
    necessary to get WebRTI to find the RTIs for a set of bugid.

    The inputs are the URL to the parent repo, the 0th changeset id of the
    repo, and the branch the active changesets are on.  All are optional.  We
    work off of the branch and root changeset first, since parent repo URLs are
    highly ambiguous, especially if the parent repo isn't the central one.

    We return a tuple of RTI type, gate name, and consolidation name, all of
    which are inputs to WebRTI.  The RTI type essentially determines whether
    the changes are destined for the trunk, marketing release, update, SRU,
    etc, depending on how the consolidation is configured in WebRTI.  The
    consolidation name is self-explanatory, though the names are defined by
    WebRTI.  The gate name is an amalgam of the two, based on a time when
    that's all the information we could actually glean.
    """

    rtiType = None
    release = None

    # Map consolidation/branch back to the "gate name" that WebRTI expects.
    # This is somewhat redundant with the consolidation name, as far as WebRTI
    # is concerned, and can lead to bad results if the gate name is incorrect.
    # However, WebRTI does use it to filter the returned RTIs to match the
    # intended type, since that's not a thing you can pass in directly.  For
    # consolidations where the gate names are not known, we have to filter that
    # client-side, anyway.
    # XXX: the gate name for AK is temporarily changed to ak-8.8-gate as
    # that's the gate name WebRTI expects.  This needs some discussion.
    gatenames = {
        ("ON", "MarketingRelease", "trunk"): "on-gate",
        ("ON", "SolarisUpdate", "sru"): "on-sru",
        ("ON", "SolarisUpdate", "update"): "on11-update",
        ("ON", "SolarisUpdate", "ak"): "ak-8.8-gate",
        ("ON", "SolarisUpdate", "s11u4"): "on11u4sru-gate",
    }

    consolidation = cset0map.get(gateCset0[:12])
    branchmap = branchmaps[consolidation]
    if branchmap and gateBranch:
        for pat in list(branchmap.keys()):
            if re.match(pat, gateBranch):
                rtiType, release = branchmap[pat]

    gateName = gatenames.get((consolidation, rtiType, release))

    if (consolidation or gateName) and rtiType:
        return rtiType, gateName, consolidation

    # If we couldn't figure out the consolidation (or gate name) and RTI type
    # from SCM information, check to see if we were given a gate path, and try
    # figuring it out from that.  The default is marketing release, so we only
    # have to test for update and SRU gate names.
    rtiType = "MarketingRelease"
    if gatePath is not None:

        #
        # The gate name should be the last component of the gate path,
        # no matter how it's accessed.
        #
        # We make a special case for both "closed" and "man", checking
        # to see if it appears to be the "usr/[closed,|man]" portion of
        # a nested repository. In that case, we really want the parent
        # repository name.
        #
        gatePath = gatePath.rstrip(os.path.sep).split(os.path.sep)
        gateName = gatePath[-1]
        try:
            if (gatePath[-2:] == ["usr", "closed"]) or \
                    (gatePath[-2:] == ["usr", "fish"]) or \
                    (gatePath[-2:] == ["usr", "man"]):
                gateName = gatePath[-3]
        except IndexError:
            pass

        # Is this an OpenSolaris gate?
        if opensolarisGateRe.search(gateName):
            rtiType = "OpenSolaris"

        # Is this a patch gate?
        if patchGateRe.search(gateName):
            rtiType = "Patch"

        # Is this an OSE (OpenSolarisExpress) gate?
        if oseGateRe.search(gateName):
            rtiType = "OSE-11"

        # Is this an S11 update gate?
        if updateGateRe.search(gateName):
            rtiType = "SolarisUpdate"

        if s11uxGateRe.search(gateName):
            rtiType = "SolarisUpdate"

        # Is this an S11 SRU gate?
        if sruGateRe.search(gateName):
            rtiType = "SolarisUpdate"

        # Is this LDOM sustaining gate?
        if ldomSustainingGateRe.search(gateName):
            rtiType = "SolarisUpdate"

        # Is this LDOM Micro gate?
        if ldomMicroGateRe.search(gateName):
            rtiType = "SolarisUpdate"

        # Is this a test gate?
        if testGateRe.search(gateName):
            rtiType = "RTI"

        # Is this an AK minor gate?
        if akMinorGateRe.search(gateName):
            rtiType = "SolarisUpdate"
    else:
        gateName = None

    return rtiType, gateName, None


def rti(bugids, gatePath=None, gateCset0="", gateBranch=None,
        output=sys.stderr):
    """Return True iff each of the specified bugids has an approved RTI.

    Required argument:
    bugids:	list of seven-digit bug ids

    Keyword arguments, used to limit the scope of the RTI search:
    gatePath: fully qualified path to gate
    """

    # Use any passed-in consolidation in favor of the auto-determined one
    expectedRTIType, gateName, consolidation = determine_rti_gate(
        gatePath, gateCset0, gateBranch)

    # Query RTI if there's a gate
    # Check the RTIs, caching them in the "rtis" dictionary
    # We do our error checking/handling here
    rtis = {}
    badRtis = []
    for cr in bugids:
        # If we don't already have an RTI object for this cr cached,
        # then go create/query it
        if cr not in list(rtis.keys()) + badRtis:
            try:
                rtis[cr] = Rti(cr, gateName, consolidation, expectedRTIType)
            except RtiOffOwan as e:
                output.write("%s\n" % e)
                return False
            except RtiException as e:
                output.write("%s\n" % e)
                badRtis.append(cr)
                continue

        crRti = rtis[cr]

        # If we've reached this point, then the RTI query succeeded, and we
        # didn't get an error back from WebRTI.  There is still some sanity
        # checking to be done, however.
        rtiNumber = crRti.rtiNumber()
        rtiTypes = crRti.rtiType()

        # check to make sure the expected RTI type matches what was found in
        # WebRTI for the bugid.
        if expectedRTIType not in rtiTypes:
            for each in rtiNumber:
                if consolidation:
                    message = (
                        "The RTI {rti} is of {rtitype} type but the "
                        "changeset fixing bug\n{cr} is on a branch that "
                        "requires a {gatetype}-type RTI.\n"
                    )
                else:
                    message = (
                        "The RTI {rti} is of {rtitype} type but the "
                        "parent gate {gatename} is a {gatetype} gate.\n"
                        "A {gatetype} RTI must be submitted to putback bug "
                        "{cr} to {gatename}.\n"
                    )
                message = message.format(
                    rti=each,
                    rtitype=rtiTypes[rtiNumber.index(each)],
                    gatename=gateName,
                    gatetype=expectedRTIType,
                    cr=cr)
                output.write(message)
                badRtis.append(cr)
                continue

        # RTI's main CR should match, otherwise gate hooks will fail.
        for i in range(0, len(rtiNumber)):
            if i > len(crRti.mainCR()):
                break
            rtiNum = rtiNumber[i]
            mainCR = crRti.mainCR()[i]
            if mainCR == cr:
                continue
            message = (
                "Error: RTI {rtinum} for CR {cr} has a mismatched "
                "main CR {maincr}\n".format(
                    rtinum=rtiNum, cr=cr, maincr=mainCR))
            if len(mainCR) == 7:
                message += "  Main CR appears to be a " \
                    "Bugtraq ID, resubmit RTI using BugDB ID\n"
            output.write(message)
            badRtis.append(cr)

        if not crRti.accepted():
            for each in rtiNumber:
                message = (
                    "Error: RTI {rti} for CR {cr} is not in "
                    "the accepted state.\n".format(cr=cr, rti=each))
                output.write(message)
            badRtis.append(cr)
            continue

    if len(badRtis) > 0:
        return False

    return True
