#!/usr/bin/python

#
# Copyright (c) 2016, 2021, Oracle and/or its affiliates.
#

#
# Check for common spelling errors.
#
# Note that only "common" misspellings should be added here to avoid more
# noise than necessary. If you only had to correct one instance, then that's
# a good indication that it doesn't need to be in this list.
#

from __future__ import print_function

from difflib import context_diff

import re
import sys

wordlist = {
    "acceptible": "acceptable",
    "accessable": "accessible",
    "accidently": "accidentally",
    "accros,accross": "across",
    "acknowlege": "acknowledge",
    "acknowleged": "acknowledged",
    "acknowlegment,acknowlegement": "acknowledgment",
    "acknowleging": "acknowledging",
    "acknowleges": "acknowledges",
    "acomodate,accomodate,acommodate": "accommodate",
    "actitivity": "activity",
    "actived": "activated",
    "adaptor": "adapter",
    "addional": "additional",
    "adresses,addreses": "addresses",
    "adress,addres": "address",
    "afterwards": "afterward",
    "alogrithms": "algorithms",
    "annonymous": "anonymous",
    "appropriatedly": "appropriately",
    "aquire": "acquire",
    "aquisition": "acquisition",
    "aquit": "acquit",
    "alot": "a lot' or 'allot",
    "amatuer": "amateur",
    "apparant": "apparent",
    "arguement": "argument",
    "assymetric,asymetric": "asymmetric",
    "auxilliary": "auxiliary",
    "beleive": "believe",
    "boundries": "boundaries",
    "bufferiing": "buffering",
    "cacheing": "caching",
    "calender": "calendar",
    "cancelability": "cancellability",
    "capabilites,capabilties": "capabilities",
    "catagory": "category",
    "chararcter,charater": "character",
    "challange": "challenge",
    "changable": "changeable",
    "changeing": "changing",
    "choosen": "chosen",
    "coincidently": "coincidentally",
    "comitted,commited": "committed",
    "compatiblity": "compatibility",
    "concatenatation,concatentation": "concatenation",
    "conext": "context",
    "correponding": "corresponding",
    "deamon": "daemon",
    "defaule": "default",
    "defintons": "definitions",
    "deleteing": "deleting",
    "delimeter": "delimiter",
    "dependant": "dependent",
    "dependants": "dependents",
    "determing": "determining",
    "directroy": "directory",
    "discrepency": "discrepancy",
    "disipline,dicipline": "discipline",
    "domian": "domain",
    "dynamicly": "dynamically",
    "eligable,elegible": "eligible",
    "encounterd": "encountered",
    "endianess": "endianness",
    "enviornment,enviroment": "environment",
    "epock": "epoch",
    "equivelent": "equivalent",
    "equivalant": "equivalent",
    "excede": "exceed",
    "exceded": "exceeded",
    "excedes": "exceeds",
    "exaust": "exhaust",
    "exausts": "exhausts",
    "exausted": "exhausted",
    "exaustion": "exhaustion",
    "exensions": "extensions",
    "existant": "existent",
    "existance": "existence",
    "experminental,experimentatl": "experimental",
    "expresssion": "expression",
    "extention": "extension",
    "filesytem": "filesystem",
    "flushs": "flushes",
    "functon,funtion,fuction": "function",
    "gaurantee": "guarantee",
    "gauranteed": "guaranteed",
    "gaurantees": "guarantees",
    "guarante,gurantee": "guarantee",
    "gaurd": "guard",
    "gaurded": "guarded",
    "gaurding": "guarding",
    "gaurds": "guards",
    "hander": "handler",
    "happend": "happened",
    "heirarchy,hieararchy,hierachy,hierarachy": "hierarchy",
    "hexidecimal": "hexadecimal",
    "hypen": "hyphen",
    "idealy": "ideally",
    "identifer,identfier": "identifier",
    "imediately": "immediately",
    "inaccessable": "inaccessible",
    "incidently": "incidentally",
    "incompatable": "incompatible",
    "indentity": "identity",
    "independant": "independent",
    "infomation,infromation": "information",
    "inputed": "inputted",
    "intented": "intended",
    "intepreted,intrepreted": "interpreted",
    "interace": "interface",
    "interpets": "interprets",
    "interpretor": "interpreter",
    "interrup": "interrupt",
    "labelled": "labeled",
    "labelling": "labeling",
    "libary": "library",
    "likly": "likely",
    "likelyhood": "likelihood",
    "loosing": "losing",
    "lower-case": "lowercase",
    "maintainance,maintainence,maintenence": "maintenance",
    "managable": "manageable",
    "manger": "manager",
    "migraton": "migration",
    "millenium,milennium,milenium": "millennium",
    "millsecond,millesecond": "millisecond",
    "millseconds,milleseconds": "milliseconds",
    "miniscule": "minuscule",
    "mulitbyte,multi-byte": "multibyte",
    "naturaly": "naturally",
    "neccessary": "necessary",
    "non existant,non existent,nonexistant,nonexistent": "non-existent",
    "noticable,noticible": "noticeable",
    "occassional": "occasional",
    "ocassionally": "occasionally",
    "occured": "occurred",
    "occurance,occurrance,occurence": "occurrence",
    "occurrs,occurres": "occurs",
    "oficial": "official",
    "omited,ommitted": "omitted",
    "opration": "operation",
    "overriden": "overridden",
    "paramenters": "parameters",
    "parmeter,paramter": "parameter",
    "particuler,particuliar": "particular",
    "perminent,permonent,permenant,permanant": "permanent",
    "persistance": "persistence",
    "posess,posses": "possess",
    "posession": "possession",
    "preceed": "precede",
    "precidence,precidance,precedance": "precedence",
    "prefered": "preferred",
    "preperation": "preparation",
    "priviledge,privlege,privilige,privalege": "privilege",
    "priviledges,privleges,priviliges,privaleges": "privileges",
    "procede": "proceed",
    "processs": "process",
    "programatically": "programmatically",
    "propertys": "properties",
    "protcol": "protocol",
    "psuedo": "pseudo",
    "publically": "publicly",
    "recieve": "receive",
    "reciept": "receipt",
    "reccomend,recomend": "recommend",
    "regardess,regardles": "regardless",
    "recommanded": "recommended",
    "refered": "referred",
    "refernce": "reference",
    "releas": "release",
    "relevent": "relevant",
    "replacable": "replaceable",
    "retreive": "retrieve",
    "retrival": "retrieval",
    "returing": "returning",
    "reurn": "return",
    "run-time": "runtime",
    "safty": "safety",
    "sart": "start",
    "searchs": "searches",
    "seperate": "separate",
    "seperated": "separated",
    "seperation": "separation",
    "shareable": "sharable",
    "similiar": "similar",
    "simultaneus,simultanious": "simultaneous",
    "simultanously": "simultaneously",
    "singe": "single",
    "smaler": "smaller",
    "specifed,specfied": "specified",
    "specificiation": "specification",
    "specificly": "specifically",
    "specifing": "specifying",
    "sring": "string",
    "struture": "structure",
    "succede,succed,suceed": "succeed",
    "succeded": "succeeded",
    "succesful,successfull": "successful",
    "superseed": "supersede",
    "suports": "supports",
    "supress": "suppress",
    "susbsystems": "subsystems",
    "suspectible": "susceptible",
    "symetric": "symmetric",
    "symetrical": "symmetrical",
    "synching": "syncing",
    "synonomous": "synonymous",
    "sytem": "system",
    "targetted": "targeted",
    "targetting": "targeting",
    "the the": "the",
    "therfore,therefor": "therefore",
    "threshhold,threhold": "threshold",
    "towards": "toward",
    "transfering": "transferring",
    "truely": "truly",
    "trys": "tries",
    "unecessary,unnecesary,unnecessery,unecessery,unnecesery": "unnecessary",
    "untill": "until",
    "unsuccesful,unsuccessfull": "unsuccessful",
    "unsufficient": "insufficient",
    "upper-case": "uppercase",
    "vaild": "valid",
    "vender": "vendor",
    "verifed": "verified",
    "verisioned": "versioned",
    "visable": "visible",
    "wierd": "weird",
    "wilcards": "wildcards",
    "writen": "written",
    "wwek": "week",
    "yeild": "yield",
    "yeilds": "yields",
}

global pattern
global mapper


def err(stream, msg, fname, line=None):
    if line:
        stream.write("%s: %d: %s\n" % (fname, line, msg))
    else:
        stream.write("%s: %s\n" % (fname, msg))


def suggest(bad, good):
    if bad == bad.lower():
        return (bad, good)

    if bad == bad.upper():
        return (bad, good.upper())
    if bad == bad.capitalize():
        return (bad, good.capitalize())
    if bad == bad.title():
        return (bad, good.title())

    return (bad.lower(), good)


def spel_chekc(fh, filename=None, output=sys.stderr, patch=False):
    ret = lineno = 0

    if not filename:
        filename = fh.name

    if filename.endswith("/SpelChk.py"):
        return 0

    if patch:
        before = list()
        after = list()

    for line in fh:
        lineno += 1

        if patch:
            before.append(line)
        rslt = list()
        match = pattern.search(line)
        while match:
            ret += 1
            m = match.group(1)
            (bad, good) = suggest(m, mapper[m.lower()])
            rslt.append((bad, good),)
            line = line.replace(m, good)
            match = pattern.search(line)

        if patch:
            after.append(line)
        else:
            for r in rslt:
                err(output, "'%s' should be '%s'" % r, filename, lineno)

    if ret > 0 and patch:
        print("".join(context_diff(before, after,
              fromfile="a/%s" % filename, tofile="b/%s" % filename)))

    return ret


# set up the word list
baddies = list()
mapper = dict()
for w in wordlist:
    for badword in w.split(","):
        baddies.append(badword)
        mapper[badword] = wordlist[w]
pattern = re.compile(r"(?i)(\b%s\b)" % r"\b|\b".join(baddies))
