#!/usr/bin/python

#
# Copyright (c) 2016, 2019, Oracle and/or its affiliates. All rights reserved.
#

#
# Check changeset tags:
#       - All tags required for current gate are specified in
#         the CRs for this changeset.
#
# This does not duplicate the work done in Comment.py.
#

import os
import sys
import traceback
import email.utils

from six.moves import configparser

from onbld.Checks.DbLookups import BugDB, defaultLookupFields
from onbld.Checks import get_user_email
from onbld.Checks.Comments import bugre


class TagRule(object):
    """Describe a single bug tag and the rules governing how that tag
    should be used."""

    def __init__(
            self, name, required=False, comp=None, subcomp=None,
            allowed_values=None):

        # The name of the tag. If the tag is of the form key=value, then
        # this just refers to the 'key' portion.
        self.name = name

        # Whether the tag must be set.
        self.required = required

        # The component this tag applies to. If empty, the tag can be applied
        # to any bug, assuming no self.subcomp restrictions.
        self.comp = comp

        # The subcomponent this tag applies to. If empty, the tag can be
        # applied to any bug, assuming no self.comp restrictions.
        self.subcomp = subcomp

        # The set of values this tag can take. If empty, there are no
        # restrictions on the value.
        self.allowed_values = allowed_values

    def __repr__(self):
        return \
            "{0}: comp: {1}, subcomp: {2}, required={3}, " \
            "allowed_values={4}".format(
                self.name, self.comp, self.subcomp, self.required,
                self.allowed_values)

    def validate(self, bugid, bug_tags, comp, subcomp):
        """Validate a bugid, checking that the tags on
        this bug meet the requirements described by this Tag object."""

        defined_tags = set()
        keyval_tags = {}
        for tag in bug_tags:
            tag_name = tag
            if isinstance(tag, list):
                keyval_tags[tag[0]] = tag[1]
                tag_name = tag[0]

            if tag_name in defined_tags:
                return False, ["Duplicate tag '%s' declared" % tag]

            defined_tags.add(tag_name)

        # this bug isn't interesting, return immediately
        if self.comp and comp != self.comp:
            return True, []

        if self.subcomp and subcomp != self.subcomp:
            return True, []

        failure_reasons = []

        def av_str(allowed_values):
            return " or ".join(
                sorted([", ".join(group) for group in sorted(allowed_values)]))

        def allowed(value, allowed_values):
            if not value:
                return False
            values = value.split(",")
            matched_group = 0
            for group in allowed_values:
                for value in values:
                    if value in group:
                        matched_group += 1
                        break
            if matched_group > 1:
                raise ValueError(
                    "tag value %s matches more than one group" % value)
            elif matched_group == 0:
                return False
            else:
                return True

        if self.name not in defined_tags and self.required:
            sub_reason = ""
            if self.comp and self.subcomp:
                sub_reason = " ('%s/%s' bugs require this tag)" % (
                    self.comp, self.subcomp)
            elif self.comp:
                sub_reason = " (component '%s' bugs require this tag)" % \
                    self.comp
            elif self.subcomp:
                sub_reason = " (subcomponent '%s' bugs require this tag)" % \
                    self.subcomp

            if self.allowed_values:
                failure_reasons.append(
                    "Tag '%s' must be set%s. Allowed values are %s" %
                    (self.name, sub_reason, av_str(self.allowed_values)))
            else:
                failure_reasons.append(
                    "Tag '%s' must be set %s" % (self.name, sub_reason))

        try:
            if self.allowed_values and self.name in defined_tags and \
                    not allowed(
                        keyval_tags.get(self.name), self.allowed_values):
                failure_reasons.append(
                    "Tag '%s' must be set to values: %s" %
                    (self.name, av_str(self.allowed_values)))
        except ValueError:
            failure_reasons.append(
                "Tag '%s' was set to a value that matches more than one "
                "group of the values: %s" %
                (self.name, av_str(self.allowed_values)))

        if failure_reasons:
            return False, failure_reasons
        return True, None


def read_tag_config(repo_root):
    """The tagchk file accepted by this code is a Python ConfigParser file
    format, that takes one or more sections, each section denoting a tag
    and the rules around which bugs must conform to those bug tags.

    We expect this file to be stored in exception_lists directory at the
    top of the gate, $WS/exception_lists/tagchk.

    Here are some examples of the possible contents of this file:

    [myproject]
    required = true

    Which says that all bugfixes being integrated into this gate must have
    the tag 'myproject'.

    We can specify tag requirements per-component or per-subcomponent:

    [frobbackport]
    component = utility
    required = true

    Which says that bugs in the utility component must have the tag
    'frobbackport'.

    We can specify comma-separated comp/subcomp values for a tag.
    [frobbackport]
    component = utility, webui/server
    required = true

    Which says that bugs in utility/* or webui/server must have the tag
    'frobbackport'

    To match a named subcomponent of all components, use the special value
    '*'. Note that normal globs or regexps are not supported.

    [frobbackport]
    component = */zones
    required = true

    Which says that bugs in eg. utility/zones, service/zones,
    documentation/zones, etc. must have the tag 'frobbackport'

    We support key=value bug tags, and can specify allowed values for a given,
    tag which must be comma-separated:

    [flavor]
    required = false
    allowed_values = vanilla, chocolate, lemon

    Which says that the bugs can have an optional tag 'flavor', but if set,
    the tag must be 'flavor=vanilla', 'flavor=chocolate' or 'flavor=lemon'.

    We allow grouping of allowed_values tags separated by '|' characters.

    A tag value can only contain values from a single matching group.

    For example, the entry:

    allowed_values = chocolate,rum,raisin|carrots,broccolli

    says that bugs can have any combination of the values in allowed_values,
    between two groups. So

    'flavor=chocolate', 'flavor=rum,raisin' and
    'flavor=chocolate,rum,raisin', 'flavor=broccoli'

    are all acceptable bug tags, but

    'flavor=chocolate,broccoli' is not acceptable.
    """
    ws = repo_root

    for suffix in ["/usr/closed", "/usr/man", "/usr/fish"]:
        if repo_root.endswith(suffix):
            ws = repo_root.replace(suffix, "")

    config_path = os.path.join(ws, "exception_lists", "bugtagchk")
    # no rules declared, so we're done.
    if not os.path.exists(config_path):
        return []

    tag_rules = []
    try:
        sc = configparser.SafeConfigParser()
        sc.read(config_path)

        for name in sc.sections():
            allowed_values = []
            required = False
            tag_name = name
            comp = None
            subcomp = None

            # allow users to specify comma-separated bug comp/subcomp lists
            comps = []

            if sc.has_option(name, "comp"):
                comp = sc.get(name, "comp").lower()
                if "," in comp:
                    comps = [c.strip() for c in comp.split(",")]
                else:
                    comps = [comp]

            if sc.has_option(name, "required"):
                try:
                    required = sc.getboolean(name, "required")
                except ValueError as e:
                    required = False
                    raise ValueError(
                        "ERROR: possible gate misconfiguration: "
                        "%s 'required' value." % config_path, e)

            if sc.has_option(name, "allowed_values"):
                allowed_value_groups = [value.strip() for value in sc.get(
                    name, "allowed_values").split("|")]
                for group in allowed_value_groups:
                    allowed_value_components = [
                        value.strip() for value in group.split(",")]
                    allowed_values.append(allowed_value_components)
                if "" in allowed_values:
                    raise ValueError(
                        "ERROR: possible null value specified in "
                        "'allowed_values' value for %s" % tag_name)

            if not comps:
                tag_rules.append(
                    TagRule(
                        tag_name, required, comp, subcomp, allowed_values))
                continue

            for item in comps:
                comp = None
                subcomp = None
                if "/" in item:
                    comp, subcomp = item.split("/", 1)
                    # A special glob to allow us to specify tags for
                    # a named subcomponent of all components.
                    if comp == "*":
                        comp = None
                else:
                    comp = item

                if not subcomp and "/" in item:
                    raise ValueError(
                        "ERROR: possible gate misconfiguration: "
                        "unable to parse 'comp/subcomp' pair %s" % item)
                tag_rules.append(
                    TagRule(
                        tag_name, required, comp, subcomp, allowed_values))
    except Exception as e:
        traceback.print_exc(e)
        raise ValueError("ERROR: possible gate misconfiguration: %s" % e)
    return tag_rules


def tagchk(
        comments, repo_root, check_db=True, output=sys.stderr,
        bugdb=None, bugdb_email=None, gatePath=None):
    """Validate CR tags against ON standards."""

    ret = 0

    if not check_db:
        return ret

    tag_rules = read_tag_config(repo_root)
    if not tag_rules:
        return ret

    bugs = {}

    for com in comments:
        # Our input must be newline-free, comments are line-wise.
        if com.find("\n") != -1:
            raise ValueError("newline in comment '%s'" % com)

        match = bugre.search(com)
        if match:
            if match.group(1) not in bugs:
                bugs[match.group(1)] = []
            bugs[match.group(1)].append(match.group(2))

    # Optimize for no bugs to check
    if len(bugs) == 0:
        return ret

    if not bugdb_email:
        bugdb_email = get_user_email()

    if not bugdb:
        # parseaddr() returns an empty string when passed None
        bugdb_email = email.utils.parseaddr(bugdb_email)[1]
        bugdb = BugDB(bugdb_email)

    results = bugdb.lookup(
        list(bugs.keys()),
        fields=defaultLookupFields)

    invalid_bugs = {}

    for crid, insts in list(bugs.items()):
        for entered in insts:
            bug_tags = results[crid]["tags"]
            synopsis = results[crid]["synopsis"]
            # bugdaemon uses old Bugster 'category/subcategory' terminology
            comp = results[crid]["category"].lower()
            subcomp = results[crid]["subcategory"].lower()

            # bugdaemon always reports a space-separated string of tags,
            # regardless of BugDB storage.
            tags_plain = bug_tags.split(" ")

            tags = []
            for tag in tags_plain:
                if "=" in tag:
                    tags.append(tag.split("=", 1))
                else:
                    tags.append(tag)

            for tag_rule in tag_rules:
                valid, reasons = \
                    tag_rule.validate(crid, tags, comp, subcomp)

                if not valid:
                    if crid in invalid_bugs:
                        existing_reasons = invalid_bugs[crid][1]
                        invalid_bugs[crid] = (
                            synopsis, existing_reasons + reasons)
                    else:
                        invalid_bugs[crid] = (synopsis, reasons)

    if invalid_bugs:
        ret = 1
        output.write("The following bugs have invalid tags:\n\n")

        for bug in invalid_bugs:
            synopsis, reasons = invalid_bugs[bug]
            output.write(
                "%s %s\n\t%s\n" % (bug, synopsis, "\n\t".join(reasons)))

    return ret
