#!/usr/bin/python

#
# Copyright (c) 2012, 2021, Oracle and/or its affiliates.
#

#
# OSNet-specific pkglint(1) checks, called as part of pkglint commands in
# usr/src/pkg/Makefile, with settings in usr/src/pkg/*pkglintrc.
#

import os
from six.moves import configparser

# pylint: disable=import-error
# This pylint disable is required whenever ON supplies modules for multiple
# versions of Python, but the pkg gate is only using one version at a time.
# This triggers a warning on the other versions, so we suppress it.
#
import pkg.lint.base as base
# pylint: enable=import-error


class OSNetActionChecker(base.ActionChecker):
    """An osnet-specific class to check actions."""

    name = "osnet.action"

    def __init__(self, config):
        self.description = _(
            "checks OS Net packages for common action errors")
        self.exception_lists = {}
        super(OSNetActionChecker, self).__init__(config)

        # Slurp in the pkglint exception list
        ws = os.getenv("WS")
        if ws is not None:
            exceptfile = configparser.RawConfigParser()
            exceptfile.read("{}/exception_lists/pkglint".format(ws))
            sections = exceptfile.sections()

            for sect in sections:
                filelist = []
                for key in sorted(exceptfile.options(sect)):
                    if key.startswith("path."):
                        filelist.append(exceptfile.get(sect, key))

                self.exception_lists[sect] = filelist

    def varmigrate(self, action, manifest, engine, pkglint_id="001"):
        """Check that we only deliver directories to /var/.migrate.

        During filesystem/minimal, a script runs to migrate unpackaged
        content from /var/.migrate to the newly mounted /var/share
        dataset.

        Directories are created if necessary under /var/share, but files
        are moved.  If packaged objects were delivered under
        /var/.migrate then subsequently migrated, they would cause
        'pkg verify/fix' errors in the image.

        Ensure /var/.migrate entries that have a salvage-from attribute
        also have an actuator (needed for self-assembly of ROZR zones
        in particular)
        """

        if "path" not in action.attrs:
            return
        path = action.attrs["path"]

        if not path.startswith("var/.migrate"):
            return

        seen_preserve = action.attrs.get("preserve")

        if action.name != "dir" and (action.name != "file" or
           not seen_preserve or (seen_preserve.lower() != "true" and
                                 seen_preserve.lower().lower() != "abandon")):
            engine.error(_(
                "Bad deliverable: %(path)s. "
                "Only directories and files with the preserve attribute can"
                " be delivered to /var/.migrate by %(pkg)s") %
                {"path": path, "pkg": manifest.fmri},
                msgid="%s%s.1" % (self.name, pkglint_id))

        seen_salvage_from = action.attrs.get("salvage-from")
        seen_actuator = action.attrs.get("refresh_fmri")

        if seen_salvage_from and not seen_actuator:
            engine.error(_(
                "/var/.migrate action %(path)s must "
                "have a refresh_fmri actuator when 'salvage-from'"
                " is set in %(pkg)s") %
                {"path": path, "pkg": manifest.fmri},
                msgid="%s%s.2" % (self.name, pkglint_id))

    varmigrate.pkglint_desc = _(
        "Only directories should be delivered to /var/.migrate")

    def varshare(self, action, manifest, engine, pkglint_id="002"):
        """Ensure that we deliver nothing underneath /var/share.

        During upgrade, from a system without an <rpool>/VARSHARE
        dataset, we create that dataset and mount it there - a populated
        mountpoint would foil our plans.

        With a mounted /var/share, since the dataset is shared across
        boot environments, we can't risk packaging operations from two
        separate boot environments trying to modify content in
        /var/share.
        """

        if "path" not in action.attrs:
            return
        path = action.attrs["path"]

        if path.rstrip("/").startswith("var/share/"):
            engine.error(_(
                "Path %(path)s cannot be delivered by "
                "%(pkg)s, reserved for unpackaged content shared "
                "across boot environments.") %
                {"path": path, "pkg": manifest.fmri},
                msgid="%s%s" % (self.name, pkglint_id))

    varshare.pkglint_desc = _(
        "Packages should not deliver content beneath /var/share.")

    def fileattr(self, action, manifest, engine, pkglint_id="003"):
        """Check that file permissions are set correctly. This checks against
        a whitelist of known files which are shipped differently, but the
        default should be root:root, with no setuid bits.

        Any deviations from this must be listed in the whitelist.
        Any whitelist entries which are not currently violations are errors.

        This check is to help us move towards a root:root ownership"""

        if "path" not in action.attrs:
            return
        path = action.attrs["path"]

        # Only check files here
        if action.name != "file":
            return

        # All should be owned by root
        owner = action.attrs.get("owner")
        if owner is not "root" and path not in self.exception_lists["owner"]:
            engine.error(_(
                "Path %(path)s is not owned by root") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))
        if owner is "root" and path in self.exception_lists["owner"]:
            engine.error(_(
                "Owner exception for %(path)s is redundant") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))

        # All files should have a group of 'root'
        group = action.attrs.get("group")
        if group is not "root" and path not in self.exception_lists["group"]:
            engine.error(_(
                "Path %(path)s group is not root") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))
        if group is "root" and path in self.exception_lists["group"]:
            engine.error(_(
                "Group exception for %(path)s is redundant") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))

        # Check for setuid
        mode = int(action.attrs.get("mode"), 8)
        if mode & 0o4000 and path not in self.exception_lists["setuid"]:
            engine.error(_(
                "Path %(path)s is setuid") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))
        if (mode & 0o4000 == 0) and path in self.exception_lists["setuid"]:
            engine.error(_(
                "Setuid exception for %(path)s is redundant") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))

        # Check for setgid
        mode = int(action.attrs.get("mode"), 8)
        if mode & 0o2000 and path not in self.exception_lists["setgid"]:
            engine.error(_(
                "Path %(path)s is setgid") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))
        if (mode & 0o2000 == 0) and path in self.exception_lists["setgid"]:
            engine.error(_(
                "Setgid exception for %(path)s is redundant") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))

        # Check for group or world writable
        # We'd like to include owner-writable too, but cannot currently as
        # many files are delivered like that.
        mode = int(action.attrs.get("mode"), 8)
        if mode & 0o22 and path not in self.exception_lists["writable"]:
            engine.error(_(
                "Path %(path)s is writable") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))
        if (mode & 0o22 == 0) and path in self.exception_lists["writable"]:
            engine.error(_(
                "Writable exception for %(path)s is redundant") %
                {"path": path}, msgid="%s%s" % (self.name, pkglint_id))

    fileattr.pkglint_desc = _(
        "Files should be root:root and not writable")

    def dirattr(self, action, manifest, engine, pkglint_id="004"):
        """Check that directory permissions are set appropriately. This checks
        for any group or world writable bits, and any of the special "setid"
        bits being set accidentally. This is checked against a whitelist of
        known directories which are shipped differently.

        Note that for historical reasons, it is almost impossible to change the
        owner/group for directories, since it will break the installation of
        non-ON packages that specify existing directories (such as /usr) with
        a group=sys (the current value)."""

        if "path" not in action.attrs:
            return
        path = action.attrs["path"]

        # Only check directories here
        if action.name != "dir":
            return

        # Check for any special bits or group/world write bits.
        modestr = action.attrs.get("mode")
        mode = int(modestr, 8)
        if mode & 0o7022 and path not in self.exception_lists["dirmode"]:
            engine.error(_(
                "Dir %(path)s is set to %(mode)s") %
                {"path": path, "mode": modestr},
                msgid="%s%s" % (self.name, pkglint_id))
        if (mode & 0o7022 == 0) and path in self.exception_lists["dirmode"]:
            engine.error(_(
                "Dirmode exception for %(path)s is redundant") %
                {"path": path},
                msgid="%s%s" % (self.name, pkglint_id))

    dirattr.pkglint_desc = _(
        "Directories should not be delivered group or world writable")

    def emptyfiles(self, action, manifest, engine, pkglint_id="005"):
        """Verify that only a known set of empty files are shipped.
        This is to stop additional files creeping in unnoticed."""

        if "path" not in action.attrs:
            return
        path = action.attrs["path"]

        # Only check files here
        if action.name != "file":
            return

        # Since pkglint is first run over the resolved manifests, we may not
        # know the size of the file at that point. So we skip here if there
        # is no size available. We will, however, catch these again when run
        # during the lint task.

        filesizestr = action.attrs.get("pkg.size")
        if filesizestr is None:
            return
        filesize = int(filesizestr)
        if filesize == 0 and path not in self.exception_lists["empty"]:
            engine.error(_(
                "File %(path)s is empty") % {"path": path},
                msgid="%s%s" % (self.name, pkglint_id))
        if filesize != 0 and path in self.exception_lists["empty"]:
            engine.error(_(
                "Empty exception for %(path)s is redundant") % {"path": path},
                msgid="%s%s" % (self.name, pkglint_id))

    emptyfiles.pkglint_desc = _(
        "Empty files should not be packaged and shipped")

    def attrfiles(self, action, manifest, engine, pkglint_id="006"):
        """Verify that any files delivered to the *_attr.d directories
        in /etc/ are named after the package, with '/' transformed to ':'"""

        if "path" not in action.attrs:
            return

        # Only check files here
        if action.name != "file":
            return

        path = action.attrs["path"]
        dirlist = [
            "etc/security/auth_attr.d",
            "etc/security/exec_attr.d",
            "etc/security/prof_attr.d",
            "etc/user_attr.d"
        ]

        dir = os.path.dirname(path)
        if dir not in dirlist:
            return

        name = os.path.basename(path)
        transformed_fmri = manifest.fmri.get_pkg_stem(
            anarchy=True, include_scheme=False).replace('/', ':')

        # This could be a strict equality test, but to allow for cases where
        # the same package delivers different fragments yet makes sense to
        # maintain the separation (perhaps being conceptually separate, or
        # delivered from different source directories), we perform a
        # startswith test instead. Thus the filename must start with the
        # package name, but can include a suffix.
        if not name.startswith(transformed_fmri):
            engine.error(_(
                "Security attr file %(path)s does not match delivering "
                "package name - should be %(pkg)s") %
                {"path": path, "pkg": transformed_fmri},
                msgid="%s%s" % (self.name, pkglint_id))

    attrfiles.pkglint_desc = _(
        "Security attr.d files should be named after the package that "
        "delivers them.")

    def restricted_dirs(self, action, manifest, engine, pkglint_id="007"):
        """Certain directories must not have content delivered to them.
        Some are reserved for customer use, some are deprecated, and some
        are just dangerous to deliver to.
        This ensures content is not delivered there."""

        if "path" not in action.attrs:
            return

        path = action.attrs["path"]

        # List of directories. Note the trailing slash should be included in
        # the path name. The second column is whether directories are allowed
        # to be delivered under the specified path, and the third is a reason
        # why content should not be delivered there.
        dirlist = [
            ("etc/driver/drv/", True, "customer-reserved"),
            ("etc/init.d/", True, "legacy"),
            ("etc/rc0.d/", True, "legacy"),
            ("etc/rc1.d/", True, "legacy"),
            ("etc/rc2.d/", True, "legacy"),
            ("etc/rc3.d/", True, "legacy"),
            ("etc/rcS.d/", True, "legacy"),
            ("etc/svc/profile/", True, "customer-reserved"),
            ("system/contract/", False, "volatile"),
            ("system/object/", False, "volatile"),
            ("system/volatile/", False, "volatile"),
            ("system/", True, "reserved"),
            ("tmp/", False, "volatile"),
            ("var/share/", False, "non-packagable"),
            ("var/svc/manifest/", True, "legacy"),
            ("var/tmp/", False, "volatile")
        ]

        for (dir, dirs_allowed, reason) in dirlist:
            if (action.name != "dir" or dirs_allowed is False) and \
                    path.startswith(dir) and \
                    path not in self.exception_lists["restricted_dirs"]:
                engine.error(_(
                    "File %(path)s delivered to a %(reason)s directory.") %
                    {"path": path, "reason": reason},
                    msgid="%s%s" % (self.name, pkglint_id))

    restricted_dirs.pkglint_desc = _(
        "Content must not be delivered to restricted directories.")


class OSNetManifestChecker(base.ManifestChecker):
    """An osnet-specific class to check actions."""

    name = "osnet.manifest"

    def test_cfg(self, manifest, engine, pkglint_id="001"):
        """If this package delivers files to /opt/onbld/test then
        it should also deliver at least one file called <foo>.cfg to
        /opt/onbld/test/test.d.

        By delivering a *.cfg file along with packaged tests, we make it
        easier for test teams to execute those tests without having to
        consult the source tree to determine how to integrate those
        tests with their test harness.

        It's hard to police tests that are delivered to places other
        than /opt/onbld/test and which also have a *.cfg file in
        $SRC/tools/test.d but have forgotten to package it.
        """

        has_onbld_tests = False
        has_cfg_file = False
        for action in manifest.gen_actions():
            if "path" not in action.attrs:
                continue
            if action.name != "file":
                return
            path = action.attrs["path"]
            if path.startswith("opt/onbld/test/"):
                has_onbld_tests = True
            else:
                continue

            if path.startswith("opt/onbld/test/test.d/") and \
                    os.path.basename(path).endswith(".cfg"):
                has_cfg_file = True

        if has_onbld_tests and not has_cfg_file:
            engine.error(_(
                "Package %(pkg)s delivers to opt/onbld/test but does not "
                "include a *.cfg file") %
                {"pkg": manifest.fmri},
                msgid="%s%s" % (self.name, pkglint_id))

    test_cfg.pkglint_desc = _(
        "ON Test packages must deliver *.cfg files.")

    def test_legacy(self, manifest, engine, pkglint_id="006"):
        """If a package has the pkg.legacy metadata in it then it
        must have an associated release_note in the manifest. As
        per PSARC/2018/081 and PSARC/2020/053.
        """

        if "pkg.legacy" in manifest and \
                manifest["pkg.legacy"].lower() == "true":

            relnote = False
            for a in manifest.gen_actions():
                if 'release-note' in a.attrs:
                    relnote = True
                    break

            if not relnote:
                engine.error(_(
                    "Legacy package %(pkg)s is missing a release note") %
                    {"pkg": manifest.fmri},
                    msgid="%s%s" % (self.name, pkglint_id))

    test_legacy.pkglint_desc = _(
        "Legacy packages must include a release note as per PSARC/2018/018.")


class ExtractLicense(base.ManifestChecker):

    """A class to extract license(s) from manifests. This is a more strict
    version of the licensing checks used by Solaris Release Engineering,
    which emits errors rather than warnings """

    name = "re.pkglint.license"

    def __init__(self, config):
        self.description = _("Extract license from manifest.")
        super(ExtractLicense, self).__init__(config)
        self.licenses = {}

    @staticmethod
    def _construct_license(ldict):
        license = {}
        for key in ["tpno", "name", "version", "description"]:
            license[key] = ldict.get("com.oracle.info.%s" % key, "")
        return license

    @staticmethod
    def _construct_set_attrs(manifest):
        result = {}
        for action in manifest.gen_actions_by_type("set"):
            map = {}
            name = ""
            for attr in action.attrs:
                if attr == "name":
                    name = action.attrs[attr]
                else:
                    val = action.attrs[attr]
                    map[attr] = val
            result[name] = map
        return result

    @staticmethod
    def _construct_notempty_attrs(manifest):
        result = {}
        for action in manifest.gen_actions():
            if action.name == "file":
                result = "hascontent"
                break
        return result

    def licensing(self, manifest, engine, pkglint_id="001"):
        licensespkg = []
        licenseslic = []
        licensepkg = {}
        licenselic = {}

        # old method of defining license info, using "set" actions
        pkgvers = ""
        if "com.oracle.info.tpno" in manifest:
            licensepkg = ExtractLicense._construct_license(manifest)
            fileaction = ExtractLicense._construct_notempty_attrs(manifest)
            licensespkg.append(licensepkg)
            licenseslic.append(licensepkg)
        else:
            if "com.oracle.info.version" in manifest:
                try:
                    pkgvers = ExtractLicense._construct_license(
                        manifest)["version"]
                except ValueError:
                    pkgvers = ""

        # extract license action attributes
        for action in manifest.gen_actions_by_type("license"):
            licenselic = ExtractLicense._construct_license(action.attrs)
            licenselic["license"] = action.attrs.get("license", "")
            licenseslic.append(licenselic)

        # manifest cannot use both set actions and license actions to define
        # tpno attribute.
        if not licensepkg and not licenselic:
            return
        pkgpath = manifest.fmri.get_pkg_stem()

        pkglint_id = "001"
        if licensepkg:
            if licensepkg["tpno"] != "" and not licenselic:
                if fileaction == "hascontent":
                    lint_id = "%s%s" % (self.name, pkglint_id)
                    engine.error(
                        _("%(fmri)s has tpno but no license action")
                        % {"fmri": pkgpath},
                        msgid=lint_id)
                    return
                return

        pkglint_id = "002"
        tpnostr = licenselic["tpno"]
        if licensespkg and licenseslic and tpnostr:
            lint_id = "%s%s" % (self.name, pkglint_id)
            engine.error(_(
                "%(fmri)s has both license and set actions for "
                "tpno") % {"fmri": pkgpath}, msgid=lint_id)

        for license in [licensepkg, licenselic]:

            if not license:
                continue
            # validation of tpno format
            pkglint_id = "003"
            tpnostr = license["tpno"]
            if tpnostr != "":
                try:
                    # force cast it to integer to compare
                    tpnoval = int(tpnostr)
                except ValueError:
                    lint_id = "%s%s.1" % (self.name, pkglint_id)
                    engine.error(_(
                        "%(pkg)s has non-integer TPNO "
                        "value of %(key)s") %
                        {"pkg": pkgpath, "key": tpnostr},
                        msgid=lint_id)

            # com.oracle.info.name, com.oracle.info.description cannot be
            # empty.
            # com.oracle.info.version may be empty if there is no version for
            # the technology
            pkglint_id = "004"
            for infofield in ["tpno", "name", "description", "version"]:
                infoval = license[infofield]
                if infoval == "":

                    if infofield == "tpno":
                        return

                    if infofield == "name":
                        lint_id = "%s%s.2" % (self.name, pkglint_id)
                        engine.error(_(
                            "%s has empty value for license attribute "
                            "com.oracle.info.%s") % (pkgpath, infofield),
                            msgid=lint_id)

                    if infofield == "description":
                        lint_id = "%s%s.3" % (self.name, pkglint_id)
                        engine.error(_(
                            "%s has empty value for license attribute "
                            "com.oracle.info.%s") %
                            (pkgpath, infofield),
                            msgid=lint_id)

                    if infofield == "version":
                        if pkgvers == "":
                            lint_id = "%s%s.4" % (self.name, pkglint_id)
                            engine.warning(_(
                                "%s has empty value for license attribute "
                                "com.oracle.info.%s") % (pkgpath, infofield),
                                msgid=lint_id)

    licensing.pkglint_desc = _("A series of license checks for Solaris RE")
