#!/usr/bin/python2.7

#
# Copyright (c) 2015, 2021, Oracle and/or its affiliates.
#

import atexit
import email.utils
import logging
import lxml.etree as etree
import os
import shutil
import StringIO
import sys
import tempfile
import time
import unittest
import urllib2

from datetime import datetime, timedelta

if not os.environ.get("SRC"):
    raise RuntimeError("$SRC must be set to run this script")
if not os.environ.get("TOOLROOT"):
    raise RuntimeError("$TOOLROOT must be set to run this script")
if os.environ.get("SRC"):
    sys.path.append(os.path.join(os.environ["SRC"], "tools"))
else:
    sys.path.append("../..")

import onbld.Checks.Cddl
import onbld.Checks.CheckMakefile
import onbld.Checks.CmtBlk
import onbld.Checks.Comments
import onbld.Checks.Copyright
import onbld.Checks.CStyle
import onbld.Checks.DbLookups
import onbld.Checks.HdrChk
import onbld.Checks.JSStyle
import onbld.Checks.JStyle
import onbld.Checks.Keywords
import onbld.Checks.Mapfile
import onbld.Checks.Notify
import onbld.Checks.Rti
import onbld.Checks.SpelChk
import onbld.Checks.Tags
import onbld.Checks.UTF8

from onbld.Checks import get_user_email
from onbld.util import cmdline_run


class CheckTest(unittest.TestCase):

    logger = logging.getLogger("onbld")
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.handlers[0].formatter = logging.Formatter(
        "%(levelname)-7s : %(message)s")
    testdir = tempfile.mkdtemp("", "checktest.")

    # Paths to mercurial using the built copy of the Cadmium extension
    CDM_PATH = os.path.join(
        os.environ.get("TOOLROOT", ""),
        "opt/onbld/lib/python/onbld/hgext/cdm.py")
    hg_cdm = "/usr/bin/hg --config 'extensions.hgext.cdm={0}'".format(CDM_PATH)

    # Cache an onbld.Checks.DbLookups.BugDB object
    __bugdb = None

    def setUp(self):

        if "ONBLD_DEBUG" in os.environ:
            self.debug = True
            CheckTest.logger.handlers[0].setLevel(logging.DEBUG)
            # Add a CR so that when running with -v, we get test descriptions
            # and the first line of debug output separated properly.
            sys.stderr.write("\n%s:\n" % self._testMethodName)
        else:
            self.debug = False

        # Create enough of a dummy workspace so that hg nits and hg pbchk
        # can do useful work later.
        self.ws = "{0}/workspace".format(self.testdir)

        if not os.path.exists(self.ws):
            parent = "{0}/on12-gate".format(self.testdir)
            os.mkdir(self.ws)
            os.mkdir(parent)
            cmdline_run(
                "{0} init".format(self.hg_cdm), cwd=parent, logger=self.logger)

            def _add_commit(ws):
                shutil.copy(os.path.join(os.environ["SRC"],
                            "prototypes", "prototype.c"), ws)
                with open("{0}/prototype.c".format(ws), "a") as temp:
                    temp.write(str(datetime.now()))
                # we need a properly formatted commit message, this will do.
                msg = \
                    "22557519 hg pbchk needs the ability to validate bug tags"
                cmdline_run(
                    "{0} add".format(self.hg_cdm), cwd=ws, logger=self.logger)
                cmdline_run("{0} commit -m '{1}'".format(
                    self.hg_cdm, msg), cwd=ws, logger=self.logger)

            _add_commit(parent)
            cmdline_run("{0} clone {1} {2}".format(
                self.hg_cdm, parent, self.ws), logger=self.logger)
            _add_commit(self.ws)

        atexit.register(shutil.rmtree, self.testdir, ignore_errors=True)

    def tearDown(self):
        pass

    def test_cstyle(self):
        """Verify that cstyle is being called correctly. Note that this isn't
        yet a comprehensive test of cstyle itself, rather a check that we can
        call cstyle correctly and deal with its return code correctly."""

        bad_c = """
/* Note lack of space between closing parenthesis and open brace for main */
void main(int argc, char* argv){
    printf("Hello world\n");
}
"""
        good_c = """
void main(int argc, char *argv) {
    printf("Hello world\n");
}
"""

        def run_cstyle(c_source):
            """Write a temporary java file, then run jsstyle on it."""
            output = StringIO.StringIO()
            fd, fname = tempfile.mkstemp(".c", "")
            with open(fname, "w") as cfile:
                cfile.write(c_source)
            with open(fname, "r") as cfile:
                ret = onbld.Checks.CStyle.cstyle(
                    cfile, output=output, )
                self.logger.debug("%s: %s" % (ret, output.getvalue()))
            os.remove(fname)
            return ret, output.getvalue()

        ret, out = run_cstyle(good_c)
        self.assert_(ret == 0, "cstyle false positive on valid c file")
        self.assert_(out == "", "unexpected output from cstyle: %s" % out)

        ret, out = run_cstyle(bad_c)
        self.assert_(ret != 0, "cstyle did not detect invalid c file")
        self.assert_(out != "", "expected some output from cstyle!")

    def test_cddl(self):
        """Verify that the check for the presence/absence of the CDDL works."""
        CDDL = """
CDDL HEADER START

The contents of this file are subject to the terms of the
Common Development and Distribution License (the "License").
You may not use this file except in compliance with the License.

You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
or http://www.opensolaris.org/os/licensing.
See the License for the specific language governing permissions
and limitations under the License.

When distributing Covered Code, include this CDDL HEADER in each
file and include the License file at usr/src/OPENSOLARIS.LICENSE.
If applicable, add the following below this CDDL HEADER, with the
fields enclosed by brackets "[]" replaced with your own identifying
information: Portions Copyright [yyyy] [name of copyright owner]

CDDL HEADER END
"""

        def cddl_(contents, expected_error, require_cddl=False):

            fd, fname = tempfile.mkstemp(".cddl", "")
            with open(fname, "w") as mapfile:
                mapfile.write(contents)

            ret = 0
            with open(fname, "r") as mapfile:
                output = StringIO.StringIO()
                ret = onbld.Checks.Cddl.cddlchk(
                    mapfile, filename="cddl.txt", output=output,
                    require_cddl=require_cddl)

            os.remove(fname)
            if expected_error:
                if isinstance(expected_error, list):
                    for error in expected_error:
                        self.assert_(error in output.getvalue())
                else:
                    self.assert_(expected_error in output.getvalue())
                self.assert_(ret > 0)
            else:
                self.assert_(
                    ret == 0, "CDDL check of %s unexpectedly failed:\n%s" %
                    (contents, output.getvalue()))

        cddl_(CDDL, "Error: CDDL comment should not be present")
        cddl_(CDDL, None, require_cddl=True)
        cddl_("there is no cddl here", None)
        cddl_(
            CDDL.replace("Covered Code", "chickens"),
            [
                "Error: Invalid line in CDDL block",
                "When distributing chickens, include this CDDL HEADER in each",
                "When distributing Covered Code, include this CDDL HEADER "
                "in each"
            ], require_cddl=True)

    def test_checkmakefile(self):
        """Check that we're spotting obvious Makefile errors."""

        def makefilechk_(
                contents, expected_error, check_error=True, all_makefiles=[]):
            """Check the given Makefile contents, ensuring that we return
            with a non-zero exit code, and that the expected error(s) are
            included in the output. If check_error is False, we verify we
            get a 0 exit code. The cadmium hooks pass all known Makefiles
            to the check, but for our purposes only passing one or two known
            Makefile paths in 'all_makefiles' is enough."""

            fd, fname = tempfile.mkstemp(".mk", "")
            with open(fname, "w") as makefile:
                makefile.write(contents + "\n")

            ret = 0
            with open(fname, "r") as makefile:
                output = StringIO.StringIO()
                self.logger.debug("input: %s" % contents)
                ret = onbld.Checks.CheckMakefile.makefilechk(
                    makefile, "Makefile", os.environ["SRC"] + "/../..",
                    output=output, allmakefiles=all_makefiles)
                self.logger.debug("%s: %s" % (ret, output.getvalue()))

            os.remove(fname)
            if expected_error:
                if isinstance(expected_error, list):
                    for error in expected_error:
                        self.assert_(error in output.getvalue())
                else:
                    self.assert_(expected_error in output.getvalue())
                if check_error:
                    self.assert_(ret > 0)
                else:
                    self.assert_(ret == 0)
            else:
                self.assert_(
                    ret == 0, "makefilechk of %s unexpectedly failed:\n%s" %
                    (contents, output.getvalue()))

        # Cadmium only ever runs CheckMakefile on files which contain
        # "Makefile" in the name. It also assumes "sane" make(1s) syntax,
        # so we don't fuzz-test for now.
        # This only warns about include files that don't appear to be Makefiles
        makefilechk_(
            "include $(SRC)/uts/README", "which is not a Makefile",
            check_error=False)
        makefilechk_("include /etc/Makefile", "not found in this repo")
        makefilechk_(".KEEP_STATE:", "Old state-file syntax found")
        makefilechk_(
            "$(NOODLES)include $(SRC)/Makefile.master",
            "unknown prefix variable ($(NOODLES)) found",
            all_makefiles="%s/Makefile.master" % os.environ["SRC"])
        makefilechk_(
            ".KEEP_STATE_FILE: $(BUILDDIR)/.make.state.cheese",
            [
                "invalid state-file syntax found",
                "state-file without prior include statement"],
        )
        makefilechk_(
            ".KEEP_STATE_FILE: $(BUILDDIR)/.make.state.$(MACH)",
            "state-file without prior include statement",
            check_error=False,
            all_makefiles="%s/cmd/Makefile.cmd" % os.environ["SRC"])
        makefilechk_(
            "include $(SRC_REL)/Makefile.master",
            "$(SRC_REL) found. Use $(SRC) instead")
        makefilechk_(
            "include $(SRC)/Makefile.master ",
            "statement has trailing whitespace")

        # The $@ target as a dependency should be referred to by $$@ to avoid
        # problems with rebuilding targets. Check that detection of that is
        # done correctly. First, some valid cases which should generate no
        # errors:
        makefilechk_(
            "foo: bar", "", check_error=False)
        makefilechk_(
            "foo: $bar", "", check_error=False)
        makefilechk_(
            "foo: $(bar)", "", check_error=False)
        makefilechk_(
            "$(foo): $(bar)", "", check_error=False)
        makefilechk_(
            "foo: $$@", "", check_error=False)
        makefilechk_(
            "$(foo): $$@", "", check_error=False)
        makefilechk_(
            "$(foo): $$(@:a=b)", "", check_error=False)

        # Now some cases we should detect:
        makefilechk_(
            "foo: $@", "'$@' detected")
        makefilechk_(
            "foo: $(@)", "'$@' detected")
        makefilechk_(
            "foo: $(@:a=b)", "'$@' detected")
        makefilechk_(
            "foo:$(@)", "'$@' detected")
        makefilechk_(
            "$(foo): $(@)", "'$@' detected")

    def _get_bugdb(self):
        """Cache a BugDB object."""

        if self.__bugdb:
            return self.__bugdb
        # Use a common BugDB object to avail of cached responses
        bugdb_email = email.utils.parseaddr(get_user_email())[1]
        bugdb = onbld.Checks.DbLookups.BugDB(bugdb_email)
        self.__bugdb = bugdb
        return bugdb

    def _comchk(self, comment, expected_output, expect_fail=True):
        """Given a list of comments and the expected output, verify that
        we get that expected output from comchk.
        """
        output = StringIO.StringIO()
        self.logger.debug(comment)
        bugdb = self._get_bugdb()
        ret = onbld.Checks.Comments.comchk(
            comment, output=output, gatePath="/ws/on12-gate", bugdb=bugdb)
        self.logger.debug("%s %s" % (ret, output.getvalue()))

        if isinstance(expected_output, basestring):
            self.assert_(
                expected_output in output.getvalue(),
                "expected output not found.\noutput:\n"
                "---\n"
                "%s\n"
                "expected output\n---\n%s\n" %
                (output.getvalue(), expected_output))
        else:
            for expected in expected_output:
                self.assert_(expected in output.getvalue())
        if expect_fail:
            self.assert_(ret > 0)
        else:
            self.assert_(ret == 0)
        return output.getvalue()

    def test_valid_comments(self):
        """A series of perfectly valid comments."""
        comments = (
            ["18179156 comchk should use mysql for ARC data"],
            [
                "18179156 comchk should use mysql for ARC data",
                "15944052 rtichk should catch bug ID mismatch in WebRTI"],
            # This comment is notable, because the putback got the
            # arc case wrong. This is the corrected version.
            [
                "PSARC/2014/423 Kstats v2 Kernel API/DDI",
                "PSARC/2015/047 Kstat V2 Kernel API/DDI improvements",
                "19569805 Need kernel support for kstats v2",
                "20449840 libkstat2 map_check_meta and "
                "mapnv_check_meta have a leak"
            ])
        for commit_message in comments:
            output = StringIO.StringIO()
            ret = onbld.Checks.Comments.comchk(
                commit_message, output=output,
                gatePath="/ws/on12-gate")
            self.logger.debug(output.getvalue())
            self.assert_(output.getvalue() == "")
            self.assert_(ret == 0)

    def test_comments_no_bugid(self):
        """comments with no bugids"""
        comments = ["this comment no id"]
        error = "These comments are neither bug nor ARC case"
        self._comchk(comments, error)

        comments = ["PSARC 2016/453 nfssnoop utility"]
        error = "Comments must include at least one bugid"
        self._comchk(comments, error)

    def test_comments_mixed_prodid(self):
        """comments with mixed productids"""
        # AK and Solaris can be pushed together
        comments = [
            "16694622 Appliance head hung on boot",
            "15380123 SUNBT6527352 cstyle can't cope with #pragma align"]
        self._comchk(comments, [], expect_fail=False)

        # AK and STC cannot be pushed together
        comments = [
            "16694622 Appliance head hung on boot",
            "26837191 Add support for REST API login tokens"]
        error = (
            "All the CRs in your changeset must have the same "
            "product ID assignment.")
        self._comchk(comments, error)

    def test_comments_wrong_synopsis(self):
        """comments with ARC cases with a bad arc/bug synopsis"""

        comment_error_pairs = (
            (
                ["15380123 life is a tale, told by an idiot"],
                "These bugs/ARC case synopsis/names don't match the "
                "database entries"),
            (
                ["PSARC/2014/344 EOF yabbadaba"],
                "These bugs/ARC case synopsis/names don't match the "
                "database entries")
        )
        for comments, error in comment_error_pairs:
            self._comchk(comments, error)

    def test_comments_spaced_bugid(self):
        """comments with no space between bugid and subject"""

        comments = ["16589750onu must die"]
        error = "These bugs are missing a single space following the ID"
        self._comchk(comments, error)

    def test_comments_extra_spaces(self):
        """comments with extra or trailing spaces in the synopsis."""

        error = "These bugs have leading or trailing spaces in the synopsis"
        comments = (
            ["16589750  onu must die"],
            ["16589750 \tonu must die"],
            ["16589750 \t onu must die"],
            ["16589750 onu must die "],
        )
        for comment in comments:
            self._comchk(comment, error)

        # when there's no single space after the bugid, we'll get both errors
        error2 = "These bugs are missing a single space following the ID"
        comments = ["16589750\tonu must die "]
        self._comchk(comments, [error, error2])

    def test_comments_dups(self):
        """comments with duplicate ids"""

        comments = [
            "16589750 onu must die",
            "18538936 BUILDTOP ignores bldrc",
            "16589750 onu must die"]
        error = "These IDs appear more than once in your comments"
        self._comchk(comments, error)

    def test_comments_blanks(self):
        """comments including blank lines"""

        comments = [
            "16589750 onu must die",
            "18538936 BUILDTOP ignores bldrc",
            "\n"]
        error = "newline in comment"
        output = StringIO.StringIO()
        raised = False
        try:
            ret = onbld.Checks.Comments.comchk(
                comments,
                check_db=False, output=output, gatePath="/ws/on12-gate")
        except ValueError as e:
            self.assert_(error in str(e))
            raised = True
        self.assert_(raised, "no exception raised")

        comments = (
            ["", "18538936 BUILDTOP ignores bldrc"],
            [" ", "18538936 BUILDTOP ignores bldrc"])
        error = "WARNING: Blank line(s) in comments"
        for comment in comments:
            self._comchk(comment, error)

    def test_comments_arc(self):
        """comments that need ARC/RTI cross checks."""

        # no ARC case: this is OK
        comments = [
            "19569805 Need kernel support for kstats v2",
            "20449840 libkstat2 map_check_meta and "
            "mapnv_check_meta have a leak"]
        self._comchk(comments, [], expect_fail=False)

        # followup push, no ARC cases: this is OK
        comments = [
            "19569805 Need kernel support for kstats v2 (fix lint)",
        ]
        self._comchk(comments, [], expect_fail=False)

        # no ARC cases, and a followup push, but too much detail.
        comments = [
            "19569805 Need kernel support for kstats v2 (fix lint)",
            "20449840 libkstat2 map_check_meta and "
            "mapnv_check_meta have a leak"
        ]
        errors = [
            "Follow up pushes should only include a single comment line "
            "with the most relevant bug, and should omit ARC cases"
        ]
        self._comchk(comments, errors)

        # A wrong ARC case
        comments = [
            "PSARC/2012/333 Authenticated Rights Profiles",
            "19569805 Need kernel support for kstats v2"]
        error = """The ARC cases in the comments do not match those in the RTI.
ARC cases in the RTI:
    PSARC 2014/423
    PSARC 2015/047

ARC cases in the comments:
    PSARC 2012/333
"""
        self._comchk(comments, error)

        # An extra ARC case
        comments = [
            "PSARC/2012/333 Authenticated Rights Profiles",
            "PSARC/2015/047 Kstat V2 Kernel API/DDI improvements",
            "19569805 Need kernel support for kstats v2",
            "20449840 libkstat2 map_check_meta and "
            "mapnv_check_meta have a leak"]
        error = """The ARC cases in the comments do not match those in the RTI.
ARC cases in the RTI:
    PSARC 2014/423
    PSARC 2015/047

ARC cases in the comments:
    PSARC 2012/333
    PSARC 2015/047
"""
        self._comchk(comments, error)

    def test_comments_non_solaris_bugid(self):
        """comments including one non-Solaris (unsupported) bugid"""
        comments = [
            "15055554 SUNBT4423105 lint should report missing/misspelled "
            "switch default lables"]
        error = "Your changeset contains CRs with an unsupported product id"
        self._comchk(comments, error)

    def test_comments_backouts(self):
        """ensure we allow valid backouts and catch invalid ones.
        Note that only gatekeepers should be pushing backouts, and the
        gate hooks should enforce this."""
        comments = ["backout 23301494: needs work"]
        self._comchk(comments, [], expect_fail=False)
        comments = [
            "backout 23301494: needs work",
            "backout 22475509/23527289/21936951/23246145/21646967: needs work"
        ]
        self._comchk(comments, [], expect_fail=False)

        err = "The following comments are badly formatted for backouts"
        comments = ["backout: needs work"]
        self._comchk(comments, err)
        comments = ["backout 123456 needs work"]
        self._comchk(comments, err)

        err = "Bugfixes cannot be pushed at the same time as backouts"
        comments = [
            "backout 23301494: needs work",
            "23301494 loading text needs to be suppressed in GRUB2 when "
            "booting from disk"
        ]
        self._comchk(comments, err)

        comments = [
            "backout 23301494: needs work",
            "20028646 grub2 should build with gcc 4.9 instead of 4.7"
        ]
        self._comchk(comments, err)

    def test_comments_pipe_symbol(self):
        """ensure we warn for pipe symbols in the synopsis.
        Note that if having a pipe symbol is correct, gatekeeper intervention
        will be required to push this changeset."""

        err = "The following comments contain a pipe symbol."
        comments = ["12345678 this contains a | symbol"]
        self._comchk(comments, err)

        comments = ["12345678 ends with a pipe |"]
        self._comchk(comments, err)

    def _rtichk(self, ids, expected_output):
        """Given a list of bug ids and the expected output, verify that
        we get that expected output from rtichk.
        """
        output = StringIO.StringIO()
        ret = onbld.Checks.Rti.rti(
            ids, output=output, gatePath="/ws/on-gate")
        if isinstance(expected_output, basestring):
            self.assert_(
                expected_output in output.getvalue(),
                "expected output not found.\noutput:\n"
                "---\n"
                "%s\n"
                "expected output\n---\n%s\n" %
                (output.getvalue(), expected_output))
        else:
            for expected in expected_output:
                self.assert_(expected in output.getvalue())
        self.logger.debug("%s %s" % (ret, output.getvalue()))
        self.assert_(not ret)

    def test_rti_not_found(self):
        """comments with a bug that doesn't have an RTI are detected"""
        # This test is brittle: if we ever do fix this bug, this test will
        # fail.
        cr_ids = ["15028115"]
        error = "RTI not found for CR: 15028115"
        self._rtichk(cr_ids, error)

    def test_followup_warn(self):
        """putbacks with more than one synopsis line, where at least one line
        looks like a followup push get an error."""

        followup_error = \
            "Follow up pushes should only include a single comment line " \
            "with the most relevant bug, and should omit ARC cases."

        # a perfectly valid followup comment
        comments = [
            "20540862 ON binaries should be delivered as PIE (fix gcc)"
        ]
        error = []
        out = self._comchk(comments, error, expect_fail=False)
        self.assert_(followup_error not in out)

        # expect a warning
        comments = [
            "PSARC/2015/212 Audit Record grouping by Tags",
            "21218238 provide a way to group audit records by "
            "tags (fix include)"]
        self._comchk(comments, followup_error)

    def test_rti_not_accepted(self):
        """comments with a bug with an unaccepted RTI are detected"""

        # first, we have to find an RTI that's submitted, but not yet accepted,
        # which means that this test is brittle: if all RTIs are accepted,
        # then the test will fail.
        now = datetime.now()
        start = now - timedelta(days=30)
        # http://spe-tools-websvcs.us.oracle.com:9100/WebRTI/rtiList?startDate=2015-04-13%200:0:0&stopDate=2015-05-13%200:0:0&gate=SOLARIS-TRUNK&consolidation=ON
        url = (
            "http://spe-tools-websvcs.us.oracle.com:9100/"
            "WebRTI/rtiList?startDate={start}&stopDate={stop}"
            "&gate=SOLARIS-TRUNK&consolidation=ON".format(
                start=start.strftime("%F%%20%H:%m:%S"),
                stop=now.strftime("%F%%20%H:%m:%S")))
        self.logger.debug("Retrieving\n%s" % url)
        try:
            op = urllib2.urlopen(url)
            xml = op.read()
            op.close()
        except urllib2.HTTPError as e:
            self.assert_(
                False,
                "Unable to run test: HTTP error for {url} got: {e}".format(
                    url=url, e=e))
        elroot = etree.fromstring(xml)
        schema = elroot.nsmap[None]
        cr_ids = []
        rti = None
        rti_id = None

        # lxml won't allow xpath selectors, so we'll dig out the first
        # S_SUBMITTED RTI by hand
        for element in elroot.findall(".//{%s}status" % (schema)):
            if element.text == "S_SUBMITTED":
                rti = element.getparent()
                rti_id = rti.find("{%s}webrti" % schema).text
                break

        if rti is None:
            self.assert_(
                False, "Unable to find any submitted RTIs, so cannot "
                "run test. This may not be a bug.")

        crs = rti.find("{%s}crs" % schema)
        for cr in crs.iterchildren():
            cr_ids.append(cr.find("{%s}id" % schema).text)

        errors = [
            "Error: RTI %s for CR %s is not in the accepted state." %
            (rti_id, cr_id) for cr_id in cr_ids
        ]
        self._rtichk(cr_ids, errors)

    def test_copyright(self):
        """Verify that copyright-statement formatting checks work."""

        # use this year so that the test cases don't rot over time.
        year = time.strftime("%Y")
        copyright_error_tuples = [
            (
                "check pre 2002 copyright with (c)",
                "Copyright (c) 2000 Sun Microsystems, Inc.",
                "ancient Sun copyright"),
            (
                "check pre 2002 copyright with by",
                "Copyright 1987 by Sun Microsystems, Inc.",
                "pre-2002 Sun copyright"),
            (
                "check last valid Sun copyright",
                "Copyright 2010, Sun Microsystems, Inc. All rights "
                "reserved.", "old Sun copyright",),
            (
                "check to make sure subject to license terms is gone",
                "Copyright (c) %s, Oracle and/or its affiliates. "
                "All rights reserved.\nUse is subject to license terms." %
                year,
                "old 'Use is subject to license terms.' message found"),
            (
                "check initial oracle copyright, pre-2020, single year",
                "Copyright (c) %s, Oracle and/or its affiliates. "
                "All rights reserved." % year, "old Oracle copyright"),
            (
                "check initial oracle copyright, pre-2020, multiple year",
                "Copyright (c) 2010, %s, Oracle and/or its affiliates. "
                "All rights reserved." % year, "old Oracle copyright"),
            (
                "check wrong single copyright year",
                "Copyright (c) 1975, Oracle and/or its affiliates.",
                "wrong copyright year 1975, should be %s" % year),
            (
                "check need comma after current year",
                "Copyright (c) 1975, %s Oracle and/or its affiliates." % year,
                "need comma after current year"),
            (
                "check for matching start and end dates",
                "Copyright (c) %s, %s, Oracle and/or its affiliates." %
                (year, year),
                "start year %s must be earlier than end year %s" %
                (year, year)),
            (
                "check start year later than end year",
                "Copyright (c) 2199, %s, Oracle and/or its affiliates." % year,
                "start year 2199 must be earlier than end year %s" % year),
            (
                "check for missing space between years",
                "Copyright (c) 1975,%s, Oracle and/or its affiliates." % year,
                "need space after initial year"),
            (
                "check for missing copyright",
                "yabadabbadoo", "no copyright message found")
        ]
        for testname, copyright, error in copyright_error_tuples:
            output = StringIO.StringIO()
            lines = copyright.split("\n")
            self.logger.debug("Running test: %s" % testname)
            ret = onbld.Checks.Copyright.copyright(
                lines, filename="no-file.txt", output=output)
            self.logger.debug("input was: %s" % copyright)
            self.logger.debug("output was: %s" % output.getvalue())
            if error:
                self.assert_(
                    error in output.getvalue(),
                    "expected error %s not in output" % error)
                self.assert_(ret > 0, "expected error %s not found!" % error)
            else:
                self.assert_(ret == 0, "error unexpectedly reported!")

    def test_hdrchk(self):
        pass

    def test_jsstyle(self):
        """Verify that jstyle is being called correctly. Note that this isn't
        a comprehensive test of jstyle itself, rather that we can call
        jsstyle correctly and deal with its return code correctly."""

        bad_js = "   print(\"hello world\")"
        good_js = "print('hello world')"

        def run_jsstyle(js_source):
            """Write a temporary java file, then run jsstyle on it."""
            output = StringIO.StringIO()
            fd, fname = tempfile.mkstemp(".js", "")
            with open(fname, "w") as jfile:
                jfile.write(js_source)
            with open(fname, "r") as jfile:
                ret = onbld.Checks.JSStyle.jsstyle(
                    jfile, output=output, )
                self.logger.debug("%s: %s" % (ret, output.getvalue()))
            os.remove(fname)
            return ret, output.getvalue()

        ret, out = run_jsstyle(good_js)
        self.assert_(ret == 0, "jsstyle false positive on valid js file")
        self.assert_(out == "", "unexpected output from jsstyle: %s" % out)

        ret, out = run_jsstyle(bad_js)
        self.assert_(ret != 0, "jsstyle did not detect invalid js file")
        self.assert_(out != "", "expected output from jsstyle!")

    def test_jstyle(self):
        """Verify that jstyle is being called correctly. Note that this isn't
        a comprehensive test of jstyle itself, rather that we can call
        jstyle correctly and deal with its return code correctly."""

        hello = "System.out.println(\"Hello world.\");"
        bad_java = \
            "public class File {public static main String([] args){%s}" % hello
        good_java = """
public class File {
    public static void
    main(String[] args)
    {
        %s
    }
}
""" % hello

        def run_jstyle(java_source):
            """Write a temporary java file, then run jsstyle on it."""
            output = StringIO.StringIO()
            fd, fname = tempfile.mkstemp(".java", "")
            with open(fname, "w") as jfile:
                jfile.write(java_source)
            with open(fname, "r") as jfile:
                ret = onbld.Checks.JStyle.jstyle(
                    jfile, output=output, )
                self.logger.debug("%s: %s" % (ret, output.getvalue()))
            os.remove(fname)
            return ret, output.getvalue()

        ret, out = run_jstyle(good_java)
        self.assert_(ret == 0, "jstyle false positive on valid java file")
        self.assert_(out == "", "unexpected output from jstyle: %s" % out)

        ret, out = run_jstyle(bad_java)
        self.assert_(ret == 1, "jstyle did not detect invalid java file")
        self.assert_(out != "", "expected output from jstyle!")

    def test_keywords(self):

        good_keywords = ["This line has no keywords"]
        bad_keywords = ["#ident	\"%Z%%M% %I% %E% SMI\""]
        output = StringIO.StringIO()
        ret = onbld.Checks.Keywords.keywords(
            good_keywords, filename="source.txt", output=output)
        self.assert_(ret == 0, "unexpected failure with good_keywords")
        self.assert_(
            output.getvalue() == "",
            "unexpected output %s" % output.getvalue())

        ret = onbld.Checks.Keywords.keywords(
            bad_keywords, filename="source.txt", output=output)
        self.assert_(ret == 1, "unexpected failure with good_keywords")
        self.assert_("contains SCCS keywords" in output.getvalue())

    def test_mapfile(self):
        """Test that our mapfile checks are functioning."""
        MAPFILE = "#\n" \
            "# MAPFILE HEADER START\n" \
            "#\n" \
            "# WARNING:  STOP NOW.  DO NOT MODIFY THIS FILE.\n" \
            "# Object versioning must comply with the rules detailed in\n" \
            "#\n" \
            "#\tusr/src/lib/README.mapfiles\n" \
            "#\n" \
            "# You should not be making modifications here until " \
            "you've read the most current\n" \
            "# copy of that file. If you need help, contact a gatekeeper " \
            "for guidance.\n" \
            "#\n" \
            "# MAPFILE HEADER END\n" \
            "#\n\n"

        def _mapfile_check(contents, expected_error):

            fd, fname = tempfile.mkstemp(".map", "")
            with open(fname, "w") as mapfile:
                mapfile.write(contents)

            ret = 0
            with open(fname, "r") as mapfile:
                output = StringIO.StringIO()
                ret = onbld.Checks.Mapfile.mapfilechk(
                    mapfile, filename="mapfile.txt", output=output)

            os.remove(fname)
            if expected_error:
                self.assert_(expected_error in output.getvalue())
                self.assert_(ret > 0)
            else:
                self.assert_(
                    ret == 0, "Mapfile check of %s unexpectedly failed:\n%s" %
                    (contents, output.getvalue()))

        _mapfile_check(
            "Ceci n'est pas une mapfile",
            "Warning: mapfile version 2 syntax expected in file")
        _mapfile_check(
            "$mapfile_version 2\nsymbol_version \n",
            "Warning: No MAPFILE block")
        _mapfile_check(
            MAPFILE + "$mapfile_version 2\nsymbol_version \n", None)
        _mapfile_check("$mapfile_version 2", None)

    def test_license(self):
        """Test that our utf8 license check works."""

        fd, fname = tempfile.mkstemp(".license", "")
        output = StringIO.StringIO()
        with open(fname, "w+b") as badlicfile:
            byte_arr = [76, 105, 99, 101, 110, 115, 101, 32, 116, 111,
                        32, 107, 105, 108, 108, 32, 246, 110, 252]
            binary_format = bytearray(byte_arr)
            badlicfile.write(binary_format)
            badlicfile.close()
            ret = onbld.Checks.UTF8.utf8chk(badlicfile, output=output)
        os.remove(fname)
        self.assert_(
            ret == 1, "UTF8 check of an ISO-8859-1 file unexpectedly passed\n")

        fd, fname = tempfile.mkstemp(".license", "")
        with open(fname, "w+b") as goodlicfile:
            byte_arr = [76, 105, 99, 101, 110, 115, 101, 32, 116, 111,
                        32, 107, 105, 108, 108, 32, 195, 182, 110, 195, 188]
            binary_format = bytearray(byte_arr)
            goodlicfile.write(binary_format)
            goodlicfile.close()
            ret = onbld.Checks.UTF8.utf8chk(goodlicfile, output=output)
        os.remove(fname)
        self.assert_(
            ret == 0, "UTF8 check of a UTF-8 file unexpectedly failed\n")

    def test_tags(self):
        """Test that our tags pbchk is functioning."""

        bugdb = self._get_bugdb()

        def run_tags(contents, expected_error):
            output = StringIO.StringIO()
            ret = onbld.Checks.Tags.tagchk(
                [line.rstrip() for line in contents],
                self.testdir, gatePath="/export/on12-gate", bugdb=bugdb,
                output=output)
            out = output.getvalue()
            if not expected_error:
                self.assert_(
                    ret == 0, "tag check for %s failed: %s" %
                    (contents, out))
            else:
                self.assert_(
                    ret == 1, "tag check for %s passed unexpectedly: %s" %
                    (contents, out))
                if isinstance(expected_error, list):
                    for error in expected_error:
                        self.assert_(
                            error in out,
                            "Expected to see \"%s\" in \"\n%s\"" %
                            (error, out))
                else:
                    self.assert_(
                        expected_error in out,
                        "Expected to see \"%s\" in \"\n%s\"" %
                        (expected_error, out))

            if out.strip():
                self.logger.debug(out)
            return out

        def write_tag_exceptions(content):
            exception_dir = "%s/exception_lists" % self.testdir
            if not os.path.isdir(exception_dir):
                os.makedirs(exception_dir)
            with open("%s/bugtagchk" % exception_dir, "w") as tagchk:
                tagchk.write(content)

        # When there are no bugs in the outgoing changeset, we don't complain
        run_tags(["This is not an ON valid commit message"], None)

        # Make sure errors are not reported when we don't require tag values
        write_tag_exceptions("[empty]\nrequired = false\n")
        run_tags(["19569805 Need kernel support for kstats v2"], None)
        write_tag_exceptions("[empty]\nrequired = false\ncomp = utility")
        run_tags(["19569805 Need kernel support for kstats v2"], None)
        write_tag_exceptions(
            "[empty]\nrequired = false\ncomp =utility\ncomp = */zones\n")
        run_tags(["19569805 Need kernel support for kstats v2"], None)
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = false",
            "allowed_values = cat, fish, mouse"]))
        run_tags(["19569805 Need kernel support for kstats v2"], None)

        # Errors are reported when we do require tag values
        write_tag_exceptions("[mytag]\nrequired = true\n")
        run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'mytag' must be set")
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            ["Tag 'empty' must be set", "Tag 'another' must be set"])
        self.assert_("Tag 'last' must be set" not in output)

        # Negative tests to ensure we don't complain about cat/subcats that
        # don't match the bugs we're fixing
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = utility",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'another' must be set")
        self.assert_("Tag 'empty' must be set" not in output)
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = */zones",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'another' must be set")
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = utility/zones",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'another' must be set")
        self.assert_("Tag 'empty' must be set" not in output)

        # Positive test to ensure we complain about comp/subcomps that
        # do match our bug
        # just the cat
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = kernel",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            [
                "Tag 'empty' must be set  "
                "(component 'kernel' bugs require this tag)",
                "Tag 'another' must be set"])
        self.assert_("Tag 'last' must be set" not in output)
        # just subcomponent
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = */kstats",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'empty' must be set  (subcomponent 'kstats' bugs "
            "require this tag)")
        self.assert_("Tag 'last' must be set" not in output)
        # cat and subcat
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = kernel/kstats",
            "[another]",
            "required = true",
            "[last]",
            "required = false"]))
        output = run_tags(
            ["19569805 Need kernel support for kstats v2"],
            [
                "Tag 'empty' must be set  "
                "('kernel/kstats' bugs require this tag)",
                "Tag 'another' must be set"])
        self.assert_("Tag 'last' must be set" not in output)

        # allow for comma-separated comp/subcomp values
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = kernel, service"]))
        output = run_tags(
            [
                "19569805 Need kernel support for kstats v2",
                "21071863 pkg/depot should use Apache 2.4"],
            [
                "Tag 'empty' must be set  "
                "(component 'kernel' bugs require this tag)",
                "Tag 'empty' must be set  "
                "(component 'service' bugs require this tag)"])
        write_tag_exceptions("\n".join([
            "[empty]",
            "required = true",
            "comp = kernel/kstats, service/pkgdepot"]))
        output = run_tags(
            [
                "19569805 Need kernel support for kstats v2",
                "21071863 pkg/depot should use Apache 2.4"],
            [
                "Tag 'empty' must be set  "
                "('kernel/kstats' bugs require this tag)",
                "Tag 'empty' must be set  "
                "('service/pkgdepot' bugs require this tag)"])

        # make sure substrings of tags don't match
        write_tag_exceptions("\n".join([
            "[kstats]",
            "required = true",
            "comp =kernel/kstats"]))
        run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'kstats' must be set")
        # and this should be fine since it's a valid tag set on this bug
        write_tag_exceptions("\n".join([
            "[kstatsv2]",
            "required = true",
            "comp =kernel/kstats"]))
        run_tags(["19569805 Need kernel support for kstats v2"], None)
        # but not this
        write_tag_exceptions("\n".join([
            "[kstatsv20]",
            "required = true",
            "comp = kernel/kstats"]))
        run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'kstatsv20' must be set")

        # Check for key/value pair tags with allowed values
        write_tag_exceptions("\n".join([
            "[kstatsv2]",
            "required = true",
            "allowed_values = mice|chickens"]))
        run_tags(
            ["19569805 Need kernel support for kstats v2"],
            "Tag 'kstatsv2' must be set to values: chickens or mice")
        # no matching category
        write_tag_exceptions("\n".join([
            "[kstatsv2]",
            "comp = utility",
            "required = true",
            "allowed_values = mice|chickens"]))
        run_tags(["19569805 Need kernel support for kstats v2"], None)

        # this should be fine
        write_tag_exceptions("\n".join([
            "[changeset]",
            "required = true",
            "allowed_values = 9924b63b337c"]))
        run_tags(["21071863 pkg/depot should use Apache 2.4"], None)

        write_tag_exceptions("\n".join([
            "[changeset]",
            "required = true",
            "allowed_values = 9924b63b337c|cows"]))
        run_tags(["21071863 pkg/depot should use Apache 2.4"], None)

        write_tag_exceptions("\n".join([
            "[changeset]",
            "required = true",
            "allowed_values = 9924b63b337cZ|cows"]))
        run_tags(
            ["21071863 pkg/depot should use Apache 2.4"],
            "Tag 'changeset' must be set to values: "
            "9924b63b337cZ or cows")

        # try a few bugs with no exception list
        write_tag_exceptions("# this is just a comment\n")
        run_tags([
            "19569805 Need kernel support for kstats v2",
            "21071863 pkg/depot should use Apache 2.4"], None)

        # try a few bugs with one bug tag not matching
        write_tag_exceptions("\n".join([
            "[changeset]",
            "comp = service",
            "required = true",
            "allowed_values = 9924b63b337cZ|cows"]))
        out = run_tags([
            "19569805 Need kernel support for kstats v2",
            "21071863 pkg/depot should use Apache 2.4"],
            "Tag 'changeset' must be set to values: "
            "9924b63b337cZ or cows")
        self.assert_("19569805 Need kernel support for kstats v2" not in out)
        self.assert_("21071863 pkg/depot should use Apache 2.4" in out)

        # find different errors with all bugs
        write_tag_exceptions("\n".join([
            "[changeset]",
            "comp = */pkgdepot",
            "required = true",
            "allowed_values = 9924b63b337cZ|cows",
            "[empty]",
            "required = true"]))
        out = run_tags([
            "19569805 Need kernel support for kstats v2",
            "21071863 pkg/depot should use Apache 2.4"],
            [
                "Tag 'changeset' must be set to values: "
                "9924b63b337cZ or cows",
                "Tag 'empty' must be set"])
        self.assert_("19569805 Need kernel support for kstats v2" in out)
        self.assert_("21071863 pkg/depot should use Apache 2.4" in out)
        # verify we reported the missing 'empty' tag twice
        self.assert_(out.count("Tag 'empty' must be set") == 2)

        # make sure matching one or more tag values is allowed work
        write_tag_exceptions("\n".join([
            "[pbchk]",
            "allowed_values = catch, all, the, bugs"]))
        out = run_tags(
            ["22557519 hg pbchk needs the ability to validate bug tags"],
            None)

        # only two matching values, which is fine
        write_tag_exceptions("\n".join([
            "[pbchk]",
            "allowed_values = catch, all"]))
        out = run_tags(
            ["22557519 hg pbchk needs the ability to validate bug tags"],
            None)
        # matches the first group, but not the 2nd
        write_tag_exceptions("\n".join([
            "[pbchk]",
            "allowed_values = catch, all, the, bugs | noisy"]))
        out = run_tags(
            ["22557519 hg pbchk needs the ability to validate bug tags"],
            None)

        # matches multiple groups, which should fail
        write_tag_exceptions("\n".join([
            "[pbchk]",
            "allowed_values = catch, all, the, noisy | annoying, bugs "]))
        out = run_tags(
            ["22557519 hg pbchk needs the ability to validate bug tags"],
            "matches more than one group of the values")

        # specify groups that the bug doesn't have
        write_tag_exceptions("\n".join([
            "[pbchk]",
            "allowed_values = chocolate| vanilla |rum, raisin"]))
        out = run_tags(
            ["22557519 hg pbchk needs the ability to validate bug tags"],
            "Tag 'pbchk' must be set to values:")

    def test_notify(self):
        """Test that notify works."""

        def run_notify(files, ws, expected_output):
            output = StringIO.StringIO()
            onbld.Checks.Notify.notify(
                [(f, None) for f in files], ws, output=output)
            out = output.getvalue()
            if expected_output:
                if isinstance(expected_output, list):
                    for text in expected_output:
                        self.assert_(
                            text in out,
                            "Expected to see \"%s\" in \"\n%s\"" %
                            (text, out))
                else:
                    self.assert_(
                        expected_output in out,
                        "Expected to see \"%s\" in \"\n%s\"" %
                        (expected_output, out))

            if out.strip():
                self.logger.debug(out)
            return out

        ws = os.environ["SRC"].replace("usr/src", "")
        run_notify(["/no/such/file"], ws, "")
        # issue a unit test notice
        run_notify(
            ["%s/usr/src/tools/scripts/build.py" % ws], ws,
            ["tests are available in the gate", "build.cfg"])

        # issue a developer notification
        out = run_notify(
            ["%s/usr/src/cmd/cmd-inet/usr.sbin/snoop/Makefile" % ws],
            ws, ["No new support should be added to snoop", "snoop.cfg"])
        self.assert_("build.cfg" not in out)

        # change two files, one with tests, the other with a developer
        # notification
        run_notify(
            ["%s/usr/src/tools/scripts/build.py" % ws,
                "%s/usr/src/cmd/cmd-inet/usr.sbin/snoop/Makefile" % ws],
            ws, [
                "No new support should be added to snoop",
                "tests are available in the gate", "build.cfg", "snoop.cfg"])

        # issue two unit-test notifications
        run_notify(
            ["%s/usr/src/tools/scripts/build.py" % ws,
                "%s/usr/src/cmd/rad/daemon/Makefile" % ws],
            ws, [
                "tests are available in the gate",
                "build.cfg", "rad-test.cfg"])

        # issue two developer notifications
        run_notify(
            ["%s/usr/src/tools/onbld/hgext/cdm.py" % ws,
                "%s/usr/src/cmd/cmd-inet/usr.sbin/snoop/Makefile" % ws],
            ws, [
                "No new support should be added to snoop",
                "we do not run them automatically", "pbchk.cfg", "snoop.cfg"])

        # change two files, one with tests, the other with a developer
        # notification
        run_notify(
            ["%s/usr/src/tools/scripts/build.py" % ws,
                "%s/usr/src/cmd/cmd-inet/usr.sbin/snoop/Makefile" % ws],
            ws, [
                "No new support should be added to snoop",
                "tests are available in the gate", "build.cfg", "snoop.cfg"])

    def test_cdm_nits(self):
        """Test that hg nits works."""
        nits_command = "{0} nits".format(self.hg_cdm)
        ret, out, err = cmdline_run(nits_command, cwd=self.ws)
        self.assert_(ret == 0, "hg nits failed")
        # sanity test that we have expected output, and that nothing from
        # pbchk-only is included
        self.assert_("Copyright check:" in out)
        self.assert_("C style check:" in out)
        self.assert_("RTI check:" not in out)

        # check that nits runs on a single file
        nits_command = "{0} nits prototype.c".format(self.hg_cdm)
        ret, out, err = cmdline_run(nits_command, cwd=self.ws)
        self.assert_(ret == 0, "hg nits failed")
        self.assert_("File permission check:" in out)
        self.assert_("Header format check:" in out)
        self.assert_("Package history checks:" not in out)

    def test_cdm_pbchk(self):
        """Test that hg pbchk works."""
        pbchk_command = "{0} pbchk".format(self.hg_cdm)
        ret, out, err = cmdline_run(pbchk_command, cwd=self.ws)
        self.assert_(ret == 0, "hg pbchk failed")
        self.assert_("Checking for uncommitted changes:" in out)
        self.assert_("Checking for merges:" in out)

    def test_dblookup(self):
        """Test that DB lookup works. This is basic sanity checking as
        other tests in this module implicitly test that DbLookups works.
        """
        bugdb = self._get_bugdb()

        # Since bug_type isn't one of the standard fields, verify that
        # we can get that value, and can correctly tell which CRs are
        # bugs, and which are RFEs
        res = bugdb.lookup(["22177665"], fields=["cr_number", "bug_type"])
        self.assert_(res["22177665"]["bug_type"] == "B")

        res = bugdb.lookup(["22869182"], fields=["cr_number", "bug_type"])
        self.assert_(res["22869182"]["bug_type"] == "E")

    def test_spelchk(self):
        """Verify that we can check for common misspellings."""

        def spelr_(contents, expected_error="", fakepath="bad.c"):
            fd, fname = tempfile.mkstemp(".c", "")
            with open(fname, "w") as f:
                f.write(contents)

            with open(fname, "r") as f:
                output = StringIO.StringIO()
                ret = onbld.Checks.SpelChk.spel_chekc(f, filename=fakepath,
                                                      output=output)

            os.remove(fname)
            if expected_error:
                self.assert_(output.getvalue() == expected_error,
                             "Expected: %s\nGot: %s\n" % (expected_error,
                                                          output.getvalue()))
            else:
                self.assert_(ret == 0,
                             "'spell' should not have produced:\n%s" %
                             output.getvalue())

        # Unit test: pass
        spelr_("This is a valid string")

        # NOTE: In the following tests, the misspelled words are
        # split in order to prevent the spell checker from flagging them.

        # Unit test: check for single error
        spelr_("I have comi" "tted a crime against spelling.",
               expected_error=""
               "bad.c: 1: 'com" "itted' should be 'committed'\n")

        # Unit test: suggestions should match the case of the original
        spelr_("Gau" "rd against typos at all times.",
               expected_error="bad.c: 1: 'Gau" "rd' should be 'Guard'\n")
        spelr_("THIS IS NOT GAU" "RANTEED",
               expected_error=""
               "bad.c: 1: 'GAU" "RANTEED' should be 'GUARANTEED'\n")

        # Unit test: check for multiple errors on the same line
        spelr_("It is a priv" "iledge to be publ" "ically ackn" "owleged.",
               expected_error=""
               "bad.c: 1: 'pri" "viledge' should be 'privilege'\n"
               "bad.c: 1: 'pub" "lically' should be 'publicly'\n"
               "bad.c: 1: 'ack" "nowleged' should be 'acknowledged'\n")

        # Unit test: multiple errors on different lines
        spelr_("""What is this, ama""" """tuer hour?
               I am exa""" """usted by these typos.
               Correct spelling is an unnec""" """essery burden.""",
               expected_error=""
               "bad.c: 1: 'ama" "tuer' should be 'amateur'\n"
               "bad.c: 2: 'exa" "usted' should be 'exhausted'\n"
               "bad.c: 3: 'unn" "ecessery' should be 'unnecessary'\n")

        # Unit test: ignore spelling errors in SpelChk.py
        spelr_("Tra" "nsfering All Res" "ponsibility", fakepath="/SpelChk.py")


class CheckTestRunner(unittest.TextTestRunner):
    """Used so we can modify the environment to allow more verbose logging.
    By default the "-v" argument just causes unittest to print test
    descriptions before test result status, but we want more verbosity than
    that."""

    def __init__(self, stream=sys.stderr, descriptions=1, verbosity=1):
        if verbosity > 1:
            os.environ["ONBLD_DEBUG"] = "true"
        # $ONBLD_BUILDTEST must not be in the environment, since that
        # influences the 'build generate' output, which can break tests
        if "ONBLD_BUILDTEST" in os.environ:
            os.environ.pop("ONBLD_BUILDTEST")

        # When running as a user with an NFS home directory, setting
        # $PYTHONNOUSERSITE can speed up test execution. See python(1).
        os.environ["PYTHONNOUSERSITE"] = "true"
        unittest.TextTestRunner.__init__(
            self, stream=stream, descriptions=descriptions,
            verbosity=verbosity)


if __name__ == '__main__':
    unittest.main(
        testRunner=CheckTestRunner(verbosity=2 if "-v" in sys.argv else 1))
