#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License version 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#
# Copyright (c) 2008, 2019, Oracle and/or its affiliates. All rights reserved.
# Copyright 2008, 2010, Richard Lowe
#

"""
Deal with Mercurial versioning.

At a basic level, code to verify that the version of Mercurial in use
is suitable for use with Cadmium, and compare that version for the
sake of adapting to Mercurial API changes.
"""

#
# It is important that this module rely on as little of Mercurial as
# is possible.
#

from mercurial import util
import mercurial.ui


class VersionMismatch(Exception):
    "Exception used to indicate a mismatch between SCM tools and Mercurial"
    pass


#
# List of versions that are explicitly acceptable to us
#
GOOD_VERSIONS = [
    "4.1", "4.7.1", "4.9.1", "5.1.1"
]


def check_version():
    """Check that we're running on a suitable version of Mercurial"""

    def versionstring(versions):
        """return the list, versions, as a vaguely grammatical string"""
        if len(versions) > 1:
            return "%s or %s" % (", ".join(versions[0:-1]), versions[-1])
        else:
            return versions[0]

    # Changed in 4.1 (d83ca854fa21)
    try:
        ui = mercurial.ui.ui.load()
    except AttributeError:
        ui = mercurial.ui.ui()

    versions = ui.configlist("cdm", "mercurial_versions", GOOD_VERSIONS)
    if not versions:
        versions = GOOD_VERSIONS
    if "all" in versions:
        return
    if "default" in versions:
        i = versions.index("default")
        versions[i:i + 1] = GOOD_VERSIONS
    if util.version() not in versions:
        # If we don't find the explicitly tested version, see if we match on
        # the major/minor pair, since Mercurial typically maintains API
        # compatibility within a minor version range.
        hgv = tuple([int(v) for v in util.version().split(".")[:2]])
        good_vers = set(tuple(_split_version(v)[:2]) for v in versions)
        if hgv not in good_vers:
            raise VersionMismatch(
                "Scm expects Mercurial version %s, actual version is %s." %
                (versionstring(versions), util.version()))


def _split_version(ver):
    """Return the Mercurial version as a list [MAJOR, MINOR, MICRO],
    if this is not a released Mercurial return None."""

    try:
        m = map(int, ver.split("."))
        # If there's only one element, it's not really a tagged version
        if len(m) <= 1:
            return None
        else:
            return m
    except ValueError:
        return None


def at_least(desired):
    """Return boolean indicating if the running version is greater
    than or equal to, the version specified by major, minor, micro"""

    hgver = _split_version(util.version().split("+")[0])
    desired = map(int, desired.split("."))

    #
    # If _split_version() returns None, we're running on a Mercurial that
    # has not been tagged as a release.  We assume this to be newer
    # than any released version.
    #
    if hgver is None:
        return True

    # Pad our versions to the same overall length, appending 0's
    while len(hgver) < len(desired):
        hgver.append(0)
    while len(desired) < len(hgver):
        desired.append(0)

    for real, req in zip(hgver, desired):
        if real != req:
            return real > req

    return True
