#!/usr/bin/python

#
# Copyright (c) 2013, 2021, Oracle and/or its affiliates.
#

# To assist 2->3 transition
from __future__ import print_function

import atexit
import collections
import copy
import datetime
import errno
import logging
import lxml.etree
import os
import platform
import pwd
import re
import shutil
import tempfile

from six.moves import configparser
from six.moves import StringIO

import onbld.util as util
import onbld.log as log

"""A clean shell environment used to build the OS/Net Consolidation.
"""

logger = logging.getLogger("onbld")

ret, out, err = util.cmdline_run("/usr/bin/uname -p")
if ret != 0 or not out:
    raise ValueError("Unable to run /usr/bin/uname!")
MACH = out.rstrip()
HOSTNAME = platform.uname()[1]

# Dmake became awkward in Studio 12.6 using the strings "sparc64" and
# "x64" which we don't otherwise need or use in the environment. Handle
# that here.
if MACH == "i386":
    MACH64 = "amd64"
    DMAKE_MACH = "x64"
else:
    MACH64 = "sparcv9"
    DMAKE_MACH = "sparc64"

# Used whenever we need the user to contact the gatekeeper alias for advice.
GATEKEEPER_MAILTO = "gatekeeper@on.us.oracle.com"

# The minimum amount of swap in gigabytes that we recommend for build machines.
MINIMUM_RECOMMENDED_SWAP = 12

# A file whose presence indicates this is a usr/fish repository. We could
# check the '0'th changeset hash, but this seems more efficient.
FISH_STANDALONE_CANARY_PATH = ".fish_repo"

# Section names for the bldrc file
SECT_PATHS = "paths"
SECT_BUILDOPTIONS = "build_options"
SECT_CODECHECKS = "code_checks"
SECT_PARTIAL = "partial_builds"

# The gkoverrides section is special - it's not required to be in
# the bldrc file (and is omitted from the SECTS list below) For the
# most part, gatelings should never need to use it.  It contains
# options to override values for gatekeeper use only.
SECT_GKOVERRIDES = "gkoverrides"

# The checkoverrides section is also special. It's also not required to
# be in the bldrc file and is also omitted from SECTS. It contains only
# check.* key/value pairs taking the values 'true' or 'false' which state
# whether a certain check should be enabled, or check.*.must_pass keys, taking
# the values 'true' if the test must pass in order for build to exit 0, or
# false if a failure in that check should not impact the build exit code.
SECT_CHECKOVERRIDES = "checkoverrides"

SECTS = [SECT_PATHS, SECT_BUILDOPTIONS, SECT_CODECHECKS, SECT_PARTIAL]

# A series of documentation comments allowing us to build our own bldrc
DOC_SECTS = {
    SECT_PATHS: """
#
# Paths used in the build.
#
""",
    SECT_BUILDOPTIONS: """
#
# General build options.
#
""",
    SECT_CODECHECKS: """
#
# Build options involved in verifying build output
#
""",
    SECT_PARTIAL: """
#
# Build options needed when building a partial tree, or producing
# binaries used for partial tree builds (eg. a missing usr/closed tree)
#
""",
    SECT_GKOVERRIDES: """
#
# Settings that are usually only needed by gatekeepers
#
""",
    SECT_CHECKOVERRIDES: """
#
# Settings that allow more control over build checks
#
"""}

BLDRC_BANNER = """
#
# This is a build config file, originally generated with
# generate_cmd:: %(generate_cmd)s
# by %(user)s on %(date)s
#
"""

CHECK_OVERRIDE_HELP = (
    "This setting makes build(1) skip the '%s' task when "
    "performing the 'nightly', 'qnightly' or 'check' tasks. The value must "
    "be one of 'true' or 'false'."
    "\n\n"
    "Note that some checks are skipped for certain types of build by default "
    "and the values of this setting will not impact that behaviour."
    "\n\n"
    "If 'check' is set to 'false' in your bldrc file (or with -C on the "
    "command line), this setting is ignored.")

CHECK_MUST_PASS_HELP = (
    "This setting determines whether a failing '%s' task should cause "
    "build(1) to exit 1 if that task is run as part of a composite build "
    "task (like 'nightly' or 'check')."
    "\n\n"
    "The value must be one of 'true' or 'false'."
    "\n\n"
    "Note that if the '%s' task is executed on it's own, the value of this "
    "key does not influence the build(1) exit code."
)

# used to differentiate several $WS/data directories and logs when building
# either osnet flavor
SERVER_PREFIX = "server-"
FISH_PREFIX = "fish-"

# A global we use to cache the computed AKV_BUILD environment variable.
# see get_build_environment(..)
AKV_BUILD = None

# A namedtuple we can use to store information about a given tool.
# The following members are used:
#
# name               the execname of the tool
# version_flag       the argument to pass to the tool to get the version no
# default            where we look for this tool by default
# minver             minimum acceptable version
# group_num          the compiletools_re regex group used to find the version
# bldrc              whether the tool should have an entry in the 'paths'
#                    section of the bldrc
#
# Allow these variable names as we use tuples like classes
# pylint: disable=C0103
ToolsTuple = collections.namedtuple(
    "ToolsTuple", "name version_flag default minver group_num bldrc")
# pylint: enable=C0103


def tastes_like_bldrc(bldrc_path):
    """Rather than doing a full bldrc parse and environment configuration,
    this exists merely to tell whether the provided file is likely to be a
    valid bldrc file. We can check its full validity later."""
    cfg = configparser.RawConfigParser()
    try:
        cfg.read(bldrc_path)
        # use a key and section name unlikely to be in a cfg file
        # by accident.
        cfg.get(SECT_BUILDOPTIONS, "cw_no_shadow")
    except configparser.Error:
        return False
    return True


class OnBldEnvironmentException(Exception):
    """An general exception we throw when working with OnBldEnvironment
    objects."""
    pass


class OnBldNoFishEnvironmentException(Exception):
    """Thrown when we've been asked to produce a usr/fish environment, but
    don't appear have a usr/fish repository."""
    pass


class Env(object):
    """This class represents details about a single UNIX environment variable
    that is known to the build system.
    """

    def __init__(self, name=None, default_value=None,
                 section=SECT_BUILDOPTIONS, docstring=""):
        """A default_value of None or "<None>" indicates that no environment
        variable should be exported for this Env object.

        A None section means the key cannot appear in a bldrc file.
        """

        self.name = name
        self.default = default_value
        if section not in SECTS and section != SECT_GKOVERRIDES \
           and section is not None:
            raise OnBldEnvironmentException("unknown section %s" % section)
        self.section = section
        self.doc = docstring


# a value used when we want to claim part of the namespace for an environment
# variable that we never want to set.
empty_env = Env(None, None, None, None)


class OnBldEnvironment(object):
    """This class contains a series of methods that can be called to generate
    a build environment (in the os.environ sense) used for the OS/Net build.

    It defines the syntax of 'bldrc' configuration files, used to influence how
    the environment is set, along with methods that can create temporary shell
    scripts, which can be sourced by $SHELL.

    User-overrides to the default environment are supported, allowing a given
    environment variable to override the default (or bldrc-influenced)
    environment, or to allow a user-specified environment variable to be
    cleared from the environment. If 'strict_environment' is True, any user
    overrides must be present in the environment."""

    def __init__(self, bldrc=None, ws_path=None, strict_environment=True):
        # We use a RawConfigParser to avoid the fancy interpolation of config
        # file entries that the other ConfigParser classes perform.
        self.__cfg = configparser.RawConfigParser()
        if bldrc:
            self.bldrc = os.path.abspath(bldrc)
            # we set self.ws_path from the value set in the bldrc file later
        else:
            self.bldrc = None
            self.ws_path = ws_path

        # Flags that influence how the build env should be configured.
        # Note that having a OnBldEnvironment with both debug and nondebug set
        # at the same time is allowed - it's a request that we perform actions
        # for both debug and non-debug builds. When we're generating the
        # os.environ in self.get_build_environment(), we choose one or the
        # other.
        self.debug = True
        self.nondebug = True
        self.common_task = False
        self.pkgmerge = False
        self.pkgsurf = True
        self.generate_splice = False
        self.archive_build = False
        # Allowed values for parfait_type are: "baseline", "upload"
        self.parfait_type = "baseline"
        self.incremental = False
        self.check = True
        self.snapshots = False
        self.build_in_place = False
        self.update_parent_ws = False
        self.log_format = log.NORMAL_FORMAT
        self.tmpdir = None
        self.strict_environment = strict_environment
        self.tools_version_check = None
        self.zipfiles = False

        # release indicates that this is a non-developer build, to be
        # delivered to customers.
        self.release = False

        # build_date and iso8601_date are properties, and are immutable once
        # set. Several build flavors (eg. debug and non-debug) should
        # have the same *_build_date.
        self.__build_date = None
        self.__iso8601_build_date = None
        self.user_env_overrides = False
        self.overrides = {}

        if self.bldrc:
            if not os.path.exists(self.bldrc):
                raise OnBldEnvironmentException(
                    "Unable to find bldrc file %s" % bldrc)
            try:
                self.__cfg.read(self.bldrc)
            except configparser.Error as e:
                raise OnBldEnvironmentException(
                    "Invalid bldrc file: %s\nA new bldrc file can be "
                    "generated by running 'build generate'" % e)

            # process bldrc overrides to the default flags
            try:
                self.parfait_type = self.__cfg.get(
                    SECT_CODECHECKS, "parfait_type")
                self.debug = self.__cfg.get(
                    SECT_BUILDOPTIONS, "debug").lower() == "true"
                self.nondebug = self.__cfg.get(
                    SECT_BUILDOPTIONS, "nondebug").lower() == "true"
                self.incremental = self.__cfg.get(
                    SECT_BUILDOPTIONS, "incremental").lower() == "true"
                self.check = self.__cfg.get(
                    SECT_BUILDOPTIONS, "check").lower() == "true"
                self.build_in_place = self.__cfg.get(
                    SECT_BUILDOPTIONS, "bip").lower() == "true"
                self.update_parent_ws = self.__cfg.get(
                    SECT_BUILDOPTIONS, "update_parent_ws").lower() == "true"
                self.snapshots = self.__cfg.get(
                    SECT_BUILDOPTIONS, "snapshots").lower() == "true"

                self.user_env_overrides = self.__cfg.get(
                    SECT_BUILDOPTIONS, "userenv") or \
                    self.__cfg.get(SECT_BUILDOPTIONS, "userunsetenv")
                self.ws_path = self.__cfg.get(SECT_PATHS, "ws")

            except configparser.Error as e:
                # if any of these are missing in the config file, something is
                # wrong.
                raise OnBldEnvironmentException(
                    "Missing or invalid bldrc option: %s" % e)

            bad_keys = []
            if self.__cfg.has_section(SECT_CHECKOVERRIDES):
                for option, val in self.__cfg.items(SECT_CHECKOVERRIDES):
                    if val not in ["true", "false"]:
                        bad_keys.append(option)
                    self.overrides[option] = val == "true"
                if bad_keys:
                    raise OnBldEnvironmentException(
                        "values should be either 'true' or 'false' "
                        "for the following keys: %s" % ", ".join(bad_keys))

            try:
                log_format = self.__cfg.get(SECT_GKOVERRIDES, "log_format")
                if log_format == "<None>":
                    self.log_format = log.NORMAL_FORMAT
                elif log_format not in log.log_format_names:
                    raise OnBldEnvironmentException(
                        "'log_format' must be one of: <None>, %s" %
                        ", ".join(list(log.log_format_names.keys())))
                else:
                    self.log_format = log.log_format_names[log_format]
            except (
                    configparser.NoSectionError, configparser.NoOptionError,
                    configparser.MissingSectionHeaderError):
                # most config files won't have gk options included.
                pass

            for section, attrs in [(SECT_BUILDOPTIONS,
                                    ["pkgmerge", "pkgsurf", "zipfiles"]),
                                   (SECT_GKOVERRIDES,
                                    ["generate_splice", "archive_build"])]:
                for key in attrs:
                    try:
                        self.__setattr__(
                            key,
                            self.__cfg.get(section, key).lower() == "true")
                    except (
                            configparser.NoSectionError,
                            configparser.NoOptionError,
                            configparser.MissingSectionHeaderError):
                        # most config files won't have gk options included.
                        pass

            altws_pkgarchive = None
            try:
                altws_pkgarchive = self.__cfg.get(
                    SECT_GKOVERRIDES, "altws_pkgarchive")
            except Exception:
                pass

            if self.pkgmerge and not \
                    ((self.debug and self.nondebug) or altws_pkgarchive):
                # We want a very specific format for the warning,
                # which build(1) logging will otherwise ruin, so print
                # that text ourselves before raising the exception.
                message = (
                    "\nOnly the following pkgmerge configurations are "
                    "supported by build(1),\nand specific bldrc settings "
                    "were not found:\n\n"
                    "* A debug/nondebug/sparc/i386 merge\n"
                    "    ('altws_pkgarchive' must be set, 'debug' and "
                    "'nondebug' must be 'true')\n\n"
                    "* A debug or nondebug sparc/i386 merge\n"
                    "    ('altws_pkgarchive' must be set)\n\n"
                    "* A debug and nondebug single-arch merge for local "
                    "packages\n"
                    "    ('debug' and 'nondebug' must be 'true')\n")
                print(message)
                raise OnBldEnvironmentException(
                    "Invalid bldrc configuration for pkgmerge.")

            if not self.debug and not self.nondebug:
                raise OnBldEnvironmentException(
                    "At least one of 'debug' and 'nondebug' must be set to "
                    "'true' in bldrc file.")

            # Verify that altws_keep if set, it is of the correct form
            altws_keep = None
            try:
                altws_keep = self.__cfg.get(
                    SECT_PATHS, "altws_keep")
            except Exception:
                pass
            if altws_keep and altws_keep != "<None>":
                try:
                    mach, path = altws_keep.split(":", 1)
                except ValueError:
                    raise OnBldEnvironmentException(
                        "'altws_keep' should be of the form "
                        "<mach>:<path to ws>/keep")
                if mach not in ["i386", "sparc"]:
                    raise OnBldEnvironmentException(
                        "the 'mach' component of 'altws_keep' must be "
                        "'i386' or 'sparc'")

        # Check if this workspace has "/ucb" somewhere in the pathname,
        # since ksh changes the /bin/echo behaviour accordingly and causes
        # strange and time-wasting build setup failures.
        if self.ws_path and "/ucb" in self.ws_path:
            raise OnBldEnvironmentException(
                "Cannot build workspace with '/ucb' in the pathname. "
                "It causes differences in how /bin/echo works, which leads "
                "to bad datafile formatting and hard-to-diagnose build setup "
                "failures.\n")

        # Some of these are environment purification, others are setting
        # defaults which we change later. This dictionary contains fixed
        # settings required for the ON build - we do not allow these to be
        # overridden by config files or existing environment variables.
        #
        # verify_bldrc_consistency() will check that any supplied bldrc file
        # has all of these keys set, though won't check for valid values.
        #

        lc_envvar_help = (
            "See http://pubs.opengroup.org/onlinepubs/7908799/xbd/envvar.html"
            "#tag_002_002"
            " this variable must never be changed from the default value "
            "'C' since the build depends on strict collation and locale "
            "settings.")

        src_subdir = "/usr/src"
        fwhooks_loc = "@WS@"

        if self.ws_path and os.path.exists(os.path.join(
                self.ws_path, FISH_STANDALONE_CANARY_PATH)):
            src_subdir = ""
            fwhooks_loc = "/ws/on-gate"

        # Set the starting value for PWD
        start_pwd = os.getcwd()

        # Environment settings that should never change and which are not
        # included in the bldrc file.
        self.__fixed_environment = [
            Env("LANG", "C", None, lc_envvar_help),
            Env("LC_COLLATE", "C", None, lc_envvar_help),
            Env("LC_CTYPE", "C", None, lc_envvar_help),
            Env("LC_MESSAGES", "C", None, lc_envvar_help),
            Env("LC_MONETARY", "C", None, lc_envvar_help),
            Env("LC_NUMERIC", "C", None, lc_envvar_help),
            Env("LC_TIME", "C", None, lc_envvar_help),
            Env("MAKE",
                "/usr/bin/env -u KEEP_STATE -u SUNPRO_DEPENDENCIES dmake",
                None,
                "Needed by make(1S) when the default make.rules file is not "
                "used.  Do not change this."),
            Env("MACH", MACH, None, "'uname -p' output for the current build "
                "machine. DO NOT override this."),
            Env("MACH64", MACH64, None, "The 64-bit architecture for the "
                "current machine. DO NOT override this."),
            Env("DMAKE_MODE", "parallel", None,
                "Determines the behaviour of dmake, set to either 'serial', "
                "'parallel', or 'distributed'. See dmake(1) for more "
                "information."),
            Env("DMAKE_OUTPUT_MODE", "TXT2", None, "Defines the "
                "format of the dmake output. See dmake(1) for examples."),
            Env("SUNW_NO_UPDATE_NOTIFY", "TRUE", None,
                "Magic variable to prevent the devpro compilers/teamware "
                "from sending mail back to devpro on every use."),
            Env("UT_NO_USAGE_TRACKING", "1", None,
                "Magic variable to prevent the devpro compilers/teamware "
                "from sending mail back to devpro on every use."),
            Env("POUND_SIGN", "#", None, "Used in the OS/Net build in "
                "Makefiles when we need a '#' character that would otherwise "
                "be interpreted as a comment, and to selectively comment-out "
                "sections from pkg(7) manifests. See further comments in "
                "usr/src/Makefile.master."),
            Env("STABS", "@TOOLROOT@/opt/onbld/bin/@MACH@/stabs", None,
                "The path to stabs."),
            Env("CTFSTABS", "@TOOLROOT@/opt/onbld/bin/@MACH@/ctfstabs", None,
                "The path to ctfstabs."),
            Env("GENOFFSETS", "@TOOLROOT@/opt/onbld/bin/genoffsets", None,
                "The path to genoffsets."),
            Env("CTFCONVERT", "@TOOLROOT@/opt/onbld/bin/@MACH@/ctfconvert",
                None, "The path to ctfconvert."),
            Env("CTFMERGE", "@TOOLROOT@/opt/onbld/bin/@MACH@/ctfmerge", None,
                "The path to ctfmerge."),
            Env("CTFDUMP", "@TOOLROOT@/opt/onbld/bin/@MACH@/ctfdump", None,
                "The path to ctfdump."),
            Env("CTFCHK", "@TOOLROOT@/opt/onbld/bin/ctfchk", None,
                "The path to ctfchk."),
            Env("ELFSIGN", "@TOOLROOT@/opt/onbld/bin/@MACH@/elfsign", None,
                "The path to elfsign or elfsigncmp. This value defaults to "
                "using elfsign from our tools area. If 'check.elfsigncmp' is "
                "enabled for the build, we use the elfsigncmp wrapper script "
                "instead."),
            Env("BOOTSIGN", "@TOOLROOT@/opt/onbld/bin/@MACH@/bootsign", None,
                "The path to bootsign."),
            Env("TOOLBUILD", "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@build.@MACH@",
                None,
                "The directory where build objects for tools only are "
                "created, prior to being linked and installed in 'toolroot'. "
                "This variable is ignored when 'bip' is set to 'true'. The "
                "supplied value is differentiated from 'build' in that it is "
                "NOT modified by build(1) depending on whether we're doing a "
                "debug or non-debug build."),
            Env("TOOLROOT",
                "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@_tools",
                None,
                "The location where the onbld tools are built. We always "
                "build non-debug tools regardless of whether we're doing a "
                "debug or non-debug OS/Net build."),
            Env("PYTHONNOUSERSITE", "true", None,
                "This prevents python(1) adding the user's site directory "
                "in ~/.local/lib/pythonx.y/site-packages to sys.path. Since "
                "there are many python processes run by the build, this can "
                "speed the build up significantly for users with remote "
                "NFS home directories."),
            Env("FW_PROTO_ROOT", None, None,
                "Used by the usr/fish build, this points to the proto area "
                "for the objects built from a matching usr/src repository. "
                "When 'fw_hooks' is set in the bldrc for a standalone "
                "usr/fish build, the usr/fish build derives 'fw_proto_root' "
                "from that."
                "\n\n"
                "When building in a combined usr/src + usr/fish "
                "workspace, FW_PROTO_ROOT is set to the value of $ROOT for "
                "usr/src."),
            Env("FW_STUB_ROOT", None, None,
                "The stub equivalent of 'fw_proto_root'. See also 'stubroot'"),
            Env("ONBLD_BRANCH_PREFIX", "", None,
                "Used by the build to indicate what flavor of workspace "
                "this build produces. This prefix is referred to in "
                "several variables in the bldrc file, so that we can build "
                "both flavors in the same workspace without namespace "
                "clashes for directories that the build writes to: $ROOT, "
                "$PKGARCHIVE, etc."
                "\n\n"
                "Note the term 'branch' is a misnomer, stemming (ha ha) from "
                "the time that the ON tree was physically branched into "
                "'server' and 'default' mercurial branches. When doing a "
                "'fish_is_present=yes' build, this value is set to '', "
                "when doing a 'fish_is_present=no build', this value is set "
                "to '%s'." % SERVER_PREFIX),
            Env("ONBLD_HG_BRANCH", "default", None,
                "Mercurial active branch name. Almost nothing in the build "
                "should rely on this, since it will obviously change "
                "depending on the release, the respin, or any feature "
                "branches that are created in a developer's private "
                "workspaces."
                "\n\n"
                "build(1) code that needs to adjust for Solaris 'server' vs. "
                "ZFSSA behaviour should use 'fish_is_present' instead."),
            Env("FISHROOT", None, None,
                "The proto area where the usr/fish files are written. "
                "When building in usr/fish, the builds write to $ROOT which "
                "gets set to the correct value. When building in usr/src, we "
                "record the fish proto area in this variable."
                "\n\n"
                "Nothing in usr/src should depend on the usr/fish proto area, "
                "since it gets populated after the 'install' task from "
                "usr/src, however in cases where we have previously "
                "built in this workspace, we use $FISHROOT to lookup any "
                "existing version.xml in order to set $AKV_BUILD correctly."),
            Env("ONBLD_PRESERVE_DC_DATASET", None, None,
                "Set to any value, this prevents the fish.destroy_dc, "
                "fish.*iso and fish.*images tasks from destroying the dc "
                "dataset, allowing developers to investigate the "
                "image-construction area. To use it, set it in the shell "
                "environment, and add it to the 'userenv' value of the "
                "bldrc file. See 'build explain userenv'."),
            Env("PARFAIT_LINK_LIMIT", "524288000", None,
                "Internal limit that Parfait uses to protect itself against "
                "massive linking tasks. genunix is already bigger than the "
                "default limit, so it is set here to 500M."),
            Env("PWD", start_pwd, None, "The PWD variable is crucial in "
                "navigating the Makefiles during the build. The expectation "
                "is that the shell in use will keep it up to date with chdir "
                "operations, but some shells like bash will only maintain a "
                "'local' variable called PWD when it is not present in the "
                "environment. This entry is present to ensure it is always "
                "in the environment.")
        ]

        # A dictionary that defines what build variables we set in the
        # environment. We want to retain exactly this order, so that we can
        # generate consistently ordered bldrc files using it. To do this, we
        # keep a list of tuples. When moving to Python 2.7 or later, this
        # can become an OrderedDict and we can use that OrderedDict when
        # creating ConfigParser instances, to simplify the code in
        # self.generate_bldrc
        self.__environment = {}
        self.__env_list = [
            # Paths
            Env("WS", self.ws_path, SECT_PATHS,
                "The full path to the top of the local ON mercurial "
                "repository. Workspace paths must not contain spaces."),

            Env("ONBLD_DATA", "@WS@/data", SECT_PATHS,
                "The full path to the directory where all build data should "
                "be written."),

            Env("ONBLD_KEEP", "@WS@/keep", SECT_PATHS,
                "The full path to the directory where all cross-build "
                "persistent content should be written, e.g. "
                "signing-material."),

            Env("BUILD_TOOLS", "/opt", SECT_PATHS,
                "This value points to the top level of a directory that "
                "contains all of the tools needed during the build which are "
                "not part of the WOS - the compilers, parfait, etc. If local "
                "tools are not installed, change this value to "
                "'/ws/on-tools'. Local tools are likely to result in "
                "faster builds, and are recommended.\n\n"
                "If setting this value, users should check that "
                "'parfait_home' is valid as the defaults reference this "
                "value."),

            Env("COMPILERPATH", "/opt/developerstudio12.6",
                SECT_PATHS, "The path used to find the compiler used for the "
                "build."),

            Env("SRC", "@WS@%s" % src_subdir, SECT_PATHS,
                "The top of the usr/src directory tree."),

            Env("ROOT", "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@",
                SECT_PATHS,
                "The top of the proto area for each build type. "
                "The supplied value is modified by build(1) depending on "
                "whether we're doing a debug or non-debug build, appending a "
                "'-nd' suffix to the value. See also the 'debug' and "
                "'nondebug' options."),

            Env("BUILD", "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@build.@MACH@",
                SECT_PATHS,
                "The directory where build objects are created, prior to "
                "being linked and installed in 'root'. This variable is "
                "ignored when 'bip' is set to 'true'. The supplied value is "
                "modified by build(1) depending on whether we're doing a "
                "debug or non-debug build, appending a '-nd' suffix to the "
                "value. See also the 'debug' and 'nondebug' options."),

            Env("NATIVEROOT",
                "@ONBLD_DATA@/proto/@ONBLD_BRANCH_PREFIX@root_@MACH@_native",
                SECT_PATHS, "The location where the native objects are built. "
                "Objects should be placed in this root directory if they are "
                "built specifically for use during the OS/Net build, and are "
                "not packaged for delivery. See $SRC/Makefile for comments "
                "related to the native build target."),

            Env("PKGARCHIVE",
                "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@pkg/@MACH@-debug",
                SECT_PATHS, "The default location where packages should be "
                "placed if built. If a non-debug build is performed, the "
                "@MACH@-debug component is replaced with @MACH-nondebug."
                "\n\n"
                "If set to the special value '<None>', packages are not "
                "published."
                "\n\n"
                "Note that for AK-Fish builds, this value is used "
                "with a '%s' prefix on the basename for AK packages. "
                "AK-Fish cpio archives are written to "
                "@PKGARCHIVE@/../archives/@MACH@-<flavor>" % FISH_PREFIX),

            Env("PATH", "@TOOLROOT@/opt/onbld/bin:"
                "@TOOLROOT@/opt/onbld/bin/@MACH@:@NATIVEROOT@/usr/bin:"
                "@PKG7_ROOT@/usr/bin:"
                "/usr/bin:/usr/sbin:@COMPILERPATH@/bin",
                SECT_PATHS, "The $PATH used while running the build."),

            Env("JAVA_ROOT", "/usr/jdk/instances/jdk1.8.0", SECT_PATHS,
                "The path used to find the java tools for OS/Net"),

            Env("LOGDIR", "@ONBLD_KEEP@/log.@MACH@", SECT_PATHS,
                "Log files written by build(1) will appear in this directory, "
                "which is created if it does not exist. The log directory "
                "should include @MACH@ to allow concurrent builds on "
                "different platforms."),

            Env("ALTWS_KEEP", None, SECT_PATHS,
                "The sibling workspace 'keep' directory that is being used to "
                "produce the packages referred to by 'altws_pkgarchive'. "
                "\n\n"
                "This is used during the 'pkgmerge' and 'findunref' tasks "
                "to determine whether to try to locate build(1) lock and "
                "state files in order to wait for specific build tasks in "
                "that workspace to complete."
                "\n\n"
                "If set to <None>, then no attempt is made to synchronise "
                "build tasks between workspaces, and 'altws_pkgarchive' is "
                "assumed to already contain built packages."
                "\n\n"
                "If set, this value should be of the form:"
                "\n\n"
                "<uname -p output for other arch>:<path to ws keep dir>"
                "\n\n"
                "We recommend that 'pkgsurf' and 'pkgmerge' are both set to "
                "'false' when building the alternate workspace, and set to "
                "true when building this workspace. Both workspaces do not "
                "need complementary 'altws_keep' strings. "
                "See also 'pkgmerge_altws_timeout'."
                "\n\n"
                "An example value when building this on i386 is:\n"
                "sparc:/net/eleven/builds/ongk/workspace/myws/keep"),

            Env("ALTWS_PKGARCHIVE", None, SECT_PATHS,
                "The packages produced for the other architecture that are "
                "used during the 'pkgmerge' task. That is, if this workspace "
                "is building 'i386', the alternative packages will be for "
                "'sparc', and visa versa."
                "\n\n"
                "If set to the value <None> then the 'pkgmerge' task will "
                "only consider 'debug' and 'nondebug' variants produced "
                "during the build that were written to 'pkgarchive' in the "
                "current workspace."
                "\n\n"
                "The value will be adjusted to find debug, or non-debug "
                "packages when necessary."
                "\n\n"
                "For example:\n"
                "/net/eleven/builds/ongk/workspace/myws/data/pkg/sparc-debug"),

            Env("FW_HOOKS", fwhooks_loc, SECT_PATHS,
                "This points to a workspace containing $ROOT and $STUBROOT "
                "binaries, used by the usr/fish gate to reference built ON "
                "commands and libraries. We assume standard ONBLD_DATA and "
                "ROOT values."
                "\n\n"
                "If the current workspace is not expected to produce a full "
                "ON build, this value should be set to a path that contains "
                "one. For example:\n"
                "/ws/on-gate"),

            Env("FW_PKG_PROTO", "/", SECT_PATHS,
                "This points to a pkg(7) image containing an install of at "
                "least pkg:/developer/opensolaris/osnet. For most users, the "
                "root directory of the build machine will suffice."
                "\n\n"
                "When constructing images with 'fish.image', 'fish.text_iso', "
                "etc. build(1) will consult the pkg(7) publishers in this "
                "image to obtain URIs for the package repositories that "
                "should be used to construct the various AK images. pkg(7) "
                "facets that are set in this image are also propagated to the "
                "AK images being built, other than 'facet.locale.*' facets, "
                "which are hardcoded by 'dc_text|ai_*.xml'"
                "\n\n"
                "For ON packages, build(1) will prefer local 'nightly' and "
                "'extra' origins from @PKGARCHIVE@ if present in the current "
                "workspace or from the workspace pointed to by @FW_HOOKS@, "
                "falling back to the 'nightly' and 'extra' origins configured "
                "in @FW_PKG_PROTO@."),

            # We have this entry just so that we can document the
            # environment variable: obviously a bldrc file containing a
            # key that sets a path to itself is a little weird, so we
            # use a 'None' section name so that it doesn't need to appear in
            # the file.
            Env("BLDRC", "", None,
                "A full path to the bldrc file used to supply build "
                "configuration. For build(1), any bldrc file supplied on the "
                "command line will override this environment variable.\n\nA "
                "new bldrc file can be generated by running "
                "'build generate'."),

            # Overrides for some build subcommands
            Env("ONBLD_PKGSERVE_PORT", None, None,
                "If set, the 'build pkgserve' task will use this port for "
                "the pkg.depot Apache instance it starts. This variable is "
                "inherited from the user environment and is not set in the "
                "bldrc file."),

            Env("ONBLD_SERVE_PORT", None, None,
                "If set, the 'build pkgserve' task will use this port for "
                "the pkg.depot Apache instance it starts. This variable is "
                "inherited from the user environment and is not set in the "
                "bldrc file."),

            # General build options
            Env("DEBUG", "true", SECT_BUILDOPTIONS, "If set to 'true', the "
                "build will produce debug binaries. It can be overridden by "
                "the -D flag to 'build -i'.  Note that 'build -i' will only "
                "setup the environment for *either* debug or non-debug, not "
                "both."),

            Env("NONDEBUG", "true", SECT_BUILDOPTIONS, "If set to 'true', the "
                "build will produce non-debug binaries. It can be overridden "
                "by the -d flag to 'build -i'.  Note that 'build -i' will "
                "only setup the environment for *either* debug or non-debug, "
                "not both.  When non-debug builds are being run, some other "
                "build variables are modified automatically, eg. $ROOT, to "
                "append '-nd' to the path."),

            Env("INCREMENTAL", "false", SECT_BUILDOPTIONS, "If set to 'true', "
                "the build will be as incremental as possible - no clobber "
                "phase will be run."),

            Env("CHECK", "true", SECT_BUILDOPTIONS, "If set to 'true', "
                "build(1) will perform a series of checks having produced the "
                "normal build output. The checks that are performed will "
                "vary, depending on the build type (debug or nondebug), "
                "the values in the [checkoverrides] section of the bldrc "
                "file and the command line flags used."
                "\n\n"
                "Failing build tasks generally cause build(1) to exit with a "
                "non-zero exit code. Some control is allowed over which build "
                "checks can cause the build to fail, and which are non-fatal "
                "using the 'check.<name>.must_pass' keys in the "
                "[checkoverrides] section. (use 'build generate -O' to "
                "include that section in the bldrc file)"),

            Env("PKGMERGE", "false", SECT_BUILDOPTIONS,
                "Whether the 'nightly' or 'qnightly' tasks should attempt to "
                "produce a merged repository using the 'pkgmerge' task "
                "during the build."
                "\n\n"
                "Merging is done for debug and non-debug variants if the "
                "build has set 'debug' and 'nondebug'. Merging is done for "
                "'arch' variants if 'altws_pkgarchive' is specified."
                "\n\n"
                "If 'debug' and 'nondebug' and 'altws_pkgarchive' are set, "
                "a four-way merge is performed (debug/nondebug/i386/sparc)"
                "\n\n"
                "To coordinate the merging of two separate builds that "
                "may be running in parallel, see 'altws_keep'."),

            Env("PKGSURF", "true", SECT_BUILDOPTIONS,
                "Whether the 'nightly' or 'qnightly' tasks should attempt to "
                "reversion published repositories using the 'pkgsurf' task."
                "\n\n"
                "See 'pkgsurfarchive'. If 'pkgmerge' is 'true', then "
                "resurfacing happens after pkgmerge and is done on the "
                "'merged' package repositories, otherwise it happens "
                "after packages are published for each build flavor."
                "\n\n"
                "Note that independently reversioning two workspaces of "
                "different architectures can result in those workspaces "
                "having repositories with different versions, which will "
                "cause a subsequent pkgmerge to fail. If you will need to "
                "merge repositories, use the 'pkgmerge' task."),

            Env("ZIPFILES", "false", SECT_BUILDOPTIONS,
                "Whether the 'nightly' or 'qnightly' tasks should attempt to "
                "produce a zipfile of the merged package repo contents. If "
                "pkgmerge is not set to true, this setting will be ignored."),

            Env("BIP", None, SECT_BUILDOPTIONS, "If set to 'true', this "
                "causes the build to 'build in place'. That is, the build "
                "will store compiled objects beneath $SRC adjacent to the "
                "source being compiled, before they are copied to $ROOT. "
                "Otherwise, objects will be stored in a separate directory "
                "structure in $WS, but outside $SRC. See 'build'."),

            Env("CW_NO_SHADOW", None, SECT_BUILDOPTIONS,
                "This controls whether we perform shadow compilation. Setting "
                "this to anything other than <None> will prevent cw(1) from "
                "doing shadow compilation.\n\n"
                "With shadow compilation, for each invocation of cw(1), "
                "objects are built first with the primary compiler, with the "
                "results eventually making it to the proto area. Then the "
                "shadow compiler is run on the sources, but the resulting "
                "objects are discarded.\n\n"
                "The shadow compiler is set to gcc if the primary is cc. The "
                "shadow compiler is cc if the primary is gcc. Disabling "
                "shadow compilation will improve the build time at the cost "
                "of not catching errors produced by the secondary compiler. "
                "See '__gnuc'."),

            Env("DMAKE_MAX_JOBS", None, SECT_BUILDOPTIONS,
                "Do more work in parallel if possible. This sets the maximum "
                "number of processes that dmake will spawn at a time when "
                "building. See dmake(1) for more details. If set to '<None>' "
                "or left unset, $HOME/.make.machines is consulted for the "
                "value for this build machine. If set to '<None>' and no "
                ".make.machines entry is present, the maximum number of jobs "
                "is set to 16 for systems with 4 or more cpus, or 4 for "
                "systems with less than 4 cpus."),

            Env("SNAPSHOTS", "false", SECT_BUILDOPTIONS, "If set to 'true', "
                "build(1) will try to take a ZFS snapshot when starting each "
                "build task if $WS resides in a separate "
                "dataset. The snapshot is named according to the task being "
                "run. Any snapshot of that name which already exists will be "
                "renamed with a datestamp appended.\n\nSnapshots are "
                "taken using a mkdir(2) inside .zfs/snapshot, which works "
                "across NFS. The user running the build will need "
                "appropriate ZFS delegations. See zfs_allow(1).\n\nIf a "
                "snapshot cannot be taken, the reason will be recorded in the "
                "log files, but is not considered a build error.\n\n"
                "The following commands configure ZFS delegations with "
                "simple permission sets to allow build snapshots, as well as "
                "more complex permissions to allow other dataset operations:"
                "\n\n"
                "# zfs allow -s @build_perms \\\n"
                "    snapshot,rename,create,destroy,mount,diff,send,receive \\"
                "\n"
                "    rpool/builds\n"
                "# zfs allow -s @build_perms_ext \\\n"
                "    rollback,clone,userprop,hold,release,promote,share \\\n"
                "    rpool/builds\n"
                "\n"
                "# zfs allow larry @build_perms     rpool/builds/larry\n"
                "# zfs allow   moe @build_perms     rpool/builds/moe\n"
                "# zfs allow curly @build_perms     rpool/builds/curly\n"
                "# zfs allow curly @build_perms_ext rpool/builds/curly\n"),

            Env("MAILTO", "", SECT_BUILDOPTIONS, "As part of 'nightly', the "
                "build will send mail notifications upon completion. If this "
                "value is set to '<None>', no mail notifications will be "
                "sent. If this value is set to an empty string (the default), "
                "then the value is determined automatically from ~/.hgrc "
                "using the '[ui] username' or '[email] from' properties. "
                "See hgrc(5). If that fails, we fall back to using "
                "ldaplist(1).\n\n"
                "Otherwise, the value can be set to a comma-separated list of "
                "email addresses. Addresses that include display names, such "
                "as:\"Joe Bloggs <joe@oracle.com>\"\n are also supported. "
                "See Python's smtplib SMTP.sendmail() documentation and "
                "RFC 822."),

            Env("USERENV", "", SECT_BUILDOPTIONS,
                "A series of semicolon-separated environment values to "
                "inherit from the user's shell, eg. MAIL;COLORTERM. Many of "
                "the sensible variables are already built-in, so only use "
                "this if the environment produced by build(1) isn't "
                "sufficient."),

            Env("USERUNSETENV", "", SECT_BUILDOPTIONS,
                "A series of semicolon-separated environment variables to "
                "explicitly *unset* in the build environment. Use this with "
                "extreme caution as it has the ability to break the build in "
                "interesting ways. See 'userenv'."),

            Env("USER", "", SECT_BUILDOPTIONS, "If not set, user is "
                "determined automatically from $LOGNAME or $USER."),

            Env("BUILD_PROJECT", "", SECT_BUILDOPTIONS, "Use this project "
                "name when running the build. See the -p option to newtask(1) "
                "and /etc/project"),

            Env("UPDATE_PARENT_WS", "false", SECT_BUILDOPTIONS, "If set to "
                "'true', then as part of 'nightly' or when running the "
                "'update_parent_proto' build task, the proto areas from this "
                "workspace will be copied to 'parent_root' and "
                "'parent_toolroot' using rsync(1)."),

            Env("AKV_GATE_ID", "", SECT_BUILDOPTIONS, "The gate "
                "identifier used by storage appliance builds. This is "
                "set to 'ak' for gatekeeper builds. If set to the empty "
                "string, we default to a sanitized value of the form "
                "@USER@_@BASEWSDIR@. Note that distro_const will truncate "
                "this value if longer than 31 characters."
                "See $SRC/Makefile.versions."),

            # Settings related to code checks
            Env("PARFAIT_HOME", "@BUILD_TOOLS@/parfait/parfait-tools-4.0",
                SECT_CODECHECKS, "The path used to find parfait, a static "
                "code analysis tool."),

            Env("PARFAIT_TYPE", "baseline", SECT_CODECHECKS, "Determines how "
                "Parfait static code analysis is handled during the build. "
                "The work is performed in two phases: building and analysis."
                "\n\n"
                "During a 'dmake install', or when using the 'install' "
                "subcommand, when Parfait is enabled, the compilers are "
                "replaced with parfait wrappers. These run on the source "
                "code, producing bytecode as well as the normal build output. "
                "When performing Parfait analysis, as part of the 'nightly', "
                "'check' or 'parfait' subcommands, that bytecode is analysed."
                "\n\n"
                "The results of the analysis are then either compared to a "
                "known baseline, or uploaded to the Parfait server to form a "
                "new baseline.\n\n"
                "The following values can be set for PARFAIT_TYPE:\n\n"
                "baseline       Perform a build or baseline analysis\n"
                "upload         Perform a build or baseline upload"
                "\n\n"
                "Note that 'check.parfait' should be set to 'false' to "
                "disable parfait."),

            Env("FORTIFY_VERS",
                "19.10",
                SECT_CODECHECKS,
                "Version of the Fortify code analysis tool."),

            Env("FORTIFY_HOME",
                "@BUILD_TOOLS@/fortify/fortify-@FORTIFY_VERS@",
                SECT_CODECHECKS, "The path used to find Fortify, a static "
                "code analysis tool."),

            Env("ELF_DATA_BASELINE_DIR", "/ws/on-gate/public/"
                "ELF-data-baseline.@MACH@", SECT_CODECHECKS, "When performing "
                "build checks and when 'elf_data_baseline_dir' is set to "
                "anything other than <None>, the ELF interface description "
                "file resulting from the build is compared to that from the "
                "specified directory. This ensures that our object versioning "
                "evolves in a backward compatible manner.\n\n"
                ""
                "You should not need to change this unless you wish to use "
                "locally cached baseline files. If you use this, it must be "
                "local (or nfs): nightly cannot copy over ssh or http.\n\n"
                ""
                "Gatekeepers: If a change to 'elf_data_baseline_dir' is due "
                "to a change in the version of Solaris being used as the "
                "baseline, you should also adjust 'elf_data_baseline_type'. "
                "This can be necessary when a new Solaris marketing gate is "
                "created, or when the gate for an older version of Solaris is "
                "shut down. Non-gatekeepers should not need to change "
                "'elf_data_baseline_dir'."),

            Env("ELF_DATA_BASELINE_AGE", "0", SECT_CODECHECKS, "If "
                "'elf_data_baseline_age' is non-0, and the age of the "
                "baseline exceeds the specified value, in days, a warning is "
                "written to the mail message and to the log file. Developers "
                "should leave this set to 0. Gatekeepers should set it to a "
                "reasonable value in order to ensure that the baseline files "
                "do not become stale due to a failure of the other gate to "
                "produce current files."),

            Env("ELF_DATA_BASELINE_TYPE", "DIRECT_PREDECESSOR",
                SECT_CODECHECKS, "'elf_data_baseline_type' describes how the "
                "baseline is related to the workspace being built. It must be "
                "set to one of the following values:\n\n"
                "	DIRECT_PREDECESSOR  Baseline is one release behind\n"
                "	DIRECT_SUCCESSOR    Baseline is one release ahead\n"
                "	PREDECESSOR         Baseline is >1 release behind\n"
                "	SUCCESSOR           Baseline is >1 release ahead\n\n"
                "For more information on the ELF checks performed, see "
                "find_elf(1), interface_check(1), interface_cmp(1) and "
                "check_rtime(1)"),

            Env("PARENT_WS", "/ws/on-gate", SECT_CODECHECKS, "The workspace "
                "parent of this workspace, used by the 'update_parent' task, "
                "or by nightly when the 'update_parent_ws' build option is "
                "being used. The value set here is intended for use as a "
                "convenient macro when defining the 'parent_root' and "
                "'parent_toolroot' options, which are used by the "
                "'update_parent' tasks.\n\n"
                "This value, if set, is also used to specify the area where "
                "certain ELF checks upload information that can then be "
                "referenced by other child workspaces of the parent."),

            Env("PARENT_ROOT",
                "@PARENT_WS@/data/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@",
                SECT_CODECHECKS, "When running the 'update_parent' build "
                "task, or by 'nightly' when the 'update_parent_ws' build "
                "option is set, then the proto area(s) from this workspace "
                "will be copied to the parent proto area after a successful "
                "build.\n"
                "\n"
                "The value here must be an absolute path to the parent proto "
                "area. ssh(1) URIs are NOT supported."),

            Env("PARENT_NATIVEROOT",
                "@PARENT_WS@/data/@ONBLD_BRANCH_PREFIX@proto/"
                "root_@MACH@_native",
                SECT_CODECHECKS, "Placeholder for future work."),

            Env("PARENT_TOOLROOT",
                "@PARENT_WS@/data/"  # joined line
                "@ONBLD_BRANCH_PREFIX@proto/root_@MACH@_tools",
                SECT_CODECHECKS, "This specifies the location of the parent "
                "tools proto area, used by the 'update_parent' build task or "
                "by 'nightly' when the 'update_parent_ws' build option is "
                "set. See also 'parent_root'.\n\n"
                "The value here must be an an absolute path to the parent "
                "tools proto area. ssh(1) URIs are NOT supported."),

            Env("PARENT_ELF_DATA_DIR", None, SECT_CODECHECKS,
                "If 'update_parent_ws' is set to 'true', then "
                "'parent_elf_data_dir specifies an alternate directory to "
                "upload the ELF data generated as part of the 'check.elf' "
                "build task (this build task is also performed by the 'check' "
                " and 'nightly' build tasks.)"
                "\n\n"
                "If 'parent_elf_data_dir' is not specied, then the default "
                "location is the value of 'parent_ws'/keep."),

            Env("REF_PROTO_LIST",
                "@PARENT_WS@/public/@ONBLD_BRANCH_PREFIX@proto_list_@MACH@",
                SECT_CODECHECKS, "A list of files that are present in a "
                "reference proto area. This list is consulted during the "
                "'check.protocmp' task to find differences between the files "
                "delivered by this workspaces versus those delivered by the "
                "parent or other reference workspace."
                "\n"
                "See $TOOLROOT/opt/onbld/bin/protocmp.terse."),

            Env("DATASPACE_PROFILING", None, SECT_CODECHECKS, "Setting this "
                "to 'yes' enables support for dataspace profiling of the "
                "kernel. See er_kernel(1)."),


            # Settings related to building partial tree builds, or producing
            # components needed for partial tree builds (eg. a missing
            # usr/closed)

            Env("CLOSED_IS_PRESENT", None, SECT_PARTIAL,
                "This controls whether to build the closed source. If "
                "set to '<None>', it will be set according to whether the "
                "usr/closed source tree is present. closed_is_present=no "
                "means we do not build the closed sources, "
                "closed_is_present=yes means we do."),

            Env("ON_CLOSED_BINS", None, SECT_PARTIAL, "If doing a build "
                "where the usr/closed source tree is not present, this is the "
                "location of those encumbered binaries. A common value "
                "for this key is:\n\n"
                "on_closed_bins = @WS@/closed\n\n"
                "where that directory contains the root_$MACH[-nd] "
                "directories extracted from the closed tarballs:\n\n"
                "    on-closed-bins.i386.tar.bz2\n"
                "    on-closed-bins-nd.i386.tar.bz2\n\n"
                "The closed tarballs are created by the 'nightly' and "
                "'closed_tarball' build(1) subcommands when performing a "
                "'tonic' build."),

            Env("MAN_IS_PRESENT", None, SECT_PARTIAL,
                "This controls whether to include man pages in the build. If "
                "set to '<None>', it will be set according to whether the "
                "usr/man source tree is present. man_is_present=no means "
                "we do not include man pages."),

            Env("FISH_IS_PRESENT", None, SECT_PARTIAL,
                "This controls whether to include build tasks in 'nightly' "
                "that produce AK Fish content and whether we build the AKFORK "
                "parts of $SRC. If set to '<None>', it will be set according "
                "to whether the usr/fish source tree is present."
                "'fish_is_present=no' means we do not build AK Fish content "
                "and instead build the 'server' flavor of ON."),

            Env("TONICBUILD", None, SECT_PARTIAL,
                "When 'tonicbuild' is set to anything other than <None>, the "
                "build will produce tarballs containing binaries resulting "
                "from the compilation of usr/closed. The binaries come from "
                "the closed proto area, $CLOSEDROOT (automatically derived "
                "from the value of 'root' in the bldrc file).\n\n"
                "The closed tarball allows other users to perform a build "
                "of that gate when usr/closed isn't present, and can be "
                "configured by the 'on_closed_bins' bldrc key.\n\n"
                "A tonic build causes build(1) to run the 'closed_tarball' "
                "subcommand as part of the 'nightly' subcommand."),


            #
            # Settings we expect no people will ever change, so we don't
            # need them to appear in the bldrc file. These are essentially
            # for gatekeeper use only.
            #
            Env("BASEWSDIR", None, SECT_GKOVERRIDES, None),

            Env("CSCOPE_CONF_PATH", "@SRC@/tools/etc/build_cscope.conf",
                SECT_GKOVERRIDES, "The config file used by the 'xref' task to "
                "build cscope databases."),

            Env("DEV_CM",
                "\"@(#)SunOS Internal Development: @USER@ @BUILD_DATE@ "
                "[@BASEWSDIR@]\"",
                SECT_GKOVERRIDES, "The string used for the SunOS banner "
                "message. See Makefile.master for more details. If not set, "
                "dev_cm is determined automatically."),

            Env("CTF_GENUNIX_LABEL", None, SECT_GKOVERRIDES, "The CTF "
                "label used by ctfmerge during genunix build."),

            Env("CTF_MODULE_LABEL", None, SECT_GKOVERRIDES, "The CTF "
                "label used by ctfmerge during kernel module building."),

            Env("PKG7_ROOT", "", SECT_GKOVERRIDES, "Where to find "
                "pkg(7) tools; usually the build system, but may be "
                "overridden to test changes or work around problems."),

            Env("PYTHONPATH", "@PKG7_ROOT@/usr/lib/python2.7/vendor-packages",
                SECT_GKOVERRIDES, "Where to find pkg(7) python modules "
                "corresponding to tools in 'pkg7_root'. Note that we only "
                "set this value in the environment if PKG7_ROOT is set to "
                "something other than the empty string, since PYTHONPATH set "
                "to 2.7 packages will upset Python 3.x scripts that don't "
                "use the python(1) -E argument, or similar."),

            Env("PKGPUBLISHER_REDIST", None, SECT_GKOVERRIDES, "Where we "
                "publish redistributable packages, set to 'on-redist' by "
                "default."),

            Env("PKGPUBLISHER_NONREDIST", None, SECT_GKOVERRIDES, "The "
                "publisher name for non-redistributable packages, set to "
                "'on-extra' by default."),

            Env("RELEASE", None, SECT_GKOVERRIDES, "The release "
                "number, set in Makefile.master to '5.11'"),

            Env("RELEASE_DATE", None, SECT_GKOVERRIDES, "The release "
                "date, set in Makefile.master to eg. 'July 2012'."),

            Env("VERSION", "@BASEWSDIR@", SECT_GKOVERRIDES,
                "Version affects customer-visible version info for mcs, motd, "
                "and uname and boot messages, especially if genunix is being "
                "delivered."),

            Env("BUILD_DATE", None, SECT_GKOVERRIDES, "If not set, build_date "
                "is determined automatically. See also 'iso8601_build_date'."),

            Env("ISO8601_BUILD_DATE", None, SECT_GKOVERRIDES,
                "If not set, iso8601_build_date is determined automatically. "
                "See also 'build_date'."),

            Env("SPRO_VROOT", "@COMPILERPATH@", SECT_GKOVERRIDES,
                "One of the many variables that influence where we source the "
                "compiler from. It should probably be left set to "
                "@COMPILERPATH@"),

            Env("STUBROOT", "@ROOT@_stub", SECT_GKOVERRIDES,
                "Define where stub objects are built. Gatelings should "
                "never need to override this. More background on STUBROOT "
                "and the build's use of them is at 15689528 (changeset "
                "e0f109815da5)  The supplied value is modified by build(1) "
                "depending on whether we're doing a debug or non-debug build, "
                "appending a '-nd' suffix to the value."),

            Env("STUBENVCPPFLAGS1", "-I@STUBROOT@/usr/include",
                SECT_GKOVERRIDES, "cpp flags for the stub object build."),

            Env("STUBENVLDLIBS1", "-L@STUBROOT@/lib -L@STUBROOT@/usr/lib",
                SECT_GKOVERRIDES, "ld flags for the stub object build."),

            Env("STUBENVLDLIBS2", "", SECT_GKOVERRIDES, "Additional ld flags "
                "for the stub object build."),

            Env("STUBENVLDLIBS3", "", SECT_GKOVERRIDES, "Yet more ld flags "
                "for the stub object build."),

            Env("ENVCPPFLAGS1", "-I@ROOT@/usr/include",
                SECT_GKOVERRIDES, "cpp flags used for the build"),

            Env("ENVCPPFLAGS2", "", SECT_GKOVERRIDES, "Additional cpp flags "
                "used for the build."),

            Env("ENVCPPFLAGS3", "", SECT_GKOVERRIDES, "Yet more cpp flags "
                "used for the build."),

            Env("ENVCPPFLAGS4", "", SECT_GKOVERRIDES, "Haven't you used "
                "enough cpp flags already? No? Ok then, here's some more."),

            Env("ENVLDLIBS1", "-L@ROOT@/lib -L@ROOT@/usr/lib",
                SECT_GKOVERRIDES, "ld flags used for the build."),

            Env("ENVLDLIBS2", "", SECT_GKOVERRIDES, "Additional ld flags used "
                "for the build."),

            Env("ENVLDLIBS3", "", SECT_GKOVERRIDES, "Yet more ld flags used "
                "for the build."),

            Env("RELEASE_TYPE", "developer", SECT_GKOVERRIDES, "This value "
                "governs whether this is an internal build for developers, or "
                "an externally-visible build such as an IDR, SRU, or a public "
                "release such as a beta or GA release."),

            Env("MAKEFLAGS", "ekr", SECT_GKOVERRIDES,
                "We use 'k' to have 'make' to do as much as it can, in case "
                "there's more than one problem. This is especially important "
                "with the gate, since multiple unrelated broken things can be "
                "integrated. We use 'e' to inherit variables from the shell "
                "environment, which works when build(1) is used to define the "
                "environment, but should be used with caution elsewhere. "
                "We use 'r' to prevent use of the system make.rules file.  "
                "See make(1S) for more information on MAKEFLAGS."),

            Env("__GNUC", None, SECT_GKOVERRIDES,
                "Setting to any value other than '<None>' causes us to use "
                "the GNU tool-chain when building instead of the compilers "
                "in 'compilerpath'. See also comments in Makefile.master."),

            Env("GCC_ROOT", "/usr/gcc/10", SECT_GKOVERRIDES,
                "Points to the location of the GNU compilers (gcc, g++)."),

            Env("GNU_ROOT", "/usr/gnu", SECT_GKOVERRIDES,
                "Points to the location of the gnu toolchain."),

            Env("RSYNC_IDENT_USER", None, SECT_GKOVERRIDES,
                "The username used to login to the remote machine when "
                "running rsync(1) by the 'update_parent' task, or by "
                "'nightly' when the 'update_parent_ws' is set. If this "
                "value is not set, we default to using the username defined "
                "in 'user'."),

            Env("RSYNC_IDENT_KEY", None, SECT_GKOVERRIDES,
                "The ssh identity file passed to 'ssh -i' when running "
                "rsync(1) as part of the 'update_parent' task, which is also "
                "invoked by 'nightly' when 'update_parent_ws' is set."),

            Env("LOG_FORMAT", None, SECT_GKOVERRIDES,
                "Setting this to 'notimestamp' causes plaintext log files "
                "produced by build(1) to omit the timestamp. HTML log files "
                "will still include the timestamp. Setting it to 'normal' "
                "will include the timestamp, and setting to '<None>' will use "
                "the default."),

            Env("USE_INTERNAL_CERT", None, SECT_GKOVERRIDES,
                "Setting this to an empty string causes the "
                "build to enable code and packaging for developer testing "
                "with the 'INTERNAL USE ONLY' certs being honored "
                "\n\n"
                "Without also setting 'use_internal_cert', users booting "
                "the resulting bits would see elfsign verification warnings "
                "during boot."),

            Env("CORES_PATTERN", "@ONBLD_KEEP@/cores/core.%f.%m.%p@%t",
                SECT_GKOVERRIDES,
                "This value specifies the pattern used by coreadm(1) when "
                "creating the build environment. By default, cores are placed "
                "in $WS/keep/cores. To prevent build(1) from changing the "
                "default cores location, this value can be set to '<None>'."),

            Env("ARCHIVE_BUILD", "false", SECT_GKOVERRIDES,
                "Enables build archival. If set to 'true', an archive will be "
                "created using 'buildid_thisbuild' as an identifier. If "
                "unset, or set to any value other than 'true', no such "
                "archive will be created."
                "\n\n"
                "Build archives are stored in 'archive_to_dir', which is "
                "usually derived from 'archive_to_ws'."
                "\n\n"
                "A build archive will contain the repositories "
                "from 'pkgarchive', the kernel deliverables from 'root', and "
                "the intermediate build objects from selected subdirectories "
                "of 'build'."
                "\n\n"
                "A build archive may subsequently be used as the baseline "
                "build for splice generation."
                "\n\n"
                "See also 'build explain generate_splice' and 'build explain "
                "buildid_thisbuild'"),

            Env("GENERATE_SPLICE", "false", SECT_GKOVERRIDES,
                "Enables splice generation. If set to 'true' a splice build "
                "will be performed. If unset, or set to any value other than "
                "'true', a normal build will be performed."
                "\n\n"
                "A splice build differs from a regular build in the following "
                "ways:"
                "\n\n"
                "During a 'dmake install' or when using the 'install' "
                "subcommand, spliceable object files (currently "
                "limited to those that are part of kernel modules) will be "
                "compared to those from the archived build identified by "
                "'buildid_baseline'. Any differences detected will trigger "
                "creation of a splice module identified by "
                "'buildid_thisbuild'."
                "\n\n"
                "Baseline build archives are searched for in "
                "'archive_from_dir', which is usually derived from "
                "'archive_to_ws'."
                "\n\n"
                "During a 'dmake install' in $SRC/pkg, or when using the "
                "'packages' subcommand, a splice dependency module will "
                "be created for 'buildid_thisbuild'. This module and all "
                "splice modules generated during the install phase will "
                "be packaged together in pkg://system/osnet-splice."
                "\n\n"
                "During a 'dmake install' in $SRC/pkg, or when using the "
                "'packages' subcommand, when publishing packages, deliverable "
                "objects will be searched for in the archived build specified "
                "by 'buildid_baseline' first, and $ROOT second."
                "When setting 'generate_splice', it is strongly recommended "
                "to also set 'archive_build'."
                "\n\n"
                "See also 'build explain archive_build', 'build explain "
                "buildid_thisbuild', and 'build explain buildid_baseline'."),

            Env("TEST_SPLICE", "true", SECT_GKOVERRIDES,
                "If 'test_splice' is set to 'true', and both 'archive_build' "
                "and 'generate_splice' are false, the uts build will "
                "create three different versions of the splicetest kmod. "
                "One built unmodified (the baseline), one built with "
                "-DKSPLICE_TEST (the first splice), and one built with "
                "-DKSPLICE_TEST and -DKSPLICE_STACKED_TEST (the second, "
                "or stacked, splice). A sanity check on the ksplice "
                "generation tools is done using these different versions "
                "of the splicetest kmod. If 'test_splice' is not 'true', "
                "or if either 'archive_build' or 'generate_splice' is 'true', "
                "only the baseline version of the splicetest kmod is "
                "built and the ksplice tools tests are not performed."),

            Env("BUILDID_BASELINE", "0", SECT_GKOVERRIDES,
                "The Ksplice ID of the basis for comparison for 'build' and "
                "'pkgarchive' lookup. This value should be a setting of "
                "'buildid_thisbuild' from a previous build."),

            Env("BUILDID_THISBUILD", "00000000", SECT_GKOVERRIDES,
                "A build identifier. If 'archive_build'='true', this will be "
                "used to tag the resulting build products. For a splice "
                "build, as indicated by 'generate_splice'='true', this should "
                "be the Ksplice ID to be generated."),

            Env("ARCHIVE_FROM_WS", "@WS@", SECT_GKOVERRIDES,
                "The root of the workspace containing archived builds "
                "for use during splice builds."),

            Env("ARCHIVE_FROM_DIR",
                "@ARCHIVE_FROM_WS@/data/"
                "@ONBLD_BRANCH_PREFIX@ksplice-@BUILDID_BASELINE@",
                SECT_GKOVERRIDES,
                "The baseline (reference) archive directory for 'build', "
                "'root', and 'pkgarchive' storage."),

            Env("KSPLICE_BASELINE_BUILD",
                "@ARCHIVE_FROM_DIR@/@ONBLD_BRANCH_PREFIX@build.@MACH@",
                SECT_GKOVERRIDES,
                "The directory where build objects forming the basis for "
                "comparison are found when 'generate_splice'='true'."),

            Env("KSPLICE_BASELINE_PKGARCHIVE",
                "@ARCHIVE_FROM_DIR@/@ONBLD_BRANCH_PREFIX@pkg/@MACH@-nondebug",
                SECT_GKOVERRIDES,
                "The directory where reference repositories for accumulating "
                "splice packages and objects can be found. This is the "
                "'ksplice_splice_pkgarchive' directory from a previous build "
                "of the same workspace."),

            Env("KSPLICE_BASELINE_ROOT",
                "@ARCHIVE_FROM_DIR@/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@",
                SECT_GKOVERRIDES,
                "The directory where the 'root' archive from a previous "
                "(baseline) build can be found."),

            Env("KSPLICE_BASELINE_KSPLICEROOT",
                "@ARCHIVE_FROM_DIR@/@ONBLD_BRANCH_PREFIX@proto/"
                "root_@MACH@_ksplice_@BUILDID_BASELINE@",
                SECT_GKOVERRIDES,
                "The directory where 'kspliceroot' from a previous (baseline) "
                "build can be found."),

            Env("ARCHIVE_TO_WS", "@WS@", SECT_GKOVERRIDES,
                "The root of the workspace into which builds "
                "should be archived when 'archive_build' is 'true'."),

            Env("ARCHIVE_TO_DIR",
                "@ARCHIVE_TO_WS@/data/"
                "@ONBLD_BRANCH_PREFIX@ksplice-@BUILDID_THISBUILD@",
                SECT_GKOVERRIDES,
                "The archival directory for 'build', 'root', and "
                "'pkgarchive' storage."),

            Env("KSPLICE_SPLICE_BUILD",
                "@ARCHIVE_TO_DIR@/@ONBLD_BRANCH_PREFIX@build.@MACH@",
                SECT_GKOVERRIDES,
                "The directory where build objects from this build should be "
                "archived."),

            Env("KSPLICE_SPLICE_PKGARCHIVE",
                "@ARCHIVE_TO_DIR@/@ONBLD_BRANCH_PREFIX@pkg/@MACH@-nondebug",
                SECT_GKOVERRIDES, "The directory where package repositories "
                "from this build should be archived."),

            Env("KSPLICE_SPLICE_ROOT",
                "@ARCHIVE_TO_DIR@/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@",
                SECT_GKOVERRIDES,
                "The directory where 'root' from this build should be "
                "archived."),

            Env("SUMMARY_FILE",
                "@ARCHIVE_TO_DIR@/"
                "@ONBLD_BRANCH_PREFIX@summary_@BUILDID_THISBUILD@_@MACH@",
                SECT_GKOVERRIDES,
                "When 'archive_build' is 'true', the file into which a"
                "summary of the current build will be written."),

            Env("KSPLICEROOT", "@ROOT@_ksplice_@BUILDID_THISBUILD@",
                SECT_GKOVERRIDES,
                "Define where ksplice objects are installed. Gatelings should "
                "never need to override this. The supplied value is modified "
                "by build(1) depending on whether we're doing a debug or "
                "non-debug build, appending a '-nd' suffix to the value."),

            Env("KSPLICE_SPLICE_KSPLICEROOT",
                "@ARCHIVE_TO_DIR@/@ONBLD_BRANCH_PREFIX@proto/"
                "root_@MACH@_ksplice_@BUILDID_THISBUILD@",
                SECT_GKOVERRIDES,
                "The directory where 'kspliceroot' from this build should be "
                "archived."),

            Env("KSPLICE_DELIVERY_EXCEPTIONS", "grub", SECT_GKOVERRIDES,
                "Space-separated list of features that should be delivered "
                "from the baseline build, without regard for whether or not "
                "they were modified during this splice build. Each item 'fu' "
                "in this list should correspond to a version-controlled file "
                "usr/src/pkg/ksplice_exceptions.'fu'"),

            Env("KSPLICE_MODIFIED_EXCEPTIONS", "default", SECT_GKOVERRIDES,
                "Space-separated list of features expected to cause spurious "
                "mixed-deliverable failures in a splice build. Each item 'fu' "
                "in this list should correspond to a version-controlled file "
                "usr/src/pkg/ksplice_exceptions.'fu'. You should always leave "
                "the default exceptions here, and append to this list."),

            Env("PKGRESOLVEIMAGE", "/", SECT_GKOVERRIDES,
                "By default package dependencies will be resolved against the "
                "other published manifests and then finally against the files "
                "on the build machine itself.  pkgresolveimage can be set to "
                "point to an alternate Solaris image for the purposes of "
                "building ON bits that are synced with an older build of "
                "Solaris than what is installed on the build machine."),

            Env("PARFAIT_SERVER_URL",
                "https://parfait.cz.oracle.com/parfait-server/projects"
                "/solaris-on/tasks/on-@ONBLD_HG_BRANCH@-@MACH@",
                SECT_CODECHECKS,
                "The URL used to contact the parfait server for baseline "
                "analysis, or new baseline upload. Note that since the "
                "active branch is included in this URL, the default will fail "
                "for any developer workspaces that include feature branches "
                "which are unknown to the parfait server. Make sure to update "
                "this value accordingly."),

            Env("PARFAIT_SERVER_PASSWORD", None, SECT_GKOVERRIDES, "The "
                "password needed to upload analysis results to "
                "'parfait_server_url'. This is only required if "
                "'parfait_type' is set to 'upload' and the server has been "
                "configured to require a password."),

            Env("PARFAIT_GUI", "yes", SECT_GKOVERRIDES, "If set to 'no', then "
                "html logs are not produced when performing parfait "
                "analysis."),

            Env("FORTIFY_SERVER",
                "https://fortifyapp2.us.oracle.com:8443/ssc", SECT_GKOVERRIDES,
                "The server used by Fortify to upload baselines and perform "
                "analysis."),

            Env("FORTIFY_READ_TOKEN",
                "@ONBLD_DATA@/fortify/fortifyread.token",
                SECT_GKOVERRIDES, "The authorization file to use "
                "when reading Fortify baselines.\n"
                "To generate a new token run\n"
                "fortifyclient -url <serverURL> token -gettoken "
                "AnalysisDownloadToken -daysToLive <days> -user onreaduser"),

            Env("FORTIFY_WRITE_TOKEN",
                "@ONBLD_DATA@/fortify/fortifywrite.token",
                SECT_GKOVERRIDES, "The authorization file to use "
                "when uploading fortify baselines and performing analysis.\n"
                "To generate a new token run\n"
                "fortifyclient -url <serverURL> token -gettoken "
                "AnalysisUploadToken -daysToLive <days> -user onwriteuser"),

            Env("FORTIFY_INTERMEDIATE_DIR",
                "@ONBLD_DATA@/fortify/inter", SECT_GKOVERRIDES,
                "The directory to store Fortify intermediate data."),

            Env("FORTIFY_PROJECT",
                "Solaris", SECT_GKOVERRIDES, "The default project name for "
                "Fortify analysis."),

            Env("FORTIFY_PROJECT_VERSION",
                "Trunk", SECT_GKOVERRIDES, "The default project version for "
                "Fortify analysis."),

            Env("PKGSURFARCHIVE",
                "@ONBLD_KEEP@/@ONBLD_BRANCH_PREFIX@pkgsurf-ref.pkg/"
                "@MACH@-debug",
                SECT_GKOVERRIDES,
                "The directory containing package repositories from a "
                "previous build which are used for the 'pkgsurf' task. "
                "\n\n"
                "By default, if 'pkgarchive' points to packages within "
                "the workspace, the 'save_packages' task will move those "
                "packages to 'pkgsurfarchive' before clobbering the "
                "packages."
                "\n\n"
                "If 'pkgmerge' is set, and 'pkgsurfarchive' points to a "
                "'nightly' directory within the workspace, use the merged "
                "repositories as the resurface reference."),

            Env("PKGMERGE_ALTWS_TIMEOUT", None, SECT_GKOVERRIDES,
                "When performing a 'pkgmerge' task, this specifies the number "
                "of seconds the build process running the merge should wait "
                "for the corresponding 'packages' and/or 'packages-nd' tasks "
                "in the alternative workspace to complete. If this is not set "
                "then the 'pkgmerge' task defaults to a 1 hour timeout."),

            # Use $BASEWSDIR rather than $VERSION because nightly builds tend
            # to modify the latter (to assign a build number or date stamp).
            # A dataset name that changes across builds will make 'clobber's
            # life harder to clean up after failed builds.
            Env("DC_DATASET", "<zpool>/dc/@USER@/@BASEWSDIR@",
                SECT_GKOVERRIDES,
                "The name of the dataset used by build(1) when invoking "
                "tasks that construct media created by distro_const(8), such "
                "as fish.image. If the dataset does not exist, build(1) will "
                "create it."
                "\n\n"
                "At runtime, the special token '<zpool>' is computed to be "
                "the value of the 'config/build_pool' property from "
                "svc:/developer/build/onbld, or if that is not set, "
                "the top-level dataset that contains this workspace. If this "
                "workspace is resident on an NFS share, then we use the "
                "top-level dataset of '/'."
                "\n\n"
                "See 'snapshots' for details of the ZFS allow permission sets "
                "that must be granted for users of this dataset. Both "
                "'@build_perms' and '@build_perms_ext are required."
                "\n\n"
                "Note that *if* the 'dc_dataset' dataset is created by "
                "build(1), then it WILL BE DESTROYED as part of the build "
                "once build artifacts have been saved to $WS or, in the case "
                "of interrupted builds, during 'clobber'. This is done so "
                "so that old data does not accumulate on build machines."
                "\n\n"
                "For developers working on the image construction task "
                "itself, setting ONBLD_PRESERVE_DC_DATASET in the "
                "environment and adding that to the 'userenv' bldrc value "
                "will prevent 'fish.destroy_dc' from removing the "
                "'dc_dataset'."
                "\n\n"
                "It is recommended that you do not use an existing dataset "
                "for 'dc_dataset' unless you promise to clean up afterwards."),

            Env("FISH_GK_IMAGES", "false", SECT_GKOVERRIDES,
                "When set to 'true', during a 'nightly' task, the "
                "'fish.gk_images' task is run, to produce AK upgrade "
                "packages as well as AI and text install media for the 'nas' "
                "product, as well as AI and upgrade packages for the 'jds' "
                "product. When set to 'false' (the default), the 'fish.image' "
                "task is run instead, producing only 'nas' AK upgrade "
                "packages."),

            Env("MANROOT",
                "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@proto/root_@MACH@_man",
                SECT_GKOVERRIDES,
                "The top of the man proto area for this architecture. "
                "The supplied value is shared between debug and non-debug "
                "builds, and is only built once by build(1)"),

            Env("FISH_GK_NIGHTLY_IMAGES", "false", SECT_GKOVERRIDES,
                "When set to 'true' during a 'nightly' task, images suitable "
                "for GK nightly builds are created.  In particular this is "
                "limited to producing fish 'nas' upgrade and AI iso images."),

            # AK Versioning
            # AK versioning uses <AKV_GATE_ID>-<AK_PRODUCT>-<AKV_VERSION_BASE>
            # where AKV_VERSION_BASE is
            # <AKV_COMPONENT>,<AKV_BUILD>-<AKV_BUILD>.
            # An example for AK 8.8 nas biweekly 20 would be
            # ak-nas-2013.06.05.8.0,1-1.20.
            Env("AKV_COMPONENT", "", SECT_GKOVERRIDES,
                "The component represents the release date for biweekly "
                "builds and release builds.  For development builds, this "
                "value is always the date of the build."),

            Env("AKV_BUILD", "", SECT_GKOVERRIDES,
                "The build number for the build.  For development builds, "
                "this defaults to the Makefile.version defaults unless there "
                "is an existing proto area version.xml.  In that case, the "
                "build number is the version.xml build number plus one."),

            Env("AKV_BRANCH", "", SECT_GKOVERRIDES,
                "This field is effectively an indicator of the type of build."
                "  The valid values are 0 (development), 1 (biweekly/release),"
                " 2 (IDR).  The default is 0 (development)."),

            Env("AKV_VERSION_BASE", "", SECT_GKOVERRIDES,
                "Override the entire version base string."),

            Env("FISH_LICENSES", "", SECT_GKOVERRIDES,
                "Set that the fish LICENSES file should be generated as part "
                "of the fish images."),

            Env("ZIPDIR", "@ONBLD_DATA@/@ONBLD_BRANCH_PREFIX@zipfiles",
                SECT_GKOVERRIDES,
                "Location to write the RE zip files. This is deliberately not "
                "suffixed with an architecture because it is intended to be "
                "used after the 4-way pkgmerge; the filenames deployed here "
                "already label the content of the zip files."),

            Env("CERT_EXPIRY_WARNING_DAYS", "90", SECT_GKOVERRIDES,
                "If an active certificate is due to expire within the next "
                "'CERT_EXPIRY_WARNING_DAYS' days, the "
                "check.certificate-expiry task will generate an error."),
        ]

        # Check for duplicates in the final environment to ensure that
        # developers haven't unintentionally added multiple Env() objects for
        # the same variable.
        duplicates = set()
        for item in self.__env_list:
            if item.name in self.__environment:
                duplicates.add(item.name)
            self.__environment[item.name] = item

        for item in self.__fixed_environment:
            if item.name in self.__environment:
                duplicates.add(item.name)
            self.__environment[item.name] = item

        if duplicates:
            raise OnBldEnvironmentException(
                "%s already exists in self.__environment or "
                "self.__fixed_environment, which constitutes a bug in "
                "environment.py. Please report this to your friendly "
                "gatekeepers at %s" %
                (", ".join(list(duplicates)), GATEKEEPER_MAILTO))

        #
        # Technically, this is unnecessary since the build will unset all
        # environment variables anyway. However, we'll keep this list and
        # use it to issue strong warnings if users try to whitelist these
        # variables and inherit them from their shell.
        #
        # Note that the use of 'LC_ALL' here actually isn't toxic since we're
        # setting the individual LC_* variables in self.__fixed_environment,
        # which will override that. However, we'll include it in the list in
        # case anyone thinks it's going to have an effect.
        #
        self.__toxic_env_vars = [
            "ARCH",
            "CLASSPATH",
            "CONFIG",
            "ENV",
            "GROUP",
            "IDR_BUILD",
            "INSTALL_BIN_PATH",
            "LC_ALL",
            "LD_OPTIONS",
            "LD_LIBRARY_PATH",
            "LD_AUDIT",
            "LD_BIND_NOW",
            "LD_BREADTH",
            "LD_CONFIG",
            "LD_DEBUG",
            "LD_FLAGS",
            "LD_LIBRARY_PATH_64",
            "LD_NOVERSION",
            "LD_ORIGIN",
            "LD_LOADFLTR",
            "LD_NOAUXFLTR",
            "LD_NOCONFIG",
            "LD_NODIRCONFIG",
            "LD_NOOBJALTER",
            "LD_PRELOAD",
            "LD_PROFILE",
            "OWNER",
            "RELEASE_BUILD",
            "REMOTE",
            "SRU_BUILD",
            "TERMCAP",
            "TERMINFO",
            "UMASK",
        ]

        for item in self.__toxic_env_vars:
            if item in self.__environment:
                raise OnBldEnvironmentException(
                    "%s already exists in self.__environment, which "
                    "constitutes a bug in build.py. Please report this to "
                    "your friendly gatekeepers at %s!" %
                    (item, GATEKEEPER_MAILTO))
            self.__environment[item] = empty_env

        # Apart from the prescribed environment variables above, the following
        # variables are deemed safe to always inherit from the user's
        # environment and do not impact the OS/Net build.
        self.__whitelist_env_vars = [
            "BLDRC",
            "DISPLAY",
            "EDITOR",
            "HISTCMD",
            "HISTFILE",
            "HISTSIZE",
            "HOME",
            "LESS",
            "LOGNAME",
            "ONBLD_PKGSERVE_PORT",
            "ONBLD_SERVE_PORT",
            "PAGER",
            "PS1",
            "PS2",
            "PS3",
            "PS4",
            "RPS1",
            "RPS2",
            "SHELL",
            "SPROMPT",
            "SSH_AGENT_PID",
            "SSH_AUTH_SOCK",
            "TERM",
            "XAUTHORITY",
            "_",
            "http_proxy",
            "no_proxy"
        ]

        # Suffixed vars get special treatment. If variables are
        # derived from suffixed vars, we don't want them to get an
        # errant "-nd" suffix stuffed into the middle somewhere.
        self.__suffix_vars = [
            "ROOT",
            "KSPLICE_BASELINE_ROOT",
            "KSPLICE_BASELINE_KSPLICEROOT",
            "KSPLICE_SPLICE_ROOT",
            "KSPLICE_SPLICE_KSPLICEROOT",
            "NATIVEROOT",
            "STUBROOT",
            "KSPLICEROOT",
            "PARENT_ROOT",
            "PARENT_NATIVEROOT",
            "BUILD",
            "KSPLICE_BASELINE_BUILD",
            "KSPLICE_SPLICE_BUILD",
            "SUMMARY_FILE",
            "FW_PROTO_STUB",
            "FW_PROTO",
            "FISHROOT"
        ]

        # Similar to the above, these also get special treatment, but
        # indicate where we have extended "-debug" or "-nondebug" suffixes
        # rather than just '' or '-nd' suffixes.
        self.__ext_suffix_vars = [
            "ALTWS_PKGARCHIVE",
            "KSPLICE_BASELINE_PKGARCHIVE",
            "KSPLICE_SPLICE_PKGARCHIVE",
            "PKGARCHIVE",
            "PKGSURFARCHIVE",
            "CPIODIR",
        ]

        self.__compiletools = [
            ToolsTuple("as", "-V", "/usr/bin/as", "12", 1, True),
            ToolsTuple("cc", "-V", "@COMPILERPATH@/bin/cc", "5.15", 1, True),
            ToolsTuple("cxx", "-V", "@COMPILERPATH@/bin/CC", "5.15", 1, True),
            ToolsTuple(
                "dmake", "-v", "@COMPILERPATH@/bin/dmake", "8.4", 1, True),
            ToolsTuple(
                "javac", "-fullversion", "@JAVA_ROOT@/bin/javac", "1.8", 1,
                True),
            ToolsTuple("ld", "-V", "/usr/bin/ld", "5.11", 1, True),
            ToolsTuple(
                "parfait", "--version", "@PARFAIT_HOME@/bin/parfait",
                "4.0", 1, False),
            ToolsTuple(
                "fortify", "-version", "@FORTIFY_HOME@/bin/sourceanalyzer",
                "19.1.0.22", 1, False)
        ]

        # Regexp matches for tool versions. The keys for this
        # dictionary correspond with the keys in self.__compiletools.
        self.__compiletools_re = {
            "dmake":
            r"dmake:.*Make (\d+\.\d+?) SunOS_%s.*" % DMAKE_MACH,

            "cc":
            r"cc: Studio 12.6 Sun C (\d+\.\d+?) SunOS_%s.*" % MACH,

            "cxx":
            r"CC: Studio 12.6 Sun C\+\+ ([\d\.]+) SunOS_%s.*" % MACH,

            "as":
            ".*Compiler Common "
            r"(\d+\.\d+) SunOS_%s (s11|st)_(\d+).*" % MACH,

            "ld":
            "ld: Software Generation Utilities - Solaris Link Editors: "
            r"(\d\.\d+?)-(\d\.\d+?)",

            "javac":
            r"javac full version \"(\d\.\d\.\d)_(\d+?)-b(\d+?)\"",

            "parfait": r"Parfait ([\d\.]+) .*",

            "fortify": r"Fortify Static Code Analyzer ([\d\.]+).[\d\.]+ .*",
        }

        self.checktools = {
            "findunref": {},
            "wsdiff": {},
        }

        # The build uses stash(1) to install files into $ROOT for the most
        # part, which sets file permissions. For things that don't or can't
        # use stash, ensure we use a consistent umask.
        os.umask(0o022)

        # No bldrc file was passed, so we build a self.__cfg
        if not self.bldrc:
            for env in self.__env_list:
                if env.section is None:
                    continue
                if not self.__cfg.has_section(env.section):
                    self.__cfg.add_section(env.section)
                if env.default is None:
                    self.__cfg.set(env.section, env.name.lower(), "<None>")
                else:
                    self.__cfg.set(env.section, env.name.lower(), env.default)
            for tool in self.__compiletools:
                # We don't require all tools to have overrides in the bldrc
                if tool.bldrc:
                    self.__cfg.set("paths", tool.name, tool.default)

    @property
    def build_date(self):
        """The $BUILD_DATE timestamp - see the BUILD_DATE value of
        self.__env_list."""
        return self.__build_date

    @build_date.setter
    def build_date(self, value):
        if not self.__build_date:
            self.__build_date = value
        else:
            raise ValueError(
                "Attempt to change build_date after it was already set to %s" %
                self.__build_date)

    @property
    def iso8601_build_date(self):
        """The $ISO8601_BUILD_DATE timestamp - see the ISO8601_BUILD_DATE value
        of self.__env_list."""
        return self.__iso8601_build_date

    @iso8601_build_date.setter
    def iso8601_build_date(self, value):
        # This should only ever happen during development
        if not self.__iso8601_build_date:
            self.__iso8601_build_date = value
        else:
            raise ValueError(
                "Attempt to change iso8601_build_date after it was already "
                "set to %s" % self.__iso8601_build_date)

    def help(self, variable):
        """The variables in our build environment rc files can get lost in
        the noise of documentation, so instead, we document them in the code,
        and allow users to lookup help programmatically."""

        env = self.__environment.get(variable.upper(), empty_env)
        if env.doc is None:
            return None, None
        return env.doc, env.section

    def keys(self):
        """Return a list of the Env objects we know about, along with their
        section name"""
        return dict(
            (key, self.__environment[key].section)
            for key in list(self.__environment.keys()))

    def get_build_environment(
            self, include_unset_vals=False, debug=None, nondebug=None,
            verbose=False, userenv_details=False, readonly=False,
            fish=False):
        """Return a dictionary of environment variable names and values which
        can be used when doing an OS/Net build, using the supplied bldrc file
        as a source.

        If 'userenv_details' is True, we return a list of the environment
        variables in the returned dictionary that the user had specified should
        be inherited from their existing shell environment (a whitelist) and a
        list of the environment variables that were removed from the returned
        dictionary by user-request (a blacklist)

        If 'include_unset_vals' is True, then the returned dictionary will
        include keys with values of None (intentionally not a basestring) to
        indicate a shell environment variable should not be set for this value.
        All other dictionary values are basestrings, and indicate a shell
        variable ought to be set for this key.

        Parameter values 'debug' and 'nondebug' override self.debug and
        self.nondebug just for producing this environment, but do not change
        the values in the underlying OnBldEnvironment object. If those values
        are 'None' then we respect the values set in OnBldEnvironment,
        producing a *debug* environment if both values are True.

        If 'readonly' is True, we return an environment that has not
        modified any part of the workspace while that environment was being
        created. The returned dictionary is mutable.
        """

        global AKV_BUILD

        newenv = {}
        whitelist_env_vars_user = []
        blacklist_env_vars_user = []

        # If we've added parameters to override the environment produced, that
        # those values make sense.
        if nondebug is not None and debug is not None and debug == nondebug:
            raise OnBldEnvironmentException(
                "Conflicting parameter values for 'debug' and 'nondebug'")

        # Otherwise, if set from the bldrc, we can allow the object to contain
        # both values - debug takes precedent
        if debug is None:
            debug = self.debug
        if nondebug is None:
            nondebug = self.nondebug

        if not debug and not nondebug:
            raise OnBldEnvironmentException(
                "At least one of 'debug' and 'nondebug' must be set.")

        # set the default clean build environment first
        for name in self.__environment:
            if self.__environment[name].default is not None or \
                    include_unset_vals:
                newenv[name] = self.__environment[name].default

        # pull in variables that we know are safe
        for name in self.__whitelist_env_vars:
            newenv[name] = os.environ.get(name)
            # some users may not have defaults set, in which case, we shouldn't
            # set the variable either
            if newenv[name] is None:
                del newenv[name]

        if self.bldrc:
            # walk the config file, setting any additional variables
            for envvar in self.__environment:
                env_item = self.__environment[envvar]
                if not env_item.section:
                    continue
                try:
                    newenv[envvar] = self.__cfg.get(
                        env_item.section, envvar.lower())
                    # allow for our special '<None>' strings that indicate
                    # a value should not be set.
                    if newenv[envvar] == "<None>":
                        if include_unset_vals:
                            newenv[envvar] = None
                        else:
                            del newenv[envvar]

                    # check to see if we've got a normalized path for anything
                    # in SECT_PATHS. Some Makefile targets do not deal with
                    # non-normalized paths terribly well
                    if env_item.section == SECT_PATHS:
                        path = newenv.get(envvar)
                        if not path:
                            continue
                        normpath = os.path.normpath(path)
                        if newenv[envvar].rstrip("/") != normpath and \
                                normpath not in [".", "/"]:
                            logger.info(
                                "Normalizing path '%s' to '%s'" %
                                (newenv[envvar], normpath))
                            newenv[envvar] = normpath

                except (
                        configparser.NoOptionError,
                        configparser.NoSectionError):
                    newenv[envvar] = env_item.default
                    if newenv[envvar] in (None, "<None>") and \
                            not include_unset_vals:
                        del newenv[envvar]

            # load any user-specified overrides
            whitelist_env_vars_user = set(self.__cfg.get(
                SECT_BUILDOPTIONS, "userenv").split(";"))

            blacklist_env_vars_user = set(self.__cfg.get(
                SECT_BUILDOPTIONS, "userunsetenv").split(";"))

            # remove empty keys
            if "" in whitelist_env_vars_user:
                whitelist_env_vars_user.remove("")

            if "" in blacklist_env_vars_user:
                blacklist_env_vars_user.remove("")

            invalid_user_overrides = set()
            missing_user_overrides = set()

            # We allow, but discourage overrides anything in the list of
            # fixed environment variables, because otherwise the user has no
            # way of overriding them.
            fixed_env = [e.name for e in self.__fixed_environment]
            for key in whitelist_env_vars_user:
                if key in fixed_env:
                    if verbose:
                        logger.warning(
                            "User override of $%s is not recommended. "
                            "Continue at your own risk." % key)
                    newenv[key] = os.environ.get(key, "")
                elif key in self.__toxic_env_vars:
                    if verbose:
                        logger.warning(
                            "User override of $%s is strongly discouraged. "
                            "Continue at your own risk." % key)
                    newenv[key] = os.environ.get(key, "")
                elif key not in self.__environment:
                    if key not in os.environ:
                        missing_user_overrides.add(key)
                    else:
                        newenv[key] = os.environ[key]
                else:
                    invalid_user_overrides.add(key)

            if missing_user_overrides and self.strict_environment:
                vars = ["$%s" % key for key in missing_user_overrides]
                raise OnBldEnvironmentException(
                    "User override of %s required, but no such "
                    "value is set in the environment." % ", ".join(vars))

            if invalid_user_overrides:
                raise OnBldEnvironmentException(
                    "The value of the following keys must be set in "
                    "individual bldrc properties, not 'userenv': %s" %
                    ", ".join(list(invalid_user_overrides)))

            self.__whitelist_env_vars.extend(list(whitelist_env_vars_user))

        # set some dynamic environment variables if they're not already set

        # the self.build_date is immutable for this object, once set.
        now = datetime.datetime.now()
        if not self.build_date:
            existing = newenv.get("BUILD_DATE")
            if existing:
                self.build_date = existing
            else:
                self.build_date = datetime.datetime.strftime(
                    now, "%Y%m%dT%H%M%S")
        newenv["BUILD_DATE"] = self.build_date

        if not self.iso8601_build_date:
            existing = newenv.get("ISO8601_BUILD_DATE")
            if existing:
                self.iso8601_build_date = existing
            else:
                self.iso8601_build_date = datetime.datetime.strftime(
                    now, "%Y-%m-%d")
        newenv["ISO8601_BUILD_DATE"] = self.iso8601_build_date

        if not newenv.get("RELEASE_DATE"):
            newenv["RELEASE_DATE"] = datetime.date.strftime(now, "%B %Y")

        if not newenv.get("BASEWSDIR"):
            newenv["BASEWSDIR"] = os.path.basename(newenv.get("WS", ""))

        user, logname, mailto = self.get_userinfo()
        if not newenv.get("USER"):
            newenv["USER"] = logname

        if not newenv.get("LOGNAME"):
            newenv["LOGNAME"] = newenv["USER"]

        if logname != newenv["USER"]:
            logger.warning(
                "The current user '%s' does not match the 'user' value, '%s' "
                "in the bldrc file." % (logname, newenv["USER"]))

        # For mailto, a value of "" implies we should auto-fill the email
        # address from the user's .hgrc.
        if "MAILTO" in newenv and newenv["MAILTO"] == "":
            if mailto == "":
                raise OnBldEnvironmentException(
                    "Unable to determine a default value for 'mailto' from "
                    "~/.hgrc. See 'build explain mailto'")
            newenv["MAILTO"] = mailto

        if not newenv.get("DEV_CM"):
            newenv["DEV_CM"] = \
                "\"@(#)SunOS Internal Development: %(USER)s %(BUILD_DATE)s " \
                "[%(BASEWSDIR)s]\"" % newenv

        # We only set PYTHONPATH if PKG7_ROOT is set to something other than
        # the empty string. Setting a Python 2.7 PYTHONPATH and trying to run
        # Python 3.x binaries that don't have '-E' in their #! args will
        # result in breakage.
        if "PYTHONPATH" in newenv and newenv["PKG7_ROOT"] == "":
            newenv["PATH"].replace("@PKG7_ROOT@/usr/bin:", "")
            del newenv["PYTHONPATH"]

        # We need to special-case some Env() items, since they should not
        # appear in the shell environment.
        for name in [
                "DEBUG", "NONDEBUG", "INCREMENTAL", "CHECK", "SNAPSHOTS",
                "PARFAIT_TYPE", "UPDATE_PARENT_WS", "LOG_FORMAT",
                "PKGMERGE", "PKGSURF"]:
            if name in newenv:
                del newenv[name]

        if (self.pkgmerge and "PKGARCHIVE" not in newenv) or \
                (self.pkgsurf and "PKGARCHIVE" not in newenv):
            raise OnBldEnvironmentException(
                "Cannot set 'pkgmerge' or 'pkgsurf' to 'true' when "
                "'pkgarchive' is set to '<None>'")

        # Check for closed, man and fish trees, and set environment variables
        for var, dir in (
                ("CLOSED_IS_PRESENT", "usr/closed"),
                ("MAN_IS_PRESENT", "usr/man"),
                ("FISH_IS_PRESENT", "usr/fish")):
            if var not in newenv or newenv[var] is None:
                if os.path.isdir("%s/%s" % (newenv["WS"], dir)):
                    newenv[var] = "yes"
                else:
                    newenv[var] = "no"
            if newenv[var] not in ["yes", "no"]:
                raise OnBldEnvironmentException(
                    "'%s' should be set to '<None>', 'yes' or 'no', not '%s'" %
                    (var.lower(), newenv[var]))

        # The presence of usr/fish or an explicit FISH_IS_PRESENT=yes/no in
        # the bldrc determines the build flavor (AK vs Solaris Server).
        if self.__cfg.get(SECT_PARTIAL, "fish_is_present") == "<None>":
            if os.path.isdir(os.path.join(newenv["WS"], "usr/fish")):
                newenv["FISH_IS_PRESENT"] = "yes"
            else:
                newenv["FISH_IS_PRESENT"] = "no"

        # Override the FISH_IS_PRESENT computation for standalone fish.
        # When building a standalone fish gate, add a clue to the environment.
        # $WS is computed initially by find_workspace(.) in
        # $SRC/tools/scripts/build.py and is is set to the top level of the
        # fish subrepo if no ON $SRC repo is found. Checking for the
        # AK-specific file off $WS is all we need to do here.
        # The FISH_STANDALONE_CANARY_PATH computation should stay as long as
        # we continue to support standalone fish builds.
        if os.path.exists(
                os.path.join(newenv["WS"], FISH_STANDALONE_CANARY_PATH)):
            newenv["FISH_IS_PRESENT"] = "yes"
            newenv["FISH_STANDALONE"] = "yes"
            # the FW_HOOKS default of @WS@ is never correct for standlone fish
            if newenv["FW_HOOKS"] == "@WS@":
                if verbose:
                    logger.warning(
                        "'fw_hooks' was set to @WS@ in the bldrc which cannot "
                        "work for standalone fish builds!\n"
                        "          Falling back to /ws/on-gate.")
                newenv["FW_HOOKS"] = "/ws/on-gate"

        # Sanity check fish_is_present now.
        elif newenv["FISH_IS_PRESENT"] == "yes" and \
                not os.path.exists(os.path.join(newenv["WS"], "usr/fish")):
            raise OnBldEnvironmentException(
                "fish_is_present is set to 'yes', but usr/fish is missing.")

        if fish or newenv.get("FISH_STANDALONE"):
            newenv = self.__get_fish_env(newenv)
        # Otherwise, check to see if the bldrc wants us to do a server build
        elif newenv.get("FISH_IS_PRESENT") == "no":
            newenv["ONBLD_BRANCH_PREFIX"] = SERVER_PREFIX
        else:
            # we're doing an AK build - set $FISHROOT so we can retrieve
            # the versions.xml file later.
            root = newenv["ROOT"]
            root_base = os.path.basename(root)
            root_dir = os.path.dirname(root)
            newenv["FISHROOT"] = os.path.join(
                root_dir, FISH_PREFIX + root_base)

        # Set mercurial branch.
        ret, out, err = util.cmdline_run("hg -R %s branch" % newenv["WS"])
        if ret != 0:
            raise OnBldEnvironmentException(
                "Cannot determine mercurial branch. %s" % err)

        branch = out.rstrip()
        if branch == "default" and newenv.get("FISH_IS_PRESENT") == "no":
            # This is a server build of the default branch. Force the "branch"
            # name to "server". This is to maintain the server analysis for
            # Parfait
            branch = "server"

        newenv["ONBLD_HG_BRANCH"] = branch

        # Modify the environment based on flags that have been set. It is
        # important to do this before we expand macros, so that variables
        # referencing macros get any new definitions.
        if self.overrides.get("check.parfait", True) and self.check:
            self.__set_parfait(newenv)

        self.__set_release_type(newenv)

        if debug:
            self.__set_debug(newenv)
        else:
            self.__set_nondebug(newenv)

        if self.incremental:
            self.__set_incremental(newenv)

        if not self.build_in_place:
            self.__unset_build_in_place(
                newenv, include_unset_vals=include_unset_vals)
        else:
            self.__set_build_in_place(newenv)

        if self.overrides.get("check.elfsigncmp", False) and self.check:
            self.__set_elfsigncmp(newenv)

        # tonic builds require the closed binaries to appear in a different
        # proto area. $TONICBUILD should be set to an empty value if a tonic
        # build is being performed.
        if newenv.get("TONICBUILD", None):
            newenv["CLOSEDROOT"] = "@ROOT@-closed"
            newenv["TONICBUILD"] = ""

        # if users are overriding SPRO_VROOT, then make sure it's the same as
        # COMPILERPATH, otherwise other default values could break
        spro_vroot = newenv["SPRO_VROOT"]
        if spro_vroot != "@COMPILERPATH@" and \
                spro_vroot != newenv["COMPILERPATH"]:
            raise OnBldEnvironmentException(
                "'compilerpath' should have the same value as 'spro_vroot'")

        # Check for spaces in WS values - spaces in paths are common forms of
        # script breakage, so we don't allow them.
        if " " in newenv["WS"]:
            raise OnBldEnvironmentException(
                "'ws' cannot point to a path that contains space characters.")
        # Likewise, check for double "/" in WS values - we don't want to break
        # the SRC_REL variable.
        if "//" in newenv["WS"]:
            raise OnBldEnvironmentException("'ws' cannot contain '//'")

        def __get_dmake_max_jobs():
            """Compute a value for DMAKE_MAX_JOBS, consulting .make.machines,
            falling back to psrinfo(1)."""
            max_jobs = None
            make_machines_path = "%s/.make.machines" % os.path.expanduser("~")
            try:
                with open(make_machines_path, "r") as make_machines:
                    for line in make_machines:
                        items = line.lstrip().rstrip().split()
                        if not items:
                            continue
                        if items[0] == HOSTNAME:
                            max_jobs = items[1].replace("max=", "")
                            break
            except Exception:
                pass
            if max_jobs:
                return max_jobs
            # Otherwise, compute max_jobs using the number of processors,
            # capping at 16. 4-way systems can still cope with 16 max jobs,
            # generally.
            ret, out, err = util.cmdline_run("/usr/sbin/psrinfo")
            max_jobs = len(out.split("\n")) - 1
            if max_jobs >= 4:
                return "16"
            if max_jobs < 4:
                return "4"
            return str(max_jobs)

        # Check that dmake_max_jobs is a positive integer
        dmake_max_jobs = newenv.get("DMAKE_MAX_JOBS")
        if not dmake_max_jobs:
            dmake_max_jobs = __get_dmake_max_jobs()
            newenv["DMAKE_MAX_JOBS"] = dmake_max_jobs
        try:
            if int(dmake_max_jobs) <= 0:
                raise ValueError()
        except ValueError:
            raise OnBldEnvironmentException(
                "invalid dmake_max_jobs value '%s'. Expecting a positive "
                "integer" % dmake_max_jobs)

        # Check for values that cause environment variables to be set to
        # empty variables.
        for key in ["__GNUC"]:
            val = newenv.get(key)
            if val is not None:
                newenv[key] = ""
            if val and val.lower() not in ["true", "", "yes"]:
                raise OnBldEnvironmentException(
                    "%s should be set to <None> to clear value in the "
                    "environment" % key.lower())

        if not self.tmpdir or not os.path.exists(self.tmpdir):
            # Set tmpdir and do our best to ensure it gets cleaned up on exit
            self.tmpdir = tempfile.mkdtemp(
                prefix="build.%s." % os.getpid(), dir="/tmp")
            atexit.register(shutil.rmtree, self.tmpdir, ignore_errors=True)

        # Set tmpdir, allowing for any user-set TMPDIR value.
        if "TMPDIR" not in newenv:
            newenv["TMPDIR"] = self.tmpdir
        else:
            self.tmpdir = newenv["TMPDIR"]

        def determine_dc_zpool(dir, last_resort=False):

            smf_cmd = "svcprop -p config/build_pool " \
                      "svc:/developer/build/onbld:default"
            ret, out, err = util.cmdline_run(smf_cmd)
            if ret == 0 and out.strip() != '""':
                return out.strip()

            cmd = "/usr/bin/df -P {0}".format(dir)
            ret, out, err = util.cmdline_run(cmd)
            if ret != 0:
                raise OnBldEnvironmentException("{0} failed: {1}".format(
                    cmd, err))
            fsline = out.split("\n")[1]
            fs = fsline.split()[0]

            # check for NFS-mounted datasets
            cmd = "/usr/bin/df -n {0}".format(dir)
            ret, out, err = util.cmdline_run(cmd)
            if ": nfs" in out:
                if dir != "/":
                    return determine_dc_zpool("/", last_resort=True)
            return fs.split("/")[0]

        if "<zpool>" in newenv["DC_DATASET"]:
            pool = determine_dc_zpool(newenv["WS"])
            newenv["DC_DATASET"] = newenv["DC_DATASET"].replace(
                "<zpool>", pool)

        # replace @MACROS@
        osenv_macro_re = re.compile("@([_A-Z0-9a-z]*)@")
        expanded_vars = {}

        def macro_expansion():
            """Expand macros present in property values"""

            for key in newenv:
                if not newenv[key]:
                    expanded_vars[key] = newenv[key]
                    continue
                # a set of all the macros used in this value
                macros = set(osenv_macro_re.findall(newenv[key]))
                if not macros:
                    expanded_vars[key] = newenv[key]
                    continue

                for macro in macros:
                    if macro in expanded_vars:
                        if key in self.__suffix_vars:
                            newenv[key] = newenv[key].replace(
                                "@%s@" % macro, re.sub(r'-nd$', '',
                                                       expanded_vars[macro]))
                        elif key in self.__ext_suffix_vars:
                            newenv[key] = newenv[key].replace(
                                "@%s@" % macro, re.sub(r'-nondebug$', '',
                                                       expanded_vars[macro]))
                        else:
                            newenv[key] = newenv[key].replace(
                                "@%s@" % macro, expanded_vars[macro])
                        expanded_vars[key] = newenv[key]

        # macro_expansion could be recursive, but it's probably safe to assume
        # that nobody has more than 20 levels of nested macros.
        for i in range(1, 20):
            macro_expansion()

        # ... but we'll check nonetheless
        remaining_macros = set()
        for key in newenv:
            if not newenv[key]:
                continue
            macros = set(osenv_macro_re.findall(newenv[key]))
            if macros:
                remaining_macros.update(macros)
        if remaining_macros:
            if len(remaining_macros) > 1:
                message = "Possible circular reference with %s?" % \
                    ", ".join(list(remaining_macros))
            else:
                message = "No definition for @%s@" % remaining_macros.pop()
            raise OnBldEnvironmentException(
                "Unexpanded macros in bldrc file! %s" % message)

        if self.nondebug:
            for envvar in self.__environment:
                env_item = self.__environment[envvar]
                if not env_item.section:
                    continue
                # check to see if the user has set the value in their
                # config file
                try:
                    value = self.__cfg.get(
                        env_item.section, envvar.lower())
                except (
                        configparser.NoOptionError,
                        configparser.NoSectionError):
                    # otherwise use the default
                    value = env_item.default

        # unset any values that the user specifically does not want
        # set in the build environment
        for key in blacklist_env_vars_user:
            if key in newenv:
                del newenv[key]

        if newenv.get("CORES_PATTERN") is not None and not readonly:
            pattern = newenv["CORES_PATTERN"]
            if not pattern:
                raise OnBldEnvironmentException(
                    "The value of 'cores_pattern' cannot be empty. "
                    "Use '<None>', or see coreadm(1) for valid patterns.")
            cores_dir = os.path.dirname(pattern)
            if cores_dir:
                try:
                    if not os.path.exists(cores_dir):
                        os.makedirs(cores_dir)
                    fd, temp = tempfile.mkstemp(
                        prefix=".build-cores-test", dir=cores_dir + "/")
                    os.remove(temp)
                except (OSError, IOError) as e:
                    raise OnBldEnvironmentException(
                        "Unable to use specified coreadm dir: %s" % e)
            ret, out, err = util.cmdline_run(
                "/usr/bin/coreadm -p %s %s" % (pattern, os.getpid()))
            if ret != 0:
                raise OnBldEnvironmentException(
                    "Unable to set coreadm pattern to %s" % pattern)
            # this isn't needed in the build environment
            del newenv["CORES_PATTERN"]

        if "LOGDIR" in newenv and not os.path.isdir(newenv["LOGDIR"]):
            try:
                os.makedirs(newenv["LOGDIR"], 0o755)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        # COMMON_BUILD_TASK is used by usr/src/Makefile to eliminate
        # duplication of effort during a nightly build.
        if self.common_task:
            newenv["COMMON_BUILD_TASK"] = ""

        # choose a sensible AKV_GATE_ID
        akv_gate_id = newenv["AKV_GATE_ID"]
        if akv_gate_id == "":
            akv_gate_id = "{0}_{1}".format(newenv["USER"], newenv["BASEWSDIR"])

        # Note that while mkisofs(8) requires the volume ID to be no longer
        # than 32 characters, distro_const will truncate if necessary.
        akv_gate_id = re.sub("[^a-zA-Z0-9_]", "_", akv_gate_id)
        newenv["AKV_GATE_ID"] = akv_gate_id

        # Clear out AKV_* versioning components that are empty.  These have
        # mostly reasonable defaults or are handled elsewhere.
        if newenv["AKV_COMPONENT"] == "":
            del newenv["AKV_COMPONENT"]

        if newenv["AKV_BUILD"] == "":
            del newenv["AKV_BUILD"]

        if newenv["AKV_BRANCH"] == "":
            del newenv["AKV_BRANCH"]

        if newenv["AKV_VERSION_BASE"] == "":
            del newenv["AKV_VERSION_BASE"]

        # Choose a sensible AKV_BUILD by incrementing the value from the
        # current proto area, if present. This allows successive AK builds
        # in the same workspace on the same day to always produce a *.pkg file
        # that can be upgraded to on a test machine running the previous build
        # from this workspace. The value is cached once set.
        #
        # Note that if 'fish_gk_images' is set or we have an existing value in
        # the environment, we do *not* bump the value from the previous build
        # on the assumption that we're either doing an official gatekeeper
        # build, or the user really wants the AKV_BUILD value they have set.
        #
        # There's a catch with this, in that if we're in a workspace where
        # a debug build was done, but not a non-debug build, and we're
        # now doing a non-debug build, we won't have the previous version.xml
        # to fall back on. This seems like a rare scenario for most developers.
        #
        if (newenv["FISH_GK_IMAGES"] == "false" and
                newenv["FISH_GK_NIGHTLY_IMAGES"] == "false" and
                "AKV_BUILD" not in newenv):

            if AKV_BUILD:
                newenv["AKV_BUILD"] = AKV_BUILD
            else:
                root = newenv.get("FISHROOT")
                if root is None:
                    root = newenv["ROOT"]
                version_path = os.path.join(
                    root,
                    "usr/lib/ak/metadata/appliance/nas/product/version.xml")
                if os.path.exists(version_path):
                    tree = lxml.etree.ElementTree(file=version_path)
                    el = tree.xpath(
                        ".//nvpair[@name='AKV_COMPONENT']/string")[0]
                    akv_comp = el.attrib["value"]
                    now_date = datetime.datetime.now().strftime("%Y.%m.%d")
                    # if we built on the same day, bump the AKV_BUILD value
                    if akv_comp == now_date:
                        el = tree.xpath(
                            ".//nvpair[@name='AKV_BUILD']/string")[0]
                        akv_branch = int(el.attrib["value"])
                        akv_branch += 1
                    else:
                        akv_branch = 1
                    newenv["AKV_BUILD"] = str(akv_branch)
                    AKV_BUILD = newenv["AKV_BUILD"]

        # Now that macros are expanded, we can expand this path. It's unlikely
        # that @CPIODIR@ was used in other macros, leave those alone anyway.
        if "CPIODIR" in newenv:
            newenv["CPIODIR"] = os.path.abspath(newenv["CPIODIR"])

        # Sanitize the dataset name since BASEWSDIR could have invalid
        # characters.
        newenv["DC_DATASET"] = re.sub("[@%]", "-", newenv["DC_DATASET"])

        if userenv_details:
            return newenv, whitelist_env_vars_user, blacklist_env_vars_user
        return newenv

    def __set_parfait(self, env):
        """Setup the environment to perform a parfait build/analysis."""

        if env.get("CLOSED_IS_PRESENT") == "no" and \
                "FISH_STANDALONE" not in env:
            raise OnBldEnvironmentException(
                "Unable to perform a parfait build without the closed tree")

        # Cause Parfait not to wrap native tool error messages with
        # explanations not to blame the parfait tool itself.
        env["PARFAIT_WRAPPER_QUIETNATIVE"] = "true"

        allowed_values = ["baseline", "upload"]
        if self.parfait_type not in allowed_values:
            raise OnBldEnvironmentException(
                "Unknown parfait_type value: %s. Allowed values are %s" %
                (self.parfait_type, ", ".join(allowed_values)))

        env["PARFAIT_NATIVEJAVAC"] = "@JAVA_ROOT@/bin/javac"
        env["PARFAIT_NATIVEJAR"] = "@JAVA_ROOT@/bin/jar"
        env["PARFAIT_BUILD"] = ""

    def __set_debug(self, env):
        """Setup the environment to perform a debug build."""

        env["DEBUG_BUILD"] = ""

        if "EXTRA_OPTIONS" in env:
            del env["EXTRA_OPTIONS"]
        if "EXTRA_CFLAGS" in env:
            del env["EXTRA_CFLAGS"]

        # If we've previously setup as a non-debug build, then reverse that
        # by removing the "-nd" suffix for environment entries that matter.
        for suffix, new_suffix, suffix_vars in [
                ("-nd", "", self.__suffix_vars),
                ("-nondebug", "-debug", self.__ext_suffix_vars)]:
            for nd_var in suffix_vars:
                var = env.get(nd_var, "")
                if not var:
                    continue
                env[nd_var] = re.sub("%s$" % suffix, new_suffix, var)

    def __set_nondebug(self, env):
        """Setup the environment to perform a non-debug build."""

        if "EXTRA_OPTIONS" in env:
            del env["EXTRA_OPTIONS"]
        if "EXTRA_CFLAGS" in env:
            del env["EXTRA_CFLAGS"]

        # Add -nd suffixes for the values that need them.
        for suffix, new_suffix, suffix_vars in [
                ("", "-nd", self.__suffix_vars),
                ("-debug", "-nondebug", self.__ext_suffix_vars)]:
            for debug_var in suffix_vars:
                var = env.get(debug_var, "")
                if not var:
                    continue
                # nothing more to do in this case
                if var.endswith(new_suffix):
                    env[debug_var] = var
                elif suffix == "-debug" and not var.endswith(new_suffix):
                    env[debug_var] = re.sub("%s$" % suffix, new_suffix, var)
                else:
                    env[debug_var] = "%s%s" % (var, new_suffix)

    def __set_release_type(self, env):
        """Setup the environment for the specified release type."""

        known_release_types = ["developer", "release", "sru", "idr"]
        rel_type = env["RELEASE_TYPE"]
        # Tidy up RELEASE_TYPE from the environment.
        del env["RELEASE_TYPE"]

        if rel_type == "developer":
            # Nothing to do here. This is the default in the Makefiles.
            return
        elif rel_type == "release":
            env["RELEASE_BUILD"] = ""
        elif rel_type == "sru":
            env["SRU_BUILD"] = ""
        elif rel_type == "idr":
            env["IDR_BUILD"] = ""
        else:
            raise OnBldEnvironmentException("release_type must be one of %s" %
                                            ", ".join(known_release_types))

        # All rel_type other than "developer" are a variant of a release build
        self.release = True

    def __set_incremental(self, env):
        """Do what it takes to do an incremental build."""
        pass

    def __unset_build_in_place(self, env, include_unset_vals=False):
        """Turn off build_in_place, making built objects appear in a separate
        directory, outside $SRC"""
        if "BIP" in env:
            if include_unset_vals:
                env["BIP"] = None
            else:
                del env["BIP"]

    def __set_build_in_place(self, env):
        """Turn on build_in_place, making built objects appear in a separate
        directory, outside $SRC"""
        env["BIP"] = "true"

    def __set_elfsigncmp(self, env):
        """Change $ELFSIGN to enable intra-build checks of elfsigned binaries
        against the unsigned version of that binary."""
        env["ELFSIGN"] = "@TOOLROOT@/opt/onbld/bin/elfsigncmp"

    def find_tools(self, log_versions=True):
        """Given a build environment dictionary, we check that the version of
        each tool is at least at the minimum level listed in
        self.__compiletools

        Versions found are emitted to the log if log_versions is True.
        Returns True if all tools were found at the required version or
        greater."""

        # If already checked, return cached value
        if self.tools_version_check is not None:
            return self.tools_version_check

        valid_toolversions = True
        env = self.get_build_environment()

        for tool in self.__compiletools:
            # We should not check for the presence of parfait if we're
            # not doing a parfait build or have disabled build checks.
            if tool.name == "parfait" and \
                    (not self.overrides.get("check.parfait", True) or
                        not self.check):
                continue
            if tool.name == "fortify" and \
                    (not self.overrides.get("check.fortify", True) or
                        not self.check):
                continue

            try:
                cmd = self.__cfg.get("paths", tool.name)
            except configparser.NoOptionError:
                cmd = tool.default

            # Expand macros used in the tools paths. These were recursively
            # expanded in env as part of get_build_environment() so we don't
            # need to go to the same lengths here.
            osenv_macro_re = re.compile("@([_A-Z0-9a-z]*)@")
            macros = osenv_macro_re.findall(cmd)
            for macro in macros:
                expanded_macro = env.get(macro)
                if not expanded_macro:
                    logger.error("unable to expand macro @%s@" % macro)
                    valid_toolversions = False
                cmd = cmd.replace("@%s@" % macro, expanded_macro)

            args = [cmd, tool.version_flag]
            if not os.path.exists(cmd):
                logger.error(
                    "unable to find command '%s' for %s" % (cmd, tool.name))
                valid_toolversions = False
                continue
            ret, out, err = util.cmdline_run(  # pylint: disable=W0612
                " ".join(args), raise_error=False, env_arg=env)

            infoline = out.rstrip().lstrip() + err.rstrip().lstrip()

            cmdre = self.__compiletools_re[tool.name]
            if not cmdre:
                logger.warning("no version regexp found for %s" % tool.name)
                valid_toolversions = False
                continue

            cmdreval = re.match(cmdre, infoline)
            if cmdreval is not None:
                # The version comparison here is only a sanity check - not all
                # tools use dotted integers and it's not necessarily true that
                # a tool version greater than the one we require is actually
                # the one we want to build with. Once gate machines use local
                # tools and have pkg:/developer/opensolaris/osnet installed,
                # and all required tools packages are included there, this
                # check will be less important.
                if cmdreval.group(tool.group_num) < tool.minver:
                    logger.error(
                        "version of required tool %s is %s, should be %s" %
                        (tool.name, cmdreval.group(tool.group_num),
                            tool.minver))
                    valid_toolversions = False
                else:
                    if log_versions:
                        logger.info(
                            "version of required tool %s is\n%s" %
                            (tool.name, infoline))
            else:
                logger.warning(
                    "Unable to determine version information for %s\n"
                    "Version string was: \"%s\"\n\n"
                    "Please check for recent flag days that may "
                    "describe this issue.\n"
                    "If you have further questions, mail %s for advice.\n"
                    "Continue at your own risk.\n" %
                    (cmd, infoline, GATEKEEPER_MAILTO))

        if not valid_toolversions:
            logger.error(
                "This could indicate an out of date config file.")

        # Cache the result for next time
        self.tools_version_check = valid_toolversions

        return valid_toolversions

    def verify_bldrc_consistency(self, bldrc=None):
        """A method to check that the supplied bldrc file is consistent with
        the definitions in self.__environment and self.__compiletools. This
        should be called by consumers before performing any operations with the
        OnBldEnvironment.
        """

        # if we weren't passed an bldrc file, we know that the default config
        # is valid, since we're in charge of it.
        if not bldrc:
            return True

        valid_rc = True
        cfgfile = configparser.SafeConfigParser()
        cfgfile.read(bldrc)
        cfg_sections = cfgfile.sections()

        compiletools = {}
        for tool in self.__compiletools:
            if tool.bldrc:
                compiletools[tool.name] = tool

        # check values from the cfg file against those in self.__environment
        for sect in cfg_sections:
            if sect == SECT_CHECKOVERRIDES:
                continue
            for key, value in cfgfile.items(
                    sect, raw=True):  # pylint: disable=W0612
                if key.upper() not in self.__environment and \
                        key not in compiletools:
                    logger.error(
                        "bldrc file error: key/value '%s=%s' is not supported "
                        "by current version of build(1)" % (key, value))
                    valid_rc = False

                env_entry = self.__environment.get(key.upper())
                if env_entry and env_entry.section != sect:
                    logger.error(
                        "bldrc file error: key/value '%s=%s' can only appear "
                        "in section %s, not %s" %
                        (key, value, env_entry.section, sect))
                    valid_rc = False

        # check self.__environment and self.__compiletools against values in
        # the cfgfile
        for envvar in self.__environment:
            env_item = self.__environment[envvar]
            # some self.__environment entries should not be written to the
            # bldrc
            if not env_item.section:
                continue
            # some self.__environment entries don't have to appear in the bldrc
            if env_item.section not in SECTS:
                continue
            if env_item.section not in cfg_sections:
                logger.error(
                    "bldrc file error: section '%s' was not found in bldrc "
                    "file" % env_item.section)
                valid_rc = False
            try:
                val = cfgfile.get(
                    env_item.section, envvar.lower())  # pylint: disable=W0612

            except (AttributeError, configparser.NoSectionError,
                    configparser.NoOptionError) as exc:
                logger.error("bldrc file error: %s" % exc)
                valid_rc = False

        for toolname in compiletools:
            try:
                val = cfgfile.get(SECT_PATHS, toolname)
            except (AttributeError, configparser.NoSectionError,
                    configparser.NoOptionError) as exc:
                logger.error("bldrc file error: %s" % exc)
                valid_rc = False
        return valid_rc

    def merge_overrides(self, runtime_overrides):
        """Given a dictionary of containing keys that map to the names of
        check.* tasks, with values of True or False, add that to the
        dictionary of overrides set in this OnBldEnvironment object."""
        # Sanity check whether we're allowed to process overrides
        if not self.check:
            raise OnBldEnvironmentException(
                "Cannot use check.* overrides when global 'check' "
                "option is 'false'")
        self.overrides.update(runtime_overrides)

        # Validate the overrides that are now in place
        for key in self.overrides:
            if key.endswith(".must_pass"):
                check_key = key.replace(".must_pass", "", 1)
                val = self.overrides.get(check_key)
                if val and val == "false":
                    raise OnBldEnvironmentException(
                        "Cannot set %s when %s is false." %
                        (key, check_key))
            elif not self.check and self.overrides[key]:
                raise OnBldEnvironmentException(
                    "'%s' cannot be set to 'true' when the global "
                    "'check' option is 'false'." % key)

    def generate_bldrc(
            self, gk_cfgfile=False, ksplice_cfgfile=False,
            cwd=None, options=None, overrides=None, cmd=None):
        """Generate a bldrc file with default values. If gk_cfgfile is True,
        then include gatekeeper-specific values. If ksplice_cfgfile is True,
        then include ksplice-specific values. If 'cwd' is not None, then use
        this for the value of 'ws'.

        'options', if passed should be a dictionary representing key/value
        pairs that are used to override the defaults that are generated.

        'overrides', if passed should be a dictionary with keys
        'check.*' or "check.*.must_pass mapping to boolean True or False.
        """

        # This code pre-dates Python 2.7, and ConfigParser wasn't built on
        # OrderedDict back then, so we needed a way to produce deterministic
        # config files. For now, the below works fine as we don't need any
        # fancy escaping of bldrc values that switching to a pure configparser
        # backend would bring.
        strio = StringIO()
        user, logname, email = self.get_userinfo()
        if not cmd:
            raise OnBldEnvironmentException(
                "No 'cmd' parameter was passed to generate_cmd")

        # ignore any $WS path
        if "usr/src/tools/scripts/build.py" in cmd[0]:
            cmd[0] = "usr/src/tools/scripts/build.py"

        generate_cmd = " ".join(cmd)
        strio.write(BLDRC_BANNER % {
            "user": "%s <%s>" % (user, email),
            "date": datetime.datetime.strftime(
                datetime.datetime.now(), log.datefmt),
            "generate_cmd": generate_cmd})
        sects = copy.copy(SECTS)
        if gk_cfgfile:
            sects.append(SECT_GKOVERRIDES)

        if overrides:
            sects.append(SECT_CHECKOVERRIDES)

        if not options:
            options = {}

        # Check that options were passed to include in the config file are
        # known to this version of build.py
        if options:
            env_keys = [item.name.lower() for item in self.__env_list]
            env_keys.extend(
                [tool.name for tool in self.__compiletools if tool.bldrc])

            invalid_options = []
            for key in options:
                if key not in env_keys:
                    invalid_options.append(key)
            if invalid_options:
                raise OnBldEnvironmentException(
                    "The following key(s) are not recognised: %s" %
                    ", ".join(invalid_options))

        for section in sects:
            strio.write(DOC_SECTS[section])
            strio.write("[%s]\n" % section)

            for env_item in self.__env_list:
                if env_item.section != section:
                    continue
                key = env_item.name.lower()
                value = options.get(key, env_item.default)
                # make the workspace the cwd if set
                if (key == "ws" and cwd and key not in options and
                        value is None):
                    value = cwd
                elif value is None:
                    value = "<None>"

                strio.write("%s = %s\n" % (key, value))

            if section == SECT_PATHS:
                for tool in self.__compiletools:
                    if not tool.bldrc:
                        continue
                    tool_val = options.get(tool.name, tool.default)
                    strio.write("%s = %s\n" % (tool.name, tool_val))

            def override_func(item):
                """A key function to allow us to sort the <task>.must_pass
                keys below the rest of the overrides."""
                if "must_pass" in item:
                    return "z" + item
                return item

            if section == SECT_CHECKOVERRIDES:
                prev_key = ""
                for key in sorted(overrides, key=override_func):
                    if "buildtest." in key and \
                            "ONBLD_BUILDTEST" not in os.environ:
                        continue
                    # Add a comment spacer when starting the must_pass subsect
                    if ".must_pass" in key and ".must_pass" not in prev_key:
                        strio.write(
                            "\n#\n# The following keys determine whether "
                            "failing checks influence the exit\n"
                            "# code of build(1)\n"
                            "#\n")
                    if overrides[key] not in [True, False]:
                        raise OnBldEnvironmentException(
                            "[overrides] values must be either 'true' or "
                            "'false'.")
                    strio.write(
                        "%s = %s\n" % (key, str(overrides[key]).lower()))
                    prev_key = key

            strio.write("\n")

        return strio.getvalue()

    def get_userinfo(self):
        """Extract user, logname and email address from $HOME/.hgrc, consulting
        $LOGNAME or $USER for the user name."""

        username = None
        useremail = None

        cmd = "/usr/bin/hg config email.from"
        ret, out, err = util.cmdline_run(cmd)
        if ret is 0:
            useremail = out.strip()

        cmd = "/usr/bin/hg config ui.username"
        ret, out, err = util.cmdline_run(cmd)
        if ret is 0:
            username = out.strip()

        # Some usernames have config file entries like:
        # username = Joe Bloggs <joe.bloggs@oracle.com>
        if username and len(username.split("<")) >= 2:
            if not useremail:
                useremail = username.split("<")[1].rstrip(">")
            username = username.split("<")[0].rstrip()

        logname = os.environ.get("LOGNAME", "")
        if not logname:
            logname = os.environ.get("USER", "")
        if not logname:
            logname = pwd.getpwuid(os.getuid())[0]

        # Make one last attempt to obtain an email address
        if useremail is None:
            cmd = "/usr/bin/ldaplist -l passwd %s | "\
                "/usr/bin/awk '/mail:/ {print $2}'" % logname
            ret, out, err = util.cmdline_run(cmd, shell=True)
            if ret == 0:
                useremail = out.rstrip()

        return username, logname, useremail

    def generate_bldenv_script(
            self, env, user_env, blacklist_user_env, outpath,
            source_dotfiles=False):
        """Create a shell script fragment that can be called from build(1).
        The contents of the build environment are determined by the
        bldrc file, and self.get_build_environment().

        If source_dotfiles is set, we source the user's shell dotfiles. This
        **should not** be used for production builds as a user's shell
        environment cannot be guaranteed to be clean of settings which may
        influence the build. This includes unsetting $PYTHONNOUSERSITE
        so that power-users can avail of local python modules.
        """

        # Shell use to set values for different shells. We deal with
        # Bourne and Csh variants.

        setenv_sh = "export %(key)s='%(val)s'\n"
        alias_sh = "alias %(key)s='%(val)s'\n"
        source_sh = ". %s\n"
        unsetenv_sh = """
# Unset all environment variables
for var in $(python -c '
import os
import re
for e in os.environ:
    if not (re.match(e, "PS[0-9]") or
            re.match(e, "RPS[0-9]") or e == "SPROMPT"):
        print(e)
'); do
        %s
        unset $var
done
"""
        bypass_sh = """        if [[ "$var" == "%s" ]]; then
                continue
        fi
"""
        setenv_csh = "setenv %(key)s '%(val)s'\n"
        alias_csh = "alias %(key)s '%(val)s'\n"
        source_csh = "source %s\n"
        unsetenv_csh = """
# Unset all environment variables. Persuading csh to deal with inline Python
# is gruesome, so we'll create a herefile for that instead.
set buildenv_py=`mktemp -t build.env.XXXXXX.py`
cat > $buildenv_py << EOF
#!/usr/bin/python
import os
import re
for e in os.environ:
    if not (re.match(e, "PS[0-9]") or
            re.match(e, "RPS[0-9]") or e == "SPROMPT"):
        print(e)
EOF

chmod 700 $buildenv_py
foreach var (`$buildenv_py`)
        %s
        unsetenv $var
end
/usr/bin/rm -f $buildenv_py
unset buildenv_py
"""
        bypass_csh = """      if ( "$var" == "%s" ) then
                 continue
        endif
"""
        # determine which of the above shell fragments to use
        # to clear the environment.
        if env["SHELL"].endswith("csh"):
            unset_env = unsetenv_csh
            set_env = setenv_csh
            bypass = bypass_csh
            alias = alias_csh
            source = source_csh
        else:
            unset_env = unsetenv_sh
            set_env = setenv_sh
            bypass = bypass_sh
            alias = alias_sh
            source = source_sh

        bypasses = "# Environment variables deemed safe to inherit\n"
        for var in self.__whitelist_env_vars:
            bypasses = "%s%s" % (bypasses, bypass % var)

        if not self.bldrc:
            bldrc_path = "[ no path ]"
        else:
            bldrc_path = self.bldrc

        shell_basename = os.path.basename(env["SHELL"])
        user, logname, email = self.get_userinfo()
        with open(outpath, "w") as env_script:
            env_script.write("#!%s\n" % env["SHELL"])
            header = """
#
# This is an OS/Net build environment created by
# %s on %s
# from bldrc file: %s
#
"""
            env_script.write(
                header % (
                    "%s <%s>" % (user, email),
                    datetime.datetime.isoformat(datetime.datetime.now()),
                    bldrc_path))

            # When in "developer mode", we want to preserve as much of the
            # user shell as possible. We do this by sourcing their dotfile
            # before writing shell fragments that obliterate their
            # environment. This is potentially toxic to the build, but it's
            # also optional, and not the default.
            if source_dotfiles:
                if "PYTHONNOUSERSITE" in env:
                    del env["PYTHONNOUSERSITE"]

                rcfile = None
                if shell_basename == "bash":
                    rcfile = "%s/.bashrc" % env["HOME"]

                elif shell_basename.startswith("ksh"):
                    rcfile = "%s/.kshrc" % env["HOME"]

                elif shell_basename in ["tcsh", "csh", "pfcsh", "pftcsh"]:
                    # Being a profile shell is irrelevant here.
                    shell_shortname = shell_basename.replace("pf", "")
                    rcfile = os.path.join(
                        env["HOME"], ".%src" % shell_shortname)

                if rcfile and os.path.exists(rcfile):
                    # Allow csh users access to more of their environment
                    if shell_basename in ["tcsh", "csh", "pfcsh", "pftcsh"]:
                        env_script.write("set tmphome = $home\n")
                        env_script.write("set home = %s\n" % env["HOME"])
                        env_script.write(source % rcfile)
                        env_script.write("set home = $tmphome\n")
                    else:
                        env_script.write(source % rcfile)

            env_script.write(unset_env % bypasses)
            for key in sorted(env.keys()):
                if key.upper() in user_env:
                    continue
                if key.upper() in self.__whitelist_env_vars:
                    continue
                # An environment entry with a 'None' value means that the
                # option should not be set in the environment.
                if env[key] is not None:
                    # when in developer mode, we add the user's current PATH
                    # to the end of the build(1) supplied path
                    if source_dotfiles and key == "PATH":
                        env_script.write(
                            set_env % {
                                "key": key,
                                "val": env[key] + ":%s" % os.environ["PATH"]})
                    else:
                        env_script.write(
                            set_env % {"key": key, "val": env[key]})

            env_script.write("\n")
            # Emit a list of the inherited or user-specified environment
            # variables.
            for key in sorted(env.keys()):
                if key.upper() in user_env:
                    env_script.write(
                        "# User-specified environment variable : "
                        "%(key)s=%(val)s\n" %
                        {"key": key.upper(), "val": env[key.upper()]})
                    continue
                if key.upper() in self.__whitelist_env_vars:
                    env_script.write(
                        "# Inheriting from the user environment: "
                        "%(key)s=%(val)s\n" %
                        {"key": key.upper(),
                            "val": "\\n".join(env[key.upper()].split("\n"))})

            if blacklist_user_env:
                env_script.write(
                    "\n# The following variables were "
                    "specifically unset by the user:\n")
                for key in blacklist_user_env:
                    env_script.write("# %s\n" % key)
            env_script.write("\nmkdir -p %s\n" % env["TMPDIR"])
            env_script.write("cd %s\n" % env["SRC"])

            # The build uses stash(1) to install files into $ROOT for the most
            # part, which sets file permissions. For things that don't or can't
            # use stash, ensure we use a consistent umask. We do this here
            # so that it overrides any dotfile settings that we may have
            # sourced.
            env_script.write("umask 0022\n")

            if shell_basename in ["csh", "tcsh", "pfcsh", "pftcsh"]:
                if source_dotfiles:
                    # try to deal with workspaces that are in symlink farms
                    if "tcsh" in shell_basename:
                        env_script.write("set symlinks=chase\n")
                    else:
                        env_script.write("set hardpaths\n")
                    #
                    # When detecting csh or tcsh and using source_dotfiles,
                    # build(1) changes $HOME, after writing this env_script,
                    # to point to the build(1) temporary dir containing this
                    # env_script, renamed to ".cshrc" or ".tcshrc", causing us
                    # to execute it at startup.
                    #
                    # This is not elegant. But, csh, so meh.
                    #
                    shell_cmd = ""
                else:
                    # csh uses -f to not read any shell rc files
                    # in which case we run the shell here and rely on
                    # build(1) to run this fragment using os.execvpe, which
                    # sets the environment
                    shell_cmd = "%s -f" % env["SHELL"]
                env_script.write("%s\n" % shell_cmd)
            else:
                # try to deal with workspaces that are in symlink farms
                env_script.write(alias % {"key": "cd", "val": "cd -P"})
                # Do our best to clean up afterwards. csh-based shells have
                # no ability to trap anything other than SIGINT, unfortunately.
                env_script.write(
                    "trap \"/usr/bin/rm -rf %s %s %s\" EXIT\n" %
                    (env["TMPDIR"], outpath, outpath))

            # Throw an exception if not using a whitelisted shell
            shells = ["bash", "csh", "ksh", "ksh93", "pfbash", "pfcsh",
                      "pfksh", "pfksh93", "pfsh", "pftcsh",
                      "pfzsh", "sh", "tcsh", "zsh", ]
            basename = env["SHELL"].split("/")[-1]
            if basename in shells:
                return
            raise OnBldEnvironmentException(
                "%s is not supported." % env["SHELL"])

    def __get_fish_env(self, env):
        """Given a standard ON build environment, apply the necessary changes
        to allow us to build the usr/fish subrepo. This should be called before
        macro expansion.

        The fish environment gets its own $ROOT area, separate from the rest
        of ON since that layout is used to construct the various AK cpio
        archives. STUBROOT is shared. The fish environment also gets its own
        PKGARCHIVE.
        """

        # trying to run this method twice would be bad, so guard against that.
        # This should never happen and indicates an error in build.py.
        if "IS_FISH_ENV" in env:
            raise OnBldEnvironmentException(
                "get_fish_env() has already been called!")

        if env.get("FISH_IS_PRESENT", "no") == "no" and \
                "FISH_STANDALONE" not in env:
            raise OnBldNoFishEnvironmentException(
                "Cannot run AK-specific build tasks: "
                "usr/fish repository is missing,  "
                "'fish_is_present' is set to 'no', "
                "or this is not a standalone usr/fish repository.")

        newenv = copy.copy(env)
        if env.get("FISH_STANDALONE", "no") == "yes":
            newenv["FISH_IS_PRESENT"] = "yes"

        root = env["ROOT"]
        root_base = os.path.basename(root)
        root_dir = os.path.dirname(root)

        pkgarchive = env["PKGARCHIVE"]
        pkgarchive_base = os.path.basename(pkgarchive)
        pkgarchive_dir = os.path.dirname(pkgarchive)

        root = os.path.join(root_dir, FISH_PREFIX + root_base)
        pkgarchive = os.path.join(
            pkgarchive_dir, FISH_PREFIX + pkgarchive_base)

        # Derive AKARCHIVEDIR and CPIODIR from PKGARCHIVE. Note we can't use
        # os.path.abspath here since we haven't yet expanded the macros, and
        # Python helpfully prepends cwd if pkgarchive_dir doesn't start with
        # '/', which is the case when we haven't expanded macros.
        ak_archivedir = os.path.join(pkgarchive_dir, "..", "archives")
        newenv["AKARCHIVEDIR"] = ak_archivedir
        newenv["CPIODIR"] = os.path.join(ak_archivedir, pkgarchive_base)
        newenv["ROOT"] = root
        newenv["PKGARCHIVE"] = pkgarchive
        newenv["SRC"] = os.path.join(env["WS"], "usr/fish")

        if newenv.get("FISH_STANDALONE"):
            newenv["SRC"] = env["WS"]
            newenv["PATH"] = env["PATH"].replace(env["TOOLROOT"], "")
            newenv["TOOLROOT"] = ""
            if newenv["FW_HOOKS"] == env["WS"]:
                raise OnBldEnvironmentException(
                    "For standalone fish repository builds, 'fw_hooks' must "
                    "be set in the bldrc file to something other than '@WS@'. "
                    "Perhaps you meant to set this to /ws/ak-9-on-gate ?")
        else:
            newenv["FW_PROTO"] = env["ROOT"]
            # Obtaining the ON stubroot is tricky. Normally set to @ROOT@_stub,
            # that would return the usr/fish stub if we use that here. Instead
            # use dirname and basename of @ROOT@ computed earlier. Assume
            # that a gatekeeper hasn't overridden 'stubroot' in the bldrc.
            newenv["FW_PROTO_STUB"] = "{0}/{1}_stub".format(
                root_dir, root_base)

        newenv["CODEMGR_WS"] = newenv["SRC"]
        newenv["IS_FISH_ENV"] = "true"
        return newenv

    def __deepcopy__(self, memo):
        """We need to provide our own deepcopy since ConfigParser objects can't
        be deepcopied in Python 2.7 (issue 16058)"""
        cls = self.__class__
        result = cls.__new__(cls)
        for key, val in list(self.__dict__.items()):
            if key == "_OnBldEnvironment__cfg":
                cfg = configparser.RawConfigParser()
                sio = StringIO()
                self.__cfg.write(sio)
                sio.seek(0)
                cfg.read_file(sio)
                setattr(result, key, cfg)
            else:
                setattr(result, key, copy.deepcopy(val, memo))
        return result
