#!/bin/bash

#
# Copyright (c) 2014, 2020, Oracle and/or its affiliates.
#

# This script creates two repositories. Repo 'a' serves as a parent repository
# to it's clone 'b'. During each test we create fresh clone 'b', make some
# modifications to it and observe whether 'hg branchchk' works as supposed

# Do not read any personal .hgrc configuration files; see hg(1)
# This means that we have to add '-u test' to every commit, otherwise
# mercurial complains
export HGRCPATH=/dev/null

# Disable any local mercurial output configurations; see hg(1)
export HGPLAIN=1

# Some sensible defaults
export PATH=/usr/bin
export LC_ALL=C

function fail {
	print -u2 "$@"
	exit 1
}

# Where to find our cadmium extension
if (( $# > 1 )); then
	fail "usage: branchchk_test [path_to_cdm.py]"
elif (( $# == 1 )); then
	CDM_PATH=$(realpath "$1")
	if [[ $? != 0 ]]; then
		fail "cannot get absolute path of $1"
	fi
else
	CDM_PATH=/opt/onbld/lib/python/onbld/hgext/cdm.py
fi

#
# Create a temporary directory for repositories created during test run and
# change current directory to it. The mktemp utility respects TMPDIR that is
# set if the test is run as part of 'test' subcommand of build(1).
#
OUTDIR=$(mktemp -d)
if [[ $? != 0 ]]; then
	fail "cannot create temporary directory"
fi
cd "$OUTDIR"
if [[ $? != 0 ]]; then
	fail "cannot change directory to $OUTDIR"
fi

function cleanup {
	cd - > /dev/null && rm -rf "$OUTDIR"
}

trap cleanup EXIT

function branchchk {
	hg --config "extensions.cdm=$CDM_PATH" branchchk
}

function glog {
	# Display 'hg log' in graphical representation. This helps understanding
	# mercurial output in the case of any issues
	hg --config 'extensions.graphlog=' glog
}

# The default repository should have three named branches, 'default', '"second"'
# and "'third'":
#
# @  default
# |
# | o  "second"
# | |
# | | o  'third'
# | |/
# | |
# | o  "second"
# |/
# |
# o  default
function init {
	hg version
	rm -rf a
	hg init a

	(
		cd a
		touch file
		hg commit -u test -A -m default
		hg branch '"second"'
		hg commit -u test -m 'create "second"'
		hg branch "'third'"
		hg commit -u test -m "create 'third'"
		hg update -C '"second"'
		touch 2
		hg commit -u test -A -m 2
		hg update -C default
		touch default
		hg commit -u test -A -m default
		glog
	)
}

# Create one changesets
function single_commit {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	touch a1
	hg commit -u test -A -m a1

	glog
	branchchk
}

# Create two changesets
function two_commits {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	touch a1
	hg commit -u test -A -m a1
	touch a2
	hg commit -u test -A -m a2

	glog
	branchchk
}

# Create two changesets with common parent
function two_nodes {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	typeset TIP
	TIP=$( hg tip --template '{rev}' )

	touch a1
	hg commit -u test -A -m a1

	hg update -C $TIP # Back to the parent change
	touch a2
	hg commit -u test -A -m a2

	glog
	branchchk
}

# Create two changesets with common parent and then
# switch to different branch
function two_nodes_diff_branch {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	typeset TIP
	TIP=$( hg tip --template '{rev}' )

	touch a1
	hg commit -u test -A -m a1

	hg update -C $TIP # Back to the parent change
	touch a2
	hg commit -u test -A -m a2

	hg update -C '"second"'

	glog
	branchchk
}

# Create several changesets, each on separate branch
function multiple_commits {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	touch a1
	hg commit -u test -A -m a1

	hg update -C '"second"'
	touch a2
	hg commit -u test -A -m a2

	hg update -C "'third'"
	touch a3
	hg commit -u test -A -m a3

	glog
	branchchk
}

# Create several changesets, each on separate branch. Make all but one of the
# created changesets have secret phase
function multiple_commits_secret {
	# Older mercurial does not have phases
	hg help phases &> /dev/null || return 0

	rm -rf b; hg clone a b; cd b

	hg update -C default
	touch a1
	hg commit -u test -A -m a1

	hg update -C '"second"'
	touch a2
	hg commit -u test -A -m a2
	hg phase --force --secret tip

	hg update -C "'third'"
	touch a3
	hg commit -u test -A -m a3
	hg phase --force --secret tip

	glog
	branchchk
}

# Create 'Y' structure, that is changeset which has two children changesets.
# Similar situation to 'two_nodes'. In this case though even the common parent
# is new local changeset. Then merge is called, but it is not committed.
function merge_uncommitted {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	touch middle
	hg commit -u test -A -m middle

	typeset TIP
	TIP=$( hg tip --template '{rev}' )

	touch a1
	hg commit -u test -A -m a1
	hg update -C $TIP

	touch a2
	hg commit -u test -A -m a2
	hg merge

	glog
	branchchk
}

# Merge two local heads and commit the merge
function merge {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	typeset TIP
	TIP=$( hg tip --template '{rev}' )

	touch a1
	hg commit -u test -A -m a1
	hg update -C $TIP

	touch a2
	hg commit -u test -A -m a2

	hg merge
	hg commit -u test -m merge

	glog
	branchchk
}

# Add new branch originating at revision 0
function from_zero {
	rm -rf b; hg clone a b; cd b

	hg update -C null # go to the state before first commit

	touch zero
	hg commit -u test -A -m zero

	glog
	branchchk
}

# Create a new named branch
function new_branch {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	hg branch err
	hg commit -u test -A -m err

	glog
	branchchk
}

# Create a new named branch with additional local changeset
function new_branch_and_commit {
	rm -rf b; hg clone a b; cd b

	hg update -C default
	hg branch err
	hg commit -u test -A -m err

	touch newfile
	hg commit -u test -A -m newfile

	glog
	branchchk
}

# Close existing named branch
function closed_branch {
	rm -rf b; hg clone a b; cd b

	hg update -C '"second"'
	hg commit -u test -m ende --close-branch

	echo remaining branches:
	hg branches

	branchchk
}

# Test when there is MQ enabled
function mq {
	rm -rf b; hg clone a b; cd b

	touch file_mq
	hg --config extensions.mq= qnew -u test -f test_mq
	hg --config extensions.mq= qapplied
	hg --config extensions.mq= qpop -a

	branchchk
}

# Make sure that branchchk works even on un-modified workspace
function clean_ws {
	rm -rf b; hg clone a b; cd b

	branchchk
}

# Branchchk should work if CWD is set to new directory created using "hg cp"
function hg_cp {
	rm -rf b; hg clone a b; cd b

	hg update -C '"default"'
	mkdir dirA
	touch dirA/file
	hg commit -u test -A -m dirA
	hg cp dirA dirB
	hg commit -u test -A -m dirB
	cd dirB

	branchchk
}

# Execute the given test.
#
# $1 - name of the test
# $2 - expected 'hg branchchk' return code
#
# If the variable DEBUG is set, be verbose. Otherwise be verbose only if the
# test fails.
function testme {
	echo "running test '$1' - expected result is $2"
	typeset ERROR_LOG_PRINTED= # remember that we already printed the verbose log
	OUT=$( "$1" 2>&1 )
	typeset RET=$?

	if [[ $2 != $RET ]]; then
		# Check return value
		echo "Error ($1) - return should be '$2' but is '$RET'"
		echo "$OUT"
		ERROR_LOG_PRINTED=1
		EXIT_STATUS=1
	fi

	if echo "$OUT" | grep Traceback >/dev/null; then
		# Make sure we catch that python crashed
		echo "Error ($1) - python crashed"
		echo "$OUT"
		ERROR_LOG_PRINTED=1
		EXIT_STATUS=1
	fi

	if [[ -n ${DEBUG:-} ]] && [[ -z "$ERROR_LOG_PRINTED" ]]; then
		echo "$OUT"
	fi

	if [[ -z "$ERROR_LOG_PRINTED" ]]; then
		echo "... OK"
	fi

	if [[ -n "${DEBUG:-}" ]]; then
		echo "=============================================================================="
	fi
}

init


EXIT_STATUS=0 # set to 1 by function testme when thest fails

#DEBUG=1 # Enable to see verbose output

testme single_commit 0
testme two_commits 0
testme two_nodes 1
testme two_nodes_diff_branch 1
# It is questionable whether merge and merge_uncommitted should not actually
# return 1. But since we have special check 'mergechk' for that, it's better
# not to warn the user twice
testme merge 0
testme merge_uncommitted 0
testme from_zero 1
testme new_branch 1
testme new_branch_and_commit 1
testme closed_branch 1
testme multiple_commits 1
testme multiple_commits_secret 0
testme clean_ws 0
testme mq 0
testme hg_cp 0

exit $EXIT_STATUS
