#!/usr/bin/ksh

#
# Copyright (c) 2008, 2021, Oracle and/or its affiliates.
#

#
# This is the cadtest front end.
#

PYTHON=/usr/bin/python2.7

DEFAULT_HG=/usr/bin/hg
DEFAULT_CDM=/opt/onbld/lib/python/onbld/hgext/cdm.py
DEFAULT_ACTIVE=/opt/onbld/bin/hg-active

HARNESSDIR=$(cd $(dirname $0) && pwd)
DTEST=${HARNESSDIR}/dtest

function usage {
	cat <<# EOT
	Usage: cadtest [ options ] -t <name> [ tests definition dir ]
	  -h		Usage (this message)
	  -d <dir>	Working directory for repos and results
	  -t <dir>	Name for the results directory
	  -k		Keep temporary files even if the tests succeeded
	  -n		Show config values and exit
	  -r		Ignore rc file (\$HOME/.hgrc)
	  -a <path>	Override location of 'hg-active'
	  -c <path>	Override location of 'cdm.py'
	  -m <path>	Override location for 'hg'

	Defaults:
	  hg defaults to $DEFAULT_HG
	  cdm.py is found by searching:
		-c path option
		Current ON workspace (\$WS/usr/src/tools/onbld/hgext/cdm.py)
		hgext.cdm setting in \$HOME/.hgrc file
		$DEFAULT_CDM
	  hg-active is found by searching:
		-a path option
		Current ON workspace (\$WS/usr/src/tools/scripts/hg-active.py)
		$DEFAULT_ACTIVE

EOT
}


while getopts a:c:d:hkm:nrt: opt; do
	case $opt in
	a)	HGACTIVE_FILE=$OPTARG
		;;
	c)	CDMPY=$OPTARG
		;;
	d)	WORKINGDIR=$OPTARG
		;;
	k)	KEEPREPOS=1
		;;
	m)	HG=$OPTARG
		;;
	n)	dryrun=1
		;;
	r)	ignore_hgrc=1
		;;
	t)	RESULTDIR=$OPTARG
		;;
	h)	usage
		exit 0
		;;
	?)	usage
		exit 2
		;;
	esac
done

shift $(($OPTIND - 1))

HG=${HG:-/usr/bin/hg}

if [[ ! -x $(which $HG) ]]; then
	print -u2 "Could not find Hg executable '$HG'"
	exit 2
fi

echo "Using HG = $HG"

# If the user asked to ignore hgrc, set HGRCPATH empty to disable it.
if [ -n "$ignore_hgrc" ]; then
	export HGRCPATH=
fi

# Avoid untrusted warnings
HG="$HG --config ui.report_untrusted=false"

# If there is no Mercurial username (or ignore_hgrc is in effect), supply one.
if [ -z "$($HG config ui.username)" ]; then
	HG="$HG --config ui.username=unit_tests"
fi

#
# If the user supplied CMDPY, use it.
# If not, and there is a cdm found in the current workspace, use it.
# Otherwise, if the user has a hgext.cdm setting, use that.
# Failing that, fall back on the copy in /opt.
#
if [ -z $CDMPY ]; then
	CDMPY=$WS/usr/src/tools/onbld/hgext/cdm.py
	if [ ! -r $CDMPY ]; then
		CDMPY=$($HG config extensions.hgext.cdm)
		if [ -z "$CDMPY" ]; then
			CDMPY=/opt/onbld/lib/python/onbld/hgext/cdm.py
		fi
	fi
fi
HG="$HG --config extensions.hgext.cdm=$CDMPY"

BASEDIR=${WORKINGDIR:-$HARNESSDIR/results/$RESULTDIR}
export BACKUPDIR=$BASEDIR/cdm.backup.$$/
export HG="$HG --config cdm.backupdir=$BACKUPDIR"

#
# If the user supplied HGACTIVE_FILE, use it.
# If not, and it is found in the current ON workspace, use it.
# Otherwise that, fall back on the copy in /opt.
#
if [ -z $HGACTIVE_FILE ]; then
	HGACTIVE_FILE=$WS/usr/src/tools/scripts/hg-active.py
fi
if [ -r $HGACTIVE_FILE ]; then
	HGACTIVE="$PYTHON $HGACTIVE_FILE"
else
	HGACTIVE_FILE=$DEFAULT_ACTIVE
	HGACTIVE=$DEFAULT_ACTIVE
fi
export HGACTIVE

export REPOS=$BASEDIR/repos.$$
export BASEWS=$REPOS/basews
export EMPTYWS=$REPOS/emptyws
export HARNESSDIR

echo "Mercurial version:"
$HG version | head -1
echo

echo "HGUSER: $($HG config ui.username)"
echo

echo "cdm under test: $CDMPY"
echo

echo "hg-active selected: $HGACTIVE_FILE"
echo

# Config options printed. If this is a dryrun, exit now without doing anything.
if [ -n "$dryrun" ]; then
	exit 0
fi

if [[ -z "$RESULTDIR" ]]; then
	print -u2 "You must specify a test result name, with -t"
	exit 2
fi

print "Creating baseline...\c "
mkdir -p $REPOS
mkdir -p $BASEWS
cd $BASEWS
$HG init
for elt in a b c d e f g h; do
    echo "This is file $elt" >> $elt
    $HG add $elt
done
$HG ci -m "Baseline"
cd $BASEDIR
$HG init $EMPTYWS
print "done"

if [[ -z $1 ]]; then
	TESTS=$HARNESSDIR/tst
else
	TESTS="$@"
fi


if ! $DTEST $TESTS; then
	echo "Tests failed"
	FAILED=1    
fi

if [[ -n "$KEEPREPOS" || -n "$FAILED" ]]; then
	echo "Workspaces are in $REPOS"
	echo "Backups are in $BACKUPDIR"
else
	    rm -rf $REPOS $BACKUPDIR
fi

if [[ -n "$FAILED" ]]; then
	exit 1
fi
exit 0
