#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License version 2
#  as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#
# Copyright (c) 2008, 2021, Oracle and/or its affiliates.
# Copyright 2008, 2010 Richard Lowe
#

"""OpenSolaris extensions to Mercurial

    This extension contains a number of commands to help you work with
the OpenSolaris consolidations.  It provides commands to check your
changes against the various style rules used for OpenSolaris, to
backup and restore your changes, to generate code reviews, and to
prepare your changes for integration.


The Parent

    To provide a uniform notion of parent workspace regardless of
filesystem-based access, Cadmium uses the highest numbered changeset
on the current branch that is also in the parent workspace to
represent the parent workspace.


The Active List

    Many Cadmium commands operate on the active list, the set of
files ("active files") you have changed in this workspace in relation
to its parent workspace, and the metadata (commentary, primarily)
associated with those changes.


NOT Files

    Many of Cadmium's commands to check that your work obeys the
various stylistic rules of the OpenSolaris consolidations (such as
those run by "hg nits") allow files to be excluded from this checking
by means of NOT files kept in the .hg/cdm/ directory of the Mercurial
repository for one-time exceptions, and in the exception_lists
directory at the repository root for permanent exceptions.  (For ON,
these would mean one in $WS and one in $WS/usr/closed).

    These files are in the same format as the Mercurial hgignore
file, a description of which is available in the hgignore(5) manual
page.


Common Tasks

  - Show diffs relative to parent workspace               - pdiffs
  - Check source style rules                              - nits
  - Run pre-integration checks                            - pbchk
  - Collapse all your changes into a single changeset     - recommit
"""

import atexit
import logging
import os
import re
import sys
import stat
import termios
import textwrap

from mercurial import util, commands, node, ui, error
from mercurial import localrepo, scmutil
import mercurial.lock as lockmod

from mercurial import cmdutil, node, patch
try:
    from mercurial.pathutil import canonpath
except ImportError:
    try:
        from mercurial.scmutil import canonpath
    except ImportError:
        canonpath = util.canonpath

# command moved to registrar from cmdutil in 4.2
try:
    from mercurial import registrar
    command = registrar.command
except (ImportError, AttributeError):
    command = cmdutil.command

from mercurial import match

#
# Adjust the load path based on the location of cdm.py and the version
# of python into which it is being loaded.  This assumes the normal
# onbld directory structure, where cdm.py is in
# lib/python(version)?/onbld/hgext/.  If that changes so too must
# this.
#
# This and the case below are not equivalent.  In this case we may be
# loading a cdm.py in python2.X/ via the lib/python/ symlink but need
# python2.Y in sys.path.
#
sys.path.insert(
    1, os.path.join(
        os.path.dirname(__file__), "..", "..", "..", "python%d.%d" %
        sys.version_info[:2]))

#
# Add the relative path from cdm.py to usr/src/tools to the load path,
# such that a cdm.py loaded from the source tree uses the modules also
# within the source tree.
#
sys.path.insert(2, os.path.join(os.path.dirname(__file__), "..", ".."))

from onbld.Scm import Version
from onbld.Scm.WorkSpace import ActiveEntry, WorkSpace
from onbld.Scm.Backup import CdmBackup
from onbld.Checks import Cddl, Comments, Copyright, CStyle, HdrChk, Akfork
from onbld.Checks import JStyle, JSStyle, Keywords, Mapfile, Notify, SpelChk
from onbld.Checks import CheckMakefile, ParfaitRemind, Rti, Tags, onOWAN
from onbld.Checks import FortifyRemind, UTF8, Pkgfmtchk
from onbld.Checks.DbLookups import BugDBEmailMissing

testedwith = " ".join(str(v) for v in Version.GOOD_VERSIONS)
# Folded URL to BugDB 10006/GATE-STUFF/OS-NET-TOOLS
buglink = "https://furl.oraclecorp.com/kppR"

try:
    Version.check_version()
except Version.VersionMismatch as badversion:
    raise error.Abort("Version Mismatch:\n %s\n" % badversion)


cmdtable = {}
command = command(cmdtable)


def yes_no(ui, msg, default):
    if default:
        prompt = " [Y/n]:"
        defanswer = "y"
    else:
        prompt = " [y/N]:"
        defanswer = "n"

    index = ui.promptchoice(msg + prompt + "$$ &Yes $$ &No",
                            default=["y", "n"].index(defanswer))
    resp = ("y", "n")[index]

    return resp[0] in ("Y", "y")


def buildfilelist(ws, parent, files):
    """Build a list of files in which we're interested.

    If no files are specified take files from the active list relative
    to "parent".

    Return a list of 2-tuples the first element being a path relative
    to the current directory and the second an entry from the active
    list, or None if an explicit file list was given."""

    if files:
        return [(path, None) for path in sorted(files)]
    else:
        active = ws.active(parent=parent)
        return [(ws.filepath(e.name), e) for e in sorted(active)]


buildfilelist = util.cachefunc(buildfilelist)


def not_check(repo, cmd):
    """return a function which returns boolean indicating whether a file
    should be skipped for CMD."""

    #
    # The ignore routines need a canonical path to the file (relative to the
    # repo root), whereas the check commands get paths relative to the cwd.
    #
    # Wrap our argument such that the path is canonified before it is checked.
    #
    def canonified_check(ignfunc):
        def f(path):
            return ignfunc(canonpath(repo.root, repo.getcwd(), path))
        return f

    ignorefiles = []

    for f in [
            repo.vfs.join("cdm/%s.NOT" % cmd),
            repo.wjoin("exception_lists/%s" % cmd)]:
        if os.path.exists(f):
            ignorefiles.append(f)

    if ignorefiles:
        pats = ['include:%s' % f for f in ignorefiles]
        ign = match.match(repo.root, "", [], pats, warn=repo.ui.warn)
        return canonified_check(ign)
    else:
        return util.never


def abort_if_dirty(ws):
    """Abort if the workspace has uncommitted changes, merges,
    branches, or has Mq patches applied"""

    if ws.modified():
        raise error.Abort("workspace has uncommitted changes")
    if ws.merged():
        raise error.Abort("workspace contains uncommitted merge")
    if ws.branched():
        raise error.Abort("workspace contains uncommitted branch")
    if ws.mq_applied():
        raise error.Abort("workspace has Mq patches applied")


#
# Adding a reference to WorkSpace from a repo causes a circular reference
# repo <-> WorkSpace.
#
# This prevents repo, WorkSpace and members thereof from being garbage
# collected.  Since transactions are aborted when the transaction object
# is collected, and localrepo holds a reference to the most recently created
# transaction, this prevents transactions from cleanly aborting.
#
# Instead, we hold the repo->WorkSpace association in a dictionary, breaking
# that dependence.
#

class wsldict(dict):
    """The wsldict class pretends to map repo objects to workspace objects, but
    maps repo object URLs to workspace objects instead, so that versions of
    mercurial that use proxy objects instead of the original repo objects, we
    can still do what we need to.
    """

    def __contains__(self, key):
        return dict.__contains__(self, key.url())

    def __getitem__(self, key):
        return dict.__getitem__(self, key.url())

    def __setitem__(self, key, value):
        dict.__setitem__(self, key.url(), value)

    def __delitem__(self, key):
        return dict.__delitem__(self, key.url())


wslist = wsldict()


def reposetup(ui, repo):
    if repo.local() and repo not in wslist:
        wslist[repo] = WorkSpace(repo)

        if ui.interactive() and sys.stdin.isatty():
            ui.setconfig("hooks", "preoutgoing.cdm_pbconfirm",
                         "python:hgext_cdm.pbconfirm")

        # Set default labels for those using the color extension
        for key, value in [
                ("log.branch", "red bold"),
                ("log.tag", "blue"),
                ("changeset.public", "yellow"),
                ("changeset.draft", "green"),
                ("changeset.secret", "magenta bold"),
                ("desc.bugid", "cyan"),
                ("desc.arc", "cyan")]:
            if ui.config("color", key) is None:
                ui.setconfig("color", key, value)


def pbconfirm(ui, repo, hooktype, source):
    def wrapper(settings=None):
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, settings)

    if source == "push":
        if not yes_no(ui, "Are you sure you wish to push?", False):
            return 1
        else:
            settings = termios.tcgetattr(sys.stdin.fileno())
            orig = list(settings)
            atexit.register(wrapper, orig)
            settings[3] = settings[3] & (~termios.ISIG)  # c_lflag
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, settings)


def get_options(repo, config_file):
    """Read options file returning dictionary of configuration items found.
    Configuration options consist of key=value pairs.
    Comments indicated by a # are ignored.
    """
    comment_char = "#"
    option_char = "="

    options = {}
    filename = os.path.join(repo.root, config_file)

    try:
        fp = open(filename, "r")
    except IOError:
        return options

    for line in fp:
        # Remove comments:
        if comment_char in line:
            # split on comment char, keep only the part before
            line, __comment = line.split(comment_char, 1)

        # Find lines with an option=value, construct options dictionary
        if option_char in line:
            option, value = line.split(option_char, 1)
            option = option.strip()
            value = value.strip()
            options[option] = value

    fp.close()

    return options


@command("pdiffs",
         [("p", "parent", "", "parent workspace"),
          ("a", "text", None, "treat all files as text"),
          ("g", "git", None, "use extended git diff format"),
          ("w", "ignore-all-space", None,
              "ignore white space when comparing lines"),
          ("b", "ignore-space-change", None,
              "ignore changes in the amount of white space"),
          ("B", "ignore-blank-lines", None,
              "ignore changes whose lines are all blank"),
          ("U", "unified", 3, "number of lines of context to show"),
          ("I", "include", [], "include names matching the given patterns"),
          ("X", "exclude", [], "exclude names matching the given patterns")],
         "hg pdiffs [OPTION...] [-p PARENT] [FILE...]")
def cdm_pdiffs(ui, repo, *pats, **opts):
    """diff workspace against its parent

    Show differences between this workspace and its parent workspace
    in the same manner as "hg diff".

    For a description of the changeset used to represent the parent
    workspace, see The Parent in the extension documentation ("hg help
    cdm").
    """

    parent = opts["parent"]

    diffs = wslist[repo].pdiff(pats, opts, parent=parent)
    if diffs:
        for chunk, label in patch.difflabel(diffs.splitlines, True):
            ui.write(chunk, label=label)


@command("list|active",
         [("p", "parent", "", "parent workspace"),
          ("a", "added", None, "show added files"),
          ("m", "modified", None, "show modified files"),
          ("r", "removed", None, "show removed files")],
         "hg list [-amrRu] [-p PARENT]")
def cdm_list(ui, repo, **opts):
    """list active files (those changed in this workspace)

    Display a list of files changed in this workspace as compared to
    its parent workspace.

    File names are displayed one-per line, grouped by manner in which
    they changed (added, modified, removed).  Information about
    renames or copies is output in parentheses following the file
    name.

    For a description of the changeset used to represent the parent
    workspace, see The Parent in the extension documentation ("hg help
    cdm").

    Output can be filtered by change type with --added, --modified,
    and --removed.  By default, all files are shown.
    """

    wanted = []

    if opts["added"]:
        wanted.append(ActiveEntry.ADDED)
    if opts["modified"]:
        wanted.append(ActiveEntry.MODIFIED)
    if opts["removed"]:
        wanted.append(ActiveEntry.REMOVED)

    act = wslist[repo].active(opts["parent"])
    chngmap = {ActiveEntry.MODIFIED: "modified",
               ActiveEntry.ADDED: "added",
               ActiveEntry.REMOVED: "removed"}

    lst = {}
    for entry in act:
        if wanted and (entry.change not in wanted):
            continue

        chngstr = chngmap[entry.change]
        if chngstr not in lst:
            lst[chngstr] = []
        lst[chngstr].append(entry)

    for chng in sorted(lst.keys()):
        ui.write(chng + ":\n")
        for elt in sorted(lst[chng]):
            if elt.is_renamed():
                ui.write("\t%s (renamed from %s)\n" % (
                    elt.name, elt.parentname))
            elif elt.is_copied():
                ui.write("\t%s (copied from %s)\n" % (
                    elt.name, elt.parentname))
            else:
                ui.write("\t%s\n" % elt.name)


@command("arcs",
         [("p", "parent", "", "parent workspace")],
         "hg arcs [-p PARENT]")
def cdm_arcs(ui, repo, parent=None):
    """show all ARC cases referenced in changeset comments"""

    act = wslist[repo].active(parent)

    # We take a set of the appropriate comments to eliminate duplicates.
    for elt in set(filter(Comments.isARC, act.comments())):
        ui.write(elt + "\n")


@command("bugs",
         [("p", "parent", "", "parent workspace")],
         "hg bugs [-p PARENT]")
def cdm_bugs(ui, repo, parent=None):
    """show all bug IDs referenced in changeset comments"""

    act = wslist[repo].active(parent)

    for elt in set(filter(Comments.isBug, act.comments())):
        ui.write(elt + "\n")


@command("comments",
         [("p", "parent", "", "parent workspace")],
         "hg comments [-p PARENT]")
def cdm_comments(ui, repo, parent=None):
    """show changeset commentary for all active changesets"""
    act = wslist[repo].active(parent)

    for elt in act.comments():
        ui.write(elt + "\n")


@command("renamed",
         [("p", "parent", "", "parent workspace")],
         "hg renamed [-p PARENT]")
def cdm_renamed(ui, repo, parent=None):
    """show renamed active files

    Renamed files are shown in the format::

       new-name old-name

    One pair per-line.
    """

    act = wslist[repo].active(parent)

    for entry in sorted(filter(lambda x: x.is_renamed(), act)):
        ui.write("%s %s\n" % (entry.name, entry.parentname))


@command("comchk",
         [("p", "parent", "", "parent workspace"),
          ("N", "nocheck", None, "do not compare comments with databases")],
         "hg comchk [-p PARENT]")
def cdm_comchk(ui, repo, **opts):
    """check active changeset comment formatting

    Check that active changeset comments conform to O/N rules.

    Each comment line must contain either one bug or ARC case ID
    followed by its synopsis, or credit an external contributor.
    """

    active = wslist[repo].active(opts.get("parent"))
    parent = wslist[repo].parent(opts.get("parent"))

    try:
        branch = active.revs[0].branch()
    except IndexError:
        # No changes, committed or uncommitted
        branch = None

    ui.write("Comments check:\n")

    check_db = not opts.get("nocheck")
    bugdb_email = ui.config("cdm", "bugdb_email")
    if not bugdb_email:
        bugdb_email = ui.config("email", "from")
    try:
        return Comments.comchk(
            active.comments(), check_db=check_db, output=ui,
            bugdb_email=bugdb_email,
            gatePath=parent, gateCset0=repo[0].hex(), gateBranch=branch)
    except BugDBEmailMissing as e:
        ui.warn(e)
        return 1


@command("cddlchk",
         [("p", "parent", "", "parent workspace")],
         "hg cddlchk [-p PARENT]")
def cdm_cddlchk(ui, repo, *args, **opts):
    """check for a valid CDDL header comment in all active files.

    Check active files to ensure that a Common Development and Distribution
    License (CDDL) block comment is *not* present.

    The logic of this check can be reversed by using the cddlchk.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "cddlchk")
    ret = 0

    ui.write("CDDL check:\n")

    for f, e in filelist:
        require_cddl = False
        if e and e.is_removed():
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            require_cddl = True

        fh = open(f, "r")
        ret |= Cddl.cddlchk(fh, output=ui, require_cddl=require_cddl)
        fh.close()
    return ret


@command("mapfilechk",
         [("p", "parent", "", "parent workspace")],
         "hg mapfilechk [-p PARENT]")
def cdm_mapfilechk(ui, repo, *args, **opts):
    """check for a valid mapfile header block in active files

    Check that all link-editor mapfiles contain the standard mapfile
    header comment directing the reader to the document containing
    Solaris object versioning rules (README.mapfile).

    Files can be excluded from this check using the mapfilechk.NOT
    file.  See NOT Files in the extension documentation ("hg help
    cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "mapfilechk")
    ret = 0

    # We are interested in examining any file that has the following
    # in its final path segment:
    #    - Contains the word "mapfile"
    #    - Begins with "map."
    #    - Ends with ".map"
    # We don"t want to match unless these things occur in final path segment
    # because directory names with these strings don"t indicate a mapfile.
    # We also ignore files with suffixes that tell us that the files
    # are not mapfiles.
    MapfileRE = re.compile(
        r".*((mapfile[^/]*)|(/map\.+[^/]*)|(\.map))$", re.IGNORECASE)
    NotMapSuffixRE = re.compile(r".*\.[ch]$", re.IGNORECASE)

    ui.write("Mapfile comment check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (not MapfileRE.match(f)) or NotMapSuffixRE.match(f):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= Mapfile.mapfilechk(fh, output=ui)
        fh.close()
    return ret


@command("makefilechk",
         [("p", "parent", "", "parent workspace")],
         "hg makefilechk [-p PARENT]")
def cdm_makefilechk(ui, repo, *args, **opts):
    """Perform basic sanity checks on Makefiles included in the
    outgoing changeset(s).
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, 'makefilechk')
    ret = 0

    MakefileRE = re.compile(r".*Makefile.*", re.IGNORECASE)
    erroff_re = re.compile(r"\\+.*-erroff", re.MULTILINE)

    # We're interested in any file which includes the word "Makefile"
    # Determine the list of Makefile* elements in the repo
    RepoMakefiles = [q for q in scmutil.revsingle(repo, None)
                     if "Makefile" in q]
    if "usr/closed" in repo.root or "usr/man" in repo.root:
        # We need to append usr/src Makefiles, too
        if Version.at_least("4.8"):
            openrepo = localrepo.instance(
                ui, os.path.join(repo.root, "../.."), False)
        else:
            openrepo = localrepo.localrepository(
                ui, os.path.join(repo.root, "../.."))
        RepoMakefiles = RepoMakefiles + \
            [q for q in scmutil.revsingle(openrepo, None) if "Makefile" in q]
    else:
        # We may need to append usr/closed Makefiles too. Some elements of
        # usr/src point to usr/closed if various environment variables
        # are set.
        closed_path = os.path.join(repo.root, "usr/closed")
        # but only if usr/closed is a separate repository
        if os.path.exists(os.path.join(closed_path, ".hg")):
            if Version.at_least("4.8"):
                closedrepo = localrepo.instance(ui, closed_path, False)
            else:
                closedrepo = localrepo.localrepository(ui, closed_path)
            RepoMakefiles = RepoMakefiles + [
                q for q in scmutil.revsingle(closedrepo, None)
                if "Makefile" in q]

        # We may need to append usr/man Makefiles too.
        man_path = os.path.join(repo.root, "usr/man")
        if os.path.exists(os.path.join(man_path, ".hg")):
            if Version.at_least("4.8"):
                manrepo = localrepo.instance(ui, man_path, False)
            else:
                manrepo = localrepo.localrepository(ui, man_path)
            RepoMakefiles = RepoMakefiles + [
                q for q in scmutil.revsingle(manrepo, None)
                if "Makefile" in q]

    RepoMakefiles = RepoMakefiles + ["usr/src/Targetdirs"]

    ui.write("Basic Makefile checks:\n")
    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (e or opts.get('honour_nots')) and exclude(f):
            ui.status('Skipping %s...\n' % f)
            continue
        elif MakefileRE.match(f):
            fh = open(f, "r")
            # get the changes for this file to determine whether to enable
            # the lint -erroff warning in CheckMakefile
            patch = wslist[repo].pdiff([f], opts, parent=opts.get("parent"))
            if patch:
                adds_erroff = erroff_re.search(patch)
            else:
                adds_erroff = None
            ret |= CheckMakefile.makefilechk(fh, os.path.abspath(f),
                                             repo.root, False,
                                             RepoMakefiles, output=ui,
                                             erroff_warnings=adds_erroff)
            fh.close()
    return ret


@command("notifychk",
         [("p", "parent", "", "parent workspace")],
         "hg notifychk[-p PARENT]")
def cdm_notifychk(ui, repo, *args, **opts):
    """check active files for associated unit-tests or developer notifications
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ui.write("Developer notifications:\n")
    Notify.notify(filelist, repo.root, output=ui)
    return 0


@command("copyright",
         [("p", "parent", "", "parent workspace")],
         "hg copyright [-p PARENT]")
def cdm_copyright(ui, repo, *args, **opts):
    """check each active file for a current and correct copyright notice

    Check that all active files have a correctly formed copyright
    notice containing the current year.

    See the Non-Formatting Considerations section of the OpenSolaris
    Developer's Reference for more info on the correct form of
    copyright notice.
    (http://hub.opensolaris.org/bin/view/Community+Group+on/devref_7#H723NonFormattingConsiderations)

    Files can be excluded from this check using the copyright.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "copyright")
    ret = 0

    ui.write("Copyright check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= Copyright.copyright(fh, output=ui)
        fh.close()
    return ret


def cdm_utf8chk(ui, repo, *args, **opts):
    """check license files used by pkg info --license are in UTF-8

    License files processed by pkg(1) must be in ASCII or UTF-8.

    Files can be excluded from this check using the utf8chk.NOT
    file.  See NOT Files in the extension documentation ("hg help
    cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "utf8chk")
    ret = 0

    # We are interested in examining any file that has the following
    # in its final path segment:
    #    - Starts with "THIRDPARTYLICENSE"
    #    - Ends with ".license" (typical for Userland)
    # We don"t want to match unless these things occur in final path segment.
    # We also ignore files with suffixes that tell us that the files
    # are not licenses.
    LicenseFileRE = re.compile(
        r".*((/THIRDPARTYLICENSE.*)|(\.license))$", re.IGNORECASE)

    ui.write("License encoding check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (not LicenseFileRE.match(f)):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= UTF8.utf8chk(fh, output=ui)
        fh.close()
    return ret


@command("spell",
         [("d", "diff", None, "generate diff output")],
         "hg spell [-d]")
def cdm_spelchk(ui, repo, *args, **opts):
    """check all files for common spelling errors"""

    diff = opts.get("diff")

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)

    # We're using the same exclude list as the "copyright" check for now, but
    # may need to switch to a separate list if we've got files with intentional
    # spelling mistakes which should still have a valid copyright
    exclude = not_check(repo, "copyright")
    ret = 0

    if not diff:
        ui.write("Spelling check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        with open(f, "r") as fh:
            ret |= SpelChk.spel_chekc(fh, output=ui, patch=diff)

    return ret


@command("hdrchk",
         [("p", "parent", "", "parent workspace")],
         "hg hdrchk [-p PARENT]")
def cdm_hdrchk(ui, repo, *args, **opts):
    """check active C header files conform to the O/N header rules

    Check that any added or modified C header files conform to the O/N
    header rules.

    See the section "HEADER STANDARDS" in the hdrchk(1) manual page
    for more information on the rules for O/N header file formatting.

    Files can be excluded from this check using the hdrchk.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "hdrchk")
    ret = 0

    ui.write("Header format check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif not f.endswith(".h"):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= HdrChk.hdrchk(fh, lenient=True, output=ui)
        fh.close()
    return ret


@command("cstyle",
         [("p", "parent", "", "parent workspace")],
         "hg cstyle [-p PARENT]")
def cdm_cstyle(ui, repo, *args, **opts):
    """check active C source files conform to the C Style Guide

    Check that any added or modified C source file conform to the C
    Style Guide.

    See the C Style Guide for more information about correct C source
    formatting.
    (http://hub.opensolaris.org/bin/download/Community+Group+on/WebHome/cstyle.ms.pdf)

    Files can be excluded from this check using the cstyle.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "cstyle")
    ret = 0

    ui.write("C style check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif not (f.endswith(".c") or f.endswith(".h")):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= CStyle.cstyle(fh, output=ui,
                             picky=True, check_posix_types=True,
                             check_continuation=True, doxygen_comments=True)
        fh.close()
    return ret


@command("pkgfmtchk",
         [("d", "show_diff", None, "Display any changes required"),
          ("f", "fixup", None, "Fix manifests"),
          ("p", "parent", "", "parent workspace")],
         "hg pkgfmtchk [-p PARENT] [-f]")
def cdm_pkgfmtchk(ui, repo, *args, **opts):
    """check active p5m source files conform to the required pkgfmt layout.
    Note there is no exception list, since the build will fail if manifests
    are not in the correct format.
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ret = 0
    check_only = True
    show_diff = False

    if opts.get("fixup"):
        check_only = False
    elif opts.get("show_diff"):
        show_diff = True
        check_only = False

    ui.write("Package manifest format check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif not f.endswith(".p5m"):
            continue

        fh = open(f, "r")
        ret |= Pkgfmtchk.pkgfmtchk(fh, output=ui,
                                   check_only=check_only,
                                   show_diff=show_diff)
        fh.close()
    return ret


@command("jsstyle",
         [("p", "parent", "", "parent workspace")],
         "hg jsstyle [-p PARENT]")
def cdm_jsstyle(ui, repo, *args, **opts):
    """check active JavaScript source files for common stylistic errors

    Files can be excluded from this check using the jsstyle.NOT file.
    See NOT files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "jsstyle")
    ret = 0

    ui.write("JavaScript style check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif not f.endswith(".js"):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        options = get_options(repo, "config/jsstyle.conf")
        kwargs = {}
        kwargs["output"] = ui
        kwargs["doxygen_comments"] = True
        if "JSSTYLE_INDENT" in options and \
                options["JSSTYLE_INDENT"] in ["space", "tab"]:
            kwargs["indent_" + options["JSSTYLE_INDENT"]] = True
        else:
            kwargs["indent_space"] = True

        if "JSSTYLE_TABSTOP" in options and \
                options["JSSTYLE_TABSTOP"] in ["2", "4", "6", "8", "10"]:
            kwargs["tabstop_" + options["JSSTYLE_TABSTOP"]] = True
        else:
            kwargs["tabstop_4"] = True

        fh = open(f, "r")
        ret |= JSStyle.jsstyle(fh, **kwargs)
        fh.close()
    return ret


@command("jstyle",
         [("p", "parent", "", "parent workspace")],
         "hg jstyle [-p PARENT]")
def cdm_jstyle(ui, repo, *args, **opts):
    """check active Java source files for common stylistic errors

    Files can be excluded from this check using the jstyle.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "jstyle")
    ret = 0

    ui.write("Java style check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif not f.endswith(".java"):
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= JStyle.jstyle(fh, output=ui, picky=True)
        fh.close()
    return ret


@command("parfait_remind",
         [("p", "parent", "", "parent workspace")],
         "hg parfait_remind [-p PARENT]")
def cdm_parfait_remind(ui, repo, *args, **opts):
    """check for unresolved Parfait issues in all active files."""

    subrepo = ""

    if repo.root.endswith("usr/closed"):
        subrepo = "closed"
    elif repo.root.endswith("usr/fish"):
        subrepo = "fish"

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ret = 0

    branch = repo[None].branch()

    ui.write("Parfait reminder:\n")

    for f, e in filelist:
        if ret != 0:
            # if this happens, there has been a problem trying to access
            # the parfait reminder database, give up now.
            break

        if e and e.is_removed():
            continue

        ret = ParfaitRemind.parfait_remind(
            f, repo.root, branch, output=ui, subrepo=subrepo)

    # we always return success - this check is purely informative and does
    # not prevent putback.
    return 0


@command("fortify_remind",
         [("p", "parent", "", "parent workspace")],
         "hg fortify_remind [-p PARENT]")
def cdm_fortify_remind(ui, repo, *args, **opts):
    """check for unresolved Fortify issues in all active files."""

    subrepo = ""

    if repo.root.endswith("usr/closed"):
        subrepo = "closed"
    elif repo.root.endswith("usr/fish"):
        subrepo = "fish"

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ret = 0

    branch = repo[None].branch()

    ui.write("Fortify reminder:\n")

    for f, e in filelist:
        if ret != 0:
            # if this happens, there has been a problem trying to access
            # the Fortify reminder database, give up now.
            break

        if e and e.is_removed():
            continue

        ret = FortifyRemind.fortify_remind(
            f, repo.root, branch, output=ui, subrepo=subrepo)

    # we always return success - this check is purely informative and does
    # not prevent putback.
    return 0


@command("permchk",
         [("p", "parent", "", "parent workspace")],
         "hg permchk [-p PARENT]")
def cdm_permchk(ui, repo, *args, **opts):
    """check the permissions of each active file

    Check that the file permissions of each added or modified file do not
    contain the executable bit.

    Files can be excluded from this check using the permchk.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "permchk")
    exeFiles = []

    ui.write("File permission check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        mode = stat.S_IMODE(os.stat(f)[stat.ST_MODE])
        if mode & stat.S_IEXEC:
            exeFiles.append(f)

    if len(exeFiles) > 0:
        ui.write(
            "Warning: the following active file(s) have executable mode "
            "(+x) permission set,\nremove unless intentional:\n")
        for fname in exeFiles:
            ui.write("  %s\n" % fname)

    return len(exeFiles) > 0


@command("tagchk",
         [("p", "parent", "", "parent workspace")],
         "hg tagchk [-p PARENT]")
def cdm_tagchk(ui, repo, **opts):
    """check modification of workspace tags

    Check for any modification of the repository's .hgtags file.

    With the exception of the gatekeepers, nobody should introduce or
    modify a repository's tags.
    """

    active = wslist[repo].active(opts.get("parent"))

    ui.write("Checking for new tags:\n")

    if ".hgtags" in active:
        tfile = wslist[repo].filepath(".hgtags")
        ptip = active.parenttip.rev()

        ui.write(
            "Warning: Workspace contains new non-local tags.\n"
            "Only gatekeepers should add or modify such tags.\n"
            "Use the following commands to revert these changes:\n"
            "  hg revert -r%d %s\n"
            "  hg commit %s\n"
            "You should also recommit before integration\n" %
            (ptip, tfile, tfile))

        return 1

    return 0


@command("branchchk",
         [("p", "parent", "", "parent workspace")],
         "hg branchchk [-p PARENT]")
def cdm_branchchk(ui, repo, **opts):
    """check for changes in number or name of branches

    Check that the workspace is not getting new head or branch, or that a
    branch is not being closed. Also check that all outgoing changesets are on
    the same branch.
    """

    err = 0
    ui.write("Checking for multiple heads (or branches):\n")

    # Get list of "outgoing" changesets
    ws = wslist[repo]
    parent = ws.parent(opts["parent"])
    outgoing = repo.changelog.nodesbetween(ws.findoutgoing(parent))[0]
    outgoing = [repo[o] for o in outgoing]

    # Iterate over all branches
    for branch in repo.branchmap():

        # Find commit which created the branch
        repo.ui.pushbuffer()
        commands.log(
            repo.ui, repo, date=None, only_branch=[branch],
            user=None, rev=None, template="{node} ")
        first_in_branch = repo.ui.popbuffer()
        if not first_in_branch:
            continue
        first_in_branch = first_in_branch.split()[-1]
        first_in_branch = repo[first_in_branch]

        # Make sure it was not our outgoing commit which created the branch
        if first_in_branch in outgoing:
            ui.write(
                "new branch was created: %s\n" %
                first_in_branch.description().splitlines()[0])
            err = 1

        heads = [
            repo[x]
            for x in repo.branchheads(branch)
            if not repo[x].obsolete()
        ]

        if len(heads) <= 1:
            continue

        # Ok, check that we don't have uncommitted merge, which will be
        # reported by mergechk
        if repo.dirstate.branch() == branch:
            if [heads[0].node(), heads[1].node()] == repo.dirstate.parents():
                continue

        ui.write("branch '%s' contains multiple heads:\n" % branch)
        err = 1
        for node in heads:
            ui.write(" -> %s\n" % node.description().splitlines()[0])

    # Now make sure that our changeset does not close any branch
    branches = set([node.branch() for node in outgoing])
    if len(branches) > 1:
        ui.write("You have commits on multiple branches:\n")
        err = 1
        for b in branches:
            ui.write(" -> %s\n" % b)

    for node in outgoing:
        if node.extra().get("close"):
            ui.write("Branch %s is getting closed\n" % node.branch())
            ui.write(" -> %s\n" % node.description().splitlines()[0])
            err = 1

    return err


@command("rtichk",
         [("p", "parent", "", "parent workspace"),
          ("N", "nocheck", None, "skip RTI check")],
         "hg rtichk [-N] [-p PARENT]")
def cdm_rtichk(ui, repo, **opts):
    """check active bug/RFEs have approved RTIs

    Check that any CR IDs mentioned in active changeset comments have
    an associated, approved Request to Integrate (RTI).

    This check only works on systems with access to webrti.us.oracle.com.

    This check maybe disabled by creating an empty file rtichk.NOT.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    if opts.get("nocheck") or os.path.exists(repo.vfs.join("cdm/rtichk.NOT")):
        ui.status("Skipping RTI checks...\n")
        return 0

    if not onOWAN():
        ui.write("RTI checks only work on OWAN, skipping...\n")
        return 0

    parent = wslist[repo].parent(opts.get("parent"))
    active = wslist[repo].active(parent)

    ui.write("RTI check:\n")

    bugs = []

    for com in active.comments():
        match = Comments.isBug(com)
        if match and match.group(1) not in bugs:
            bugs.append(match.group(1))

    try:
        branch = active.revs[0].branch()
    except IndexError:
        # No changes, committed or uncommitted
        branch = None
    cset0 = repo[0].hex()

    # RTI normalizes the gate path for us
    return int(not Rti.rti(bugs, gatePath=parent, gateCset0=cset0,
                           gateBranch=branch, output=ui))


@command("keywords",
         [("p", "parent", "", "parent workspace")],
         "hg keywords [-p PARENT]")
def cdm_keywords(ui, repo, *args, **opts):
    """check active files for SCCS keywords

    Check that any added or modified files do not contain SCCS keywords
    (#ident lines, etc.).

    Files can be excluded from this check using the keywords.NOT file.
    See NOT Files in the extension documentation ("hg help cdm").
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    exclude = not_check(repo, "keywords")
    ret = 0

    ui.write("Keywords check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue
        elif (e or opts.get("honour_nots")) and exclude(f):
            ui.status("Skipping %s...\n" % f)
            continue

        fh = open(f, "r")
        ret |= Keywords.keywords(fh, output=ui)
        fh.close()
    return ret


@command("akfork",
         [("p", "parent", "", "parent workspace")],
         "hg akfork [-p PARENT]")
def cdm_akfork(ui, repo, *args, **opts):
    """check active files for AK-divergence keywords and warn the developer
    to test both flavours.
    """

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ret = 0

    ui.write("AKfork check:\n")

    for f, e in filelist:
        if e and e.is_removed():
            continue

        fh = open(f, "r")
        ret |= Akfork.akfork(fh, output=ui)
        fh.close()

    if ret:
        ui.write("Be sure to test both AK and server bits.\n")

    return ret


#
# NB:
#    There's no reason to hook this up as an invokable command, since
#    we have "hg status", but it must accept the same arguments.
#
def cdm_outchk(ui, repo, **opts):
    """Warn the user if they have uncommitted changes"""

    ui.write("Checking for uncommitted changes:\n")

    st = wslist[repo].modified()
    if st:
        ui.write("Warning: the following files have uncommitted changes:\n")
        for elt in st:
            ui.write("   %s\n" % elt)
        return 1
    return 0


def cdm_mergechk(ui, repo, **opts):
    """Warn the user if their workspace contains merges"""

    active = wslist[repo].active(opts.get("parent"))

    ui.write("Checking for merges:\n")

    merges = filter(
        lambda x: len(x.parents()) == 2 and x.parents()[1], active.revs)

    if merges:
        ui.write("Workspace contains the following merges:\n")
        for rev in merges:
            desc = rev.description().splitlines()
            ui.write(
                "  %s:%s\t%s\n" % (
                    rev.rev() or "working", str(rev),
                    desc and desc[0] or "*** uncommitted change ***"))
        return 1
    return 0


@command("histchk",
         [("p", "parent", "", "parent workspace")],
         "hg histchk [-p PARENT]")
def cdm_defunctchk(ui, repo, *args, **opts):

    prefix = "osnet"
    HistoryRE = re.compile(r"(.*/)*%s-(.*-)*history" % prefix)
    CurrentHistoryRE = re.compile(r"(.*/)*%s-(.*-)*history\.current" % prefix)

    parent = opts.get("parent")
    allfiles = buildfilelist(wslist[repo], parent, args)
    allhistoryfiles = [f for f, e in allfiles if HistoryRE.match(f)]
    currenthistoryfiles = [f for f in allhistoryfiles
                           if CurrentHistoryRE.match(f)]
    statichistoryfiles = [f for f in allhistoryfiles
                          if f not in currenthistoryfiles]
    ui.write("Package history checks:\n")

    returnstatus = 0

    if statichistoryfiles:
        ui.write("Warning: the following history files were changed:\n")
        for s in statichistoryfiles:
            ui.write("   %s\n" % s)
        returnstatus = 1

    for h in currenthistoryfiles:
        hdiffs = wslist[repo].pdiff([h], opts, parent=parent).split("\n")[4:]
        removed_lines = [d for d in hdiffs if d.startswith("-")]

        if removed_lines:
            ui.status("%s: %d lines removed or changed\n" %
                      (h, len(removed_lines)))
            returnstatus = 1

        for d in range(1, len(hdiffs) - 1):
            if hdiffs[d - 1].startswith("+") and not hdiffs[d].startswith("+"):
                ui.status("%s: history file changes should be append-only\n" %
                          h)
                returnstatus = 1
                break

    return returnstatus


@command("bugtagchk",
         [("p", "parent", "", "parent workspace")],
         "hg bugtagchk [-p PARENT]")
def cdm_bugtagchk(ui, repo, *args, **opts):
    """check active changeset bug tags.

    Check that active changeset bug tags conform to ON rules.

    Each comment line must contain either one bug or ARC case ID
    followed by its synopsis, or credit an external contributor.

    Bug tags are obtained from bug database and compared against the
    rules imposed by the gate we doing the pbchk against.
    """

    active = wslist[repo].active(opts.get("parent"))
    parent = wslist[repo].parent(opts.get("parent"))

    ui.write("Bug tags check:\n")

    check_db = not opts.get("nocheck")
    bugdb_email = ui.config("cdm", "bugdb_email")
    if not bugdb_email:
        bugdb_email = ui.config("email", "from")
    try:
        return Tags.tagchk(
            active.comments(), repo.root, check_db=check_db, output=ui,
            bugdb_email=bugdb_email, gatePath=parent)
    except BugDBEmailMissing as e:
        ui.warn(e)
        return 1


def run_checks(ws, cmds, *args, **opts):
    """Run CMDS (with OPTS) over active files in WS"""

    ret = 0

    for cmd in cmds:
        name = cmd.func_name.split("_")[1]
        if not ws.ui.configbool("cdm", name, True) \
                or skip_cmd(ws.repo.root, name):
            ws.ui.status("Skipping %s check...\n\n" % name)
        else:
            ws.ui.pushbuffer()
            result = cmd(ws.ui, ws.repo, honour_nots=True, *args, **opts)
            output = ws.ui.popbuffer()

            ret |= result

            if not ws.ui.quiet or result != 0:
                ws.ui.write(output, "\n")
    return ret


def skip_cmd(root, name):
    """With a given repository root path, and a command name, determine whether
    this command should be skipped by reading a flat file in
    <root>/exception_lists/cdm
    If the command name is listed on a line in this file, the checks for that
    command should be skipped.
    """
    exception_list_path = os.path.join(root, "exception_lists/cdm")
    if os.path.exists(exception_list_path):
        exceptions = []
        with open(exception_list_path, "r") as el:
            exceptions = [line.strip() for line in el.readlines()]
        if name in exceptions:
            return True
    return False


@command("nits",
         [("p", "parent", "", "parent workspace")],
         "hg nits [-p PARENT]")
def cdm_nits(ui, repo, *args, **opts):
    """check for stylistic nits in active files

    Check each active file for basic stylistic errors and other
    outstanding issues.

    The following checks are run over each active file (see "hg help
    <check>" for more information about each):

      - bugtagchk    (Bug tags)
      - cddlchk      (CDDL block comments)
      - copyright    (copyright statements)
      - cstyle       (C source style)
      - hdrchk       (C header style)
      - exceptionchk (validate removed files against exception lists)
      - jstyle       (java source style)
      - jsstyle      (JavaScript source style)
      - makefile     (Makefiles)
      - mapfilechk   (link-editor mapfiles)
      - pkgfmtchk    (p5m files)
      - notifychk    (report test availability and developer notifications)
      - permchk      (file permissions)
      - keywords     (SCCS keywords)
      - parfait_remind (Existing parfait issues reminder)
      - fortify_remind (Existing Fortify issues reminder)
      - akfork       (Check for AK fork divergence keywords)
      - utf8chk      (Check license files are in UTF-8)

    With the global -q/--quiet option, only provide output for those
    checks which fail.
    """

    cmds = [
        cdm_bugtagchk,
        cdm_cddlchk,
        cdm_copyright,
        cdm_cstyle,
        cdm_exceptionchk,
        cdm_hdrchk,
        cdm_spelchk,
        cdm_jstyle,
        cdm_jsstyle,
        cdm_makefilechk,
        cdm_mapfilechk,
        cdm_pkgfmtchk,
        cdm_notifychk,
        cdm_permchk,
        cdm_keywords,
        cdm_parfait_remind,
        cdm_fortify_remind,
        cdm_akfork,
        cdm_utf8chk]

    return run_checks(wslist[repo], cmds, *args, **opts)


def subrepo_list(ui):
    """Return the default list of ON subrepos.
    This includes the '.' entry as well.
    """

    default_subrepos = ["usr/man", "usr/closed", "usr/fish", "."]
    return ui.configlist("cdm", "onsub.subrepos", default_subrepos)


@command("onsub",
         [],
         "hg onsub command...")
def cdm_onsub(ui, repo, *args, **opts):
    """execute shell commands in each subrepository

    Subrepos here are in the ON sense, not actual Mercurial subrepos.
    With no arguments, runs the preamble command (see below).
    Commands with options should be quoted: hg onsub "ls -l".

    Environment variables exported to commands:

    HG_REPO
        Absolute path of the repo where onsub is executed.

    HG_SUBPATH
        Relative path of the current subrepo.
        Automatically appended to "hg onsub reparent /repo/path".

    Configuration entries that can be set in [cdm] section of ~/.hgrc:

    onsub.subrepos
        List of subrepos to search for.
        Default value: "usr/man","usr/closed","usr/fish",".".

    onsub.preamble
        Command to automatically prepend to the command specified
        as an argument. Default value: echo ${HG_SUBPATH};
        (note the trailing semicolon).

    onsub.use_hgcmds
        When set to True (default), hg commands can be used directly,
        e.g.:

            hg onsub tip

        is equivalent to:

            hg onsub hg tip
    """

    from mercurial import hg, commands
    try:
        from mercurial.utils import procutil
        hgexec = procutil.hgexecutable()
    except ImportError:
        hgexec = util.hgexecutable()

    # reparents get special treatment
    reparent = False

    if len(args) <= 0:
        cmdargs = []
    else:
        # split quoted commands
        cmdargs = " ".join(args).split(" ")
        reparent = (cmdargs[0] == "reparent")

        # use hg commands unless configured otherwise
        if ui.configbool("cdm", "onsub.use_hgcmds", True):
            try:
                c = cmdutil.findcmd(cmdargs[0], commands.table)
                cmdargs.insert(0, hgexec)
            except error.UnknownCommand:
                pass

    subrepos = subrepo_list(ui)

    default_preamble = "echo ${HG_SUBPATH};"
    preamble = ui.config("cdm", "onsub.preamble", default_preamble)

    for dir in subrepos:
        path = os.path.join(repo.root, dir)
        try:
            subrepo = hg.repository(ui, path=path)
        except Exception:
            subrepo = None
        if subrepo:
            env = {"HG_REPO": repo.root, "HG_SUBPATH": dir}
            # for reparent, append subpath to the last arg
            if reparent and dir != "." and len(cmdargs) > 1:
                cmd = preamble + " ".join(cmdargs[0:-1]) + \
                    " " + os.path.join(cmdargs[-1], dir)
            else:
                cmd = preamble + " ".join(cmdargs)
            ui.system(cmd, environ=env, cwd=path)


@command("pbchk",
         [("p", "parent", "", "parent workspace"),
          ("N", "nocheck", None, "skip RTI and bug/ARC database checks")],
         "hg pbchk [-N] [-p PARENT]")
def cdm_pbchk(ui, repo, **opts):
    """run pre-integration checks on this workspace

    Check this workspace for common errors prior to integration.

    The following checks are run over the active list (see "hg help
    <check>" for more information about each):

      - branchchk      (addition/modification of branches)
      - bugtagchk      (Bug tags)
      - cddlchk        (CDDL block comments)
      - comchk         (changeset descriptions)
      - copyright      (copyright statements)
      - cstyle         (C source style)
      - exceptionchk   (validate removed files against exception lists)
      - hdrchk         (C header style)
      - jstyle         (java source style)
      - jsstyle        (JavaScript source style)
      - keywords       (SCCS keywords)
      - mapfilechk     (link-editor mapfiles)
      - makefile       (Makefiles)
      - pkgfmtchk      (p5m files)
      - notifychk      (report test availability and developer notifications)
      - parfait_remind (print the number of open parfait issues)
      - fortify_remind (print the number of open Fortify issues)
      - utf8chk        (license encoding)
      - permchk        (file permissions)
      - rtichk         (requests to integrate)
      - spell          (check for commonly misspelled words)
      - tagchk         (addition/modification of tags)
      - akfork         (Check for AK fork divergence keywords)

    Additionally, the workspace is checked for outgoing merges (which
    should be removed with "hg recommit"), and uncommitted changes.

    With the global -q/--quiet option, only provide output for those
    checks which fail.
    """

    #
    # The current ordering of these is that the commands from cdm_nits
    # run first in the same order as they would in cdm_nits, then the
    # pbchk specifics are run.
    #
    cmds = [
        cdm_bugtagchk,
        cdm_cddlchk,
        cdm_copyright,
        cdm_cstyle,
        cdm_exceptionchk,
        cdm_hdrchk,
        cdm_spelchk,
        cdm_jstyle,
        cdm_jsstyle,
        cdm_makefilechk,
        cdm_mapfilechk,
        cdm_pkgfmtchk,
        cdm_notifychk,
        cdm_permchk,
        cdm_keywords,
        cdm_parfait_remind,
        cdm_fortify_remind,
        cdm_utf8chk,
        cdm_akfork,
        cdm_comchk,
        cdm_tagchk,
        cdm_branchchk,
        cdm_rtichk,
        cdm_outchk,
        cdm_mergechk,
        cdm_defunctchk
    ]

    return run_checks(wslist[repo], cmds, **opts)


@command("recommit|reci",
         [("p", "parent", "", "parent workspace"),
          ("m", "message", "", "use <text> as commit message"),
          ("l", "logfile", "", "read commit message from file"),
          ("u", "user", "", "record user as committer")],
         "hg recommit [-m TEXT] [-l FILE] [-u USER] [-p PARENT]")
def cdm_recommit(ui, repo, **opts):
    """replace outgoing changesets with a single equivalent changeset

    Replace all outgoing changesets with a single changeset containing
    equivalent changes.  This removes uninteresting changesets created
    during development that would only serve as noise in the gate.

    Any changed file that is now identical in content to that in the
    parent workspace (whether identical in history or otherwise) will
    not be included in the new changeset.  Any merges information will
    also be removed.

    If no files are changed in comparison to the parent workspace, the
    outgoing changesets will be removed, but no new changeset created.

    recommit will refuse to run if the workspace contains more than
    one outgoing head, even if those heads are on the same branch.  To
    recommit with only one branch containing outgoing changesets, your
    workspace must be on that branch and at that branch head.

    recommit will prompt you to take a backup if your workspace has
    been changed since the last backup was taken.  In almost all
    cases, you should allow it to take one (the default).

    recommit cannot be run if the workspace contains any uncommitted
    changes, applied Mq patches, or has multiple outgoing heads (or
    branches).
    """

    ws = wslist[repo]

    if not os.getcwd().startswith(repo.root):
        raise error.Abort("recommit is not safe to run with -R")

    abort_if_dirty(ws)

    wlock = repo.wlock()
    lock = repo.lock()

    try:
        parent = ws.parent(opts["parent"])
        between = repo.changelog.nodesbetween(ws.findoutgoing(parent))[2]
        heads = set(between) & set(repo.heads())

        if len(heads) > 1:
            ui.warn("Workspace has multiple outgoing heads (or branches):\n")
            for head in sorted(map(repo.changelog.rev, heads), reverse=True):
                ui.warn("\t%d\n" % head)
            raise error.Abort("you must merge before recommitting")

        active = ws.active(parent)

        if filter(lambda b: len(b.parents()) > 1, active.bases()):
            raise error.Abort("Cannot recommit a merge of two non-outgoing "
                              "changesets")

        if len(active.revs) <= 0:
            raise error.Abort("no changes to recommit")

        if len(active.files()) <= 0:
            ui.warn(
                "Recommitting %d active changesets, but no active files\n" %
                len(active.revs))

        #
        # During the course of a recommit, any file bearing a name
        # matching the source name of any renamed file will be
        # clobbered by the operation.
        #
        # As such, we ask the user before proceeding.
        #
        bogosity = [f.parentname for f in active if f.is_renamed() and
                    os.path.exists(repo.wjoin(f.parentname))]
        if bogosity:
            ui.warn("The following file names are the original name of a "
                    "rename and also present\n"
                    "in the working directory:\n")

            for fname in bogosity:
                ui.warn("  %s\n" % fname)

            if not yes_no(ui, "These files will be removed by recommit."
                          "  Continue?",
                          False):
                raise error.Abort("recommit would clobber files")

        user = opts["user"] or ui.username()
        comments = "\n".join(active.comments())

        message = cmdutil.logmessage(ui, opts) or ui.edit(comments, user)
        if not message:
            raise error.Abort("empty commit message")

        bk = CdmBackup(ui, ws, backup_name(ui, repo.root))
        if bk.need_backup():
            if yes_no(ui, "Do you want to backup files first?", True):
                bk.backup()

        oldtags = repo.tags()
        clearedtags = [
            (name, nd, repo.changelog.rev(nd), local)
            for name, nd, local in active.tags()]

        ws.squishdeltas(active, message, user=user)
    finally:
        lock.release()
        wlock.release()

    if clearedtags:
        ui.write("Removed tags:\n")
        for name, nd, rev, local in sorted(
                clearedtags, key=lambda x: x[0].lower()):
            ui.write(
                "  %5s:%s:\t%s%s\n" %
                (rev, node.short(nd), name, (local and " (local)" or "")))

        # We need to invalidate the tags cache because squishdeltas()
        # invalidated the caches of the localrepo that the WorkSpace object has
        # cached, but not the filtered repoview proxy we have here.  Ick.
        repo.invalidatecaches()
        for ntag, nnode in sorted(repo.tags().items(),
                                  key=lambda x: x[0].lower()):
            if ntag in oldtags and ntag != "tip":
                if oldtags[ntag] != nnode:
                    ui.write("tag '%s' now refers to revision %d:%s\n" %
                             (ntag, repo.changelog.rev(nnode),
                              node.short(nnode)))


def do_eval(cmd, files, root, changedir=True):
    if not changedir:
        os.chdir(root)

    for path in sorted(files):
        dirn, base = os.path.split(path)

        if changedir:
            os.chdir(os.path.join(root, dirn))

        os.putenv("workspace", root)
        os.putenv("filepath", path)
        os.putenv("dir", dirn)
        os.putenv("file", base)
        os.system(cmd)


@command("eval",
         [("p", "parent", "", "parent workspace"),
          ("r", "remain", None, "do not change directory")],
         "hg eval [-p PARENT] [-r] command...")
def cdm_eval(ui, repo, *command, **opts):
    """run specified command for each active file

    Run the command specified on the command line for each active
    file, with the following variables present in the environment:

      :$file:      -  active file basename.
      :$dir:       -  active file dirname.
      :$filepath:  -  path from workspace root to active file.
      :$workspace: -  full path to workspace root.

    For example:

      hg eval "echo $dir; hg log -l3 $file"

    will show the last the 3 log entries for each active file,
    preceded by its directory.
    """

    act = wslist[repo].active(opts["parent"])
    cmd = " ".join(command)
    files = [x.name for x in act if not x.is_removed()]

    do_eval(cmd, files, repo.root, not opts["remain"])


@command("apply",
         [("p", "parent", "", "parent workspace"),
          ("r", "remain", None, "do not change directory")],
         "hg apply [-p PARENT] [-r] command...")
def cdm_apply(ui, repo, *command, **opts):
    """apply specified command to all active files

    Run the command specified on the command line over each active
    file.

    For example "hg apply 'wc -l'" will output a count of the lines in
    each active file.
    """

    act = wslist[repo].active(opts["parent"])

    if opts["remain"]:
        appnd = " $filepath"
    else:
        appnd = " $file"

    cmd = " ".join(command) + appnd
    files = [x.name for x in act if not x.is_removed()]

    do_eval(cmd, files, repo.root, not opts["remain"])


@command("reparent",
         [],
         "hg reparent PARENT")
def cdm_reparent(ui, repo, parent):
    """reparent your workspace

    Update the "default" path alias that is used as the default source
    for "hg pull" and the default destination for "hg push" (unless
    there is a "default-push" alias).  This is also the path all
    Cadmium commands treat as your parent workspace.
    """

    def append_new_parent(parent):
        fp = repo.vfs("hgrc", "a", atomictemp=True)
        if fp._fp.tell() != 0:
            fp.write("\n")
        fp.write("[paths]\n"
                 "default = %s\n\n" % parent)
        fp.close()

    def update_parent(path, line, parent):
        line = line - 1  # The line number we"re passed will be 1-based
        fp = None

        try:
            fp = open(path)
            data = fp.readlines()
        finally:
            if fp and not fp.closed:
                fp.close()

        #
        # line will be the last line of any continued block, go back
        # to the first removing the continuation as we go.
        #
        while data[line][0].isspace():
            data.pop(line)
            line -= 1

        assert data[line].startswith("default")

        data[line] = "default = %s\n" % parent
        if data[-1] != "\n":
            data.append("\n")

        fp = util.atomictempfile(path, "w", 0o644)
        fp.write("".join(data))
        fp.close()

    from mercurial import config
    parent = ui.expandpath(parent)

    if not os.path.exists(repo.vfs.join("hgrc")):
        append_new_parent(parent)
        return

    cfg = config.config()
    cfg.read(repo.vfs.join("hgrc"))
    source = cfg.source("paths", "default")

    if not source:
        append_new_parent(parent)
        return
    else:
        path, target = source.rsplit(":", 1)

        if path != repo.vfs.join("hgrc"):
            raise error.Abort(
                "Cannot edit path specification not in repo hgrc\n"
                "default path is from: %s" % source)

        update_parent(path, int(target), parent)


def backup_name(ui, fullpath):
    """Create a backup directory name based on the specified path.

    In most cases this is the basename of the path specified, but
    certain cases are handled specially to create meaningful names"""

    subrepos = subrepo_list(ui)

    fullpath = fullpath.rstrip(os.path.sep).split(os.path.sep)

    #
    # If a path is "special", we append the basename of the path to
    # the path element preceding the constant, special, part.
    #
    # Such that for instance:
    #     /foo/bar/onnv-fixes/usr/closed
    #  has a backup name of:
    #     onnv-fixes-closed
    #
    for elt in subrepos:
        elt = elt.split(os.path.sep)
        pathpos = len(elt)

        if fullpath[-pathpos:] == elt:
            return "%s-%s" % (fullpath[-pathpos - 1], elt[-1])
    else:
        return fullpath[-1]


@command("backup|bu",
         [("t", "if-newer", None, "only backup if workspace files are newer")],
         "hg backup [-t]")
def cdm_backup(ui, repo, if_newer=False):
    """backup workspace changes and metadata

    Create a backup copy of changes made in this workspace as compared
    to its parent workspace, as well as important metadata of this
    workspace.

    NOTE: Only changes as compared to the parent workspace are backed
    up.  If you lose this workspace and its parent, you will not be
    able to restore a backup into a clone of the grandparent
    workspace.

    By default, backups are stored in the cdm.backup/ directory in
    your home directory.  This is configurable using the cdm.backupdir
    configuration variable, for example:

      hg backup --config cdm.backupdir=/net/foo/backups

    or place the following in an appropriate hgrc file::

      [cdm]
      backupdir = /net/foo/backups

    Backups have the same name as the workspace in which they were
    taken, with "-closed" appended in the case of O/N's usr/closed.
    """

    name = backup_name(ui, repo.root)
    bk = CdmBackup(ui, wslist[repo], name)

    wlock = repo.wlock()
    lock = repo.lock()

    try:
        if if_newer and not bk.need_backup():
            ui.status("backup is up-to-date\n")
        else:
            bk.backup()
    finally:
        lock.release()
        wlock.release()


@command("restore",
         [("g", "generation", "", "generation number")],
         "hg restore [-g GENERATION] BACKUP")
def cdm_restore(ui, repo, backup, **opts):
    """restore workspace from backup

    Restore this workspace from a backup (taken by "hg backup").

    If the specified backup directory does not exist, it is assumed to
    be relative to the cadmium backup directory (~/cdm.backup/ by
    default).

    For example::

      % hg restore on-rfe - Restore the latest backup of ~/cdm.backup/on-rfe
      % hg restore -g3 on-rfe - Restore the 3rd backup of ~/cdm.backup/on-rfe
      % hg restore /net/foo/backup/on-rfe - Restore from an explicit path
    """

    if not os.getcwd().startswith(repo.root):
        raise error.Abort("restore is not safe to run with -R")

    abort_if_dirty(wslist[repo])

    if opts["generation"]:
        gen = int(opts["generation"])
    else:
        gen = None

    if os.path.exists(backup):
        backup = os.path.abspath(backup)

    wlock = repo.wlock()
    lock = repo.lock()

    try:
        bk = CdmBackup(ui, wslist[repo], backup)
        bk.restore(gen)
    finally:
        lock.release()
        wlock.release()


@command("webrev",
         [("C", "C", "", "ITS priority file"),
          ("c", "c", "", "generate webrev for changeset"),
          ("D", "D", None, "delete remote webrev"),
          ("I", "I", "", "ITS configuration file"),
          ("i", "i", "", "include file"),
          ("N", "N", None, "suppress comments"),
          ("n", "n", None, "do not generate webrev"),
          ("O", "O", None, "OpenSolaris mode"),
          ("o", "o", "", "output directory"),
          ("p", "p", "", "use specified parent"),
          ("t", "t", "", "upload target"),
          ("U", "U", None, "upload the webrev"),
          ("w", "w", "", "use wx active file"),
          ("z", "z", None, "compress the webrev in ZIP format")],
         "hg webrev [WEBREV_OPTIONS]")
def cdm_webrev(ui, repo, **opts):
    """generate web-based code review and optionally upload it

    Generate a web-based code review using webrev(1) and optionally
    upload it.  All known arguments are passed through to webrev(1).
    """

    webrev_args = ""
    for key in list(opts.keys()):
        if opts[key]:
            if isinstance(opts[key], bool):
                webrev_args += "-" + key + " "
            else:
                webrev_args += "-" + key + " " + opts[key] + " "

    retval = os.system("webrev " + webrev_args)
    if retval != 0:
        return retval - 255

    return 0


@command("exceptionchk",
         [("p", "parent", "", "parent workspace")],
         "hg exceptionchk [-p PARENT]")
def cdm_exceptionchk(ui, repo, *args, **opts):
    """Validate any file removals against the exception list. If there is
    a direct match, flag it, as the check.exceptions task may well fail
    during the build.

    Wildcard entries in the exception list are problematic, and so are
    skipped in this check."""

    filelist = buildfilelist(wslist[repo], opts.get("parent"), args)
    ret = 0

    ui.write("Exception list check:\n")

    lists_to_check = [
        "cddlchk",
        "copyright",
        "cstyle",
        "hdrchk",
        "jsstyle",
        "keywords",
        "makefilechk",
        "mapfilechk",
        "permchk"
    ]

    for f, e in filelist:
        if e and e.is_removed():
            file = canonpath(repo.root, repo.getcwd(), f)
            for elist in lists_to_check:
                elist_fullpath = repo.wjoin("exception_lists/%s" % elist)
                if os.access(elist_fullpath, os.F_OK):
                    with open(elist_fullpath) as excpt:
                        if file in [line.strip()
                                    for line in excpt.readlines()]:
                            ui.write("%s: Entry in '%s' exception list should "
                                     "be removed\n" % (file, elist))

    return ret
