#!/usr/bin/python

#
# Copyright (c) 2013, 2021, Oracle and/or its affiliates.
#

import cgi
import collections
import copy
import datetime
import os
import re
import time
import traceback

from six.moves import cStringIO as StringIO

# Deal with the rename of _levelNames (2.x) to _levelToName (3.x)
import logging
try:
    logging._levelToName = logging._levelNames
except AttributeError:
    pass


"""The logging infrastructure used when building the OS/Net Consolidation.
"""

# Create a log levels for the sorts of output we'd like to categorize.
# We have exactly 10 levels here, defined in the gap between logging.INFO and
# logging.DEBUG.
STATE_LEVEL = logging.INFO - 1
# Summaries of completed tasks
SUMMARY_LEVEL = STATE_LEVEL - 1
# A special status, generally only used for "*** Error " messages from dmake
CMDERR_LEVEL = SUMMARY_LEVEL - 1
# Warning messages about tasks being run
TWARNING_LEVEL = CMDERR_LEVEL - 1
# Interesting information about tasks being run
TINFO_LEVEL = TWARNING_LEVEL - 1
# A level to log commands that are run
CMDLINE_LEVEL = TINFO_LEVEL - 1
# the following correspond to stderr and stdout from commands.
CMDOUT_LEVEL = CMDLINE_LEVEL - 1
CMD_LEVEL = CMDOUT_LEVEL - 1

# TASK_LEVEL is intentionally above DEBUG, but below INFO. Setting a log
# threshold to task level should get us all command line and stdout/stderr
# output from a given Task.
TASK_LEVEL = logging.DEBUG + 1
# A means to insert blank lines into log files. This is intentionally below
# DEBUG level, since it generally doesn't add any useful information.
SPACE_LEVEL = logging.DEBUG - 1

levels = (
    (STATE_LEVEL, "STATE"),
    (SUMMARY_LEVEL, "SUMMARY"),
    (TWARNING_LEVEL, "TWARN"),
    (TINFO_LEVEL, "TINFO"),
    (CMDLINE_LEVEL, "CMDLINE"),
    (CMDERR_LEVEL, "CMDERR"),
    (CMDOUT_LEVEL, "CMDOUT"),
    (CMD_LEVEL, "CMD"),
    (SPACE_LEVEL, "SPACE"),
    (TASK_LEVEL, "TASK"),
    # shorten the level name for critical so that it fits in 7 char-wide boxes
    (logging.CRITICAL, "CRIT"),
)

NORMAL_FORMAT = 0
NOTIMESTAMP_FORMAT = 1
log_format_names = {
    "normal": NORMAL_FORMAT,
    "notimestamp": NOTIMESTAMP_FORMAT
}

# Used by loggers when replacing 'asctime' format strings
datefmt = "%Y%m%dT%H%M%S"

# Map the format identifier to the log format string. produce_extract(..)
# is sensitive to these log formats and must be updated if new log formats are
# added, or these ones are changed.
log_formats = {
    NORMAL_FORMAT: "[%(asctime)s] %(levelname)-7s : %(message)s",
    NOTIMESTAMP_FORMAT: "%(levelname)-7s : %(message)s",
}

# Map the format identifier to the number of chars before the plaintext log
# message, used when we need to extract just the log message from the log file.
log_prefix_cols = {
    NORMAL_FORMAT: 29,
    NOTIMESTAMP_FORMAT: 10
}

for level, level_name in levels:
    if len(level_name) > 7:
        # As an aid to developers of this module, check for log level length
        raise ValueError(
            "Log levels should be no more than 7 characters, otherwise log "
            "files become harder to read.")
    logging.addLevelName(level, level_name)

logger = logging.getLogger("onbld")
# we want the parent logger to grab everything
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
# only info messages make it to the console by default
ch.setLevel(logging.INFO)

# A default formatter
formatter = logging.Formatter(log_formats[NORMAL_FORMAT], datefmt=datefmt)
nots_formatter = logging.Formatter(log_formats[NOTIMESTAMP_FORMAT])
ch.setFormatter(formatter)
logger.addHandler(ch)


def destroy_logfiles(logger):
    """Destroy any log files this logger may have written to the
    *only* the FileHandlers defined in this module.
    """
    for hdl in copy.copy(logger.handlers):
        hdl.close()
        if isinstance(hdl, RenameFileHandler):
            if os.path.exists(hdl.baseFilename):
                os.remove(hdl.baseFilename)
            # Remove the handler so any subsequent
            # log messages to this logger don't cause
            # the files to recreated.
            logger.removeHandler(hdl)


class StringIOHandler(logging.StreamHandler):
    """A logging.Handler that writes all log messages to a local StringIO
    object. Loggers using this handler Might Get Loud. (and use lots of memory,
    so only use it in cases where we don't expect large logs)"""
    def __init__(self):
        self.stringio = StringIO()
        super(StringIOHandler, self).__init__(stream=self.stringio)


class RenameFileHandler(logging.FileHandler):
    """A FileHandler that renames any existing logfiles before continuing."""

    def __init__(self, filename, mode="a", encoding=None, delay=0):
        """Create a FileHandler. Most of the work is done in the super
        class."""

        if os.path.exists(filename):
            # Rename the existing log to its creation time. Note that other
            # classes can implement their own log rotation (eg. the 'nightly'
            # subcommand for build(1) does this while also using this class)

            ctime = os.stat(filename).st_ctime
            ts = datetime.datetime.fromtimestamp(ctime).strftime(datefmt)
            os.rename(filename, "%s.%s" % (filename, ts))

        super(RenameFileHandler, self).__init__(
            filename, mode=mode, encoding=encoding, delay=delay)


class HtmlRenameFileHandler(RenameFileHandler):
    """A RenameFileHandler that writes html. LogRecord objects for this handler
    are expected to be formatted with HtmlFormatter, but we need to write some
    content to the html file before emitting log records, and it seems cleaner
    to do it here, rather than in the HtmlFormatter itself."""

    header = \
        """<html xmlns="http://www.w3.org/1999/xhtml">
      <head>
        <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
        <title>Build log</title>
        <style type="text/css">
    /* timestamps */
    a.ts {color:LightSlateGrey;}
    /* tasks */
    span.tk {color:MidnightBlue;}
    /* log level names */
    span.lv {color:green;}
    /* warnings */
    span.wg {color:red; white-space: pre-wrap;}
    /* commands */
    span.cd {color:grey; white-space: pre-wrap;}
    /* stderr from a command */
    /* cw(1) when printing the individual compiler output */
    span.cw {color:CadetBlue; white-space: pre-wrap;}
    /* command output */
    span.ou {color:Crimson;
        background-color:Beige; white-space: pre-wrap;
    }
    /* command errors */
    span.er {color:white;
        background-color:red; white-space: pre-wrap;}
    /* possible directory output */
    span.di {color:black;}
    /* higher level logging errors, WARNING, ERROR, CRITICAL */
    span.crit {color:red;
    font-weight:bold;
    background-color:Beige; white-space: pre-wrap;}
    pre.crit {color:red;
    background-color:Beige; white-space: pre-wrap;}
        </style>
        </head>
    <body>
    <p>
"""

    footer = """
    </p>
    </body>
    </html>
    """

    def __init__(self, filename, mode="a", encoding=None, delay=0):
        super(HtmlRenameFileHandler, self).__init__(
            filename, mode, encoding, delay)
        self.stream.write(self.header)

    def close(self):
        self.stream.write(self.footer)
        super(HtmlRenameFileHandler, self).close()


class HtmlFormatter(logging.Formatter):
    """Format log messages using pretty html syntax. Uses the CSS classes
    defined by HtmlFileHandler."""

    def __init__(self, fmt=None, datefmt=datefmt, force_date=False):
        """Create a HtmlFormatter - most of the work is done in the
        superclass."""

        super(HtmlFormatter, self).__init__(fmt=fmt, datefmt=datefmt)

        # a counter we increment in case several log messages arrive
        # with the same timestamp. Used, so that we can create unique
        # html anchors for different messages with the same timestamp.
        # (we want the anchor names to be human-friendly: seconds since
        # the epoch, with milisecond precision is not human-friendly, hence
        # this approach)
        self.ts_counter = 0
        # our current timestamp
        self.cur_ts = None

        # intentional quoting of '%' below for 'levelname' and 'message' since
        # we string-replace with the CSS class name we wish to use per log
        # format.
        if force_date:
            self.ts_anchor = \
                "[<a name=\"%%(anchor)s\" href=\"#%%(anchor)s\" " \
                "class=\"ts\">%%(force_date)s</a>] "
        else:
            self.ts_anchor = \
                "[<a name=\"%%(anchor)s\" href=\"#%%(anchor)s\" " \
                "class=\"ts\">%%(asctime)s</a>] "
        self.log_template = "<span class=\"lv\">%%(levelname)s:</span> " \
            "<span class=\"%s\">%%(message)s</span>" \
            "<br/>"
        # any time we're printing exception tracebacks, they get lots of
        # visibility
        self.exc_template = "\n<pre class=\"crit\">%%(exc_tb)s</pre>"

        self.cmderr_fmt = (self.ts_anchor + self.log_template) % "er"
        self.out_fmt = (self.ts_anchor + self.log_template) % "ou"
        self.cmd_fmt = (self.ts_anchor + self.log_template) % "cd"
        self.cw_fmt = (self.ts_anchor + self.log_template) % "cw"
        self.dir_fmt = (self.ts_anchor + self.log_template) % "di"
        self.task_fmt = (self.ts_anchor + self.log_template) % "tk"
        self.warn_fmt = (self.ts_anchor + self.log_template) % "wg"
        self.std_fmt = (self.ts_anchor + self.log_template) % "mg"
        self.crit_fmt = (self.ts_anchor + self.log_template) % "crit"
        self.crit_exc_fmt = (
            self.ts_anchor + self.log_template + self.exc_template) % "crit"

    def format(self, record):
        """Format 'record' - a logging.LogRecord object."""

        record.asctime = self.formatTime(record, self.datefmt)
        msg = cgi.escape(record.getMessage())
        record.message = msg

        if "force_date" in record.__dict__:
            timestamp = record.force_date
        else:
            timestamp = record.asctime

        if not self.cur_ts:
            anchor = timestamp
            self.cur_ts = timestamp
        elif self.cur_ts == timestamp:
            self.ts_counter += 1
            anchor = "%s_%s" % (timestamp, self.ts_counter)
        else:
            self.cur_ts = timestamp
            anchor = record.asctime
            self.ts_counter = 0

        record.__dict__["anchor"] = anchor

        def hyperlink_log(msg):
            """check to see if our message points to a new logfile, in which
            case we can make it a hyperlink"""
            if msg.endswith(".log.html") or msg.endswith(".log"):
                parts = msg.split(" ")
                filename = os.path.basename(parts[-1])
                if not filename.endswith(".html"):
                    filename = "%s.html" % filename
                parts[-1] = '<a href="%s">%s</a>' % (filename, parts[-1])
                return " ".join(parts)
            else:
                return msg

        # try to format what's likely to be cw(1) stderr
        if record.levelno == CMD_LEVEL and msg.startswith("+ "):
            return self.cw_fmt % record.__dict__
        # possibly just a directory name being emitted, but it's expensive
        # to see if the dir really exists, so just look for " ", which is good
        # enough.
        elif record.levelno == CMD_LEVEL and " " not in msg:
            return self.dir_fmt % record.__dict__

        elif record.levelno == CMD_LEVEL:
            return self.cmd_fmt % record.__dict__

        elif record.levelno == CMDOUT_LEVEL:
            return self.out_fmt % record.__dict__

        # more severe command errors (for dmake output, this is "*** Error")
        elif record.levelno == CMDERR_LEVEL:
            return self.cmderr_fmt % record.__dict__

        elif record.levelno == TASK_LEVEL:
            return self.task_fmt % record.__dict__

        elif record.levelno == TINFO_LEVEL:
            record.message = hyperlink_log(msg)
            return self.task_fmt % record.__dict__

        elif record.levelno == TWARNING_LEVEL:
            record.message = hyperlink_log(msg)
            return self.warn_fmt % record.__dict__

        # higher level Python logging module statuses are the most noticeable,
        # and we print full stack traces if we can.
        elif record.levelno in [
            logging.ERROR, logging.WARNING, logging.CRITICAL
        ]:
            if record.exc_info:
                record.exc_tb = "".join(
                    traceback.format_tb(record.exc_info[2]))
                return self.crit_exc_fmt % record.__dict__
            return self.crit_fmt % record.__dict__
        # we just want to write a separator in the log file
        elif record.levelno == SPACE_LEVEL:
            return "<hr/><br/>\n<br/>\n<br/>\n"
        # otherwise print less noticeable output
        else:
            return self.std_fmt % record.__dict__


class PkgHtmlFormatter(HtmlFormatter):
    """A Formatter to tailor the html log to deal with the output of the
    packages build."""

    def format(self, record):
        """Uses mostly the same code as HtmlFormatter, but deemphasises any
        lines of of the output starting with "info:"."""

        fmt = super(PkgHtmlFormatter, self).format(record)
        if record.message.startswith("info:"):
            return self.cmd_fmt % record.__dict__
        else:
            return fmt


class SpacesFormatter(logging.Formatter):
    """A simple formatter that allows us to insert blank separators when
    emitting certain types of log record."""

    def format(self, record):
        if record.levelno == SPACE_LEVEL:
            return "                ------------------------------------\n\n\n"
        else:
            return super(SpacesFormatter, self).format(record)


def switch_logs(logger, logsuffix="", old_logsuffix="", level=TASK_LEVEL):
    """A convenience function to have the given logging.Logger switch to a new
    file with a different suffix. 'logsuffix' is the new suffix to use, eg.
    '-nd'. 'old_logsuffix' is what the log currently uses. 'level' is the
    log level at which the new Handler should handle messages.
    """

    def new_logname(path, filesuffix, logsuffix, old_logsuffix):
        """
        'path'          full path to the log file
        'filesuffix'    file suffix for log files: '.log' or '.log.html'
        'old_logsuffix' current logname suffix: "" or "-nd"
        'logsuffix'     new logname suffix: "" or "-nd"
        """
        dirname = os.path.dirname(path)
        basename = os.path.basename(path)
        # remove the file suffix
        name = re.sub("%s$" % filesuffix, "", basename)
        # remove the old suffix
        if old_logsuffix:
            name = re.sub("%s$" % old_logsuffix, "", name)
        return os.path.join(
            dirname, "%s%s%s" % (name, logsuffix, filesuffix))

    # A list of tuples describing the changes we're making to the log handlers
    # in the form (<old handler>, old filename, <new handler>, new filename)
    # the filenames are just the basenames, rather than the full paths
    # (since they're referenced in the log files, and we rotate logs into their
    # own subdirs)
    handler_changes = []
    for handler in logger.handlers:
        formatter = handler.formatter

        if isinstance(handler, HtmlRenameFileHandler):
            old_name = handler.stream.name
            name = new_logname(old_name, ".log.html", logsuffix, old_logsuffix)
            new_handler = HtmlRenameFileHandler(name)
            new_handler.setFormatter(formatter)
            new_handler.setLevel(level)
            handler_changes.append(
                (handler, os.path.basename(old_name), new_handler,
                    os.path.basename(name)))

        elif isinstance(handler, RenameFileHandler):
            old_name = handler.stream.name
            name = new_logname(old_name, ".log", logsuffix, old_logsuffix)
            new_handler = RenameFileHandler(name)
            new_handler.setFormatter(formatter)
            new_handler.setLevel(level)
            handler_changes.append(
                (handler, os.path.basename(old_name),
                    new_handler, os.path.basename(name)))

    # Switch our handlers, logging messages before and after we do so.
    for old_handler, old_name, new_handler, new_name in handler_changes:
        logger.log(TINFO_LEVEL, "Switching log handler to %s" % new_name)
    for old_handler, old_name, new_handler, new_name in handler_changes:
        logger.removeHandler(old_handler)
        logger.addHandler(new_handler)
    for old_handler, old_name, new_handler, newname in handler_changes:
        logger.log(TINFO_LEVEL, "Continuing log handling from %s" % old_name)


def close_logs(logger):
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            logger.removeHandler(handler)


def produce_extract(
        logger, log_path, bufsize=20, log_format=NORMAL_FORMAT, elogger=None):
    """Tries to produce a useful extract by parsing a plaintext log file. The
    output is written to new log files, named <name>.extract.log and
    <name>.extract.log.html given a 'log_path' argument "<name>.log".

    'logger' is a logger where we record the name of the extract being
    produced, or None, if no log messages should be recorded.

    'bufsize' is the number of log entries we write before and after we hit a
    given "interesting" log message.

    'log_format' defines the format of the log messages we emit, either
    NORMAL_FORMAT or NOTIMESTAMP_FORMAT are accepted. If log formats change,
    this method will need to be updated.

    'elogger' is the logger where we write extracted log content to. If set to
    None, then we create new plaintext and html log files and write a message
    to the created elogger saying which log file we're extracting from.

    If any log records were emitted while constructing the extract, then we
    return the basename of the plaintext log file generated.
    """

    # This works by keeping a ringbuffer of log entries of size 'bufsize'.
    # As we walk through the log file, we add entries to that ringbuffer. As
    # soon as we encounter a report-worthy log entry (based on its log level),
    # we write all buffered log entries to our logger, then continue on to log
    # the subsequent 'bufsize' number of entries (regardless of their log
    # level). Between each cluster of log entries, we emit a log spacer to
    # separate-out incidents.
    #
    # This function depends on log files being written in the formats listed in
    # log_formats above.

    level_map = {}

    original_log_base = os.path.basename(log_path)
    original_log_dir = os.path.dirname(log_path)

    log_name = original_log_base.replace(".log", "")
    log_stem = "%s/%s" % (original_log_dir, log_name)

    created_elogger = False
    if not elogger:
        created_elogger = True
        elogger = logging.getLogger("%s.extract" % log_name)
        handler = RenameFileHandler("%s.extract.log" % log_stem)
        if log_format == NOTIMESTAMP_FORMAT:
            formatter = SpacesFormatter(log_formats[NOTIMESTAMP_FORMAT])
        else:
            formatter = SpacesFormatter(
                "[%(force_date)s] %(levelname)-7s : %(message)s")
            handler.setFormatter(formatter)
            elogger.addHandler(handler)

        # We always write timestamps for html logs
        handler = HtmlRenameFileHandler("%s.extract.log.html" % log_stem)
        if log_format == NOTIMESTAMP_FORMAT:
            formatter = HtmlFormatter()
        else:
            formatter = HtmlFormatter(force_date=True)
        handler.setFormatter(formatter)
        elogger.addHandler(handler)

    if logger:
        logger.log(
            TINFO_LEVEL, "Producing extract log at %s.extract.log" % log_name)

    # whether we've emitted any log messages at all
    recorded_incident = False

    # note that since SPACE_LEVEL is defined below logging.DEBUG, we should
    # not use Logger.debug() in this function.
    elogger.setLevel(SPACE_LEVEL)

    for level_no, level_label in levels:
        level_map[level_label] = level_no
    # Since we do lookups from level names to level IDs, logging.getLevelName()
    # isn't enough. Private access to internals of 'logging' is not pleasant,
    # but there's not much else we can do here if our log extract needs to
    # include levels defined in the 'logging' module itself.
    level_map.update(logging._levelToName)

    logbuffer = collections.deque(maxlen=bufsize)

    # Having dumped our buffer, we also want to write subsequent log entries.
    # We set this to 0 initially, then set to bufsize once we hit an error
    subsequent_entries = 0

    # Having dumped a window of the log, we want to insert a series of blank
    # lines into the log to separate-out error incidents
    written_space = True

    # Keep the last directory name seen, writing it at the start of each
    # incident. The name of the directory isn't always accurate, but we do
    # our best.
    dir = None

    # the log levels we consider interesting enough to cause us to start
    # logging
    report_levels = [CMDERR_LEVEL, TWARNING_LEVEL, logging.ERROR]

    with open(log_path, "r") as log_file:

        # record the level and timestamp, useful as globals in cases where we
        # encounter log messages with embedded '\n' characters.
        level_no = TINFO_LEVEL
        tstamp = ""
        extra = {}
        # the column of the log entry that contains the text of the message
        msg_col = log_prefix_cols[log_format] - 1

        for line in log_file:
            items = line.split()
            level = ""
            if log_format == NOTIMESTAMP_FORMAT:
                # the format of a log entry is:
                # [log level] [colon] [log message]
                if len(items) == 0 or items[0] not in level_map:
                    message = line.rstrip()
                elif len(items) == 3 and items[2].startswith("/"):
                    level = items[0]
                    dir = items[2]
                else:
                    level = items[0]
                    message = line[msg_col:].rstrip()

            else:
                # the format of a log entry is:
                # [timestamp]  [log level] [colon] [log message]
                if len(items) == 0 or not items[0].startswith("[2"):
                    message = line.rstrip()
                elif items[1] not in level_map:
                    message = line.rstrip()
                elif len(items) == 4 and items[3].startswith("/"):
                    level = items[1]
                    dir = items[3]
                    if items[0].startswith("[2"):
                        tstamp = items[0].replace("[", "").replace("]", "")
                else:
                    # This has a Year 2999 problem, but that's probably ok.
                    if items[0].startswith("[2"):
                        tstamp = items[0].replace("[", "").replace("]", "")

                    level = items[1]
                    message = line[msg_col:].rstrip()

                # when re-logging this message, we want to preserve the
                # original timestamp. Python logging doesn't allow us to
                # create LogRecords with specific timestamps, so we achieve
                # this with a custom LogRecord property, and expect our
                # Formatter objects to render it in place of the actual log
                # timestamp.
                extra = {"force_date": tstamp}

            level_no = level_map.get(level, TINFO_LEVEL)

            # We've hit a report-worthy log entry, and are now dumping the
            # subsequent log entries.
            if subsequent_entries > 0:
                # if we encounter another report-worthy entry while draining
                # subsequent log entries, then reset our counter
                if level_no in report_levels:
                    subsequent_entries = bufsize
                elogger.log(level_no, message, extra=extra)
                subsequent_entries = subsequent_entries - 1
                continue
            else:
                logbuffer.append((level_no, extra, message))
                if not written_space:
                    elogger.log(SPACE_LEVEL, "", extra=extra)
                    written_space = True

            if level_no in report_levels:
                # Write a message saying where these logs are coming from.
                # if we were passed an elogger, then we assume the caller has
                # taken care of this.
                if not recorded_incident and created_elogger:
                    elogger.log(
                        TINFO_LEVEL, "Log contents extracted from %s" %
                        original_log_base, extra={
                            "force_date": time.strftime(datefmt)})
                    elogger.log(SPACE_LEVEL, "", extra=extra)

                if dir:
                    elogger.log(CMD_LEVEL, dir, extra=extra)

                for level_no, extra, msg in logbuffer:
                    elogger.log(level_no, msg, extra=extra)
                logbuffer = collections.deque(maxlen=bufsize)
                subsequent_entries = bufsize
                written_space = False
                recorded_incident = True

        if created_elogger and recorded_incident:
            close_logs(elogger)
            return "%s.extract.log" % log_name
        elif created_elogger and not recorded_incident:
            try:
                logger.log(
                    TINFO_LEVEL, "No extract messages found. Removing %s" %
                    "%s.extract.log" % log_name)
                close_logs(elogger)
                os.remove("%s.extract.log" % log_stem)
                os.remove("%s.extract.log.html" % log_stem)
            except OSError:
                pass
