#!/usr/bin/python3.7 -B

#
# Copyright (c) 2014, 2021, Oracle and/or its affiliates.
#

import atexit
import copy
import datetime
import logging
import os
import re
import shutil
import sys
import tarfile
import tempfile
import time
import unittest
import fnmatch

# Ensure we get the onbld package and can run build.py directly from the
# workspace.
if not (os.path.exists("../../scripts/build.py") or os.path.exists(
        os.path.join(os.environ.get("SRC"), "tools/scripts/build.py"))):
    raise RuntimeError(
        "This test can only run if $SRC is set or "
        "./../scripts/build.py exists")
if os.environ.get("SRC"):
    sys.path.append(os.path.join(os.environ["SRC"], "tools"))
    SCRIPTS_PATH = os.path.join(os.environ["SRC"], "tools/scripts")
else:
    sys.path.append("../..")
    SCRIPTS_PATH = os.path.abspath("../../scripts")
    os.environ["SRC"] = os.path.abspath(os.path.join(SCRIPTS_PATH, "../.."))

os.environ["PATH"] = "%s:%s" % (os.environ["PATH"], SCRIPTS_PATH)

import onbld.build
import onbld.environment
import onbld.log
import onbld.util

BUILD_CHECKS = onbld.build.get_known_checks()
# a series of 'build generate' arguments to flag all build checks as
# non-fatal, useful when doing nightly simulations where we don't
# actually care about failing checks.
IGNORE_ALL_CHECKS = " ".join(
    ["-D %s.must_pass=false" % check for check in BUILD_CHECKS])


class BuildTest(unittest.TestCase):

    logger = logging.getLogger("onbld")
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.handlers[0].formatter = logging.Formatter(
        "%(levelname)-7s : %(message)s")
    testdir = tempfile.mkdtemp("", "buildunittest.")

    def setUp(self):

        if "ONBLD_DEBUG" in os.environ:
            self.debug = True
            BuildTest.logger.handlers[0].setLevel(logging.DEBUG)
            # Add a CR so that when running with -v, we get test descriptions
            # and the first line of debug output separated properly.
            sys.stderr.write("\n%s:\n" % self._testMethodName)
        else:
            self.debug = False

        self.workspace = "%s/workspace" % self.testdir
        self.mach = onbld.environment.MACH
        if self.mach == "i386":
            self.altmach = "sparc"
        else:
            self.altmach = "i386"

        self.create_workspace(self.workspace)
        atexit.register(shutil.rmtree, self.testdir, ignore_errors=True)

    def tearDown(self):
        # note that we only destroy the workspace between test runs
        if os.path.exists(self.workspace):
            shutil.rmtree(self.workspace)

    def wait(self, message):
        self.logger.error(self.workspace)
        self.logger.error("SLEEPING: %s" % message)
        time.sleep(10000000)

    def create_workspace(self, dir, fail=False, noisy=True, fish=True):
        """Creates a fake workspace in dir, enough to fool build(1) into
        accepting it as valid, though not all build tasks are guaranteed to
        function in this workspace.
        If fail is true, then we build the workspace by ensuring that any
        builds done with this workspace will report a fail. If fail is False,
        we attempt to reuse a previously cached version of the non-failing
        workspace.
        If noisy is true, we emit '*** Error' messages, as part of the make
        targets, to fool anything in build(1) that may mistake those for
        failing commands - only the command exit codes are important, not
        the command output."""

        shutil.rmtree(dir, ignore_errors=True)
        ws_tar = "%s/ws.tar" % self.testdir
        if os.path.exists(ws_tar) and not fail and noisy:
            # extract the workspace from our tarball
            try:
                tar = tarfile.open(ws_tar, "r:gz")
                tar.extractall(dir)
            finally:
                if tar:
                    tar.close()
            return

        os.makedirs("%s/usr/src" % dir)
        os.makedirs("%s/usr/closed" % dir)
        os.makedirs("%s/usr/man" % dir)
        os.makedirs("%s/usr/src/pkg" % dir)
        os.makedirs("%s/usr/src/tools/test.d" % dir)

        os.symlink("../Makefile", "%s/usr/src/pkg/Makefile" % dir)
        open("%s/usr/closed/.empty" % dir, "w").close()
        open("%s/usr/man/.empty" % dir, "w").close()

        if fish:
            fish_dir = os.path.join(dir, "usr/fish/pkgdefs")
            os.makedirs(fish_dir)
            os.symlink(
                "../../src/Makefile", "%s/usr/fish/pkgdefs/Makefile" % dir)
            open(
                "%s/%s" % (
                    fish_dir,
                    onbld.environment.FISH_STANDALONE_CANARY_PATH), "w"
            ).close()

        with open("%s/usr/src/Makefile.master" % dir, "w") as make_master:
            make_master.write(
                "RELEASE_MAJOR= 5\n"
                "RELEASE_MINOR= 12\n"
                "RELEASE=$(RELEASE_MAJOR).$(RELEASE_MINOR)\n")

        # We include the fail == "silent" to make sure nothing in build(1)
        # depends on seeing dmake error codes. A failing command ought to
        # be enough because build(1) calls things other than just dmake, and
        # we need to be able to detect those failing too.
        if fail == "silent":
            command = "\texit 1\n"
        elif fail:
            command = "@echo \\*\\*\\* Error code 1\n\texit 1\n"
        elif noisy:
            # the 'Error code 1' here is another check to ensure we don't
            # care about that string appearing anywhere. The exit code is
            # what matters.
            command = "@echo \\*\\*\\* Error code 1\n\texit 0\n"
        else:
            command = "exit 0\n"

        makefile_text = (
            "\n"
            "install all bldtools setup check ctf-check lint pmodes:\n"
            "\t%s\n" % command)
        with open("%s/usr/src/Makefile" % dir, "w") as makefile:
            makefile.write(makefile_text)

        real_src = os.environ.get("SRC")
        curdir = os.getcwd()
        if not real_src:
            while curdir != "/" and not real_src:
                makefile_path = os.path.join(curdir, "usr/src/Makefile.master")
                if os.path.exists(makefile_path):
                    real_src = curdir
                else:
                    curdir = os.path.dirname(curdir)
        if curdir == "/":
            raise RuntimeError("Unable to find workspace $SRC!")

        # Pull copies of the Makefiles that build(1) expects to be able to
        # access in order to perform workspace version checks.
        for makefile in [
                "Makefile.buildnum", "Makefile.gk", "Makefile.nightlyid"]:
            makefile_path = os.path.join(real_src, makefile)
            shutil.copyfile(
                makefile_path, "{0}/usr/src/{1}".format(dir, makefile))

        # Pull copies of some test.cfg files needed later
        for cfg in ["snoop.cfg", "crypto-fips.cfg"]:
            cfg_path = os.path.join(real_src, "tools/test.d", cfg)
            shutil.copyfile(
                cfg_path, "{0}/usr/src/tools/test.d/{1}".format(
                    dir, cfg))

        # hg add the files in the repos, ending with "." so it doesn't pickup
        # files in the nested repositories.
        for repo in ["usr/man", "usr/closed", "."]:
            self.cmdrun("hg init %s/%s" % (dir, repo))
            self.cmdrun("hg -R %s/%s add" % (dir, repo))
            self.cmdrun(
                "hg -R %s/%s commit -m 'buildtest workspace %s'" %
                (dir, repo, repo))
            self.cmdrun(
                "hg -R %s/%s tag -m 'buildtest tag %s' on12_01" %
                (dir, repo, repo))

        if not fail and fish:
            try:
                tar = tarfile.open(ws_tar, "w:gz")
                tar.add(dir, ".")
            finally:
                if tar:
                    tar.close()

    def cmdrun(
            self, cmd, exit=0, logger=False, env_arg=None, shell=True,
            cwd=None):
        """Run a command using onbld.util.cmdline_run."""

        if self.debug:
            self.logger.debug("Running: %s" % cmd)

        if not cwd:
            cwd = self.workspace

        if logger:
            return onbld.util.cmdline_run(
                cmd, logger=self.logger, shell=shell, raise_error=True,
                exit=exit, env_arg=env_arg, cwd=cwd)
        else:
            return onbld.util.cmdline_run(
                cmd, shell=shell, raise_error=True, exit=exit, env_arg=env_arg,
                cwd=cwd)

    def build(
            self, cmd_args, exit=0, env_arg=None, shell=True, cwd=None,
            allow_reexec=False):
        """A wrapper to run build(1). Note that by default, we run build(1)
        with the test workspace as the current working directory so that we
        automatically pickup the bldrc file at the root of the workspace.

        We attempt to use the workspace copy of build.py first, falling back to
        the copy of build(1) on the test system otherwise.
        """
        if cwd is None:
            cwd = self.workspace

        if allow_reexec:
            no_reexec = ""
        else:
            no_reexec = "-W"

        buildpy_path = os.path.join(SCRIPTS_PATH, "build.py")
        if os.path.exists(buildpy_path):
            cmd_name = buildpy_path
        else:
            raise RuntimeError("Unable to find build.py")
        ret, out, err = self.cmdrun(
            "%s %s %s" % (cmd_name, no_reexec, cmd_args),
            exit=exit, logger=False, env_arg=env_arg, shell=shell, cwd=cwd)

        # insert a delay if we're running tstamp to make findunref testing
        # more reliable.
        if cmd_args == "tstamp":
            time.sleep(1)

        self.out = out
        self.err = err
        self.ret = ret

    def test_generate(self):
        """Basic tests for the generate subcommand."""

        self.build("generate")
        self.assertTrue("build_tools = /opt" in self.out)
        self.assertTrue("gkoverrides" not in self.out)
        self.assertTrue("dev_cm = " not in self.out)
        self.assertTrue("check.elf = " not in self.out)
        self.assertTrue("generate_cmd::" in self.out)

        self.build("generate -k")
        self.assertTrue("gkoverrides" in self.out)
        self.assertTrue("dev_cm = " in self.out)

        self.build("generate -O")
        self.assertTrue("checkoverrides" in self.out)
        self.assertTrue("gkoverrides" not in self.out)
        self.assertTrue("check.elf = true" in self.out)
        # verify our generate_cmd was saved
        self.assertTrue("generate -O" in self.out)

        # we should add the checkoverrides section automatically
        self.build("generate -D check.elf=false")
        self.assertTrue("checkoverrides" in self.out)
        self.assertTrue("check.elf = false" in self.out)
        self.assertTrue("check.install-noise = true" in self.out)
        # this defaults to false, so generate should pick up on that
        self.assertTrue("check.elfsigncmp = false" in self.out)
        # verify our generate_cmd was saved
        self.assertTrue("generate -D check.elf=false" in self.out)

        # We should add the gkoverrides section automatically
        self.build("generate -D version=foo")
        self.assertTrue("gkoverrides" in self.out)
        self.assertTrue("dev_cm = " in self.out)

        # Make sure we can use '=' chars in values
        self.build("generate -D 'envcppflags2=-Ifoo -D ZVAL=noodles'")
        self.assertTrue("envcppflags2 = -Ifoo -D ZVAL=noodles" in self.out)

        self.build(
            "generate -D mailto=tim.s.foster@oracle.com -D bip=true")
        self.assertTrue("mailto = tim.s.foster@oracle.com" in self.out)
        self.assertTrue("bip = true" in self.out)
        # verify our generate_cmd was saved
        self.assertTrue(
            "generate -D mailto=tim.s.foster@oracle.com -D bip=true"
            in self.out)

        usage_str = "build generate [-kO] [-D <key>=<val>]"
        usage_cmds = [
            ("generate --help", usage_str),
            ("generate --h", usage_str),
        ]
        for (cmd, expected_err) in usage_cmds:
            self.build(cmd, exit=2)
            self.assertTrue(expected_err in self.err)

        error_cmds = [
            ("generate -skjdfhsdf", "option -s not recognized"),
            ("generate -D foo", "-D arguments take the form "),
            ("generate -D check.parfait=noodles",
                "values should be either 'true' or 'false'"),
        ]
        for (cmd, expected_err) in error_cmds:
            self.build(cmd, exit=1, )
            self.assertTrue(expected_err in self.err)

        self.build("generate -D noodles=bar", exit=1)
        self.assertTrue("unknown bldrc key" in self.err)

        self.build("generate -D dmake_output_mode=txt", exit=1)
        self.assertTrue("cannot be changed in the bldrc file" in self.err)

    def test_regenerate(self):
        self.build("regenerate", exit=1)
        self.build("regenerate --noodles", exit=1)
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf=false -D check=false > %s/bldrc" % self.workspace)
        self.build("regenerate %s/bldrc" % self.workspace)
        self.assertTrue(
            "generate -D build_tools=/opt -D check.elf=false -D check=false"
            in self.out)
        self.assertTrue("build_tools = /opt" in self.out)
        self.assertTrue("check.elf = false" in self.out)

        # try an old bldrc file with no generate_cmd:: comment string
        with open("%s/bldrc" % self.workspace, "w") as old_bldrc:
            for line in self.out.split("\n"):
                if "generate_cmd::" in line:
                    continue
                old_bldrc.write("%s\n" % line)
        self.build("regenerate %s/bldrc" % self.workspace, exit=1)

        # Make sure we replace the path to build(1) when regenerating.
        # First write a bldrc file with a weird path, then verify that
        # path is not present in the regenerated bldrc.
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf=false -D check=false")
        with open("%s/weird_bldrc" % self.workspace, "w") as weird_bldrc:
            for line in self.out.split("\n"):
                if "generate_cmd::" in line:
                    weird_bldrc.write(
                        "# generate_cmd:: /opt/noodles/build generate "
                        "-D build_tools=/opt -D check.elf=false "
                        "-D check=false\n")
                else:
                    weird_bldrc.write("%s\n" % line)
        self.build("regenerate %s/weird_bldrc" % self.workspace)
        self.assertTrue("/opt/noodles/build" not in self.out)

        # try a malicious bldrc file with an evil bldrc comment string
        with open("%s/bldrc" % self.workspace, "w") as bad_bldrc:
            bad_bldrc.write("# generate_cmd:: evil command!\n")
        self.build("regenerate %s/bldrc" % self.workspace, exit=1)

    def test_basic_environment(self):
        """Basic tests to ensure build(1) is setting an appropriate
        environment and inheriting environment values properly"""

        env = copy.deepcopy(os.environ)
        env["NOODLES"] = "yum"
        env["SPAGHETTI"] = "nice"
        if "PASTA" in env:
            del env["PASTA"]

        self.build(
            "generate -D user=joeb -D mailto=joe.bloggs@oracle.com "
            "-D build_tools=/opt -D userenv=SPAGHETTI "
            "-D check.rti_ready=false "
            "-D userunsetenv=UT_NO_USAGE_TRACKING > %s/bldrc" %
            self.workspace)

        for cmd in [
                "buildtest.environment 2>&1",
                "-qi env"]:
            self.build(cmd, env_arg=env)
            self.assertTrue("USER=joeb" in self.out)
            self.assertTrue("MAILTO=joe.bloggs@oracle.com" in self.out)
            self.assertTrue("SPAGHETTI=nice" in self.out)
            self.assertTrue("NOODLES=yum" not in self.out)
            self.assertTrue("UT_NO_USAGE_TRACKING=1" not in self.out)
            self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
            self.assertTrue("IS_FISH_ENV=true" not in self.out)

        # Verify that we fail when env variable is not proper section.
        self.build("generate > %s/bldrc" % self.workspace)
        with open("%s/bldrc" % self.workspace, "a") as bldrc:
            bldrc.write("\n[bogus_code_checks]\nparfait_server_password = xxx")

        self.build("buildtest.environment", env_arg=env, exit=1)
        self.assertTrue(
            "bldrc file error: key/value 'parfait_server_password=xxx' can"
            " only appear in section gkoverrides, not bogus_code_checks"
            in self.err)

        # Make sure we fail-fast when a userenv value is not set in the
        # environment.
        self.build(
            "generate -D user=joeb -D mailto=joe.bloggs@oracle.com "
            "-D build_tools=/opt -D 'userenv=SPAGHETTI;PASTA' "
            "-D check.rti_ready=false "
            "-D userunsetenv=UT_NO_USAGE_TRACKING > %s/bldrc" %
            self.workspace)
        self.build("buildtest.environment", env_arg=env, exit=1)
        self.assertTrue(
            "User override of $PASTA required, but no such value is set"
            in self.err)
        self.assertTrue(
            "User override of $SPAGHETTI required but no such value is set"
            not in self.err)

        # When using build -i, we also enforce this userenv check, but
        # allow -e to ignore it.
        self.build("-qi /usr/bin/true", env_arg=env, exit=1)
        self.assertTrue(
            "User override of $PASTA required, but no such value is set"
            in self.err)
        self.build("-qie /usr/bin/true", env_arg=env, exit=0)
        self.assertTrue(
            "User override of $PASTA required, but no such value is set"
            not in self.err)

        self.build(
            "generate -D build_tools=/opt "
            "-D userenv=MAILTO > %s/bldrc" % self.workspace)
        self.build("buildtest.environment", exit=1)
        self.assertTrue(
            "Error: The value of the following keys must be set "
            "in individual bldrc\nproperties, not 'userenv': MAILTO"
            in self.err)

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        # test we can get a fish environment when we ask for it
        self.build("-iA env | grep FISH")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("IS_FISH_ENV=true" in self.out)

        # make sure we can run a command when we're not actually in a
        # workspace, but pointing to a bldrc which has a workspace
        # set instead
        self.build(
            "generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build(
            "-qi -f %s/bldrc /usr/bin/true" %
            self.workspace, cwd="/tmp")
        self.build("buildtest.pass %s/bldrc" % self.workspace, cwd="/tmp")

    def test_fish_env(self):
        """Test that fish_is_present=yes|no|<None> works as expected"""
        # create a workspace and specify we want a server build,
        # make sure that's functional, make sure we're not choosing a
        # FISH_BUILD environment
        self.create_workspace(self.workspace, noisy=False, fish=False)
        # verify auto-detection is working
        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build(
            "-qi -f %s/bldrc /usr/bin/true" %
            self.workspace, cwd="/tmp")
        self.build(
            "buildtest.environment %s/bldrc 2>&1" % self.workspace, cwd="/tmp")
        self.assertTrue("FISH_IS_PRESENT=no" in self.out)
        self.assertTrue("IS_FISH_ENV=true" not in self.out)
        self.assertTrue("FISH_STANDALONE=yes" not in self.out)
        self.build(
            "generate -D fish_is_present=no "
            "-D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build(
            "-qi -f %s/bldrc /usr/bin/true" %
            self.workspace, cwd="/tmp")
        self.build(
            "buildtest.environment %s/bldrc 2>&1" % self.workspace, cwd="/tmp")
        self.assertTrue("FISH_IS_PRESENT=no" in self.out)
        self.assertTrue("IS_FISH_ENV=true" not in self.out)
        self.assertTrue("FISH_STANDALONE=yes" not in self.out)

        # we shouldn't be able to get a fish env
        self.build("-iA /usr/bin/true", exit=1)

        # verify we can't produce a fish build when usr/fish is missing
        self.build(
            "generate -D fish_is_present=yes "
            "-D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.pass", exit=1)

        # create a fish env and sanity check that it really is one
        self.create_workspace(self.workspace, fish=True)
        self.build(
            "generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" not in self.out)

        # now explicitly ask for a fish environment when usr/fish is present
        self.build(
            "generate -D fish_is_present=yes -D build_tools=/opt "
            "> %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" not in self.out)

        # ask for a non-fish environment even though usr/fish is present
        self.build(
            "generate -D fish_is_present=no -D build_tools=/opt "
            "> %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("FISH_IS_PRESENT=no" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" not in self.out)

    def test_standalone_fish_env(self):
        """Test that we can auto-detect a standalone fish workspace."""
        shutil.rmtree(self.workspace)
        os.mkdir(self.workspace)
        os.mkdir("%s/.hg" % self.workspace)
        open(
            "%s/%s" %
            (self.workspace,
                onbld.environment.FISH_STANDALONE_CANARY_PATH), "w").close()
        self.build("generate -D check.parfait=false > bldrc")
        self.build("buildtest.environment 2>&1")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" in self.out)

        self.build("-i env | grep FISH")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" in self.out)

        # asking for a specific environment also works
        self.build("-iA env | grep FISH")
        self.assertTrue("FISH_IS_PRESENT=yes" in self.out)
        self.assertTrue("FISH_STANDALONE=yes" in self.out)

    def test_flavor(self):
        """Tests that we change build variables depending on the type of build
           requested."""

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))
        self.build("buildtest.environment")
        self.assertTrue(
            "ROOT=%s/proto/root-%s\n" % (self.workspace, self.mach))
        self.assertTrue(
            "PACKAGES=%s/packages/%s/nightly\n" % (self.workspace, self.mach))
        self.assertTrue(
            "STUBROOT=%s/proto/root_%s_stub\n" % (self.workspace, self.mach))
        # this should be fixed for both debug and nondebug
        self.assertTrue("LOGDIR=%s/log.%s" % (self.workspace, self.mach))
        self.assertTrue(
            "TOOLS_PROTO=%s/usr/src/tools/proto/root_%s-nd" %
            (self.workspace, self.mach))

        self.build("-i true")
        self.assertTrue("Build type     is DEBUG" in self.err)
        self.build("-id true")
        self.assertTrue("Build type     is DEBUG" in self.err)
        self.build("-iD true")
        self.assertTrue("Build type     is non-DEBUG" in self.err)

        self.build("buildtest.environment -d 2>&1")
        self.assertTrue(
            "ROOT=%s/proto/root-%s-nd" % (self.workspace, self.mach))
        self.assertTrue(
            "PACKAGES=%s/packages/%s-nd/nightly" % (self.workspace, self.mach))
        self.assertTrue(
            "STUBROOT=%s/proto/root_%s-nd_stub" % (self.workspace, self.mach))
        # this should be fixed for both debug and nondebug
        self.assertTrue("LOGDIR=%s/log.%s" % (self.workspace, self.mach))
        self.assertTrue(
            "TOOLS_PROTO=%s/usr/src/tools/proto/root_%s-nd" %
            (self.workspace, self.mach))

        # check the generate -D options are respected
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s -D debug=true -D nondebug=false "
            "> %s/bldrc" % (self.workspace, self.workspace))
        self.build("-i true")
        self.assertTrue("Build type     is DEBUG" in self.err)
        self.assertTrue("Build type     is non-DEBUG" not in self.err)
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s -D debug=false -D nondebug=true "
            "> %s/bldrc" % (self.workspace, self.workspace))
        self.build("-i true")
        self.assertTrue("Build type     is non-DEBUG" in self.err)
        self.assertTrue("Build type     is DEBUG" not in self.err)

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s -D root=%s/foo > %s/bldrc" %
            (self.workspace, self.testdir, self.workspace))
        self.build("buildtest.environment -d 2>&1")
        self.assertTrue("ROOT=%s/foo\n" % self.testdir in self.out)
        self.build("buildtest.environment -D 2>&1")
        self.assertTrue("ROOT=%s/foo-nd" % self.testdir in self.out)

    def test_overrides(self):
        """Tests to see that overrides are processed correctly."""

        self.build(
            "generate -D check.elf=false -D build_tools=/opt "
            "> %s/bldrc" % self.workspace)

        self.build("buildtest.environment 2>&1")
        self.assertTrue("check.elf:false" in self.out)
        self.assertTrue("check.elf:true" not in self.out)
        self.assertTrue("check.parfait:true" in self.out)
        self.assertTrue("check.protocmp:true" in self.out)
        self.assertTrue("check.lint:true" in self.out)

        # make sure lint is now disabled
        self.build("buildtest.environment -O check.lint=false 2>&1")
        self.assertTrue("check.elf:false" in self.out)
        self.assertTrue("check.elf:true" not in self.out)
        self.assertTrue("check.parfait:true" in self.out)
        self.assertTrue("check.protocmp:true" in self.out)
        self.assertTrue("check.lint:false" in self.out)

        # disable all checks with -C
        self.build("buildtest.environment -C")
        self.assertTrue("check.parfait" not in self.out)
        self.assertTrue("check.lint" not in self.out)
        self.assertTrue("PARFAIT_BUILD=" not in self.out)

        # Ensure we can't process conflicting bldrc files with check=false
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf=false -D check=false > %s/bldrc" % self.workspace)
        self.build(
            "buildtest.environment -O check.lint=false 2>&1", exit=1)
        self.build("buildtest.environment -C -O check.lint=true", exit=1)

        # Ensure .must_pass also works
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf=false > %s/bldrc" % self.workspace)
        self.build("buildtest.environment -c 2>&1")
        self.assertTrue("check.elf.must_pass:true" in self.out)
        self.build("buildtest.environment -O check.elf.must_pass=true 2>&1")
        self.assertTrue("check.elf.must_pass:true" in self.out)
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf.must_pass=false > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("check.elf.must_pass:false" in self.out)
        self.build("buildtest.environment -O check.elf.must_pass=true 2>&1")
        self.assertTrue("check.elf.must_pass:true" in self.out)

        # Ensure we deal with disabled checks
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false > %s/bldrc" % self.workspace)
        self.build("buildtest.environment -O check.lint=false 2>&1", exit=1)
        self.assertTrue("check.parfait" not in self.out)
        self.assertTrue("PARFAIT_BUILD=" not in self.out)

        self.build("buildtest.environment -O check.lint=true 2>&1", exit=1)

    def test_parfait(self):
        """Tests to see that we enable/disable parfait correctly."""

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("PARFAIT_BUILD=" in self.out)

        self.build("buildtest.environment -P 2>&1")
        self.assertTrue("check.parfait:false" in self.out)
        self.assertTrue("PARFAIT_BUILD=" not in self.out)

        self.build(
            "generate -D build_tools=/opt "
            "-D check.parfait=false > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("check.parfait:false" in self.out)
        self.assertTrue("PARFAIT_BUILD=" not in self.out)

        self.build(
            "buildtest.environment -p baseline %s/bldrc 2>&1" % self.workspace)
        self.assertTrue("check.parfait:true" in self.out)
        self.assertTrue("PARFAIT_BUILD=" in self.out)

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("-i env")
        self.assertTrue("PARFAIT_BUILD=" in self.out)
        self.assertTrue("Parfait build  is enabled" in self.err)
        self.build("-iP env")
        self.assertTrue("PARFAIT_BUILD=" not in self.out)
        self.assertTrue("Parfait build  is disabled" in self.err)

        self.build(
            "generate -D build_tools=/opt "
            "-D check.parfait=false > %s/bldrc" % self.workspace)
        self.build("-ip env")
        self.assertTrue("PARFAIT_BUILD=" in self.out)
        self.assertTrue("Parfait build  is enabled" in self.err)

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false > %s/bldrc" % self.workspace)
        self.build("-i env")
        self.assertTrue("PARFAIT_BUILD=" not in self.out)
        self.assertTrue("Parfait build  is disabled" in self.err)

        # -C is mutually exclusive with -P
        self.build("buildtest.environment -CP", exit=1)

    def test_bip(self):
        """Checks that we set BIP correctly."""

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")
        self.assertTrue("BIP=" not in self.out)

        self.build("buildtest.environment -b 2>&1")
        self.assertTrue("BIP=true" in self.out)

        self.build(
            "generate -D build_tools=/opt "
            "-D bip=yes > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1")

        self.build("buildtest.environment -b 2>&1")
        self.assertTrue("BIP=true" in self.out)

        self.build("buildtest.environment -B %s/bldrc 2>&1" % self.workspace)
        self.assertTrue("BIP=true" not in self.out)

        self.build("-iBf %s/bldrc env" % self.workspace)
        self.assertTrue("BIP=true" not in self.out)
        self.assertTrue("Build in place is disabled" in self.err)

        self.build("-ibf %s/bldrc env" % self.workspace)
        self.assertTrue("BIP=true" in self.out)
        self.assertTrue("Build in place is enabled" in self.err)

    def test_exit_codes(self):
        """Exit codes should reflect task status."""

        self.build("buildtest.pass")
        self.build("buildtest.fail", exit=1)
        self.build("check.buildtest.fail1", exit=1)
        self.build("check.buildtest.fail2", exit=1)
        self.build("buildtest.stacktrace", exit=1)
        self.assertTrue("Not the comfy chair!" in self.err)
        self.assertTrue(
            "Please report this to your friendly gatekeepers" in self.err)

    def test_ie(self):
        """Test that build -ie adds $PATH to the end of the build path
        and that we unset $PYTHONNOUSERSITE."""

        def get_env(var, output):
            for line in output.split("\n"):
                if line.startswith("%s=" % var):
                    return line

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("-iq env")
        ipath = get_env("PATH", self.out)
        nousersite = get_env("PYTHONNOUSERSITE", self.out)
        self.assertTrue(nousersite)

        self.build("-iqe env")
        iepath = get_env("PATH", self.out)
        nousersite = get_env("PYTHONNOUSERSITE", self.out)
        self.assertTrue(not nousersite)

        self.assertTrue(ipath != iepath, "%s\n\n!=\n\n%s" % (ipath, iepath))
        self.assertTrue(iepath.endswith(os.environ["PATH"]))
        self.assertTrue("CONTINUE AT YOUR OWN RISK" in self.err)

    def test_explain(self):
        """Test that the explain subcommand works."""

        # basic usage
        self.build("explain version")
        self.assertTrue("[gkoverrides] version:" in self.out)
        self.assertTrue("Version affects customer-visible" in self.out)
        self.assertTrue("[paths] build:" not in self.out)

        # help for all sections
        self.build("explain")
        self.assertTrue("[None] ctfmerge:" in self.out)
        self.assertTrue("[paths] build:" in self.out)
        for section in onbld.environment.SECTS:
            self.assertTrue("[%s]" % section in self.out)
        for section in ["None", "gkoverrides", "checkoverrides"]:
            self.assertTrue("[%s]" % section in self.out)

        # help for sections that aren't in the __environment dict
        self.assertTrue("[checkoverrides] check.protocmp:" in self.out)
        self.assertTrue(
            "This setting makes build(1) skip the 'check.protocmp' task when"
            in self.out)
        self.assertTrue(
            "[checkoverrides] check.protocmp.must_pass:" in self.out)
        self.assertTrue(
            "This setting determines whether a failing 'check.protocmp' task"
            in self.out)

        self.build("explain check.protocmp")
        self.assertTrue(
            "This setting makes build(1) skip the 'check.protocmp' task when"
            in self.out)
        self.build("explain check.protocmp.must_pass")
        self.assertTrue(
            "This setting determines whether a failing 'check.protocmp' task"
            in self.out)

        # unknown keys
        self.build("explain noodles", exit=1)

    def test_rti_ready(self):
        """Test that the rti_ready task catches errors."""

        self.create_workspace(self.workspace)

        # Make sure our default rti_readiness is clean
        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("check.rti_ready -v")
        shutil.copy(
            "%s/bldrc" % self.workspace, "%s/bldrc.clean" % self.workspace)

        # try disabling packages
        self.build(
            "generate -D build_tools=/opt -D pkgsurf=false "
            "-D pkgarchive='<None>' > %s/bldrc" % self.workspace)
        self.build("check.rti_ready -Pv", exit=1)
        self.assertTrue(
            "'pkgarchive' was set to '<None>' in the bldrc file" in self.err)
        self.assertTrue(
            "check.parfait was disabled" in self.err)

        # try pkgmerge/pkgsurf without packages
        self.build(
            "generate -D build_tools=/opt -D pkgmerge=true "
            "-D pkgarchive='<None>' > %s/bldrc" % self.workspace)
        self.build("check.rti_ready -Pv", exit=1)
        self.assertTrue(
            "Cannot set 'pkgmerge' or 'pkgsurf' to 'true' when 'pkgarchive' "
            "is" in self.err)

        # try disabling a build check
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf=false > %s/bldrc" % self.workspace)
        self.build("check.rti_ready -v", exit=1)
        self.assertTrue("check.elf was disabled" in self.err)

        # try changing a build must_pass setting
        self.build(
            "generate -D build_tools=/opt "
            "-D check.elf.must_pass=false > %s/bldrc" % self.workspace)
        self.build("check.rti_ready -v", exit=1)
        self.assertTrue(
            "check.elf.must_pass default was changed to false" in self.err)

        # try both
        self.build("check.rti_ready -P -v", exit=1)
        self.assertTrue(
            "check.parfait was disabled" in self.err)
        self.assertTrue(
            "check.elf.must_pass default was changed to false" in self.err)

        # check that we catch source trees with out-of-sync repositories
        shutil.copy(
            "%s/bldrc.clean" % self.workspace, "%s/bldrc" % self.workspace)
        self.cmdrun("hg -R %s tag on12_60" % self.workspace)
        self.cmdrun("hg -R %s/usr/man tag on12_60" % self.workspace)
        self.cmdrun("hg -R %s/usr/closed tag on12_59" % self.workspace)
        self.build("check.rti_ready -v", exit=1)
        self.assertTrue(
            "This workspace contains repositories that have been "
            "synced to different OS/Net builds" in self.err)
        self.cmdrun("hg -R %s/usr/closed tag on12_60" % self.workspace)
        self.build("check.rti_ready")

    def test_build_date(self):
        """Tests that we're setting build_date and iso8601_build_date."""

        self.build(
            "generate -D build_tools=/opt "
            "-D build_date='A long time ago' "
            "-D iso8601_build_date='Some other time' > %s/bldrc" %
            self.workspace)
        self.build(
            "buildtest.environment 2>&1| grep BUILD_DATE")
        self.assertTrue("BUILD_DATE=A long time ago" in self.out)
        self.assertTrue("ISO8601_BUILD_DATE=Some other time" in self.out)

        # checking for the auto-computed values is brittle, so'll just
        # make sure the value contains this year and assume we've got
        # better things to do on December 31st at midnight.
        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.environment 2>&1| grep BUILD_DATE")
        year = datetime.datetime.now().year
        self.assertTrue("ISO8601_BUILD_DATE=%s-" % year in self.out)
        self.assertTrue("BUILD_DATE=%s" % year in self.out)

    def test_buildtest(self):
        """Ensure that none of our test tasks have crept into help, explain
        or generate output."""

        for cmd, exit in [
                ("generate", 0),
                ("generate -O", 0),
                ("explain", 0),
                ("help -v", 2)]:
            self.build(cmd, exit=exit)
            self.assertTrue("buildtest" not in self.out)
        # check that we get buildtest values for generate when running with
        # $ONBLD_BUILDTEST
        self.build(
            "generate -O -D build_tools=/opt",
            env_arg={"ONBLD_BUILDTEST": "true"})
        self.assertTrue("check.buildtest.fail1.must_pass = true" in self.out)
        self.assertTrue("check.buildtest.fail1 = true" in self.out)

    def test_build_i_args(self):
        """Ensure that we can run build -i with unquoted and quoted args."""
        self.build(
            "generate -D build_tools=/opt "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        # A single quoted argument
        self.build("-qi 'uname -a'")
        self.assertTrue("SunOS" in self.out)
        self.assertTrue(self.mach in self.out)

        # Two separate arguments - build(1) should be passing all of these
        # through. If we just get SunOS in the output, then something's wrong.
        # Use 'shell' here instead of -i for some variety.
        self.build("shell -q uname -a")
        self.assertTrue("SunOS" in self.out)
        self.assertTrue(self.mach in self.out)

    def test_nightly_task_inclusion(self):
        """Run the 'build nightly' task in a dummy workspace and ensure
        appropriate build tasks are started. We don't emulate enough of
        the workspace to cause all build tasks to succeed, but we should
        still start them all and should disable the tasks that should
        be disabled."""

        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D check.rti_ready=false " +
            IGNORE_ALL_CHECKS +
            " -D pkgsurf=false "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        def log_presence(logs, log_exists=True):
            dir = "%s/log.%s/last" % (self.workspace, self.mach)
            for logfile in logs:
                if log_exists:
                    self.assertTrue(
                        os.path.exists("%s/%s" % (dir, logfile)),
                        "%s is missing from log dir" % logfile)
                else:
                    self.assertTrue(
                        not os.path.exists("%s/%s" % (dir, logfile)),
                        "%s was unexpectedly found in log dir" % logfile)

        self.build("nightly")
        log_presence([
            "nightly.log", "nightly.log.html", "setup.log", "setup-nd.log",
            "packages.log", "packages-nd.log.html",
            "check.parfait.log", "check.parfait-nd.log.html",
            "check.findunref.log", "mail_msg"])
        # wsdiff is disabled by default
        log_presence(["check.wsdiff.log"], log_exists=False)

        # Check that a "last" symlink has been created
        logdir = "%s/log.%s" % (self.workspace, self.mach)
        old_last_linkdest = os.readlink("%s/last" % logdir)
        logdir_contents = os.listdir(logdir)

        # Run a debug-only build
        self.build("nightly -d")

        # Check that our logs have actually been rotated
        new_logdir_contents = os.listdir(logdir)
        self.assertTrue(logdir_contents != new_logdir_contents)
        self.assertTrue(os.readlink("%s/last" % logdir) != old_last_linkdest)

        log_presence([
            "install.log", "setup.log", "packages.log", "mail_msg"])
        log_presence([
            "install-nd.log", "setup-nd.log", "packages-nd.log"],
            log_exists=False)

        self.build("nightly -D")
        log_presence([
            "install-nd.log", "setup-nd.log", "packages-nd.log", "mail_msg"])
        log_presence([
            "install.log", "setup.log", "packages.log"], log_exists=False)

        # unlink the "last" link and point to a non-existent directory, to
        # make sure that the link rotation can cope with that
        os.unlink("%s/last" % logdir)
        os.symlink("%s/dummy" % logdir, "%s/last" % logdir)

        self.build("nightly -dD")
        log_presence([
            "nightly.log", "nightly.log.html", "setup.log", "setup-nd.log",
            "packages.log", "packages-nd.log.html",
            "check.parfait.log", "check.parfait-nd.log.html",
            "check.findunref.log", "mail_msg"])

        # now disable all checks, and ensure that we don't see check.parfait or
        # check.findunref
        self.build("nightly -C")
        log_presence([
            "nightly.log", "nightly.log.html", "setup.log", "setup-nd.log",
            "mail_msg"])
        log_presence([
            "check.parfait.log", "check.parfait-nd.log.html",
            "check.findunref.log", "check.wsdiff.log"],
            log_exists=False)

        # Just disable parfait
        self.build("nightly -P")
        log_presence([
            "nightly.log", "nightly.log.html", "setup.log", "setup-nd.log",
            "mail_msg", "check.findunref.log.html"])
        log_presence([
            "check.parfait.log", "check.parfait-nd.log.html",
            "check.wsdiff.log"], log_exists=False)

        # and now disable parfait and check.findunref and enable check.wsdiff
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D check.rti_ready=false "
            "-D check.parfait=false "
            "-D check.findunref=false "
            "-D check.wsdiff=true " +
            IGNORE_ALL_CHECKS +
            " -D pkgsurf=false "
            " -D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        self.build("nightly")
        log_presence([
            "nightly.log", "nightly.log.html", "setup.log", "setup-nd.log",
            "mail_msg", "check.cstyle.log", "check.wsdiff.log"])
        log_presence([
            "check.parfait.log", "check.parfait-nd.log.html",
            "check.findunref.log", "check.rti_ready.log"], log_exists=False)

    def test_nightly_fail(self):
        """Run the 'build nightly' task in a dummy workspace configured to
        always fail."""

        self.create_workspace(self.workspace, fail=True)
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' " +
            IGNORE_ALL_CHECKS +
            " -D ws=%s > %s/bldrc" % (self.workspace, self.workspace))
        self.build("nightly", exit=1)
        self.assertTrue("CMDERR  : *** Error code 1" in self.err)
        self.assertTrue("SUMMARY : 'tools' failed" in self.err)

        # Make sure we don't really depend on "*** Error code 1" appearing
        # and that all we look for is the command exit code.
        self.create_workspace(self.workspace, fail="silent")
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' " +
            IGNORE_ALL_CHECKS +
            " -D ws=%s > %s/bldrc" % (self.workspace, self.workspace))
        self.build("nightly -v", exit=1)
        self.assertTrue("SUMMARY : 'tools' failed" in self.err)

        # check that we bailed out as expected
        self.assertTrue("Bailing out now" in self.err)
        # We should never have tried to run check.parfait since the failure in
        # setup should have prevented us from getting any further.
        self.assertTrue("SUMMARY : 'check.parfait'" not in self.err)

    def test_timestamp(self):
        """Test that the log_format value of the bldrc file works."""

        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("here")
        with open(
                "%s/log.%s/here.log" % (self.workspace, self.mach)) as logfile:
            for line in logfile:
                self.assertTrue(
                    line.startswith(
                        "[%s" % datetime.datetime.now().strftime("%Y%m%d")))
        self.build(
            "generate -D build_tools=/opt "
            "-D log_format=notimestamp > %s/bldrc" % self.workspace)
        self.build("here")
        with open(
                "%s/log.%s/here.log" % (self.workspace, self.mach)) as logfile:
            for line in logfile:
                self.assertTrue(
                    not line.startswith(
                        "[%s" % datetime.datetime.now().strftime("%Y%m%d")))

    def test_hg(self):
        """Test that hg info is accurately recorded. We use a mixture of
        just calling onbld.build.get_hg_info(..) and buildtest.hginfo, because
        we'd like both to work."""

        # we're not doing any branch testing here
        env = {"ONBLD_HG_BRANCH": "default"}

        # get info
        info = onbld.build.get_hg_info(
            self.workspace, env, force_cache_refresh=True)
        new_info = onbld.build.get_hg_info(self.workspace, env)
        self.assertTrue(info == new_info)

        # modify a file in the workspace
        with open("%s/usr/src/Makefile" % self.workspace, "a") as makefile:
            makefile.write("\n")

        # we only spot changes when forcing a cache refresh
        new_info = onbld.build.get_hg_info(self.workspace, env)
        new_info_refresh = onbld.build.get_hg_info(
            self.workspace, env, force_cache_refresh=True)
        self.assertTrue(new_info[self.workspace] == info[self.workspace])
        self.assertTrue(
            new_info_refresh[self.workspace] != info[self.workspace])

        # check that hg branch tip doesn't show the updated data
        new_info_tip = onbld.build.get_hg_info(
            self.workspace, env, hg_identify=False, force_cache_refresh=True)
        self.assertTrue(
            new_info_tip[self.workspace] != new_info_refresh[self.workspace])

        # revert the file. Having forced a cache refresh of the
        # 'hg_identify=False' data above, that should also clear the cache
        # for the hg_identify=True data as well.
        self.cmdrun("hg revert %s/usr/src/Makefile" % self.workspace)
        new_info = onbld.build.get_hg_info(self.workspace, env)
        self.assertTrue(new_info[self.workspace] == info[self.workspace])

        # remove some subtrees
        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("buildtest.hginfo")
        self.assertTrue(info["%s/usr/closed" % self.workspace][0] in self.err)
        shutil.rmtree("%s/usr/closed/.hg" % self.workspace)

        self.build("buildtest.hginfo")
        self.assertTrue(
            info["%s/usr/closed" % self.workspace][0] not in self.err)
        shutil.rmtree("%s/usr/man/.hg" % self.workspace)
        self.build("buildtest.hginfo")
        self.assertTrue(info["%s/usr/man" % self.workspace][0] not in self.err)

    def test_pkgarchive_removal(self):
        """Ensure that when we run the 'packages' task, we always end up with
        a fresh PKGARCHIVE directory. When incremental builds do package
        publication properly, this test will start failing."""

        for incremental_flag in ["", "-i"]:
            self.create_workspace(self.workspace, noisy=False)

            pkg_archive = os.path.join(
                self.workspace, "data", "pkg", "%s-debug" % self.mach)
            pkg_archive_nd = os.path.join(
                self.workspace, "data", "pkg", "%s-nondebug" % self.mach)

            # make a package archive, and dump some files in there.
            os.makedirs(pkg_archive)
            os.makedirs(pkg_archive_nd)
            open(os.path.join(pkg_archive, "foo"), "w").close()
            open(os.path.join(pkg_archive_nd, "foo"), "w").close()

            self.build(
                "generate -D build_tools=/opt "
                "-D mailto='<None>' "
                "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))
            self.build("packages %s" % incremental_flag)

            # check that the files we left there originally were removed
            # as part of "build packages"
            self.assertTrue(
                not os.path.exists(os.path.join(pkg_archive, "foo")))
            self.assertTrue(
                not os.path.exists(os.path.join(pkg_archive_nd, "foo")))

    def test_cores(self):
        """Ensure that the build has coreadm settings set to save core files
        into the workspace cores directory. This also tests the check.cores
        task to ensure it looks for cores in the right place."""

        # For this test to run in a known environment, we might need to
        # change the coreadm settings for the process running the test.
        # We could do this in the main unit test setUp/tearDown, but it's
        # only needed for one test, so that would be overkill.

        ret, out, err = self.cmdrun("coreadm")
        if "per-process core dumps: disabled" in out:
            self.logger.debug(
                "Skipping test: per-process core dumps are disabled")
            return

        unit_test_pid = os.getpid()
        ret, out, err = self.cmdrun("coreadm %s" % unit_test_pid)
        original_pattern = out.strip().split()[1]
        if original_pattern != "core":
            self.logger.info(
                "default coreadm settings for unit test process "
                "(%s) changed" % unit_test_pid)
            self.cmdrun("coreadm -p core %s" % unit_test_pid)

        with open("%s/dump_core.c" % self.workspace, "w") as core_c:
            core_c.write("main;\n")
        try:
            ret, out, err = self.cmdrun(
                "cc -o %s/dump_core -w %s/dump_core.c" %
                (self.workspace, self.workspace))
        except onbld.util.UnexpectedExitCodeException:
            self.logger.debug(
                "Skipping test: unable to compile unit test dump_core.c: "
                "%s" % out)
            return

        with open("%s/usr/src/Makefile" % self.workspace, "w") as makefile:
            makefile.write("\n")
            makefile.write(
                "install all bldtools setup check ctf-check lint pmodes:\n")
            makefile.write("\t%s/dump_core\n" % self.workspace)

        self.build(
            "generate -D build_tools=/opt "
            "-D debug=false "
            "-D check=false "
            "-D mailto='<None>' -k "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))
        self.assertTrue(not os.path.exists("%s/keep/cores" % self.workspace))

        ws_cores = "%s/keep/cores" % self.workspace
        # run a build which is known to core dump. We run the check.cores
        # task, because this build core dumps in 'setup' before we get to run
        # checks. This also tests that the 'nightly' task does include the
        # tstamp task
        self.build("nightly -v", exit=1)
        # fake some NFS cruft in the cores directory
        nfs_cruft_path = "%s/.nfs12345" % ws_cores
        open(nfs_cruft_path, "w").close()

        # we have exactly 2 entries, one for setup and one piece of NFS cruft.
        self.assertTrue(len(os.listdir("%s" % ws_cores)) == 2)
        self.build("check.cores -v", exit=1)
        self.assertTrue(os.path.exists(ws_cores))
        self.assertTrue(self.err.count("core.dump_core.") == 1)

        ret, out, err = self.cmdrun("file %s/*" % ws_cores)
        self.assertTrue("cores/core.dump_core" in out)
        self.assertTrue("ELF " in out)

        # This time, we should have several files in our cores directory, but
        # check.find_cores should only identify one of them (the one
        # created by this setup task, not the previous nightly build)
        self.build("tstamp")
        self.build("setup", exit=1)
        self.build("check.cores -v", exit=1)

        self.assertTrue(len(os.listdir(ws_cores)) == 3)
        self.assertTrue(self.err.count("core.dump_core.") == 1)

        # check that our interactive shell has this setting
        self.build("-qi coreadm '$$'")
        self.assertTrue("cores/core.%f.%m.%p@%t\tdefault" in self.out)

        # check that after clobbering, we have no cores at all and check.cores
        # passes.
        self.build("clobber")
        self.assertTrue(len(os.listdir(ws_cores)) == 0)
        self.build("tstamp")
        open(nfs_cruft_path, "w").close()
        self.build("check.cores")
        os.remove(nfs_cruft_path)
        os.rmdir(ws_cores)

        # check that we can create cores outside the workspace
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D debug=false "
            "-D cores_pattern=%s/temp_cores/core.%%f.%%p "
            "-D ws=%s > %s/bldrc" %
            (self.testdir, self.workspace, self.workspace))
        self.build("setup", exit=1)
        self.build("tstamp")
        self.build("setup", exit=1)
        self.build("check.cores -v", exit=1)

        self.assertTrue(len(os.listdir("%s/temp_cores" % self.testdir)) == 2)
        self.assertTrue(self.err.count("core.dump_core.") == 1)

        # clobber should not have removed the cores, since the cores dir is
        # outside the workspace.
        self.build("clobber")
        self.assertTrue(
            len(os.listdir("%s/temp_cores" % self.testdir)) == 2,
            "Expected to find 2 cores outside the workspace")

        shutil.rmtree("%s/temp_cores" % self.testdir)

        # This time, specify '<None>' for cores_pattern resulting in the
        # default behaviour, to write cores in the directory.
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D debug=false "
            "-D cores_pattern='<None>' "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        self.build("setup -v", exit=1)
        self.build("check.cores -v", exit=1)

        # check that a core file was written to the workspace rather than
        # a cores directory
        self.assertTrue(not os.path.exists(ws_cores))
        ret, out, err = self.cmdrun("file %s/usr/src/*" % self.workspace)
        self.assertTrue("core:\t" in out)
        self.assertTrue("ELF " in out)

        # make sure coreadm(1) inside a build shell agrees
        self.build("-qi coreadm '$$'")
        self.assertTrue("\tcore\tdefault" in self.out)

        # Ensure that we fail with an empty cores pattern
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D debug=false "
            "-D cores_pattern= "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        self.build("here", cwd="%s/usr/src" % self.workspace, exit=1)
        self.assertTrue(
            "Error: The value of 'cores_pattern' cannot be empty." in self.err)

        # Ensure we fail early if we can't create the cores directory
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D debug=false "
            "-D cores_pattern=/dev/null/cores.%%p "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))
        self.build("here", cwd="%s/usr/src" % self.workspace, exit=1)
        self.assertTrue(
            "Error: Unable to use specified coreadm dir" in self.err)

        if original_pattern != "core":
            self.logger.info(
                "Restoring original coreadm pattern %s" % original_pattern)
            self.cmdrun(
                "coreadm -p %s %s" % (original_pattern, unit_test_pid))

    def test_nightly_tstamp(self):
        """Ensure we record hg information in the build timestamp file."""

        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' " +
            IGNORE_ALL_CHECKS +
            " -D debug=false "
            "-D pkgsurf=false "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        self.build("nightly")
        self.assertTrue(
            os.path.exists(
                "%s/keep/.build.%s.tstamp" % (self.workspace, self.mach)))
        self.assertTrue(
            os.path.exists(
                "%s/keep/.build.%s.tstamp.prev" % (self.workspace, self.mach)))
        # These should always be identical at the end of a successful build.
        # During a build, assuming new changesets have appeared in the
        # workspace, the timestamp file contents will differ.
        old = open(
            "%s/keep/.build.%s.tstamp.prev" %
            (self.workspace, self.mach)).read()
        new = open(
            "%s/keep/.build.%s.tstamp" % (self.workspace, self.mach)).read()
        self.assertTrue(old == new, "Old and new tstamp files differed!")

        # save the prev timestamp so that it doesn't get clobbered by the
        # next successful nightly.
        shutil.copy(
            "%s/keep/.build.%s.tstamp.prev" % (self.workspace, self.mach),
            "%s/keep/.build.%s.tstamp.old" % (self.workspace, self.mach))

        # Make a change to a file in the workspace, and verify that
        # the new build timestamp changes
        with open("%s/usr/src/Makefile" % self.workspace, "a") as makefile:
            makefile.write("#\n")

        self.build("nightly")
        new = open(
            "%s/keep/.build.%s.tstamp" % (self.workspace, self.mach)).read()
        new_rotated = open(
            "%s/keep/.build.%s.tstamp.prev" %
            (self.workspace, self.mach)).read()
        self.assertTrue(
            new != old, "build timestamps did not change across changes")

        self.logger.debug("old version info:")
        self.logger.debug(old)

        self.logger.debug("new version info:")
        self.logger.debug(new)

        # Check that versions were recorded and the files were correctly
        # rotated.
        ret, out, err = self.cmdrun("hg log -r tip --template '{node|short}'")
        self.assertTrue(out in old)
        self.assertTrue(out + "+" not in old)
        self.assertTrue(out + "+" in new)
        self.assertTrue(out + "+" in new_rotated)

        # usr/closed and usr/man changesets should not have changed
        for ws in ["usr/closed", "usr/man"]:
            cmd = "hg -R %s/%s log -r tip --template '{node|short}'" % \
                (self.workspace, ws)
            ret, out, err = self.cmdrun(cmd)
            self.assertTrue(out in old)
            self.assertTrue(out in new)
            self.assertTrue(out in new_rotated)

    def test_findunref(self):
        """Check that our findunref task is working. To do this, we simulate
        a nightly build. First we run the 'tstamp' task, which gets run at the
        start of nightly and is used by findunref. We touch(1) various files
        in our dummy workspace so that their access times get updated, which
        findunref can then report on."""

        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D debug=false "
            "-D ws=%s > %s/bldrc" % (self.workspace, self.workspace))

        # For this build task, we create a few dummy files
        dummy_files = ["usr/src/foo", "usr/src/bar"]
        for dummy_path in dummy_files:
            with open("%s/%s" % (self.workspace, dummy_path), "w") as dummy:
                dummy.write("%s\n" % dummy_path)
            self.cmdrun(
                "hg -R %(ws)s add %(ws)s/%(dummy)s" %
                {"ws": self.workspace, "dummy": dummy_path})
            self.cmdrun(
                "hg -R %(ws)s commit -m 'add %(dummy)s' %(ws)s/%(dummy)s" %
                {"ws": self.workspace, "dummy": dummy_path})

        tstamp_path = "%s/keep/.build.%s.tstamp" % (self.workspace, self.mach)

        # need to create a dummy exception_list for findunref, we can't use
        # /dev/null since build(1) hardcodes the findunref arguments. Note
        # that tools are always built debug only.
        unref_srcdir = \
            "%s/data/build.%s/usr/src/tools/findunref" % (
                self.workspace, self.mach)
        os.makedirs(unref_srcdir)
        open("%s/exception_list" % unref_srcdir, "w").close()

        # need to install the build machine copy of findunref since our dummy
        # workspace doesn't have any tools built
        toolsroot_onbld = \
            "%s/data/proto/root_%s_tools/opt/onbld/bin/%s" % \
            (self.workspace, self.mach, self.mach)
        os.makedirs(toolsroot_onbld)
        shutil.copy(
            "/opt/onbld/bin/%s/findunref" % self.mach, toolsroot_onbld)

        # this should fail because we don't have a tstamp
        self.build("check.findunref -v", exit=1)
        self.assertTrue(
            "findunref: fatal: cannot stat timestamp file" in self.err)
        self.build("tstamp")
        tstamp_stat = os.stat(tstamp_path)

        # now touch all the files
        self.cmdrun(
            "find %s/usr -exec touch {} + > /dev/null" % self.workspace)
        self.build("check.findunref -v")

        # touch all the files except one, which we give an old date to.
        # Expect to see findunref complain about that old file.
        self.cmdrun(
            "find %s/usr -exec /usr/bin/touch {} +" % self.workspace)
        os.utime(
            "%s/usr/src/foo" % self.workspace,
            (tstamp_stat.st_atime - 1, tstamp_stat.st_mtime - 1))

        self.build("check.findunref -v", exit=1)
        self.assertTrue("> ./usr/src/foo" in self.err)
        self.assertTrue("./usr/src/bar" not in self.err)

        # Now check the same with the bar file, and see if we get
        # the fixed noise appearing as a result of the previous findunref
        self.build("tstamp")
        tstamp_stat = os.stat(tstamp_path)
        self.cmdrun("find %s/usr -exec /usr/bin/touch {} +" % self.workspace)
        os.utime(
            "%s/usr/src/bar" % self.workspace,
            (tstamp_stat.st_atime - 1, tstamp_stat.st_mtime - 1))

        self.build("check.findunref -v", exit=1)
        self.assertTrue("> ./usr/src/bar" in self.err)
        self.assertTrue("< ./usr/src/foo" in self.err)

        # touch all files and we should see only the bar file going away
        self.build("tstamp")
        self.cmdrun("find %s/usr -exec /usr/bin/touch {} +" % self.workspace)
        # technically, this should be a pass, since we're fixing findunref
        # but because a difference was found, the test will fail
        self.build("check.findunref -v", exit=1)
        self.assertTrue("< ./usr/src/bar" in self.err)
        self.assertTrue("./usr/src/foo" not in self.err)

        # touch all files and findunref should now pass
        self.build("tstamp")
        self.cmdrun("find %s/usr -exec /usr/bin/touch {} +" % self.workspace)
        self.build("check.findunref -v")

    def test_pass(self):
        """Check that a nightly pass returns cleanly."""
        self.build("generate -D build_tools=/opt "
                   "-D mailto='<None>' > %s/bldrc" % self.workspace)
        self.build("buildtest.nightly_pass", exit=0)

    def test_fail(self):
        """Check that a nightly fail returns non-zero."""
        self.build("generate -D build_tools=/opt "
                   "-D mailto='<None>' > %s/bldrc" % self.workspace)
        self.build("buildtest.nightly_fail", exit=1)

    def test_must_pass(self):
        """Check that must_pass references in bldrc files are functional."""
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' > %s/bldrc" % self.workspace)
        # nightly_conditional runs several tests, two of which fail,
        # of those, fail1 must pass by default, and fail2 can fail but is
        # non-fatal by default
        self.build("buildtest.nightly_conditional", exit=1)
        self.build(
            "buildtest.nightly_conditional "
            "-O check.buildtest.fail1.must_pass=false", exit=0)
        self.assertTrue(
            "'check.buildtest.fail1' failed (non-fatal)" in self.err)
        self.assertTrue(
            "'check.buildtest.fail2' failed (non-fatal)" in self.err)

        self.build(
            "buildtest.nightly_conditional "
            "-O check.buildtest.fail1.must_pass=false "
            "-O check.buildtest.fail2.must_pass=true", exit=1)
        self.assertTrue(
            "'check.buildtest.fail1' failed (non-fatal)" in self.err)
        self.assertTrue("'check.buildtest.fail2' failed," in self.err)

        self.build(
            "buildtest.nightly_conditional "
            "-O check.buildtest.fail1.must_pass=true "
            "-O check.buildtest.fail2.must_pass=true", exit=1)
        self.assertTrue("'check.buildtest.fail1' failed," in self.err)
        self.assertTrue("'check.buildtest.fail2' failed," in self.err)

        # this time, we use the bldrc file rather than -O to build to ensure
        # options are being passed from the bldrc correctly. We set
        # ONBLD_BUILDTEST to include the buildtest values in the 'generate'
        # output.
        env = {"ONBLD_BUILDTEST": "true"}
        self.build(
            "generate -D build_tools=/opt "
            "-D mailto='<None>' "
            "-D check.buildtest.fail1.must_pass=false "
            "-D check.buildtest.fail2.must_pass=true "
            "> %s/bldrc" % self.workspace, env_arg=env)
        self.build("buildtest.nightly_conditional", exit=1)
        self.assertTrue(
            "'check.buildtest.fail1' failed (non-fatal)" in self.err)
        self.assertTrue("'check.buildtest.fail2' failed," in self.err)

    def test_unittest_basics(self):
        """Check that the build(1) unit test framework functions."""

        def write_test(path, contents):
            with open(path, "w") as test_cfg:
                test_cfg.write(contents)

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))

        # ensure we can deal with no test configurations
        self.build("check.tests")
        self.build("check.all_tests")

        bad_path = "%s/usr/src/tools/test.d/bad.cfg" % self.workspace
        write_test(bad_path, "simple]\n")
        self.build("check.all_tests -v", exit=1)
        self.assertTrue("Unable to parse test config" in self.err)
        self.assertTrue(bad_path in self.err)

        # add a single test that passes
        good_path = "%s/usr/src/tools/test.d/good.cfg" % self.workspace
        write_test(
            good_path, "[good_test]\n" +
            "script=/usr/bin/touch\n" +
            "args=@WS@/yay_we_ran\n")

        # we should still fail since we've got a bad config
        self.build("check.all_tests -v", exit=1)
        self.assertTrue("Unable to parse test config" in self.err)
        self.assertTrue(bad_path in self.err)
        self.assertTrue(
            not os.path.exists(os.path.join(self.workspace, "yay_we_ran")))
        os.remove(bad_path)

        self.build("check.all_tests -v")
        self.assertTrue("/usr/bin/touch" in self.err)
        self.assertTrue(
            os.path.exists(os.path.join(self.workspace, "yay_we_ran")))
        os.remove(good_path)

        write_test(
            bad_path, """
[good_test]
script=/usr/bin/touch
synopsis=This is a good test
# note that we *do not* encourage tests to write to $WS, but for
# this unit-test test, it's ok.
args=@WS@/yay_we_ran

[bad_test]
script=/usr/bin/false
synopsis=This will fail

# This check that we do actually verify privileges(7). We intentionally
# use a privilege that doesn't exist to force us to skip this test.
[ugly_test]
script=/usr/bin/touch
args=@WS@/does_not_exist
synopsis=This test can never run
privileges=there_is_no_spoon

# a non-executable file
[non_exec]
script=/etc/passwd

# a non-existent file
[non_exist]
script=/usr/bin/there_is_no_spoon

""")
        self.build("check.all_tests -v", exit=1)
        self.assertTrue("This is a good test" in self.err)
        self.assertTrue("there_is_no_spoon" in self.err)
        self.assertTrue(
            "insufficient privileges - required there_is_no_spoon" in self.err)
        self.assertTrue("Skipped test ugly_test" in self.err)
        self.assertTrue("TWARN   : Test non_exist from " in self.err)
        self.assertTrue("TWARN   : Test non_exec from " in self.err)

        self.assertTrue(
            not os.path.exists(os.path.join(self.workspace, "does_not_exist")))

        # we should catch duplicate section names
        bad_path2 = "%s/usr/src/tools/test.d/bad2.cfg" % \
            self.workspace
        write_test(
            bad_path2, """
[good_test]
script=/usr/bin/touch
synopsis=This is a good test
# note that we *do not* encourage tests to write to $WS, but for
# this unit-test test, it's ok.
args=@WS@/yay_we_ran
""")
        self.build("check.all_tests -v", exit=1)
        self.assertTrue(
            "Section names must be unique but duplicates "
            "were discovered: good_test" in self.err)

        os.remove(bad_path2)
        os.remove(bad_path)

        # this should pass, and we should ignore the skipped test
        write_test(good_path, """
[hurrah_test]
script=/usr/bin/touch
synopsis=Another test that passes
privileges=basic
args=@WS@/this_ran_too

[ignored_test]
script=/usr/bin/true
synopsis=This test can never run
privileges=there_is_no_spoon

[non_exec]
script=/etc/passwd

[non_exist]
script=/usr/bin/there_is_no_spoon
""")
        self.build("check.all_tests -v", exit=1)
        self.assertTrue("Another test that passes" in self.err)
        self.assertTrue("there_is_no_spoon" in self.err)
        self.assertTrue("Skipped test ignored_test" in self.err)
        self.assertTrue("TWARN   : Test non_exist from " in self.err)
        self.assertTrue("TWARN   : Test non_exec from " in self.err)

        self.assertTrue(
            "insufficient privileges - required there_is_no_spoon" in self.err)
        self.assertTrue(
            os.path.exists(os.path.join(self.workspace, "this_ran_too")))
        os.remove(good_path)

    def test_unittest_changes(self):
        """Test that we run tests due to files or hg changesets which match
        our test patterns, and that tests only run if files were changed
        since we last did a build."""

        def write_test(path, contents):
            with open(path, "w") as test_cfg:
                test_cfg.write(contents)

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))

        test_path = "%s/usr/src/tools/test.d/test.cfg" % self.workspace
        write_test(test_path, """
[my_test]
script = /usr/bin/touch
args = @WS@/ran_my_test
# only run the test if my_file changes
pattern.basic = my_file
""")
        test_ran_indicator = os.path.join(self.workspace, "ran_my_test")
        open(os.path.join(self.workspace, "before_tstamp"), "w").close()
        self.build("tstamp")
        self.build("check.tests -v")
        self.assertTrue(not os.path.exists(test_ran_indicator))
        dummy_path = os.path.join(self.workspace, "dummy")
        open(dummy_path, "w").close()
        open(os.path.join(self.workspace, "dummy"), "w").close()
        self.cmdrun("hg add %s" % dummy_path)
        self.build("check.tests -v")
        self.assertTrue(not os.path.exists(test_ran_indicator))
        # Ensure that we spotted the changed file, but not the file
        # that was last modified before we built.
        self.assertTrue("dummy" in self.err)
        self.assertTrue("before_tstamp" not in self.err)

        self.build("tstamp")
        # modify a file that should trigger the test
        my_file_path = os.path.join(self.workspace, "my_file")
        open(my_file_path, "w").close()
        self.cmdrun("hg add %s" % my_file_path)
        self.build("check.tests -v")
        self.assertTrue(os.path.exists(test_ran_indicator))
        self.assertTrue("my_file" in self.err)
        self.assertTrue("dummy" not in self.err)
        self.assertTrue("before_tstamp" not in self.err)

        write_test(test_path, """
[my_hg_test]
script = /usr/bin/touch
args = @WS@/ran_hg_test
# only run the test if my_file changes
pattern.basic = my_hg_file
""")
        test_ran_indicator = os.path.join(self.workspace, "ran_hg_test")
        self.build("tstamp")
        ignore_hg_path = os.path.join(self.workspace, "unrelated")
        open(ignore_hg_path, "w").close()
        self.cmdrun("hg add %s" % ignore_hg_path)
        self.cmdrun("hg commit -m 'foo' %s" % ignore_hg_path)
        self.build("check.tests -v")
        self.assertTrue(not os.path.exists(test_ran_indicator))

        trigger_hg_path = os.path.join(self.workspace, "my_hg_file")
        open(trigger_hg_path, "w").close()
        self.cmdrun("hg add %s" % trigger_hg_path)
        self.cmdrun("hg commit -m 'bar' %s" % trigger_hg_path)
        self.build("check.tests -v")
        self.assertTrue(os.path.exists(test_ran_indicator))
        os.remove(test_ran_indicator)

        self.build("tstamp")
        self.build("check.tests -v")
        # no new hg changesets, so nothing should run
        self.assertTrue(not os.path.exists(test_ran_indicator))
        open(trigger_hg_path, "w").write("foo")
        # now we've modified the trigger file, the test should run
        self.build("check.tests -v")
        self.assertTrue(os.path.exists(test_ran_indicator))
        self.assertTrue("my_hg_file" in self.err)
        os.remove(test_ran_indicator)

    def create_build_repo(
            self, pkgs_path, nightly_mfs, extra_mfs,
            flavors=None, mach=None):
        """Create package repositories similar to the ones done by the
        build(1) packages task, and publish packages to them."""

        if not mach:
            mach = self.mach

        if not flavors:
            flavors = [
                "{0}-debug".format(mach),
                "{0}-nondebug".format(mach)]

        for flavor in flavors:
            pkg_flavor = "{0}/{1}".format(pkgs_path, flavor)
            if not os.path.isdir(pkg_flavor):
                os.makedirs(pkg_flavor, 0o777)
            repo = "{0}/repo.osnet".format(pkg_flavor)
            self.cmdrun("pkgrepo create {0}".format(repo))
            self.cmdrun(
                "pkgrepo -s {0} set publisher/prefix=nightly".format(repo))
            self.cmdrun(
                "pkgrepo -s {0} add-publisher localizable".format(repo))
            for nightly_mf in nightly_mfs:
                self.cmdrun(
                    "pkgsend publish -s {0} {1}".format(repo, nightly_mf))

            if not extra_mfs:
                return
            repo = "{0}/{1}/repo.osnet-internal".format(
                pkgs_path, flavor)
            self.cmdrun("pkgrepo create {0}".format(repo))
            self.cmdrun(
                "pkgrepo -s {0} set publisher/prefix=extra".format(
                    repo))
            for extra_mf in extra_mfs:
                self.cmdrun(
                    "pkgsend publish -s {0} {1}".format(repo, extra_mf))

    def test_pkgsurf(self):
        """Test that the pkgsurf task can resurface two repositories.
        This is not an exhaustive test of pkgmerge, it merely tests that
        build(1) is accessing the correct repositories."""

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))

        mf1 = """
set name=pkg.fmri value=pkg:/buildtest/test@1.0
file /usr/bin/ls path=usr/bin/ls owner=root group=sys mode=0755
"""
        # the build uses a -c argument to exclude the resurfacing of
        # the system/osnet-splice package. Verify that continues to work.
        mf1_splice = """
set name=pkg.fmri value=pkg:/system/osnet-splice@1.0
"""

        extra_name = "consolidation/osnet/osnet-internal-incorporation"

        mf1_extra = """
set name=pkg.fmri value=pkg:/{0}@1.0
file /usr/bin/cat path=usr/bin/cat owner=root group=sys mode=0755
""".format(extra_name)

        mf2 = """
set name=pkg.fmri value=pkg:/buildtest/test@2.0
file /usr/bin/ls path=usr/bin/ls owner=root group=sys mode=0755
"""

        mf2_splice = """
set name=pkg.fmri value=pkg:/system/osnet-splice@2.0
"""

        # note the intentional lack of content change here, relying
        # instead on the pkgsurf task to explicitly forbid resurfacing
        # of the internal incorporation
        mf2_extra = """
set name=pkg.fmri value=pkg:/{0}@2.0
file /usr/bin/cat path=usr/bin/cat owner=root group=sys mode=0755
""".format(extra_name)

        def write_temp(suffix, content):
            fd, path = tempfile.mkstemp(suffix, dir=self.testdir)
            open(path, "w").write(content)
            return path

        mf1_path = write_temp("_mf1.p5m", mf1)
        mf1_splice_path = write_temp("_mf1_splice.p5m", mf1_splice)

        mf2_path = write_temp("_mf2.p5m", mf2)
        mf2_splice_path = write_temp("_mf2_splice.p5m", mf2_splice)

        mf1_extra_path = write_temp("_extra_mf1.p5m", mf1_extra)
        mf2_extra_path = write_temp("_extra_mf2.p5m", mf2_extra)

        # Create our reference pkgsurf repositories
        pkgsurf_ref = "{0}/keep/pkgsurf-ref.pkg".format(self.workspace)
        self.create_build_repo(
            pkgsurf_ref, [mf1_path, mf1_splice_path], [mf1_extra_path])
        # Create the repositories published by this build
        pkgpub = "{0}/data/pkg/".format(self.workspace)
        self.create_build_repo(
            pkgpub, [mf2_path, mf2_splice_path], [mf2_extra_path])
        self.build("pkgsurf -v")

        flavors = [
            "{0}-debug".format(self.mach),
            "{0}-nondebug".format(self.mach)]

        # check for resurfaced repos
        for flavor in flavors:
            repo = "{0}/{1}/repo.osnet".format(pkgpub, flavor)
            ret, out, err = self.cmdrun("pkgrepo -s {0} list".format(repo))
            self.assertTrue(re.search(r" buildtest/test\s*1.0:", out))
            self.assertTrue(not re.search(r" buildtest/test\s*2.0:", out))
            # make sure we did not resurface system/osnet-splice
            self.assertTrue(re.search(r"system/osnet-splice\s*2.0:", out))

            repo = "{0}/{1}/repo.osnet-internal".format(pkgpub, flavor)
            ret, out, err = self.cmdrun("pkgrepo -s {0} list".format(repo))
            # we should not have resurfaced here since mf2_extra delivers
            # only the "incorporation"
            self.assertTrue(re.search(r" {0}\s*2.0:".format(extra_name), out))
            self.assertTrue(
                not re.search(r" {0}\s*1.0:".format(extra_name), out))

        shutil.rmtree(pkgsurf_ref)
        shutil.rmtree(pkgpub)

        # check that we can resurface when the internal repos are missing
        # and do this for debug packages only.
        pkgsurf_ref = "{0}/keep/pkgsurf-ref.pkg".format(self.workspace)
        self.create_build_repo(
            pkgsurf_ref, [mf1_path, mf1_splice_path], None,
            flavors=["{0}-debug".format(self.mach)])
        pkgpub = "{0}/data/pkg".format(self.workspace)
        self.create_build_repo(
            pkgpub, [mf2_path, mf2_splice_path], None,
            flavors=["{0}-debug".format(self.mach)])
        self.build("pkgsurf -v", exit=1)
        # build thinks the closed repo should be there, so we fail
        self.build("pkgsurf -vd", exit=1)
        # now do a open-only build
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s -D closed_is_present=no > %s/bldrc" %
            (self.workspace, self.workspace))
        self.build("pkgsurf -vd")

        # check that when we have no reference resurface repositories,
        # (the case when we have a freshly checked-out workspace) we exit
        # with a 0 exit code and a mild warning.
        shutil.rmtree(pkgsurf_ref)
        self.build("pkgsurf -vd")
        self.assertTrue("no such reference repository" in self.err)

        # when we have no repositories, we exit with an error
        shutil.rmtree(pkgpub)
        self.build("pkgsurf -vd", exit=1)

    def test_pkgmerge(self):
        """Test that we can pkgmerge debug/non-debug/sparc/x86 repositories.
        The pkgmerge task does more than this test can exercise, optionally
        waiting for an "alternative architecture" build to complete before
        running. We don't have a good way to run background tasks in this
        test framework, so that functionality needs to be tested manually.
        """

        def write_temp(suffix, content):
            fd, path = tempfile.mkstemp(suffix, dir=self.testdir)
            open(path, "w").write(content)
            return path

        # the pkgmerge support in build has sanity tests to ensure we're
        # trying to merge published ON repositories, so the package names
        # here matter.
        mf = """
set name=pkg.fmri value=pkg:/system/core-os@1.0
file {0} path=usr/bin/ls owner=root group=sys mode=0755
file /usr/bin/cat path=usr/bin/cat owner=root group=sys mode=0755
"""

        mf_extra = """
set name=pkg.fmri value=pkg:/test/gk-test@1.0
file {0} path=usr/bin/cat owner=root group=sys mode=0755
"""
        mf_badver = mf.replace("@1.0", "@2.0")

        # Create a workspace that we pretend is of the other architecture
        alt_ws_path = tempfile.mkdtemp("_altws", dir=self.testdir)
        self.create_workspace(alt_ws_path)

        # map build flavors to the binaries we actually for those flavors
        # allowing us to fake alternate architecture deliveries
        mach_map = {
            "{0}-debug".format(self.mach): "/usr/bin/ls",
            "{0}-nondebug".format(self.mach): "/usr/bin/uname"
        }
        altmach_map = {
            "{0}-debug".format(self.altmach): "/usr/sbin/prtconf",
            "{0}-nondebug".format(self.altmach): "/usr/bin/id"
        }

        mach_flavors = [
            "{0}-debug".format(self.mach),
            "{0}-nondebug".format(self.mach)]
        altmach_flavors = [
            "{0}-debug".format(self.altmach),
            "{0}-nondebug".format(self.altmach)]

        # Create sparc and x86 debug and non-debug repositories
        for ws, mach, bin_map, flavors in [
                (self.workspace, self.mach, mach_map, mach_flavors),
                (alt_ws_path, self.altmach, altmach_map, altmach_flavors)]:
            pkgpub = os.path.join(ws, "data", "pkg")
            for flavor in flavors:
                mf_path = write_temp("_mf.p5m", mf.format(bin_map[flavor]))
                mf_extra_path = write_temp(
                    "_extra_mf.p5m", mf_extra.format(bin_map[flavor]))
                self.create_build_repo(
                    pkgpub, [mf_path], [mf_extra_path], flavors=[flavor],
                    mach=mach)

        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))

        # we can't merge when we are only building a single flavor and
        # there's no altws_pkgarchive specified
        self.build("pkgmerge -vd", exit=1)
        self.build("pkgmerge -vD", exit=1)

        # A default 2-way debug/non-debug merge
        self.build("pkgmerge -v")
        self.assertTrue(
            "2-way debug/nondebug single-arch pkgmerge" in self.err)

        merged = os.path.join(
            self.workspace, "data", "pkg",
            "merged-{0}".format(self.mach), "repo.osnet")
        ret, out, err = self.cmdrun(
            "pkgrepo -s {0} contents core-os".format(merged))
        # we have 3 file actions
        self.assertTrue(out.count("file ") == 3)
        # we have at least produced some variants
        self.assertTrue("variant.debug.osnet=true" in out)
        self.assertTrue("variant.debug.osnet=false" in out)
        # we have 2 usr/bin/ls actions
        self.assertTrue(out.count("usr/bin/ls") == 2)
        # we have 1 usr/bin/cat action
        # (since the same content was delivered for debug and nondebug)
        self.assertTrue(out.count("usr/bin/cat") == 1)

        merged = os.path.join(
            self.workspace, "data", "pkg", "merged-{0}".format(self.mach),
            "repo.osnet-internal")
        ret, out, err = self.cmdrun(
            "pkgrepo -s {0} contents gk-test".format(merged))
        # we have at least produced some variants
        self.assertTrue("variant.debug.osnet=true" in out)
        self.assertTrue("variant.debug.osnet=false" in out)
        self.assertTrue("variant.arch=i386" not in out)
        self.assertTrue("variant.arch=sparc" not in out)
        # we have 2 usr/bin/cat file actions
        self.assertTrue(out.count("file ") == 2)
        self.assertTrue(out.count("usr/bin/cat") == 2)

        # when using a single workspace NFS-shared to two build machines,
        # one of them will not need to do a pkgmerge, since their pkgarchive
        # path will be the same as the altws_pkgarchive path. This should not
        # be considered a failure.
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s "
            "-D altws_pkgarchive=@PKGARCHIVE@ > %s/bldrc" %
            (self.workspace, self.workspace))
        self.build("pkgmerge -v")
        self.assertTrue(
            "Not doing a pkgmerge for this build since the debug "
            "pkgarchive path is identical to the altws_pkgarchive path"
            in self.err)

        # a 4-way debug/nondebug/sparc/i386 merge
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s "
            "-D altws_pkgarchive=%s/data/pkg/%s-debug > %s/bldrc" %
            (self.workspace, alt_ws_path, self.altmach, self.workspace))

        merged = os.path.join(
            self.workspace, "data", "pkg", "merged", "repo.osnet")
        self.build("pkgmerge -v")
        self.assertTrue("4-way debug/nondebug/sparc/i386 pkgmerge" in self.err)
        ret, out, err = self.cmdrun(
            "pkgrepo -s {0} contents core-os".format(merged))
        # we have 5 file actions
        self.assertTrue(out.count("file ") == 5)
        # we have at least produced some variants
        self.assertTrue("variant.debug.osnet=true" in out)
        self.assertTrue("variant.debug.osnet=false" in out)
        self.assertTrue("variant.arch=i386" in out)
        self.assertTrue("variant.arch=sparc" in out)

        # we have 4 usr/bin/ls actions
        self.assertTrue(out.count("usr/bin/ls") == 4)
        # we have 1 usr/bin/cat action
        # (since the same content was delivered for debug and nondebug
        #  for both architectures)
        self.assertTrue(out.count("usr/bin/cat") == 1)

        # delete the non-debug altws packages and verify we fail
        shutil.rmtree(os.path.join(
            alt_ws_path, "data", "pkg", "{0}-nondebug".format(self.altmach),
            "repo.osnet-internal"))
        self.build("pkgmerge -v", exit=1)

        merged = os.path.join(
            self.workspace, "data", "pkg", "merged-debug", "repo.osnet")
        # a 2-way sparc/i386 single-flavor merge
        self.build(
            "generate -D build_tools=/opt -D debug=true -D nondebug=false "
            "-D check=false -D ws=%s "
            "-D altws_pkgarchive=%s/data/pkg/%s-debug > %s/bldrc" %
            (self.workspace, alt_ws_path, self.altmach, self.workspace))

        self.build("pkgmerge -v")
        self.assertTrue("2-way sparc/i386 single-flavor pkgmerge" in self.err)
        ret, out, err = self.cmdrun(
            "pkgrepo -s {0} contents core-os".format(merged))
        # we have at least produced some variants
        self.assertTrue("variant.arch=i386" in out)
        self.assertTrue("variant.arch=sparc" in out)
        self.assertTrue("variant.debug.osnet=true" not in out)
        self.assertTrue("variant.debug.osnet=false" not in out)

        self.assertTrue(out.count("file ") == 3)
        self.assertTrue(out.count("usr/bin/ls") == 2)
        self.assertTrue(out.count("usr/bin/cat") == 1)

        # delete our internal-only repo and verify we fail
        shutil.rmtree(merged)
        shutil.rmtree(os.path.join(
            self.workspace, "data", "pkg",
            "{0}-nondebug".format(self.mach),
            "repo.osnet-internal"))
        self.build(
            "generate -D build_tools=/opt "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))
        self.build("pkgmerge -v", exit=1)

        # but it works works for an open-only build
        self.build(
            "generate -D build_tools=/opt -D closed_is_present=no "
            "-D check=false -D ws=%s > %s/bldrc" %
            (self.workspace, self.workspace))
        self.build("pkgmerge -v")

        # finally, verify that when we publish a newer package version
        # and attempt to merge, that we exit with an error
        repo = "{0}/data/pkg/{1}-nondebug/repo.osnet".format(
            self.workspace, self.mach)
        badver_path = write_temp("_mf.p5m", mf_badver.format(
            mach_map["{0}-debug".format(self.mach)]))
        self.cmdrun("pkgsend publish -s {0} {1}".format(repo, badver_path))
        self.build("pkgmerge -v", exit=1)

    def test_buildshell(self):
        """Basic test to ensure that build(1) is rejecting inappropriate
        user shells"""

        # ensure a few common shells work
        env = copy.deepcopy(os.environ)
        shells = [
            "/usr/bin/bash", "/usr/bin/ksh93", "/usr/bin/sh", "/usr/bin/pfsh"]
        for sh in shells:
            env["SHELL"] = sh
            self.build(
                "generate -D build_tools=/opt > %s/bldrc" % self.workspace)
            self.build("-i /usr/bin/true", env_arg=env, exit=0)
            self.assertTrue(
                "Error: {0} is not supported.".format(sh) not in self.err)

        env["SHELL"] = "/usr/bin/fakesh"
        self.build("generate -D userenv=SHELL > %s/bldrc" % self.workspace)
        self.build("-i /usr/bin/true", env_arg=env, exit=1)
        self.assertTrue("Error: /usr/bin/fakesh is not supported." in self.err)

    def test_test(self):
        """Test the 'test' subcommand. Note that test_unittest_basics and
        test_unittest_changes exercise more of the framework"""

        # relies on snoop.cfg and crypto-fips.cfg being copied
        # to our test workspace in setup()
        self.build("test snoop")
        self.assertTrue("Running test snoop" in self.err)
        self.build("test snoop -v")
        self.assertTrue("No test script associated with snoop" in self.err)

        self.build("test snoop -v crypto-fips-140")
        self.assertTrue("No test script associated with snoop" in self.err)
        self.assertTrue(
            "No test script associated with crypto-fips-140" in self.err)

        self.build("test noodles", exit=1)
        self.assertTrue("Unknown tests: noodles" in self.err)
        self.build("test noodles carrots", exit=1)
        self.assertTrue("Unknown tests: carrots noodles" in self.err)

        # Verify that we list all known tests when given no args
        self.build("test")
        self.assertTrue("crypto-fips.cfg" in self.out)
        self.assertTrue("snoop.cfg" in self.out)
        self.assertTrue("Warn about PSARC 2010/429 EOF of snoop" in self.out)
        self.assertTrue(
            "Crypto Framework FIPS 140-2 Boundary Change note." in self.out)

    def test_no_workspace(self):
        """Verify that only certain build subcommands can run without
        a workspace."""

        self.build("test -v")
        self.assertTrue("Warn about PSARC 2010/429 EOF of snoop" in self.out)
        shutil.rmtree(self.workspace)

        # If our test machine has a test.d directory in /opt which has content,
        # we expect build to pass, otherwise it should fail, but in all cases,
        # it shouldn't complain about a missing workspace. We need to work a
        # little harder to get the expected exit status right.
        expected_exit = 1
        if os.path.exists("/opt/onbld/test/test.d/"):
            for file in os.listdir("/opt/onbld/test/test.d/"):
                if fnmatch.fnmatch(file, '*.cfg'):
                    expected_exit = 0
                    break
        self.build("test -v", exit=expected_exit, cwd=self.testdir)

        self.assertTrue(
            "Unable to determine a valid workspace" not in self.err)
        self.build("shell", exit=1, cwd=self.testdir)
        self.assertTrue("Unable to determine a valid workspace" in self.err)

    def test_pythonpath(self):
        """Verify we only set PYTHONPATH when PKG7_ROOT is set."""
        self.build("generate -D build_tools=/opt > %s/bldrc" % self.workspace)
        self.build("-iq env")
        self.assertTrue("PYTHONPATH=" not in self.out)
        self.assertTrue("PKG7_ROOT=" in self.out)

        self.build(
            "generate -D pkg7_root=/tmp/a -D build_tools=/opt > %s/bldrc" %
            self.workspace)
        self.build("-iq env")
        self.assertTrue(
            "PYTHONPATH=/tmp/a/usr/lib/python2.7/vendor-packages" in self.out)
        self.assertTrue("PKG7_ROOT=/tmp/a" in self.out)

        self.build(
            "generate -D pythonpath=/foo -D build_tools=/opt> %s/bldrc" %
            self.workspace)
        self.build("-iq env")
        self.assertTrue("PYTHONPATH=" not in self.out)

    def test_reexec(self):
        """Verify we attempt to re-exec the workspace copy of build.py"""

        # make a fake build.py wrapper in our workspace that we can check for
        scripts = os.path.join(self.workspace, "usr/src/tools/scripts")
        os.makedirs(scripts)
        fake_build_path = os.path.join(scripts, "build.py")
        with open(fake_build_path, "w") as fake_build:
            fake_build.write(
                "#!/usr/bin/python\n"
                "print('We are running fake build')\n")

        os.chmod(fake_build_path, 0o755)

        # testing the test suite itself - by default always runs
        # the version from this real workspace.
        self.build("buildtest.pass")
        self.assertTrue("re-exec" not in self.err)

        # verify the dummy build script in our test workspace is ok
        ret, out, err = self.cmdrun(fake_build_path)
        self.assertTrue("We are running fake build" in out)

        # verify we attempt to reexec it
        self.build("buildtest.pass", allow_reexec=True)
        self.assertTrue("We are running fake build" in self.out)
        # verify our debug log message isn't emitted by default
        self.assertTrue("re-exec build:" not in self.err)

        os.environ["ONBLD_DEBUG_REEXEC_LOG"] = "true"
        self.build("buildtest.pass", allow_reexec=True)
        self.assertTrue("We are running fake build" in self.out)
        self.assertTrue("re-exec build:" in self.err)
        del os.environ["ONBLD_DEBUG_REEXEC_LOG"]

        # check we get the right exit code, normally 2 for help, but
        # not with our faked version
        self.build("help", allow_reexec=True, exit=0)
        self.assertTrue("We are running fake build" in self.out)
        self.assertTrue("re-exec build:" not in self.err)

        # verify we don't attempt to reexec
        self.build("buildtest.pass", allow_reexec=False)
        self.assertTrue("We are running fake build" not in self.out)

        # again check we really get the right exit code
        self.build("help", allow_reexec=False, exit=2)
        self.assertTrue("We are running fake build" not in self.err)

        self.build("help", allow_reexec=True, cwd="/tmp", exit=2)
        self.assertTrue(
            "Skipping re-exec since no workspace found" in self.err)

        os.unlink(fake_build_path)
        self.build("buildtest.pass", allow_reexec=True)
        self.assertTrue(
            "Skipping re-exec since no workspace build.py found" in self.err)

    def test_akv_values(self):
        """Verify that AKV_* and akv override values are handled correctly."""

        self.build("generate", exit=0)
        # akv_gate_id is a build option override whereas the others are GK
        # overrides.  Only akv_gate_id should show up in a "normal" generate.
        self.assertTrue("akv_branch" not in self.out)
        self.assertTrue("akv_build" not in self.out)
        self.assertTrue("akv_component" not in self.out)
        self.assertTrue("akv_gate_id" in self.out)
        self.assertTrue("akv_version_base" not in self.out)

        # Verify that the values show up when generating with gk options
        # included.
        self.build("generate -k", exit=0)
        self.assertTrue("akv_branch" in self.out)
        self.assertTrue("akv_build" in self.out)
        self.assertTrue("akv_component" in self.out)
        self.assertTrue("akv_gate_id" in self.out)
        self.assertTrue("akv_version_base" in self.out)

        self.build("generate > {0}/bldrc".format(self.workspace), exit=0)
        self.build("-i env", exit=0)
        # AKV_BRANCH, AKV_COMPONENT, and AKV_VERSION_BASE really have no use
        # outside of gatekeeper builds.  They allow setting of the AK package
        # version string for release purposes.  They have appropriate defaults
        # in the build itself (Makefile.versions).  These should be unset
        # unless we specifically override them in the build.
        self.assertTrue("AKV_BRANCH" not in self.out)
        self.assertTrue("AKV_COMPONENT" not in self.out)
        self.assertTrue("AKV_VERSION_BASE" not in self.out)
        # AKV_GATE_ID and AKV_BUILD may or may not be set depending on the
        # build.

        # Validate that the override values do indeed show up in the
        # environment.
        self.build("generate -D akv_gate_id=ak -D akv_branch=1.20 "
                   "-D akv_build=1 -D akv_component=2013.06.05.8.0 "
                   "-D akv_version_base=2013.06.05.8.0,1-1.20 "
                   "> {0}/bldrc".format(self.workspace), exit=0)
        self.build("-i env", exit=0)
        # Expect all of these to be present and set to the values passed into
        # generate.
        self.assertTrue("AKV_BRANCH=1.20" in self.out)
        self.assertTrue("AKV_BUILD=1" in self.out)
        self.assertTrue("AKV_COMPONENT=2013.06.05.8.0" in self.out)
        self.assertTrue("AKV_GATE_ID=ak" in self.out)
        self.assertTrue("AKV_VERSION_BASE=2013.06.05.8.0,1-1.20" in self.out)


class BuildTestRunner(unittest.TextTestRunner):
    """Used so we can modify the environment to allow more verbose logging.
    By default the "-v-" argument just causes unittest to print test
    descriptions before test result status, but we want more verbosity than
    that."""

    def __init__(self, stream=sys.stderr, descriptions=1, verbosity=1):
        if verbosity > 1:
            os.environ["ONBLD_DEBUG"] = "true"
        # $ONBLD_BUILDTEST must not be in the environment, since that
        # influences the 'build generate' output, which can break tests
        if "ONBLD_BUILDTEST" in os.environ:
            os.environ.pop("ONBLD_BUILDTEST")

        # When running as a user with an NFS home directory, setting
        # $PYTHONNOUSERSITE can speed up test execution. See python(1).
        os.environ["PYTHONNOUSERSITE"] = "true"
        unittest.TextTestRunner.__init__(
            self, stream=stream, descriptions=descriptions,
            verbosity=verbosity)


if __name__ == '__main__':
    unittest.main(
        testRunner=BuildTestRunner(verbosity=2 if "-v" in sys.argv else 1))
