#!/usr/bin/python

#
# Copyright (c) 2013, 2021, Oracle and/or its affiliates.
#

import errno
import logging
import os
import textwrap
import time
import threading
import shlex
import socket
import subprocess
import signal
from resource import getrlimit, setrlimit, RLIMIT_CORE

import onbld.log

util_logger = logging.getLogger("onbld.util")

"""Utility functions used when building the OS/Net Consolidation.
"""


class UnexpectedExitCodeException(Exception):
    """An exception to capture program failures."""
    def __init__(self, cmdline, exit_state, retcode, output, comment):
        self.cmdline = cmdline
        self.exit_state = exit_state
        self.retcode = retcode
        self.output = output
        self.comment = comment


def debugcmd(cmdline):
    wrapper = textwrap.TextWrapper(
        initial_indent="$ ",
        subsequent_indent="\t",
        break_long_words=False,
        break_on_hyphens=False)
    res = wrapper.wrap(cmdline.strip())
    util_logger.debug(" \\\n".join(res))


def debugresult(retcode, expected, output):
    if output.strip() != "":
        util_logger.debug(output.strip())
    if not isinstance(expected, list):
        expected = [expected]
    if retcode is None or retcode != 0 or retcode not in expected:
        util_logger.debug(
            "[exited %s, expected %s]" %
            (retcode, ", ".join(str(e) for e in expected)))


def reset_pipe():
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def cmdline_run(
        cmdline, comment="", exit=0, logger=None, env_arg=None, cwd=None,
        raise_error=False, shell=False, stdout=None, stderr=None,
        cmdout_level=onbld.log.CMDOUT_LEVEL):
    """If a logging.Logger object is passed in 'logger', we create a
    LoggerThread to read the output of the command when run, writing each line
    of stderr/stdout with onbld.CMDOUT_LEVEL level messages to the provided
    logger and return an integer exit code.

    The command line to be run is logged with onbld.CMDLINE_LEVEL when using a
    logger.

    If 'stderr' and 'stdout' are not None, we run the command using the
    supplied file descriptors or file objects to capture the output of the
    command.

    'cwd' is the path to the working directory the command is to execute with.

    If 'shell' is True, we pass that parameter to Popen - note that this is
    not encouraged and should be used sparingly.

    If logger, or stderr/stdout file descriptors are not passed, then we return
    the retcode, stderr and stdout, otherwise we just return the retcode. If
    exceptions were thrown while trying to run the command, a 99 exit code is
    returned.

    'cmdout_level' is the log level at which to emit output to the log.
    Usually leaving this to the default is sufficient, but when running a
    command where any output at all indicates an error, this can be helpful
    when producing log extracts.
    """

    class LoggerThread(threading.Thread):
        """A class we use to allow us to log messages as they arrive in from
        a subprocess.Popen(..) method call, rather than waiting for all output
        via proc.communicate()"""

        # state definitions for our LoggerThread.
        CMDOUT_STATE = 0
        CMD_STATE = 1

        def __init__(self, logger, cmdout_level=cmdout_level):
            """Create a LoggingThread object, using the logging.Logger passed
            in 'logger'.

            'out_level' is the log level used to write normal cmd output.
            """

            super(LoggerThread, self).__init__()
            self.logger = logger
            self.cmdout_level = cmdout_level

            self.read_fd, self.write_fd = os.pipe()
            self.pipe_fobj = os.fdopen(self.read_fd)
            self.daemon = True

            # A variable to let us know when all input has been read from the
            # pipe and it's safe to return.
            self.drained = False

            # We keep state, to try to cope with shell commands that end with
            # continuation characters like '\'
            self.log_state = None
            self.start()

        def run(self):
            """Read text from our pipe, write to our logger."""

            while True:
                # Read a line of text from the pipe
                line = self.pipe_fobj.readline()
                if not line:
                    self.drained = True
                    self.pipe_fobj.close()
                    return

                line = line.rstrip()
                # Do a rough job of categorizing command output:
                # everything that starts with a '/' or could be a cw(1)
                # subcommand gets logged at "CMD_LEVEL", dmake target warnings
                # and errors are "CMDERR_LEVEL", everything else is
                # treated as "CMDOUT_LEVEL".
                dmake_error = "*** Error"
                dmake_warn = "dmake: Warning: Don't know how to make target"
                dmake_fatal = \
                    "dmake: Fatal error: Don't know how to make target"

                if (line.startswith("/") or line.startswith("+ /") or
                        self.log_state == LoggerThread.CMD_STATE):
                    self.logger.log(onbld.log.CMD_LEVEL, line)
                    self.log_state = LoggerThread.CMD_STATE
                elif (line.startswith(dmake_error) or
                        line.startswith(dmake_warn) or
                        line.startswith(dmake_fatal)) and \
                        "(ignored)" not in line:
                    # This is quite specific to dmake(1). If there's need for
                    # more control over the log-level categorization, or a need
                    # to inspect the state of the LoggerThread, we should
                    # instead allow clients to pass a Visitor-like object that
                    # can be used to inspect log messages and potentially
                    # change log-levels for those messages as they are
                    # received.
                    self.logger.log(onbld.log.CMDERR_LEVEL, line)
                    self.log_state = LoggerThread.CMD_STATE
                else:
                    self.logger.log(self.cmdout_level, line)
                    self.log_state = LoggerThread.CMDOUT_STATE

                # determine our next state
                seen_continuation = line.endswith("\\")
                if (seen_continuation and
                        self.log_state == LoggerThread.CMD_STATE):
                    self.log_state == LoggerThread.CMD_STATE
                else:
                    self.log_state = LoggerThread.CMDOUT_STATE

        def fileno(self):
            return self.write_fd

        def close(self):
            os.close(self.write_fd)

    if env_arg:
        newenv = env_arg
    else:
        newenv = os.environ.copy()

    if cwd:
        newenv["PWD"] = cwd

    if shell:
        cmd = cmdline
    else:
        cmd = shlex.split(cmdline)

    retcode = 99
    if stdout and stderr:
        try:
            p = subprocess.Popen(
                cmd, env=newenv, stdout=stdout, stderr=stderr, cwd=cwd,
                preexec_fn=reset_pipe, shell=shell, universal_newlines=True)
            p.wait()
            return p.returncode
        except KeyboardInterrupt:
            raise
        except Exception:
            return 99

    elif not logger:
        try:
            p = subprocess.Popen(
                cmd, env=newenv, stdout=subprocess.PIPE, preexec_fn=reset_pipe,
                stderr=subprocess.PIPE, cwd=cwd, shell=shell,
                universal_newlines=True)
            output, errout = p.communicate()
            retcode = p.returncode
        except KeyboardInterrupt:
            raise
        except Exception:
            output = ""
            errout = ""
            retcode = 99
        debugresult(retcode, exit, output + errout)

        if raise_error and retcode != exit:
            raise UnexpectedExitCodeException(
                cmdline, exit, retcode, output + errout, comment)

        return retcode, output, errout

    else:
        if cwd:
            logger.log(onbld.log.CMDLINE_LEVEL, "%s (CWD=%s)" % (cmdline, cwd))
        else:
            logger.log(onbld.log.CMDLINE_LEVEL, cmdline)

        # Ideally we'd use two logger threads, one for stderr, another for
        # stdout, allowing us to assign different priorities depending on which
        # fd the output had been written to. In practise this turns out to be
        # difficult, as dmake(1) seems to buffer its stderr and stdout in
        # interesting ways, causing us to be unable to log stderr and stdout
        # in the right order to a single log file. For now, we just use a
        # single LoggerThread.
        log = LoggerThread(logger, cmdout_level=cmdout_level)

        try:
            p = subprocess.Popen(
                cmd, env=newenv, stdout=log, stderr=log, cwd=cwd, bufsize=-1,
                preexec_fn=reset_pipe, shell=shell, universal_newlines=True)
            retcode = p.wait()
            if retcode != exit:
                logger.log(
                    onbld.log.TWARNING_LEVEL,
                    "Command '%s' did not return %s" % (cmdline, exit))
        except KeyboardInterrupt:
            raise
        except Exception as e:
            logger.log(
                onbld.log.TWARNING_LEVEL, "Error running command %s: %s" %
                (cmdline, e))
            if isinstance(e, OSError):
                # pylint bug #1004, fixed in pylint 1.7.0
                # pylint: disable=no-member
                if e.errno == errno.ENOENT and newenv.get("PATH"):
                    # pylint: enable=no-member
                    logger.log(
                        onbld.log.TWARNING_LEVEL, "$PATH was %s" %
                        newenv["PATH"])
        finally:
            # Our LoggerThread uses a pipe, allowing subprocess.Popen to write
            # output from the process to the pipe. LoggerThread reads from that
            # pipe, emitting log messages for each line read. However, it can
            # still be dealing with buffered output, and still writing log
            # messages even when 'p.wait()' above has returned, indicating
            # the process has finished. This is ugly, but closing the write end
            # of the pipe, then spinning till we've read everything that's
            # arriving in seems to work.
            log.close()
            while not log.drained:
                time.sleep(0.02)

        return retcode


def get_hostnames():
    """Return a list of hostnames this system is known by."""
    # socket.gethostname() isn't enough - for VPN users, it's useful to include
    # all domain names we respond to. There's probably a better way to do this.
    cmd = "/usr/sbin/ipadm show-addr -p -o addr"
    ret, out, err = cmdline_run(cmd)
    hostnames = []
    for addr in out.split():
        # ignore the netmask
        addr = addr.split("/")[0]
        if ":" in addr or "?" in addr:
            continue
        elif "->" in addr:
            ipaddr = addr.split("->")[1]
        else:
            ipaddr = addr
        hostnames.append(socket.getfqdn(ipaddr))
    return hostnames


def raise_corelimit():
    """Raise the core ulimit to the current hard limit.
    Produces a warning, but otherwise we continue with the build."""
    softlim, hardlim = getrlimit(RLIMIT_CORE)
    try:
        setrlimit(RLIMIT_CORE, (hardlim, hardlim))
    except Exception:
        return False
    return True
