/*
 * Copyright (c) 2013, 2019, Oracle and/or its affiliates. All rights reserved.
 */

#ifndef _SYS_VERIFIED_BOOT_H
#define	_SYS_VERIFIED_BOOT_H

#include <sys/types.h>
#include <sys/param.h>
#include <sys/crypto/elfsign.h>

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * This defines Verified Boot settings and keys passed from the Hypervisor.
 * Version 0 is the original struct that is being used in SPARC firmware.
 * Version 1 is a newer version that includes DN hash information.
 * Version 1 is currently used in kernel zones.
 * Note: Kernel thread stack sizes are generally smaller than userland stacks.
 * So avoid placing this on the stack in deeply-nested or commonly-used
 * functions.
 */
#define	VERIFY_MODULUS_LEN			256 /* bytes in RSA-2048 key */
#define	VERIFY_DN_HASH_LEN			32 /* bytes in SHA-256 digest */

/* For hash algorithm OID check */
#define	MAX_OID_LEN				100
#define	OID_SHA256WITHRSAENCRYPTION		"1.2.840.113549.1.1.11"

/* Version 0 */
#define	VERIFY_PUBLIC_KEYS_CONFIGURED_V0	7 /* fits in 2KB */

#ifdef	USE_INTERNAL_CERT
#define	VERIFY_PUBLIC_KEYS_PRECONFIGURED	3
#else
#define	VERIFY_PUBLIC_KEYS_PRECONFIGURED	2
#endif	/* USE_INTERNAL_CERT */

#define	MAX_VERIFY_PUBLIC_KEYS_V0	(VERIFY_PUBLIC_KEYS_CONFIGURED_V0 + \
					VERIFY_PUBLIC_KEYS_PRECONFIGURED)

typedef struct verified_boot_info {
	uint8_t boot_policy;
	uint8_t module_policy;
	uint8_t key_count;
	uint8_t version;	/* 0 */
	uint8_t align[4];
	uint8_t public_key_data[][VERIFY_MODULUS_LEN];
} verified_boot_info_v0_t;

typedef uchar_t
    verify_public_keys_t[MAX_VERIFY_PUBLIC_KEYS_V0][VERIFY_MODULUS_LEN];

/* Version 1 */
#define	VERIFY_PUBLIC_KEYS_CONFIGURED_V1	14	/* fits in 4KB */
#define	MAX_VERIFY_PUBLIC_KEYS_V1	(VERIFY_PUBLIC_KEYS_CONFIGURED_V1 + \
					VERIFY_PUBLIC_KEYS_PRECONFIGURED)
typedef struct verified_boot_cert {
	unsigned char key[VERIFY_MODULUS_LEN];
	unsigned char dn_hash[VERIFY_DN_HASH_LEN];
} verified_boot_cert_t;

typedef struct verified_boot_info_v1 {
	uint64_t policy;
	uint64_t key_count;
	verified_boot_cert_t cert[VERIFY_PUBLIC_KEYS_CONFIGURED_V1];
} verified_boot_info_v1_t;

typedef verified_boot_cert_t
    verify_public_certs_t[MAX_VERIFY_PUBLIC_KEYS_V1];

typedef enum verify_policy {
    VERIFY_INVALID, VERIFY_NONE, VERIFY_WARNING, VERIFY_ENFORCE
} verify_policy_t;

#ifdef	sun4v
#define	MAX_CERT_COUNT	(VERIFY_PUBLIC_KEYS_CONFIGURED_V0 - \
			VERIFY_PUBLIC_KEYS_PRECONFIGURED)
#else
#define	MAX_CERT_COUNT	VERIFY_PUBLIC_KEYS_CONFIGURED_V1
#endif	/* sun4v */

/*
 * Sanity checking macros
 */

/* Sanity check the ELF file header to verify it's correct. */
#define	VERIFY_SANITY_CHECK_ELF_HDR(debug_printf, ehdr, filesize, errcode) \
	if (((ehdr)->e_ident[EI_MAG0] != ELFMAG0) || \
	    ((ehdr)->e_ident[EI_MAG1] != ELFMAG1) || \
	    ((ehdr)->e_ident[EI_MAG2] != ELFMAG2) || \
	    ((ehdr)->e_ident[EI_MAG3] != ELFMAG3)) { \
		debug_printf("Error: malformed kernel ELF header, " \
		    "e_ident = 0x%x\n", *(uint32_t *)(ehdr)); \
		(errcode) = -1; \
	} else if ((ehdr)->e_ident[EI_CLASS] != ELFCLASS64) { \
		debug_printf("Error: kernel not ELF64, " \
		    "e_ident[EI_CLASS] = 0x%x\n", (ehdr)->e_ident[EI_CLASS]); \
		(errcode) = -1; \
	} else if ((ehdr)->e_shentsize != sizeof (Elf64_Shdr)) { \
		debug_printf("section header size error: " \
		    "e_shentsize = %u\n", (ehdr)->e_shentsize); \
		(errcode) = -1; \
	} else if (((ehdr)->e_shstrndx == 0) || \
	    (((ehdr)->e_shstrndx >= (ehdr)->e_shnum) && \
	    ((ehdr)->e_shstrndx != SHN_XINDEX))) { \
		debug_printf( \
		    "e_shstrndx invalid: %u > e_shnum  %u\n", \
		    (ehdr)->e_shstrndx, (ehdr)->e_shnum); \
		(errcode) = -1; \
	} else if (((ehdr)->e_shentsize == 0) || ((ehdr)->e_shnum == 0)) { \
		debug_printf( \
		    "section header check error: e_shentsize = %u, " \
		    "e_shnum = %u\n", (ehdr)->e_shentsize, (ehdr)->e_shnum); \
		(errcode) = -1; \
	} else if (((ehdr)->e_shstrndx != SHN_XINDEX) && ((ehdr)->e_shoff + \
	    ((ehdr)->e_shnum * (ehdr)->e_shentsize) > (filesize))) { \
		debug_printf("section end %lu > filesize %u\n", \
		    (ehdr)->e_shoff + ((ehdr)->e_shnum * (ehdr)->e_shentsize), \
		    (filesize)); \
		(errcode) = -1; \
	} else if ((ehdr)->e_phoff + \
	    ((ehdr)->e_phnum * (ehdr)->e_phentsize) > (filesize)) { \
		debug_printf("program segment end %lu > filesize %u\n", \
		    (ehdr)->e_phoff + ((ehdr)->e_phnum * (ehdr)->e_phentsize), \
		    (filesize)); \
		(errcode) = -1; \
	} else { \
		(errcode) = 0; \
	}


/* Checks only for unix/wanboot (not other kernel modules). */
#define	VERIFY_SANITY_UNIX_WANBOOT_ELF_HDR(debug_printf, ehdr, errcode) \
	if ((ehdr)->e_type != ET_EXEC) { \
		debug_printf("Error: boot file not ET_EXEC, e_type = 0x%x\n", \
		    (ehdr)->e_type); \
		(errcode) = -1; \
	} else if (((ehdr)->e_phnum == 0) || ((ehdr)->e_phoff == 0)) { \
		debug_printf( \
		    "Error: no program headers in kernel module\n"); \
		(errcode) = -1; \
	} else { \
		(errcode) = 0; \
	}

#define	VERIFY_SANITY_CHECK_SIGNATURE_SECTION(debug_printf, \
    sig_offset, sig, sig_len, dn, dn_len, sig_oid, sig_oid_len, \
    filesize, errcode) \
	if ((sig_offset) + (sig_len) > (filesize)) { \
		debug_printf( \
		    "signature section 0x%zx+size 0x%zx > file size 0x%x\n", \
		    (sig_offset), (sig_len), (filesize)); \
		(errcode) = -1; \
	} else if (((sig_len) > SIG_MAX_LENGTH) || ((sig_len) == 0)) { \
		debug_printf(": .SUNW_SIGNATURE length error, %ld\n", \
		    (sig_len)); \
		(errcode) = -1; \
	} else if ((dn) == NULL) { \
		debug_printf("NULL signature DN error\n"); \
		(errcode) = -1; \
	} else if (((dn_len) == 0) || ((dn_len) > MAXNAMELEN)) { \
		debug_printf("filesig DN length error, %ld\n", \
		    (dn_len)); \
		(errcode) = -1; \
	} else if ((sig) == NULL) { \
		debug_printf("NULL signature error\n"); \
		(errcode) = -1; \
	} else if (((sig_len) != VERIFY_MODULUS_LEN)) { \
		debug_printf("filesig signature length error, %ld\n", \
		    (sig_len)); \
		(errcode) = -1; \
	} else if ((sig_oid) == NULL) { \
		debug_printf("NULL signature OID error\n"); \
		(errcode) = -1; \
	} else if (((sig_oid_len) > MAX_OID_LEN)) { \
		debug_printf("filesig OID length error " \
		    "sig_oid_len %ld\n", (sig_oid_len)); \
		(errcode) = -1; \
	} else { \
		(errcode) = 0; \
	}

verify_policy_t vboot_get_policy(void);
const char *vboot_get_policy_string(void);

#ifdef	__cplusplus
}
#endif

#endif /* _SYS_VERIFIED_BOOT_H */
