/*
 * Copyright (c) 2003, 2020, Oracle and/or its affiliates.
 */

#ifndef _SYS_ZONE_H
#define	_SYS_ZONE_H

#include <sys/types.h>
#include <sys/mutex.h>
#include <sys/param.h>
#include <sys/rctl.h>
#include <sys/ipc_rctl.h>
#include <sys/pset.h>
#include <sys/tsol/label.h>
#include <sys/cred.h>
#include <sys/netstack.h>
#include <sys/uadmin.h>
#include <sys/ksynch.h>
#include <sys/socket_impl.h>
#include <sys/modhash.h>
#include <sys/paths.h>
#include <sys/nvpair.h>
#include <netinet/in.h>
#include <sys/secext.h>
#include <sys/kidmap_zone.h>
#include <sys/uuid.h>
#include <sys/time.h>
#include <sys/sysinfo.h>

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * NOTE
 *
 * The contents of this file are private to the implementation of
 * Solaris and are subject to change at any time without notice.
 * Applications and drivers using these interfaces may fail to
 * run on future releases.
 */

/* Available both in kernel and for user space */

/* zone id restrictions and special ids */
#define	MAX_ZONEID	9999
#define	MIN_USERZONEID	1	/* lowest user-creatable zone ID */
#define	MIN_ZONEID	0	/* minimum zone ID on system */
#define	GLOBAL_ZONEID	0
#define	ZONEID_WIDTH	4	/* for printf */
#define	SYS_ZONE_PREFIX	"SYS"	/* system-reserved zone prefix. */

/*
 * Special zoneid_t token to refer to all zones.
 */
#define	ALL_ZONES	 (-1)

/*
 * Special token to refer to all zones bound to the default resource
 * pool.
 */
#define	ALL_ZONES_DEFAULT_POOL (-2)

/* system call subcodes */
#define	ZONE_CREATE		0
#define	ZONE_DESTROY		1
#define	ZONE_GETATTR		2
#define	ZONE_ENTER		3
#define	ZONE_LIST		4
#define	ZONE_SHUTDOWN		5
#define	ZONE_LOOKUP		6
#define	ZONE_BOOT		7
/* 8 is reserved to match the S10 subcodes */
#define	ZONE_SETATTR		9
#define	ZONE_ADD_DATALINK	10
#define	ZONE_DEL_DATALINK	11
#define	ZONE_CHECK_DATALINK	12
#define	ZONE_LIST_DATALINK	13
#define	ZONE_LIST_DEFUNCT	14
#define	ZONE_GETATTR_DEFUNCT	15
#define	ZONE_GET_DEVANN		16
#define	ZONE_MARK_AUX_STATE	17
#define	ZONE_UNMARK_AUX_STATE	18
#define	ZONE_GET_AUX_STATE	19
#define	ZONE_LINK_CONSOLE	20
#define	ZONE_UNLINK_CONSOLE	21
#define	ZONE_LIST_DATASET	22
#define	ZONE_WRITE_HOSTDATA	23
#define	ZONE_GET_BOOT_DEVICE	24
#define	ZONE_EMPTY		25
#define	ZONE_ADD_DATASET	26
#define	ZONE_DEL_DATASET	27
#define	ZONE_INFO_NVL		28
#define	ZONE_OP			29
#define	ZONE_OP_END		30
#define	ZONE_OP_UPDATE		31
#define	ZONE_OP_STATE_INIT	32
#define	ZONE_RSS_TIMEOUT	33


/* zone attributes */
#define	ZONE_ATTR_ROOT		1
#define	ZONE_ATTR_NAME		2
#define	ZONE_ATTR_STATUS	3
#define	ZONE_ATTR_PRIVSET	4
#define	ZONE_ATTR_UNIQID	5
#define	ZONE_ATTR_POOLID	6
#define	ZONE_ATTR_INITPID	7
#define	ZONE_ATTR_SLBL		8
#define	ZONE_ATTR_INITNAME	9
#define	ZONE_ATTR_BOOTARGS	10
#define	ZONE_ATTR_BRAND		11
#define	ZONE_ATTR_PHYS_MCAP	12
#define	ZONE_ATTR_SCHED_CLASS	13
#define	ZONE_ATTR_FLAGS		14
#define	ZONE_ATTR_HOSTID	15
#define	ZONE_ATTR_FS_ALLOWED	16
#define	ZONE_ATTR_NETWORK	17
#define	ZONE_ATTR_MAC_PROFILE	18
#define	ZONE_ATTR_INIT_RECOVER	19
#define	ZONE_ATTR_VCPUS		20
#define	ZONE_ATTR_DEDICATED_CPU	21
#define	ZONE_ATTR_NETID		22
#define	ZONE_ATTR_DIO_DEVMAP	23
#define	ZONE_ATTR_BOOT_DEVICE	24
#define	ZONE_ATTR_CONFIGURE_NETWORK 25
#define	ZONE_ATTR_NETAUTO	26
#define	ZONE_ATTR_BOOTARGS_CFG	27
#define	ZONE_ATTR_BOOTARGS_TRAN 28
#define	ZONE_ATTR_NPROCS	29
#define	ZONE_ATTR_VBOOT_INFO	30
#define	ZONE_ATTR_MIG_CLASS	31
#define	ZONE_ATTR_PAGESIZE_POLICY	32
#define	ZONE_ATTR_FEATURES	33
#define	ZONE_ATTR_PHYS_FREE	34
#define	ZONE_ATTR_MRP_KEY	35
#define	ZONE_ATTR_MRP_SVC	36
#define	ZONE_ATTR_HW_PROVIDER	37

/*
 * When getting/setting scheduling class name of a zone by
 * ZONE_ATTR_SCHED_CLASS, we use the following string for the default scheduling
 * class (i.e. not set to specific scheduling class).
 */
#define	ZONE_DFLT_SCHED_CLASS_NAME	"default"

/* In support of dataset aliasing */
#define	ZONE_INVISIBLE_SOURCE	"$globalzone"

/* Start of the brand-specific attribute namespace */
#define	ZONE_ATTR_BRAND_ATTRS	32768

#define	ZONE_FS_ALLOWED_MAX	1024
#define	ZONE_VCPUSTR_MAX	20
#define	ZONE_BOOTDEVSTR_MAX	1024
#define	ZONE_HWPROVIDERSTR_MAX	20

#define	ZONE_EVENT_CHANNEL		"com.sun:zones:status"
#define	ZONE_EVENT_STATUS_CLASS		"status"
#define	ZONE_EVENT_STATUS_SUBCLASS	"change"

#define	ZONE_EVENT_UNINITIALIZED	"uninitialized"
#define	ZONE_EVENT_INITIALIZED		"initialized"
#define	ZONE_EVENT_READY		"ready"
#define	ZONE_EVENT_RUNNING		"running"
#define	ZONE_EVENT_SHUTTING_DOWN	"shutting_down"

#define	ZONE_CB_NAME		"zonename"
#define	ZONE_CB_NEWSTATE	"newstate"
#define	ZONE_CB_OLDSTATE	"oldstate"
#define	ZONE_CB_TIMESTAMP	"when"
#define	ZONE_CB_ZONEID		"zoneid"

/*
 * The maximum number of ZBEs.
 * Keep in sync with ZONE_MAXZBES in usr/src/lib/brand/shared/zone/common.ksh
 */
#define	ZONE_MAXZBES		500

/*
 * Exit values that may be returned by scripts or programs invoked by various
 * zone commands.
 *
 * These are defined as:
 *
 *	ZONE_SUBPROC_OK
 *	===============
 *	The subprocess completed successfully.
 *
 *	ZONE_SUBPROC_UNAVAILABLE
 *	=====================
 *	The zone storage is not accessible or the zone's pkg(7) image is out
 *	of sync with the global zone and as such the zone should enter the
 *	"unavailable" state.
 *
 *	ZONE_SUBPROC_USAGE
 *	==================
 *	The subprocess failed with a usage message, or a usage message should
 *	be output in its behalf.
 *
 *	ZONE_SUBPROC_TRYAGAIN
 *	========================
 *	The subprocess did not complete, but the actions performed by the
 *	subprocess require no recovery actions by the user.
 *
 *	For example, if the subprocess were called by "zoneadm install," the
 *	installation of the zone did not succeed but the user need not perform
 *	a "zoneadm uninstall" before attempting another install.
 *
 *	ZONE_SUBPROC_FATAL
 *	==================
 *	The subprocess failed in a fatal manner, usually one that will require
 *	some type of recovery action by the user.
 *
 *	For example, if the subprocess were called by "zoneadm install," the
 *	installation of the zone did not succeed and the user will need to
 *	perform a "zoneadm uninstall" before another install attempt is
 *	possible.
 *
 *	The non-success exit values are large to avoid accidental collision
 *	with values used internally by some commands (e.g. "Z_ERR" and
 *	"Z_USAGE" as used by zoneadm.)
 */
#define	ZONE_SUBPROC_OK			0
#define	ZONE_SUBPROC_MAX_ZBES_REACHED	251
#define	ZONE_SUBPROC_UNAVAILABLE	252
#define	ZONE_SUBPROC_USAGE		253
#define	ZONE_SUBPROC_TRYAGAIN		254
#define	ZONE_SUBPROC_FATAL		255

#ifdef _SYSCALL32
typedef struct {
	caddr32_t zone_name;
	caddr32_t zone_root;
	caddr32_t zone_privs;
	size32_t zone_privssz;
	caddr32_t rctlbuf;
	size32_t rctlbufsz;
	caddr32_t extended_error;
	caddr32_t zfsbuf;
	size32_t  zfsbufsz;
	int match;			/* match level */
	uint32_t doi;			/* DOI for label */
	caddr32_t label;		/* label associated with zone */
	int flags;
} zone_def32;
#endif
typedef struct {
	const char *zone_name;
	const char *zone_root;
	const struct priv_set *zone_privs;
	size_t zone_privssz;
	const char *rctlbuf;
	size_t rctlbufsz;
	int *extended_error;
	const char *zfsbuf;
	size_t zfsbufsz;
	int match;			/* match level */
	uint32_t doi;			/* DOI for label */
	const bslabel_t *label;		/* label associated with zone */
	int flags;
} zone_def;

/* extended error information */
#define	ZE_UNKNOWN	0	/* No extended error info */
#define	ZE_CHROOTED	1	/* tried to zone_create from chroot */
#define	ZE_AREMOUNTS	2	/* there are mounts within the zone */
#define	ZE_LABELINUSE	3	/* label is already in use by some other zone */
#define	ZE_ARESHARES	4	/* there are GZ shares in the zone's root */

/*
 * zone_status values
 *
 * You must modify zone_status_names in mdb(1)'s genunix module
 * (genunix/zone.c) when you modify this enum.
 *
 * See zone_op_state_t for how these states map to zone op states.
 */
typedef enum {
	ZONE_IS_UNINITIALIZED = 0,
	ZONE_IS_INITIALIZED,
	ZONE_IS_READY,
	ZONE_IS_BOOTING,
	ZONE_IS_RUNNING,
	ZONE_IS_SHUTTING_DOWN,
	ZONE_IS_EMPTY,
	ZONE_IS_DOWN,
	ZONE_IS_DYING,
	ZONE_IS_DEAD
} zone_status_t;
#define	ZONE_MIN_STATE		ZONE_IS_UNINITIALIZED
#define	ZONE_MAX_STATE		ZONE_IS_DEAD

/*
 * Valid commands which may be issued by zoneadm to zoneadmd.  The kernel also
 * communicates with zoneadmd, but only uses Z_REBOOT and Z_HALT.
 * While these are enums, do not renumber them as the interface between
 * zoneadmd and several (brand) scripts uses the enum as number argument.
 * Add new commands before Z_AFTER_LAST_CMD.  Zoneadmd has a zone_cmd_t to
 * string routine (z_cmd_name()); it needs to follow the changes in this file.
 *
 * Do not forget migrate_state_check() when adding to this list.
 */
typedef enum zone_cmd {
	Z_READY = 0,
	Z_MIN_CMD = 0,
	Z_BOOT,
	Z_SHUTDOWN,
	Z_REBOOT,
	Z_HALT,
	Z_UNINSTALL,
	Z_MOUNT,
	Z_SUSPEND,
	Z_UNMOUNT,
	Z_SAVECORE,
	Z_INFO,
	Z_APPLY,
	Z_RESTART,
	Z_LOG,
	Z_ATTACH,
	Z_SYSBOOT,
	Z_DETACH,
	Z_MIGRATE,
	Z_MIGRATE_CHECK,
	Z_SETPROM,
	Z_GETPROM,
	Z_CONFIG_STALE,
	Z_RENAME,
	Z_UNSTALL,
	Z_INSTALL,
	Z_CLONE,
	Z_MIGRATE_THREAD,
	Z_MIGRATE_CRYPTO,
	Z_ZONEADMD_EXIT,
	Z_SUSPEND_CHECK,
	Z_ADD_VF,
	Z_REMOVE_VF,
	Z_GET_POST_INSTALL_OP,
	Z_AFTER_LAST_CMD		/* must be last */
} zone_cmd_t;

/*
 * zoneadm set arguments
 */
#define	ZONE_VAR_NAME_LEN	256
#define	ZONE_VAR_VALUE_LEN	2048
typedef struct {
	/* Be sure it has space for '=' and '\0'. */
	char buf[ZONE_VAR_NAME_LEN + ZONE_VAR_VALUE_LEN + 2];
	boolean_t clear;
} zone_cmd_promargs_t;

/* zonecfg_digest_t must be kept in sync with uu_fnv1a_64_t */
typedef uint64_t zonecfg_digest_t;

/*
 * Defined type and flags for zoneadmd flags.
 */
typedef enum {
	Z_CMD_FLAG_NONE			= 1 << 0,
	Z_CMD_FLAG_FORCE		= 1 << 1,
	Z_CMD_FLAG_RW			= 1 << 2,
	Z_CMD_FLAG_TRANSIENT_RW		= 1 << 3,
	Z_CMD_FLAG_REBOOT		= 1 << 4,
	Z_CMD_FLAG_NO_RESUME		= 1 << 5,
	Z_CMD_FLAG_INIT_FAILED		= 1 << 6,
	Z_CMD_FLAG_SAVECORE_LIVE	= 1 << 7,
	Z_CMD_FLAG_QUIET		= 1 << 8,
	Z_CMD_FLAG_DRYRUN		= 1 << 9,
	Z_CMD_FLAG_LOG_DEBUG		= 1 << 10,
	Z_CMD_FLAG_LOG_NODEBUG		= 1 << 11,
	Z_CMD_FLAG_INIT_HOSTDATA	= 1 << 12,
	Z_CMD_FLAG_TAKEOVER		= 1 << 13,
	Z_CMD_FLAG_INSTALLING		= 1 << 14,
	Z_CMD_FLAG_MIGRATING		= 1 << 15,
	Z_CMD_FLAG_TRANSIENT_BOOTARGS	= 1 << 16,
	Z_CMD_FLAG_STALL		= 1 << 17,
	Z_CMD_FLAG_STOR_CREATE_MISSING	= 1 << 18,
	Z_CMD_FLAG_FORCE_STOR_DES_ALL	= 1 << 19,
	Z_CMD_FLAG_FINAL_BOOTARGS	= 1 << 20,
	Z_CMD_FLAG_SAVECORE_SUSPEND	= 1 << 21,
	Z_CMD_FLAG_MIGRATED		= 1 << 22,
	Z_CMD_FLAG_PRE_INSTALL		= 1 << 23,
	Z_CMD_FLAG_FOR_CLONE		= 1 << 24
} admdflags_t;

/*
 * The union holding the post install op data.
 * If other calls to zoneadmd require additional rval data, add a new member
 * to this union.
 */
typedef union {
	struct {
		boolean_t pi_reboot; /* if false then the zone halted */
		boolean_t pi_failed; /* the halt/reboot failed */
	} post_install_op;
} zone_cmd_rval_data_t;

/*
 * The structure of a request to zoneadmd.
 */
typedef struct zone_cmd_arg {
	/* unique "generation number" */
	uint64_t uniqid;
	/* requested action */
	zone_cmd_t cmd;
	/* flags passed down from zonecfg/zoneadm */
	admdflags_t flags;
	/* locale in which to render messages */
	char locale[MAXPATHLEN];
	union {
		/* arguments passed to zone_boot() */
		char bootbuf[BOOTARGS_MAX];
		/* for set-prom and get-prom */
		zone_cmd_promargs_t promargs;
		/* for migrate */
		struct {
			int nr_threads;
			boolean_t suspend_at_start;
		} migrate;
		/* for HA failover interface */
		datalink_id_t linkid;
		/* for other things that lack definition in system headers */
		struct {
			int size;
			uchar_t buf[BOOTARGS_MAX - sizeof (int)];
		} generic;
	} args;
	/* post install op data */
	zone_cmd_rval_data_t rval_data;
	/* LZR baseline digest */
	zonecfg_digest_t digest;
} zone_cmd_arg_t;

#if __STDC_VERSION__ >= 199901L
/*
 * Used for "set" request responses from zoneadmd.
 */
typedef struct {
	size_t		result_size;
	boolean_t	is_nvlist;
	char		result[];
} zone_set_result_t;
#endif

/* This is the property created for devices visible in non-global zone. */
#define	ZONE_PROP_NAME		"zone"

/*
 * Structure of zoneadmd's response to a request.  A NULL return value means
 * the caller should attempt to restart zoneadmd and retry.
 */

typedef enum {
	ZCE_OK = 0,
	ZCE_ERR,
	ZCE_ENOTSUP,
	ZCE_INVALID_KERNEL_CALL,
	ZCE_INVALID_REQ,
	ZCE_INVALID_STATE,
	ZCE_MODIFIED_LCFG,
	ZCE_NO_ZONE,
	ZCE_PENDING_OP,
	ZCE_STATE_CHANGE,
	ZCE_BUSY
} zone_cmd_err_t;

/*
 * Log message header. The message string follows immediately after the header.
 * The next header can follow on the next 64-bit aligned address after the end
 * of the string.
 */
typedef struct zlog_msg {
	int zlm_level;		/* message priority, see syslog.h */
	size_t zlm_len;		/* length of message's string */
} zlog_msg_t;

/*
 * zone_cmd_rval_t:
 *		       |---------------|
 *		       |	       |
 *		       ~	       ~
 *		    /  |---------------|
 *		    |  |  zlog_msg_t   | = rval_log
 *		    |  |---------------|
 *   rval_log_size  |  | A message\0   |
 *		    |  |---------------|
 *		    |  |  zlog_msg_t   |
 *		    |  |---------------|
 *		    |  | A bit longer  |
 *		    |  | error	       |
 *		    |  | message\0     |
 *		    \  |---------------|
 *		       |  zlog_msg_t   | <-- zlog_t.next_msg
 *		       |---------------|
 */
typedef struct zone_cmd_rval {
	/* return value of request */
	zone_cmd_err_t rval;
	/* post install data */
	zone_cmd_rval_data_t rval_data;
	/* zlog size in bytes */
	size_t rval_log_size;
	/* log messages (variable size) */
	zlog_msg_t rval_log;
} zone_cmd_rval_t;

/*
 * Possible actions to take to recover when the init process for a zone fails.
 */
typedef enum {
	Z_RESTART_INIT = 0,
	Z_REBOOT_ZONE,
	Z_HALT_ZONE
} zone_init_t;

#define	ZONE_STATE_MAXSTRLEN	14

/*
 * Zone Operation event model.
 *
 * Zone operations:
 *
 * ZOP_NONE: No operation is occurring.
 *
 * ZOP_INIT: Set on zone_op_state_init().
 *
 * ZOP_CREATING: Set on zonecfg create.
 *
 * ZOP_INSTALLING: Set on zoneadm attach, clone or install.
 *
 * ZOP_UNINSTALLING: Set on zoneadm detach or uninstall.
 *
 * ZOP_READYING: Set on zoneadm ready.
 *
 * ZOP_BOOTING: Set on zoneadm boot or on the way up of a zone reboot.
 *
 * ZOP_HALTING: Set on zoneadm halt or shutdown.
 *
 * ZOP_MARKING: Set on zoneadm mark.
 *
 * ZOP_DELETING: Set on zonecfg delete.
 */
typedef enum {
	ZOP_INVALID = -1,
	ZOP_NONE = 0,
	ZOP_INIT,
	ZOP_CREATING,
	ZOP_INSTALLING,
	ZOP_UNINSTALLING,
	ZOP_READYING,
	ZOP_BOOTING,
	ZOP_HALTING,
	ZOP_DELETING,
	ZOP_MARKING,
	ZOP_MAX
} zone_op_t;
#define	ZONE_OP_VALID(x) \
	(((x) > ZOP_INVALID) && ((x) < ZOP_MAX))

/*
 * Zone OP states:
 *
 * The states meaningful from an admin perspective.
 *
 * ZOPS_CONFIGURED: on-disk state. A zone gets into this state as result of the
 * successful operation ZOP_UNINSTALLING or as result of zoneadm mark.
 *
 * ZOPS_INCOMPLETE: on-disk state. A zone gets into this state as result of a
 * failed attempt to perform ZOP_INSTALLING or as result of zoneadm mark.
 *
 * ZOPS_UNAVAILABLE: on-disk state. A zone gets into this state as result of
 * a failed attempt to perform ZOP_INSTALLING or as result of zoneadm mark.
 *
 * ZOPS_INSTALLED: on-disk state. A zone gets into this state as result of the
 * successful operations ZOP_INSTALLING, ZOP_HALTING or failed attempt to
 * perform ZOP_BOOTING or as result of zoneadm mark.
 *
 * ZOPS_READY: in-kernel state. A zone gets into this state as result of the
 * successful operation ZOP_READYING.
 *
 * ZOPS_RUNNING: in-kernel state. A zone gets into this state as result of the
 * successful operation ZOP_BOOTING.
 *
 * ZOPS_DOWN: in-kernel state. A zone gets into this state as result of
 * a failed attempt to perform ZOP_HALTING.
 *
 *
 * There are two other enumerations for zone states:
 *
 * zone_status_t values describe in-kernel states that are implementation
 * details.
 *
 * zone_state_t values describe on-disk and in-kernel states that are
 * implemention detail also consumed in brand scripts. These values are
 * currently used to represent state in zoneadm(8) and in libzonecfg's
 * zone_get_state(). Eventually zoneadm(8) and other administrative interfaces
 * will use only zone_op_state_t values.
 *
 * Mappings of zone_state_t and zone_status_t into Zone OP states:
 *
 * ZOPS_CONFIGURED:	ZONE_STATE_CONFIGURED
 *
 * ZOPS_INCOMPLETE:	ZONE_STATE_INCOMPLETE
 *
 * ZOPS_UNAVAILABLE:	ZONE_STATE_UNAVAILABLE
 *
 * ZOPS_INSTALLED:	ZONE_STATE_INSTALLED
 * 			ZONE_STATE_MOUNTED
 * 			ZONE_IS_UNINITIALIZED
 * 			ZONE_IS_INITIALIZED
 * 			ZONE_IS_DOWN
 * 			ZONE_IS_DYING
 * 			ZONE_IS_DEAD
 *
 * ZOPS_READY:		ZONE_STATE_READY
 * 			ZONE_IS_READY
 *
 * ZOPS_RUNNING:	ZONE_STATE_RUNNING
 * 			ZONE_IS_BOOTING
 *			ZONE_IS_RUNNING
 *
 * ZOPS_DOWN:		ZONE_STATE_SHUTTING_DOWN
 * 			ZONE_STATE_DOWN
 * 			ZONE_IS_SHUTTING_DOWN
 * 			ZONE_IS_EMPTY
 *
 * ZOPS_INVALID:	indication that there is no mapping or that we do not
 *			have a valid zone state or zone operation code.
 *			NB: the kernel itself will never produce, maintain
 *			or deal with ZOPS_INVALID.
 */
typedef enum {
	ZOPS_INVALID = -1,
	ZOPS_CONFIGURED = 0,
	ZOPS_INCOMPLETE,
	ZOPS_UNAVAILABLE,
	ZOPS_INSTALLED,
	ZOPS_READY,
	ZOPS_RUNNING,
	ZOPS_DOWN,
	ZOPS_MAX
} zone_op_state_t;
#define	ZONE_ZOPS_VALID(x) \
	(((x) > ZOPS_INVALID) && ((x) < ZOPS_MAX))

/*
 * Keys for ZONE_INFO_NVL
 */
#define	ZONE_INFO_NVL_STATE	"zone_info_nvl_state"
#define	ZONE_INFO_NVL_OP	"zone_info_nvl_op"
#define	ZONE_INFO_NVL_LAST_OP	"zone_info_nvl_last_op"
#define	ZONE_INFO_NVL_AUXSTATE	"zone_info_nvl_auxstate"

/*
 * The zone support infrastructure uses the zone name as a component
 * of unix domain (AF_UNIX) sockets, which are limited to 108 characters
 * in length, so ZONENAME_MAX is limited by that. This size includes the
 * terminating NUL.
 */
#define	ZONENAME_MAX		64

#define	GLOBAL_ZONENAME		"global"

/*
 * Extended Regular expression (see regex(7)) which matches all valid zone
 * names.
 */
#define	ZONENAME_REGEXP		"[a-zA-Z0-9][-_.a-zA-Z0-9]{0,62}"

/*
 * Where the zones support infrastructure places temporary files.
 * ZONES_TMPDIR is for files that can be read by other users (mode 755).
 * CONTAINER_STATE_DIR is for files that can be read by other users (mode 755).
 * ZONES_PRIVATE_TMPDIR are for files only accessible by root (mode 700).
 * We must lock everyone but root out of ZONES_PRIVATE_TMPDIR since anyone
 * can open any UNIX domain socket, regardless of its file system permissions.
 *
 * The mode of ZONES_TMPDIR and CONTAINER_STATE_DIR is the default zones
 * directory mode (755).
 * The mode of ZONE_PRIVATE_TMPDIR is private only to root.
 *
 * The zonepath directory should have the private mode (700). The directories
 * above zonepath should be 755 (default zones directory mode) since
 * this could also be ZONES_TMPDIR in the case of kernel zones and must be
 * accessible for ZOSS roots by non-root user ids.
 */
#define	ZONES_TMPDIR			_PATH_SYSVOL "/zones"
#define	CONTAINER_STATE_DIR		_PATH_SYSVOL "/zones/state"
#define	ZONES_PRIVATE_TMPDIR		_PATH_SYSVOL "/zones/_private"

#define	ZONES_DEFAULT_MODE		(S_IRWXU | S_IRGRP | S_IROTH | \
    S_IXGRP | S_IXOTH)
#define	ZONES_PRIVATE_MODE		(S_IRWXU)

/*
 * The path to the door used by clients to communicate with zoneadmd.
 */
#define	ZONE_DOOR_PATH		ZONES_TMPDIR "/%s.zoneadmd_door"

/*
 * Zones Delegated Restarter door path, always in the global zone.
 */
#define	ZDR_DOOR_PATH		_PATH_SYSVOL "/zdr_door"

/*
 * Zones restarter door call commands.
 */
typedef enum {
	ZDA_CMD_ZONEADMD,		/* Starts zoneadmd(8) */
	ZDA_CMD_ZONEADMD_READY,		/* zoneadmd(8) is ready for calls */
	ZDA_CMD_ZONE_OP_STATE_SYNC,	/* Zones OP events synchronization */
	ZDA_CMD_ZONE_IMPORT,		/* Import zone and clear .xml.  */
	ZDA_CMD_ZONE_REMOVE,		/* Remove a zone. */
	ZDA_CMD_GET_ZSTATE,		/* Zone status from the index file */
	ZDA_CMD_INVALID			/* Validation flag */
} zdr_door_cmd_t;

/*
 * The structure to request the Zones Delegated Restarter to
 * start zoneadmd(8) and also for the kernel upcall to the restarter.
 */
typedef struct zdr_door_args {
	zdr_door_cmd_t	zda_cmd;
	zone_cmd_t	zda_zone_cmd;
	char		zda_zonename[ZONENAME_MAX];
	boolean_t	zda_debug;
} zdr_door_args_t;

typedef struct zdr_door_ret {
	zone_op_state_t	zdrr_zops;
	int		zdrr_err;
	int		zdrr_errno;
} zdr_door_ret_t;

/* zone_flags */

/*
 * The following flags are set when the zone is created and are never modified.
 * Threads that test for these flags don't have to hold zone_lock.
 */
typedef enum zone_flags {
	ZF_HASHED_LABEL		= 0x2,	/* zone has a unique label */
	ZF_IS_SCRATCH		= 0x4,	/* scratch zone */
	ZF_NET_EXCL		= 0x8,	/* Zone has an exclusive IP stack */
	ZF_LOCKDOWN		= 0x10,	/* mwac is enabled for this zone */
	ZF_LOCKDOWN_TEMP	= 0x20,	/* mwac can be disabled */
	ZF_TRANSIENT_RW		= 0x40,	/* Read-only zone transient r/w */
	ZF_FORCED_RW		= 0x80,	/* Read-only zone forced r/w */
	ZF_SHARED_LABEL		= 0x100, /* zone has a  non-unique label */
	ZF_GLOBAL_TIME		= 0x200, /* zone can set global/system time */
	ZF_INIT_TRUSTED		= 0x400, /* Init is started in the TP */
	ZF_TRUSTED_PATH_UNSAFE	= 0x800 /* TPD_UNSAFE should be set in init */
} zone_flags_t;

/* zone_create flags */
typedef enum zone_create_flags {
	ZCF_NET_EXCL		= 0x1,	/* Create a zone with exclusive IP */
	ZCF_SHARED_LABEL	= 0x2,	/* Create a zone with a shared label */
	/* Create a zone which can set global/system time */
	ZCF_GLOBAL_TIME		= 0x4,
	ZCF_INIT_TRUSTED	= 0x8,	/* Init is run in the Trusted Path. */
	ZCF_TRUSTED_PATH_UNSAFE	= 0x10	/* Trusted Path is Unsafe. */
} zone_create_flags_t;

/* zone network address property types (zn_type) */
typedef enum zone_addrtype {
	ZONE_NETWORK_ADDRESS = 0,
	ZONE_NETWORK_DEFROUTER,
	ZONE_NETWORK_ALLOWED_MACADDRESS,
	ZONE_NETWORK_ALLOWED_VLANID
} zone_addrtype_t;

#define	ZONE_NET_ADDRNAME		"address"
#define	ZONE_NET_RTRNAME		"route"
#define	ZONE_NET_ALLOWED_MACADDRNAME	"allowed-mac-address"
#define	ZONE_NET_ALLOWED_VLANIDNAME	"allowed-vlan-id"

typedef struct zone_net_addr {
	sa_family_t	zna_family;
	uint16_t	zna_plen;
	union {
		struct in6_addr znau_addr6;
		struct in_addr znau_addr4;
	} zna_addru;
} zone_net_addr_t;
#define	zna_addr6	zna_addru.znau_addr6
#define	zna_addr4	zna_addru.znau_addr4

/*
 * accounts for IPoIB VNICs and we are not using MAXMACADDRLEN since
 * if it changes, then we will break the ABI.
 */
#define	ZONE_MAXMACADDRLEN	20

typedef struct zone_net_macaddr {
	uint8_t znm_addr[ZONE_MAXMACADDRLEN];
	uint8_t znm_len;
} zone_net_macaddr_t;

/* max nelems of zone_net_addr_t */
#define	ZONE_NET_ADDR_MAX	64

/* max nelems of zone_net_macaddr_t */
#define	ZONE_NET_MACADDR_MAX	64

/* max nelems of uint8_t required to represent VLAN_ID_MAX VLAN IDs in bits */
#define	ZONE_NET_VID_BITARR_MAX	512

/*
 * Note that this is an ABI for kernel zone guest/host, and cannot change
 * incompatibly. `zn_type' can be extended to retrieve new information
 * in the future.
 */
typedef struct zone_net_data {
	zone_addrtype_t	zn_type;
	datalink_id_t	zn_linkid;
	/*
	 * zn_nelems below captures the number of allowed MAC/IP addresses or
	 * ZONE_NET_VID_BITARR_MAX number of uint8_t to hold allowed VLAN IDs
	 * in a bit array.
	 */
	uint_t		zn_nelems;
	zone_net_addr_t zn_addrs[1];
	/*
	 * based on zn_type either
	 *   (zn_nelems - 1) of zone_net_addr_t follows or
	 *   (zn_nelems) of zone_net_macaddr_t or uint8_t follows
	 */
} zone_net_data_t;

typedef struct zone_net_id {
	datalink_id_t	zni_linkid;
	uint32_t	zni_id;
} zone_net_id_t;

typedef struct zone_configure_net {
	datalink_id_t	zcn_linkid;
	boolean_t	zcn_configure;
} zone_configure_net_t;

typedef struct zone_dl_auto {
	datalink_id_t	zda_linkid;
	boolean_t	zda_auto_link;
} zone_dl_auto_t;

/*
 * Device annotations.
 */
#define	ZONE_ALLOW_PARTITION	"allow-partition"
#define	ZONE_ALLOW_RAW_IO	"allow-raw-io"
#define	ZONE_ALLOW_MHD		"allow-mhd"
#define	ZONE_BOOTPRI		"bootpri"
#define	ZONE_DEV_ID		"id"
#define	ZONE_DIO_DEV		"directio-device"
#define	ZONE_DIO_TYPE		"directio-type"
#define	ZONE_GZ_ONLINE		"gz_online"
#define	ZONE_MATCH_DEV		"match"
#define	ZONE_IOV_STATE		"iov-state"
#define	ZONE_SURI_DEV		"suri-device"
#define	ZONE_WHOLEDISK		"whole-disk"
#define	ZONE_MEDIA_TYPE		"media"

/* Media type for anet configured for a zone. */
#define	ZMT_IB			0

typedef enum {
	ZIS_SUSPEND,
	ZIS_HA_FAILOVER
} zone_iov_state_t;

/*
 * Type of direct IO device that is configured for a zone. This type will be
 * passed from host to guest, which, combined with ZONE_DEV_ID, is used to
 * identify a direct IO device. Thus, it becomes an interface b/w guest and
 * host. In order to keep old guest compatible with new host, don't remove
 * entry, but only grow this list.
 */
typedef enum dio_dev_type {
	DIO_T_DEVICE = 0,
	DIO_T_ANET,
	DIO_T_IB_VHCA,
	DIO_T_MAX
} dio_dev_type_t;

typedef enum {
	ZDT_DISK = 0,
	ZDT_NET,
	ZDT_VPCI
} zone_dev_type_t;

#ifdef _SYSCALL32
typedef struct zone_devann32 {
	dev32_t zda_dev;
	caddr32_t zda_buf;
	size32_t zda_bufsize;
} zone_devann32_t;
#endif

typedef struct zone_devann {
	dev_t zda_dev;
	caddr_t zda_buf;
	size_t zda_bufsize;
} zone_devann_t;

/*
 * For dataset aliasing.  See comments above zone_dataset_t.
 */
typedef enum zone_ds_cfg {
	ZONE_DS_CFG_INVALID = 0,
	ZONE_DS_CFG_PLATFORM,		/* Configured via platform.xml */
	ZONE_DS_CFG_DELEGATED,		/* Configured via zonecfg */
	ZONE_DS_CFG_MAX
} zone_ds_cfg_t;
#define	ZONE_DS_CFG_VALID(x) \
	(((x) > ZONE_DS_CFG_INVALID) && ((x) < ZONE_DS_CFG_MAX))

/*
 * in-kernel state of the virtual zpool, used for dataset and
 * zpool resources.
 */
typedef enum zone_ds_state {
	ZONE_DS_STATE_INVALID = 0,
	ZONE_DS_STATE_IMPORTED,
	ZONE_DS_STATE_EXPORTED,
	ZONE_DS_STATE_MAX
} zone_ds_state_t;
#define	ZONE_DS_STATE_VALID(x) \
	(((x) > ZONE_DS_STATE_INVALID) && ((x) < ZONE_DS_STATE_MAX))

#define	ZONE_DS_ALIAS_STR	"alias"
#define	ZONE_DS_CFGSRC_STR	"cfgsrc"
#define	ZONE_DS_STATE_STR	"state"
#define	ZONE_DS_ZPOOL_STR	"zpool"

typedef struct zonehrtoffset {
	volatile memtime_sec_t	zto_offset;
} zonehrtoffset_t;

typedef struct zonetimeoffset {
	zonehrtoffset_t			*zonehrtoffset;
	struct anon_map 		*zto_offset_amp;
	/*
	 * zto_lock, zto_cv, zto_io_pending is only used for setting the
	 * zone offset, it is used to make write in memory and write to
	 * disk synchronous.
	 */
	kmutex_t			zto_lock;
	kcondvar_t			zto_cv;
	boolean_t			zto_io_pending;
} zonetimeoffset_t;

/*
 * Migration classes supported for a kernel zone.
 * Some of them are generic classes and can be configured in cpu-arch
 * zone config property. New classes should be added to support new CPUs.
 * The new class should be added in the end before ZMIG_CLASS_MAX and
 * don't change the order of previously defined classes.
 */
#ifdef	__sparc
typedef	enum zone_mig_class {
	/* ZMIG_CLASS_UNKNOWN is used to indicate an invalid mig class. */
	ZMIG_CLASS_UNKNOWN = 0,
	/* ZMIG_CLASS_NATIVE is mapped to real class during booting. */
	ZMIG_CLASS_NATIVE,
	/* Real mig classes. */
	ZMIG_CLASS_GENERIC,
	ZMIG_CLASS_MIG_CLS1,
	ZMIG_CLASS_SPARC64_CLS1,
	ZMIG_CLASS_SPARC64PLUS,
	ZMIG_CLASS_T4,
	ZMIG_CLASS_T5,
	ZMIG_CLASS_M5,
	ZMIG_CLASS_M6,
	ZMIG_CLASS_M7, /* For T7 and M7. */
	ZMIG_CLASS_M8, /* For T8 and M8. */
	ZMIG_CLASS_SPARC64PLUS2,
	ZMIG_CLASS_SPARC64,
	ZMIG_CLASS_MAX
} zone_mig_class_t;

#define	MIG_CLS_VALID(cls)	\
	(((cls) >= ZMIG_CLASS_GENERIC) && ((cls) < ZMIG_CLASS_MAX))

#else
typedef	enum zone_mig_class {
	/* ZMIG_CLASS_UNKNOWN is used to indicate an invalid mig class. */
	ZMIG_CLASS_UNKNOWN = 0,
	/* ZMIG_CLASS_NATIVE is mapped to real class during booting. */
	ZMIG_CLASS_NATIVE,
	/* Real mig classes. */
	ZMIG_CLASS_NEHALEM,
	ZMIG_CLASS_NEHALEM_NATIVE,
	ZMIG_CLASS_WESTMERE,
	ZMIG_CLASS_WESTMERE_NATIVE,
	ZMIG_CLASS_SANDYBRIDGE,
	ZMIG_CLASS_SANDYBRIDGE_NATIVE,
	ZMIG_CLASS_IVYBRIDGE,
	ZMIG_CLASS_IVYBRIDGE_NATIVE,
	ZMIG_CLASS_HASWELL,
	ZMIG_CLASS_HASWELL_NATIVE,
	ZMIG_CLASS_BROADWELL,
	ZMIG_CLASS_BROADWELL_NATIVE,
	ZMIG_CLASS_SKYLAKE,
	ZMIG_CLASS_SKYLAKE_NATIVE,
	ZMIG_CLASS_MAX
} zone_mig_class_t;

/*
 * ZMIG_CLASS_NATIVE is considered as a valid class and will be used as a
 * fallback when we failed to map it to a real mig class.
 */
#define	MIG_CLS_VALID(cls)	\
	(((cls) >= ZMIG_CLASS_NATIVE) && ((cls) < ZMIG_CLASS_MAX))

#endif	/* __sparc */

/* Good enough for 64 features. */
typedef	uint64_t zone_host_feature_t;

/*
 * Page size policy we can configure for a kernel zone under capped-memory in
 * pagesize-policy property.
 */
typedef	enum zone_pgsz_policy {
	PGSZ_POLICY_UNKNOWN = 0,
	PGSZ_POLICY_COMPATIBLE,
	PGSZ_POLICY_SMALLEST_ONLY,
	PGSZ_POLICY_LARGEST_AVAILABLE,
	PGSZ_POLICY_LARGEST_ONLY
} zone_pgsz_policy_t;

#ifdef _KERNEL

#include <sys/random_impl.h>
#include <sys/taskq.h>

/*
 * We need to protect the definition of 'list_t' from userland applications and
 * libraries which may be defining ther own versions.
 */
#include <sys/list.h>
#include <sys/credit.h>

#define	ZONE_MAX_SWAP_CHUNK	(64)

/*
 * uniqid of the Global Zone.
 *
 * We initialize its value to be higher than the highest valid zoneid
 * to keep the uniqid space from overlapping with the zoneid_t space
 */
#define	GLOBAL_ZONEUNIQID	(MAX_ZONEID + 1)

/* Invalid zone ID */
#define	INVALID_ZONEID		(MAX_ZONEID + 2)

/*
 * A zone (or pool) may not have a specific scheduling class set up. In that
 * case ZONE_SCHED_CLASS_UNSET is used.
 */
#define	ZONE_SCHED_CLASS_UNSET	((id_t)(-1))

struct pool;
struct brand;

/*
 * Each of these constants identifies a kernel subsystem that acquires and
 * releases zone references.  Each subsystem that invokes
 * zone_hold_ref() and zone_rele_ref() should specify the
 * zone_ref_subsys_t constant associated with the subsystem.  Tracked holds
 * help users and developers quickly identify subsystems that stall zone
 * shutdowns indefinitely.
 *
 * NOTE: You must modify zone_ref_subsys_names in usr/src/uts/common/os/zone.c
 * when you modify this enumeration.
 */
typedef enum zone_ref_subsys {
	ZONE_REF_NFS,			/* NFS client */
	ZONE_REF_NFSV4,			/* NFSv4 client */
	ZONE_REF_SMBFS,			/* SMBFS */
	ZONE_REF_LOFI,			/* LOFI devices */
	ZONE_REF_VFS,			/* VFS infrastructure */
	ZONE_REF_IPC,			/* IPC infrastructure */
	ZONE_REF_NFSSRV,		/* NFS server */
	ZONE_REF_SHAREFS,		/* SHAREFS */
	ZONE_REF_NUM_SUBSYS		/* This must be the last entry. */
} zone_ref_subsys_t;

/*
 * zone_ref represents a general-purpose references to a zone.  Each zone's
 * references are linked into the zone's zone_t::zone_ref_list.  This allows
 * debuggers to walk zones' references.
 */
typedef struct zone_ref {
	struct zone	*zref_zone; /* the zone to which the reference refers */
	list_node_t	zref_linkage; /* linkage for zone_t::zone_ref_list */
} zone_ref_t;

/*
 * Dataset aliasing and read only zone roots make use of zone_dataset_t
 * list nodes.  Each list node corresponds to an nvlist item passed
 * into the kernel via zone_create().  That nvlist looks like:
 *
 *  <dataset name> => {
 *	alias => <alias>
 *	cfgsrc => <ZONE_DS_CFG_PLATFORM|ZONE_DS_CFG_DELEGATED>
 *	zpool => <B_TRUE|B_FALSE>
 *  }
 *
 * For example, a zone that has the platform dataset in the 'zones' pool and
 * another delegated dataset in the 'tank' pool would look like:
 *
 * 'zones/zone1/rpool'
 * => { alias => 'rpool'; cfgsrc => ZONE_DS_CFG_PLATFORM ; zpool => 0 }
 * 'tank/zone1'
 * => { alias => 'tank'; cfgsrc => ZONE_DS_CFG_DELEGATED ; zpool => 0 }
 *
 * For example, a delegated zpool resource 'tank' will have its top-level
 * zpool dataset delegated like:
 *
 * 'tank'
 * => { alias => 'tank'; cfgsrc => ZONE_DS_CFG_DELEGATED ; zpool => 1 }
 *
 * All of the delegated datasets are subject to aliasing.  See
 * zone_dataset_alias() and zone_dataset_unalias().  Only those configured as
 * platform datasets are subject to mandatory write access controls (MWAC).
 * See zone_dataset_visible().
 *
 * To prevent a pool containing a delegated dataset from being exported,
 * a hold is taken on each zpool using the ZFS_IOC_POOL_HOLD ioctl.  As the
 * delegation is removed (e.g. during zone shutdown) the hold is released
 * using the ZFS_IOC_POOL_RELE ioctl.  If ZFS reference tracking is enabled,
 * the reference counting code will store the address of a zone_dataset_t
 * instance.  The zd_zone field exists to make it easier to figure out which
 * zone holds the reference.
 *
 * A virtual zpool (delegated dataset) has two kernel states to support LZR:
 *
 * ZONE_DS_STATE_IMPORTED
 * The delegated dataset is visible to the NGZ.
 * Any delegated dataset including platform datasets spliced into the NGZ
 * during NGZ boot is automatically in this state.
 * A virtual zpool in this state may be exported by the NGZ except for
 * NGZs platform datasets, aka. the virtual 'rpool'.
 * After a virtual zpool is imported by the NGZ its kernel state switches
 * to ZONE_DS_STATE_IMPORTED.
 *
 * ZONE_DS_STATE_EXPORTED
 * The delegated dataset is not visible to the NGZ.
 * The delegated dataset can be removed from the running NGZ via LZR.
 * Any delegated dataset added to the NGZ via LZR will be in this state
 * and can be observed as importable virtual zpool and imported in the NGZ.
 * A virtual zpool that is exported by the NGZ switches its kernel state
 * to ZONE_DS_STATE_EXPORTED.
 *
 * A delegated dataset in state ZONE_DS_STATE_EXPORTED cannot be observed
 * anymore by zone_dataset_visible(). Also one cannot unalias a dataset
 * anymore with zone_dataset_unalias() once it has been removed from the
 * NGZs namespace. However we still allow to alias a delegated dataset
 * back with zone_dataset_alias() as long as its still around in the
 * zones dataset list. Once it has been removed entirely by LZR from
 * the zones dataset list, the alias will become ZONE_INVISIBLE_SOURCE.
 */
typedef struct zone_dataset {
	char		*zd_dataset;	/* Real dataset name */
	char		*zd_alias;	/* Aliased zpool name */
	zone_ds_cfg_t	zd_cfgsrc;	/* How was it configured */
	struct zone	*zd_zone;	/* The zone to which this belongs */
	zone_ds_state_t	zd_state;	/* virtual zpool state */
	boolean_t	zd_zpool;	/* zpool resource */
	boolean_t	zd_delete_pend;	/* dataset removal pending */
	boolean_t	zd_export_pend;	/* virt. zpool export pending */
	list_node_t	zd_linkage;
} zone_dataset_t;

/*
 * structure for zone kstats
 */
typedef struct zone_kstat {
	kstat_named_t zk_zonename;
	kstat_named_t zk_usage;
	kstat_named_t zk_value;
} zone_kstat_t;

/*
 * A structure for collecting zone CPU stats data. It contains a subset of the
 * fields in cpu_sys_stats_t and is sized as a power of 2 so we can quickly
 * index zone->zone_cpu_stats[] in the syscall path. Future additions must
 * ensure it remains that way.
 *
 * While CMS_IDLE microstate counter in msacct array is not used to update the
 * zone kstat (it does not make sense to attribute the idle time to a specific
 * zone), to have it makes code simpler in new_cpu_mstate().
 */
typedef struct zone_cpu_stats {
	volatile hrtime_t msacct[NCMSTATES]; /* CPU microstates accounting */
	uint64_t zcs_syscall;		/* system calls */
	uint64_t zcs_sysread;		/* read() + readv() system calls */
	uint64_t zcs_syswrite;		/* write() + writev() system calls */
	uint64_t zcs_sysfork;		/* forks */
	uint64_t zcs_sysvfork;		/* vforks */
	uint64_t zcs_sysexec;		/* execs */
	uint64_t zcs_sysspawn;		/* spawns */
	uint64_t zcs_readch;		/* bytes read by rdwr() */
	uint64_t zcs_writech;		/* bytes written by rdwr() */
	uint64_t zcs_msg;		/* msg count (msgrcv() + msgsnd()) */
	uint64_t zcs_sema;		/* semaphore ops count (semop()) */
	uint64_t zcs_namei;		/* pathname lookups */
	uint64_t zcs_procovf;		/* failed forks */
	uint64_t _zcs_pad[15];		/* padding to the next power-of-two */
} zone_cpu_stats_t;

struct cpucap;

typedef struct zone {
	/*
	 * zone_name is never modified once set.
	 */
	char		*zone_name;	/* zone's configuration name */
	/*
	 * zone_nodename and zone_domain are never freed once allocated.
	 */
	char		*zone_nodename;	/* utsname.nodename equivalent */
	char		*zone_domain;	/* srpc_domain equivalent */
	/*
	 * zone_hostid is used for per-zone hostid emulation.
	 * Currently it isn't modified after it's set (so no locks protect
	 * accesses), but that might have to change when we allow
	 * administrators to change running zones' properties.
	 *
	 * The global zone's zone_hostid must always be HW_INVALID_HOSTID so
	 * that zone_get_hostid() will function correctly.
	 */
	uint32_t	zone_hostid;	/* zone's hostid, HW_INVALID_HOSTID */
					/* if not emulated */
	/*
	 * zone_lock protects the following fields of a zone_t:
	 * 	zone_ref
	 * 	zone_cred_ref
	 *	zone_status
	 * 	zone_subsys_ref
	 * 	zone_ref_list
	 * 	zone_ntasks
	 *	zone_kthreads
	 * 	zone_flags
	 * 	zone_zsd
	 *	zone_pfexecd
	 *	zone_mwac_white_list
	 *	zone_mwac_black_list
	 *	zone_brand_name
	 *	zone_brand
	 *	zone_dl_list
	 *	zone_auxstate
	 *	zone_idmap
	 *	zone_directio_dev_map
	 */
	kmutex_t	zone_lock;
	/*
	 * zone_linkage is the zone's linkage into the active or
	 * death-row list.  The field is protected by zonehash_lock.
	 */
	list_node_t	zone_linkage;
	zoneid_t	zone_id;	/* ID of zone */
	uint_t		zone_ref;	/* count of zone_hold()s on zone */
	uint_t		zone_cred_ref;	/* count of zone_hold_cred()s on zone */
	/*
	 * Fixed-sized array of subsystem-specific reference counts
	 * The sum of all of the counts must be less than or equal to zone_ref.
	 * The array is indexed by the counts' subsystems' zone_ref_subsys_t
	 * constants.
	 */
	uint_t		zone_subsys_ref[ZONE_REF_NUM_SUBSYS];
	list_t		zone_ref_list;	/* list of zone_ref_t structs */
	/*
	 * zone_rootvp and zone_rootpath can never be modified once set.
	 */
	struct vnode	*zone_rootvp;	/* zone's root vnode */
	char		*zone_rootpath;	/* Path to zone's root + '/' */
	ushort_t	zone_flags;	/* misc flags */
	zone_status_t	zone_status;	/* ZONE_IS_{READY|RUNNING|etc} */
	nvlist_t	*zone_auxstate;	/* auxiliary state info */
	size_t		zone_auxstate_len; /* len of aux state in string form */
	uint_t		zone_ntasks;	/* number of tasks executing in zone */
	kmutex_t	zone_nlwps_lock; /* protects zone_nlwps, and *_nlwps */
					/* counters in projects and tasks */
					/* that are within the zone */
	rctl_qty_t	zone_nlwps;	/* number of lwps in zone */
	rctl_qty_t	zone_nlwps_ctl; /* protected by zone_rctls->rcs_lock */
	rctl_qty_t	zone_shmmax;	/* System V shared memory usage */
	ipc_rqty_t	zone_ipc;	/* System V IPC id resource usage */

	uint_t		zone_rootpathlen; /* strlen(zone_rootpath) + 1 */
	uint32_t	zone_shares;	/* FSS shares allocated to zone */
	rctl_set_t	*zone_rctls;	/* zone-wide (zone.*) rctls */
	kmutex_t	zone_mem_lock;	/* protects zone_locked_mem and */
					/* kpd_locked_mem for all */
					/* projects in zone. */
					/* Also protects zone_max_swap */
					/* grab after p_lock, before rcs_lock */
	rctl_qty_t	zone_locked_mem;	/* bytes of locked memory in */
						/* zone */
	/* True if locked mem is set to a non-default value. */
	boolean_t	zone_locked_mem_ctl_set;
	rctl_qty_t	zone_locked_mem_ctl;	/* Current locked memory */
						/* limit.  Protected by */
						/* zone_rctls->rcs_lock */
	kmutex_t	zone_adi_metadata_lock;	/* protects zone_adi_metadata */
						/* kpd_adi_metadata for all */
						/* projects in zone. */
	rctl_qty_t	zone_adi_metadata;	/* bytes of adi metadata */
	rctl_qty_t	zone_adi_metadata_ctl;	/* named adi memory limit, */
						/* zone_rctls->rcs_lock */
	credit_t	*zone_max_swap_credit;	/* credit backer for swap */
	uint64_t	zone_max_swap_failures; /* # times hit limit */
	rctl_qty_t	zone_max_swap; /* bytes of swap reserved by zone */
	/* True if max swap if set to a non-default value. */
	boolean_t	zone_max_swap_ctl_set;
	rctl_qty_t	zone_max_swap_ctl;	/* current swap limit. */
						/* Protected by */
						/* zone_rctls->rcs_lock */
	kmutex_t	zone_rctl_lock;	/* protects zone_max_lofi */
	rctl_qty_t	zone_max_lofi; /* lofi devs for zone */
	rctl_qty_t	zone_max_lofi_ctl;	/* current lofi limit. */
						/* Protected by */
						/* zone_rctls->rcs_lock */
	list_t		zone_zsd;	/* list of Zone-Specific Data values */
	kcondvar_t	zone_cv;	/* used to signal state changes */
	struct proc	*zone_zsched;	/* Dummy kernel "zsched" process */
	pid_t		zone_proc_initpid; /* pid of "init" for this zone */
	char		*zone_initname;	/* fs path to 'init' */
	int		zone_boot_err;  /* for zone_boot() if boot fails */
	char		*zone_bootargs;	/* arguments passed via zone_boot() */
	char		*zone_macprofile; /* mac-profile */
	uint64_t	zone_phys_mcap;	/* physical memory cap */
	zone_pgsz_policy_t	zone_pgsz_policy;	/* kernel zones only */
	key_t		zone_mrp_key;	/* kernel zones only */
	char		*zone_mrp_svc;	/* kernel zones only */
	kthread_t	*zone_kthreads;	/* kernel threads in zone */
	struct priv_set	*zone_privset;	/* limit set for zone */
	/*
	 * zone_vfslist is protected by vfs_list_lock().
	 */
	struct vfs	*zone_vfslist;	/* list of FS's mounted in zone */
	uint64_t	zone_uniqid;	/* unique zone generation number */
	struct cred	*zone_kcred;	/* kcred-like, zone-limited cred */
	/*
	 * zone_pool is protected by pool_lock().
	 */
	struct pool	*zone_pool;	/* pool the zone is bound to */
	hrtime_t	zone_pool_mod;	/* last pool bind modification time */
	/* zone_psetid is protected by cpu_lock */
	psetid_t	zone_psetid;	/* pset the zone is bound to */
	/*
	 * The following two can be read from the context of the zone
	 * (curproc->p_zone) without holding any locks.  They are updated
	 * under cpu_lock.
	 */
	int		zone_ncpus;	   /* zone's idea of ncpus */
	int		zone_ncpus_online; /* zone's idea of ncpus_online */
	/*
	 * List of ZFS datasets exported to this zone.
	 * The R/W lock zone_datasets_lock protects access to the list.
	 */
	krwlock_t	zone_datasets_lock;
	list_t		zone_datasets;		/* list of datasets */
	uint64_t	zone_spa_config_generation; /* virt. zpool config gen */

	ts_label_t	*zone_slabel;	/* zone sensitivity label */
	int		zone_match;	/* require label match for packets */
	tsol_mlp_list_t zone_mlps;	/* MLPs on zone-private addresses */
	/* how to recover from an init failure */
	zone_init_t	zone_init_recover;
	char		*zone_brand_name;	/* zone brand name */
	struct brand	*zone_brand;		/* zone brand module ptr */
	void		*zone_brand_data;	/* store brand specific data */
	id_t		zone_defaultcid;	/* dflt scheduling class id */
	kstat_t		*zone_swapresv_kstat;
	kstat_t		*zone_lockedmem_kstat;
	kstat_t		*zone_adi_metadata_kstat;
	list_t		zone_dl_list;
	netstack_t	*zone_netstack;
	struct cpucap	*zone_cpucap;	/* CPU caps data */
	/*
	 * Solaris Auditing per-zone audit context
	 */
	struct au_kcontext	*zone_audit_kctxt;
	/*
	 * For private use by mntfs.
	 */
	struct mntelem	*zone_mntfs_db;
	krwlock_t	zone_mntfs_db_lock;

	struct klpd_reg		*zone_pfexecd;

	char		*zone_fs_allowed;
	char		*zone_vcpus;
	char		*zone_boot_device;
	struct avl_tree		*zone_mwac_white_list;
	struct avl_tree		*zone_mwac_black_list;
	rctl_qty_t	zone_nprocs;	/* number of processes in the zone */
	rctl_qty_t	zone_nprocs_ctl;	/* current limit protected by */
						/* zone_rctls->rcs_lock */
	kstat_t		*zone_nprocs_kstat;
	mod_hash_t	*zone_devann_hash;
	kmutex_t	zone_devann_lock;
	time_t		zone_destroy_time;  /* when the zone went defunct */
	secext_conf_t	zone_security;	/* security extensions for the zone */
	/* array of CPU stats indexed by cpu_seqid */
	zone_cpu_stats_t	*zone_cpu_stats;
	/* array of accumulated CPU stats indexed by cpu_seqid */
	zone_cpu_stats_t	*zone_cpu_stats_accum;
	/*
	 * Array of flags.  Not part of zone_cpu_stats_t due to its size
	 * constraints.
	 */
	boolean_t	*zone_cpu_stats_cleared;
	/*
	 * Array of cpu kstat ids indexed by cpu_seqid. -1 is stored when
	 * there's no kstat for the cpu_seqid. Creation and destruction of cpu
	 * kstats are done under cpu_lock, and thus modification of
	 * zone_cpu_kstat_kids is also protected by cpu_lock. So grabbing
	 * cpu_lock and comparing zone_cpu_kstat_kids[cpu_seqid] with -1 is a
	 * valid way to check that cpu kstat exists or not.
	 */
	kid_t		*zone_cpu_kstat_kids;
	/*
	 * Accumulated per zone kstat for data from CPU stats that are reset to
	 * 0 on crossing boundaries. Like zone_cpu_kstat_kids,
	 * creation/destruction of kstat and this field must be protected by
	 * cpu_lock.
	 */
	kstat2_t		*zone_accum_kstat;
	/* Memory kstats. They are initialized when the zone's brand is set. */
	kstat2_t		*zone_mem_usage_kstat;
	kstat2_t		*zone_mem_cap_kstat;
	/*
	 * Used for fixing mstate time to ignore time spent servicing
	 * interrupts.
	 */
	hrtime_t	**zone_cpu_intracct;
	/*
	 * Per-zone ID mapping data
	 */
	idmap_zone_specific_t	*zone_idmap;

	/* dev/devices mapping for direct-io devices */
	nvlist_t	*zone_directio_dev_map;

	char		*zone_cfg_bootargs;
	char		*zone_tran_bootargs;

	/*
	 * Per zone structure which consists of info about zone's time offset
	 * from the global zone, page sharing info, etc.
	 */
	zonetimeoffset_t	*zone_offset_time;

	/* verified-boot info */
	void		*zone_vboot;

	zone_mig_class_t zone_mig_class;

	zone_host_feature_t	zone_features;

	time_t			zone_boot_time;
	int		zone_core_flag;
	krwlock_t	zone_core_lock;

	vminfo_t	zone_vminfo;   /* zone's vminfo_t structure */
	pgcnt_t		zone_phys_free; /* zone's freemem from rcapd */
	/* Last update time of zone_phys_free */
	hrtime_t	zone_phys_update_time;
	taskq_t		*zone_system_taskq;	/* per-zone system taskq */
	list_node_t	zone_capped_linkage;
	char		*zone_hw_provider;
} zone_t;

#define	ZONE_HRTOFFSET	(curzone->zone_offset_time->zonehrtoffset)
#define	ZONE_OFFSETAMP	(curzone->zone_offset_time->zto_offset_amp)
#define	ZONE_OFFSET_LOCKPTR	(&curzone->zone_offset_time->zto_lock)
#define	ZONE_OFFSET_CVPTR	(&curzone->zone_offset_time->zto_cv)
#define	ZONE_SETS_GLOBAL_TIME	((curzone->zone_flags & ZF_GLOBAL_TIME))

#define	ZONE_INTR_ACCT(zone, cpu, mstate)	\
	((zone)->zone_cpu_intracct[(cpu)->cpu_seqid][(mstate)])

#define	ZONE_CURZ_INTR_ACCT(cpu, mstate)		\
	ZONE_INTR_ACCT(curzone, (cpu), (mstate))

extern zone_t zone0;
extern zone_t *global_zone;
extern uint_t maxzones;

extern long zone(int, void *, void *, void *, void *);
extern void zone_zsd_init(void);
extern void zone_init(void);
extern void zone_hold(zone_t *);
extern void zone_rele(zone_t *);
extern void zone_init_ref(zone_ref_t *);
extern void zone_hold_ref(zone_t *, zone_ref_t *, zone_ref_subsys_t);
extern void zone_rele_ref(zone_ref_t *, zone_ref_subsys_t);
extern void zone_cred_hold(zone_t *);
extern void zone_cred_rele(zone_t *);
extern void zone_task_hold(zone_t *);
extern void zone_task_rele(zone_t *);
extern zone_t *zone_find_all_by_id(zoneid_t);
extern zone_t *zone_find_by_id(zoneid_t);
extern zone_t *zone_find_by_label(const ts_label_t *);
extern zone_t *zone_find_by_name(const char *);
extern zone_t *zone_find_by_any_path(const char *, boolean_t);
extern zone_t *zone_find_by_path(const char *);
extern zoneid_t getzoneid(void);
extern int zone_datalink_walk(zoneid_t, int (*)(datalink_id_t, void *), void *);
extern int zone_check_datalink(zoneid_t *, datalink_id_t);
extern taskq_t *zone_get_systemtaskq(zone_t *const);

/*
 * Zone's vminfo update hook into clock()
 */
extern void (*zone_vminfo_callout)();

/*
 * Zone-specific data (ZSD) APIs
 */
/*
 * The following is what code should be initializing its zone_key_t to if it
 * calls zone_getspecific() without necessarily knowing that zone_key_create()
 * has been called on the key.
 */
#define	ZONE_KEY_UNINITIALIZED	0

typedef uint_t zone_key_t;

extern void	zone_key_create(zone_key_t *, void *(*)(zoneid_t),
    void (*)(zoneid_t, void *), void (*)(zoneid_t, void *));
extern int	zone_key_delete(zone_key_t);
extern void	*zone_getspecific(zone_key_t, zone_t *);
extern int	zone_setspecific(zone_key_t, zone_t *, const void *);
extern void	zone_zsd_create_sync(zone_t *const, zone_key_t);

/*
 * The definition of a zsd_entry is truly private to zone.c and is only
 * placed here so it can be shared with mdb.
 *
 * State maintained for each zone times each registered key, which tracks
 * the state of the create, shutdown and destroy callbacks.
 *
 * zsd_flags is used to keep track of pending actions to avoid holding locks
 * when calling the create/shutdown/destroy callbacks, since doing so
 * could lead to deadlocks.
 */
typedef enum zsd_flag {
	ZSD_CREATE_NEEDED	= 0x0001,
	ZSD_CREATE_INPROGRESS	= 0x0002,
	ZSD_CREATE_COMPLETED	= 0x0004,
	ZSD_SHUTDOWN_NEEDED	= 0x0010,
	ZSD_SHUTDOWN_INPROGRESS	= 0x0020,
	ZSD_SHUTDOWN_COMPLETED	= 0x0040,
	ZSD_DESTROY_NEEDED	= 0x0100,
	ZSD_DESTROY_INPROGRESS	= 0x0200,
	ZSD_DESTROY_COMPLETED	= 0x0400
} zsd_flag_t;

struct zsd_entry {
	zone_key_t		zsd_key;	/* Key used to lookup value */
	void			*zsd_data;	/* Caller-managed value */
	/*
	 * Callbacks to be executed when a zone is created, shutdown, and
	 * destroyed, respectively.
	 */
	void			*(*zsd_create)(zoneid_t);
	void			(*zsd_shutdown)(zoneid_t, void *);
	void			(*zsd_destroy)(zoneid_t, void *);
	list_node_t		zsd_linkage;
	zsd_flag_t		zsd_flags;
	kcondvar_t		zsd_cv;
};

#define	ZSD_SHUTDOWN_ALL	\
	(ZSD_SHUTDOWN_NEEDED|ZSD_SHUTDOWN_INPROGRESS|ZSD_SHUTDOWN_COMPLETED)
#define	ZSD_DESTROY_ALL	\
	(ZSD_DESTROY_NEEDED|ZSD_DESTROY_INPROGRESS|ZSD_DESTROY_COMPLETED)
#define	ZSD_ALL_INPROGRESS \
	(ZSD_CREATE_INPROGRESS|ZSD_SHUTDOWN_INPROGRESS|ZSD_DESTROY_INPROGRESS)

/*
 * Macros to help with zone visibility restrictions.
 */

/*
 * Is process in the global zone?  Or is the zone the global zone)?
 * NB: The userland zfs implementation has its own definitions of these.
 * See sys/zfs_context.h.
 */
#define	INGLOBALZONE(p) \
	((p)->p_zone == global_zone)
#define	ISGLOBALZONE(z) \
	((z) == global_zone)
/*
 * Can process view objects in given zone?
 */
#define	HASZONEACCESS(p, zoneid) \
	((p)->p_zone->zone_id == (zoneid) || INGLOBALZONE(p))

/*
 * Convenience macro to see if a resolved path is visible from within a
 * given zone. The path argument must have previously been resolved.
 *
 * zone_rootpath has a trailing '/', and zone_rootpathlen is
 * (strlen(zone_rootpath) + 1), so we compare the first
 * (zone_rootpathlen - 2) bytes, and then check if the next character is
 * either '\0' (ie, path is the rootpath),
 * or '/' (path is a subdirectory of the zone's rootpath).
 */
#define	ZONE_PATH_VISIBLE(path, zone)			\
	(strncmp((path), (zone)->zone_rootpath,		\
	    (zone)->zone_rootpathlen - 2) == 0 &&	\
	((path)[(zone)->zone_rootpathlen - 2] == '/' ||	\
	(path)[(zone)->zone_rootpathlen - 2] == '\0'))

extern char *zone_path_translate(char *path, zone_t *zone);

/*
 * Special processes visible in all zones.
 */
#define	ZONE_SPECIALPID(x)	 ((x) == 0 || (x) == 1)
/*
 * The zone is an immutable zone.
 */
#define	ZONE_IMMUTABLE(zone)	(((zone)->zone_flags & ZF_LOCKDOWN) != 0)

/*
 * Zone-safe version of thread_create() to be used when the caller wants to
 * create a kernel thread to run within the current zone's context.
 */
extern kthread_t *zthread_create(caddr_t, size_t, void (*)(), void *, size_t,
    pri_t, const char *);
extern void zthread_exit(void);

extern void zproc_enter(zone_t *);

/*
 * Functions for an external observer to register interest in a zone's status
 * change.  Observers will be woken up when the zone status equals the status
 * argument passed in (in the case of zone_status_timedwait, the function may
 * also return because of a timeout; zone_status_wait_sig may return early due
 * to a signal being delivered; zone_status_timedwait_sig may return for any of
 * the above reasons).
 *
 * Otherwise these behave identically to cv_timedwait(), cv_wait(), and
 * cv_wait_sig() respectively.
 */
extern clock_t zone_status_timedwait(zone_t *, clock_t, zone_status_t);
extern clock_t zone_status_timedwait_sig(zone_t *, clock_t, zone_status_t);
extern void zone_status_wait(zone_t *, zone_status_t);
extern int zone_status_wait_sig(zone_t *, zone_status_t);

/*
 * Get the status  of the zone (at the time it was called).  The state may
 * have progressed by the time it is returned.
 */
extern zone_status_t zone_status_get(zone_t *);

/*
 * Safely get the hostid of the specified zone (defaults to machine's hostid
 * if the specified zone doesn't emulate a hostid).  Passing NULL retrieves
 * the global zone's (i.e., physical system's) hostid.
 */
extern uint32_t zone_get_hostid(zone_t *);

/*
 * Get the "kcred" credentials corresponding to the given zone.
 */
extern struct cred *zone_get_kcred(zoneid_t);

/*
 * Get/set the pool the zone is currently bound to.
 */
extern struct pool *zone_pool_get(zone_t *);
extern void zone_pool_pset_set(zone_t *, struct pool *, psetid_t);

/*
 * Get the pset the zone is currently using.
 */
extern psetid_t zone_pset_get(zone_t *);
extern int zone_pset_visible(zone_t *, psetid_t);

/*
 * Get the number of cpus/online-cpus visible from the given zone.
 */
extern int zone_ncpus_get(zone_t *);
extern int zone_ncpus_online_get(zone_t *);

/*
 * Returns true if the named pool/dataset is visible in the given zone.
 */
extern int zone_dataset_visible(zone_t *, const char *, int *);

/*
 * Translates dataset names to/from aliases.
 */
extern int zone_dataset_alias(zone_t *, const char *, char *, size_t);
extern int zone_dataset_unalias(zone_t *, const char *, char *, size_t,
    boolean_t *);
extern int zone_dataset_unalias_force(zone_t *, const char *, char *, size_t,
    boolean_t *);

/*
 * Zone delegated dataset iterator function and callback data structure.
 */
typedef struct zone_ds_cb_data {
	void		*zdscb_data;	/* Opaque callback data */
	boolean_t	zdscb_imponly;	/* Return importable pools only */
} zone_ds_cb_data_t;

typedef void (*zone_ds_iter_f)(zone_dataset_t *, void *);
extern void zone_iter_datasets(zone_t *, zone_ds_iter_f, void *);

/*
 * zone version of kadmin()
 */
extern int zone_kadmin(int, int, const char *, cred_t *, uint32_t, boolean_t);
extern void zone_shutdown_global(void);

extern int zone_walk(int (*)(zone_t *, void *), void *);

extern int zone_getattr_common(zone_t *, int, void *, size_t, ssize_t *, int);

extern ulong_t zone_count_all_swap_failures(void);

/*
 * Used by the syscall and context switch paths to cause processes of a
 * shutting down zone to kill themselves.
 */
extern volatile unsigned int zone_shutdown_count;
extern void zone_proc_exit_on_shutdown(void);

/*
 * Device annotations.
 *
 * The devnames zone profile can include PROFILE_TYPE_ANNOTATE entries,
 * which consist of a string key/value pair.  On lookup, devnames
 * annotates the dev_t for the sdev_node by calling into
 * zone_devann_insert() for each annotation entry.  Annotations are
 * per-zone and per-dev_t.  Drivers wishing to check for annotations
 * call zone_devann_find().  When a minor node is removed, the DDI code
 * calls zone_devann_clear() to remove all references to the matching dev_t.
 */

extern int zone_devann_present(zone_t *, dev_t);
extern char *zone_devann_find(zone_t *, dev_t, const char *);
extern int zone_devann_insert(zone_t *, dev_t, const char *, const char *,
    boolean_t);
extern int zone_devann_remove(zone_t *, dev_t, const char *);
extern void zone_devann_clear_all(dev_t);
extern void zone_devann_clear(zone_t *, dev_t);

/*
 * Auxiliary state
 */
int zone_mark_aux_state(zone_t *, const char *);
int zone_unmark_aux_state(zone_t *, const char *);
int zone_get_aux_state(zone_t *, char *, size_t, size_t *);
boolean_t zone_has_aux_state(zone_t *, const char *);

/*
 * Zone CPU stat related functions.
 */
void zcs_add_cur_to_accum(zone_cpu_stats_t *, zone_cpu_stats_t *, hrtime_t *);

/*
 * Set non-global zone's virtual time
 */
extern int zone_set_virtual_time(memtime_sec_t);
extern zonetimeoffset_t *zone_offset_add(zone_t *, memtime_sec_t);

/*
 * Export/import virtual zpools in a NGZ.
 */
typedef enum zone_vzp_cmd {
	ZONE_VZP_INVALID = 0,
	ZONE_VZP_IMPORT,		/* Import a virtual zpool */
	ZONE_VZP_EXPORT_START,		/* Start virtual zpool export */
	ZONE_VZP_EXPORT_CANCEL,		/* Cancel virtual zpool export */
	ZONE_VZP_EXPORT_END,		/* Finish virtual zpool export */
	ZONE_VZP_MAX
} zone_vzp_cmd_t;
#define	ZONE_VZP_CMD_VALID(x) \
	(((x) > ZONE_VZP_INVALID) && ((x) < ZONE_VZP_MAX))

extern int zone_virt_zpool_manage(zoneid_t, zone_vzp_cmd_t,
    const char *);
extern uint64_t zone_virt_zpool_config_gen(zone_t *);

/*
 * Check if rcapd is updating appropriately.
 * zone_phys_free should be updated everytime rcapd runs.
 */
extern boolean_t zone_valid_rcapd_updates(hrtime_t);

#endif	/* _KERNEL */

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_ZONE_H */
