/*
 * Copyright (c) 2013, 2017, Oracle and/or its affiliates. All rights reserved.
 */

#ifndef	_SYS_ZVBLK_COMMON_H
#define	_SYS_ZVBLK_COMMON_H

#include <sys/dkio.h>

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * o There are two block descriptor types
 *	o first
 *	o continuation
 *
 * o For the sake of simplicity, we require that:
 *	+ A descriptor does not cross page boundary and is PAGESIZE aligned.
 *	+ Vector array cannot wrap to a continuation descriptor.
 *	+ GPA array of a vector can wrap to next continuation descriptor.
 *	+ We do not support chained I/O request. Next Desc PA in a descriptor
 *	  must point to a continuation descriptor or be zero.
 *	+ We do not support holes b/w GPA arrays that belong to two consecutive
 *	  vectors.
 *
 * o Bit Numbering Convention
 *	o |63   56|55   48|47   40|39   32|31   24|23   16|15   08|07   00|
 *
 *
 * FIRST DESCRIPTOR
 * ----------------
 *
 * +------+-------+----------------+---------------------------------------+
 * |      |  BIT  |                |                                       |
 * |OFFSET| RANGE | NAME           | DESCRIPTION                           |
 * +------+-------+----------------+---------------------------------------+
 * | 0x00 | 07:00 | Version        | 0 for the current version             |
 * | 0x01 | 07:00 | Desc Type      | 0 - First descriptor                  |
 * | 0x02 | 15:00 | Reserved       | Always 0                              |
 * | 0x04 | 31:00 | Reserved       | Always 0                              |
 * | 0x08 | 63:00 | Next Desc PA   | 0 - No further descriptors            |
 * |      |       |                | > 0 - Guest PA of next descriptor.    |
 * |      |       |                |       Must be aligned to 8 byte       |
 * |      |       |                |       boundary.                       |
 * +------+-------+----------------+---------------------------------------+
 * | 0x10 | 63:00 | I/O Request ID | Identifier to echo back to guest      |
 * |      |       |                | Used two bytes for now                |
 * +------+-------+----------------+---------------------------------------+
 * | 0x18 | 07:00 | Operation      | 0 - Invalid setting                   |
 * |      |       |                | 1 - Read                              |
 * |      |       |                | 2 - Write                             |
 * +------+-------+----------------+---------------------------------------+
 * | 0x19 | 07:00 | Desc Count     | Total number of linked descriptors    |
 * +------+-------+----------------+---------------------------------------+
 * | 0x1a | 15:00 | Vector Count   | Number of logically contiguous buffers|
 * |      |       |                | An array of pointers to the vectors   |
 * |      |       |                | follows. A vector is an array of      |
 * |      |       |                | guest page physical addresses.        |
 * +------+-------+----------------+---------------------------------------+
 * | 0x1c | 31:00 | Page Count     | Total number of pages in all vectors  |
 * +------+-------+----------------+---------------------------------------+
 * | 0x20 | 63:00 | Block Number   | Disk block number for operation       |
 * +------+-------+----------------+---------------------------------------+
 * | 0x28 | 63:00 | Size           | Total size of operation               |
 * +------+-------+----------------+---------------------------------------+
 * | 0x30 | 07:00 | Operation      | 0 - Invalid setting                   |
 * |      |       | Result         | 1 - OK                                |
 * |      |       |                | 2 - Error                             |
 * | 0x31 | 07:00 | Reserved       | Always 0                              |
 * | 0x32 | 15:00 | Reserved       | Always 0                              |
 * | 0x34 | 31:00 | Reserved       | Always 0                              |
 * +------+-------+----------------+---------------------------------------+
 * | 0x38 | 63:00 | Reserved       | Always 0                              |
 * +------+-------+----------------+---------------------------------------+
 * | 0x40 | 07:00 | V1 Desc Index  | Descriptor index (0..n-1) where the   |
 * |      |       |                | array of guest page physical addresses|
 * |      |       |                | starts.                               |
 * | 0x41 | 07:00 | Reserved       | Always 0                              |
 * | 0x42 | 15:00 | V1 Desc Offset | Offset into the descriptor where the  |
 * |      |       |                | array of guest page physical addresses|
 * |      |       |                | starts.                               |
 * | 0x44 | 31:00 | V1 Size        | Size of the vector. This along with   |
 * |      |       |                | the page offset specifies the number  |
 * |      |       |                | of pages in the vectors page array.   |
 * +------+-------+----------------+---------------------------------------+
 *                                   ...
 * +------+-------+----------------+---------------------------------------+
 * |      | 07:00 | Vn Desc Index  |                                       |
 * |      | 07:00 | Vn Reserved    | Additional PA array pointers as       |
 * |      | 15:00 | Vn Desc Offset | needed.                               |
 * |      | 31:00 | Vn Size        |                                       |
 * +------+-------+----------------+---------------------------------------+
 * |      | 63:00 | Page Address   | Guest physical address of the page.   |
 * +------+-------+----------------+---------------------------------------+
 *                                   ...
 * +------+-------+----------------+---------------------------------------+
 * |      | 63:00 | Page Address   | Additional guest PAs as needed        |
 * +------+-------+----------------+---------------------------------------+
 *
 *
 * CONTINUATION DESCRIPTOR
 * -----------------------
 *
 * +------+-------+----------------+---------------------------------------+
 * |      |  BIT  |                |                                       |
 * |OFFSET| RANGE | NAME           | DESCRIPTION                           |
 * +------+-------+----------------+---------------------------------------+
 * | 0x00 | 07:00 | Version        | 0 for the current version             |
 * | 0x01 | 07:00 | Desc Type      | 1 - continuation descriptor           |
 * | 0x02 | 15:00 | Reserved       | Always 0                              |
 * | 0x04 | 31:00 | Reserved       | Always 0                              |
 * | 0x08 | 63:00 | Next Desc PA   | 0 - No further descriptors            |
 * |      |       |                | > 0 - Guest PA of next descriptor.    |
 * |      |       |                |       Must be aligned to 8 byte       |
 * +------+-------+----------------+---------------------------------------+
 * | 0x10 | 63:00 | Page Address   | Guest physical address of the page.   |
 * |      |       |                | The first page in vector includes the |
 * |      |       |                | offset for that vector.               |
 * +------+-------+----------------+---------------------------------------+
 *                                   ...
 * +------+-------+----------------+---------------------------------------+
 * |      | 63:00 | Page Address   | Additional guest PAs as needed        |
 * +------+-------+----------------+---------------------------------------+
 */

/* Descriptor types in I/O header. */
#define	ZVBLK_IODESC_TYPE_FIRST	0x0 /* First descriptor */
#define	ZVBLK_IODESC_TYPE_CONT	0x1 /* Continuation descriptor */

/* Make sure vector array fit into first desc page. */
#define	ZVBLK_MAX_VECTOR_NUM	\
	((PAGESIZE - sizeof (zvblk_first_desc_t)) / sizeof (zvblk_vec_desc_t))
/* Max I/O size, hard-coded to 1MB. */
#define	ZVBLK_MAX_SIZE	(1024 * 1024)
/*
 * Max number of pages - assume that, in average, each guest data page
 * provides at least DEV_BSIZE bytes of data.
 */
#define	ZVBLK_MAX_GPA_NUM	(ZVBLK_MAX_SIZE / DEV_BSIZE)
/*
 * Max number of descriptor pages - 1 first desc
 * page + cont desc pages for GPAs.
 */
#define	ZVBLK_MAX_DESC_NUM	\
	(sizeof (uint64_t) * ZVBLK_MAX_GPA_NUM / PAGESIZE + 1 + 1)

typedef struct zvblk_desc_header {
	uint8_t		zih_ver;
	uint8_t		zih_desc_type;
	/* Padding for uint64_t alignment */
	uint8_t		zih_reserved[6];
	uint64_t	zih_next_desc_gpa;
} zvblk_desc_header_t;

typedef	uint16_t zgbuf_id_t;

/*
 * When adding a new field by removing reserved/unused field, descriptor version
 * needs to be bumped up. When adding valid flags or values for existing field
 * there is no need to change the version. Just remember to update the mask
 * or related info if needed.
 */
typedef struct zvblk_first_desc {
	zvblk_desc_header_t 	zid_hdr;
	uint8_t			zid_pad[6];
	zgbuf_id_t		zid_id;
	uint8_t			zid_op;
	uint8_t			zid_desc_num;
	uint16_t		zid_vec_num;
	uint32_t		zid_page_num;
	uint64_t		zid_block;
	uint64_t		zid_size;
	uint8_t			zid_result;
	uint8_t			zid_xop;	/* Version 1 */
	uint8_t			zid_xerror;	/* Version 1 */
	/* Padding for uint64_t alignment */
	uint8_t			zid_reserved[5];
	uint64_t		zid_unused;
	/* variable number of uint64_t data starts */
	uint64_t		zid_data[1];
} zvblk_first_desc_t;

typedef struct zvblk_cont_desc {
	zvblk_desc_header_t 	zicd_hdr;
	/* variable number of uint64_t data starts */
	uint64_t		zicd_data[1];
} zvblk_cont_desc_t;

/* Make sure the whole structure will fit into a uint64_t. */
typedef struct zvblk_vec_desc {
	uint8_t			zvd_desc_index;
	uint8_t			zvd_reserved;
	uint16_t		zvd_desc_offset;
	uint32_t		zvd_size;
} zvblk_vec_desc_t;

#define	ZVBLK_OP_READ		0x01
#define	ZVBLK_OP_WRITE		0x02
#define	ZVBLK_OP_FLUSH_WCACHE	0x03
#define	ZVBLK_OP_FLAG_MASK	0x80 /* We have only one flag so far. */
/*
 * Can be OR'ed with other ops to indicate the op should be done on
 * return of the hypercall.
 */
#define	ZVBLK_OP_FLAG_SYNC	0x80

#define	ZVBLK_XOP_NOP		0x00
#define	ZVBLK_XOP_DPO		0x01
#define	ZVBLK_XOP_FUA		0x02
#define	ZVBLK_XOP_FLAG_MASK	0X03	/* Unison of valid flags */

#define	ZVBLK_DOOR_LOCK		0x01
#define	ZVBLK_DOOR_UNLOCK	0x02

#define	ZVBLK_RES_UNKNOWN	0x00
#define	ZVBLK_RES_OKAY		0x01
#define	ZVBLK_RES_FAIL		0x02
#define	ZVBLK_RES_NOTSUP	0x04
/*
 * The ZVBLK_RES_RESERVATION_CONFLICT indicates that the Reservation Conflict
 * is detected on the backend device.
 */
#define	ZVBLK_RES_RESERVATION_CONFLICT	0x08
/*
 * The backend target driver, i.e., sd, returns EIO for the failed IO.
 */
#define	ZVBLK_RES_EIO			0x10

/*
 * Maximum memory size allowed for user request.
 */
#define	ZVBLK_MAXIMUM_MEM_SIZE	1024*1024

#define	ZVBLK_IO_INTR		0x0
#define	ZVBLK_MEDIA_INTR	0x1

/* Or'able bits to present all media events to guest. */
#define	ZVBLK_MEDIA_STATE_CHANGED	0x1
#define	ZVBLK_MEDIA_EJECT_REQUESTED	0x2

/* Return value from ZVBLK_IO hypercall. */
#define	ZVBLK_IO_ISSUED	0	/* I/O request issued to backend. */
#define	ZVBLK_IO_DONE	1	/* I/O request is done. */
#define	ZVBLK_IO_FAILED	((uint64_t)-1)	/* Failed to handled I/O request. */

/*
 * Interface used in guest userspace.
 */
#define	ZVBLK_IOC_BASE		(('z' << 24) | ('V' << 16) | ('B') << 8)
#define	ZVBLK_IOC_BOOTABLE	(ZVBLK_IOC_BASE | 1)

/*
 * Zvblk hypercalls.
 */
#define	ZVBLK_FEATURE		(0)	/* Check feature. */
#define	ZVBLK_GET_INFO_V1	(1)	/* Get device/media info. */
#define	ZVBLK_CDROM_DOOR	(2)	/* Lock/Unlock door */
#define	ZVBLK_CDROM_EJECT	(3)	/* Eject media */
#define	ZVBLK_IO		(4)	/* IO. */
#define	ZVBLK_GETWCE		(5)	/* Get write cache status. */
#define	ZVBLK_SETWCE		(6)	/* Set write cache status. */
#define	ZVBLK_GET_MEDIA_EVENT	(7)	/* Get media event information. */
#define	ZVBLK_SET_COMPLETE_RING	(8)	/* Set complete ring buffer. */
#define	ZVBLK_GET_BOOTABLE	(9)	/* Is bootpri set in zonecfg? */
#define	ZVBLK_SCSI_RESV		(10)	/* Perform SCSI reservation */
#define	ZVBLK_GET_INFO_V2	(11)	/* Deprecated */
#define	ZVBLK_GET_CONFIG	(12)	/* Get device/media info */

/* Change this and add code to zks.c if new hypercalls are added */
#define	ZKS_BLK_LAST		ZVBLK_GET_CONFIG


/*
 * The following section describes the data structure used in ZVBLK_GET_INFO_V1.
 * The data structure contains only basic info of zvblk. All extended properties
 * should be obtained through ZVBLK_GET_CONFIG nvlist.
 */
#define	ZVBLK_MAX_DEVID_STR_LEN	(512)

/* All kinds of info we can get from host. */
#define	ZVBLK_DEV_SIZE		0x1
#define	ZVBLK_DEV_BSIZE		0x2
#define	ZVBLK_DEV_TYPE		0x4
#define	ZVBLK_DEV_ID		0x8
#define	ZVBLK_DISKID		0x10

typedef struct zvblk_info {
	uint64_t	zdi_size;	/* Size in bytes. */
	uint32_t	zdi_lblksize;	/* Logical block size. */
	uint32_t	zdi_pblksize;	/* Physical block size. */
	boolean_t	zdi_removable;	/* Is it removable? */
	dk_disk_id_t	zdi_diskid;	/* Disk ID data. */
	/* Device ID string. */
	char		zdi_devidstr[ZVBLK_MAX_DEVID_STR_LEN];
} zvblk_info_t;

/*
 * The following section describes the nvlist used in ZVBLK_GET_CONFIG. All
 * extended properties of zvblk should be passed using the nvlist.
 */
/* Extended info property name. */
#define	ZVBLK_INFO_IODESC_VER	"zvblk_info_iodesc_ver"
#define	ZVBLK_INFO_BOOTABLE	"zvblk_info_bootable"

#ifdef	__cplusplus
}
#endif

#endif /* _SYS_ZVBLK_COMMON_H */
