/*
 * Copyright (c) 2013, 2020, Oracle and/or its affiliates.
 */

#ifndef _SYS_ZVMM_HCALL_H
#define	_SYS_ZVMM_HCALL_H

#ifndef	_ASM
#include <sys/types.h>
#endif	/* _ASM */

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Hypercall feature versioning
 *
 * An initial set of hypercalls for each extender is guaranteed to be present,
 * and to behave the same (see the relevant host extender). After this set, any
 * new hypercall, or new feature using a slot in the hypercall space for a
 * particular extender, must be tested for.
 *
 * For each such feature or hypercall, the guest driver will call
 * ZV<foo>_FEATURE with the hypercall number as an argument.  The host extender
 * can use this as indication that the guest extender can handle the relevant
 * feature.  If it's supported, the host extender should return 0.  Otherwise,
 * it should return ENOENT.
 *
 * If the guest extender gets "ENOENT" it must not use the feature, and must
 * fall back to other methods.  *EVERY* new hypercall beyond the base set must
 * be tested for.
 *
 * This way, we can arbitrarily backport features as needed, as long as the
 * enumeration space is never renumbered.
 *
 * If a feature isn't available upon resume the guest should gracefully deal
 * with the lack of feature and continue. In some cases it may not be possible
 * to resume without a previously used feature. Features which once used and
 * then become required for correct operation of the guest should be recorded
 * in the extender required features array. The suspend/resume code will save
 * these features into the suspend image and refuse to resume a guest on a host
 * which does not support them. See ZVCNTRL_SPIN_WAIT in zvcntrl for an
 * example.
 */

/*
 * The 1st argument to any hypercall is the target extender ID (as per the
 * "ext-id" property).  Most devices are dynamic and discovered via enumeration,
 * however the following are fixed, to allow for early use in boot.
 */
#define	ZVMM_CNTRL_EXTID (0)
#define	ZVMM_CONS_EXTID	(1)
#define	ZVMM_LDC_EXTID	(2) /* Only available on sun4v */

#define	ZVMM_IRQ_NONE	0xffff
#define	ZVMM_IRQ_LAST	0x10000
#define	ZVMM_ITARG_ALL	(-1)
#define	ZVMM_ITARG_NONE	(-2)

#define	ZVCNTRL_FEATURE		(0)
#define	ZVCNTRL_ENUM_V1		(1)

#define	ZVCNTRL_BIND		(2)	/* 3rd arg is irq_cookie */
				/* 4th is interrupt vector or 0 to unbind */

#define	ZVCNTRL_GET_TOD		(3)	/* Return the time-of-day. */

#define	ZVCNTRL_GET_FREQ	(4)	/* Return the CPU frequency. */
#define	ZVCNTRL_REBOOT		(5)	/* reboot the machine */
#define	ZVCNTRL_HALT		(6)	/* poweroff the machine */
#define	ZVCNTRL_GET_HRTIME	(7)	/* gets hrtime */
#define	ZVCNTRL_CLK_IRQ		(8)	/* returns the IRQ the clock uses */
#define	ZVCNTRL_CLK_RESOLUTION	(9)	/* report clk resolution */
#define	ZVCNTRL_CLK_REPROGRAM	(10)	/* program clock to go off */

/* this only affect masking on the calling CPU */
#define	ZVCNTRL_IRQ_ENABLE	(11)	/* enables the arg IRQ */
#define	ZVCNTRL_IRQ_DISABLE	(12)	/* disable the arg IRQ */

/* this reports pending on any CPU */
#define	ZVCNTRL_IRQ_PENDING	(13)	/* report if an IRQ is pending */
#define	ZVCNTRL_ALLOC_EXT_IRQ	(14)	/* allocate an IRQ for extender use. */
/* useful for causing a VMEXIT to check for pending interrupts */
#define	ZVCNTRL_NOOP		(15)	/* no arguments */

/*
 * Start a secondary CPU. Takes the Solaris real mode platter phys addr as
 * an argument, but we have a data structure there.
 */
#define	ZVCNTRL_START_CPU	(16)	/* arg is zvg_start_cpu_t */

#define	ZVCNTRL_ALLOC_IRQ	(17)	/* allocates an IRQ for IPI use */
#define	ZVCNTRL_SEND_IPI	(18)	/* 2 args: cpu id, irq num */
/* useful for waiting when doing polled I/O */
#define	ZVCNTRL_PAUSE		(19)	/* no arguments */

#define	ZVCNTRL_GET_EVENT	(20)	/* ask zvcntrl for events to handle */
#define	ZVCNTRL_COMPLETE_EVENT	(21)	/* tell zvcntrl we've handled one */

/* event codes that are retrieved with ZVCTNRL_EVENT */
#define	ZVCNTRL_EVENT_NONE	(0)
/* Obsoleted ZVCNTRL_EVENT_CPUADD (1) */
/* Obsoleted ZVCNTRL_EVENT_CPUREM (2) */
#define	ZVCNTRL_EVENT_DEVADD	(3)	/* add device to guest */
#define	ZVCNTRL_EVENT_DEVREM	(4)	/* remove device from guest */
#define	ZVCNTRL_EVENT_FREELIST	(5)	/* generate free list */
#define	ZVCNTRL_EVENT_CPUADD_BATCH  (6) /* add list of cpus to guest */
#define	ZVCNTRL_EVENT_CPUREM_BATCH  (7) /* remove list of cpus from guest */


#define	ZVCNTRL_SUSPEND		(22)	/* places CPU in suspend */

/*
 * A work queue is a counter that allows guest CPUs to sleep
 * in the host, waiting for work to be posted by another guest CPU without
 * requiring the use of guest interrupts. These hcalls are only used on x86,
 * the equivalent sparc guest routines didn't require special hypercalls.
 */
#define	ZVCNTRL_POST		(23)	/* 3rd arg target vcpu */
#define	ZVCNTRL_WAIT		(24)
#define	ZVCNTRL_CONSUME		(25)	/* 3rd arg which vcpu */

/* Distinguish between different work queues */
#define	ZVCNTRL_WORKQ_XCALL	(0)	/* For x-calls */
#define	ZVCNTRL_WORKQ_CBE	(1)	/* For cbe x-calls  */
#define	ZVCNTRL_WORKQ_CPU	(2)	/* For cpu onlining */

/*
 * Identifies which IRQ to use for a perf counter overflow
 */
#define	ZVCNTRL_CPC_IRQ		(26)	/* returns perf ctr overflow irq */
#define	ZVCNTRL_FREE_EXT_IRQ	(27)	/* free an IRQ from extender use. */

#define	ZVCNTRL_DEV_RESUME_V1	(28)	/* two args, dev type, dev addr */
					/* returns extender id */

#define	ZVCNTRL_UNUSED_29	(29)	/* unused value */
#define	ZVCNTRL_GET_ZONEATTR	(30)	/* Get zone attribute */
#define	ZVCNTRL_SET_TOD		(31)	/* sets the time-of-day. */

#define	ZVCNTRL_MARK_AUX_STATE	(32)	/* Set auxiliary state */
#define	ZVCNTRL_UNMARK_AUX_STATE	(33)	/* Set auxiliary state */
#define	ZVCNTRL_GET_AUX_STATE	(34)
#define	ZVCNTRL_DEBUG_INFO_PA	(35)	/* Set debug_info guest pa */
#define	ZVCNTRL_READ_PILQ	(36)	/* get 1st entry in q for pil on cpu */
#define	ZVCNTRL_DELAY		(37)	/* Guest vcpu spinning on a lock */
#define	ZVCNTRL_GET_RANDOM_BYTES	(38)	/* Returns 64 bit random. */
#define	ZVCNTRL_CLK_DISABLE	(39)	/* disable clock */
#define	ZVCNTRL_VSTATS_QUERY	(40)	/* Query for count of supported */
					/* vcpu counters */
#define	ZVCNTRL_VSTATS_SHARE_PA	(41)	/* Share vcpu software counter memory */

#define	ZVCNTRL_SPIN_WAIT	(42)	/* Wait until a resource is released */
#define	ZVCNTRL_SPIN_RELEASE	(43)	/* Resource is released */
#define	ZVCNTRL_SPIN_WOULD_WAIT	(44)	/* Check if a vcpu has run recently */
#define	ZVCNTRL_ADD_CRIT_SECTION	(45)	/* Critial code section */

#define	ZVCNTRL_ENUM_V2		(46)	/* New ENUM, support dev id string */
#define	ZVCNTRL_DEV_RESUME_V2	(47)	/* New DEV_RESUME */
					/* support dev id string */

#define	ZVCNTRL_WRITE_HOSTDATA	(48)	/* Write hostdata to bootable disks */
#define	ZVCNTRL_SET_MARKER	(49)	/* write debug marker to log */
#define	ZVCNTRL_GET_TOD_HIRES	(50)	/* High-resolution TOD */
#define	ZVCNTRL_GET_CONTIG_SZ	(51)	/* host contig mem size KZ can get */
#define	ZVCNTRL_GET_NUM_IRQ	(52)	/* Number of IRQs for each extender */
#define	ZVCNTRL_ZVPCI_SUP_FEATURE	(53) /* zvpci supports feature hcall */
#define	ZVCNTRL_CPU_STATUS_AREA	(54)	/* cpu status area for pv locking */

/* 55 obsolete ZVCNTRL_MAP_PILQ hypercall - don't reuse */

#define	ZVCNTRL_CRASHDUMP	(56)	/* request host take a core dump */
#define	ZVCNTRL_EOI_EDGE_TRIGGERED (57)	/* zvnex fixed to be edge triggered */
#define	ZVCNTRL_UPDATE_FREELIST (58)	/* guests update free list */
#define	ZVCNTRL_UNSCALED_FREQ	(59)	/* query unscaled TSC rate */
#define	ZVCNTRL_STOP_CPU	(60)	/* Stop a CPU (used by the LZR) */
#define	ZVCNTRL_COMPLETE_EVENT_V2 (61)	/* Complete event with resp. buffer */

/*
 * This feature hypercall is used by guest implicit DLMP, a.k.a iDLMP,
 * checking whether the host supports iDLMP. Guest iDLMP uses this
 * feature to determine when to bring up a datalink once all necessary
 * devices are ready and attached in the kernel. It needs to be a zvcntrl
 * hypercall since there could be case that all datalinks are direct I/O
 * without the presence of zvnet extender, e.g. on a host that do not
 * understand iDLMP.
 *
 * The ZVCNTRL_IDLMP_SUPPORT feature will be saved to suspend image as a
 * required feature, depending on the presence of iov-enabled anet
 * (anet:iov property set to "auto" or "on") before kernel zone
 * suspension. zvcntrl_update_iov_devs() is used to keep track of the
 * count for iov-enabled anets, and to set/clear the required feature
 * accordingly.
 */
#define	ZVCNTRL_IDLMP_SUPPORT	(62)	/* host supports iDLMP */

#define	ZVCNTRL_OPTIMIZED_PERFCNT	(63) /* Improved perf counter support */
#define	ZVCNTRL_PRECISE_OVF	(64)	/* CPC precise overflow support */
#define	ZVCNTRL_READ_ENTROPY	(65)	/* Provide entropy. */

#define	ZVCNTRL_IRQ_PAGE	(66)	/* x86: zvmm_mach_hcall.h */
#define	ZVCNTRL_IBRS_MSR	(67)	/* x86: IBRS/IBPB MSRs known */
#define	ZVCNTRL_ARCHCAP_MSR	(68)	/* x86: Arch Capability MSR known */
#define	ZVCNTRL_FLUSH_MSR	(69)	/* x86: Flush MSR known */
#define	ZVCNTRL_TSX_DISABLE	(70)	/* x86: TSX is disabled */

/* Change this and add code to zks.c if new hypercalls are added */
#define	ZKS_CNTRL_LAST		ZVCNTRL_TSX_DISABLE

/*
 * Zvmm vstat performance counters.
 *
 * Used to export various hypervisor statistics into the guest.
 * The guest first needs to query for the maximum number of supported counters
 * using ZVCNTRL_VSTATS_QUERY. It then shares a piece of memory using
 * ZVCNTRL_VSTATS_SHARE_PA.
 * Finally it can enable the selected counter by setting vstat_ctrl in the
 * shared memory. As vstats are updated on a guest switch, its value
 * (vstat_value) is only guaranteed to be updated only after a guest switch
 * happens.  It's therefore wise to issue a ZVCNTRL_NOOP hypercall after
 * changing any vstat counter settings. For speed & efficiency the guest
 * should keep the VSTAT_END_MARK counter at the end of the counter array.
 * The host does not look at any counters beyond this mark.
 *
 * Vstat counters are never removed but they may be made obsolete. Obsolete
 * counters will always be zero. All counter numbers must be preserved when
 * backporting and all counters up to the backported one must exist.
 * If needed zero-counters must be created.
 */
typedef struct cpu_vstat {
	uint64_t	vstat_ctrl;	/* enabled counter or disabled */
	volatile uint64_t vstat_value;	/* counter value written by host */
} cpu_vstat_t;

/*
 * Values for vstat_ctrl. All entries past the first VSTAT_END_MARK counter
 * in the ZVCNTRL_VSTAT_SHARE_PA area are also disabled no matter what is in
 * vstat_ctrl. VSTAT_DISABLED counters are not updated.
 * When new types of counted events are added, they go at the end and
 * VSTAT_NR_CTRS is updated.
 */
#define	VSTAT_END_MARK	(UINT64_MAX)	/* Don't update following counters */
#define	VSTAT_DISABLED	(0)		/* Don't update this counter */
#define	VSTAT_STOLEN	(1)		/* Time stolen by host */
#define	VSTAT_NR_CTRS	(2)		/* Number of counters */

#ifdef _KERNEL

#if !defined(sparc)

/*
 * For Intel/AMD we use different instructions for hypercall, so
 * the first time through has to be an init function to set it up
 */
extern uint64_t (*zvg_hypercall)(uint64_t, uint64_t, ...);
extern uint64_t zvg_hypercall_init(uint64_t, uint64_t, ...);

#else	/* !sparc */

#ifndef	_ASM
extern uint64_t zvg_hypercall(uint64_t, uint64_t, ...);
#endif	/* _ASM */

#endif	/* !sparc */

#endif /* _KERNEL */

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_ZVMM_HCALL_H */
