/*
 * Copyright (c) 2013, 2016, Oracle and/or its affiliates. All rights reserved.
 */

/* Shared interface between zvsdir guest and host. */

#ifndef	_SYS_ZVSDIR_COMMON_H
#define	_SYS_ZVSDIR_COMMON_H

#include <sys/param.h>
#include <sys/time.h>

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Each zvnode_handle_t represent a single zvnode which, in turn, represent
 * a single vnode on host side and has been opened by guest. So, we can have
 * at most 2^ZVNODE_ID_SHIFT nodes opened per shared directroy file system.
 */
typedef uint64_t zvnode_handle_t;
#define	ZVNODE_HANDLE_SHIFT	16 /* # of bits for a valid handle */
#define	ZVNODE_MAX_HANDLE	((1 << ZVNODE_HANDLE_SHIFT) - 1)
#define	ZVNODE_INVAL_HANDLE	ZVNODE_MAX_HANDLE
#define	ZVNODE_HANDLE_MASK	ZVNODE_MAX_HANDLE

/*
 * Host can raise async report of a zvnode to guest/zvsfs for different
 * reasons OR'ed together. There can be 16 reasons so far.
 * Note that zvnode_reason and zvnode_handle will be OR'ed into one uint64_t
 * variable when passed from host to guest. So make sure they can fit.
 */
typedef uint16_t zvnode_reason_t;
#define	ZVNODE_REASON_ATTR_CHG		(0x1)
#define	ZVNODE_REASON_DELETED		(0x2)
#define	ZVNODE_REASON_RENAMED		(0x4)

/*
 * Hypercall interfaces.
 */

/*
 * Hypercall commands.
 */
/* Check feature. */
#define	ZVSDIR_FEATURE		(0)
/* Connect to host. */
#define	ZVSDIR_CONNECT		(1)
/* Disconnect from host. */
#define	ZVSDIR_DISCONNECT	(2)
/* Send I/O request to host. */
#define	ZVSDIR_START_IO		(3)
/* Get chained finished I/O requests from host. */
#define	ZVSDIR_GET_COMPLETE_IO	(4)
/* Get node report from host. */
#define	ZVSDIR_GET_REPORT	(5)
/* Change this and add code to zks.c if new hypercalls are added */
#define	ZKS_SDIR_LAST	ZVSDIR_GET_REPORT

/*
 * Below are data structures defined for passing arguments and results
 * between guest and host in each hypercall.
 * Note that they must fit in one page since we only pass one page
 * between host and guest for now during hypercalls.
 */
#define	HCALL_BUF_SIZE	PAGESIZE

/*
 * zvfs_info_t contains mostly a subset of statvfs64_t, which is
 * needed from underlying file system.
 */
typedef struct zvfs_info {
	unsigned long	fi_maxnamesize;	/* max file name size */
	unsigned long	fi_blksize;	/* fs blk size */
	fsblkcnt64_t	fi_totalblks;	/* tobal # of blk */
	fsblkcnt64_t	fi_blkavail;	/* # of blk available */
	fsfilcnt64_t	fi_totalfiles;	/* total # of files */
	fsfilcnt64_t	fi_fileavail;	/* # of file nodes available */
	char		fi_desc_str[MAXNAMELEN]; /* description string */
} zvfs_info_t;

/*
 * Supported node types.
 * Will add more types if it makes sense.
 */
typedef enum zvnode_type {
	UNKNOWN,
	ZVDIR,
	ZVREG
} zvnode_type_t;

/*
 * zvnode_info_t contains a subset of vattr_t, which is needed from
 * underlying file system.
 */
typedef struct zvnode_info {
	zvnode_type_t	ni_type;	/* vnode type */
	mode_t		ni_mode;	/* file access mode */
	ino64_t		ni_nodeid;	/* node id */
	off64_t		ni_size;	/* file size in bytes */
	timestruc_t	ni_atime;	/* time of last access */
	timestruc_t	ni_mtime;	/* time of last modification */
	timestruc_t	ni_ctime;	/* time of last status change */
	blksize_t	ni_blksize;	/* fundamental block size */
	blkcnt64_t	ni_nblocks;	/* # of blocks allocated */
} zvnode_info_t;

/*
 * Define our own version of dirent64_t.
 */
typedef struct zvnode_dir_entry {
	ino64_t		nd_ino;		/* "inode number" of entry */
	off64_t		nd_off;		/* offset of disk directory entry */
	uint16_t	nd_reclen;	/* length of this record */
	char		nd_name[1];	/* file name */
} zvnode_dir_entry_t;

typedef struct zvsdir_node_open {
	zvnode_handle_t	zno_dirhdl;	/* in: handle of the target dir node */
	char		zno_name[MAXNAMELEN]; /* in: file name in the dir */
	zvnode_handle_t	zno_hdl;	/* out: handle of file node */
} zvsdir_node_open_t;

#define	ZVSDIR_NODEINFO_ATTR	0x1
#define	ZVSDIR_NODEINFO_NAME	0x2
typedef struct zvsdir_node_info {
	zvnode_handle_t	zni_handle;	/* in: node handle */
	uint64_t	zni_flag;	/* in: select what info to get */
	zvnode_info_t	zni_info;	/* out: node attr */
	char		zni_name[MAXNAMELEN]; /* out: node file name */
} zvsdir_node_info_t;

#define	ZVSDIR_MAX_READDIR_BUF_LEN	(HCALL_BUF_SIZE -	\
	offsetof(zvsdir_hcall_t, u.zh_readdir.znrd_entry))
typedef struct zvsdir_node_readdir {
	zvnode_handle_t	znrd_dirhdl; /* in: handle for the target dir node */
	uint64_t	znrd_length; /* in: data buffer length */
	uint64_t	znrd_offset; /* in&out: offset to read next dir entry */
	uint64_t	znrd_bytes_read;  /* out: # of bytes read in this I/O */
	boolean_t	znrd_eof;    /* out: B_TRUE == eof */
	zvnode_dir_entry_t znrd_entry[1]; /* out: dir entries buffer */
} zvsdir_node_readdir_t;

#define	ZVSDIR_MAX_READ_BUF_LEN		(HCALL_BUF_SIZE -	\
	offsetof(zvsdir_hcall_t, u.zh_read.znr_data))
typedef struct zvsdir_node_read {
	zvnode_handle_t	znr_hdl;	/* in: handle for the target node */
	uint64_t	znr_length;	/* in: data buffer length */
	uint64_t	znr_offset;	/* in: offset to start reading */
	uint64_t	znr_bytes_read; /* out: # of bytes read in this I/O */
	char		znr_data[1];	/* out: data buffer */
} zvsdir_node_read_t;

typedef struct zvsdir_node_close {
	zvnode_handle_t	znc_handle;	/* in: handle for the target node */
} zvsdir_node_close_t;

typedef enum zvsdir_op {
	ZVSDIR_OP_FS_GETINFO,
	ZVSDIR_OP_NODE_OPEN,
	ZVSDIR_OP_NODE_CLOSE,
	ZVSDIR_OP_NODE_GETINFO,
	ZVSDIR_OP_NODE_READDIR,
	ZVSDIR_OP_NODE_READ
} zvsdir_op_t;

typedef struct zvsdir_hcall_header {
	zvsdir_op_t	zhh_op;		/* in: operation code */
	int		zhh_error;	/* out: operation result */
	uint64_t	zhh_op_id;	/* in: id echoed back from host */
	uint64_t	zhh_next_op_id; /* out: chained id */
} zvsdir_hcall_header_t;

typedef struct zvsdir_hcall {
	zvsdir_hcall_header_t	zh_header;
	union {
		zvfs_info_t		zh_fs_info;
		zvsdir_node_open_t	zh_open;
		zvsdir_node_close_t	zh_close;
		zvsdir_node_info_t	zh_info;
		zvsdir_node_readdir_t	zh_readdir;
		zvsdir_node_read_t	zh_read;
	} u;
} zvsdir_hcall_t;

#ifdef	__cplusplus
}
#endif

#endif /* _SYS_ZVSDIR_COMMON_H */
