/*
 * Description
 * ImaLib.c - Implements a sample common IMA library
 *
 * License:
 * The contents of this file are subject to the SNIA Public License
 * Version 2.0(the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *
 * /http://www.snia.org/English/Resources/Code/OpenSource.html
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 *  the License for the specific language governing rights and limitations
 *  under the License.
 *
 * The Original Code is  SNIA HBA API and IMA general header file
 *
 * The Initial Developer of the Original Code is:
 * Benjamin F. Kuo, Troika Networks, Inc. (benk@troikanetworks.com)
 * David Dillard       VERITAS Software        david.dillard@veritas.com
 *
 * Contributor(s):
 * Jeff Ding, Adaptec, Inc. (jding@corp.adaptec.com)
 *
 *   Changes:
 *  09/24/2003 Initial Draft
 *  (for other changes... see the CVS logs)
 *
 *  12/15/2003 corrected the defined parameter in IMA_SetPhbaIsnsDiscovery().
 *  lower case the computer name as iscsi name in IMA_GenerateNodeName().
 *
 *  01/21/2005 Updated to support IMA 1.1.3.
 */

/*
 * Copyright (c) 2006, 2017, Oracle and/or its affiliates. All rights reserved.
 */

#include <unistd.h>
#include <pthread.h>
#define	_XOPEN_SOURCE /* glibc2 needs this */
#include <dlfcn.h>
#include <stdarg.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <syslog.h>

#include "libsun_ima.h"
#include "ima.h"
#include "ima-plugin.h"

#define	LIBRARY_PROPERTY_SUPPORTED_IMA_VERSION 2
#define	LIBRARY_PROPERTY_IMPLEMENTATION_VERSION L"2.0.0"
#define	LIBRARY_PROPERTY_VENDOR L"QLogic, Inc."
#define	DEFAULT_NODE_NAME_FORMAT "iqn.1986-03.com.sun.central.%s"

#define	LIBRARY_FILE_NAME L"libima.so"
#define	EUOS_ERROR IMA_ERROR_UNEXPECTED_OS_ERROR

#define	IMA_NODENAME_LENGTH 256

IMA_PLUGIN_INFO  plugintable[IMA_MAX_NUM_PLUGINS];
int number_of_plugins = -1;
static IMA_NODE_NAME    sharedNodeName;
static IMA_NODE_ALIAS   sharedNodeAlias;

pthread_mutex_t libMutex = PTHREAD_MUTEX_INITIALIZER;

void InitLibrary();
void ExitLibrary();

static void libSwprintf(wchar_t *wcs,
	const wchar_t *lpszFormat, ...) {
	va_list args;
	va_start(args, lpszFormat);
	(void) vswprintf(wcs, 255, lpszFormat, args);
	va_end(args);
}

void so_init(void);
void so_fini(void);
static int os_createmutex(pthread_mutex_t *mutex);
static void os_obtainmutex(pthread_mutex_t *mutex);
static void os_releasemutex(pthread_mutex_t *mutex);
static void os_destroymutex(pthread_mutex_t *mutex);
static IMA_STATUS getSolarisNodeProps(IMA_NODE_PROPERTIES *nodeProps);
static IMA_STATUS getSolarisSharedNodeName(IMA_NODE_NAME name);
static IMA_STATUS getSolarisSharedNodeAlias(IMA_NODE_ALIAS alias);
static IMA_STATUS setSolarisSharedNodeName(const IMA_NODE_NAME name);
static IMA_STATUS setSolarisSharedNodeAlias(const IMA_NODE_ALIAS alias);

#pragma init(so_init)
#pragma fini(so_fini)

void so_init() {
	InitLibrary();
}
void so_fini() {
	ExitLibrary();
}

static IMA_STATUS getSolarisNodeProps(IMA_NODE_PROPERTIES *nodeProps) {
	int i;
	IMA_STATUS status = IMA_ERROR_UNKNOWN_ERROR;
	IMA_GetNodePropertiesFn PassFunc;
	IMA_OID nodeOid;

	bzero((void *)&nodeOid, sizeof (IMA_OID));
	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (strstr(plugintable[i].PluginPath,
		    "libsun_ima.so") == NULL) {
			continue;
		}
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc =
			    (IMA_GetNodePropertiesFn) dlsym(
			    plugintable[i].hPlugin,
			    "IMA_GetNodeProperties");
			if (PassFunc != NULL) {
				status = PassFunc(nodeOid, nodeProps);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		break;
	}

	os_releasemutex(&libMutex);
	return (status);
}

static IMA_STATUS getSolarisSharedNodeName(IMA_NODE_NAME name) {
	IMA_STATUS status = IMA_ERROR_UNKNOWN_ERROR;
	IMA_NODE_PROPERTIES nodeProps;

	status = getSolarisNodeProps(&nodeProps);
	if (status != IMA_STATUS_SUCCESS) {
		return (status);
	}
	bcopy((void *)&nodeProps.name, name, sizeof (IMA_NODE_NAME));
	return (status);
}

static IMA_STATUS getSolarisSharedNodeAlias(IMA_NODE_ALIAS alias) {
	IMA_STATUS status = IMA_ERROR_UNKNOWN_ERROR;
	IMA_NODE_PROPERTIES nodeProps;

	status = getSolarisNodeProps(&nodeProps);
	if (status != IMA_STATUS_SUCCESS) {
		return (status);
	}
	bcopy((void *)&nodeProps.alias, alias, sizeof (IMA_NODE_ALIAS));
	return (status);
}

static IMA_STATUS setSolarisSharedNodeName(const IMA_NODE_NAME name) {
	int i;
	IMA_STATUS status = IMA_ERROR_UNKNOWN_ERROR;
	IMA_SetNodeNameFn PassFunc;
	IMA_OID nodeOid;

	bzero((void *)&nodeOid, sizeof (IMA_OID));
	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (strstr(plugintable[i].PluginPath,
		    "libsun_ima.so") == NULL) {
			continue;
		}
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc =
			    (IMA_SetNodeNameFn) dlsym(plugintable[i].hPlugin,
			    "IMA_SetNodeName");
			if (PassFunc != NULL) {
				status = PassFunc(nodeOid, name);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		break;
	}

	os_releasemutex(&libMutex);
	return (status);
}

static IMA_STATUS setSolarisSharedNodeAlias(const IMA_NODE_ALIAS alias) {
	int i;
	IMA_STATUS status = IMA_ERROR_UNKNOWN_ERROR;
	IMA_SetNodeAliasFn PassFunc;
	IMA_OID nodeOid;

	bzero((void *)&nodeOid, sizeof (IMA_OID));
	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (strstr(plugintable[i].PluginPath,
		    "libsun_ima.so") == NULL) {
			continue;
		}
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc =
			    (IMA_SetNodeAliasFn) dlsym(plugintable[i].hPlugin,
			    "IMA_SetNodeAlias");
			if (PassFunc != NULL) {
				status = PassFunc(nodeOid, alias);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		break;
	}

	os_releasemutex(&libMutex);
	return (status);
}

/* Create the mutex.  Return 1 if successful, 0 otherwise */
int os_createmutex(pthread_mutex_t *mutex) {
	if (pthread_mutex_init(mutex, NULL) == 0)
		return (1);
	return (0);
}

void os_obtainmutex(pthread_mutex_t *mutex) {
	(void) pthread_mutex_lock(mutex);
}

void os_releasemutex(pthread_mutex_t *mutex) {
	(void) pthread_mutex_unlock(mutex);
}

void os_destroymutex(pthread_mutex_t *mutex) {
	(void) pthread_mutex_destroy(mutex);
}

void InitLibrary() {

	FILE *imaconf;
	char fullline[512]; /* Full line read in from IMA.conf */
	char pluginname[64]; /* Read in from file IMA.conf */
	char pluginpath[256]; /* Read in from file IMA.conf */
	char imaConfFilePath[256];
	char *charPtr;
	IMA_UINT i = 0;

	os_obtainmutex(&libMutex);
	if (number_of_plugins != -1) {
		os_releasemutex(&libMutex);
		return;
	}

	number_of_plugins = 0;
	sharedNodeAlias[0] = 0;

	/* Open configuration file from known location */
	(void) strlcpy(imaConfFilePath, "/etc/ima.conf", 256);

	if ((imaconf = fopen(imaConfFilePath, "r")) == NULL) {
		os_releasemutex(&libMutex);
		return;
	}
	/* Read in each line and load library */
	while ((imaconf != NULL) &&
	    (fgets(fullline, sizeof (fullline), imaconf))) {
		if ((fullline[0] != '#') && (fullline[0] != '\n')) {
			/* Take out the '\n' */
			if ((charPtr = (char *)strchr(fullline, '\n')) != NULL)
				*charPtr = '\0';

			/* look for the first tab */
			if ((charPtr = (char *)strchr(fullline, '\t')) == NULL)
				charPtr = (char *)strchr(fullline, ' ');

			/* Set Null termination for library name if found */
			if (charPtr != NULL) {
				*charPtr++ = '\0';
				/*
				 * Skip spaces and tabs until
				 * the next character found
				 */
				while ((*charPtr == ' ') || (*charPtr == '\t'))
					charPtr++;
			}
			else
				continue; /* May be invalid entry */

			/* Copy library name and path */
			(void) strlcpy(pluginname, fullline, 64);
			(void) strlcpy(pluginpath, charPtr, 256);

			/*
			 * Continue to the next line if library name or
			 * path is invalid
			 */
			if ((strlen(pluginname) == 0) ||
			    (strlen(pluginpath) == 0))
				continue;

			/* Load the DLL now */
			plugintable[i].hPlugin = dlopen(pluginpath, RTLD_LAZY);
			if (plugintable[i].hPlugin != NULL) {
				typedef int (*InitializeFn)();
				InitializeFn PassFunc;

				(void) strlcpy(
				    (char *)&plugintable[i].PluginName,
				    (char *)&pluginname, 64);
				(void) strlcpy((char *)
				    &plugintable[i].PluginPath,
				    (char *)&pluginpath, 256);
				plugintable[i].ownerId = i + 1;

				PassFunc = (InitializeFn)
				    dlsym(
				    plugintable[i].hPlugin, "Initialize");
				if (PassFunc != NULL) {
					PassFunc(plugintable[i].ownerId);
				}

				plugintable[i].number_of_vbcallbacks = 0;
				plugintable[i].number_of_pccallbacks = 0;
				(void) os_createmutex(
				    &(plugintable[i].pluginMutex));
				i++;
			}
		}
	}
	(void) fclose(imaconf);
	number_of_plugins = i;
	os_releasemutex(&libMutex);
}


void ExitLibrary() {
	IMA_UINT j;
	int i;

	os_obtainmutex(&libMutex);
	if (number_of_plugins == -1) {
		os_releasemutex(&libMutex);
		return;
	}

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].hPlugin != NULL) {
			TerminateFn ExitPassFunc;

			os_obtainmutex(&(plugintable[i].pluginMutex));
			for (j = 0; j < plugintable[i].number_of_vbcallbacks;
			    j++) {
#define	IMA_DFOBC_STR "IMA_DeregisterForObjectVisibilityChangesFn"
				IMA_DeregisterForObjectVisibilityChangesFn
				    PassFunc;

				PassFunc =
				    (IMA_DeregisterForObjectVisibilityChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    IMA_DFOBC_STR);
				if (PassFunc != NULL) {
					PassFunc(plugintable[i].vbcallback[j]);
				}
#undef IMA_DFOBC_STR
			}
			plugintable[i].number_of_vbcallbacks = 0;

			for (j = 0; j < plugintable[i].number_of_pccallbacks;
			    j++) {
				IMA_DeregisterForObjectPropertyChangesFn
				    PassFunc;
				PassFunc =
				    (IMA_DeregisterForObjectPropertyChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_DeregisterForObjectPropertyChangesFn");
				if (PassFunc != NULL) {
					PassFunc(plugintable[i].pccallback[j]);
				}
			}
			plugintable[i].number_of_pccallbacks = 0;

			ExitPassFunc = (TerminateFn)
			    dlsym(plugintable[i].hPlugin, "Terminate");
			if (ExitPassFunc != NULL) {
				ExitPassFunc();
			}
			/* Unload DLL from memory */
			(void) dlclose(plugintable[i].hPlugin);
			os_releasemutex(&(plugintable[i].pluginMutex));
			os_destroymutex(&(plugintable[i].pluginMutex));
		}
	}
	number_of_plugins = -1;
	os_releasemutex(&libMutex);
	os_destroymutex(&libMutex);
}


static void VisibilityCallback(
    IMA_BOOL becomingVisible,
    IMA_OID objectId) {
	int i;
	IMA_UINT j;
	os_obtainmutex(&libMutex);
	for (i = 0; i < number_of_plugins; i++) {
		if ((plugintable[i].hPlugin != NULL) &&
		    (objectId.ownerId == plugintable[i].ownerId)) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			for (j = 0;
			    j < plugintable[i].number_of_vbcallbacks;
			    j++) {
				(plugintable[i].vbcallback[j])
				    (becomingVisible, objectId);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
	}
	os_releasemutex(&libMutex);

}

static void PropertyCallback(
    IMA_OID objectId) {
	int i;
	IMA_UINT j;

	os_obtainmutex(&libMutex);
	for (i = 0; i < number_of_plugins; i++) {
		if ((plugintable[i].hPlugin != NULL) &&
		    (objectId.ownerId == plugintable[i].ownerId)) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			for (j = 0;
			    j < plugintable[i].number_of_pccallbacks;
			    j++) {
				(plugintable[i].pccallback[j])(objectId);
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
	}
	os_releasemutex(&libMutex);
}

/*
 * Gets the date and time, in the form of an IMA_DATETIME, from the build
 * script when compiled.
 */
static void GetBuildTime(IMA_DATETIME* pdatetime) {

#if defined(BUILD_DATE)
	if (strptime(BUILD_DATE, "%Y/%m/%d %T %Z", pdatetime) == NULL) {
		(void) memset(pdatetime, 0, sizeof (IMA_DATETIME));
	}
#else
	(void) memset(pdatetime, 0, sizeof (IMA_DATETIME));
#endif
}



/*
 * Gets the properties of the IMA library that is being used.
 *
 * @param pProps A pointer to an @ref IMA_LIBRARY_PROPERTIES structure
 *    allocated by the caller.  On successful return this structure will
 *    contain the properties of the IMA library.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *     an error occurred.
 * @retval IMA_SUCCESS Returned if the library properties were successfully
 *    returned.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pProps is NULL or
 *    specifies a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetLibraryProperties(
    IMA_LIBRARY_PROPERTIES *pProps) {

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	/* Fill in the library properties. */
	GetBuildTime(&pProps->buildTime);
	pProps->supportedImaVersion = LIBRARY_PROPERTY_SUPPORTED_IMA_VERSION;
	libSwprintf(pProps->implementationVersion, L"%ls",
	    LIBRARY_PROPERTY_IMPLEMENTATION_VERSION);
	libSwprintf(pProps->vendor, L"%ls", LIBRARY_PROPERTY_VENDOR);

	libSwprintf(pProps->fileName, LIBRARY_FILE_NAME);

	return (IMA_STATUS_SUCCESS);
}


/*
 * Gets a list of the object IDs of all currently loaded plugins.
 *
 * @param ppList A pointer to a pointer to an @ref IMA_OID_LIST.
 *    On successful return this will contain a pointer to an @ref
 *    IMA_OID_LIST which contains the object IDs of all of the plugins
 *    currently loaded by the library.
 * @return An IMA_STATUS indicating if the operation was successful
 *    or if an error occurred.
 * @retval IMA_SUCCESS Returned if the plugin ID list was successfully
 *    returned.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a ppList is NULL or
 *    specifies a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetPluginOidList(
    IMA_OID_LIST **ppList) {
	int i;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);

	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (number_of_plugins - 1) * sizeof (IMA_OID));

	if ((*ppList) == NULL)
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);

	(*ppList)->oidCount = number_of_plugins;

	for (i = 0; i < number_of_plugins; i++) {

		(*ppList)->oids[i].objectType = IMA_OBJECT_TYPE_PLUGIN;
		(*ppList)->oids[i].ownerId = plugintable[i].ownerId;
		(*ppList)->oids[i].objectSequenceNumber = 0;

	}
	os_releasemutex(&libMutex);
	return (IMA_STATUS_SUCCESS);
}




/*
 * Gets the properties of the specified vendor plugin.
 *
 * @param pluginId The ID of the plugin whose properties are being retrieved.
 * @param pProps A pointer to an @ref IMA_PLUGIN_PROPERTIES structure
 *    allocated by the caller.  On successful return this will contain the
 *    properties of the plugin specified by pluginId.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the plugin properties were successfully
 *    returned.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a pluginId does not
 *    specify any valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if @a pluginId does not
 *    specify a plugin object.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a pluginId refers to a
 *     plugin, but not one that is currently loaded.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pProps is NULL or
 *    specify a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetPluginProperties(
    IMA_OID pluginOid,
    IMA_PLUGIN_PROPERTIES *pProps) {
	IMA_GetPluginPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((pluginOid.objectType != IMA_OBJECT_TYPE_PLUGIN) ||
	    (pluginOid.objectSequenceNumber != 0))
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == pluginOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPluginPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPluginProperties");
				if (PassFunc != NULL) {
					status = PassFunc(pluginOid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}




/*
 * Gets the object ID for the plugin associated with the specified object ID.
 *
 * @param objectId The object ID of an object that has been received from
 *    a previous library call.
 * @param pPluginId A pointer to an @ref IMA_OID structure allocated by the
 *    caller.  On successful return this will contain the object ID of the
 *    plugin associated with the object specified by @a objectId.  This
 *    can then be used to work with the plugin, e.g., to get the
 *    properties of the plugin or the send the plugin an IOCtl.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the associated plugin ID was
 *    successfully returned.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pPluginId is NULL
 *    or specifes a memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a objectId specifies
 *    an object not owned by a plugin, but instead one that is owned by
 *    the library.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a objectId specifies
 *    an object with an invalid type.
 */
IMA_API IMA_STATUS IMA_GetAssociatedPluginOid(
    IMA_OID objectId,
    IMA_OID *pPluginId) {
	int i;
	IMA_STATUS status;

	if (pPluginId == NULL || objectId.ownerId == RL_LIBRARY_SEQNUM)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (objectId.objectType != IMA_OBJECT_TYPE_UNKNOWN &&
	    objectId.objectType != IMA_OBJECT_TYPE_PLUGIN &&
	    objectId.objectType != IMA_OBJECT_TYPE_NODE &&
	    objectId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    objectId.objectType != IMA_OBJECT_TYPE_PHBA &&
	    objectId.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL &&
	    objectId.objectType != IMA_OBJECT_TYPE_PORTAL_GROUP &&
	    objectId.objectType != IMA_OBJECT_TYPE_LNP &&
	    objectId.objectType != IMA_OBJECT_TYPE_PNP &&
	    objectId.objectType != IMA_OBJECT_TYPE_TARGET &&
	    objectId.objectType != IMA_OBJECT_TYPE_LU &&
	    objectId.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    objectId.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INVALID_OBJECT_TYPE);

	os_obtainmutex(&libMutex);

	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (objectId.ownerId == plugintable[i].ownerId) {
			pPluginId->objectType = IMA_OBJECT_TYPE_PLUGIN;
			pPluginId->ownerId = plugintable[i].ownerId;
			pPluginId->objectSequenceNumber = 0;
			status = IMA_STATUS_SUCCESS;
		}

	}
	os_releasemutex(&libMutex);
	return (status);
}




/*
 * Gets the object ID of the shared node.
 *
 * @param pSharedNodeId A pointer to an @ref IMA_OID structure allocated by
 *    the caller.  On successful return it will contain the object ID of the
 *    shared node of the currently executing system is placed.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the shared node ID has been successfully
 *    retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pSharedNodeId is NULL
 *    or specifies a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetSharedNodeOid(
    IMA_OID *pSharedNodeId) {
	if (pSharedNodeId == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	pSharedNodeId->objectType = IMA_OBJECT_TYPE_NODE;
	pSharedNodeId->ownerId = RL_LIBRARY_SEQNUM;
	pSharedNodeId->objectSequenceNumber = RL_SHARED_NODE_SEQNUM;
	return (IMA_STATUS_SUCCESS);
}


IMA_API IMA_STATUS IMA_GetObjectType(
    IMA_OID oid,
    IMA_OBJECT_TYPE *pObjectType) {
	IMA_STATUS status;
	int i;

	if (pObjectType == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_UNKNOWN &&
	    oid.objectType != IMA_OBJECT_TYPE_PLUGIN &&
	    oid.objectType != IMA_OBJECT_TYPE_NODE &&
	    oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL &&
	    oid.objectType != IMA_OBJECT_TYPE_PORTAL_GROUP &&
	    oid.objectType != IMA_OBJECT_TYPE_LNP &&
	    oid.objectType != IMA_OBJECT_TYPE_PNP &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_LU &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INVALID_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			*pObjectType = oid.objectType;
			status = IMA_STATUS_SUCCESS;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}



/*
 * Gets the properties of the specified iSCSI node.
 * @param nodeId The ID of the node to get the properties of.
 * @param pProps A pointer to an @ref IMA_NODE_PROPERTIES structure
 *    which on successfully return
 *    will contain the properties of the specified node.
 * @return An IMA_STATUS indicating if the operation was successful or
 *    if an error occurred.
 * @retval IMA_SUCCESS Returned if the node properties have been
 *    successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pProps is NULL
 *    or specifies a memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a nodeId does
 *     not specify any valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if @a nodeId does
 *    not specify a node object.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a nodeId does not
 *    specify a node which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetNodeProperties(
    IMA_OID nodeOid,
    IMA_NODE_PROPERTIES *pProps) {
	IMA_GetNodePropertiesFn PassFunc;
	int i;
	IMA_STATUS status;
	char fullline[IMA_NODENAME_LENGTH];
	char nodename[IMA_NODENAME_LENGTH];

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (nodeOid.objectType != IMA_OBJECT_TYPE_NODE)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	if ((nodeOid.ownerId == RL_LIBRARY_SEQNUM) &&
	    (nodeOid.objectSequenceNumber == RL_SHARED_NODE_SEQNUM)) {
		pProps->runningInInitiatorMode = IMA_TRUE;
		pProps->runningInTargetMode = IMA_TRUE;
		pProps->nameAndAliasSettable = IMA_TRUE;

		if (sharedNodeName[0] == 0) {
			if (getSolarisSharedNodeName(sharedNodeName) !=
			    IMA_STATUS_SUCCESS) {
				(void) gethostname((char *)fullline,
					IMA_NODENAME_LENGTH);
				(void) snprintf(nodename, IMA_NODENAME_LENGTH,
				    DEFAULT_NODE_NAME_FORMAT, fullline);
				(void) mbstowcs(sharedNodeName,
				    nodename, IMA_NODENAME_LENGTH);
			}
		}

		if (sharedNodeName[0] != 0) {
			libSwprintf(pProps->name, L"%ls", sharedNodeName);
			pProps->nameValid = IMA_TRUE;
		}
		else
			pProps->nameValid = IMA_FALSE;

		if (sharedNodeAlias[0] == 0) {
			(void) getSolarisSharedNodeAlias(sharedNodeAlias);
		}

		if (sharedNodeAlias[0] != 0) {
			libSwprintf(pProps->alias, L"%ls", sharedNodeAlias);
			pProps->aliasValid = IMA_TRUE;
		}
		else
			pProps->aliasValid = IMA_FALSE;

		return (IMA_STATUS_SUCCESS);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == nodeOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetNodePropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetNodeProperties");

				if (PassFunc != NULL) {
					status = PassFunc(nodeOid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}




/*
 * Sets the name of the specified node.
 *
 * @param nodeId The object ID of the node whose name is being set.
 * @param newName The new name of the node.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the node name was successfully changed.
 * @retval IMA_STATUS_REBOOT_NECESSARY Returned if a reboot is necessary
 *    before the setting of the name actually takes affect.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a newname is NULL, or
 *    specifies a memory area to which data cannot be written, or has a
 *    length of 0.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a nodeId does not
 *    specify any valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if @a nodeId does not
 *    specify a node object.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a nodeId does not specify a
 *    node which is currently known to the system.
 * @retval IMA_ERROR_NAME_TOO_LONG Returned if @a newName contains too many
 *    characters.
 */
IMA_API IMA_STATUS IMA_SetNodeName(
    IMA_OID nodeOid,
    const IMA_NODE_NAME newName) {
	IMA_SetNodeNameFn PassFunc;
	int i;
	IMA_STATUS status;

	if (newName == NULL || wcslen(newName) == 0)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (wcslen(newName) > IMA_NODE_NAME_LEN - 1)
		return (IMA_ERROR_NAME_TOO_LONG);

	if (nodeOid.objectType != IMA_OBJECT_TYPE_NODE)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	if ((nodeOid.ownerId == RL_LIBRARY_SEQNUM) &&
	    (nodeOid.objectSequenceNumber == RL_SHARED_NODE_SEQNUM)) {
		if (setSolarisSharedNodeName(newName) != IMA_STATUS_SUCCESS) {
			return (IMA_ERROR_UNKNOWN_ERROR);
		}
		os_obtainmutex(&libMutex);
		libSwprintf(sharedNodeName, L"%ls", newName);
		os_releasemutex(&libMutex);
		return (IMA_STATUS_SUCCESS);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == nodeOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetNodeNameFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetNodeName");

				if (PassFunc != NULL) {
					status = PassFunc(nodeOid, newName);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}




/*
 * Generates an unique node name for the currently running system.
 *
 * @param generatedname On successful return contains the generated node
 *    name.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a generatedname is NULL
 *    or specifies a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GenerateNodeName(
    IMA_NODE_NAME generatedname) {
	char computername[IMA_NODENAME_LENGTH];
	char nodename[IMA_NODENAME_LENGTH];

	if (generatedname == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (getSolarisSharedNodeName(generatedname) != IMA_STATUS_SUCCESS) {
		(void) gethostname(computername, IMA_NODENAME_LENGTH);
		(void) snprintf(nodename, IMA_NODENAME_LENGTH,
		    DEFAULT_NODE_NAME_FORMAT, computername);
		(void) mbstowcs(generatedname, nodename, IMA_NODENAME_LENGTH);
	}
	return (IMA_STATUS_SUCCESS);
}


/*
 * Sets the alias of the specified node.
 *
 * @param nodeId The object ID of the node whose alias is being set.
 * @param newAlias A pointer to a Unicode string which contains the new node
 *    alias.If this parameter is NULL then the current alias is deleted, in
 *    which case the specified node no longer has an alias.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the node's alias has been successfully set.
 * @retval IMA_STATUS_REBOOT_NECESSARY A reboot is necessary before
 *    the setting of the
 *    alias actually takes affect.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a nodeId does not
 *    specify any valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if @a nodeId does not
 *    specify a node object.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a nodeId does not specify
 *    a node which is currently known to the system.
 * @retval IMA_ERROR_NAME_TOO_LONG Returned if @a newAlias contains too many
 *               characters.
 */
IMA_API IMA_STATUS IMA_SetNodeAlias(
    IMA_OID nodeOid,
    const IMA_NODE_ALIAS newAlias) {
	IMA_SetNodeAliasFn PassFunc;
	int i;
	IMA_STATUS status;

	if (newAlias == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (wcslen(newAlias) > IMA_NODE_ALIAS_LEN - 1)
		return (IMA_ERROR_NAME_TOO_LONG);

	if (nodeOid.objectType != IMA_OBJECT_TYPE_NODE)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	if ((nodeOid.ownerId == RL_LIBRARY_SEQNUM) &&
	    (nodeOid.objectSequenceNumber == RL_SHARED_NODE_SEQNUM)) {
		if (setSolarisSharedNodeAlias(newAlias) != IMA_STATUS_SUCCESS) {
			return (IMA_ERROR_UNKNOWN_ERROR);
		}
		os_obtainmutex(&libMutex);
		if (wcslen(newAlias) > 0 && newAlias != NULL)
			libSwprintf(sharedNodeAlias, L"%ls", newAlias);
		else
			libSwprintf(sharedNodeAlias, L"%ls", L"");

		os_releasemutex(&libMutex);
		return (IMA_STATUS_SUCCESS);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == nodeOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetNodeAliasFn)
				    dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetNodeAlias");

				if (PassFunc != NULL) {
					status = PassFunc(nodeOid, newAlias);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}




/*
 * Gets a list of the object IDs of all the logical HBAs in the system.
 *
 * @param ppList A pointer to a pointer to an @ref IMA_OID_LIST structure.
 *    on successful return this will contain a pointer to an
 *    @ref IMA_OID_LIST which contains the object IDs of all of the
 *    LHBAs currently in the system.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the LHBA ID list has been successfully
 *    returned.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a ppList is NULL or
 *   specifies a
 *   memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetLhbaOidList(
    IMA_OID_LIST **ppList) {
	IMA_GetLhbaOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL) {
		return (IMA_ERROR_INVALID_PARAMETER);
	}

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetLhbaOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetLhbaOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					totalIdCount += ppOidList->oidCount;
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}

			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			break;
		}
	}


	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (totalIdCount - 1) * sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}
	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetLhbaOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetLhbaOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					for (j = 0;
					    (j < ppOidList->oidCount) &&
					    (totalIdCount <
					    (*ppList)->oidCount);
					    j++) {
						(*ppList)->oids[totalIdCount].
						    objectType
						    = ppOidList->oids[j].
						    objectType;
						(*ppList)->oids[totalIdCount].
						    objectSequenceNumber =
						    ppOidList->oids[j].
						    objectSequenceNumber;
						(*ppList)->oids[totalIdCount].
						    ownerId =
						    ppOidList->oids[j].ownerId;
						totalIdCount++;
					}
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			free(*ppList);
			break;
		}

	}
	os_releasemutex(&libMutex);
	return (status);
}




/*
 * Gets the properties of the specified logical HBA.
 *
 * @param lhbaId The object ID of the LHBA whose properties are being
 *    retrieved.
 * @param pProps A pointer to an @ref IMA_LHBA_PROPERTIES structure.
 *    On successful
 *    return this will contain the properties of the LHBA specified by
 *    @a lhbaId.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the properties of the specified LHBA
 *    have been successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a pProps is NULL or
 *    specify a memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if @a lhbaId does not
 *    specify any valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if @a lhbaId does not
 *    specify a LHBA.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a lhbaId does not
 *    specify a LHBA which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetLhbaProperties(
    IMA_OID lhbaId,
    IMA_LHBA_PROPERTIES *pProps) {
	IMA_GetLhbaPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetLhbaPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetLhbaProperties");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}




/*
 * Gets a list of the object IDs of all the physical HBAs in the system.
 *
 * @param ppList A pointer to a pointer to an @ref IMA_OID_LIST structure.
 *    on successful return this will contain a pointer to an
 *    @ref IMA_OID_LIST which contains the object IDs of all of the
 *    PHBAs currently in the system.
 * @return An IMA_STATUS indicating if the operation was successful or if
 *    an error occurred.
 * @retval IMA_SUCCESS Returned if the PHBA ID list has been successfully
 *    returned.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a ppList is NULL or
 *    specify a memory area to which data cannot be written.
 * @retval IMA_SUCCESS Returned if the properties of the specified PHBA
 *    have been successfully retrieved.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a phbaId does not
 *    specify a PHBA which is currently known to the system.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if @a ppList is NULL or
 *    specify a memory area to which data cannot be written.
 */
IMA_API IMA_STATUS IMA_GetPhbaOidList(
    IMA_OID_LIST **ppList) {
	IMA_GetPhbaOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetPhbaOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetPhbaOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					totalIdCount += ppOidList->oidCount;
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			break;
		}

	}


	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (totalIdCount - 1) * sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}

	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetPhbaOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetPhbaOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					for (j = 0;
					    (j < ppOidList->oidCount) &&
					    (totalIdCount <
					    (*ppList)->oidCount);
					    j++) {
						(*ppList)->oids[totalIdCount].
						    objectType =
						    ppOidList->oids[j].
						    objectType;
						(*ppList)->oids[totalIdCount].
						    objectSequenceNumber =
						    ppOidList->oids[j].
						    objectSequenceNumber;
						(*ppList)->oids[totalIdCount].
						    ownerId =
						    ppOidList->oids[j].ownerId;
						totalIdCount++;
					}
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			free(*ppList);
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


/*
 * Gets the general properties of a physical HBA.
 *
 * @param phbaId The object ID of the PHBA whose
 *    properties are being queried.
 * @param pProps A pointer to an @ref
 *    IMA_PHBA_PROPERTIES structure.  On successful
 *    return this will contain the properties of
 *    the PHBA specified by @a phbaId.
 * @return An IMA_STATUS indicating if the
 *    operation was successful or if an error
 *    occurred.
 * @retval IMA_SUCCESS Returned if the properties
 *    of the specified PHBA have been
 *    successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned
 *    if @a pProps is NULL or specifies a
 *    memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned
 *    if @a phbaId does not specify any
 *    valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned
 *    if @a phbaId does not specify a
 *    PHBA.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned
 *    if @a phbaId does not specify a PHBA
 *    which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetPhbaProperties(
    IMA_OID phbaId,
    IMA_PHBA_PROPERTIES *pProps) {
	IMA_GetPhbaPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPhbaPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPhbaProperties");

				if (PassFunc != NULL) {
					status = PassFunc(phbaId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

/*
 * Frees a previously allocated IMA_OID_LIST structure.
 *
 * @param pList A pointer to an @ref IMA_OID_LIST
 *    structure allocated by the
 *    library.  On successful return the memory
 *    allocated by the list is freed.
 * @return An IMA_STATUS indicating if the operation
 *    was successful or if an error occurred.
 * @retval IMA_SUCCESS Returned if the specified object
 *    ID list was successfully freed.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned
 *    if @a pList is NULL or specifies a
 *    memory area from which data cannot be read.
 */
IMA_API IMA_STATUS IMA_FreeMemory(
    void *pMemory) {
	if (pMemory == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);
	free(pMemory);
	return (IMA_STATUS_SUCCESS);
}




IMA_API IMA_STATUS IMA_GetNonSharedNodeOidList(
    IMA_OID_LIST **ppList) {
	IMA_GetNonSharedNodeOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetNonSharedNodeOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetNonSharedNodeOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					totalIdCount += ppOidList->oidCount;
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			break;
		}

	}

	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (totalIdCount - 1) * sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}

	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetNonSharedNodeOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetNonSharedNodeOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					for (j = 0;
					    (j < ppOidList->oidCount) &&
					    (totalIdCount < (
					    *ppList)->oidCount);
					    j++) {
						(*ppList)->oids[
						    totalIdCount].objectType =
						    ppOidList->oids[j].
						    objectType;
						(*ppList)->oids[totalIdCount].
						    objectSequenceNumber =
						    ppOidList->oids[j].
						    objectSequenceNumber;
						(*ppList)->oids[
						    totalIdCount].
						    ownerId =
						    ppOidList->oids[j].
						    ownerId;
						totalIdCount++;
					}
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			free(*ppList);
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}



/*
 * Gets the first burst length properties of
 * the specified logical HBA.
 *
 * @param lhbaId The object ID of the logical HBA
 *    to get the first burst length
 *    properties of.
 * @param pProps A pointer to a min/max values
 *    structure.
 * @return An IMA_STATUS indicating if the operation
 *    was successful or if an error
 *    occurred.
 * @retval IMA_SUCCESS Returned if the first burst
 *    length properties have been
 *    successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned
 *    if @a pProps is NULL or specifies a
 *    memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned
 *    if @a lhbaId does not specify any
 *    valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned
 *    if @a lhbaId does not specify a LHBA.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned
 *    @a lhbaId does not specify a LHBA
 *    which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetFirstBurstLengthProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetFirstBurstLengthPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetFirstBurstLengthPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetFirstBurstLengthProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

/*
 * Gets the max burst length properties of the
 * specified logical HBA.
 *
 * @param lhbaId The object ID of the logical HBA to
 * get the max burst length properties of.
 * @param pProps A pointer to an @ref IMA_MIN_MAX_VALUE
 *    structure allocated by the
 *    caller.  On successful return this structure
 *    will contain the max
 *    burst length properties of this LHBA.
 * @return An IMA_STATUS indicating if the operation
 *    was successful or if an error occurred.
 * @retval IMA_SUCCESS Returned if the max burst
 *    length properties have been
 *    successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned
 *    if @a pProps is NULL or specifies a
 *    memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned
 *    if @a lhbaId does not specify any
 *    valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned
 *    if @a lhbaId does not specify a HBA.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned
 *    if @a lhbaId does not specify a LHBA
 *    which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetMaxBurstLengthProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetMaxBurstLengthPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetMaxBurstLengthPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetMaxBurstLengthProperties");
				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


/*
 * Gets the maximum receive data segment length properties
 * of the specified logical HBA.
 *
 * @param lhbaId The object ID of the logical HBA to
 *    get the max receive data
 *    segment length properties of.
 * @param pProps A pointer to an @ref IMA_MIN_MAX_VALUE
 *    structure allocated by the caller.
 *    On successful return this structure will contain the max
 *    receive data segment length properties of this LHBA.
 * @return An IMA_STATUS indicating if the operation
 *    was successful or if an error occurred.
 * @retval IMA_SUCCESS Returned if the max receive
 *    data segment length properties
 *    have been successfully retrieved.
 * @retval IMA_ERROR_INVALID_PARAMETER Returned if
 *    @a pProps is NULL or specifies a
 *    memory area to which data cannot be written.
 * @retval IMA_ERROR_INVALID_OBJECT_TYPE Returned if
 *    @a lhbaId does not specify any
 *    valid object type.
 * @retval IMA_ERROR_INCORRECT_OBJECT_TYPE Returned if
 *    a lhbaId does not specify a LHBA.
 * @retval IMA_ERROR_OBJECT_NOT_FOUND Returned if @a
 *    lhbaId does not specify a LHBA
 *    which is currently known to the system.
 */
IMA_API IMA_STATUS IMA_GetMaxRecvDataSegmentLengthProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetMaxRecvDataSegmentLengthPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;
#define	IMA_GMRDSLPFN IMA_GetMaxRecvDataSegmentLengthPropertiesFn
#define	IMA_GMRDSLP "IMA_GetMaxRecvDataSegmentLengthProperties"

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GMRDSLPFN)
				    dlsym(plugintable[i].hPlugin,
				    IMA_GMRDSLP);

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
#undef IMA_GMRDSLPFN
#undef IMA_GMRDSLP
	return (status);
}



/* --------------------------------------------- */
IMA_API IMA_STATUS IMA_PluginIOCtl(
    IMA_OID pluginOid,
    IMA_UINT command,
    const void *pInputBuffer,
    IMA_UINT inputBufferLength,
    void *pOutputBuffer,
    IMA_UINT *pOutputBufferLength) {
	IMA_PluginIOCtlFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pInputBuffer == NULL || inputBufferLength == 0 ||
	    pOutputBuffer == NULL || pOutputBufferLength == NULL ||
	    *pOutputBufferLength == 0)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (pluginOid.objectType != IMA_OBJECT_TYPE_PLUGIN)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == pluginOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_PluginIOCtlFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_PluginIOCtl");

				if (PassFunc != NULL) {
					status = PassFunc(
					    pluginOid, command,
					    pInputBuffer, inputBufferLength,
					    pOutputBuffer, pOutputBufferLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}




IMA_API IMA_STATUS IMA_GetNetworkPortalOidList(
    IMA_OID lnpId,
    IMA_OID_LIST **ppList) {
	IMA_GetNetworkPortalOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;
	int i;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lnpId.objectType != IMA_OBJECT_TYPE_LNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lnpId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetNetworkPortalOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetNetworkPortalOidList");

				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					IMA_UINT listSize;
					listSize = sizeof (IMA_OID_LIST);
					status = PassFunc(lnpId, &ppOidList);
					if (IMA_SUCCESS(status)) {

						*ppList = (IMA_OID_LIST*)
						    calloc(1,
						    sizeof (IMA_OID_LIST)
						    + (ppOidList->
						    oidCount - 1)*
						    sizeof (IMA_OID));

						if ((*ppList) == NULL) {
							return (EUOS_ERROR);
						}
						else
							(void) memcpy((*ppList),
							    ppOidList,
							    listSize
							    + (ppOidList->
							    oidCount - 1)*
							    sizeof (IMA_OID));
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetFirstBurstLength(
    IMA_OID lhbaId,
    IMA_UINT firstBurstLength) {
	IMA_SetFirstBurstLengthFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetFirstBurstLengthFn)
				    dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetFirstBurstLength");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, firstBurstLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetMaxBurstLength(
    IMA_OID lhbaId,
    IMA_UINT maxBurstLength) {
	IMA_SetMaxBurstLengthFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetMaxBurstLengthFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetMaxBurstLength");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, maxBurstLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetMaxRecvDataSegmentLength(
    IMA_OID lhbaId,
    IMA_UINT maxRecvDataSegmentLength) {
	IMA_SetMaxRecvDataSegmentLengthFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetMaxRecvDataSegmentLengthFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetMaxRecvDataSegmentLength");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId,
					    maxRecvDataSegmentLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetMaxConnectionsProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetMaxConnectionsPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetMaxConnectionsPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetMaxConnectionsProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetMaxConnections(
    IMA_OID lhbaId,
    IMA_UINT maxConnections) {
	IMA_SetMaxConnectionsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetMaxConnectionsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetMaxConnections");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, maxConnections);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetDefaultTime2RetainProperties(
    IMA_OID lhbaId,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetDefaultTime2RetainPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDefaultTime2RetainPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDefaultTime2RetainProperties");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDefaultTime2Retain(
    IMA_OID lhbaId,
    IMA_UINT defaultTime2Retain) {
	IMA_SetDefaultTime2RetainFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetDefaultTime2RetainFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDefaultTime2Retain");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, defaultTime2Retain);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetDefaultTime2WaitProperties(
    IMA_OID lhbaId,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetDefaultTime2WaitPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDefaultTime2WaitPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDefaultTime2WaitProperties");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDefaultTime2Wait(
    IMA_OID lhbaId,
    IMA_UINT defaultTime2Wait) {
	IMA_SetDefaultTime2WaitFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetDefaultTime2WaitFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDefaultTime2Wait");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, defaultTime2Wait);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetMaxOutstandingR2TProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetMaxOutstandingR2TPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetMaxOutstandingR2TPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetMaxOutstandingR2TProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetMaxOutstandingR2T(
    IMA_OID lhbaId,
    IMA_UINT maxOutstandingR2T) {
	IMA_SetMaxOutstandingR2TFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA &&
	    lhbaId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetMaxOutstandingR2TFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetMaxOutstandingR2T");

				if (PassFunc != NULL) {
					status = PassFunc(
					    lhbaId, maxOutstandingR2T);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetErrorRecoveryLevelProperties(
    IMA_OID oid,
    IMA_MIN_MAX_VALUE *pProps) {
	IMA_GetMaxOutstandingR2TPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetErrorRecoveryLevelPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetErrorRecoveryLevelProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetErrorRecoveryLevel(
    IMA_OID oid,
    IMA_UINT errorRecoveryLevel) {
	IMA_SetErrorRecoveryLevelFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetErrorRecoveryLevelFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetErrorRecoveryLevel");

				if (PassFunc != NULL) {
					status = PassFunc(
					    oid, errorRecoveryLevel);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetInitialR2TProperties(
    IMA_OID oid,
    IMA_BOOL_VALUE *pProps) {
	IMA_GetInitialR2TPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetInitialR2TPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetInitialR2TProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetInitialR2T(
    IMA_OID oid,
    IMA_BOOL initialR2T)
{
	IMA_SetInitialR2TFn PassFunc;
	int i;
	IMA_STATUS status;

	if (initialR2T != IMA_TRUE &&
	    initialR2T != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetInitialR2TFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetInitialR2T");

				if (PassFunc != NULL) {
					status = PassFunc(oid, initialR2T);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetImmediateDataProperties(
    IMA_OID oid,
    IMA_BOOL_VALUE *pProps) {
	IMA_GetImmediateDataPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetImmediateDataPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetImmediateDataProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetImmediateData(
    IMA_OID oid,
    IMA_BOOL immediateData) {
	IMA_SetImmediateDataFn PassFunc;
	int i;
	IMA_STATUS status;

	if (immediateData != IMA_TRUE &&
	    immediateData != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetImmediateDataFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetImmediateData");

				if (PassFunc != NULL) {
					status = PassFunc(oid, immediateData);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetDataPduInOrderProperties(
    IMA_OID oid,
    IMA_BOOL_VALUE *pProps) {
	IMA_GetDataPduInOrderPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDataPduInOrderPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDataPduInOrderProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDataPduInOrder(
    IMA_OID oid,
    IMA_BOOL dataPduInOrder) {
	IMA_SetDataPduInOrderFn PassFunc;
	int i;
	IMA_STATUS status;

	if (dataPduInOrder != IMA_TRUE &&
	    dataPduInOrder != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetDataPduInOrderFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDataPduInOrder");

				if (PassFunc != NULL) {
					status = PassFunc(oid, dataPduInOrder);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetDataSequenceInOrderProperties(
    IMA_OID oid,
    IMA_BOOL_VALUE *pProps) {
	IMA_GetDataSequenceInOrderPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDataSequenceInOrderPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDataSequenceInOrderProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDataSequenceInOrder(
    IMA_OID oid,
    IMA_BOOL dataSequenceInOrder) {
	IMA_SetDataSequenceInOrderFn PassFunc;
	int i;
	IMA_STATUS status;

	if (dataSequenceInOrder != IMA_TRUE &&
	    dataSequenceInOrder != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetDataSequenceInOrderFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDataSequenceInOrder");

				if (PassFunc != NULL) {
					status = PassFunc(
					    oid, dataSequenceInOrder);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetStatisticsCollection(
    IMA_OID oid,
    IMA_BOOL enableStatisticsCollection) {
	IMA_SetStatisticsCollectionFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableStatisticsCollection != IMA_TRUE &&
	    enableStatisticsCollection != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetStatisticsCollectionFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetStatisticsCollection");

				if (PassFunc != NULL) {
					status = PassFunc(
					oid, enableStatisticsCollection);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetNetworkPortStatus(
    IMA_OID portOid,
    IMA_NETWORK_PORT_STATUS *pStatus) {
	IMA_GetNetworkPortStatusFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pStatus == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (portOid.objectType != IMA_OBJECT_TYPE_PNP &&
	    portOid.objectType != IMA_OBJECT_TYPE_LNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == portOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetNetworkPortStatusFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetNetworkPortStatus");

				if (PassFunc != NULL) {
					status = PassFunc(portOid, pStatus);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetTargetOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {
	IMA_GetTargetOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;
	int i;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_LNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetTargetOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetTargetOidList");

				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					IMA_UINT listSize;
					listSize = sizeof (IMA_OID_LIST);
					status = PassFunc(oid, &ppOidList);
					if (IMA_SUCCESS(status)) {
						*ppList =
						    (IMA_OID_LIST*)calloc(1,
						    sizeof (IMA_OID_LIST) +
						    ((ppOidList->oidCount - 1)*
						    sizeof (IMA_OID)));

						if ((*ppList) == NULL) {
							return (EUOS_ERROR);
						}
						else
							(void) memcpy((*ppList),
							    ppOidList, listSize
							    + (ppOidList->
							    oidCount - 1)*
							    sizeof (IMA_OID));
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_RemoveStaleData(
    IMA_OID lhbaId) {
	IMA_RemoveStaleDataFn PassFunc;
	int i;
	IMA_STATUS status;

	if (lhbaId.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_RemoveStaleDataFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RemoveStaleData");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaId);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetIsnsDiscovery(
    IMA_OID phbaId,
    IMA_BOOL enableIsnsDiscovery,
    IMA_ISNS_DISCOVERY_METHOD discoveryMethod,
    const IMA_HOST_ID *iSnsHost) {
	IMA_SetIsnsDiscoveryFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableIsnsDiscovery != IMA_TRUE &&
	    enableIsnsDiscovery != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (enableIsnsDiscovery == IMA_TRUE && iSnsHost == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (discoveryMethod != IMA_ISNS_DISCOVERY_METHOD_STATIC &&
	    discoveryMethod != IMA_ISNS_DISCOVERY_METHOD_DHCP &&
	    discoveryMethod != IMA_ISNS_DISCOVERY_METHOD_SLP)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA &&
	    phbaId.objectType != IMA_OBJECT_TYPE_LHBA) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetIsnsDiscoveryFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetIsnsDiscovery");

				if (PassFunc != NULL) {
					status = PassFunc(phbaId,
					    enableIsnsDiscovery,
					    discoveryMethod, iSnsHost);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetSlpDiscovery(
    IMA_OID phbaId,
    IMA_BOOL enableSlpDiscovery) {
	IMA_SetSlpDiscoveryFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableSlpDiscovery != IMA_TRUE &&
	    enableSlpDiscovery != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA &&
	    phbaId.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetSlpDiscoveryFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetSlpDiscovery");

				if (PassFunc != NULL) {
					status = PassFunc(
					    phbaId,
					    enableSlpDiscovery);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetStaticDiscovery(
    IMA_OID phbaId,
    IMA_BOOL enableStaticDiscovery) {
	IMA_SetStaticDiscoveryFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableStaticDiscovery != IMA_TRUE &&
	    enableStaticDiscovery != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA &&
	    phbaId.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetStaticDiscoveryFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetStaticDiscovery");

				if (PassFunc != NULL) {
					status = PassFunc(
					    phbaId,
					    enableStaticDiscovery);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetSendTargetsDiscovery(
    IMA_OID phbaId,
    IMA_BOOL enableSendTargetsDiscovery) {
	IMA_SetSendTargetsDiscoveryFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableSendTargetsDiscovery != IMA_TRUE &&
	    enableSendTargetsDiscovery != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA &&
	    phbaId.objectType != IMA_OBJECT_TYPE_LHBA) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetSendTargetsDiscoveryFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetSendTargetsDiscovery");

				if (PassFunc != NULL) {
					status = PassFunc(
					    phbaId,
					    enableSendTargetsDiscovery);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

/*
 * this forces plugins to rescan all iscsi targets on this
 * ipaddress/port and return a
 * list of discovered targets.
 * ERROR/todo:
 * according to IMA spec., pTargetOidList is allocated by
 * the caller for library to return data,
 * how does a caller know how much space it will be?
 * pTargetOidList should be allocated by the library/plugin
 * like IMA_GetLnpOidList
 */
IMA_API IMA_STATUS IMA_AddPhbaStaticDiscoveryTarget(
    IMA_OID phbaOid,
    const IMA_TARGET_ADDRESS targetAddress,
    IMA_OID_LIST **pTargetOidList) {
	IMA_AddPhbaStaticDiscoveryTargetFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;
	int i;
	IMA_STATUS status;

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {

		if (plugintable[i].ownerId == phbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_AddPhbaStaticDiscoveryTargetFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_AddPhbaStaticDiscoveryTarget");

				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					IMA_UINT listSize;
					listSize =
					    sizeof (IMA_OID_LIST);
					status = PassFunc(phbaOid,
					    targetAddress, &ppOidList);
					if (IMA_SUCCESS(status)) {

						(*pTargetOidList) =
						    (IMA_OID_LIST*)
						    calloc(1, listSize +
						    (ppOidList->oidCount-1)*
						    sizeof (IMA_OID));

						if ((*pTargetOidList) == NULL) {
							status =
							    EUOS_ERROR;
						}
						(void) memcpy((*pTargetOidList),
						    ppOidList,
						    listSize +
						    (ppOidList->oidCount-1)*
						    sizeof (IMA_OID));
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_RemovePhbaStaticDiscoveryTarget(
    IMA_OID phbaOid,
    IMA_OID targetOid) {
	IMA_RemovePhbaStaticDiscoveryTargetFn PassFunc;
	int i;
	IMA_STATUS status;

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == targetOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_RemovePhbaStaticDiscoveryTargetFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RemovePhbaStaticDiscoveryTarget");

				if (PassFunc != NULL) {
					status = PassFunc(phbaOid, targetOid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetPnpOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {
	IMA_GetPnpOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;
	int i;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_LNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {

		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPnpOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPnpOidList");

				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;

					status = PassFunc(oid, &ppOidList);
					if (IMA_SUCCESS(status)) {
						IMA_UINT listSize;
						listSize =
						    sizeof (IMA_OID_LIST);
						*ppList = (IMA_OID_LIST*)
						    calloc(1, listSize +
						    (ppOidList->oidCount-1)*
						    sizeof (IMA_OID));

						if ((*ppList) == NULL) {
							status =
							    EUOS_ERROR;
						}
						else
							(void) memcpy((*ppList),
							    ppOidList,
							    listSize +
							    (ppOidList->
							    oidCount - 1)*
							    sizeof (IMA_OID));
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetPhbaDownloadProperties(
    IMA_OID phbaId,
    IMA_PHBA_DOWNLOAD_PROPERTIES *pProps) {
	IMA_GetPhbaDownloadPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPhbaDownloadPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPhbaDownloadProperties");

				if (PassFunc != NULL) {
					status = PassFunc(phbaId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_IsPhbaDownloadFile(
    IMA_OID phbaId,
    const IMA_WCHAR *pFileName,
    IMA_PHBA_DOWNLOAD_IMAGE_PROPERTIES *pProps) {
	IMA_IsPhbaDownloadFileFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pFileName == NULL || pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_IsPhbaDownloadFileFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_IsPhbaDownloadFile");

				if (PassFunc != NULL) {
					status = PassFunc(
					    phbaId, pFileName, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_PhbaDownload(
    IMA_OID phbaId,
    IMA_PHBA_DOWNLOAD_IMAGE_TYPE imageType,
    const IMA_WCHAR *pFileName) {
	IMA_PhbaDownloadFn PassFunc;
	int i;
	IMA_STATUS status;

	if (phbaId.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	if (imageType != IMA_DOWNLOAD_IMAGE_TYPE_FIRMWARE &&
	    imageType != IMA_DOWNLOAD_IMAGE_TYPE_OPTION_ROM &&
	    imageType != IMA_DOWNLOAD_IMAGE_TYPE_ALL &&
	    imageType != IMA_DOWNLOAD_IMAGE_TYPE_BOOTCODE)
	    return (IMA_ERROR_INVALID_PARAMETER);

	if (pFileName == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == phbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_PhbaDownloadFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_PhbaDownload");

				if (PassFunc != NULL) {
					status = PassFunc(
					    phbaId, imageType, pFileName);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetNetworkPortalProperties(
    IMA_OID networkPortalId,
    IMA_NETWORK_PORTAL_PROPERTIES *pProps) {
	IMA_GetNetworkPortalPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (networkPortalId.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == networkPortalId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetNetworkPortalPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetNetworkPortalProperties");

				if (PassFunc != NULL) {
					status = PassFunc(
					    networkPortalId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetNetworkPortalIpAddress(
    IMA_OID networkPortalId,
    const IMA_IP_ADDRESS NewIpAddress) {
	IMA_SetNetworkPortalIpAddressFn PassFunc;
	int i;
	IMA_STATUS status;

	if (networkPortalId.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == networkPortalId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetNetworkPortalIpAddressFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetNetworkPortalIpAddress");

				if (PassFunc != NULL) {
					status = PassFunc(
					    networkPortalId, NewIpAddress);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetLnpOidList(
    IMA_OID_LIST **ppList) {
	IMA_GetLnpOidListFn PassFunc;
	IMA_FreeMemoryFn    FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetLnpOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetLnpOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					totalIdCount += ppOidList->oidCount;
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			break;
		}

	}


	*ppList = (IMA_OID_LIST*)calloc(1,
	    sizeof (IMA_OID_LIST) + (totalIdCount - 1)* sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}

	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			PassFunc = (IMA_GetLnpOidListFn)
			    dlsym(plugintable[i].hPlugin,
			    "IMA_GetLnpOidList");
			if (PassFunc != NULL) {
				IMA_OID_LIST *ppOidList;
				status = PassFunc(&ppOidList);
				if (status == IMA_STATUS_SUCCESS) {
					for (j = 0; (j < ppOidList->oidCount) &&
					    (totalIdCount <
					    (*ppList)->oidCount);
					    j++) {
						(*ppList)->oids[totalIdCount].
						    objectType =
						    ppOidList->oids[j].
						    objectType;
						(*ppList)->oids[totalIdCount].
						    objectSequenceNumber =
						    ppOidList->oids[j].
						    objectSequenceNumber;

						(*ppList)->oids[totalIdCount].
						    ownerId =
						    ppOidList->oids[j].ownerId;
						totalIdCount++;
					}
					FreeFunc = (IMA_FreeMemoryFn)
					    dlsym(plugintable[i].hPlugin,
					    "IMA_FreeMemory");
					if (FreeFunc != NULL) {
						FreeFunc(ppOidList);
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS) {
			free(*ppList);
			break;
		}

	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetLnpProperties(
    IMA_OID lnpId,
    IMA_LNP_PROPERTIES *pProps) {
	IMA_GetLnpPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lnpId.objectType != IMA_OBJECT_TYPE_LNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lnpId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetLnpPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetLnpProperties");

				if (PassFunc != NULL) {
					status = PassFunc(lnpId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetPnpProperties(
    IMA_OID pnpId,
    IMA_PNP_PROPERTIES *pProps) {
	IMA_GetPnpPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (pnpId.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == pnpId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPnpPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPnpProperties");

				if (PassFunc != NULL) {
					status = PassFunc(pnpId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetPnpStatistics(
    IMA_OID pnpId,
    IMA_PNP_STATISTICS *pStats) {
	IMA_GetPnpStatisticsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pStats == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (pnpId.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == pnpId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPnpStatisticsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPnpStatistics");

				if (PassFunc != NULL) {
					status = PassFunc(pnpId, pStats);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetTargetProperties(
    IMA_OID targetId,
    IMA_TARGET_PROPERTIES *pProps) {
	IMA_GetTargetPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (targetId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == targetId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetTargetPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetTargetProperties");

				if (PassFunc != NULL) {
					status = PassFunc(targetId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API	IMA_STATUS IMA_GetSessionProperties(
    IMA_OID sessionId,
    IMA_SESSION_PROPERTIES *pProps) {
	IMA_GetSessionPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (sessionId.objectType != IMA_OBJECT_TYPE_SESSION &&
	    sessionId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == sessionId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetSessionPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetSessionProperties");

				if (PassFunc != NULL) {
					status = PassFunc(sessionId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API	IMA_STATUS IMA_GetConnectionProperties(
    IMA_OID connectionId,
    IMA_CONNECTION_PROPERTIES *pProps) {
	IMA_GetConnectionPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (connectionId.objectType != IMA_OBJECT_TYPE_CONNECTION)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == connectionId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetConnectionPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetConnectionProperties");

				if (PassFunc != NULL) {
					status = PassFunc(connectionId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetTargetErrorStatistics(
    IMA_OID targetId,
    IMA_TARGET_ERROR_STATISTICS *pStats) {
	IMA_GetTargetErrorStatisticsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pStats == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (targetId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == targetId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetTargetErrorStatisticsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetTargetErrorStatistics");

				if (PassFunc != NULL) {
					status = PassFunc(targetId, pStats);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetLuOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {
	IMA_GetLuOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;
	int i;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {

		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetLuOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetLuOidList");

				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;

					status = PassFunc(oid, &ppOidList);
					if (IMA_SUCCESS(status)) {
						IMA_UINT listSize;
						listSize =
						    sizeof (IMA_OID_LIST);
						*ppList = (IMA_OID_LIST*)
						    calloc(1, listSize +
						    (ppOidList->oidCount - 1)*
						    sizeof (IMA_OID));

						if ((*ppList) == NULL) {
							status = EUOS_ERROR;
						}
						else
							(void) memcpy((*ppList),
							    ppOidList,
							    listSize +
							    (ppOidList->
							    oidCount - 1)*
							    sizeof (IMA_OID));
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetLuOid(
    IMA_OID targetId,
    IMA_UINT64 lun,
    IMA_OID *pluId) {
	IMA_GetLuOidFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pluId == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);


	if (targetId.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
		status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == targetId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(
				    &(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetLuOidFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetLuOid");

				if (PassFunc != NULL) {
					status =
					    PassFunc(targetId, lun, pluId);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetLuProperties(
    IMA_OID luId,
    IMA_LU_PROPERTIES *pProps) {
	IMA_GetLuPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (luId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == luId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetLuPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetLuProperties");

				if (PassFunc != NULL) {
					status = PassFunc(luId, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetStatisticsProperties(
    IMA_OID oid,
    IMA_STATISTICS_PROPERTIES *pProps) {
	IMA_GetStatisticsPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_LU &&
	    oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);


	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetStatisticsPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetStatisticsProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetDeviceStatistics(
    IMA_OID oid,
    IMA_DEVICE_STATISTICS *pStats) {
	IMA_GetDeviceStatisticsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pStats == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LU &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDeviceStatisticsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDeviceStatistics");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pStats);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_LuInquiry(
    IMA_OID deviceId,
    IMA_BOOL evpd,
    IMA_BOOL cmddt,
    IMA_BYTE pageCode,

    IMA_BYTE *pOutputBuffer,
    IMA_UINT *pOutputBufferLength,

    IMA_BYTE *pSenseBuffer,
    IMA_UINT *pSenseBufferLength) {
	IMA_LuInquiryFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pOutputBuffer == NULL || pOutputBufferLength == NULL ||
	    *pOutputBufferLength == 0 ||
	    (pSenseBuffer == NULL && pSenseBufferLength != NULL &&
	    *pSenseBufferLength != 0))
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((evpd != IMA_TRUE && evpd != IMA_FALSE) ||
	    (cmddt != IMA_TRUE && cmddt != IMA_FALSE))
		return (IMA_ERROR_INVALID_PARAMETER);

	if (deviceId.objectType != IMA_OBJECT_TYPE_TARGET &&
	    deviceId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == deviceId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_LuInquiryFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_LuInquiry");

				if (PassFunc != NULL) {
					status =
					    PassFunc(deviceId, evpd,
					    cmddt, pageCode,
					    pOutputBuffer, pOutputBufferLength,
					    pSenseBuffer, pSenseBufferLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_LuReadCapacity(
    IMA_OID deviceId,
    IMA_UINT cdbLength,
    IMA_BYTE *pOutputBuffer,
    IMA_UINT *pOutputBufferLength,

    IMA_BYTE *pSenseBuffer,
    IMA_UINT *pSenseBufferLength) {
	IMA_LuReadCapacityFn PassFunc;
	int i;
	IMA_STATUS status;

	if (cdbLength != 10 && cdbLength != 16)
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((pOutputBuffer == NULL || pOutputBufferLength == NULL ||
	    *pOutputBufferLength == 0) ||
	    (pSenseBuffer == NULL && pSenseBufferLength != NULL &&
	    *pSenseBufferLength != 0))
		return (IMA_ERROR_INVALID_PARAMETER);

	if (deviceId.objectType != IMA_OBJECT_TYPE_TARGET &&
	    deviceId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == deviceId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_LuReadCapacityFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_LuReadCapacity");

				if (PassFunc != NULL) {
					status = PassFunc(deviceId, cdbLength,
					    pOutputBuffer, pOutputBufferLength,
					    pSenseBuffer, pSenseBufferLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_LuReportLuns(
    IMA_OID deviceId,
    IMA_BOOL sendToWellKnownLun,
    IMA_BYTE selectReport,

    IMA_BYTE *pOutputBuffer,
    IMA_UINT *pOutputBufferLength,

    IMA_BYTE *pSenseBuffer,
    IMA_UINT *pSenseBufferLength) {
	IMA_LuReportLunsFn PassFunc;
	int i;
	IMA_STATUS status;

	if ((pOutputBuffer == NULL || pOutputBufferLength == NULL ||
	    *pOutputBufferLength == 0) ||
	    (pSenseBuffer == NULL && pSenseBufferLength != NULL &&
	    *pSenseBufferLength != 0))
		return (IMA_ERROR_INVALID_PARAMETER);

	if (sendToWellKnownLun != IMA_TRUE && sendToWellKnownLun != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (deviceId.objectType != IMA_OBJECT_TYPE_TARGET &&
	    deviceId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == deviceId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_LuReportLunsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_LuReportLuns");

				if (PassFunc != NULL) {
					status = PassFunc(deviceId,
					    sendToWellKnownLun, selectReport,
					    pOutputBuffer, pOutputBufferLength,
					    pSenseBuffer, pSenseBufferLength);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_ExposeLu(
    IMA_OID luId) {
	IMA_ExposeLuFn PassFunc;
	int i;
	IMA_STATUS status;

	if (luId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INVALID_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == luId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_ExposeLuFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_ExposeLu");

				if (PassFunc != NULL) {
					status = PassFunc(luId);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_UnexposeLu(
    IMA_OID luId) {
	IMA_UnexposeLuFn PassFunc;
	int i;
	IMA_STATUS status;

	if (luId.objectType != IMA_OBJECT_TYPE_LU)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == luId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_UnexposeLuFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_UnexposeLu");

				if (PassFunc != NULL) {
					status = PassFunc(luId);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetPhbaStatus(
    IMA_OID hbaId,
    IMA_PHBA_STATUS *pStatus) {
	IMA_GetPhbaStatusFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pStatus == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (hbaId.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == hbaId.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetPhbaStatusFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetPhbaStatus");

				if (PassFunc != NULL) {
					status = PassFunc(hbaId, pStatus);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_RegisterForObjectVisibilityChanges(
    IMA_OBJECT_VISIBILITY_FN pClientFn) {
	IMA_RegisterForObjectVisibilityChangesFn PassFunc;
	int i;
	IMA_UINT j;
	IMA_STATUS status;

	if (pClientFn == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);

	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			if (plugintable[i].number_of_vbcallbacks >=
			    IMA_MAX_CALLBACK_PER_PLUGIN) {
				os_releasemutex(&(plugintable[i].pluginMutex));
				continue;
			}

			/* check if registered already */
			for (j = 0;
			    j < plugintable[i].number_of_vbcallbacks; j++) {
				if (plugintable[i].vbcallback[j] == pClientFn) {
					status = IMA_STATUS_SUCCESS;
					break;
				}
			}
			if (status != IMA_STATUS_SUCCESS) {

				PassFunc =
				    (IMA_RegisterForObjectVisibilityChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RegisterForObjectVisibilityChanges");

				if (PassFunc != NULL) {
					status = PassFunc(VisibilityCallback);
					if (status == IMA_STATUS_SUCCESS) {
						j = plugintable[i].
						    number_of_vbcallbacks;
						plugintable[i].vbcallback[j] =
						    pClientFn;
						plugintable[i].
						    number_of_vbcallbacks++;
					}

				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS)
			break;

	}
	os_releasemutex(&libMutex);
	return (status);

}


IMA_API IMA_STATUS IMA_DeregisterForObjectVisibilityChanges(
    IMA_OBJECT_VISIBILITY_FN pClientFn) {
	IMA_DeregisterForObjectVisibilityChangesFn PassFunc;
	int i;
	IMA_UINT j;
	IMA_STATUS status;

	if (pClientFn == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);

	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			/* check if deregistered already */
			status = IMA_STATUS_SUCCESS;
			for (j = 0;
			    j < plugintable[i].number_of_vbcallbacks; j++) {
				if (plugintable[i].vbcallback[j] == pClientFn) {
					/*
					 * use IMA_ERROR_UNKNOWN_ERROR
					 * as a flag
					 */
					status = IMA_ERROR_UNKNOWN_ERROR;
					break;
				}
			}

			if (status != IMA_STATUS_SUCCESS) {

				PassFunc =
				    (IMA_DeregisterForObjectVisibilityChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_DeregisterForObjectVisibilityChanges");
				if (PassFunc != NULL) {
					status = PassFunc(VisibilityCallback);
					if (status == IMA_STATUS_SUCCESS) {
						/*
						 * where plugintable[i].
						 * vbcallback[j] == pClientFn
						 */
						for (; j <
						    plugintable[i].
						    number_of_vbcallbacks;
						    j++) {
							plugintable[i].
							    vbcallback[j] =
							    plugintable[i].
							    vbcallback[j+1];

						}
						plugintable[i].
						    number_of_vbcallbacks--;
					}
				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS)
			break;
	}
	os_releasemutex(&libMutex);
	return (status);

}


IMA_API IMA_STATUS IMA_RegisterForObjectPropertyChanges(
    IMA_OBJECT_PROPERTY_FN pClientFn) {
	IMA_RegisterForObjectPropertyChangesFn PassFunc;
	int i;
	IMA_UINT j;
	IMA_STATUS status;

	if (pClientFn == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);

	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			if (plugintable[i].number_of_pccallbacks >=
			    IMA_MAX_CALLBACK_PER_PLUGIN) {
				os_releasemutex(&(plugintable[i].pluginMutex));
				continue;
			}

			/* check if registered already */
			for (j = 0;
			    j < plugintable[i].number_of_pccallbacks;
			    j++) {
				if (plugintable[i].pccallback[j] ==
				    pClientFn) {
					status = IMA_STATUS_SUCCESS;
					break;
				}
			}
			if (status != IMA_STATUS_SUCCESS) {

				PassFunc =
				    (IMA_RegisterForObjectPropertyChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RegisterForObjectPropertyChanges");

				if (PassFunc != NULL) {
					status = PassFunc(PropertyCallback);
					if (status == IMA_STATUS_SUCCESS) {
						j = plugintable[i].
						    number_of_pccallbacks;
						plugintable[i].pccallback[j] =
						    pClientFn;
						plugintable[i].
						    number_of_pccallbacks++;
					}

				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS)
			break;

	}
	os_releasemutex(&libMutex);
	return (status);

}


IMA_API IMA_STATUS IMA_DeregisterForObjectPropertyChanges(
    IMA_OBJECT_PROPERTY_FN pClientFn) {
	IMA_DeregisterForObjectPropertyChangesFn PassFunc;
	int i;
	IMA_UINT j;
	IMA_STATUS status;

	if (pClientFn == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_STATUS_SUCCESS;
	for (i = 0; i < number_of_plugins; i++) {
		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin != NULL) {
			os_obtainmutex(&(plugintable[i].pluginMutex));
			/* check if deregistered already */
			status = IMA_STATUS_SUCCESS;
			for (j = 0;
			    j < plugintable[i].number_of_pccallbacks;
			    j++) {
				if (plugintable[i].pccallback[j] ==
				    pClientFn) {
					/*
					 * use IMA_ERROR_UNKNOWN_ERROR
					 * as a flag
					 */
					status = IMA_ERROR_UNKNOWN_ERROR;
					break;
				}
			}

			if (status != IMA_STATUS_SUCCESS) {

				PassFunc =
				    (IMA_DeregisterForObjectPropertyChangesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_DeregisterForObjectPropertyChanges");

				if (PassFunc != NULL) {
					status = PassFunc(PropertyCallback);
					if (status == IMA_STATUS_SUCCESS) {
					/*
					 * where plugintable[i].vbcallback[
					 * j] == pClientFn
					 */
						for (; j < plugintable[i].
						    number_of_pccallbacks;
						    j++) {
							plugintable[i].
							    pccallback[j]
							    = plugintable[i].
							    pccallback[j+1];

						}
						plugintable[i].
						    number_of_pccallbacks--;
					}

				}
			}
			os_releasemutex(&(plugintable[i].pluginMutex));
		}
		if (status != IMA_STATUS_SUCCESS)
			break;

	}
	os_releasemutex(&libMutex);
	return (status);

}


IMA_API IMA_STATUS IMA_GetIpProperties(
    IMA_OID oid,
    IMA_IP_PROPERTIES *pProps) {
	IMA_GetIpPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetIpPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetIpProperties");
				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetIpConfigMethod(
    IMA_OID oid,
    IMA_BOOL enableDhcpIpConfiguration) {
	IMA_SetIpConfigMethodFn PassFunc;
	int i;
	IMA_STATUS status;

	if (enableDhcpIpConfiguration != IMA_TRUE &&
	    enableDhcpIpConfiguration != IMA_FALSE)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetIpConfigMethodFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetIpConfigMethod");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    enableDhcpIpConfiguration);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetSubnetMask(
    IMA_OID oid,
    IMA_IP_ADDRESS subnetMask) {
	IMA_SetSubnetMaskFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetSubnetMaskFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetSubnetMask");

				if (PassFunc != NULL) {
					status = PassFunc(oid, subnetMask);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDnsServerAddress(
    IMA_OID oid,
    const IMA_IP_ADDRESS *primaryDnsServerAddress,
    const IMA_IP_ADDRESS *alternateDnsServerAddress) {
	IMA_SetDnsServerAddressFn PassFunc;
	int i;
	IMA_STATUS status;

	if (primaryDnsServerAddress == NULL &&
	    alternateDnsServerAddress != NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (primaryDnsServerAddress != NULL &&
	    alternateDnsServerAddress != NULL &&
	    memcmp(primaryDnsServerAddress->ipAddress,
	    alternateDnsServerAddress->ipAddress,
	    sizeof (primaryDnsServerAddress->ipAddress)) == 0)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetDnsServerAddressFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDnsServerAddress");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    primaryDnsServerAddress,
					    alternateDnsServerAddress);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_SetDefaultGateway(
    IMA_OID oid,
    IMA_IP_ADDRESS defaultGateway) {
	IMA_SetDefaultGatewayFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetDefaultGatewayFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetDefaultGateway");

				if (PassFunc != NULL) {
					status = PassFunc(oid, defaultGateway);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetSupportedAuthMethods(
    IMA_OID lhbaOid,
    IMA_BOOL getSettableMethods,
    IMA_UINT *pMethodCount,
    IMA_AUTHMETHOD *pMethodList) {
	IMA_GetSupportedAuthMethodsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pMethodCount == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaOid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetSupportedAuthMethodsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetSupportedAuthMethods");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaOid,
					    getSettableMethods,
					    pMethodCount, pMethodList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetInUseInitiatorAuthMethods(
    IMA_OID lhbaOid,
    IMA_UINT *pMethodCount,
    IMA_AUTHMETHOD *pMethodList) {
	IMA_GetInUseInitiatorAuthMethodsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pMethodCount == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaOid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetInUseInitiatorAuthMethodsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetInUseInitiatorAuthMethods");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaOid,
					    pMethodCount, pMethodList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetInitiatorAuthParms(
    IMA_OID lhbaOid,
    IMA_AUTHMETHOD method,
    IMA_INITIATOR_AUTHPARMS *pParms) {
	IMA_GetInitiatorAuthParmsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pParms == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaOid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	if (method != IMA_AUTHMETHOD_NONE &&
	    method != IMA_AUTHMETHOD_CHAP &&
	    method != IMA_AUTHMETHOD_SRP &&
	    method != IMA_AUTHMETHOD_KRB5 &&
	    method != IMA_AUTHMETHOD_SPKM1 &&
	    method != IMA_AUTHMETHOD_SPKM2)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetInitiatorAuthParmsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetInitiatorAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaOid,
					    method, pParms);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetInitiatorAuthMethods(
    IMA_OID lhbaOid,
    IMA_UINT methodCount,
    const IMA_AUTHMETHOD *pMethodList) {
	IMA_SetInitiatorAuthMethodsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (methodCount == 0 || pMethodList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaOid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetInitiatorAuthMethodsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetInitiatorAuthMethods");

				if (PassFunc != NULL) {
					status = PassFunc(lhbaOid,
					    methodCount, pMethodList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetInitiatorAuthParms(
    IMA_OID lhbaOid,
    IMA_AUTHMETHOD method,
    const IMA_INITIATOR_AUTHPARMS *pParms) {

	IMA_SetInitiatorAuthParmsFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pParms == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (method != IMA_AUTHMETHOD_NONE &&
	    method != IMA_AUTHMETHOD_CHAP &&
	    method != IMA_AUTHMETHOD_SRP &&
	    method != IMA_AUTHMETHOD_KRB5 &&
	    method != IMA_AUTHMETHOD_SPKM1 &&
	    method != IMA_AUTHMETHOD_SPKM2)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (lhbaOid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == lhbaOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_SetInitiatorAuthParmsFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_SetInitiatorAuthParms");

				if (PassFunc != NULL) {
					status =
					    PassFunc(
					    lhbaOid, method, pParms);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetStaticDiscoveryTargetOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {
	IMA_GetStaticDiscoveryTargetOidListFn PassFunc;
	int i;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetStaticDiscoveryTargetOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetStaticDiscoveryTargetOidList");
				if (PassFunc != NULL) {
					status = PassFunc(oid, ppList);
				}

				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetDiscoveryProperties(
    IMA_OID oid,
    IMA_DISCOVERY_PROPERTIES *pProps) {
	IMA_GetDiscoveryPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetDiscoveryPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDiscoveryProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_AddDiscoveryAddress(
    IMA_OID oid,
    const IMA_TARGET_ADDRESS discoveryAddress,
    IMA_OID *pDiscoveryAddressOid) {
	IMA_AddDiscoveryAddressFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_AddDiscoveryAddressFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_AddDiscoveryAddress");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    discoveryAddress,
					    pDiscoveryAddressOid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_AddStaticDiscoveryTarget(
    IMA_OID oid,
    const IMA_STATIC_DISCOVERY_TARGET staticDiscoveryTarget,
    IMA_OID *pStaticDiscoveryTargetOid) {
	IMA_AddStaticDiscoveryTargetFn PassFunc;
	int i;
	IMA_STATUS status;
	IMA_STATIC_DISCOVERY_TARGET emptyTarget;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PNP)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	(void) memset(&emptyTarget, 0, sizeof (IMA_STATIC_DISCOVERY_TARGET));
	if (memcmp(&staticDiscoveryTarget, &emptyTarget,
	    sizeof (IMA_STATIC_DISCOVERY_TARGET)) == 0)
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_AddStaticDiscoveryTargetFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_AddStaticDiscoveryTarget");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    staticDiscoveryTarget,
					    pStaticDiscoveryTargetOid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_CommitHbaParameters(IMA_OID oid,
    IMA_COMMIT_LEVEL commitLevel)
{
	IMA_CommitHbaParametersFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_CommitHbaParametersFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_CommitHbaParameters");

				if (PassFunc != NULL) {
					status = PassFunc(oid, commitLevel);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_RemoveStaticDiscoveryTarget(
    IMA_OID oid) {
	IMA_RemoveStaticDiscoveryTargetFn PassFunc;
	int i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_RemoveStaticDiscoveryTargetFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RemoveStaticDiscoveryTarget");

				if (PassFunc != NULL) {
					status = PassFunc(oid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetStaticDiscoveryTargetProperties(
    IMA_OID staticDiscoveryTargetOid,
    IMA_STATIC_DISCOVERY_TARGET_PROPERTIES *pProps) {
	IMA_GetStaticDiscoveryTargetPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (staticDiscoveryTargetOid.objectType !=
	    IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId ==
		    staticDiscoveryTargetOid.ownerId) {

			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetStaticDiscoveryTargetPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetStaticDiscoveryTargetProperties");

				if (PassFunc != NULL) {
					status = PassFunc(
					    staticDiscoveryTargetOid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetDiscoveryAddressOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {

	IMA_GetDiscoveryAddressOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((oid.objectType != IMA_OBJECT_TYPE_LHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_PNP)) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetDiscoveryAddressOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDiscoveryAddressOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(oid, &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						totalIdCount +=
						    ppOidList->oidCount;
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				break;
			}
		}
	}

	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (totalIdCount - 1)* sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}
	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetDiscoveryAddressOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetDiscoveryAddressOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(oid, &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						for (j = 0;
						    (j < ppOidList->oidCount) &&
						    (totalIdCount <
						    (*ppList)->oidCount);
						    j++) {
#define	OBJ_SEQ_NUM ppOidList->oids[j].objectSequenceNumber
							(*ppList)->oids
							    [totalIdCount].
							    objectType =
							    ppOidList->oids[j].
							    objectType;
							(*ppList)->oids[
							    totalIdCount].
							    objectSequenceNumber
							    = OBJ_SEQ_NUM;
							(*ppList)->oids[
							    totalIdCount].
							    ownerId =
							    ppOidList->
							    oids[j].ownerId;
							totalIdCount++;
#undef OBJ_SEQ_NUM
						}
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				free(*ppList);
				break;
			}
		}
	}

	os_releasemutex(&libMutex);
	return (status);

}

IMA_API IMA_STATUS IMA_GetSessionOidList(
    IMA_OID oid,
    IMA_OID_LIST **ppList) {

	IMA_GetSessionOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((oid.objectType != IMA_OBJECT_TYPE_LHBA) &&
	    (oid.objectType != IMA_OBJECT_TYPE_TARGET)) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetSessionOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetSessionOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(oid, &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						totalIdCount +=
						    ppOidList->oidCount;
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}

				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				break;
			}
		}
	}

	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST) +
	    (totalIdCount - 1)* sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}
	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetSessionOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetSessionOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(oid, &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						for (j = 0;
						    (j < ppOidList->oidCount) &&
						    (totalIdCount <
						    (*ppList)->oidCount);
						    j++) {

#define	OBJ_SEQ_NUM ppOidList->oids[j].objectSequenceNumber
							(*ppList)->oids[
							    totalIdCount].
							    objectType =
							    ppOidList->oids[j].
							    objectType;
							(*ppList)->oids[
							    totalIdCount].
							    objectSequenceNumber
							    = OBJ_SEQ_NUM;
							(*ppList)->oids[
							    totalIdCount].
							    ownerId =
							    ppOidList->oids[j].
							    ownerId;
							totalIdCount++;
#undef OBJ_SEQ_NUM
						}
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				free(*ppList);
				break;
			}
		}
	}

	os_releasemutex(&libMutex);
	return (status);

}

IMA_API IMA_STATUS IMA_GetConnectionOidList(
    IMA_OID sessionOid,
    IMA_OID_LIST **ppList) {

	IMA_GetConnectionOidListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	int i;
	IMA_UINT j;
	IMA_UINT totalIdCount;
	IMA_STATUS status;

	if (ppList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (sessionOid.objectType != IMA_OBJECT_TYPE_SESSION &&
	    sessionOid.objectType != IMA_OBJECT_TYPE_TARGET) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	/* Get total id count first */
	totalIdCount = 0;

	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == sessionOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetConnectionOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetConnectionOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(sessionOid,
					    &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						totalIdCount +=
						    ppOidList->oidCount;
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}

				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				break;
			}
		}
	}


	*ppList = (IMA_OID_LIST*)calloc(1, sizeof (IMA_OID_LIST)
	    + (totalIdCount - 1)* sizeof (IMA_OID));

	if ((*ppList) == NULL) {
		os_releasemutex(&libMutex);
		return (IMA_ERROR_UNEXPECTED_OS_ERROR);
	}
	(*ppList)->oidCount = totalIdCount;

	/* 2nd pass to copy the id lists */
	totalIdCount = 0;
	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == sessionOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetConnectionOidListFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetConnectionOidList");
				if (PassFunc != NULL) {
					IMA_OID_LIST *ppOidList;
					status = PassFunc(sessionOid,
					    &ppOidList);
					if (status == IMA_STATUS_SUCCESS) {
						for (j = 0; (
						    j < ppOidList->oidCount) &&
						    (totalIdCount <
						    (*ppList)->oidCount);
						    j++) {
#define	OBJ_SEQ_NUM ppOidList->oids[j].objectSequenceNumber
							(*ppList)->
							    oids[totalIdCount].
							    objectType =
							    ppOidList->
							    oids[j].objectType;
							(*ppList)->
							    oids[totalIdCount].
							    objectSequenceNumber
							    = OBJ_SEQ_NUM;
							(*ppList)->
							    oids[totalIdCount].
							    ownerId =
							    ppOidList->oids[j].
							    ownerId;
							totalIdCount++;
#undef OBJ_SEQ_NUM
						}
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppOidList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			if (status != IMA_STATUS_SUCCESS) {
				free(*ppList);
				break;
			}
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}

IMA_API IMA_STATUS IMA_RemoveDiscoveryAddress(
    IMA_OID discoveryAddressOid) {

	IMA_RemoveDiscoveryAddressFn PassFunc;
	int i;
	IMA_STATUS status;

	if (discoveryAddressOid.objectType !=
	    IMA_OBJECT_TYPE_DISCOVERY_ADDRESS) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == discoveryAddressOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_RemoveDiscoveryAddressFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_RemoveDiscoveryAddress");

				if (PassFunc != NULL) {
					status = PassFunc(discoveryAddressOid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetIpsecProperties(
    IMA_OID oid,
    IMA_IPSEC_PROPERTIES *pProps) {
	IMA_GetIpsecPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_PNP &&
	    oid.objectType != IMA_OBJECT_TYPE_LHBA) {
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);
	}

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetIpsecPropertiesFn)
				    dlsym(plugintable[i].hPlugin,
				    "IMA_GetIpsecProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetAddressKeys(
    IMA_OID targetOid,
    IMA_ADDRESS_KEYS **ppKeys) {
	IMA_GetAddressKeysFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	IMA_STATUS status;
	int i;

	if (ppKeys == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (targetOid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);

	status = IMA_ERROR_OBJECT_NOT_FOUND;
	for (i = 0; i < number_of_plugins; i++) {

		if (plugintable[i].ownerId == targetOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc = (IMA_GetAddressKeysFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetAddressKeys");

				if (PassFunc != NULL) {
					IMA_ADDRESS_KEYS *ppKeysList;
					IMA_UINT addrSize;
					addrSize = sizeof (IMA_ADDRESS_KEYS);
					status =
					    PassFunc(targetOid, &ppKeysList);
					if (IMA_SUCCESS(status)) {

						*ppKeys =
						    (IMA_ADDRESS_KEYS*)calloc(1,
						    addrSize +
						    (ppKeysList->addressKeyCount
						    - 1) * addrSize);
						if ((*ppKeys) == NULL) {
							status = EUOS_ERROR;
						} else {
							(void) memcpy((*ppKeys),
							    ppKeysList,
							    addrSize +
							    (ppKeysList->
							    addressKeyCount-1)*
							    addrSize);

						}
						FreeFunc = (IMA_FreeMemoryFn)
						    dlsym(
						    plugintable[i].hPlugin,
						    "IMA_FreeMemory");
						if (FreeFunc != NULL) {
							FreeFunc(ppKeysList);
						}
					}
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetDiscoveryAddressProperties(
    IMA_OID oid,
    IMA_DISCOVERY_ADDRESS_PROPERTIES *pProps) {

	IMA_GetDiscoveryAddressPropertiesFn PassFunc;
	int i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDiscoveryAddressPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetDiscoveryAddressProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS QIMA_SetUpdateInterval(
    IMA_OID pluginOid, time_t interval) {
	QIMA_SetUpdateIntervalFn updFunc;
	int i;
	IMA_STATUS status;

	if (interval <= 1)
		return (IMA_ERROR_INVALID_PARAMETER);

	if ((pluginOid.objectType != IMA_OBJECT_TYPE_PLUGIN) ||
	    (pluginOid.objectSequenceNumber != 0))
		return (IMA_ERROR_INVALID_PARAMETER);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == pluginOid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				updFunc = (QIMA_SetUpdateIntervalFn) dlsym(
				    plugintable[i].hPlugin,
				    "QIMA_SetUpdateInterval");

				if (updFunc != NULL) {
					status = updFunc(pluginOid, interval);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}

IMA_API IMA_STATUS IMA_GetDigestProperties(
		IMA_OID oid,
		IMA_DIGEST_PROPERTIES *pProps)
{
	IMA_GetDigestPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDigestPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetDigestProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}

IMA_API IMA_STATUS IMA_GetDataDigestValues(
		IMA_OID oid,
		IMA_UINT *pDigestValueCount,
		IMA_DIGEST_TYPE *pDigestValueList)
{
	IMA_GetDataDigestValuesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetDataDigestValuesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetDataDigestValues");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    pDigestValueCount,
					    pDigestValueList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetDataDigestValues(
		IMA_OID oid,
		IMA_UINT digestValueCount,
		const IMA_DIGEST_TYPE *pDigestValueList)
{
	IMA_SetDataDigestValuesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (digestValueCount == 0 ||
	    pDigestValueList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetDataDigestValuesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetDataDigestValues");

				if (PassFunc != NULL) {
					status = PassFunc(oid, digestValueCount,
					    pDigestValueList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetHeaderDigestValues(
		IMA_OID oid,
		IMA_UINT digestValueCount,
		const IMA_DIGEST_TYPE *pDigestValueList)
{
	IMA_SetHeaderDigestValuesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (digestValueCount == 0 ||
	    pDigestValueList == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetHeaderDigestValuesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetHeaderDigestValues");

				if (PassFunc != NULL) {
					status = PassFunc(oid, digestValueCount,
					    pDigestValueList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetHeaderDigestValues(
		IMA_OID oid,
		IMA_UINT *pDigestValueCount,
		IMA_DIGEST_TYPE *pDigestValueList)
{
	IMA_GetHeaderDigestValuesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	/* pDigestValueCount can be NULL */
	if (pDigestValueCount == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetHeaderDigestValuesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetHeaderDigestValues");

				if (PassFunc != NULL) {
					status = PassFunc(oid,
					    pDigestValueCount,
					    pDigestValueList);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetIFMarkerProperties(
		IMA_OID oid,
		IMA_BOOL_VALUE *pProps)
{
	IMA_GetIFMarkerPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetIFMarkerPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetIFMarkerProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetIFMarkerProperties(
		IMA_OID oid,
		IMA_BOOL ifMarker)
{
	IMA_SetIFMarkerPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetIFMarkerPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetIFMarkerProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, ifMarker);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetOFMarkerProperties(
		IMA_OID oid,
		IMA_BOOL_VALUE *pProps)
{
	IMA_GetOFMarkerPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetOFMarkerPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetOFMarkerProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetOFMarkerProperties(
		IMA_OID oid,
		IMA_BOOL ofMarker)
{
	IMA_SetOFMarkerPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetOFMarkerPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetOFMarkerProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, ofMarker);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);

}

IMA_API  IMA_STATUS IMA_GetIFMarkIntProperties(
		IMA_OID oid,
		IMA_MARKER_INT_PROPERTIES *pProps)
{
	IMA_GetIFMarkIntPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetIFMarkIntPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetIFMarkIntProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API  IMA_STATUS IMA_SetIFMarkIntProperties(
		IMA_OID oid,
		IMA_MARKER_INT ifMarkInt)
{
	IMA_SetIFMarkIntPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetIFMarkIntPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetIFMarkIntProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, ifMarkInt);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetOFMarkIntProperties(
		IMA_OID oid,
		IMA_MARKER_INT_PROPERTIES *pProps)
{
	IMA_GetOFMarkIntPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetOFMarkIntPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetOFMarkIntProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, pProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetOFMarkIntProperties(
		IMA_OID oid,
		IMA_MARKER_INT ofMarkInt)
{
	IMA_SetOFMarkIntPropertiesFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetOFMarkIntPropertiesFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetOFMarkIntProperties");

				if (PassFunc != NULL) {
					status = PassFunc(oid, ofMarkInt);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetInitiatorLocalAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_INITIATOR_AUTHPARMS *pParms,
		IMA_BOOL *enabled)
{
	IMA_GetInitiatorLocalAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP ||
	    pParms == NULL || enabled == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetInitiatorLocalAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetInitiatorLocalAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method,
					    pParms, enabled);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetInitiatorLocalAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		const IMA_INITIATOR_AUTHPARMS *pParms,
		IMA_BOOL enabled)
{
	IMA_SetInitiatorLocalAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP || pParms == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetInitiatorLocalAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetInitiatorLocalAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method,
					    pParms, enabled);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetMutualLocalAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_TARGET_AUTHPARMS *pParms)
{
	IMA_GetMutualLocalAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP || pParms == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetMutualLocalAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetMutualLocalAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method, pParms);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetMutualLocalAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		const IMA_TARGET_AUTHPARMS *pParms,
		IMA_BOOL mutualAuthEnabled)
{
	IMA_SetMutualLocalAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP || pParms == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetMutualLocalAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetMutualLocalAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method, pParms,
					    mutualAuthEnabled);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}
			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_AddLhbaMutualAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_TARGET_AUTHPARMS pParms)
{
	IMA_AddLhbaMutualAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (pParms.chapParms.name == NULL || method != IMA_AUTHMETHOD_CHAP)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_AddLhbaMutualAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_AddLhbaMutualAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method, pParms);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_RemoveLhbaMutualAuthParms(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_TARGET_AUTHPARMS pParms)
{
	IMA_RemoveLhbaMutualAuthParmsFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_RemoveLhbaMutualAuthParmsFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_RemoveLhbaMutualAuthParms");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method, pParms);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetLhbaMutualAuthParmsList(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_TARGET_AUTHPARMS_LIST **ppParmsList)
{
	IMA_GetLhbaMutualAuthParmsListFn PassFunc;
	IMA_FreeMemoryFn FreeFunc;

	IMA_STATUS status;
	IMA_UINT i;
	IMA_TARGET_AUTHPARMS_LIST *ppParms;
	IMA_UINT headerSize;
	IMA_UINT entrySize;

	if (ppParmsList == NULL || method != IMA_AUTHMETHOD_CHAP)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId != oid.ownerId)
			continue;

		status = IMA_ERROR_UNEXPECTED_OS_ERROR;
		if (plugintable[i].hPlugin == NULL)
			break;

		os_obtainmutex(&(plugintable[i].pluginMutex));
		PassFunc = (IMA_GetLhbaMutualAuthParmsListFn) dlsym(
		    plugintable[i].hPlugin, "IMA_GetLhbaMutualAuthParmsList");

		if (PassFunc == NULL) {
			os_releasemutex(&(plugintable[i].pluginMutex));
			break;
		}

		headerSize = sizeof (IMA_TARGET_AUTHPARMS_LIST);
		entrySize = sizeof (IMA_TARGET_AUTHPARMS);
		status = PassFunc(oid, method, &ppParms);
		if (IMA_SUCCESS(status)) {
			*ppParmsList = (IMA_TARGET_AUTHPARMS_LIST*)
			    calloc(1, headerSize +
			    (ppParms->authparmsCount - 1) * entrySize);
			if ((*ppParmsList) == NULL) {
				status = EUOS_ERROR;
			} else {
				(void) memcpy((*ppParmsList), ppParms,
				    headerSize + (ppParms->
				    authparmsCount-1)* entrySize);
			}
			FreeFunc = (IMA_FreeMemoryFn)
			    dlsym(plugintable[i].hPlugin, "IMA_FreeMemory");
			if (FreeFunc != NULL) {
				FreeFunc(ppParms);
			}
		}
		os_releasemutex(&(plugintable[i].pluginMutex));
		break;
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_GetMutualLocalAuth(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_BOOL *mutualAuthEnabled)
{
	IMA_GetMutualLocalAuthFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP ||
	    mutualAuthEnabled == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetMutualLocalAuthFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetMutualLocalAuth");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method,
					    mutualAuthEnabled);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetMutualLocalAuth(
		IMA_OID oid,
		IMA_AUTHMETHOD method,
		IMA_BOOL mutualAuthEnabled)
{
	IMA_SetMutualLocalAuthFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (method != IMA_AUTHMETHOD_CHAP)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_TARGET &&
	    oid.objectType != IMA_OBJECT_TYPE_DISCOVERY_ADDRESS &&
	    oid.objectType != IMA_OBJECT_TYPE_STATIC_DISCOVERY_TARGET)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetMutualLocalAuthFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetMutualLocalAuth");

				if (PassFunc != NULL) {
					status = PassFunc(oid, method,
					    mutualAuthEnabled);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}


IMA_API IMA_STATUS IMA_GetRadiusAccess(
		IMA_OID oid,
		IMA_RADIUS_PROPERTIES *radiusProps)
{
	IMA_GetRadiusAccessFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (radiusProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_GetRadiusAccessFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_GetRadiusAccess");

				if (PassFunc != NULL) {
					status = PassFunc(oid, radiusProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_SetRadiusAccess(
		IMA_OID oid,
		IMA_RADIUS_PROPERTIES *radiusProps)
{
	IMA_SetRadiusAccessFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (radiusProps == NULL)
		return (IMA_ERROR_INVALID_PARAMETER);

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_NETWORK_PORTAL)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_SetRadiusAccessFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_SetRadiusAccess");

				if (PassFunc != NULL) {
					status = PassFunc(oid, radiusProps);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}

IMA_API IMA_STATUS IMA_PersistHbaParameters(
		IMA_OID oid)
{
	IMA_PersistHbaParametersFn PassFunc;
	IMA_UINT i;
	IMA_STATUS status;

	if (oid.objectType != IMA_OBJECT_TYPE_LHBA &&
	    oid.objectType != IMA_OBJECT_TYPE_PHBA)
		return (IMA_ERROR_INCORRECT_OBJECT_TYPE);

	os_obtainmutex(&libMutex);
	status = IMA_ERROR_OBJECT_NOT_FOUND;

	for (i = 0; i < number_of_plugins; i++) {
		if (plugintable[i].ownerId == oid.ownerId) {
			status = IMA_ERROR_UNEXPECTED_OS_ERROR;
			if (plugintable[i].hPlugin != NULL) {
				os_obtainmutex(&(plugintable[i].pluginMutex));
				PassFunc =
				    (IMA_PersistHbaParametersFn) dlsym(
				    plugintable[i].hPlugin,
				    "IMA_PersistHbaParameters");

				if (PassFunc != NULL) {
					status = PassFunc(oid);
				}
				os_releasemutex(&(plugintable[i].pluginMutex));
			}

			break;
		}
	}
	os_releasemutex(&libMutex);
	return (status);
}
