#!/usr/bin/python

#
# Copyright (c) 2013, 2021, Oracle and/or its affiliates.
#

# To assist 2->3 transition
from __future__ import print_function

import collections
import copy
import datetime
import email.mime.text
import errno
import glob
import hashlib
import inspect
import itertools
import logging
import multiprocessing
import os
import re
import simplejson
import shutil
import smtplib
import socket
import stat
import sys
import tarfile
import tempfile
import time
import xml.etree.ElementTree
import operator
import zipfile

from six.moves import configparser
from six.moves import urllib

import onbld.log as log
import onbld.util as util
from onbld.log import STATE_LEVEL, SUMMARY_LEVEL, TWARNING_LEVEL, TINFO_LEVEL,\
    CMDLINE_LEVEL, TASK_LEVEL
from onbld.environment import GATEKEEPER_MAILTO, MINIMUM_RECOMMENDED_SWAP, \
    OnBldNoFishEnvironmentException, FISH_STANDALONE_CANARY_PATH, \
    SECT_GKOVERRIDES

"""Build tasks that can be used to build the OS/Net Consolidation.
"""

# cache the package version that would be produced by this workspace
__PKGVERS = None


# cache the results of get_hg_info
__HG_BRANCH_TIP_INFO = None
__HG_IDENTIFY_INFO = None


def get_hg_info(ws, env, hg_identify=True, force_cache_refresh=False):
    """Determine the top-most changesets on each branch for all repositories
    in the given workspace and count how many hg heads are in each repository.

    If 'hg_identify' is True, we use 'hg identify' to determine the changesets
    that the working directory is updated to. When 'hg_identify' is False we
    just use 'hg log -r "max(branch(<branch name>))"' to determine
    the tip-most changeset for this branch, which is faster but may not reflect
    which changeset the workspace is 'hg update'd to and won't show whether
    source files in the workspace have been modified."""

    global __HG_IDENTIFY_INFO
    global __HG_BRANCH_TIP_INFO

    if force_cache_refresh:
        __HG_IDENTIFY_INFO = None
        __HG_BRANCH_TIP_INFO = None

    if hg_identify:
        if __HG_IDENTIFY_INFO and ws in __HG_IDENTIFY_INFO:
            return __HG_IDENTIFY_INFO
        id_cmd = "identify"
    else:
        if __HG_BRANCH_TIP_INFO and ws in __HG_BRANCH_TIP_INFO:
            return __HG_BRANCH_TIP_INFO
        id_cmd = "log -r 'max(branch(.))' --template '{node|short}'"

    get_tip_cmd = "/usr/bin/hg -R %s --cwd %s %s"
    # The space in the template argument is important, for splitting the heads.
    # Using '.' gets us the heads on this branch only
    get_heads_cmd = (
        "/usr/bin/hg -R %s --cwd %s log -r 'heads(branch(.))' "
        "--template '{node|short} ' .")

    hg_info = {}
    for repo in [
            ws,
            "%s/usr/closed" % ws,
            "%s/usr/man" % ws,
            "%s/usr/fish" % ws]:

        if not os.path.isdir(os.path.join(repo, ".hg")):
            continue
        ret, out, err = util.cmdline_run(
            get_tip_cmd % (repo, repo, id_cmd), env_arg=env)
        if ret != 0:
            raise RuntimeError(
                "Unable to run hg commands for %s!\n%s\n%s\n" %
                (repo, out, err))

        # 'hg heads' can fail if we're running it in a repository
        # that contains no changesets (eg. has just been created with
        # 'hg init') so we ignore the return code.
        ret1, out1, err1 = util.cmdline_run(
            get_heads_cmd % (repo, repo), env_arg=env)

        # record the tip and number of heads for this repository
        heads = len(out1.strip().split(" "))
        hg_info[repo] = (out.strip(), heads)

    # cache results
    if hg_identify:
        __HG_IDENTIFY_INFO = hg_info
    else:
        __HG_BRANCH_TIP_INFO = hg_info

    return hg_info


def get_known_checks():
    """Return a map of the check.* tasks that are defined, mapping to the
    class that defines that task"""
    known_checks = {}
    tlist = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for task_tuple in tlist:
        if not issubclass(task_tuple[1], Task):
            continue
        task = task_tuple[1]
        if not task.task_name.startswith("check."):
            continue
        known_checks[task.task_name] = task
    return known_checks


def get_pkgvers(env):
    """Return the $(PKGVERS) component of the packages that would be published
    by this build with the given onbld os.environ."""

    # return a previously computed value if we can
    global __PKGVERS
    if __PKGVERS:
        return __PKGVERS

    # the command lines we use to retrieve package information
    make_cmd = "dmake -f Makefile.gk pkgvers"
    try:
        make_env = copy.deepcopy(env)
        make_env["NEED_NIGHTLYID"] = "yes"
        ret, out, err = util.cmdline_run(
            make_cmd, env_arg=make_env, cwd=env["SRC"])
        if ret != 0:
            return None
    except KeyboardInterrupt:
        raise
    except Exception:
        return None
    # The version string emitted by the build Makefiles looks like:
    # 5.12,5.12-5.12.0.0.0.88.12345
    # to make comparison easier, build(1) only uses the
    # $(PKGVERS_BRANCH) component. See $SRC/Makefile.buildnum.
    __PKGVERS = out.rstrip().split("-")[1]
    return __PKGVERS


class TaskException(Exception):
    """An exception that gets raised when bad things happen during tasks.
    Generally, we prefer to log messages and move on, but sometimes raising
    an exception is the only way to be sure."""
    pass


class TaskLockedException(TaskException):
    """An exception that gets raised when we're unable to obtain a lock on this
    workspace."""

    def __init__(self, message=None):
        super(TaskLockedException, self).__init__(message)
        self.host = None
        self.pid = None
        self.host = None
        self.user = None
        self.lock_name = None


class MissingToolsTaskException(TaskException):
    """An exception that gets raised when we're unable to find the external
    build tools that this task requires."""
    pass


class Task(object):
    """This class is used to represent a single task performed as part of
    the OS/Net build.

    Tasks typically execute with the environment obtained from self.bldenv.

    Logging is handled with a per-Task logging.Logger object.

    Two onbld.log.RenameFileHandlers are assigned to each logger, one for
    plaintext logs, one for HTML-formatted logs, with the filenames being
    partially defined by 'self.name'. Logging is hierarchical, so log messages
    sent to "onbld.build.mytask" will also get logged against "onbld" unless
    'propagate_logs' is False.

    The log filename is also influenced by OnBldEnvironment.debug and
    OnBldEnvironment.nondebug values, which define the type of build we're
    performing. Sometimes it can be convenient to switch build-types
    mid-task, in which case onbld.log.switch_logs(..) can be used. Note that
    this results in the start of the Task being logged to one file, with the
    end of the Task being logged to another file. We leave breadcrumbs as
    appropriate.

    NightlyTask avoids this by creating a new Task per build-type, and running
    all tasks for a single build-type in sequence, which is the approach we
    encourage subclasses to use.

    The log FileHandlers will process LogRecords of level 'level' and above.

    Subclasses should implement _execute() and optionally _execute_start() and
    _execute_finish(), taking care to also call the super-class implementation
    of _execute_start and _execute_finish.

    Subclasses should set a sensible 'task_name', which allows their
    functionality to be automatically exposed through a build(1) subcommand
    of the same name. build(1) will fail immediately if the same task_name is
    used by more than one class. Subclasses should also set a 'task_synopsis',
    used in the build(1) usage message.
    """

    # The name of this task, used to name our log files as well as to refer to
    # the task in log messages.
    task_name = "task"
    task_synopsis = None

    # Some tasks perform so little work, that a ZFS snapshot for the task
    # would be overkill. If we shouldn't take a snapshot, subclasses can set
    # this to False.
    snapshot = True

    # Most tasks produce small logs, in which case writing log extracts
    # in the case of Task failure doesn't really help much. Subclasses such as
    # InstallTask set this to True.
    produce_log_extract = False

    # Most tasks require external (possibly unpackaged) build tools to be
    # present at certain versions. Some Tasks do not require any external
    # tools, beyond those present on a build system as part of the
    # pkg:/developer/opensolaris/osnet group package. Subclasses can choose to
    # override this.
    check_tools = True

    # This determines the number of context lines that are printed around
    # a single TWARNING-level message when showing log extracts in
    # the mail_msg generation. For most purposes, 10 is sufficient.
    mail_msg_summary_fragsize = 10

    # By default, build tasks should not run when another build task is running
    # in this workspace. If set to True, this will cause build tasks to ensure
    # the workspace lock is held by this process before continuing with the
    # build task.
    require_workspace_lock = True

    # Our check.* tasks can be marked as enabled_by_default which determines
    # the value used for that task in 'build generate -O' output.
    enabled_by_default = True

    # Whether this task has to pass
    must_pass = True

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=False, name=None):
        """Create a Task. 'bldenv' is an onbld.environment.OnBldEnvironment
        describing the environment the task should execute with. 'level'
        defines which level logs messages will be written at for this task.
        'propagate_logs' defines whether messages written to the Task-specific
        logger should propagate up the logging hierarchy.

        If 'no_logfiles' is True, we don't create FileHandlers for the task,
        resulting in no log files being generated. Log messages will still be
        emitted, but we won't have a permanent record of them unless a logger
        further up the hierarchy decides to capture them.

        'name', can be set to the name of this build task, needed if multiple
        Task objects of the same name are run during the same build which
        otherwise would result in log filenames being rotated.
        """

        # The human-friendly name of this task.
        if name:
            self.name = name
        else:
            self.name = self.__class__.task_name

        # A non-zero exit status indicates something has gone wrong.
        self.exit_status = 0

        # Indicate whether this task was interrupted while running
        self.interrupted = False

        # Some tasks act as predicates, running a series of commands to
        # determine whether subsequent Tasks should run. These predicates
        # can be marked as satisfied or not.
        self.satisfied = True
        self.bldenv = bldenv

        # We discourage build(1) tasks from taking separate options per
        # task to reduce the potential for different tasks assigning
        # different meanings to the same options. Some tasks, like
        # HereTask can use the arguments for their own purposes however.
        self.args = args

        # Some Tasks need different environments depending on whether the Task
        # is running for debug or non-debug builds.  We set a default in
        # self.env for convenience, but subclasses may use self.bldenv to
        # obtain a fresh environment for a particular build type. Note that
        # the default environment is 'debug' if both debug and non-debug are
        # set to True in self.bldenv.
        self.env = bldenv.get_build_environment()

        # The task name should reflect the branch flavor we're updated to
        branch_prefix = self.env["ONBLD_BRANCH_PREFIX"]
        if branch_prefix:
            self.name = "{0}{1}".format(branch_prefix, self.name)

        # Tasks that have never started or finished are marked with None
        self.start_time = None
        self.finish_time = None
        self.start_sinceepoch = None
        self.finish_sinceepoch = None

        self.level = level

        # For convenience, set some frequently-used values from self.env.
        self.ws = self.env["WS"]
        self.src = self.env["SRC"]
        self.mach = self.env["MACH"]
        self.keep = self.env["ONBLD_KEEP"]
        self.branch = self.env["ONBLD_HG_BRANCH"]

        # Names of our log files, initialized in __setup_logging()
        self.log_name = None
        self.logextract_name = None

        # If we take snapshots, this is the name of the suffix we use
        # (making it easier to see which snapshots from different Tasks are
        # related to each other)
        self.snapshot_suffix = None

        # When trying to take snapshots, we can choose whether to enable
        # the warning that reports whether snapshots could be taken. For
        # top-level tasks such as nightly, it's useful to see that warning
        # but if snapshots are enabled, but we can't take snapshots, seeing
        # that message for every build(1) task is overkill.
        self.warn_on_snapshot_errors = False

        # When creating a task from build(1) we record the command line used
        # which we log before starting the task.
        self.command_line_args = None

        # We can override the default logic for this class, allowing us to
        # require that the task should succeed, to influence the exit status
        # of build(1)
        self.must_pass = self.bldenv.overrides.get(
            "%s.must_pass" % self.task_name, self.__class__.must_pass)

        # Tasks can be added to other tasks, for now just by subclasses
        # of CompositeTask. We set _parent task so we can build a view
        # of all tasks involved in the current build(1) invocation.
        self._parent_task = None

        if self.check_tools and not bldenv.find_tools(log_versions=False):
            if not self.bldenv.bldrc:
                raise MissingToolsTaskException(
                    "Unable to find required build tools. Since a "
                    "configuration file was not passed, the defaults were "
                    "used. Use 'build generate' to create a configuration "
                    "file to specify where the build tools are located.")
            else:
                raise MissingToolsTaskException(
                    "Unable to find required build tools")

        # Quick santity check to ensure this is actually a workspace
        ws_make_master = os.path.join(self.ws, "usr/src/Makefile.master")
        fish_standalone = os.path.join(self.ws, FISH_STANDALONE_CANARY_PATH)
        if not (os.path.exists(ws_make_master) or
                os.path.exists(fish_standalone) and
                os.path.isdir(os.path.join(self.ws, ".hg"))):
            if not self.bldenv.bldrc:
                raise TaskException(
                    "Unable to locate a valid workspace. either run "
                    "build(1) from the top level of a workspace, or use "
                    "'build generate' to create a configuration file, and "
                    "change the 'ws' key to specify the workspace "
                    "location.")
            else:
                raise TaskException(
                    "'ws' in the config file does not appear to point at "
                    "a valid workspace.")

        if self.require_workspace_lock:
            self._acquire_lock()
        self.__setup_logging(
            propagate_logs=propagate_logs, no_logfiles=no_logfiles,
            log_format=bldenv.log_format)
        self.mark_pending()

    def __setup_logging(self, propagate_logs=True, no_logfiles=False,
                        log_format=log.NORMAL_FORMAT):
        """Setup logging for the task. Some tasks are common across build types
        - we run them only once for both debug and non-debug builds, in which
        case, we don't need the "-nd" distinction in the log file. We change
        the task name to indicate the build type, which then gets used by
        self.take_snapshot()

        'log_format' is one of the log formats defined in onbld.log that
        determines the format of log entries.
        """

        if self.bldenv.nondebug and not self.bldenv.debug:
            self.name = "%s-nd" % self.name
            self.logger = logging.getLogger(
                "%s.%s-nd" %
                (self.__class__.__module__, self.__class__.__name__))
        else:
            self.logger = logging.getLogger(
                "%s.%s" % (self.__class__.__module__, self.__class__.__name__))

        # if the task failed and we produced a log file extract, record the
        # log name here
        self.logextract_name = None
        self.logger.propagate = propagate_logs
        self.logger.setLevel(self.level)
        self.log_dir = self.env["LOGDIR"]
        if no_logfiles:
            self.produce_log_extract = False
            return

        # record the name of the plaintext logfile we'll be writing
        self.log_name = "%s.log" % self.name
        log_path = os.path.join(self.log_dir, "%s.log" % self.name)

        if not os.path.exists(self.log_dir):
            try:
                os.makedirs(self.log_dir, 0o755)
            except OSError as e:
                raise TaskException(
                    "Unable to create log directory %s: %s\n"
                    "Check the 'logdir' property in the bldrc file." %
                    (self.log_dir, e))

        formatter = logging.Formatter(
            log.log_formats[log_format], datefmt=log.datefmt)
        fh = log.RenameFileHandler(log_path)
        fh.setFormatter(formatter)
        fh.setLevel(self.level)
        self.logger.addHandler(fh)

        html_fh = log.HtmlRenameFileHandler("%s.html" % log_path)
        html_fh.setFormatter(log.HtmlFormatter(
            datefmt=log.datefmt))
        html_fh.setLevel(self.level)
        self.logger.addHandler(html_fh)

    def _acquire_lock(self, dir=None):
        """Acquire lock for this task. Locking is done with a symlink that
        points to a non-existent file as a target. The target contains
        information to tell us what process on what hostname is currently
        holding the lock.

        The lock file includes the machine architecture so that two Tasks can
        run from the same workspace at the same time on different
        architectures.

        If 'dir' is set, we try to acquire the lock in the given directory
        path, which must already exist. Otherwise we default to
        os.path.join(self.env["ONBLD_KEEP"], self.mach, "locks)

        For this implementation, the lock is not Task-specific. The lock is
        deemed to be acquired if any Task in the current process already
        owns the lock. This avoids complexity if a user runs:

         % build nightly &
            [wait till 'install' subtask has completed and 'packages'
             subtask starts]
         % build clobber

        The 'build clobber' should fail because an existing process holds the
        lock. On the other hand, a 'build check' would also fail, despite some
        of the 'check' subtasks being largely independent of the rest of the
        build Tasks.  When parallel Task execution is supported, this locking
        mechanism needs to be updated."""

        acquired = False
        if dir:
            lock_dir = dir
        else:
            lock_dir = os.path.join(
                self.env["ONBLD_KEEP"], self.mach, "locks")

        # there is no Task-specific lock: all Tasks use the same lock
        self.__lock_name = "%s/build.%s.lock" % (lock_dir, self.mach)
        # some dhcp supplied hostnames contain dots
        this_host = socket.gethostname().split(".")[0]
        self.__lock_targ = (
            "%(hostname)s.%(logname)s.%(pid)s" %
            {"hostname": this_host, "logname": self.env["USER"],
                "pid": os.getpid()})

        host = ""
        user = ""
        pid = ""
        failed_reason = None
        while not acquired:
            try:
                if not os.path.exists(lock_dir):
                    os.makedirs(lock_dir, 0o755)

                os.symlink(self.__lock_targ, self.__lock_name)
                acquired = True
            except OSError as e:
                if e.errno != errno.EEXIST:
                    failed_reason = e
                    break
                # unable to create lock symlink - let's investigate why by
                # reading the current target of the symlink
                host, user, pid = os.readlink(self.__lock_name).split(".")
                try:
                    # If this process on this host already owns the lock, we
                    # consider it as "acquired".
                    if pid == str(os.getpid()) and \
                            host == this_host:
                        acquired = True
                        break
                    os.kill(int(pid), 0)
                    break
                except OSError as e:
                    # pid doesn't exist, we've got a stale lock, remove it and
                    # continue our while-loop.
                    if e.errno == errno.ESRCH and host == socket.gethostname():
                        os.remove(self.__lock_name)
                    else:
                        break
        if not acquired:
            if host and user and pid:
                tle = TaskLockedException(
                    "Unable to acquire lock. %s already held by %s on %s, "
                    "pid = %s" % (self.__lock_name, user, host, pid))
                tle.lock_name = self.__lock_name
                tle.pid = pid
                tle.user = user
                tle.host = host
                raise tle
            if failed_reason:
                tle = TaskLockedException(
                    "Unable to acquire lock %s: %s" %
                    (self.__lock_name, failed_reason))
            else:
                tle = TaskLockedException(
                    "Unable to acquire lock %s" % self.__lock_name)
            raise tle

    def _mark_state(self, state):
        """Build state files indicate the current state of each build task
        as we execute. If a build task has not finished executing, its state
        is "pending". If a build task has completed, but failed, its state is
        "fail". When the special state "clear" is used, we remove all state
        files.

        State files are used as a stable way for build processes on different
        machines to determine the basic state of a build, without having to
        parse build logs."""

        lock_dir = os.path.join(self.env["ONBLD_KEEP"], self.mach, "locks")
        state_file = os.path.join(
            lock_dir, "{0}.{1}".format(self.name, state))
        # always touch the state file before removing the old state
        if state != "clear":
            if not os.path.exists(lock_dir):
                os.makedirs(lock_dir)
            open(state_file, "w").close()
        for remove_state in ["pass", "fail", "pending"]:
            if remove_state == state:
                continue
            try:
                state_file = os.path.join(
                    lock_dir, "{0}.{1}".format(self.name, remove_state))
                os.remove(state_file)
            except OSError:
                pass

    def mark_pass(self):
        self._mark_state("pass")

    def mark_fail(self, recursive=False):
        self._mark_state("fail")

    def mark_pending(self):
        self._mark_state("pending")

    def mark_clear(self):
        self._mark_state("clear")

    def wait_task(
            self, task_name, remote_ws_keep, remote_mach, timeout=(60 * 60),
            poll_interval=15):
        """Wait for the 'task_name' Task, presumed running in workspace
        'remote_ws_keep' to complete. The default timeout is an hour.
        """
        remote_lock_dir = os.path.join(
            remote_ws_keep, remote_mach, "locks")
        remote_lock = os.path.join(
            remote_lock_dir, "build.{0}.lock".format(remote_mach))
        task_pending = os.path.join(
            remote_lock_dir, "{0}.pending".format(task_name))
        task_fail = os.path.join(
            remote_lock_dir, "{0}.fail".format(task_name))
        task_pass = os.path.join(
            remote_lock_dir, "{0}.pass".format(task_name))
        task_fail_message = "Previous task {0} failed".format(task_name)

        # make sure we can see the remote workspace:
        if not os.path.isdir(remote_ws_keep):
            raise TaskException(
                "Unable to access remote workspace {0}".format(remote_ws_keep))

        # If the workspace isn't locked, there's nothing to wait for.
        # This is prone to problems where the build has never started on
        # the remote machine.
        if not os.path.islink(remote_lock):
            self.logger.log(
                TINFO_LEVEL, "Remote lock {0} not held - assuming task "
                "completed unless it was marked as having failed.".format(
                    remote_lock))
            if os.path.exists(task_fail):
                raise TaskException(task_fail_message)
            return

        # workspace is locked.
        if not os.path.exists(task_pending) and \
                not (os.path.exists(task_fail) or os.path.exists(task_pass)):
            raise TaskException(
                "Remote workspace {0} has no pending task for {1}".format(
                    remote_ws_keep, task_name))

        self.logger.log(
            TINFO_LEVEL, "Waiting for '{0}' task in {1} to complete.".format(
                task_name, remote_ws_keep))
        now = time.time()
        expiry = now + timeout
        while now < expiry:
            if os.path.exists(task_pass):
                return
            elif os.path.exists(task_fail):
                raise TaskException(task_fail_message)
            else:
                time.sleep(poll_interval)
                now = time.time()
        raise TaskException(
            "Timeout waiting for {0} in {1}".format(task_name, remote_ws_keep))

    def _release_lock(self, dir=None):
        """Release our lock. Note that when a Task finishes, it does not
        usually release the global lock - this is intentional, since
        CompositeTasks may require the workspace lock to be held for the
        duration of several Tasks.

        When all build tasks are completed, the caller should invoke
        Task.unlock() should be called to release the lock.
        """

        if not self.require_workspace_lock:
            return

        # Check the lock still exists
        if not os.path.islink(self.__lock_name):
            return
        try:
            # Check we hold the lock by attempting to re-acquire it
            self._acquire_lock(dir=dir)
        except TaskLockedException as e:
            raise TaskLockedException(
                "Unable to release lock which we have not acquired: %s" % e)
        os.remove(self.__lock_name)

    def unlock(self, clear_statefiles=False):
        if clear_statefiles:
            self.mark_clear()
        self._release_lock()

    def run(self):
        """Runs the Task, calling _execute_start, _execute and _execute_finish
        in order, dealing with error conditions as appropriate. Subclasses
        should probably not override this method."""

        def log_exception(e, missing_fish=False):
            """Uncaught exceptions in Tasks tend to indicate a bug in
            build.py. Tasks should not usually raise exceptions in their
            _execute*() methods, but should catch and log all errors
            appropriately."""

            self.exit_status = 1
            self.mark_fail()

            if missing_fish:
                self.logger.error(e.message)
                return

            self.logger.log(
                logging.ERROR, "Uncaught exception during task execution: "
                "%s." % e, exc_info=True)
            self.logger.log(
                logging.ERROR, "Please report this to your friendly "
                "gatekeepers at %s" % GATEKEEPER_MAILTO)

        try:
            self._execute_start()
        except KeyboardInterrupt:
            return
        except OnBldNoFishEnvironmentException as e:
            log_exception(e, missing_fish=True)
            return
        except Exception as e:
            # if we haven't got off the ground, just return immediately
            log_exception(e)
            return
        try:
            if self.command_line_args:
                self.logger.log(
                    TINFO_LEVEL, "build(1) command line: %s" %
                    self.command_line_args)
            self._execute()
        except KeyboardInterrupt:
            self.logger.log(
                TWARNING_LEVEL, "Keyboard interrupt caught: stopping task "
                "execution.")
            self.exit_status = 1
            self.interrupted = True
            self.mark_fail()
            # raise this so that we stop all subsequent tasks as well
            raise
        except OnBldNoFishEnvironmentException as e:
            log_exception(e, missing_fish=True)
            return
        except Exception as e:
            log_exception(e)
        finally:
            # clean up if we can.
            try:
                self._execute_finish()
            except KeyboardInterrupt:
                pass
            except OnBldNoFishEnvironmentException as e:
                log_exception(e, missing_fish=True)
            except Exception as e:
                log_exception(e)

    def _execute(self):
        """Execute the task. Tasks should handle all exceptions raised during
        execution, with appropriate logging."""

        # Intentionally not implemented here: subclasses should be defining
        # what the task does.
        raise TaskException(
            "Attempt made to execute a Task with no implementation!")

    def _execute_start(self):
        """A method intended to run before execute(). Subclasses can implement
        this, but should also call this implementation to log timestamps and
        take a ZFS snapshot before starting the task. Tasks should handle all
        exceptions raised during _execute_start methods, with appropriate
        logging."""

        self.start_sinceepoch = time.time()
        self.start_time = datetime.datetime.fromtimestamp(
            self.start_sinceepoch)

        self.logger.log(STATE_LEVEL, "Starting task '%s'" % self.name)
        self.logger.log(TINFO_LEVEL, "bldrc file: %s" % self.bldenv.bldrc)
        if self.snapshot and self.bldenv.snapshots:
            self.take_snapshot()

    def _execute_finish(self):
        """A method that runs after execute(). Subclasses can implement this,
        but should also call this version. Tasks should handle all exceptions
        raised during execution, with appropriate logging."""

        self.finish_sinceepoch = time.time()
        self.finish_time = datetime.datetime.fromtimestamp(
            self.finish_sinceepoch)

        self.finish_time = datetime.datetime.now()
        task_length = str(self.finish_time - self.start_time).split(".")[0]

        self.logger.log(STATE_LEVEL, "Finishing task '%s'" % self.name)
        if self.exit_status == 0:
            self.mark_pass()
            self.logger.log(
                SUMMARY_LEVEL, "'%s' completed successfully and took %s." %
                (self.name, task_length))
        elif not self.must_pass:
            self.mark_fail()
            self.logger.log(
                SUMMARY_LEVEL, "'%s' failed (non-fatal) and took %s." %
                (self.name, task_length))
        else:
            self.mark_fail()
            self.logger.log(
                SUMMARY_LEVEL, "'%s' failed, and took %s." %
                (self.name, task_length))

        if self.exit_status != 0 and self.produce_log_extract:
            self.logextract_name = log.produce_extract(
                self.logger, "%s/%s" % (self.log_dir, self.log_name),
                log_format=self.bldenv.log_format)

            # If we've rotated the logger, we'll have both *.log and *-nd.log
            # files, so we need to check for extractable messages in the *-nd
            # log.
            nd_log = "%s/%s-nd.log" % (self.log_dir, self.name)
            if self.bldenv.debug and self.bldenv.nondebug and \
                    os.path.exists(nd_log):
                log.produce_extract(
                    self.logger, nd_log, log_format=self.bldenv.log_format)
            log.close_logs(self.logger)

    def summary(self, max_lines=30):
        """Return a string containing a summary of the task. Typically this
        involves digging through the log files to find interesting output if
        the task has failed, otherwise, it could provide details about the
        task output.

        The default implementation is to return all log messages produced by
        the Task using onbld.log.produce_extract() up to a maximum of 30 lines.
        """

        # If the Task doesn't use a log file, this implementation can't produce
        # a summary.
        if not self.log_name:
            return ""

        logger = logging.getLogger("summary.%s" % self.task_name)
        summary_handler = log.StringIOHandler()
        logger.addHandler(summary_handler)
        orig_log_path = os.path.join(self.log_dir, self.log_name)
        if not os.path.exists(orig_log_path):
            return ""

        log.produce_extract(
            None, orig_log_path, elogger=logger,
            bufsize=self.mail_msg_summary_fragsize)

        # If we've rotated the logger, we'll have both *.log and *-nd.log
        # files, so we need to check for extractable messages in the *-nd log.
        nd_log = "%s/%s-nd.log" % (self.log_dir, self.name)
        if self.bldenv.debug and self.bldenv.nondebug and \
                os.path.exists(nd_log):
            log.produce_extract(
                self.logger, nd_log, log_format=self.bldenv.log_format)
        lines = summary_handler.stringio.getvalue().split("\n")
        if len(lines) > max_lines:
            blank = ".\n.\n.\n(additional lines truncated)\n"
            return "\n".join(lines[:max_lines]) + blank
        return "\n".join(lines)

    def take_snapshot(self, snap_suffix=None):
        """Try to take a ZFS snapshot of self.ws, based on the Task
        name. If 'snap_suffix' is set, use this string as part of the snapshot
        name."""

        if not self.bldenv.snapshots:
            return
        if snap_suffix:
            suffix = snap_suffix
        elif self.snapshot_suffix:
            suffix = self.snapshot_suffix
        else:
            suffix = self.start_time.strftime(log.datefmt)

        snapdir = "%s/.zfs/snapshot/build-%s.%s" % (
            self.ws, self.name, suffix)
        try:
            if not os.path.exists(snapdir):
                os.mkdir(snapdir)
            else:
                renamed = "%s_renamed_%s" % (
                    snapdir, self.start_time.strftime(log.datefmt))
                os.rename(snapdir, renamed)
                self.logger.log(
                    TINFO_LEVEL, "renamed old snapshot at %s" % renamed)
                os.mkdir(snapdir)
            self.logger.log(TINFO_LEVEL, "new snapshot taken %s" % snapdir)
        except OSError as e:
            if self.warn_on_snapshot_errors:
                self.logger.log(
                    TINFO_LEVEL, "Unable to create snapshot at %s: %s" %
                    (snapdir, e))

    def get_known_tasks(self):
        if self._parent_task:
            return self._parent_task.get_known_tasks()
        return set([self])


class CompositeTask(Task):
    """A Task that is comprised of a series of other Tasks, executed
    in order. Typically composite tasks create copies of the OnBldEnvironment
    object they are passed in order to run debug and non-debug Tasks in a
    defined sequence, preventing the need for task execution to straddle
    log filename changes."""

    task_name = "composite"
    task_synopsis = None

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL,
            propagate_logs=True, no_logfiles=False, fail_fast=True,
            name="composite"):
        """Create a CompositeTask.  If fail_fast is set to True, we exit as
        soon as any of our Tasks fail, rather than executing all Tasks, and
        returning at the end."""
        super(CompositeTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            no_logfiles=no_logfiles)
        self.tasks = []
        self.fail_fast = fail_fast

    def add_task(self, task, index=None):
        """Add a task to the list of tasks we're going to execute. If
        necessary, 'index' can specify an integer specifying where we should
        insert the new task in our list of tasks to execute (task execution
        starts at index 0, and counts upwards)

        'task' should either be an instance of Task, or a 'str' that will
        get logged to self.logger when processing that list item.
        """

        if isinstance(task, Task):
            task._parent_task = self
        if index is None:
            self.tasks.append(task)
        else:
            self.tasks.insert(index, task)

    def mark_fail(self, recursive=False):
        super(CompositeTask, self).mark_fail()
        if recursive:
            for task in self.tasks:
                if isinstance(task, str):
                    continue
                task.mark_fail()

    def get_known_tasks(self):
        """Return an unordered set of all tasks due to be executed by
        this task."""
        # we have a parent, delegate responsibility to it
        if self._parent_task:
            return self._parent_task.get_known_tasks()

        # we have no parent
        def find_children(task):
            # For very heavily nested trees of CompositeTasks,
            # this could hurt due to function recursion, but that's
            # not a common scenario.
            children = set()
            if isinstance(task, CompositeTask):
                children.add(task)
                for t in task.tasks:
                    children.update(find_children(t))
            else:
                children.add(task)
            return children

        return find_children(self)

    def _execute(self):
        """Execute each of our tasks in turn, logging errors as we go."""

        if not self.tasks:
            raise TaskException(
                "No tasks have been added to this CompositeTask.")

        for task in self.tasks:
            # all related tasks should have related ZFS snapshot names
            if self.snapshot_suffix:
                task.snapshot_suffix = self.snapshot_suffix

            # We allow self.tasks to contain either Task objects, or a 'str'
            if isinstance(task, str):
                self.logger.log(TINFO_LEVEL, task)
                continue

            if task.log_name:
                self.logger.log(
                    TINFO_LEVEL, "Starting next task - logging to %s" %
                    task.log_name)
            else:
                self.logger.log(
                    TINFO_LEVEL, "Starting next task - %s" % task.name)
            try:
                task.run()
            except KeyboardInterrupt:
                self.exit_status = 1
                self.interrupted = True
                self.mark_fail()
                task_index = self.tasks.index(task)
                for t in self.tasks[task_index:]:
                    if isinstance(t, str):
                        continue
                    t.mark_fail(recursive=True)
                raise
            except Exception as e:
                self.exit_status = 1
                self.logger.log(
                    TWARNING_LEVEL, "Exception running task '%s': %s" %
                    (task.name, e), exc_info=e)

            # the str() method of timedelta objects gives us microsecond
            # resolution, which is overkill. Ignore the microseconds.
            task_length = str(task.finish_time - task.start_time).split(".")[0]
            if task.exit_status == 0:
                self.logger.log(
                    TINFO_LEVEL, "'%s' completed successfully and took %s." %
                    (task.name, task_length))
            elif not task.must_pass:
                self.logger.log(
                    TWARNING_LEVEL, "'%s' failed (non-fatal) and took %s." %
                    (task.name, task_length))
            else:
                self.logger.log(
                    TWARNING_LEVEL, "'%s' failed, and took %s." %
                    (task.name, task_length))

                if task.logextract_name:
                    self.logger.log(
                        TINFO_LEVEL, "Log extract at %s" %
                        task.logextract_name)

                if self.fail_fast:
                    self.logger.log(TWARNING_LEVEL, "Bailing out now.")
                    break
                else:
                    self.logger.log(
                        TWARNING_LEVEL, "Task %s failed." % task.name)
            if not task.satisfied:
                self.logger.log(
                    STATE_LEVEL, "Predicate task '%s' was not satisfied - no "
                    "more work to do." % task.name)
                break

        # determine our exit status
        for task in self.tasks:
            if isinstance(task, str):
                continue
            if task.exit_status != 0 and task.must_pass:
                self.exit_status = 1
        return self.exit_status


class HgPullTask(Task):
    """Performs a hg pull for all repositories in this workspace."""

    task_name = "hgpull"
    task_synopsis = "Do a simple hg pull for all repositories in this "\
        "workspace"
    check_tools = False

    def _execute(self):
        """Execute the task."""

        # a list of repositories in this workspace
        repositories = [
            (self.ws, True),
            ("%s/usr/closed" % self.ws,
                self.env["CLOSED_IS_PRESENT"] == "yes"),
            ("%s/usr/man" % self.ws, self.env["MAN_IS_PRESENT"] == "yes"),
            ("%s/usr/fish" % self.ws, self.env["FISH_IS_PRESENT"] == "yes"),
        ]

        if not os.path.exists(os.path.join(self.ws, ".hg")):
            self.logger.log(
                "%s does not appear to be a Mercurial repository!" % self.ws)
            self.exit_status = 1
            return

        # record the number of repositories we managed to update
        self.updated_repos = 0
        for repo, should_exist in repositories:
            if not should_exist:
                continue

            if not os.path.isdir(repo):
                self.logger.error("unable to find dir %s" % repo)
                self.exit_status = 1
                continue

            # Some gates don't have separate repositories for subtrees
            if not os.path.isdir(os.path.join(repo, ".hg")):
                self.logger.log(
                    TINFO_LEVEL,
                    "%s is not a separate repository, skipping" % repo)
                continue

            cmd = "/usr/bin/hg -R %s pull -uv" % repo
            ret = util.cmdline_run(cmd, env_arg=self.env, logger=self.logger)
            if ret != 0:
                self.exit_status = 1
                self.logger.log(TWARNING_LEVEL, "Pull task failed!")


class PredicateHgPullTask(HgPullTask):
    """Checks for at least one changed repository after doing hg pulls,
    flagging the task as unsatisfied otherwise. We also look for new heads
    being added to the local repository, which also causes the task to be
    flagged as unsatisfied."""

    task_name = "pull"
    task_synopsis = "Do a hg pull and report new changesets/heads"

    def _execute(self):
        self.satisfied = False
        # The hg pull command uses the checked-in changesets to determine
        # hg version information, rather than using 'hg identify'
        old_hg_info = get_hg_info(self.ws, self.env, hg_identify=False)
        # run the hg update
        super(PredicateHgPullTask, self)._execute()
        if self.exit_status != 0:
            self.logger.log(TWARNING_LEVEL, "hg pull failed")
            self.exit_status = 1
            return

        # don't use the cache of previous hg info
        new_hg_info = get_hg_info(
            self.ws, self.env, hg_identify=False, force_cache_refresh=True)

        # check if any repositories have been updated, making sure no new heads
        # were brought in by the merge.
        updated_repos = 0
        heads_mismatch = []
        for repo, info in list(new_hg_info.items()):
            if old_hg_info[repo][1] != info[1]:
                heads_mismatch.append(repo)
            if old_hg_info[repo][0] != info[0]:
                self.logger.log(
                    TINFO_LEVEL, "Old %s tip: %s" %
                    (repo, old_hg_info[repo][0]))
                self.logger.log(
                    TINFO_LEVEL, "New %s tip: %s" %
                    (repo, new_hg_info[repo][0]))
                updated_repos += 1
            else:
                self.logger.log(
                    TINFO_LEVEL, "%s remains at changeset %s" %
                    (repo, info[0]))
        if heads_mismatch:
            self.logger.log(
                logging.WARNING, "New heads were added to %s which might need "
                "to be merged. Giving up now." % self.ws)
        elif updated_repos > 0:
            self.logger.log(TINFO_LEVEL, "New changesets were pulled.")
            self.satisfied = True
        else:
            self.logger.log(TINFO_LEVEL, "No changesets were pulled.")


class ClobberTask(Task):

    """Performs a 'make clobber' in $SRC and $SRC/tools, as well as
    removing any .make.* files found beneath $SRC. This also removes
    $ROOT, $TOOLROOT, $NATIVEROOT, $PKGARCHIVE directories, and their
    STUB and KSPLICE equivalents, if they exist for any of the build
    types defined in the bldenv.

    If the coreadm directory for this process is within $WS, it will also
    delete old cores in that directory.
    """

    task_name = "clobber"
    task_synopsis = "Do a workspace clobber"
    produce_log_extract = True
    check_tools = False

    def _execute(self):

        if self.env["FISH_IS_PRESENT"] == "yes":
            # If a previous build failed to clean up, do so now.
            # Note that this happens whether we're doing an incremental
            # build or not - we aren't yet using distro_const(8) checkpoints.
            destroy_dc_dataset(self.env["DC_DATASET"], self.logger)

        if self.bldenv.incremental:
            self.logger.log(TINFO_LEVEL, "No clobber (incremental build)")
            return 0

        if self.bldenv.build_in_place:
            self.logger.log(TINFO_LEVEL, "Doing a make clobber")
            cmd = "dmake clobber"
            self.exit_status += util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd=self.src)
            self.logger.log(TINFO_LEVEL, "Doing make clobber of tools")
            self.exit_status += util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd="%s/tools" %
                self.src)

        # Get back to a clean workspace as much as possible to catch
        # problems that only occur on fresh workspaces.
        # Remove all .make.state* files, libraries, and .o's that may
        # have been omitted from clobber.  A couple of libraries are
        # under source code control, so leave them alone.
        # We should probably blow away temporary directories too.
        if self.env.get("CLOSED_IS_PRESENT") == "yes":
            closed = "%s/../closed" % self.src
        else:
            closed = ""

        if self.env.get("FISH_IS_PRESENT") == "yes" and \
                not self.env.get("FISH_STANDALONE"):
            fish = "%s/../fish" % self.src
        else:
            fish = ""

        cmd = (
            "/usr/bin/find %(src)s %(closed)s %(fish)s "
            "\\( -name .hg -o -name 'interfaces.*' \\) "
            "-prune -o "
            "\\( -name '.make.*' -o -name 'lib*.a' -o -name 'lib*.so*' "
            "-o -name '*.o' \\) \\! -type d -print | "
            "/usr/bin/xargs rm -f " %
            {"src": self.src, "closed": closed, "fish": fish})
        # given that the exciting command line above uses pipes, we need to use
        # shell=True
        self.logger.log(TINFO_LEVEL, "Clearing sources in %s" % self.ws)
        self.exit_status += util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.src,
            shell=True)

        def clobber_dirs(debug=False):
            """Remove a series of directories."""

            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            ret_dir_rm = 0

            # For each call to clobber_dirs(), we will first remove
            # $BUILD, and later $TOOLBUILD/usr/src/tools.
            #
            # When debug=True, $BUILD == $TOOLBUILD. In this case, the
            # removal of $TOOLBUILD/usr/src/tools is redundant with
            # the removal of $BUILD.
            #
            # When debug=False, $BUILD != $TOOLBUILD. When
            # clobber_dirs() was previously called with debug=True,
            # the removal of $TOOLBUILD/usr/src/tools is redundant
            # with the removal of $BUILD from the *previous* call to
            # clobber_dirs().
            #
            # For a nondebug-only clobber task, however, there is no
            # such previous call, and the directories are not nested,
            # so the second removal is necessary to properly clobber
            # the tools portion of the build tree.
            #
            # The path removal loop below handles this correctly, and
            # only operates on existing directories.

            paths = [env[key] for key in [
                "BUILD", "TOOLROOT", "NATIVEROOT", "STUBROOT", "PKGARCHIVE",
                "KSPLICE_SPLICE_PKGARCHIVE", "KSPLICEROOT",
                "MANROOT"] if env.get(key)]

            if env.get("TOOLBUILD"):
                paths.append("%s/usr/src/tools" % env["TOOLBUILD"])

            if env["FISH_IS_PRESENT"] == "yes":
                fish_env = self.bldenv.get_build_environment(
                    debug=debug, nondebug=nondebug, fish=True)
            else:
                fish_env = {}

            # If checks are enabled and if a proto area already exists, we
            # rename that to $ROOT.prev so that build checks can reference it.
            # We account for both the open and closed proto areas.
            if self.bldenv.check:
                self.logger.log(
                    TINFO_LEVEL, "Build checks are enabled: "
                    "renaming old proto dir to [proto].prev")

                rename_paths = [env["ROOT"], "%s-closed" % env["ROOT"]]
                if fish_env:
                    rename_paths.append(fish_env["ROOT"])

                for rename_path in rename_paths:
                    if os.path.isdir(rename_path):
                        prev_proto = "%s.prev" % rename_path
                        try:
                            if os.path.isdir(prev_proto):
                                self.logger.log(
                                    TINFO_LEVEL, "Removing old proto dir %s" %
                                    prev_proto)
                                shutil.rmtree(prev_proto)
                            os.rename(rename_path, "%s.prev" % rename_path)
                        except OSError as e:
                            self.logger.log(
                                TWARNING_LEVEL, "Unable to rename old "
                                "$ROOT.prev dir: %s" % e, exc_info=True)
                            # Rather than removing the old proto directory,
                            # we should bail out now.
                            return 1
            else:
                paths.append(env["ROOT"])
                paths.append("%s-closed" % env["ROOT"])
                if fish_env:
                    for key in ["CPIODIR", "PKGARCHIVE", "ROOT", "STUBROOT"]:
                        paths.append(fish_env[key])

            for path in paths:
                if os.path.exists(path):
                    self.logger.log(TINFO_LEVEL, "Removing %s" % path)
                    try:
                        if os.path.islink(path):
                            os.unlink(path)
                        else:
                            shutil.rmtree(path)
                    except OSError as e:
                        self.logger.log(
                            TWARNING_LEVEL, "Unable to remove dir: %s" %
                            e, exc_info=True)
                        ret_dir_rm = 1
            return ret_dir_rm

        self.logger.log(TINFO_LEVEL, "Removing proto and pkg areas")
        if self.bldenv.debug:
            self.exit_status += clobber_dirs(debug=True)

        if self.bldenv.nondebug:
            self.exit_status += clobber_dirs()

        if self.bldenv.pkgmerge:
            pkgmerge_type, pkgmerge_dest, \
                debug_pkgs_local, debug_pkgs_altws, \
                nondebug_pkgs_local, nondebug_pkgs_altws = \
                _compute_pkgmerge_params(self.bldenv)
            if os.path.exists(pkgmerge_dest):
                self.logger.log(TINFO_LEVEL, "Removing %s" % pkgmerge_dest)
                try:
                    shutil.rmtree(pkgmerge_dest)
                except OSError as e:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to remove dir: %s" %
                        e, exc_info=True)
                    self.exit_status += 1

        if self.bldenv.overrides.get("check.parfait", True):
            self.logger.log(
                TINFO_LEVEL, "Clobbering old parfait bytecode files")
            cmd = (
                "/usr/bin/find %(src)s %(closed)s -name '*.bc' "
                "-type f -exec rm -f {} +" %
                {"src": self.src, "closed": closed})
            self.exit_status += util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd=self.ws)

        # If we have a coreadm pattern for this process that indicates
        # a directory and if that directory is beneath $WS, we remove its
        # contents.
        ret, out, err = util.cmdline_run("/usr/bin/coreadm %s" % os.getpid())
        if ret == 0:
            cores_pattern = out.strip().split()[1]
            cores_dir = os.path.dirname(cores_pattern)
            if cores_dir.startswith(self.ws):
                self.logger.log(
                    TINFO_LEVEL, "Clobbering old cores files from %s" %
                    cores_dir)
                for entry in os.listdir(cores_dir):
                    core_path = "%s/%s" % (cores_dir, entry)
                    if os.path.isfile(core_path):
                        try:
                            self.logger.log(
                                TINFO_LEVEL, "removing %s" % core_path)
                            os.remove(core_path)
                        except OSError as e:
                            self.logger.log(
                                TWARNING_LEVEL, e)


class ToolsTask(Task):
    """Perform a tools build for this workspace."""

    task_name = "tools"
    task_synopsis = "Do a 'dmake bldtools', for 'setup', " \
        "'install', and 'here'"
    produce_log_extract = True

    def _execute(self):
        env = self.bldenv.get_build_environment(
            debug=False, nondebug=True)

        ret = util.cmdline_run("dmake bldtools", env_arg=env,
                               logger=self.logger, cwd=self.src)

        self.take_snapshot(snap_suffix="%s" % self.name)

        self.exit_status = ret


class SetupTask(Task):
    """Perform a setup build for this workspace."""

    task_name = "setup"
    task_synopsis = "Do a 'dmake setup', required for 'install' and 'here'"
    produce_log_extract = True

    def _execute(self):
        ret = 0
        bldenv = self.bldenv

        def do_setup(debug=False):
            nondebug = not debug
            env = bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            return util.cmdline_run("dmake setup",
                                    env_arg=env, logger=self.logger,
                                    cwd=self.src)

        if bldenv.debug:
            ret += do_setup(debug=True)

        if bldenv.debug and bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug setup finished, starting non-debug setup")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TASK_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if bldenv.nondebug:
            ret += do_setup()

        self.exit_status = ret


class InstallTask(Task):
    """Perform an install build for this workspace."""

    task_name = "install"
    task_synopsis = "Build OS sources, writing to the proto area"
    produce_log_extract = True

    def _execute(self):

        # Sanity-check the build environment.
        env_errors = False

        # We can't have ON_CLOSED_BINS set in a Tonic build
        if "TONICBUILD" in self.env:

            if "ON_CLOSED_BINS" in self.env:
                self.logger.log(
                    TWARNING_LEVEL, "tonic builds must not have "
                    "on_closed_bins set.")
                env_errors = True

        if env_errors:
            self.exit_status = 1
            return

        def do_install(debug=False):
            """Run a dmake install for the current environment and generate
            license files if we're doing a 'tonic' build."""

            result = 0
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            ret = util.cmdline_run(
                "dmake install", env_arg=env, logger=self.logger, cwd=self.src)
            if ret != 0:
                result = 1
            return result

        if self.bldenv.debug:
            self.exit_status += do_install(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug install finished, starting non-debug "
                "install")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TASK_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            self.exit_status += do_install()

    def _execute_start(self):
        """Describe the environment in more detail for Install."""
        super(InstallTask, self)._execute_start()
        # After build(1) has been around for a while, we might be able to
        # remove this.
        for key in sorted(self.env):
            if key == "ONRC":
                continue
            self.logger.log(TINFO_LEVEL, "%s = %s" % (key, self.env[key]))


class FishInstallTask(Task):
    task_name = "fish.install"
    task_synopsis = "Build Fish sources, writing to the fish proto area"
    produce_log_extract = True

    def _execute(self):

        def do_install(debug=False):
            """Run a dmake install for the current environment and generate
            license files we're doing a 'tonic' build."""

            result = 0
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug, fish=True)

            ret = util.cmdline_run(
                "dmake -e install", env_arg=env, logger=self.logger,
                cwd=env["SRC"])
            if ret != 0:
                result = 1
            return result

        if self.bldenv.debug:
            self.exit_status += do_install(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug fish install finished, starting non-debug "
                "install")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TASK_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            self.exit_status += do_install()


class ManInstallTask(Task):
    """Perform a man install build for this workspace."""

    task_name = "man"
    task_synopsis = "Build OS man pages, writing to $MANROOT"
    produce_log_extract = True

    def _execute(self):

        if self.env.get("MAN_IS_PRESENT", "yes") == "no":
            self.logger.log(TINFO_LEVEL, "No man repo, skipping man build")
            self.exit_status = 0
            return

        man_repo = "{0}/usr/man".format(self.ws)
        ret = util.cmdline_run(
            "dmake install", env_arg=self.env, logger=self.logger,
            cwd=man_repo)
        if ret != 0:
            self.exit_status = 1


class FishPackagesTask(Task):
    task_name = "fish.packages"
    task_synopsis = "Build Fish IPS package archives"
    produce_log_extract = True

    def _execute(self):

        def do_pkgs(debug=False):
            """Run a dmake install in pkgdefs for the current environment."""

            result = 0
            ret = 0
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug, fish=True)

            pkg_dir = env["PKGARCHIVE"]
            if os.path.exists(pkg_dir):
                try:
                    self.logger.log(
                        TINFO_LEVEL, "removing existing package content "
                        "%s" % pkg_dir)
                    shutil.rmtree(pkg_dir)
                except OSError as e:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to remove existing "
                        "package content %s: %s" % (pkg_dir, e),
                        exc_info=True)
                    ret = 1

            ret += util.cmdline_run(
                "dmake -e mkfu", env_arg=env, logger=self.logger,
                cwd=env["SRC"])
            ret += util.cmdline_run(
                "dmake -e install", env_arg=env, logger=self.logger,
                cwd=os.path.join(env["SRC"], "pkgdefs"))
            if ret != 0:
                result = 1
            return result

        if self.bldenv.debug:
            self.exit_status += do_pkgs(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug fish packages finished, starting "
                "non-debug packages")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TASK_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            self.exit_status += do_pkgs()


def _extract_ak_tools(cpiodir, tmpdir, logger):
    """When we're building without a built usr/fish proto area
    (for example, during Solaris RE image construction) pull them from the
    cpio archives instead. This could use some finessing.
    """

    ak_root = os.path.join(tmpdir, "ak_root")
    if not os.path.exists(ak_root):
        os.makedirs(ak_root)

    generic_files = [
        "usr/lib/ak/tools/*",
        "usr/lib/ak/licenses/*",
        "usr/bin/akfwlist",
        "usr/bin/aknv",
        "*libnvxml*"]

    # A map of cpio archive names to the content we need from them, and where
    # that content usually resides, relative to $ROOT
    cpio_map = {
        # ak_dc_pkgimg_ckpt wants to get KernelZone-hc-topology.xml
        # and expects to see it in $ROOT/appliance/SUNW,gourami
        # Same for OCI-hc-topology.xml in $ROOT/appliance/SUNW,nori
        "SUNW,gourami.proto.gz": (
            ["usr/platform/*"], "appliance/SUNW,gourami"),
        "SUNW,nori.proto.gz": (
            ["usr/platform/*"], "appliance/SUNW,nori"),
        "generic.usr.gz": (generic_files, ".")
    }

    ret = 0
    for archive in cpio_map:
        root_dest = os.path.join(ak_root, cpio_map[archive][1])
        try:
            os.makedirs(root_dest)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        cmd = "pigz -dc {0}/{1} | cpio -idmu {2}".format(
            cpiodir, archive, " ".join(cpio_map[archive][0]))
        logger.log(TINFO_LEVEL, "Extracting AK tools to {0}".format(ak_root))
        ret += util.cmdline_run(
            cmd, cwd=root_dest, logger=logger, shell=True)

    # sanity-check
    mkak_path = os.path.join(ak_root, "usr/lib/ak/tools/mkak")
    if not os.path.exists(mkak_path) or ret != 0:
        raise Exception("Error extracting temporary AK tooling!")
    return ak_root


class FishMkakTask(Task):
    """This build task is capable of invoking mkak(1) to produce ZFSSA Upgrade
    packages for the 'nas' product. Subclasses of this class produce AI iso
    images and Text iso images.

    Publisher configuration for the built image comes from the 'fw_pkg_proto'
    IPS user-image."""

    task_name = "fish.images"
    task_synopsis = "Build 'nas' Fish upgrade images only"
    produce_log_extract = False
    check_tools = False
    # Specify the mkak flags we want this task to run. The default flavor
    # just produces AK upgrade images.
    additional_mkak_args = []

    def _execute(self):

        def get_ips_configuration(fw_pkg_proto):
            """Get publisher configuration from directory containing
            an IPS user-image. Return an OrderedDict of the form:
            {publisher: [(origin, proxy), ..], ... }

            Note that the proxy support here is in hopes we get a fix
            for 15808976 soon.
            """
            cmd = "pkg -R {0} publisher -Hn -F tsv".format(fw_pkg_proto)
            self.logger.log(TINFO_LEVEL, cmd)
            ret, out, err = util.cmdline_run(cmd)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Error retrieving publisher "
                    "information from {0} !".format(fw_pkg_proto))
                return []

            publishers = collections.OrderedDict()
            # we get tab-separated values in the following order:
            # PUBLISHER	STICKY	SYSPUB	ENABLED	TYPE	STATUS	URI	PROXY
            for line in out.strip().split("\n"):
                toks = line.split()
                if len(toks) < 8:
                    continue
                pub = toks[0]

                # We never use the 'extra' publisher from the fw_pkg_proto.
                # The appliance should not be using packages from here anyway,
                # but when fw_pkg_proto=/, we would otherwise get the WOS
                # 'extra' publisher from build machines (running WOS bits)
                if pub == "extra":
                    continue

                origin = toks[6]
                proxy = toks[7] if toks[7] != "-" else "None"
                publishers.setdefault(pub, []).append((origin, proxy))

            return publishers

        def save_and_cleanup(env, debug=False):
            """If we're being run as standalone task, copy data into the
            PKGARCHIVE and destroy the DC dataset now. 'nightly' normally
            does this as part of 'fish.save_artifacts' and
            'fish.destroy_dc' to allow the potential of reusing the data
            in the DC area for the assembly of multiple images at some point,
            but for single-task invocations ("build fish.ai_iso"), this
            is more convenient for the user.
            """
            # If there is more than one task in this build invocation,
            # then we're not running standalone.
            known_tasks = self.get_known_tasks()
            if len(known_tasks) > 1 or "ONBLD_PRESERVE_DC_DATASET" in env:
                return 0

            runtime_bldenv = copy.deepcopy(self.bldenv)
            if debug:
                runtime_bldenv.debug = True
                runtime_bldenv.nondebug = False
            else:
                runtime_bldenv.debug = False
                runtime_bldenv.nondebug = True
            self.logger.log(
                TINFO_LEVEL,
                "non-composite job running - saving artifacts to %s" %
                env["PKGARCHIVE"])
            save_artifacts = FishSaveArtifactsTask(runtime_bldenv)
            save_artifacts.run()
            if save_artifacts.exit_status != 0:
                self.logger.log(
                    TWARNING_LEVEL, "save_artifacts failed - bailing out now")
                return save_artifacts.exit_status

            # The FishLicensesTask needs the dc dataset to be present.
            if env["FISH_LICENSES"] == "true":
                return 0

            self.logger.log(TINFO_LEVEL, "destroying DC dataset now")
            return destroy_dc_dataset(
                self.env["DC_DATASET"], self.logger)

        def do_mkak(debug=False):
            """
            Run mkak for this build flavor.

            Where possible, we derive all IPS configuration from the
            pkg(8) configuration in 'fw_pkg_proto' rather than relying on
            site-specific conf files.

            This decouples this part of the build from hardcoded dependencies
            on http://ipkg. We can later spend more time on mkakproto and have
            it create a sensible configuration by default.
            """

            # since we always attempt to run mkak from the proto area,
            # this will never work when running the test suite.
            # return immediately if so.
            ret, out, err = util.cmdline_run(
                "/usr/bin/pargs %s" % os.getppid(), shell=True)
            if "buildtest.py" in out:
                return 0

            mkak_args = []
            label = "" if debug else "-nd"

            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug, fish=True)
            onenv = None
            if env.get("FISH_STANDALONE"):
                env["ONROOT"] = os.path.join(
                    env["FW_HOOKS"], "data/proto/root_{0}{1}".format(
                        self.mach, label))
            else:
                onenv = self.bldenv.get_build_environment(
                    debug=debug, nondebug=not debug)
                env["ONROOT"] = onenv["ROOT"]

            # Check that we can see AK CPIO archives before going any further.
            sample_cpio = os.path.join(env["CPIODIR"], "generic.usr.gz")
            if not os.path.exists(sample_cpio):
                self.logger.log(
                    TWARNING_LEVEL,
                    "Cpio archive missing at %s. Unable to complete "
                    "image assembly" % sample_cpio)
                return 1

            mkak_args.append("-z {0}".format(env["DC_DATASET"]))
            mkak_args.append("-f {0}".format(os.path.abspath(env["CPIODIR"])))
            mkak_args.append("-v")

            if debug:
                mkak_args.append("-d")

            # If this command fails, mkak will bail out later anyway.
            # Add onbld:dataset to distinguish from manual ak:build=mkak runs.
            cmd = "/usr/sbin/zfs create -o onbld:dataset=true -p {0}".format(
                env["DC_DATASET"])
            ret = util.cmdline_run(cmd, logger=self.logger)

            # Determine the pkg(7) origins for the eventual AK image.
            # ON origins:
            if onenv:
                pkgarchive = onenv["PKGARCHIVE"]
            else:
                if debug:
                    pkgdir = "{0}-debug".format(self.mach)
                else:
                    pkgdir = "{0}-nondebug".format(self.mach)

                # Look for packages in the FW_HOOKS location, assuming it's
                # a built ON workspace.
                pkgarchive = os.path.join(
                    env["FW_HOOKS"], "data", "pkg", pkgdir)

            # We prefer local ON packages, so set those first.
            if os.path.exists(pkgarchive):
                nightly = os.path.join(pkgarchive, "repo.osnet")
                extra = os.path.join(pkgarchive, "repo.osnet-internal")
                mkak_args.append(
                    "-o '{0} {1} None'".format("nightly", nightly))
                mkak_args.append("-o '{0} {1} None'".format("extra", extra))

            pub_config = get_ips_configuration(env["FW_PKG_PROTO"])

            # A potential gotcha for the unwary
            if "nightly" in pub_config and os.path.exists(pkgarchive):
                origins = [origin for origin, proxy in pub_config["nightly"]]

                # If ON gate machines are running 'nightly' bits on their
                # build machines, assume the gatekeepers know what they're
                # doing and lower the verbosity of this message.
                if env["FW_PKG_PROTO"] == "/":
                    log_level = TINFO_LEVEL
                else:
                    log_level = TWARNING_LEVEL
                self.logger.log(
                    log_level, "'fw_pkg_proto' origins were found that "
                    "provide 'nightly' packages.")
                self.logger.log(
                    log_level, "We also found ON packages at {0}".format(
                        pkgarchive))
                self.logger.log(
                    log_level, "Unless this was intentional, make sure "
                    "that the ON packages in this workspace are at versions "
                    "greater than those found at:\n{0}".format(
                        "\n".join(origins)))

            # If we don't have 'solarisstudio' and 'ak' publishers, mkak will
            # supply reasonable defaults, we warn if 'solaris' is not present.
            required_origins = ["solaris"]
            for pub in required_origins:
                if pub not in pub_config:
                    self.logger.log(
                        TWARNING_LEVEL, "No origins configured for '{0}' at "
                        "{1}".format(pub, env["FW_PKG_PROTO"]))

            for pub in pub_config:
                # special-case on-onbld. It's typically configured on build
                # machines, but should never get the chance of contributing
                # packages to the appliance.
                if pub == "on-onbld":
                    continue
                for origin, proxy in pub_config[pub]:
                    mkak_args.append("-o '{0} {1} {2}'".format(
                        pub, origin, proxy))

            mkak_path = "{0}/usr/lib/ak/tools/mkak".format(env["ROOT"])
            if not os.path.exists(mkak_path):

                self.logger.warning("")
                self.logger.warning("***************************************")
                self.logger.warning("This script is being run outside the")
                self.logger.warning("confines of an ON workspace!")
                self.logger.warning("This is typically used only by Release")
                self.logger.warning("Engineering!")
                self.logger.warning("Hit Ctrl-C unless this was intentional.")
                self.logger.warning("***************************************")
                self.logger.warning("")

                try:
                    ak_root = _extract_ak_tools(
                        env["CPIODIR"], env["TMPDIR"], self.logger)
                except Exception as e:
                    self.logger.log(TWARNING_LEVEL, e)
                    return 1

                # We assume we've booted a system with the right bits
                # to run distro_const with. mkak uses ONROOT to pull the
                # reference DC manifests.
                env["ONROOT"] = "/"
                env["ROOT"] = ak_root
                mkak_path = os.path.join(ak_root, "usr/lib/ak/tools/mkak")

            cmd = "{0} {1}".format(mkak_path, " ".join(mkak_args))

            cmd += " " + " ".join(self.additional_mkak_args)
            ret = util.cmdline_run(cmd, logger=self.logger, env_arg=env)
            ret += save_and_cleanup(env, debug=debug)
            return ret

        if self.bldenv.debug:
            self.exit_status += do_mkak(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug fish images finished, starting "
                "non-debug images")

        if self.bldenv.nondebug:
            self.exit_status += do_mkak()


class FishMkakGkImagesTask(CompositeTask):
    task_name = "fish.gk_images"
    task_synopsis = \
        "Build Fish images appropriate for gk builds"
    produce_log_extract = False
    check_tools = False

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        super(FishMkakGkImagesTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)

        self.add_task(FishMkakNasTask(bldenv))


class FishMkakGkNightlyImagesTask(FishMkakTask):
    task_name = "fish.gk_nightly_images"
    task_synopsis = "Build gk nightly images (nas upgrade & AI iso)"
    produce_log_extract = False
    additional_mkak_args = ["-N"]


class FishMkakJdsImagesTask(FishMkakTask):
    task_name = "fish.jds_images"
    task_synopsis = "Build 'jds' Fish upgrade images only"
    additional_mkak_args = ["-p jds"]


class FishMkakTextIsoTask(FishMkakTask):
    task_name = "fish.txt_iso"
    task_synopsis = "Build 'nas' Fish text iso images only"
    additional_mkak_args = ["-I", "-U"]


class FishMkakAiIsoTask(FishMkakTask):
    task_name = "fish.ai_iso"
    task_synopsis = "Build 'nas' Fish AI iso images only"
    additional_mkak_args = ["-N", "-U"]


class FishMkakJdsTextIsoTask(FishMkakTask):
    task_name = "fish.jds_txt_iso"
    task_synopsis = "Build 'jds' Fish text iso images only"
    additional_mkak_args = ["-p jds", "-I", "-U"]


class FishMkakJdsAiIsoTask(FishMkakTask):
    task_name = "fish.jds_ai_iso"
    task_synopsis = "Build 'jds' Fish AI iso images only"
    additional_mkak_args = ["-p jds", "-N", "-U"]


class FishMkakNasTask(FishMkakTask):
    task_name = "fish.nas_all_images"
    task_synopsis = "Build all 'nas' Fish images"
    additional_mkak_args = ["-I", "-N"]


class FishMkakJdsGkTask(FishMkakTask):
    task_name = "fish.jds_gk_images"
    task_synopsis = "Build 'jds' Fish images appropriate for gk builds"
    produce_log_extract = False
    additional_mkak_args = ["-p jds", "-N"]


class FishMkakJdsAllTask(FishMkakTask):
    task_name = "fish.jds_all_images"
    task_synopsis = "Build all 'jds' Fish images"
    produce_log_extract = False
    additional_mkak_args = ["-p jds", "-I", "-N"]


class FishMkakConfTask(FishMkakTask):
    task_name = "fish.conf"
    task_synopsis = "Verify mkak options"
    additional_mkak_args = ["-c", "-U"]


def _save_artifacts(mkak_out, pkgarchive, logger, onroot="/"):
    """Process mkak -O stdout and save discovered artifacts to $PKGARCHIVE."""

    ret = 0
    media_paths = {}

    for line in mkak_out.split("\n"):
        toks = line.split(":", 1)
        if len(toks) != 2:
            continue
        media_paths[toks[0]] = toks[1].strip()

    for media in ["ai_iso", "text_usb", "text_iso", "update_media"]:
        if media not in media_paths:
            logger.log(
                TINFO_LEVEL, "{0} does not exist...".format(media))
            continue
        path = media_paths[media]
        if os.path.exists(path):
            dest = os.path.join(pkgarchive, os.path.basename(path))

            if os.path.exists(dest):
                logger.log(
                    TINFO_LEVEL, "Removing existing {0}".format(dest))
                os.remove(dest)

            logger.log(
                TINFO_LEVEL, "Copying {0} {1} to {2}".format(
                    media, path, dest))
            try:
                if not os.path.exists(pkgarchive):
                    os.makedirs(pkgarchive)

                shutil.copy(path, dest)

                # As a developer convenience, we add the osnet file
                # to built images. We don't do this for officially
                # released images where we can't easily tell where the
                # etc/versions/osnet file should come from when there
                # is no ON $ROOT directory present.
                osnet_path = os.path.join(onroot, "etc/versions/osnet")
                if onroot != "/" and os.path.exists(osnet_path):
                    cmd = "runat {0} cp {1} osnet".format(dest, osnet_path)
                    util.cmdline_run(cmd, logger=logger)
            except OSError as e:
                logger.log(
                    TWARNING_LEVEL, "Unable to copy {0} artifacts from "
                    "{1} to {2}: {3}".format(media, path, dest, e))
                ret = 1
    return ret


class FishLicensesTask(Task):
    """The FishLicensesTask handles generating the AK LICENSES file."""
    task_name = "fish.licenses"
    task_synopsis = "Generate a fish LICENSES file from \'dc_dataset\' "
    produce_log_extract = False
    check_tools = False

    def _execute(self):
        # We always attempt to run from the proto area, and this will never
        # work when running the test suite.  Return immediately if so.
        ret, out, err = util.cmdline_run("/usr/bin/pargs {}"
                                         .format(os.getppid(), shell=True))
        if "buildtest.py" in out:
            return

        try:
            if self.bldenv.debug:
                self.logger.log(TINFO_LEVEL, ("Starting debug AK LICENSES "
                                              "file generation."))
                self.__setup(True)
                self.__generate_licenses()
            if self.bldenv.nondebug:
                self.logger.log(TINFO_LEVEL, ("Starting nondebug AK LICENSES "
                                              "file generation."))
                self.__setup(False)
                self.__generate_licenses()
            self.exit_status = 0
        except Exception as e:
            self.logger.log(TWARNING_LEVEL, e)
            self.exit_status = 1

    def __set_env(self, debug):
        """Set the task environment."""
        self.logger.log(TASK_LEVEL,
                        ("setting task environment to fish {}"
                         .format("debug" if debug else "nondebug")))
        self.env = self.bldenv.get_build_environment(debug=debug,
                                                     nondebug=not debug,
                                                     fish=True)

    def __set_fish_root_path(self):
        """Get the path to the fish proto root or the extracted tools
        root."""
        if "ROOT" in self.env and os.path.exists(self.env["ROOT"]):
            self.logger.log(TINFO_LEVEL, "ROOT: {}".format(self.env["ROOT"]))
            return

        self.logger.log(TWARNING_LEVEL,
                        ("Detected environment where we are running outside of"
                         " a \"normal\" workspace.  This is normally a RE "
                         "build, but if not, we have an issue."))
        ak_root = _extract_ak_tools(self.env["CPIODIR"], self.env["TMPDIR"],
                                    self.logger)
        self.env["ROOT"] = ak_root
        self.logger.log(TINFO_LEVEL, "ROOT: {}".format(self.env["ROOT"]))

    def __set_mkak_path(self):
        """Set the AK_MKAK_PATH in the environment."""
        mkak_path = os.path.join(self.env["ROOT"], "usr/lib/ak/tools/mkak")
        if not os.path.exists(mkak_path):
            raise TaskException("mkak not located at {}".format(mkak_path))
        self.env["AK_MKAK_PATH"] = mkak_path
        self.logger.log(TINFO_LEVEL,
                        "AK_MKAK_PATH set to {}".format(mkak_path))

    def __set_aklicenses_path(self):
        """Set the AK_LICENSES_PATH in the environment."""
        akl_path = os.path.join(self.env["ROOT"],
                                "usr/lib/ak/tools/aklicenses")
        if not os.path.exists(akl_path):
            raise TaskException("aklicenses not found at {}".format(akl_path))
        self.env["AK_LICENSES_PATH"] = akl_path
        self.logger.log(TINFO_LEVEL,
                        "AK_LICENSES_PATH set to {}".format(akl_path))

    def __set_ak_media_path(self):
        """Retrieve the location of the ak media.  This must have been
        generated by the FishMkakTask previously."""
        mkak_path = self.env["AK_MKAK_PATH"]
        mkak_args = []
        mkak_args.append("-p nas")
        mkak_args.append("-O")
        mkak_args.append("-I")
        mkak_args.append("-N")
        mkak_args.append("-o None")
        mkak_args.append("-z {0}".format(self.env["DC_DATASET"]))
        mkak_args.append("-f {0}".format(self.env["CPIODIR"]))
        cmd = "{0} {1}".format(mkak_path, " ".join(mkak_args))
        self.logger.log(TASK_LEVEL, "Running {}".format(cmd))
        ret, out, err = util.cmdline_run(cmd, env_arg=self.env)
        self.logger.log(TASK_LEVEL,
                        ("mkak cmd: [{}]\nmkak out: [{}]\nmkak err: [{}]"
                         .format(cmd, out, err)))
        if ret != 0:
            raise TaskException("mkak failed\n[{}]\n[{}]\n[{}]"
                                .format(cmd, out, err))

        media_paths = dict()
        for line in out.split("\n"):
            tokens = line.split(":", 1)
            if len(tokens) != 2:
                continue
            media_paths[tokens[0]] = tokens[1].strip()

        osgz_path = os.path.join(media_paths["ak_media"], "os.gz")
        if not os.path.exists(osgz_path):
            raise TaskException("os.gz not found at {}".format(osgz_path))

        self.env["AK_MEDIA_PATH"] = media_paths["ak_media"]
        self.env["AK_ARCHIVE_OS_PATH"] = osgz_path
        self.logger.log(TINFO_LEVEL,
                        ("AK_MEDIA_PATH set to {}, "
                         "AK_ARCHIVE_OS_PATH set to {}"
                         .format(media_paths["ak_media"], osgz_path)))

    def __setup(self, debug):
        self.__set_env(debug)
        self.__set_fish_root_path()
        self.__set_mkak_path()
        self.__set_aklicenses_path()
        self.__set_ak_media_path()
        self.__setup_complete = True

    def __generate_licenses(self):
        if not self.__setup_complete:
            raise TaskException("Task setup not complete.  This is most likely"
                                " a programmer error.")

        if not os.path.exists(self.env["PKGARCHIVE"]):
            raise TaskException("PKGARCHIVE ({}) does not exist"
                                .format(self.env["PKGARCHIVE"]))

        aklicenses_path = self.env["AK_LICENSES_PATH"]
        fish_path = self.env["ROOT"]
        dest = os.path.join(self.env['PKGARCHIVE'], "LICENSES")
        os_archive_path = self.env["AK_ARCHIVE_OS_PATH"]
        cmd = "{} {} false {} {}".format(aklicenses_path, fish_path, dest,
                                         os_archive_path)
        self.logger.log(TASK_LEVEL, "Running {}".format(cmd))
        ret, out, err = util.cmdline_run(cmd)
        self.logger.log(TASK_LEVEL,
                        ("aklicenses cmd: [{}]\nout: [{}]\nerr: [{}]"
                         .format(cmd, out, err)))
        if ret != 0:
            raise TaskException("Failed to generate fish LICENSES file.\n"
                                "[{}]\n[{}]\n[{}]".format(cmd, out, err))

        self.logger.log(TINFO_LEVEL,
                        ("AK LICENSE file generation complete.  Results "
                         "found at {}.".format(dest)))


class FishSaveArtifactsTask(Task):
    task_name = "fish.save_artifacts"
    task_synopsis = "Save all build artifacts from the 'dc_dataset' directory"
    produce_log_extract = False
    check_tools = False

    def _execute(self):

        def do_save_artifacts(debug=False):

            # since we always attempt to run mkak from the proto area,
            # this will never work when running the test suite.
            # return immediately if so.
            ret, out, err = util.cmdline_run(
                "/usr/bin/pargs %s" % os.getppid(), shell=True)
            if "buildtest.py" in out:
                return 0

            label = "" if debug else "-nd"
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug, fish=True)
            onenv = None
            if env.get("FISH_STANDALONE"):
                env["ONROOT"] = os.path.join(
                    env["FW_HOOKS"], "data/proto/root_{0}{1}".format(
                        self.mach, label))
            else:
                onenv = self.bldenv.get_build_environment(
                    debug=debug, nondebug=not debug)
                env["ONROOT"] = onenv["ROOT"]

            mkak_path = os.path.join(env["ROOT"], "usr/lib/ak/tools/mkak")
            if not os.path.exists(mkak_path):
                try:
                    ak_root = _extract_ak_tools(
                        env["CPIODIR"], env["TMPDIR"], self.logger)
                except Exception as e:
                    self.logger.log(TWARNING_LEVEL, e)
                    return 1

                env["ONROOT"] = "/"
                env["ROOT"] = ak_root
                mkak_path = os.path.join(ak_root, "usr/lib/ak/tools/mkak")

            for product in ["nas"]:
                mkak_args = []
                mkak_args.append("-p {0}".format(product))
                mkak_args.append("-O")
                mkak_args.append("-I")
                mkak_args.append("-N")
                # We don't need to compute publisher information for this job
                # so just pass a dummy to appease mkak.
                mkak_args.append("-o None")
                mkak_args.append("-z {0}".format(env["DC_DATASET"]))
                mkak_args.append("-f {0}".format(env["CPIODIR"]))
                mkak_path = "{0}/usr/lib/ak/tools/mkak".format(env["ROOT"])
                cmd = "{0} {1}".format(mkak_path, " ".join(mkak_args))
                ret, out, err = util.cmdline_run(cmd, env_arg=env)
                if ret != 0:
                    if not os.path.exists(mkak_path):
                        self.logger.log(
                            TWARNING_LEVEL, "{0} does not exist".format(
                                mkak_path))
                        return 1
                    self.logger.log(TWARNING_LEVEL, "mkak did not return 0")
                    self.logger.log(TWARNING_LEVEL, cmd)
                    self.logger.log(TWARNING_LEVEL, out)
                    self.logger.log(TWARNING_LEVEL, err)
                    return 1
                ret += _save_artifacts(
                    out, env["PKGARCHIVE"], self.logger, env["ONROOT"])
            return ret

        if self.bldenv.debug:
            self.exit_status += do_save_artifacts(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Processing non-debug artifacts..")

        if self.bldenv.nondebug:
            self.exit_status += do_save_artifacts()


def destroy_dc_dataset(dataset, logger):
    """This is obviously a destructive action: if we have content at or
    beneath 'dc_dataset', that will be removed by this task. We use recursive
    destruction, destroying the top-most dataset we find and issue
    at most a single 'zfs destroy' command.

    This depends on the fact that mkak makes its own top-level 'container'
    dataset for each set of images it produces.
    """

    # build(1)-created datasets have the property onbld:dataset=true
    cmd = "/usr/sbin/zfs list -H -r -o name,onbld:dataset {0}".format(dataset)
    ret, out, err = util.cmdline_run(cmd)

    # if anything went wrong, bail out immediately, but not in a way that
    # causes the build to stop.
    if ret != 0:
        logger.log(
            TINFO_LEVEL, "Unable to list dataset {0}: {1}\n{2}\n{3}".format(
                dataset, cmd, out, err))
        return 0

    # recursively destroy data beneath our dataset
    for line in out.split("\n"):
        if not line.endswith("\ttrue"):
            continue
        toks = line.split()
        cmd = "/usr/sbin/zfs destroy -r {0}".format(toks[0])
        ret = util.cmdline_run(cmd, logger=logger)
        if ret != 0:
            logger.log(
                TWARNING_LEVEL,
                "Unable to destroy dataset using '{0}'\n{1}\n{2}".format(
                    cmd, out, err))
        return ret
    return 1


class FishDestroyDcDataset(Task):
    """This task runs after we've recovered all build products from the
    DC_DATASET dataset. If any of the FishMKakTasks fail, they'll cause the
    build to bail out, and we intentionally won't run this task, allowing
    developers to investigate the cause of the failure.

    When a new build is attempted, ClobberTask will ensure that any remaining
    DC_DATASET content is removed, assuming the DC_DATASET value hasn't been
    changed between builds.
    """

    task_name = "fish.destroy_dc"
    task_synopsis = \
        "Remove datasets tagged 'onbld:dataset' at/under 'dc_dataset'"
    produce_log_extract = False
    check_tools = False

    def _execute(self):
        if "ONBLD_PRESERVE_DC_DATASET" in self.env:
            self.logger.log(
                TINFO_LEVEL,
                "ONBLD_PRESERVE_DC_DATASET was set, not destroying.")
            return
        self.exit_status = destroy_dc_dataset(
            self.env["DC_DATASET"], self.logger)


class FishReBuildTask(CompositeTask):
    task_name = "fish.re_build"
    task_synopsis = \
        "Runs AK image construction tasks for Release Engineering"
    produce_log_extract = False
    check_tools = False

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        #
        # An RE build allows us to assemble the AK images without requiring
        # a full ON source workspace to be present.
        # To do this type of build, create a directory ./data, populated with
        # the $WS/data/pkg and $WS/data/archives directories from a full ON
        # build that includes the AK bits, then either do:
        #
        # build generate -D ws=`pwd` > bldrc
        #
        # or setup an fw_pkg_proto user image that points to the package
        # repositories containing integrated packages from all consolidations,
        # copy the $WS/data/archives directory as above, then do:
        #
        # build generate -D ws=`pwd` -D fw_pkg_proto=`pwd`/fw_pkg_proto > bldrc
        #
        # Then build using "build fish.re_build". This results in the AK
        # tools being extracted from ./data/archives/$MACH-<flavour>,
        # the build creates a dummy workspace in /tmp, and image-assembly
        # proceeds as usual.
        #
        super(FishReBuildTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)
        self.add_task(FishMkakGkImagesTask(bldenv))
        self.add_task(FishSaveArtifactsTask(bldenv))
        if self.env["FISH_LICENSES"] == "true":
            self.add_task(FishLicensesTask(bldenv))
        self.add_task(FishDestroyDcDataset(bldenv))


def _add_repository_metadata(
        pkg_dir, ws, hg_info, env, logger, repo=None, build_flavor="unknown"):
    """Write some metadata to the standard ON package repositories in
    pkg_dir if they exist.
    If pkg_dir is None, use the repo argument, which is the full path to
    a package repository."""

    # our changeset value can be returned in the form
    # "<hg hash> <tag>", so split on spaces to get the node.
    src_changeset = hg_info.get(ws, ["None"])[0].split(" ")[0]
    closed_changeset = hg_info.get(
        "%s/usr/closed" % ws, ["None"])[0].split(" ")[0]
    man_changeset = hg_info.get(
        "%s/usr/man" % ws, ["None"])[0].split(" ")[0]
    fish_changeset = hg_info.get(
        "%s/usr/fish" % ws, ["None"])[0].split(" ")[0]
    build_machine = socket.gethostname()

    repos = []
    if pkg_dir:
        for repository in ["repo.osnet", "repo.osnet-internal"]:
            dest_repo = os.path.join(pkg_dir, repository)
            if os.path.isdir(dest_repo):
                repos.append(dest_repo)
    else:
        # this indicates we've called this method incorrectly
        if not repo:
            raise ValueError(
                "pkg_dir or repo must be set when calling this function")
        repos = [repo]

    if env.get("FISH_IS_PRESENT", "no") == "yes":
        osnet_flavor = "default"
    else:
        osnet_flavor = "server"

    pairs = [
        ("osnet/build-machine", build_machine),
        ("osnet/version", env["VERSION"]),
        ("osnet/src-path", env["SRC"]),
        ("osnet/src-changeset", src_changeset),
        ("osnet/closed-changeset", closed_changeset),
        ("osnet/man-changeset", man_changeset),
        ("osnet/osnet-flavor", osnet_flavor),
        ("osnet/build-flavor", build_flavor)]

    if osnet_flavor == "default":
        pairs.append(("osnet/fish-changeset", fish_changeset))

    for path in repos:
        cmd = "pkgrepo -s {0} set {1}".format(
            path, " ".join(
                ["{0}={1}".format(key, value) for key, value in pairs]))
        util.cmdline_run(cmd, env_arg=env, logger=logger)


class PackagesTask(Task):
    """Performs a packages build for this workspace."""

    task_name = "packages"
    task_synopsis = "Publish packages to local pkg(7) repositories"
    produce_log_extract = True
    mail_msg_summary_fragsize = 40

    def _execute(self):
        """Remove existing PKGARCHIVE dir and publish packages."""

        if "PKGARCHIVE" not in self.env:
            self.logger.log(
                TINFO_LEVEL, "Not publishing packages since PKGARCHIVE was "
                "not set")
            return
        # temporarily switch to a html logging.Formatter better suited to deal
        # with the usr/src/pkg build.
        saved_html_formatter = None
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler) and \
                    handler.stream.name.endswith(".html"):
                saved_html_formatter = handler.formatter
                handler.formatter = log.PkgHtmlFormatter(datefmt=log.datefmt)

        # Used later to write package repository metadata
        hg_info = get_hg_info(self.ws, self.env)

        def do_pkgs(debug=False):
            """Publish debug or nondebug packages."""

            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            ret = 0

            # remove existing package repository
            pkg_dir = env["PKGARCHIVE"]
            temp_pkgs_dir = os.path.join(
                self.src, "pkg", "packages.%s" % env["MACH"])
            for path in [pkg_dir, temp_pkgs_dir]:
                try:
                    if not os.path.exists(path):
                        continue
                    self.logger.log(
                        TINFO_LEVEL, "removing existing package content "
                        "%s" % path)
                    shutil.rmtree(path)
                except OSError as e:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to remove existing "
                        "package content %s: %s" % (path, e),
                        exc_info=True)
                    ret += 1

            # publish the packages
            ret += util.cmdline_run(
                "dmake install", env_arg=env, logger=self.logger,
                cwd=os.path.join(self.src, "pkg"))
            _add_repository_metadata(
                pkg_dir, self.ws, hg_info, self.env, self.logger,
                build_flavor="debug" if debug else "nondebug")

            return ret

        if self.bldenv.debug:
            self.exit_status += do_pkgs(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug packages published, starting non-debug "
                "pkgs")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TASK_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            self.exit_status += do_pkgs()

        # replace the default formatter
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler) and \
                    handler.stream.name.endswith(".html"):
                handler.formatter = saved_html_formatter

    def summary(self, max_lines=60):
        """Use a custom summary for packages, since failures are typically
        hidden by lots of noise."""

        if self.exit_status == 0:
            return ""

        logfile_path = os.path.join(self.log_dir, self.log_name)
        if not os.path.exists(logfile_path):
            return ""

        short_summary = []
        previous_line = ""

        cols = log.log_prefix_cols[self.bldenv.log_format] - 1
        with open(logfile_path, "r") as logfile:
            for full_line in open(logfile_path, "r").readlines():
                line = full_line[cols:]
                if line.startswith("info: "):
                    if not previous_line.startswith("info: ") and \
                            previous_line:
                        short_summary.append("")
                    previous_line = line
                    continue

                if previous_line.startswith("info: "):
                    short_summary.append(previous_line.strip())

                short_summary.append(line.strip())
                previous_line = line

                if len(short_summary) > max_lines:
                    short_summary.append(".\n.\n(additional lines truncated)")
                    break

        return "\n".join(short_summary)


class ZipTask(Task):
    """Produce zipfiles for integration.
    These will be in a format suitable for RE."""

    task_name = "zipfiles"
    task_synopsis = "Create Zip files for RE integration"
    snapshot = False
    check_tools = False

    def do_zip(self, zipfile, dir):
        self.logger.log(
            TINFO_LEVEL, "Creating ZIP file %s.zip from %s" %
            (zipfile, dir))
        shutil.make_archive(zipfile, "zip", dir)

    def _execute(self):
        # The names of the zip files to produce, without the trailing ".zip"
        # which shutil will add for us (except for the foss archive, which is
        # just a plain copy).
        __ON_SERVER_ZIP = "osnet-server"
        __ON_STORAGE_ZIP = "osnet-storage"
        __ON_AK_ZIP = "osnet-ak"
        __ON_FOSS_ZIP = "osnet.zip"

        zdir = self.env['ZIPDIR']

        pkgmerge_type, pkgmerge_dest, \
            debug_pkgs_local, debug_pkgs_altws, \
            nondebug_pkgs_local, nondebug_pkgs_altws = \
            _compute_pkgmerge_params(self.bldenv)

        if not debug_pkgs_local or not debug_pkgs_altws or \
                not nondebug_pkgs_local or not nondebug_pkgs_altws or \
                not self.bldenv.pkgmerge:
            self.logger.log(TWARNING_LEVEL, "Zip files should be created with "
                            "debug and non-debug bits pkgmerged from both "
                            "architectures.")

        repodir = os.path.join(pkgmerge_dest, "repo.osnet")

        if self.env.get("FISH_IS_PRESENT") == "yes":
            # Zip up ON (default) IPS packages
            try:
                self.do_zip(os.path.join(zdir, __ON_STORAGE_ZIP), repodir)
            except FileNotFoundError:
                self.exit_status = 1
                self.logger.log(TWARNING_LEVEL,
                                "ERROR: AK ON IPS repo not present")

            # Zip up AK CPIO archives
            fishenv = self.bldenv.get_build_environment(fish=True)
            cpiodir = fishenv['AKARCHIVEDIR']
            try:
                self.do_zip(os.path.join(zdir, __ON_AK_ZIP), cpiodir)
            except FileNotFoundError:
                self.exit_status = 1
                self.logger.log(TWARNING_LEVEL,
                                "ERROR: AK CPIO archives not present")
        else:
            # Zip up ON (server) IPS packages
            try:
                self.do_zip(os.path.join(zdir, __ON_SERVER_ZIP), repodir)
            except FileNotFoundError:
                self.exit_status = 1
                self.logger.log(TWARNING_LEVEL,
                                "ERROR: ON IPS repo not present")

        # Copy FOSS archive from its location in the build area.
        foss_src = os.path.join(self.env['TOOLBUILD'], "usr", "src", "tools",
                                "foss", "osnet.zip")
        foss_dst = os.path.join(zdir, __ON_FOSS_ZIP)
        self.logger.log(TINFO_LEVEL, "Copying %s to %s" % (foss_src, foss_dst))
        try:
            shutil.copy(foss_src, foss_dst)
        except FileNotFoundError:
            self.exit_status = 1
            self.logger.log(TWARNING_LEVEL,
                            "ERROR: FOSS archive not present")


class TstampTask(Task):
    """Timestamps the start of the normal build; the findunref tool and
    parfait incremental build use it.  This doesn't do much work, so we
    don't bother taking a snapshot before proceeding."""

    task_name = "tstamp"
    task_synopsis = "Update a build timestamp file"
    snapshot = False

    def _execute(self):
        tstamp_path = "%s/.build.%s.tstamp" % (self.keep, self.mach)
        hg_info = get_hg_info(self.ws, self.env)
        with open(tstamp_path, "w") as tfile:
            for repo in hg_info:
                tfile.write("%s %s\n" % (repo, hg_info[repo][0]))
            os.utime(tstamp_path, None)

        #
        # Creating the timestamp happens very early in the build, early enough
        # that we don't know if the build in the remote workspace has had a
        # chance to start yet, and we could end up reading a timestamp from
        # an older build
        #
        # Rather than drive on and risk pkgmerge failing because the packages
        # built in this workspace don't match those build in the remote
        # workspace, we introduce a delay here for the workspaces to become
        # in sync. (i.e. hg update in the remote workspace hasn't completed
        # there yet)
        #
        # Yay, distributed systems.
        #
        if self.bldenv.pkgmerge and self.env.get("ALTWS_KEEP"):
            ALTWS_TIMEOUT = 60 * 60
            ALTKEEP_TIMEOUT = 5 * 60
            alt_mach, alt_ws_keep = self.env.get("ALTWS_KEEP", "").split(":")
            if alt_mach == self.mach:
                return

            # Allow 5 minutes for the alternate ws keep directory to appear.
            timeout = 0
            while not os.path.exists(alt_ws_keep) and \
                    timeout < ALTKEEP_TIMEOUT:
                time.sleep(10)
                timeout += 10
            if not os.path.exists(alt_ws_keep):
                self.logger.error(
                    "Unable to access alternate workspace %s" % alt_ws_keep)
                self.exit_status = 1
                return
            if timeout > 0:
                self.logger.log(
                    TINFO_LEVEL,
                    "Alternate workspace %s generation took %s seconds" %
                    (alt_ws_keep, timeout))

            # enter a loop to wait up to an hour for the tstamp task
            # in the remote workspace to contain the same hg information
            # as the data in this workspace.
            timeout = 0
            found = False
            ws_changesets = set([hg_info[key][0].split()[0]
                                 for key in hg_info])
            alt_hg_info = {}
            alt_tstamp = os.path.join(
                alt_ws_keep, ".build.{0}.tstamp".format(alt_mach))

            while timeout < ALTWS_TIMEOUT and not found:
                if timeout > 0:
                    self.logger.log(
                        TINFO_LEVEL, "Checking hg identification at %s" %
                        alt_tstamp)

                # every 30 seconds, check for new hg info
                time.sleep(30)
                timeout += 30

                if os.path.exists(alt_tstamp):
                    for line in open(alt_tstamp, "r").readlines():
                        toks = line.strip().split()
                        # assume 1 head for now.
                        alt_hg_info[toks[0]] = (" ".join(toks[1:]), 1)

                alt_changesets = set(
                    [alt_hg_info[key][0].split()[0] for key in alt_hg_info])
                if ws_changesets == alt_changesets:
                    found = True
                    self.logger.log(
                        TINFO_LEVEL,
                        "Workspace at %s in sync with revisions: %s" %
                        (alt_ws_keep, alt_changesets))
                else:
                    self.logger.log(
                        TINFO_LEVEL,
                        "Workspace at %s is still using older "
                        "revisions: %s" % (alt_ws_keep, alt_hg_info))

            if timeout >= ALTWS_TIMEOUT and not found:
                self.logger.error(
                    "Alternate workspace has not been updated to the "
                    "same mercurial revisions as this workspace.")
                self.logger.info("This workspace: %s" % hg_info)
                self.logger.info("Alt workspace: %s" % alt_hg_info)
                self.exit_status = 1
                return


class RTIReadyTask(Task):
    """Check that relevant build options were set in the bldrc file that are
    needed in a build before it should be submitted for an RTI.
    """
    task_name = "check.rti_ready"
    task_synopsis = "Check that this build config can be submitted for RTI"
    snapshot = False
    must_pass = False

    def _execute(self):
        rti_ready = True

        # a series of keys that should be set in the bldrc file
        required_keys = [
            ("check", "Build checks were not performed"),
            ("debug", "A debug build was not performed"),
            ("nondebug", "A non-debug build was not performed"),
        ]

        known_checks = get_known_checks()

        # warn if the user has changed any check.* defaults
        for key, val in list(self.bldenv.overrides.items()):
            if key.endswith(".must_pass"):
                check = key.replace(".must_pass", "")
                if check not in known_checks:
                    continue
                if known_checks[check].must_pass and not val:
                    required_keys.append(
                        (key, "%s default was changed to false" % key))
            if key in known_checks and known_checks[key].enabled_by_default:
                if not val:
                    required_keys.append((key, "%s was disabled" % key))

        def key_cmp(item):
            """A 'comparator' function, actually a key mapping function."""
            if "must_pass" in item[0]:
                return (1, item[0])
            return (0, item[0])

        for var, message in sorted(required_keys, key=key_cmp):
            value = getattr(
                self.bldenv, var, self.bldenv.overrides.get(var, True))
            if not value:
                self.logger.log(TWARNING_LEVEL, message)
                rti_ready = False

        if self.bldenv.user_env_overrides:
            self.logger.log(
                TWARNING_LEVEL, "'userenv' or 'userunsetenv' were specified "
                "in the bldrc file: %s" % self.bldenv.user_env_overrides)
            rti_ready = False

        # A series of environment variables we're interested in. The tuples are
        # of the form: (name, error message, is this is allowed in the env?)
        check_env = [
            ("BIP", "'bip' was specified in the bldrc file", False),
            ("CW_NO_SHADOW", "'cw_no_shadow' was specified in the bldrc "
                "file", False),
            ("REF_PROTO_LIST", "protocmp check wasn't run against parent "
                "proto area", True),
            ("ELF_DATA_BASELINE_DIR", "ELF interface checks were not "
                "compared to a baseline build", True),
            ("PKGARCHIVE",
                "'pkgarchive' was set to '<None>' in the bldrc file", True)
        ]

        for var, message, allowed in check_env:
            if var in self.env and not allowed:
                self.logger.log(TWARNING_LEVEL, message)
                rti_ready = False
            elif var not in self.env and allowed:
                self.logger.log(TWARNING_LEVEL, message)
                rti_ready = False

        # Check hg heads against the branch tips of the workspace and warn if
        # there are discrepancies. This will not fail the task, but can be
        # useful to see in the logs.
        hg_info = get_hg_info(self.ws, self.env, hg_identify=False)
        hg_info_id = get_hg_info(self.ws, self.env, hg_identify=True)
        different_changesets = False
        for repo in hg_info:
            cs = hg_info[repo][0].split(" ")[0]
            cs_id = hg_info_id[repo][0].split(" ")[0]
            if cs != cs_id:
                different_changesets = True
                self.logger.log(
                    TWARNING_LEVEL, "%s shows different changesets "
                    "for branch tip vs. identify: %s %s" % (repo, cs, cs_id))

        if different_changesets:
            self.logger.log(
                TWARNING_LEVEL, "This workspace may not be at branch-tip "
                "changesets, or may contain modified sources.")

        # Gather the most recent biweekly st_xxx(.yyy) tags for all
        # repositories. Each repository should contain the same build tag,
        # we ignore the respins and branch indicators.
        known_builds = set()
        repo_tags = {}
        for repo in hg_info:
            found_tag = False
            tag_cmd = "hg -R %s tags" % repo
            ret, out, err = util.cmdline_run(tag_cmd, env_arg=self.env)
            for line in out.split("\n"):
                if found_tag:
                    break
                line = line.strip()
                if not line:
                    continue
                if re.search("^st_[0-9]+|^on12_[0-9]+", line):
                    found_tag = True
                    tag = line.split(" ")[0]
                    m = re.match("^[a-z0-9]+_([0-9]+)", tag)
                    if m and len(m.groups()) == 1:
                        known_builds.add(m.group(1))
                        repo_tags[repo] = tag
                    else:
                        found_tag = False

            if not found_tag:
                rti_ready = False
                self.logger.log(
                    TWARNING_LEVEL, "No st_xxx or on12_xxx tags were "
                    "found in %s!" % repo)

        if len(known_builds) > 1:
            rti_ready = False
            self.logger.log(
                TWARNING_LEVEL, "This workspace contains repositories that "
                "have been synced to different OS/Net builds.")
            self.logger.log(
                TWARNING_LEVEL, "Check that the $SRC, usr/closed, usr/fish "
                "and usr/man repositories are in sync:")
            for key in sorted(repo_tags):
                repo = key.replace(self.ws, "").lstrip("/")
                if not repo:
                    repo = "usr/src"
                self.logger.log(
                    TWARNING_LEVEL, "{0:<10} {1}".format(
                        repo, repo_tags[key]))

        # Perform basic checks to see if this build machine is capable of
        # building the bits in this workspace.
        vers_cmd = (
            "pkg contents -H -a name=pkg.fmri -t set -o value "
            "pkg:/system/core-os")

        ws_vers = get_pkgvers(self.env)
        if not ws_vers:
            rti_ready = False
            self.logger.log(
                TWARNING_LEVEL, "Unable to determine the package version "
                "of the packages that would be published by this build.")
        ret, out, err = util.cmdline_run(vers_cmd, env_arg=self.env)
        if ret != 0:
            rti_ready = False
            self.logger.log(
                TWARNING_LEVEL, "Unable to determine the package version "
                "of pkg:/system/core-os present on this build machine.")
        else:
            # Determine the package version from the output of vers_cmd,
            # which looks similar to this:
            # pkg:/system/core-os@5.12,5.11-5.12.0.0.0.26.1:20130715T143314
            coreos_pkg = out
            vers = coreos_pkg.rstrip().split("@")[1]
            vers_index = vers.index("-") + 1
            build_vers = vers[vers_index:].split(":")[0]

            # Arrays containing the version components for the build
            # machine and workspace, respectively.
            bvers = build_vers.split(".")
            wsvers = ws_vers.split(".")

            # Check the build release, include a temporary allowance for 5.12
            # build systems.
            if bvers[0:2] == ["5", "12"]:
                pass
            elif bvers[0:2] != wsvers[0:2]:
                rti_ready = False
                self.logger.log(
                    TWARNING_LEVEL, "This build machine is running "
                    "%s and should not be used to build a workspace for "
                    "version %s." % (coreos_pkg, ws_vers))

            # Verify the build machine version. These are guidelines rather
            # than requirements: in some cases it's possible to build and test
            # packages produced when these checks fail, but more care is
            # needed while installing or testing the bits, so these checks
            # are designed to give a heads-up to RTI advocates so that they
            # ask the right questions.

            # Treat s12_128, the last s12 build, as if it was st_000,
            # negative build numbers here are fine and will result in
            # warnings.
            if bvers[0:2] == ["5", "12"]:
                bvers[5] = str(int(bvers[5]) - 128)

            # The build machine should be no more than 2 builds behind the
            # workspace we're building.
            if int(bvers[5]) < (int(wsvers[5]) - 2):
                rti_ready = False
                self.logger.log(
                    TWARNING_LEVEL, "This build machine is running %s "
                    "which is more than two builds older than the "
                    "packages that would be produced (%s) and should not "
                    "be used to build this workspace." %
                    (coreos_pkg, ws_vers))

        # If the build machine version is higher than the workspace
        # we're building, this will result in the build depending on
        # non-ON WOS packages that would not be present on a test
        # machine running those lower-versioned WOS bits.
        pkgroot = self.env.get("PKGRESOLVEIMAGE")
        if pkgroot != "/":
            self.logger.log(
                TINFO_LEVEL, "The alternate image %s will be used to "
                "resolve package dependencies." % pkgroot)
        elif int(wsvers[5]) < int(bvers[5]):
            rti_ready = False
            self.logger.log(
                TWARNING_LEVEL, "This build machine is running %s "
                "which is newer than the packages that would be "
                "produced (%s.) This can result in complications "
                "during testing as the build will depend on non-ON "
                "WOS packages that are unavailable on test machines "
                "installed with that that lower WOS release." %
                (coreos_pkg, ws_vers))

        # tuples of: (packages we think should be on build machines, whether
        # those packages are strictly required, whether we should check
        # the latest version is installed)
        build_pkgs = [
            ("developer/opensolaris/osnet", False, True),
            ("developer/build/onbld-internal",
                self.env.get("FISH_IS_PRESENT") == "yes", False)]

        pkg_vers_cmd = "pkg contents -H -o value -a name=pkg.fmri"
        pkg_latest_cmd = "pkg contents -rH -o value -a name=pkg.fmri"

        for fmri, required, check_latest in build_pkgs:
            pkg_vers_cmd += " %s " % fmri
            pkg_latest_cmd += " %s@latest " % fmri

            ret_vers, out_vers, err_vers = \
                util.cmdline_run(pkg_vers_cmd, env_arg=self.env)
            ret_latest, out_latest, err_latest = \
                util.cmdline_run(pkg_latest_cmd, env_arg=self.env)

            if ret_vers != 0:
                if fmri not in out_vers:
                    if required:
                        self.logger.log(
                            TWARNING_LEVEL, "This build machine MUST have %s "
                            "installed" % fmri)
                        self.exit_status = 1
                    else:
                        self.logger.log(
                            TWARNING_LEVEL, "This build machine does not "
                            "have %s installed" % fmri)

                if check_latest:
                    if ret_latest != 0:
                        self.logger.log(
                            TWARNING_LEVEL,
                            "unable to obtain @latest pkg versions of %s" %
                            fmri)
                        self.logger.log(TWARNING_LEVEL, err_latest)
                        continue
                    elif sorted(
                            out_vers.strip()) != \
                            sorted(out_latest.strip()):
                        # this crude check is not enough to fail rti_ready,
                        # but can be useful to have in the logs.
                        self.logger.log(
                            TWARNING_LEVEL, "This build machine has older "
                            "versions of a build package than the latest "
                            "available version. Your CBE may be out of date.\n"
                            "Installed:\n%s\n\n"
                            "Available:\n%s\n" %
                            (out_vers, out_latest))

        # Check how much swap is configured if we're not running unit-tests.
        ret, out, err = util.cmdline_run(
            "/usr/bin/pargs %s" % os.getppid())
        if "buildtest.py" in out:
            self.logger.log(
                TWARNING_LEVEL, "Skipping swap size rti_check since build(1) "
                "unit tests are running.")
        else:
            swap_cmd = "/usr/sbin/swap -l"
            swap_size = 0
            ret, out, err = util.cmdline_run(swap_cmd)
            if ret != 0:
                # This should never happen.
                self.rti_ready = False
                self.logger.log(
                    TWARNING_LEVEL, "Unable to determine how much swap is "
                    "configured on this build machine. Proceed with caution.")
            else:
                # The output of swap is not parseable, but recently changed to
                # grow a new column ("encrypted"). We (temporarily) need to
                # cope with both output formats. This change can be removed
                # once it becomes the minimum CBE.
                swapre = re.compile(r'[a-z]*')
                for line in out.split("\n"):
                    # swap -l prints a column-header that starts with
                    # 'swapfile' which we can ignore.
                    if line.startswith("swapfile"):
                        continue
                    if not line:
                        continue
                    swapval = line.strip().split()[-1]
                    if swapre.match(swapval):
                        swapval = line.strip().split()[-2]
                    swap_size += int(swapval)

                # convert that number of 512-byte blocks to gigabytes
                swap_gb = int(round(float(swap_size) *
                              512 / 1024 / 1024 / 1024))

                if swap_gb < MINIMUM_RECOMMENDED_SWAP:
                    rti_ready = False
                    self.logger.log(
                        TWARNING_LEVEL, "This build machine appears to "
                        "have only %d GB of swap configured. The minimum "
                        "recommended is %d GB. Your builds may fail, "
                        "depending on the amount of physical memory "
                        "available, 'dmake_max_jobs' settings, or number of "
                        "concurrent builds running." %
                        (swap_gb, MINIMUM_RECOMMENDED_SWAP))

        if self.bldenv.pkgmerge and self.env.get("ALTWS_KEEP"):
            # verify that the alternate workspace has the same identity as
            # this one. We use the workspace .build.$MACH.tstamp file to
            # determine the changesets in the alternate workspace, which
            # is not precise, but we don't want to have to depend on NFS
            # access to $SRC.
            alt_mach, alt_ws_keep = self.env.get("ALTWS_KEEP", "").split(":")
            alt_hg_info = {}
            alt_tstamp = os.path.join(
                alt_ws_keep, ".build.{0}.tstamp".format(alt_mach))
            if os.path.exists(alt_tstamp):
                for line in open(alt_tstamp, "r").readlines():
                    toks = line.strip().split()
                    # assume 1 head for now.
                    alt_hg_info[toks[0]] = (" ".join(toks[1:]), 1)

            ws_changesets = set([hg_info_id[key][0] for key in hg_info])
            alt_changesets = set([alt_hg_info[key][0] for key in alt_hg_info])

            if ws_changesets != alt_changesets:
                def log_hginfo(info):
                    for key in sorted(info):
                        self.logger.log(
                            TWARNING_LEVEL, "{0} = {1}".format(
                                key, info[key][0]))

                self.logger.log(
                    TWARNING_LEVEL, "A cross-architecture pkgmerge is "
                    "configured, but this workspace, and the alternate "
                    "workspace at {0} may not be in sync and do not contain "
                    "the same hg(1) changesets at 'tip':".format(alt_ws_keep))
                self.logger.log(TWARNING_LEVEL, "This workspace:")
                log_hginfo(hg_info_id)
                self.logger.log(
                    TWARNING_LEVEL, "Workspace at {0} :".format(alt_ws_keep))
                log_hginfo(alt_hg_info)
                rti_ready = False

        # For RTI submission, a clobber build should be used so that everything
        # is built and checked consistently.
        if self.bldenv.incremental is True:
            self.logger.log(TWARNING_LEVEL, "Incremental builds must not be "
                            "used for RTI submission")
            rti_ready = False

        if not rti_ready:
            self.exit_status = 1
            self.logger.log(
                TWARNING_LEVEL, "One or more bldrc file settings were found "
                "that suggest this build is not ready for RTI")
        else:
            self.logger.log(
                TINFO_LEVEL, "All bldrc settings required for RTI submission "
                "were enabled.")


class NightlyTask(CompositeTask):
    """Do a nightly build. Nightly is composed of a series of other tasks.
    We complete an entire debug build before starting a non-debug build, which
    means that all tasks started here will start and finish within the same
    log files, rather than needing to use onbld.log.switch_logs()."""

    task_name = "nightly"
    task_synopsis = "Do a build, running several other subcommands"

    def __init__(self, bldenv, args=None, fail_fast=True, level=TASK_LEVEL,
                 propagate_logs=True, no_logfiles=False,
                 avoid_task_creation=False):
        """Add Tasks that make up a nightly build. We have
        'avoid_task_creation' to allow easier subclassing for CompositeTasks
        that wish to use 'nightly'-like mail_msg creation."""
        self.core_os_ver = None
        self.entire_ver = None
        self.osnet_ver = None
        self.build_machine = socket.gethostname()
        self.warn_on_snapshot_errors = True

        super(NightlyTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            no_logfiles=no_logfiles, fail_fast=fail_fast)

        if "FISH_STANDALONE" in self.env and self.task_name == "nightly":
            raise TaskException(
                "This workspace appears to be a standalone Fish repository. "
                "Use 'build fish' rather than 'build nightly' to build it.")

        def add_common_build_tasks(
                runtime_bldenv, label, skip_pkgsurf=True,
                skip_man=False):
            """Add the tasks that need to run individually for both debug
            and non-debug builds, but which can't share the same environment.
            (eg. $ROOT is different for debug and non-debug builds)"""

            runtime_bldenv.common_task = True
            self.add_task(SetupTask(runtime_bldenv))

            if self.env["MAN_IS_PRESENT"] == "yes" and not skip_man:
                self.add_task(ManInstallTask(runtime_bldenv))

            self.add_task(InstallTask(runtime_bldenv))
            if self.env["FISH_IS_PRESENT"] == "yes":
                self.add_task(FishInstallTask(runtime_bldenv))

            self.add_task(PackagesTask(runtime_bldenv))
            if self.env["FISH_IS_PRESENT"] == "yes":
                self.add_task(FishPackagesTask(runtime_bldenv))

            if not skip_pkgsurf:
                self.add_task(PkgsurfTask(runtime_bldenv))
                self.add_task(PkgdiffTask(runtime_bldenv))

            elif runtime_bldenv.pkgsurf:
                self.add_task("Skipping pkgsurf until after pkgmerge runs")

            if self.env["FISH_IS_PRESENT"] == "yes":
                if self.env["FISH_GK_IMAGES"] == "true":
                    self.add_task(FishMkakGkImagesTask(runtime_bldenv))
                elif self.env["FISH_GK_NIGHTLY_IMAGES"] == "true":
                    self.add_task(FishMkakGkNightlyImagesTask(runtime_bldenv))
                else:
                    self.add_task(FishMkakTask(runtime_bldenv))
                self.add_task(FishSaveArtifactsTask(runtime_bldenv))
                if self.env["FISH_LICENSES"] == "true":
                    self.add_task(FishLicensesTask(runtime_bldenv))
                    # This is needed since it gets skipped by the
                    # FishMkakTask when "FISH_LICENSES" is true.
                    self.add_task(FishDestroyDcDataset(runtime_bldenv))

            if self.bldenv.overrides.get("check.parfait", True) and \
                    self.bldenv.check:
                # if Parfait analysis fails, we still want to continue with the
                # rest of the build
                self.add_task(
                    ParfaitAnalyseTask(runtime_bldenv))
            else:
                self.add_task(
                    "Skipping check.parfait due to checkoverrides setting")

        def get_pkg_fmri(fmri):
            ret, pkg_ver, err = util.cmdline_run(
                "pkg contents -H -a name=pkg.fmri -t set -o value %s" % fmri,
                env_arg=self.env)
            return pkg_ver.rstrip()

        self.core_os_ver = get_pkg_fmri("pkg:/system/core-os")
        self.entire_ver = get_pkg_fmri("pkg:/entire")
        self.osnet_ver = get_pkg_fmri("pkg:/developer/opensolaris/osnet")
        self.onbld_ver = get_pkg_fmri("pkg:/developer/build/onbld")
        self.onbld_int_ver = get_pkg_fmri(
            "pkg:/developer/build/onbld-internal") or "(not installed)"
        self.pkgvers = get_pkgvers(self.env)

        ret, self.linkers_ver, err = util.cmdline_run("/usr/bin/ld -V")
        self.linkers_ver = self.linkers_ver.rstrip()

        ret, task_id, err = util.cmdline_run(
            "/usr/bin/ps -o taskid= -p %s" % os.getpid())

        self.add_task("Build workspace  %s" % self.ws)
        self.add_task("Build conf file  %s" % self.bldenv.bldrc)
        self.add_task("Build project    %s" % self.env["BUILD_PROJECT"])
        self.add_task("Build taskid     %s" % task_id.strip())
        self.add_task("Build pkgvers    %s" % self.pkgvers)
        self.add_task("")

        hg_info = get_hg_info(self.ws, self.env)
        self.add_task("Workspace hg(1) changesets:")
        self.add_task(".           node=%s" % hg_info[self.ws][0])

        # We can only emit hg information for subtrees if they exist.
        if "%s/usr/closed" % self.ws in hg_info:
            self.add_task(
                "usr/closed  node=%s" % hg_info["%s/usr/closed" % self.ws][0])
        else:
            self.add_task(
                "usr/closed  node=[repository not present]")

        if "%s/usr/man" % self.ws in hg_info:
            self.add_task(
                "usr/man     node=%s" % hg_info["%s/usr/man" % self.ws][0])
        else:
            self.add_task("usr/man     node=[repository not present]")

        if "%s/usr/fish" % self.ws in hg_info:
            self.add_task(
                "usr/fish    node=%s" % hg_info["%s/usr/fish" % self.ws][0])
        else:
            self.add_task("usr/fish    node=[repository not present]")

        self.add_task("")
        self.add_task("Build machine details:")
        self.add_task(
            "Identity             %s (%s)\n" % (self.build_machine, self.mach))
        self.add_task(
            "WOS pkg              %s\n" % self.entire_ver)
        self.add_task(
            "core-os pkg          %s\n" % self.core_os_ver)
        self.add_task(
            "osnet pkg            %s\n" % self.osnet_ver)
        self.add_task(
            "onbld pkg            %s\n" % self.onbld_ver)
        self.add_task(
            "onbld-internal pkg   %s\n" % self.onbld_int_ver)
        self.add_task(
            "linkers              %s\n" % self.linkers_ver)

        if avoid_task_creation:
            return

        log_tstamp = self.bldenv.pkgmerge and self.env.get("ALTWS_KEEP")
        self.add_task(TstampTask(self.bldenv, no_logfiles=not log_tstamp))

        if self.bldenv.check and self.bldenv.overrides.get(
                "check.rti_ready", RTIReadyTask.enabled_by_default):
            self.add_task(RTIReadyTask(self.bldenv))

        self.add_task(RecordCompositeTask(self.bldenv))

        if self.bldenv.pkgsurf and not self.bldenv.generate_splice:
            self.add_task(SavePackagesTask(self.bldenv))
        elif not self.bldenv.pkgsurf:
            self.add_task(
                "Skipping save_packages since we are not pkgsurfing")
        else:
            self.add_task(
                "Skipping save_packages, using baseline as reference")

        if self.bldenv.incremental:
            self.add_task("Skipping clobber for incremental build.")
        else:
            self.add_task(ClobberTask(self.bldenv))

        # We don't do per-build-flavor pkgsurfing if we will be
        # pkgmerging later, because then we risk potentially trying
        # to merge debug and nondebug packages that have different
        # versions, which would fail. Instead, we pkgsurf after
        # pkgmerge has completed.
        skip_pkgsurf = (not self.bldenv.pkgsurf) or self.bldenv.pkgmerge

        # By default, perform the man page build if MAN_IS_PRESENT=yes
        skip_man = False

        self.add_task(ToolsTask(self.bldenv))

        if self.bldenv.debug:
            runtime_bldenv = copy.deepcopy(self.bldenv)
            runtime_bldenv.debug = True
            runtime_bldenv.nondebug = False
            add_common_build_tasks(runtime_bldenv, "DEBUG", skip_pkgsurf)

        if self.bldenv.nondebug and self.bldenv.debug:
            self.add_task("Debug build completed, starting non-debug build.")
            # we don't need to switch logs for nightly - the log file is short
            # enough that splitting it into separate files is unnecessary.

            # We've already done a man page build, so no need to build it again
            if self.env["MAN_IS_PRESENT"] == "yes":
                skip_man = True

        if self.bldenv.nondebug:
            runtime_bldenv = copy.deepcopy(bldenv)
            runtime_bldenv.debug = False
            runtime_bldenv.nondebug = True
            add_common_build_tasks(
                runtime_bldenv, "non-DEBUG", skip_pkgsurf, skip_man)

        if self.env["FISH_IS_PRESENT"] == "yes":
            # We've already saved the content we care about - clean up.
            # Developers who are debugging image construction should invoke
            # the raw fish.image|*_iso tasks and clean up manually.
            self.add_task(FishDestroyDcDataset(bldenv))

        if "TONICBUILD" in self.env:
            self.add_task(GenerateTPLTask(self.bldenv))
            self.add_task(ClosedBinsTask(bldenv))

        if self.bldenv.pkgmerge:
            self.add_task(PkgmergeTask(bldenv))

        # post-merge pkgsurfing if necessary
        if self.bldenv.pkgsurf and self.bldenv.pkgmerge:
            self.add_task(PkgsurfTask(bldenv))
            self.add_task(PkgdiffTask(bldenv))

        # See if zip file creation is requested, but only if pkgmerge is
        # also set:
        if self.bldenv.zipfiles:
            if self.bldenv.pkgmerge:
                self.add_task(ZipTask(bldenv))
            else:
                self.add_task(
                    "Skipping zipfile creation since we are not pkgmerging")

        self.add_task(CheckCompositeTask(bldenv))

        if self.bldenv.archive_build:
            self.add_task(ArchiveTask(bldenv))

        if self.bldenv.update_parent_ws:
            self.add_task(UpdateParentTask(bldenv))

    def get_mail_subject(self):
        return "%(status)s: %(vers)s %(mach)s %(pkgvers)s (%(host)s)" % {
            "status": self.status_desc,
            "vers": self.env["VERSION"],
            "host": self.build_machine,
            "mach": self.mach,
            "pkgvers": self.pkgvers}

    def generate_mail_msg(self):
        """Produce the mail_msg file for this build, and send it if mailto was
        set."""

        mail_msg_path = os.path.join(self.log_dir, "mail_msg")
        with open(mail_msg_path, "w") as mail_msg:
            # milliseconds are overkill.
            start_time = str(self.start_time).split(".")[0]
            finish_time = str(self.finish_time).split(".")[0]
            build_length = str(self.finish_time - self.start_time).split(
                ".")[0]

            if self.interrupted:
                self.status_desc = "fail (interrupted)"
                self.exit_status = 1
            elif self.exit_status == 0:
                self.status_desc = "pass"
                # Walk the task list to see whether we've ignored any
                # failures, in which case we reclassify it as a pass, but
                # with issues that ought to prevent an RTI from being granted
                for task in self.tasks:
                    if isinstance(task, str):
                        continue
                    elif task.exit_status != 0 and task.must_pass:
                        self.status_desc = "fail"
                    elif task.exit_status != 0 and not task.must_pass:
                        self.status_desc = "pass (RTI issues)"
            else:
                self.status_desc = "fail"

            if "FISH_STANDALONE" in self.env:
                flavor = "fish standalone"
            elif self.env["FISH_IS_PRESENT"] == "yes":
                flavor = "default"
            else:
                flavor = "server"

            mail_msg.write("Build summary for %s %s %s build of %s\n\n" % (
                flavor, self.mach, self.task_name, self.ws))

            mm_fmt = "{0:<16}: {1}\n"

            mail_msg.write(mm_fmt.format("Build status", self.status_desc))
            mail_msg.write(mm_fmt.format("Started", start_time))
            mail_msg.write(mm_fmt.format("Completed", finish_time))
            mail_msg.write(mm_fmt.format("Total time", build_length))
            mail_msg.write(mm_fmt.format("Build config", self.bldenv.bldrc))
            mail_msg.write(mm_fmt.format("Build version", self.env["VERSION"]))
            mail_msg.write(mm_fmt.format("Build pkgvers", self.pkgvers))

            mail_msg.write("\n")
            no_repo = ["[repository not present]"]
            hg_info = get_hg_info(self.ws, self.env)

            subrepos = ["usr/closed", "usr/man"]
            if self.env["FISH_IS_PRESENT"] == "yes":
                subrepos.append("usr/fish")

            mail_msg.write(
                mm_fmt.format(". changeset", hg_info[self.ws][0]))
            for subrepo in subrepos:
                repo = os.path.join(self.ws, subrepo)
                mail_msg.write(
                    mm_fmt.format(
                        subrepo + " cs", hg_info.get(repo, no_repo)[0]))

            mail_msg.write("\n")
            mail_msg.write(mm_fmt.format(
                "Start timestamp", self.start_sinceepoch))

            mail_msg.write("\n")
            mail_msg.write("Build machine details:\n")
            mail_msg.write(
                mm_fmt.format("Identity", "{0} ({1})".format(
                    self.build_machine, self.mach)))
            mail_msg.write(mm_fmt.format("WOS pkg", self.entire_ver))
            mail_msg.write(mm_fmt.format("core-os pkg", self.core_os_ver))
            mail_msg.write(mm_fmt.format("osnet pkg", self.osnet_ver))
            mail_msg.write(mm_fmt.format("onbld pkg", self.onbld_ver))
            mail_msg.write(mm_fmt.format("onbld-internal pkg",
                           self.onbld_int_ver))
            # Only add the linkers version to the mail if they are non-stock,
            # as the version of stock linkers is implicit in the version of
            # the core-os, and does not normally add useful information.
            # We can easily detect test linkers because the version string
            # has a third colon delimiter followed by extra information.
            if self.linkers_ver.count(":") > 2:
                mail_msg.write(mm_fmt.format("linkers", self.linkers_ver))

            mail_msg.write("\n\n")

            # use a wider column width from here on
            mm_fmt = "{0:<22}: {1}\n"

            def emit_task_summary(tasks):
                """Recursively print a high-level summary for these tasks."""
                for task in tasks:
                    if not isinstance(task, Task):
                        continue
                    # if this is a disabled task, we'll report it later
                    if not self.bldenv.overrides.get(
                            task.task_name, True):
                        continue
                    if not task.start_time:
                        status = "fail (never started)"
                    elif task.exit_status == 0:
                        status = "pass"
                    elif not task.must_pass:
                        status = "fail (non-fatal)"
                    else:
                        status = "*fail*"
                    mail_msg.write(mm_fmt.format(task.name, status))
                    if isinstance(task, CompositeTask):
                        emit_task_summary(task.tasks)

            def emit_task_details(tasks):
                """Recursively print details for these tasks."""
                for task in tasks:
                    if not isinstance(task, Task):
                        continue

                    summary = task.summary()
                    if summary:
                        mail_msg.write("--- %s ---\n" % task.name)
                        mail_msg.write("(%s)\n" % task.task_synopsis)
                        mail_msg.write("\n%s\n" % summary)
                    if isinstance(task, CompositeTask):
                        emit_task_details(task.tasks)

            def emit_disabled_tasks():
                """Show all disabled tasks if checks were enabled."""
                if not self.bldenv.check:
                    return
                written_newline = False
                known_checks = get_known_checks()
                for check in known_checks:
                    if not self.bldenv.overrides.get(check, True):
                        if not written_newline:
                            mail_msg.write("\n")
                            written_newline = True
                        # only mention checks that are normally enabled
                        # by default
                        if known_checks[check].enabled_by_default:
                            mail_msg.write(mm_fmt.format(check, "disabled"))

            mail_msg.write("=== Task summary ===\n")
            emit_task_summary(self.tasks)
            emit_disabled_tasks()
            mail_msg.write("\n\n")

            mail_msg.write("=== Task details ===\n")
            emit_task_details(self.tasks)

        if not self.env.get("MAILTO"):
            return
        self.logger.log(
            TINFO_LEVEL, "Sending mail_msg to %s" % self.env["MAILTO"])
        with open(mail_msg_path, "r") as mail_msg:
            msg = email.mime.text.MIMEText(mail_msg.read())
            recipients = self.env["MAILTO"].split(",")
            # Assume that the person sending the mail is the user who
            # started the build, and is the first recipent.
            sender = recipients[0]
            subj = self.get_mail_subject()
            msg["From"] = sender
            msg["To"] = self.env["MAILTO"]
            msg["Subject"] = subj
            msg["X-ON-Build-Notification"] = "true"
            try:
                s = smtplib.SMTP("localhost")
                s.sendmail(sender, recipients, msg.as_string())
                s.quit()
            except Exception as smtp_error:
                # fallback to using mailhost.us.oracle.com
                try:
                    self.logger.log(
                        TINFO_LEVEL,
                        "Unable to send mail using localhost: %s" % smtp_error)
                    self.logger.log(
                        TINFO_LEVEL,
                        "Sending mail using mailhost.us.oracle.com.")
                    s = smtplib.SMTP("mailhost.us.oracle.com")
                    s.sendmail(sender, recipients, msg.as_string())
                    s.quit()
                except Exception as smtp_error1:
                    self.logger.log(
                        TINFO_LEVEL,
                        "Unable to send mail using mailhost.us.oracle.com: "
                        "%s" % smtp_error1)
                    # Accept that we can't send mail. Log a message at a high
                    # level to defeat "build -q" in case we're running with
                    # cron, and it might have better luck sending mail.
                    self.logger.error(
                        "Unable to send complete \"%s\" build mail "
                        "notification to %s: %s" %
                        (self.status_desc, self.env["MAILTO"], smtp_error1))
                    self.logger.error(
                        "Was sendmail enabled on the build machine?")

    def _execute_start(self):
        super(NightlyTask, self)._execute_start()
        self.logger.log(
            TINFO_LEVEL, "Start timestamp  %s" % self.start_sinceepoch)

    def _execute_finish(self):
        """Copy our log files to their own directory. If the build
        was successful, we also rename the build timestamp file."""

        super(NightlyTask, self)._execute_finish()
        try:
            self.generate_mail_msg()
            tstamp_path = "%s/.build.%s.tstamp" % (self.keep, self.mach)
            if self.exit_status == 0 and os.path.exists(tstamp_path):
                shutil.copy(tstamp_path, "%s.prev" % tstamp_path)
        except KeyboardInterrupt:
            # Overzealous Ctrl-Cs shouldn't result in message asking users to
            # mail gatekeepers.
            pass
        except Exception as e:
            self.logger.log(
                TWARNING_LEVEL, "Unable to generate mail_msg: %s" % e,
                exc_info=True)
            # This is severe enough to warrant a message suggesting the user
            # get in touch with the gatekeepers, so raise this and let
            # _execute_finish deal with it.
            raise

        try:
            log_basename = "log.{0}".format(
                self.start_time.strftime(log.datefmt))
            log_subdir = os.path.join(self.log_dir, log_basename)
            os.makedirs(log_subdir)
            for entry in os.listdir(self.log_dir):
                if entry.endswith(".log") or entry.endswith(".log.html") or \
                        entry.startswith("parfait_analysis_") or \
                        entry.startswith("wsdiff.results") or \
                        entry == "mail_msg" or \
                        entry == "proto_list_%s" % self.mach:
                    basename = os.path.basename(entry)
                    os.rename(
                        "%s/%s" % (self.log_dir, entry), "%s/%s" %
                        (log_subdir, basename))

            # create a handy 'last' symlink
            last_link = os.path.join(self.log_dir, "last")
            if os.path.lexists(last_link):
                os.remove(last_link)
            os.symlink(log_basename, last_link)

            if self.bldenv.archive_build:
                archive_log_dir = os.path.join(
                    self.env["ARCHIVE_TO_DIR"],
                    os.path.basename(self.log_dir))
                if (os.path.exists(archive_log_dir)):
                    archive_log_subdir = os.path.join(archive_log_dir,
                                                      log_basename)
                    archive_last_link = os.path.join(archive_log_dir, "last")
                    shutil.rmtree(archive_log_subdir, ignore_errors=True)
                    shutil.copytree(log_subdir, archive_log_subdir)
                    if os.path.lexists(archive_last_link):
                        os.remove(archive_last_link)
                    os.symlink(log_basename, archive_last_link)

        except Exception as e:
            # we've renamed our log files, so they may not capture
            # this error - we'll log at a high level so that the
            # problem propagates up to the root logger.
            self.logger.error(
                "Exception finishing nightly task: %s" % e, exc_info=True)
            raise


class QnightlyTask(NightlyTask):
    """Just like Nightly, except that this does a mercurial pull first,
    comparing repository tips and new heads from before and after the pull
    task before deciding to proceed."""

    task_name = "qnightly"
    task_synopsis = \
        "Do a nightly build only if new hg changesets are available"

    def get_mail_subject(self):
        for task in self.tasks:
            if isinstance(task, str):
                continue
            if task.name in ["pull", "pull-nd"] and not task.satisfied:
                return "no build: %(vers)s %(mach)s %(pkgvers)s (%(host)s)" % {
                    "vers": self.env["VERSION"],
                    "host": self.build_machine,
                    "mach": self.mach,
                    "pkgvers": self.pkgvers}

        return super(QnightlyTask, self).get_mail_subject()

    def __init__(self, bldenv, args=None, fail_fast=True, level=TASK_LEVEL,
                 propagate_logs=True):

        super(QnightlyTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)
        predicate_pull = PredicateHgPullTask(bldenv)
        self.add_task(predicate_pull, index=0)


class GKParfaitTask(NightlyTask):
    """Perform a Parfait build for the gatekeepers. This limits the
    build to setup and install."""

    task_name = "gk.parfait"
    task_synopsis = "Run the minimum necessary tasks to generate parfait DB"
    produce_log_extract = True

    def __init__(self, bldenv, args=None, fail_fast=True, level=TASK_LEVEL,
                 propagate_logs=True):
        # specify avoid_task_creation so we don't inherit all of the Tasks
        # normally added by 'nightly'
        super(GKParfaitTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast, avoid_task_creation=True)

        # Sanity checks
        if not self.bldenv.overrides.get("check.parfait", True):
            raise Exception("Parfait must be enabled for the {0} task".format(
                self.task_name))
        if self.env["MAN_IS_PRESENT"] == "yes":
            raise Exception("usr/man must not be present for the {0} "
                            "task".format(self.task_name))
        if "CW_NO_SHADOW" not in self.env:
            self.logger.log(
                TWARNING_LEVEL, "cw_no_shadow should be set for "
                "the {0} task".format(self.task_name))

        def add_common_build_tasks(runtime_bldenv, label):
            """Add the tasks needed for debug and non-debug Parfait runs."""

            runtime_bldenv.common_task = True
            self.add_task(SetupTask(runtime_bldenv))
            self.add_task(InstallTask(runtime_bldenv))
            if self.env["FISH_IS_PRESENT"] == "yes":
                self.add_task(FishInstallTask(runtime_bldenv))
            self.add_task(ParfaitAnalyseTask(runtime_bldenv))

        self.add_task(TstampTask(self.bldenv, no_logfiles=True))
        self.add_task(ClobberTask(self.bldenv))
        self.add_task(ToolsTask(self.bldenv))

        if self.bldenv.debug:
            runtime_bldenv = copy.deepcopy(self.bldenv)
            runtime_bldenv.debug = True
            runtime_bldenv.nondebug = False
            add_common_build_tasks(runtime_bldenv, "DEBUG")

        if self.bldenv.nondebug and self.bldenv.debug:
            self.add_task(
                "Debug Parfait build completed, starting non-debug build.")

        if self.bldenv.nondebug:
            runtime_bldenv = copy.deepcopy(bldenv)
            runtime_bldenv.debug = False
            runtime_bldenv.nondebug = True
            add_common_build_tasks(runtime_bldenv, "non-DEBUG")


class FishCompositeTask(NightlyTask):
    """Perform only a Fish build for this workspace and generate and send a
    mail_msg on completion. For a full ON nightly build, we expect that
    NightlyTask will instead invoke the required Fish Tasks."""

    task_name = "fish"
    task_synopsis = "Build only the fish subrepo"
    produce_log_extract = False

    def __init__(self, bldenv, args=None, fail_fast=True, level=TASK_LEVEL,
                 propagate_logs=True):
        # specify avoid_task_creation so we don't inherit all of the Tasks
        # normally added by 'nightly'
        super(FishCompositeTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast, avoid_task_creation=True)

        self.add_task(TstampTask(self.bldenv, no_logfiles=True))

        def add_common_build_tasks(runtime_bldenv, label):
            """Add the tasks that need to run individually for both debug
            and non-debug builds, but which can't share the same environment.
            (eg. $ROOT is different for debug and non-debug builds)"""

            runtime_bldenv.common_task = True
            self.add_task(FishInstallTask(runtime_bldenv))
            if self.bldenv.overrides.get("check.parfait", True) and \
                    self.bldenv.check:
                self.add_task(ParfaitAnalyseTask(runtime_bldenv))
            else:
                self.add_task(
                    "Skipping check.parfait due to checkoverrides setting")
            self.add_task(FishPackagesTask(runtime_bldenv))
            if self.env["FISH_GK_IMAGES"] == "true":
                self.add_task(FishMkakGkImagesTask(runtime_bldenv))
            elif self.env["FISH_GK_NIGHTLY_IMAGES"] == "true":
                self.add_task(FishMkakGkNightlyImagesTask(runtime_bldenv))
            else:
                self.add_task(FishMkakTask(runtime_bldenv))
            self.add_task(FishSaveArtifactsTask(runtime_bldenv))
            if self.env["FISH_LICENSES"] == "true":
                self.add_task(FishLicensesTask(runtime_bldenv))
                # This is needed since it gets skipped by the
                # FishMkakTask when "FISH_LICENSES" is true.
                self.add_task(FishDestroyDcDataset(runtime_bldenv))

        if self.bldenv.incremental:
            self.add_task("Skipping clobber for incremental build.")
        else:
            self.add_task(ClobberTask(self.bldenv))

        if self.bldenv.debug:
            runtime_bldenv = copy.deepcopy(self.bldenv)
            runtime_bldenv.debug = True
            runtime_bldenv.nondebug = False
            add_common_build_tasks(runtime_bldenv, "DEBUG")

        if self.bldenv.nondebug and self.bldenv.debug:
            self.add_task(
                "Debug fish build completed, starting non-debug build.")

        if self.bldenv.nondebug:
            runtime_bldenv = copy.deepcopy(bldenv)
            runtime_bldenv.debug = False
            runtime_bldenv.nondebug = True
            add_common_build_tasks(runtime_bldenv, "non-DEBUG")

        self.add_task(CheckCompositeTask(bldenv))
        self.add_task(FishDestroyDcDataset(bldenv))


class CheckCompositeTask(CompositeTask):
    """Runs a series of checks on the build output. 'fail_fast' is False for
    these checks, because we want to execute all Tasks, rather than bailing out
    on the first failed task."""

    task_name = "check"
    task_synopsis = "Run a series of checks on the source and proto trees"

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        super(CheckCompositeTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)

        def add_check(task):
            """Determine whether this is a check we should run."""
            if isinstance(task, Task):
                enabled_task = self.bldenv.overrides.get(
                    task.task_name, task.enabled_by_default)
                if not enabled_task:
                    self.add_task(
                        "Skipping %s due to checkoverrides setting" %
                        task.task_name)
                    # the task is no longer pending
                    task.mark_clear()
                    # Since the creation of the Task object may have
                    # instantiated an empty logfile, we'll remove that
                    # if it exists. Some tasks do not have log files.
                    if not task.log_name:
                        return
                    logfile = os.path.join(task.log_dir, task.log_name)
                    if os.path.exists(logfile):
                        try:
                            os.remove(logfile)
                        except OSError:
                            self.add_task("Unable to remove %s!" % logfile)
                    return
            self.add_task(task)
            self.add_task("")

        if not self.bldenv.check:
            self.add_task("Skipping all build checks")
            self.exit_status = 0
        else:
            if self.bldenv.generate_splice:
                add_check(SpliceCheckTask(self.bldenv))

            add_check(FortifyCheckTask(self.bldenv))

            add_check(ProtoCmpCheckTask(self.bldenv))
            if self.env.get("FISH_STANDALONE"):
                add_check("Skipping elf check for fish build.")
            else:
                add_check(ELFVerificationCheckTask(self.bldenv))

            if self.bldenv.nondebug:
                if self.bldenv.incremental:
                    add_check("Skipping ctf-check for incremental build.")
                elif self.env.get("FISH_STANDALONE"):
                    add_check("Skipping ctf-check for fish build.")
                else:
                    add_check(CtfCheckTask(self.bldenv))
                add_check("Skipping lint check for non-debug build.")

            if self.bldenv.debug:
                add_check(LintCheckTask(self.bldenv))
                add_check("Skipping ctf-check for debug build.")

            add_check(CstyleCheckTask(self.bldenv))

            add_check(FindCoresCheckTask(self.bldenv))
            if self.bldenv.incremental:
                add_check("Skipping findunref for incremental build.")
            elif self.env.get("FISH_STANDALONE"):
                add_check(
                    "Skipping findunref for fish build. "
                    "See check.cstyle instead.")
            else:
                add_check(FindUnrefCheckTask(self.bldenv))

            if self.env.get("FISH_STANDALONE"):
                add_check("Skipping paths and pmodes check for fish build.")
            else:
                add_check(PathsCheckTask(self.bldenv))
                add_check(PmodesCheckTask(self.bldenv))

            add_check(HgCommitCheckTask(self.bldenv))
            add_check(ExceptionListCheckTask(self.bldenv))

            if self.env.get("FISH_STANDALONE"):
                add_check("Skipping noise comparison checks for fish build.")
            elif self.bldenv.incremental:
                add_check(
                    "Skipping noise comparison checks for incremental build.")
            else:
                add_check(InstallNoiseCheckTask(self.bldenv))
                add_check(WsdiffCheckTask(self.bldenv))
                add_check(CertificateExpiryCheckTask(self.bldenv))
                if self.bldenv.nondebug:
                    add_check(
                        "Skipping lint noise check for non-debug build.")
                if self.bldenv.debug:
                    add_check(LintNoiseCheckTask(self.bldenv))

    def summary(self):
        """This task has no useful summary since it just aggregates several
        tasks together."""
        return ""


class CtfCheckTask(Task):
    """Check CTF data using the 'ctf-check' target."""

    task_name = "check.ctf"
    task_synopsis = "Check CTF data in the non-debug proto area"
    snapshot = False

    def _execute(self):
        """Check CTF data."""

        if self.bldenv.nondebug:
            # we can't use self.env because that's debug
            env = self.bldenv.get_build_environment(
                debug=False, nondebug=True)
            self.logger.log(
                TINFO_LEVEL, "Checking CTF data in non-debug proto area")
            self.exit_status = util.cmdline_run(
                "dmake ctf-check", env_arg=env, logger=self.logger,
                cwd=self.src)
        else:
            self.logger.log(TINFO_LEVEL, "Skipping ctf-check of debug build")


class CstyleCheckTask(Task):
    """Run cstyle(1) and hdrchk(1) using the 'check' target in $SRC and
    $SRC/tools."""

    task_name = "check.cstyle"
    task_synopsis = "Do cstyle and hdrchck across the source tree"
    snapshot = False
    produce_log_extract = True

    def _execute(self):
        """Do a make check. We only do this once, preferring the debug
        environment."""

        self.logger.log(TINFO_LEVEL, "Doing cstyle/hdrchk of usr/src")
        ret = util.cmdline_run(
            "dmake check", env_arg=self.env, logger=self.logger, cwd=self.src)
        self.exit_status += ret

        if self.env["FISH_IS_PRESENT"] == "yes" and \
                not self.env.get("FISH_STANDALONE"):
            fish_env = self.bldenv.get_build_environment(fish=True)
            self.logger.log(TINFO_LEVEL, "Doing cstyle/hdrchk of usr/fish")
            ret = util.cmdline_run(
                "dmake check", env_arg=fish_env, logger=self.logger,
                cwd=fish_env["SRC"])
            self.exit_status += ret


class FindCoresCheckTask(Task):
    """Looks for core files dumped by build processes and log them along with
    their timestamps."""

    task_name = "check.cores"
    task_synopsis = "Look for core files dumped by build processes"
    snapshot = False
    # we shouldn't blow up the build because of compiler bugs
    must_pass = False

    def _execute(self):
        """Look for cores."""

        self.logger.log(TINFO_LEVEL, "Looking for core files")

        # If we have a specific coreadm pattern for this process that indicates
        # a directory, we can just list the contents of that directory.
        ret, out, err = util.cmdline_run("/usr/bin/coreadm %s" % os.getpid())
        if ret == 0:
            cores_pattern = out.strip().split()[1]
            cores_dir = os.path.dirname(cores_pattern)
            if cores_dir:
                if os.path.isdir(cores_dir):
                    new_cores = []
                    st_tstamp = os.stat(
                        "%s/.build.%s.tstamp" % (self.keep, self.mach))
                    for core_file in os.listdir(cores_dir):
                        # ignore nfs cruft
                        if core_file.startswith(".nfs"):
                            continue
                        core_path = "%s/%s" % (cores_dir, core_file)
                        core_st = os.stat(core_path)
                        if core_st.st_mtime > st_tstamp.st_mtime:
                            new_cores.append(
                                "%-64s%12s" % (core_path, core_st.st_size))

                    if new_cores:
                        self.logger.log(
                            TWARNING_LEVEL,
                            "Core files were found in the workspace:\n"
                            "%-64s%12s\n%s" %
                            ("PATH", "SIZE", "\n".join(new_cores)))
                        self.exit_status = 1
                    return

        # otherwise, do more work to search for all cores because we didn't
        # have a cores directory configured for this process.
        if self.env.get("CLOSED_IS_PRESENT") == "no":
            closed = ""
        else:
            closed = "%s/../closed" % self.src

        cmd = (
            "/usr/bin/find %(src)s %(closed)s -name core -a -type f "
            "-exec file {} + -exec ls -l {} + | grep core" %
            {"src": self.src, "closed": closed})
        ret = util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.src,
            shell=True, exit=1, cmdout_level=TWARNING_LEVEL)

        # We need the pipeline to end with a grep for 'core' in order to get
        # a useful exit code that tells us whether we've encountered a core
        # file or not. The exit code in 'ret' will be from the grep command,
        # 0 if core files were found, so invert that.
        if ret == 0:
            self.exit_status = 1
            self.logger.log(
                TWARNING_LEVEL, "Core files were found in the workspace")


class FindUnrefCheckTask(Task):
    """See findunref(1)."""

    task_name = "check.findunref"
    task_synopsis = "Find unused files in the source tree"
    snapshot = False
    # since findunref uses the previous build in this workspace to compare
    # against, unless that build is close to this one, unrelated unref fixes
    # can cause a false negative.
    must_pass = False

    def _execute(self):
        """Run findunref."""

        if self.bldenv.incremental:
            self.logger.log(
                TINFO_LEVEL, "Skipping findunref check since this is an "
                "incremental build.")
            return

        branch_prefix = self.env["ONBLD_BRANCH_PREFIX"]

        unref_current = "%(keep)s/%(prefix)sunref-%(mach)s.out" % \
            {"keep": self.keep, "prefix": branch_prefix, "mach": self.mach}
        unref_prev = "%(keep)s/%(prefix)sunref-%(mach)s.ref" % \
            {"keep": self.keep, "prefix": branch_prefix, "mach": self.mach}

        if os.path.exists(unref_current):
            shutil.copyfile(unref_current, unref_prev)

        self.logger.log(
            TINFO_LEVEL, "Diff of unreferenced files (since last build)")

        # Ideally we'd pipe these commands together, but since Python doesn't
        # let us do the shell equivalent of 'set -o pipefail' we want to catch
        # places where findunref fails, but sort doesn't, so we use separate
        # commands and a temporary file.
        unref_tmp = tempfile.mkstemp(".out", "findunref.")[1]

        if "BIP" in self.env:
            exceptions_path = "%s/tools/findunref/exception_list" % self.src
        else:
            exceptions_path = "%s/usr/src/tools/findunref/exception_list" % \
                self.env["TOOLBUILD"]
        cmd = (
            "findunref -S hg -t %(keep)s/.build.%(mach)s.tstamp -s usr "
            "%(ws)s %(exceptions_path)s > %(unref_tmp)s" %
            {"exceptions_path": exceptions_path, "mach": self.mach,
                "unref_tmp": unref_tmp,
                "keep": self.keep,
                "ws": self.ws})
        ret = util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.src,
            shell=True)
        ret += util.cmdline_run(
            "sort %(unref_tmp)s > %(unref)s" %
            {"unref": unref_current, "unref_tmp": unref_tmp},
            env_arg=self.env, logger=self.logger, cwd=self.src,
            shell=True)
        os.remove(unref_tmp)

        if not os.path.exists(unref_prev):
            shutil.copyfile(unref_current, unref_prev)

        cmd = "/usr/bin/diff %s %s" % (unref_prev, unref_current)
        self.logger.log(TINFO_LEVEL, cmd)
        nret, out, err = util.cmdline_run(cmd, env_arg=self.env)
        ret += nret
        if out:
            self.logger.log(
                TWARNING_LEVEL, "Differences found between %s and %s" %
                (unref_prev, unref_current))
            for line in out.split("\n"):
                self.logger.log(TWARNING_LEVEL, line)

        # the task will fail if we detect differences in the unref output
        self.exit_status = ret


class SpliceCheckTask(Task):
    """Validates the contents of the splice build relative to the baseline."""

    task_name = "check.splice"
    task_synopsis = "Compare splice build repositories to baseline"

    def _execute(self):
        def do_splicecheck(debug=False, pkgmerge=False):
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug)

            if pkgmerge:
                pkgmerge_type, pkgmerge_dest, \
                    debug_pkgs_local, debug_pkgs_altws, \
                    nondebug_pkgs_local, nondebug_pkgs_altws = \
                    _compute_pkgmerge_params(self.bldenv)
                file_suffix = os.path.basename(pkgmerge_dest)
            elif debug:
                file_suffix = "debug"
            else:
                file_suffix = "nondebug"

            return util.cmdline_run(
                "dmake analyse_splice CMP_SFX={0}".format(file_suffix),
                env_arg=env, logger=self.logger,
                cwd=os.path.join(self.src, "pkg"))

        if self.bldenv.pkgmerge:
            # compare our already-merged repositories
            # no need to do this separately for debug/nondebug
            self.exit_status = do_splicecheck(pkgmerge=True)
        else:
            if self.bldenv.debug:
                self.exit_status += do_splicecheck(debug=True)

            if self.bldenv.debug and self.bldenv.nondebug:
                self.logger.log(
                    TINFO_LEVEL,
                    "Debug splice analysis finished, starting non-debug")
                log.switch_logs(
                    self.logger, logsuffix="-nd", old_logsuffix="",
                    level=TASK_LEVEL)

            if self.bldenv.nondebug:
                self.exit_status += do_splicecheck()

        if self.exit_status > 0:
            self.exit_status = 1


class ProtoCmpCheckTask(Task):
    """Runs a protolist and protocmp on $ROOT. We only need to do this once for
    a given build: either debug or non-debug is fine."""

    task_name = "check.protocmp"
    task_synopsis = "Run protolist and protocmp on a built workspace"

    def _execute(self):

        # usr/src protocmp checks
        if not self.env.get("FISH_STANDALONE"):
            proto_list_path = "%s/%sproto_list_%s" % (
                self.env["LOGDIR"], self.env["ONBLD_BRANCH_PREFIX"], self.mach)
            self.logger.log(TINFO_LEVEL, "Creating protolist system-file")
            cmd = "protolist %s > %s" % (self.env["ROOT"], proto_list_path)

            ret = util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd=self.src,
                shell=True)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to create protolist file")
                self.exit_status = 1
                return

            self.logger.log(
                TINFO_LEVEL, "Validating manifests against proto area")
            ret = util.cmdline_run(
                "dmake protocmp", env_arg=self.env, logger=self.logger,
                cwd="%s/pkg" % self.src)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Problems found comparing package "
                    "manifests against the proto area in %s" %
                    self.env["ROOT"])
                self.exit_status = 1

            # even if the above command fails, we'll still try to compare
            # against the reference proto list.
            if "REF_PROTO_LIST" in self.env:
                if not os.path.exists(self.env["REF_PROTO_LIST"]):
                    # The ref_proto_list is generally stored in
                    # /ws/on-gate/public.
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to access reference proto "
                        "list at %s" % self.env["REF_PROTO_LIST"])
                    self.exit_status = 1
                    return

                cmd = (
                    "protocmp.terse -gu "
                    "\"Files in yesterday's proto area, but not today's:\" "
                    "\"Files in today's proto area, but not yesterday's:\" "
                    "\"Files that changed between yesterday and today:\" "
                    "-d %(ref_proto_list)s %(proto_list_path)s" %
                    {"ref_proto_list": self.env["REF_PROTO_LIST"],
                        "proto_list_path": proto_list_path})
                ret = util.cmdline_run(
                    cmd, env_arg=self.env, logger=self.logger, cwd=self.ws)

                if ret != 0:
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Problems found when comparing the parent "
                        "proto list in %s with our proto area in %s" %
                        (self.env["REF_PROTO_LIST"], self.env["ROOT"]))
                    self.exit_status = 1
            else:
                self.logger.log(
                    TINFO_LEVEL, "Skipping usr/src ref proto checks, "
                    "'ref_proto_list' not set in bldrc")

        # usr/fish protocmp checks
        if self.env["FISH_IS_PRESENT"] == "yes":
            def do_fish_protocmp(debug=True):
                if debug:
                    nondebug = False
                    label = "debug"
                else:
                    nondebug = True
                    label = "nondebug"
                fish_env = self.bldenv.get_build_environment(
                    debug=debug, nondebug=nondebug, fish=True)
                src = fish_env["SRC"]
                self.logger.log(
                    TINFO_LEVEL, "Doing %s protocmp of %s" % (label, src))
                ret = util.cmdline_run(
                    "dmake protocmp", env_arg=fish_env, logger=self.logger,
                    cwd=os.path.join(src, "pkgdefs"))
                self.exit_status += ret

            if self.bldenv.debug:
                do_fish_protocmp(debug=True)
            if self.bldenv.nondebug:
                do_fish_protocmp(debug=False)


class PathsCheckTask(Task):
    """Runs checkpaths(1) on a built workspace."""

    task_name = "check.paths"
    task_synopsis = "Run checkpaths(1) on a built workspace"
    snapshot = False

    def _execute(self):
        cmdline = "checkpaths -b %s" % self.env["ROOT"]
        cwd = "%s/pkg" % self.src
        self.logger.log(CMDLINE_LEVEL, "%s (CWD=%s)" % (cmdline, cwd))
        ret, out, err = util.cmdline_run(cmdline, env_arg=self.env, cwd=cwd)
        if out or err:
            self.exit_status = 1

        for line in (out + err).split("\n"):
            line = line.rstrip().replace(self.env["WS"] + "/", "")
            if not line:
                continue
            self.logger.log(TWARNING_LEVEL, line.rstrip())


class PmodesCheckTask(Task):
    """Do pmodes checks."""

    task_name = "check.pmodes"
    task_synopsis = "Run a pmodes check on a built workspace"
    snapshot = False

    def _execute(self):
        ret = util.cmdline_run(
            "dmake pmodes", env_arg=self.env, logger=self.logger,
            cwd="%s/pkg" % self.src)
        if ret != 0:
            self.logger.log(TWARNING_LEVEL, "Pmodes check failed.")
            self.exit_status = 1


class FortifyCheckTask(Task):
    """Scan Python, Javascript, HTML, CSS, and XML"""

    task_name = "check.fortify"
    task_synopsis = "Run a Fortify scan on Python and other files "
    enabled_by_default = False

    def _execute(self):
        ret, py3path, err = util.cmdline_run(
            "python3 -c \"import sys;print(\':{}\'.join(sys.path) + \
                ':'.join(sys.path))\"".format(self.env["ROOT"]),
            env_arg=self.env)

        shutil.rmtree(self.env["FORTIFY_INTERMEDIATE_DIR"], ignore_errors=True)
        # it takes 6 hours and 40GB RAM to process usr/**/*.py
        ret = util.cmdline_run(
            "{}/bin/sourceanalyzer "
            "-project-root {} "
            "-b SolarisTrunk "
            "usr/**/*.py "
            "usr/**/*.js "
            "usr/**/*.html "
            "usr/**/*.css "
            "usr/**/*.xml "
            "-python-version 3 -python-path \"{}\"".format(
                self.env["FORTIFY_HOME"], self.env["FORTIFY_INTERMEDIATE_DIR"],
                py3path[:-1]),
            env_arg=self.env, logger=self.logger,
            cwd=self.env["WS"])
        if ret != 0:
            self.logger.log(TWARNING_LEVEL, "Fortify check failed "
                            "in analysis.")
            self.exit_status = 1

        # Scan the sources and save results in an fpr file
        ret = util.cmdline_run(
            "{}/bin/sourceanalyzer "
            "-project-root {} "
            "-b SolarisTrunk "
            "-scan "
            "-f data/fortify/solaris.fpr "
            "-XX:SoftRefLRUPolicyMSPerMB=3000 -Xmx70G "
            "-XX:+CMSClassUnloadingEnabled -XX:+UseParallelGC -Xss16M".format(
                self.env["FORTIFY_HOME"],
                self.env["FORTIFY_INTERMEDIATE_DIR"]),
            env_arg=self.env, logger=self.logger,
            cwd=self.env["WS"])
        if ret != 0:
            self.logger.log(TWARNING_LEVEL, "Fortify check failed in scan.")
            self.exit_status = 1


class FortifyUploadTask(Task):
    """Upload Fortify results to Fortify server"""

    task_name = "check.fortifyupload"
    task_synopsis = "Upload Fortify results to Fortify server"
    enabled_by_default = False

    def _execute(self):
        local_file = self.env["WS"] + "/data/fortify/solaris.fpr"

        w_token_file = open(self.env["FORTIFY_WRITE_TOKEN"], "r")
        w_token = w_token_file.readline().strip()
        w_token_file.close()

        cmd = "{0}/bin/fortifyclient -url {1} --authtoken {2} " \
            "uploadFPR -file {3} -project {4} -version {5} ".format(
                self.env["FORTIFY_HOME"],
                self.env["FORTIFY_SERVER"],
                w_token,
                local_file,
                self.env["FORTIFY_PROJECT"],
                self.env["FORTIFY_PROJECT_VERSION"])

        self.logger.log(TINFO_LEVEL, "Executing {}".format(cmd))
        ret, result, err = util.cmdline_run(cmd, env_arg=self.env)

        if ret != 0:
            self.exit_status = 1
            self.logger.log(TWARNING_LEVEL, "Fortify results upload failed!")


class ELFVerificationCheckTask(Task):
    """Run a series of checks to verify ELF versioning, ABI information, and
    the ELF runtime. Since this is an elf check, let's ope the gate is elfy.
    """

    task_name = "check.elf"
    task_synopsis = "Run a series of checks on built ELF objects"

    def _execute(self):
        elf_ddir = "%s/ELF-data.%s" % (self.keep, self.mach)

        # if we have an elf data directory, rotate and use this one
        shutil.rmtree("%s.ref" % elf_ddir, ignore_errors=True)
        if os.path.isdir(elf_ddir):
            os.rename(elf_ddir, "%s.ref" % elf_ddir)
        os.makedirs(elf_ddir)

        # Call find_elf to produce a list of the ELF objects in the proto area.
        # This list is passed to check_rtime and interface_check, preventing
        # them from separately calling find_elf to do the same work twice.
        cmd = "find_elf -fr %s > %s/object_list" % (self.env["ROOT"], elf_ddir)
        util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.ws, shell=True)

        self.logger.log(TINFO_LEVEL, "Check versioning and ABI information")

        # Produce interface description for the proto. Report errors.
        cmd = (
            "interface_check -o -w %s -f object_list -i interface "
            "-E interface.err" % elf_ddir)
        util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.ws)
        iferr_path = "%s/interface.err" % elf_ddir

        if os.path.isfile(iferr_path) and os.path.getsize(iferr_path) > 0:
            self.exit_status = 1
            self.logger.log(TWARNING_LEVEL, "Errors from interface_check:")
            with open(iferr_path, "r") as iferr:
                for line in iferr:
                    self.logger.log(TWARNING_LEVEL, line.rstrip())

        def elf_data_baseline(elf_data_baseline_dir, elf_data_baseline_type):
            """Compare the new interface description file to that from the
            baseline gate."""
            # Put a copy of our interface_cmp exceptions file in the ELF data
            # directory for the benefit of older gates using us as their
            # baseline.
            shutil.copy(
                "%s/exception_lists/interface_cmp" % self.ws,
                "%s/interface_cmp.except" % elf_ddir)

            base_efile = "%s/interface_cmp.except" % elf_data_baseline_dir
            base_ifile = "%s/interface" % elf_data_baseline_dir

            # The order in which we specify the interface files and other
            # options to interface_cmp depend on the relationship between the
            # baseline and the workspace.
            icmp_dflag = ""
            icmp_eflag = ""
            if elf_data_baseline_type == "DIRECT_SUCCESSOR":
                old_ifile = "%s/interface" % elf_ddir
                new_ifile = base_ifile
                icmp_dflag = "-d"
                icmp_eflag = "-e %s" % base_efile
            elif elf_data_baseline_type == "PREDECESSOR":
                old_ifile = base_ifile
                new_ifile = "%s/interface" % elf_ddir
            elif elf_data_baseline_type == "SUCCESSOR":
                old_ifile = "%s/interface" % elf_ddir
                new_ifile = base_ifile
                icmp_eflag = "-e %s" % base_efile
            else:  # DIRECT_PREDECESSOR
                if elf_data_baseline_type != "DIRECT_PREDECESSOR":
                    self.logger.log(
                        TWARNING_LEVEL, "unrecognised 'elf_data_baseline' "
                        "value. Assuming DIRECT_PREDECESSOR")
                old_ifile = base_ifile
                new_ifile = "%s/interface" % elf_ddir
                icmp_dflag = "-d"

            self.logger.log(
                TINFO_LEVEL, "Compare versioning and ABI information to "
                "baseline")

            # The baseline file is often a symlink from one gate's public
            # area pointing at another gate. If so, then display both
            # the given path and its resolved canonical form, in the
            # "name -> target" style used by /bin/ls for symlinks. This
            # helps gatelings see where the baseline really comes from.
            base_realpath = os.path.realpath(base_ifile)
            if base_ifile == base_realpath:
                self.logger.log(TINFO_LEVEL, "Baseline: %s" % base_ifile)
            else:
                self.logger.log(
                    TINFO_LEVEL, "Baseline: %s -> %s" %
                    (base_ifile, base_realpath))

            icmp_path = "%s/interface.cmp" % elf_ddir
            icmp_cmd = (
                "interface_cmp %(icmp_dflag)s %(icmp_eflag)s -o "
                "%(old_ifile)s %(new_ifile)s > %(icmp_path)s" %
                {"icmp_dflag": icmp_dflag, "icmp_eflag": icmp_eflag,
                    "old_ifile": old_ifile, "new_ifile": new_ifile,
                    "icmp_path": icmp_path})

            if os.path.isfile(base_ifile):
                try:
                    baseline_age = int(
                        self.env.get("ELF_DATA_BASELINE_AGE", 0))
                except (ValueError, TypeError):
                    self.logger.log(
                        TWARNING_LEVEL, "'elf_data_baseline_age' should be an "
                        "integer! Using '0' instead of %s" %
                        self.env["ELF_DATA_BASELINE_AGE"])
                    baseline_age = 0

                if baseline_age > 0:
                    cmd = (
                        "find %(base_ifile)s \\! -mtime -%(baseline_age)s "
                        "-exec echo baseline {} is more than "
                        "%(baseline_age)s days old \\;" % {
                            "base_ifile": base_ifile,
                            "baseline_age": baseline_age})
                    ret, out, err = util.cmdline_run(
                        cmd, env_arg=self.env, cwd=self.ws)

                    # any output here indicates a problem we'll want to report
                    if out or err:
                        self.logger.log(log.CMD_LEVEL, cmd)
                        for val in [out, err]:
                            for line in val.splitlines():
                                self.logger.log(TWARNING_LEVEL, line)
                        self.exit_status = 1

                util.cmdline_run(
                    icmp_cmd, env_arg=self.env, logger=self.logger,
                    cwd=self.ws, shell=True)

                if os.path.exists(icmp_path) and \
                        os.path.getsize(icmp_path) > 0:
                    self.exit_status = 1
                    with open(icmp_path, "r") as icmp_file:
                        for line in icmp_file:
                            self.logger.log(TWARNING_LEVEL, line.rstrip())

        # If ELF_DATA_BASELINE_DIR is defined, compare the new interface
        # description file to that from the baseline gate. Issue a warning if
        # the baseline is not present, and keep going.
        if "ELF_DATA_BASELINE_DIR" in self.env:
            if os.path.exists(self.env["ELF_DATA_BASELINE_DIR"]):
                elf_data_baseline(
                    self.env["ELF_DATA_BASELINE_DIR"],
                    self.env["ELF_DATA_BASELINE_TYPE"])
            else:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to access elf data baseline dir "
                    "at %s" % self.env["ELF_DATA_BASELINE_DIR"])
                self.exit_status = 1
        else:
            self.logger.log(
                TINFO_LEVEL, "Skipping ELF interface baseline checks")

        self.logger.log(TINFO_LEVEL, "Check ELF runtime attributes")

        # Use the -s option to ensure that stab sections are removed
        # following CTF generation in release objects.
        if self.bldenv.release:
            rtime_sflag = "-s"
        else:
            rtime_sflag = ""

        cmd = (
            "check_rtime -i -m -v %(rtime_sflag)s -o -w %(elf_ddir)s "
            "-D object_list -f object_list -E runtime.err "
            "-I runtime.attr.raw" %
            {"rtime_sflag": rtime_sflag, "elf_ddir": elf_ddir})
        util.cmdline_run(
            cmd, env_arg=self.env, logger=self.logger, cwd=self.ws)

        # check_rtime -I output needs to be sorted in order to
        # compare it to that from previous builds.
        util.cmdline_run(
            "/usr/bin/sort %s/runtime.attr.raw > %s/runtime.attr" %
            (elf_ddir, elf_ddir), env_arg=self.env, logger=self.logger,
            cwd=self.ws, shell=True)
        try:
            os.remove("%s/runtime.attr.raw" % elf_ddir)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass

        # Report errors
        runtime_err_path = "%s/runtime.err" % elf_ddir
        if os.path.exists(runtime_err_path) and \
                os.path.getsize(runtime_err_path) > 0:
            self.exit_status = 1
            self.logger.log(TWARNING_LEVEL, "Errors from check_rtime:")
            with open(runtime_err_path, "r") as runtime_err:
                for line in runtime_err:
                    self.logger.log(TWARNING_LEVEL, line.rstrip())

        # If there is an ELF-data directory from a previous build, then diff
        # the attr files. These files contain information about dependencies,
        # versioning, and runpaths. There is some overlap with the ABI checking
        # done above, but this also flushes out non-ABI interface differences
        # along with the other information.
        if os.path.exists("%s.ref/runtime.attr" % elf_ddir):
            ret, out, err = util.cmdline_run(
                "/usr/bin/diff %s.ref/runtime.attr %s/runtime.attr" %
                (elf_ddir, elf_ddir), env_arg=self.env, cwd=self.ws)
            if out:
                self.logger.log(
                    TWARNING_LEVEL, "Differences found in ELF runtime "
                    "attributes since last build")
                for line in out.split("\n"):
                    self.logger.log(TWARNING_LEVEL, line)

        # Finally, upload the results to the parent workspace if necessary
        if self.bldenv.update_parent_ws:
            if "PARENT_ELF_DATA_DIR" in self.env:
                elf_dest = "%s/ELF-data.%s" % (
                    self.env["PARENT_ELF_DATA_DIR"], self.mach)
            else:
                elf_dest = os.path.join(
                    self.env["PARENT_WS"], "keep/ELF-data.%s" % self.mach)
            self.logger.log(
                TINFO_LEVEL, "Copying ELF-data.%s to %s" %
                (self.mach, elf_dest))
            try:
                if not os.path.isdir(elf_dest):
                    os.makedirs(elf_dest, 0o755)
                # These files are used asynchronously by other builds for ABI
                # verification, as above for elf_data_baseline(). As such, we
                # require the file replacement to be atomic. Copy the data to a
                # temp file in the same filesystem and then rename into place.
                for name in os.listdir(elf_ddir):
                    shutil.copy(
                        "%s/%s" % (elf_ddir, name),
                        "%s/%s.new" % (elf_dest, name))
                    os.rename(
                        "%s/%s.new" % (elf_dest, name),
                        "%s/%s" % (elf_dest, name))
            except Exception as e:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to copy ELF data from %s to %s: "
                    "%s" % (elf_ddir, elf_dest, e))
                self.exit_status = 1


class InstallNoiseCheckTask(Task):
    """Looks for install noise, and diffs against the previous build."""

    task_name = "check.install-noise"
    task_synopsis = "Looks for noise in the install.log file"
    must_pass = False

    def _execute(self):
        """Run the task."""

        def do_noisecheck(debug=False):
            if debug:
                label = "DEBUG"
                # the full plaintext log file.
                logfile_path = os.path.join(
                    self.log_dir, "%sinstall.log" %
                    self.env["ONBLD_BRANCH_PREFIX"])
                # a copy of the log file with all timestamps and status levels
                # stripped out
                logfile_nots_path = os.path.join(
                    self.keep, "%sinstall-%s.out" %
                    (self.mach, self.env["ONBLD_BRANCH_PREFIX"]))
                # the log file containing uniquified, sorted noise.
                noise_logfile_path = os.path.join(
                    self.keep, "%snoise-%s.out" %
                    (self.env["ONBLD_BRANCH_PREFIX"], self.mach))
            else:
                label = "non-DEBUG"
                logfile_path = os.path.join(
                    self.log_dir, "%sinstall-nd.log" %
                    self.env["ONBLD_BRANCH_PREFIX"])
                logfile_nots_path = os.path.join(
                    self.keep, "%sinstall-nd-%s.out" %
                    (self.env["ONBLD_BRANCH_PREFIX"], self.mach))
                noise_logfile_path = os.path.join(
                    self.keep, "%snoise-nd-%s.out" %
                    (self.env["ONBLD_BRANCH_PREFIX"], self.mach))

            # If nightly has rotated the log files, or we don't have a recent
            # install log, then we can't go any further.
            if not os.path.exists(logfile_path):
                self.logger.log(
                    TWARNING_LEVEL, "Not running %s noise checks: unable to "
                    "find %s" % (label, logfile_path))
                return 1

            noise_ref_path = "%s.ref" % noise_logfile_path
            # rotate the old install-noise log file
            if os.path.exists(noise_logfile_path):
                noise_ref_path = "%s.ref" % noise_logfile_path
                if os.path.exists(noise_ref_path):
                    os.remove(noise_ref_path)
                os.rename(noise_logfile_path, noise_ref_path)

            # determine the column where the text of the log message starts,
            # then remove the timestamp and log status level from the log
            cols = log.log_prefix_cols[self.bldenv.log_format] - 1
            try:
                with open(logfile_nots_path, "w") as logfile_notimestamp:
                    for line in open(logfile_path, "r").readlines():
                        if "TINFO   : " in line:
                            continue
                        if "STATE   : " in line:
                            continue
                        if "SUMMARY : " in line:
                            continue
                        logfile_notimestamp.write(line[cols:])
            except IOError as e:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to produce temporary no-timestamp "
                    "log file: %s" % e)
                self.exit_status = 1
                return 1

            # A list of regular expressions which match non-deterministic
            # build log output - that is, output which can differ between
            # builds. Ideally we'd look for times and dates as well, and other
            # content that can change between builds, but this will do for now.
            # When comparing logs, we remove lines that match these regexps,
            # then compare the remaining lines of the log, looking for
            # noise differences.
            #
            # When checking for matches, if any regexps in the list match
            # no content from the log file, we issue a warning, indicating that
            # this list is out of date and needs to be updated.
            regexp_list = [
                "^/",
                "ar: creating",
                "ar: writing",
                "conflicts:",
                "^Writing",
                "spellin1\\: ",
                "^adding:",
                "^info:",
                "\\.c:$",
                "::",
                "^\\+",
            ]

            # Machine-specific regexps. We only have no sparc-specific noise
            # yet.
            regexp_list_i386 = [
                "#define GRUB_EXTENDED_VERSION",
            ]

            if self.mach == "i386":
                regexp_list.extend(regexp_list_i386)

            # This regexp matches content only when we're looking at a debug
            # log file.
            if debug:
                regexp_list.append("chars, width")

            # store precompiled re objects for our regexps
            compiled_regexps = {}
            # count how many matches we get for each regexp
            matches = {}

            for regexp in regexp_list:
                comp = re.compile(regexp)
                compiled_regexps[regexp] = comp
                matches[regexp] = 0

            noise_logfile = None
            try:
                noise_tmp_logfile = open("%s.tmp" % noise_logfile_path, "w")
                # walk through the log file, writing lines that don't
                # match any of our regular expressions, counting the number of
                # matches per regexps as we go.
                with open(logfile_nots_path) as logfile:
                    for line in logfile:
                        matched = False
                        content = line.rstrip()
                        for regexp in compiled_regexps:
                            if compiled_regexps[regexp].search(content):
                                matches[regexp] += 1
                                matched = True
                                break
                        if not matched and content:
                            noise_tmp_logfile.write(line)
                    noise_tmp_logfile.close()
                # we don't need to keep the no-timestamps log file anymore
                os.remove(logfile_nots_path)

                # if any regular expressions matched no lines anywhere in the
                # log file, then we have redundant code here. The regexp
                # should be removed from this script, or the user should
                # investigate why the regexp no longer matches.
                for regexp in matches:
                    if matches[regexp] == 0:
                        self.logger.log(
                            TWARNING_LEVEL, "Regexp %s matched no content in "
                            "the log when performing the %s build noise "
                            "check. If the build does not produce messages of "
                            "this type any more, then this check should be "
                            "removed from build.py." % (regexp, label))

                # sort and uniquify the noise file
                cmd = "sort -u %s > %s" % (
                    "%s.tmp" % noise_logfile_path, noise_logfile_path)
                self.logger.log(
                    TINFO_LEVEL, "Generating noise log file %s" %
                    noise_logfile_path)
                ret = util.cmdline_run(
                    cmd, logger=self.logger, env_arg=self.env, shell=True)

                os.remove("%s.tmp" % noise_logfile_path)

                if os.path.exists(noise_ref_path):
                    cmd = "diff %s %s" % (noise_ref_path, noise_logfile_path)
                    ret, out, err = util.cmdline_run(cmd, env_arg=self.env)
                    if ret != 0:
                        self.logger.log(
                            TWARNING_LEVEL, "Differences found in %s build "
                            "noise, comparing %s and %s:" %
                            (label, noise_ref_path, noise_logfile_path))
                        for line in out.split("\n"):
                            self.logger.log(TWARNING_LEVEL, line)
                        return 1
                return 0

            finally:
                if noise_logfile:
                    noise_logfile.close()
            return 0

        if self.bldenv.debug:
            self.exit_status += do_noisecheck(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug noise check finished, starting non-debug "
                "noise check")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TINFO_LEVEL)

        if self.bldenv.nondebug:
            self.exit_status += do_noisecheck()


def gather_lint_noise(lint_log_path, task):
    """Return a set of strings containing lint warnings from the supplied
    lint_log_path. Use the Task object task for access to the logger and
    task configuration. This function is used by the LintCheckTask and
    LintNoiseCheckTask."""
    lint_nots_path = "%s/%slint-noise-%s.nots" % (
        task.keep, task.env["ONBLD_BRANCH_PREFIX"], task.mach)

    # In order to produce a deterministic log of lint warnings, we need
    # to remove any timestamps present in the lint log. Determine the
    # column where the text of the log message starts based on the
    # log_format, then remove the timestamp and log status level.
    cols = log.log_prefix_cols[task.bldenv.log_format] - 1
    try:
        with open(lint_log_path, "r") as lint_log:
            with open(lint_nots_path, "w") as lint_nots:
                for line in lint_log.readlines():
                    if "TINFO   : " in line:
                        continue
                    if "STATE   : " in line:
                        continue
                    if "SUMMARY : " in line:
                        continue
                    lint_nots.write(line[cols:])
    except IOError as e:
        task.logger.log(
            TWARNING_LEVEL, "Unable to produce temporary no-timestamp log "
            "file: %s" % e)
        task.exit_status = 1
        return

    lint_noise_regexps = [
        "(library construction)",
        ": global crosschecks",
        "Ignoring unknown host",
        "\\.c:$",
        "\\.s:$",
        "========================"
    ]

    lint_noise = set()
    # Gather lint noise: any lines containing colons that don't match
    # the regular expressions in lint_noise_regexps. Unlike install log
    # noise, lint logs are small enough that we can do this in memory.
    with open(lint_nots_path, "r") as lint_nots:
        for line in lint_nots:
            if ":" not in line:
                continue
            seen_match = False
            for regexp in lint_noise_regexps:
                if not seen_match and re.search(regexp, line):
                    seen_match = True
            if not seen_match:
                lint_noise.add(line)

    os.remove(lint_nots_path)
    return lint_noise


class LintCheckTask(Task):
    """Run lint(1) over the source tree using the 'lint' target."""

    task_name = "check.lint"
    task_synopsis = "Do a 'dmake lint' on the source tree"
    snapshot = False
    produce_log_extract = True

    def _execute(self):
        """ Perform lint checks."""

        if self.bldenv.debug:
            # Remove all .ln files to ensure a full reference file
            # but leave lint libraries in place.
            cmd = (
                "/usr/bin/find %s -name .hg -prune -o "
                "-type f -name '*.ln' ! -name 'llib*.ln' "
                "-print -exec rm -f {} +" %
                self.env["BUILD"])
            ret = util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd=self.src)
            self.logger.log(
                TINFO_LEVEL, "Start 'dmake lint' of %s" % self.src)
            ret = util.cmdline_run(
                "dmake lint", env_arg=self.env, logger=self.logger,
                cwd=self.src)

            if self.env.get("FISH_IS_PRESENT") == "yes" and \
                    not self.env.get("FISH_STANDALONE"):
                fish_env = self.bldenv.get_build_environment(
                    debug=True, nondebug=False, fish=True)
                self.logger.log(
                    TINFO_LEVEL,
                    "Start 'dmake lint' of %s" % fish_env["SRC"])
                ret += util.cmdline_run(
                    "dmake lint", env_arg=fish_env, logger=self.logger,
                    cwd=fish_env["SRC"])

            self.exit_status = ret

            lint_log_path = "%s/%scheck.lint.log" % (
                self.log_dir, self.env["ONBLD_BRANCH_PREFIX"])
            lint_noise = gather_lint_noise(lint_log_path, self)

            reported_lint_warnings = False
            for noise in sorted(lint_noise):
                # This ought to be updated so that we don't grep out
                # pkglint warnings, which do not match this pattern.
                if re.search("(warning|lint):", noise, flags=re.IGNORECASE):
                    if not reported_lint_warnings:
                        self.exit_status += 1
                        reported_lint_warnings = True
                        self.logger.log(
                            TWARNING_LEVEL,
                            "The following lint warnings were reported:")
                    self.logger.log(TWARNING_LEVEL, noise.strip())
        else:
            self.logger.log(
                TINFO_LEVEL, "Skipping 'dmake lint' of %s for non-debug "
                "build" % self.src)


class LintNoiseCheckTask(Task):
    """Diffs lint noise against the previous build."""

    task_name = "check.lint-noise"
    task_synopsis = "Looks for noise in the check.lint.log file"
    must_pass = False

    def _execute(self):
        """Run the task."""

        if self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Skipping lint noise check for non-debug build")

        if not self.bldenv.debug:
            return

        lint_log_path = "%s/%scheck.lint.log" % (
            self.log_dir, self.env["ONBLD_BRANCH_PREFIX"])

        # if nightly has rotated our lint_log_path, or it doesn't exist,
        # we can't go any further
        if not os.path.exists(lint_log_path):
            self.logger.log(
                TWARNING_LEVEL, "Unable to check for lint noise: %s does not "
                "exist" % lint_log_path)
            self.exit_status = 1
            return

        lint_noise_path = "%s/%slint-noise-%s" % (
            self.keep, self.env["ONBLD_BRANCH_PREFIX"], self.mach)
        lint_noise_ref_path = "%s.ref" % lint_noise_path

        # rotate the log files
        if os.path.exists(lint_noise_path):
            os.rename(lint_noise_path, lint_noise_ref_path)

        lint_noise = gather_lint_noise(lint_log_path, self)

        if lint_noise:
            with open(lint_noise_path, "w") as lint_noise_file:
                for noise in sorted(lint_noise):
                    lint_noise_file.write(noise)
        else:
            self.logger.log(TINFO_LEVEL, "No lint noise detected.")

        if os.path.exists(lint_noise_ref_path) and lint_noise:
            cmd = "diff %s %s" % (lint_noise_ref_path, lint_noise_path)
            self.logger.log(TINFO_LEVEL, cmd)
            ret, out, err = util.cmdline_run(cmd, env_arg=self.env)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Differences found in lint noise, "
                    "comparing %s and %s:" %
                    (lint_noise_ref_path, lint_noise_path))
                for line in out.split("\n"):
                    self.logger.log(TWARNING_LEVEL, line)
                self.exit_status = 1


class WsdiffCheckTask(Task):
    """Use wsdiff(1) to compare two proto areas."""

    task_name = "check.wsdiff"
    task_synopsis = "Run wsdiff(1) to compare this and the previous proto area"
    enabled_by_default = False
    must_pass = False

    def _execute(self):
        """Run the task."""

        def do_wsdiff(debug=False):
            """Run wsdiffs."""
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)

            proto = env["ROOT"]
            prev_proto = "%s.prev" % env["ROOT"]
            if not os.path.isdir(prev_proto) or not os.path.isdir(proto):
                self.logger.log(
                    TWARNING_LEVEL, "wsdiffs can only run if %s and %s both "
                    "exist" % (proto, prev_proto))
                return 1

            mfarg = ""
            if "BIP" in self.env:
                mfdir = "%s/pkg/packages.%s" % (env["SRC"], self.mach)
            else:
                mfdir = "%s/usr/src/pkg/packages.%s" % (
                    env["BUILD"], self.mach)
            if os.path.isdir(mfdir):
                mfarg = "-m %s" % mfdir

            self.logger.log(
                TINFO_LEVEL, "Getting object changes since last build")

            results_path = os.path.join(env["LOGDIR"], "wsdiff.results")
            if os.path.exists(results_path):
                ctime = os.stat(results_path).st_ctime
                ts = datetime.datetime.fromtimestamp(ctime).strftime(
                    log.datefmt)
                os.rename(results_path, "%s.%s" % (results_path, ts))

            cmd = (
                "wsdiff -s -r %(results_path)s "
                "%(mfarg)s %(prev_proto)s %(proto)s" % {
                    "results_path": results_path,
                    "mfarg": mfarg, "prev_proto": prev_proto, "proto": proto})
            return util.cmdline_run(cmd, logger=self.logger, env_arg=env)

        if self.bldenv.debug:
            ret = do_wsdiff(debug=True)
            if ret > 0:
                self.exit_status = ret

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "wsdiffs run for debug proto area, starting "
                "wsdiffs for non-debug proto area")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TINFO_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            ret = do_wsdiff()
            if ret > 0:
                self.exit_status = ret


class ElfsigncmpCheckTask(Task):
    """This task does no work. It exists as a placeholder to determine
    whether $ELFSIGN in the environment should be set to 'elfsign' or
    'elfsigncmp', the latter of which causes futher checks to be performed
    on the signed binary during the 'install' task.

    We create no logfiles for this task, nor do we create snapshots because
    the task itself does nothing - it will always pass.
    """

    task_name = "check.elfsigncmp"
    task_synopsis = "Determines whether elfsigncmp is used to sign binaries"
    snapshot = False
    enabled_by_default = False

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(ElfsigncmpCheckTask, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs)

    def _execute(self):
        pass


class HgCommitCheckTask(Task):
    """This task will fail if it encounters any uncommitted changes in the
    $SRC, usr/closed or usr/man repositories. It is enabled by default,
    but marked as non-fatal."""

    task_name = "check.uncommitted"
    task_synopsis = "Look for untracked files in the workspace"
    must_pass = False

    def _execute(self):
        """Run the task."""

        # The Build In Place build litters $SRC with build objects - we cannot
        # run this check there.
        if "BIP" in self.env:
            self.logger.log(
                TWARNING_LEVEL, "This is a BIP build: unable to run %s" %
                self.task_name)
            self.logger.log(
                TWARNING_LEVEL, "Please be careful to commit all changes.")
            return

        # show modified, added, removed, deleted (but tracked) and untracked
        # files.
        cmd = "hg -R %s status -mardu"
        repos = get_hg_info(self.ws, self.env, hg_identify=False)
        uncommitted = {}

        for repo in repos:
            repos[repo] = None
            hg_env = copy.deepcopy(self.env)
            hg_env["HG_PLAIN"] = "true"
            self.logger.log(TINFO_LEVEL, "Running %s" % (cmd % repo))
            ret, out, err = util.cmdline_run(cmd % repo, env_arg=hg_env)
            if len(out) > 0:
                uncommitted[repo] = out
                self.exit_status = 1

        if self.exit_status:
            self.logger.log(
                TWARNING_LEVEL, "Changes were detected in this workspace "
                "which suggest either incorrect modifications to $SRC or "
                "that no 'hg commit' has happened yet.")
            self.logger.log(TWARNING_LEVEL, "See the following lists.")
            for repo in uncommitted:
                self.logger.log(TWARNING_LEVEL, "Changes in %s" % repo)
                for line in uncommitted[repo].split("\n"):
                    self.logger.log(TINFO_LEVEL, line)


class ParfaitAnalyseTask(Task):
    """Perform parfait analysis."""

    task_name = "check.parfait"
    task_synopsis = "Run parfait analysis on a built workspace"
    mail_msg_summary_fragsize = 40
    # There are bugs in parfait analysis in 1.4, so we can't require
    # this to pass at the moment.
    must_pass = False

    def _execute(self):
        """Run the task."""

        # If we're running this task, then we're obviously using parfait.
        # If the user has performed a build separately without having parfait
        # turned on, and attempts to analyse non-existent bytecode, this might
        # cause confusion, so we warn in that case.
        if not self.bldenv.overrides.get("check.parfait", True) or \
                not self.bldenv.check:
            self.logger.log(
                TWARNING_LEVEL, "Attempting to perform parfait analysis "
                "when 'check.parfait' was set to 'false'.")
            self.logger.log(
                TWARNING_LEVEL, "This task will assume a parfait build was "
                "performed, and will continue with baseline analysis.")
            self.bldenv.parfait_type = "baseline"

        # 16 jobs is probably enough, but use dmake_max_jobs if that's lower
        parfait_max_jobs = 16
        dmake_max_jobs = int(self.env.get("DMAKE_MAX_JOBS", 0))
        if dmake_max_jobs < parfait_max_jobs:
            parfait_max_jobs = dmake_max_jobs

        def do_analyse(debug=False, logdir="/"):
            """Perform parfait analysis for either a debug or nondebug build
            choosing to either upload new results, or analyse against a known
            baseline.  The return code is 3 when something goes wrong, since
            Parfait has it's own return codes, which we want to be compatible
            with (see record_parfait_results())"""

            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)

            # Parfait still uses $CODEMGR_WS, so add it to our env copy
            env["CODEMGR_WS"] = env["WS"]

            if debug:
                label = "DEBUG"
                suffix = ""
            else:
                label = "non-DEBUG"
                suffix = "-nd"

            if not os.path.exists(logdir):
                os.makedirs(logdir, 0o755)
            parfait_logfmt = "text"
            args = []
            uri = "%s%s" % (env["PARFAIT_SERVER_URL"], suffix)

            if self.bldenv.parfait_type == "upload":
                try:
                    # When the gatekeepers switch parfait versions, they modify
                    # the current workspace, and build and re-upload with the
                    # same baseline ID. It's important not to use 'hg identify'
                    # here, because then the Parfait baseline identifiers will
                    # be different.
                    hg_info = get_hg_info(self.ws, env, hg_identify=False)
                    hg_tip = hg_info[self.ws][0]
                except RuntimeError as e:
                    self.logger.log(TWARNING_LEVEL, e)
                    hg_tip = None

                if not hg_tip:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to determine hg tip changeset "
                        "for %s" % self.ws)
                    self.exit_status = 1
                    return 3

                self.logger.log(
                    TINFO_LEVEL, "Parfait analysis upload %s: %s" %
                    (label, uri))
                args.append("-s %s -r %s" % (uri, hg_tip))

                parfait_server_passwd = self.env.get("PARFAIT_SERVER_PASSWORD")
                if parfait_server_passwd:
                    args.append("-P %s" % parfait_server_passwd)

            elif self.bldenv.parfait_type == "baseline":
                args.append("-b %s" % uri)

            args.append("-j %d" % parfait_max_jobs)
            args.append("-f %s" % parfait_logfmt)
            args.append("-z %s" % self.ws)
            args.append("-c %s/tools/etc/parfait-solaris.conf" % self.src)
            args.append("-X *.linktest.bc")
            args.append("-W")
            args.append(
                "--exclude-from=%s/exception_lists/parfait_exception_list" %
                self.ws)
            args.append("-o %s/analysis" % logdir)

            if self.env.get("PARFAIT_GUI", "yes") != "no":
                args.append("-g %s/gui --ignore-missing-source" % logdir)

            if self.bldenv.incremental:
                args.append(
                    "--newer %s/.build.%s.tstamp" % (self.keep, self.mach))

            parfait_ws = "%s/usr" % self.ws
            if self.bldenv.build_in_place:
                build_ws = ""
            else:
                build_ws = "%s/usr" % self.env["BUILD"]

            # Parfait output is too verbose for our logs, so we choose to only
            # log the stderr, capturing stdout and stderr to separate files,
            # then rolling up the logs at the end.
            cmd = "%s/bin/parfait %s %s %s" % (
                env["PARFAIT_HOME"], " ".join(args), parfait_ws, build_ws)

            err_path = os.path.join(logdir, "stderr")
            out_path = os.path.join(logdir, "stdout")
            err_file = open(err_path, "w")
            out_file = open(out_path, "w")

            try:
                ret = util.cmdline_run(
                    cmd, env_arg=env, stdout=out_file, stderr=err_file)
            finally:
                err_file.close()
                out_file.close()
            return ret, cmd, out_path, err_path

        def record_parfait_results(exit_code, logdir, cmd, out_path, err_path):
            """Record the parfait results to our logger. If no bugs were
            introduced since the baseline, we treat the parfait analysis
            as succeeding, despite the fact that baseline may include
            many unfixed defects."""

            modified_exit_code = exit_code

            if exit_code == 0:
                self.logger.log(
                    TINFO_LEVEL, "Analysis completed successfully, and no "
                    "bugs were found in the analysed programs.")
            elif exit_code == 1:
                # Intentionally do nothing here - the fact that parfait found
                # errors will be obvious from the rest of the log reporting.
                pass
            elif exit_code == 2:
                with open(os.path.join(logdir, "stdout"), "r") as err:
                    for line in err:
                        self.logger.log(TWARNING_LEVEL, line.rstrip())
                self.logger.log(
                    TWARNING_LEVEL, "An error was encountered in processing "
                    "command line options or the user-supplied configuration "
                    "file.")
            else:
                self.logger.log(
                    TWARNING_LEVEL, "An internal error was detected in "
                    "parfait during processing. Please contact %s "
                    "immediately." % GATEKEEPER_MAILTO)

            err_log = os.path.join(logdir, "stderr")
            if os.path.exists(err_log):
                with open(err_log, "r") as err:
                    for line in err:
                        self.logger.log(TINFO_LEVEL, line.rstrip())

            analysis_log = os.path.join(logdir, "analysis")
            zarro_boogs = "0 defects introduced since the baseline."

            if not os.path.exists(analysis_log):
                return exit_code

            defect_summary = None
            defect_details = []
            reported_files = set()

            # Gather a list of all the files mentioned in this analysis, so
            # developers can easily see whether they need to read the details.
            with open(analysis_log, "r") as analysis:
                for line in analysis:
                    words = line.split()
                    for word in words:
                        if word.startswith("usr/src") or \
                                word.startswith("usr/closed") or \
                                word.startswith("usr/fish") or \
                                word.startswith(
                                    "build.%s/usr" % self.mach) or \
                                word.startswith(
                                    "build.%s-nd/usr" % self.mach):
                            reported_files.add(word)

                    if "defects introduced since the baseline" in line:
                        defect_summary = line.strip()

            with open(analysis_log, "r") as analysis:
                for line in analysis:
                    if "defects not shown" not in line:
                        defect_details.append(line.rstrip())
                    if self.bldenv.parfait_type == "baseline" and \
                            line.startswith(zarro_boogs):
                        modified_exit_code = 0

            if reported_files:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Parfait analysis mentioned the following files:")
                self.logger.log(TWARNING_LEVEL, "")
                for reported_file in sorted(reported_files):
                    self.logger.log(TWARNING_LEVEL, reported_file)
                self.logger.log(TWARNING_LEVEL, "")

            if defect_summary and modified_exit_code != 0:
                self.logger.log(TWARNING_LEVEL, defect_summary)
                self.logger.log(TWARNING_LEVEL, "")

            for line in defect_details:
                self.logger.log(TWARNING_LEVEL, line)

            self.logger.log(TINFO_LEVEL, "")
            self.logger.log(TINFO_LEVEL, "Parfait execution details:")
            self.logger.log(CMDLINE_LEVEL, cmd)
            self.logger.log(
                TINFO_LEVEL, "Parfait stdout saved to: %s" % out_path)
            self.logger.log(
                TINFO_LEVEL, "Parfait stderr saved to: %s" % err_path)

            return modified_exit_code

        if self.bldenv.debug:
            parfait_logdir = "%s/parfait_analysis_%s" % (
                self.env["LOGDIR"], self.mach)
            ret, cmd, out_path, err_path = do_analyse(
                debug=True, logdir=parfait_logdir)
            ret = record_parfait_results(
                ret, parfait_logdir, cmd, out_path, err_path)

            if ret > 0:
                self.exit_status = ret

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug analysis finished, starting non-debug "
                "analysis")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TINFO_LEVEL)
            self.take_snapshot(snap_suffix="%s-nd" % self.name)

        if self.bldenv.nondebug:
            parfait_logdir = "%s/parfait_analysis_%s%s" % (
                self.env["LOGDIR"], self.mach, "-nd")
            ret, cmd, out_path, err_path = do_analyse(logdir=parfait_logdir)
            ret = record_parfait_results(
                ret, parfait_logdir, cmd, out_path, err_path)

            if ret > 0:
                self.exit_status = ret


class FishCheckTask(Task):
    """Run various checks using the 'check' target in usr/fish.
    This has very low granularity at present and we might be able to break it
    down to a series of subtasks."""

    task_name = "check.fish"
    task_synopsis = "Do checks across the fish subrepo"
    snapshot = False
    produce_log_extract = True

    def _execute(self):
        """Do a make check. We only do this once, preferring the debug
        environment."""

        self.logger.log(TINFO_LEVEL, "Doing fish checks")

        ret = 0
        if self.bldenv.debug:
            env = self.bldenv.get_build_environment(
                debug=True, nondebug=False, fish=True)
            ret = util.cmdline_run(
                "dmake check", env_arg=env, logger=self.logger, cwd=env["SRC"])
            self.logger.log(TINFO_LEVEL, "Running protocmp checks")
            ret += util.cmdline_run(
                "dmake protocmp", env_arg=env, logger=self.logger,
                cwd=os.path.join(env["SRC"], "pkgdefs"))
            if ret > 0:
                self.exit_status = ret

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug checks and protocmp finished, "
                "starting non-debug protocmp.")
            log.switch_logs(
                self.logger, logsuffix="-nd", old_logsuffix="",
                level=TINFO_LEVEL)

        if self.bldenv.nondebug:
            env = self.bldenv.get_build_environment(
                debug=False, nondebug=True, fish=True)
            ret += util.cmdline_run(
                "dmake protocmp", env_arg=env, logger=self.logger,
                cwd=os.path.join(env["SRC"], "pkgdefs"))
            if ret > 0:
                self.exit_status = ret

        self.exit_status = ret


class RecordTask(Task):
    """Template Task for recording"""
    task_name = "record.template"
    task_synopsis = "Record Template"
    require_workspace_lock = False
    must_pass = False
    snapshot = False
    enabled_by_default = True
    produce_log_extract = False
    check_tools = False
    run_on_subrepos = False
    log_cmds = []

    def _execute(self):

        if self.run_on_subrepos:
            repositories = [
                (self.ws, True),
                ("%s/usr/closed" % self.ws,
                    self.env["CLOSED_IS_PRESENT"] == "yes"),
                ("%s/usr/man" % self.ws,
                    self.env["MAN_IS_PRESENT"] == "yes"),
                ("%s/usr/fish" % self.ws,
                    self.env["FISH_IS_PRESENT"] == "yes"),
            ]
        else:
            repositories = [(self.ws, True)]

        ret = 0
        for repo, present in repositories:
            if not present:
                continue
            for cmd in self.log_cmds:
                if self.run_on_subrepos:
                    run_cmd = cmd % repo
                else:
                    run_cmd = cmd

                ret += util.cmdline_run(run_cmd,
                                        env_arg=self.env,
                                        logger=self.logger,
                                        cwd=self.src,
                                        cmdout_level=TINFO_LEVEL)
                self.logger.log(TINFO_LEVEL, "\n")

        self.exit_status = ret


class RecordCompositeTask(CompositeTask):
    """Task for recording various things that may impact the build."""
    task_name = "record"
    task_synopsis = "Record information about SCM and Host"
    check_tools = False

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        super(RecordCompositeTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)

        self.add_task(RecordCBETask(bldenv))
        self.add_task(RecordMercurialStateTask(bldenv))
        self.add_task(RecordBldrcTask(bldenv))


class RecordCBETask(RecordTask):
    """This task records the packaging state of the machine we are running
    on.  This is effectively the CBE (Common Build Environment).
    We need to capture sufficient information from IPS to be able to
    re-install an equivalent machine."""

    task_name = "record.cbe"
    task_synopsis = "Record the IPS configuration of the machine"
    #
    # What are we running and where did we get it from. Record the
    # most useful information first and the most verbose last.
    log_cmds = ["/usr/bin/pkg info entire",
                "/usr/bin/pkg publisher",
                "/usr/bin/pkg property",
                "/usr/bin/pkg facet",
                "/usr/bin/pkg variant",
                "/usr/bin/pkg list -v"]


class RecordMercurialStateTask(RecordTask):
    """This task records the state of the Mercurial workspace.
    The intent to determine at what point in the past the changes were forked
    from the upstream gate, or when the workspace was last resynced to the
    gate and rebased. The only reliable way to do that is to check with the
    gate but that is a) expensive and b) not always possible.  So just
    record some key information about the current state."""

    task_name = "record.hg-state"
    task_synopsis = "Record the state of the Source Code (Mercurial) repo"
    run_on_subrepos = True

    # Hardcoding 100 as the number of changesets, 10 is too small
    # for some usecases, so rather than guess what is big enough lets
    # just go for 100.  This doesn't need to be configurable in bldrc
    log_cmds = ["hg -R %s branch",
                "hg -R %s status",
                "hg -R %s paths default",
                "hg -R %s log -b . -r 'reverse(last(0::., 100))'"]


class RecordBldrcTask(RecordTask):
    """This task records the bldrc that was passed in."""
    task_name = "record.bldrc"
    task_synopsis = "Record the bldrc content"

    def _execute(self):
        ret = 0
        if self.bldenv.bldrc:
            ret = util.cmdline_run("cat %s" % (self.bldenv.bldrc),
                                   env_arg=self.env, logger=self.logger,
                                   cwd=self.src, cmdout_level=TINFO_LEVEL)
        else:
            self.logger.log(SUMMARY_LEVEL, "No bldrc was found recording "
                            "'build generate > bldrc'")
            self.logger.log(TINFO_LEVEL,
                            self.bldenv.generate_bldrc(self, cwd=self.src,
                                                       cmd=[self.task_name]))
        self.exit_status = ret


class CertificateExpiryCheckTask(Task):
    """This task informs us when the code signing certificates used
    by Verfied Boot have expired or are about to do so.
    """

    task_name = "check.certificate-expiry"
    task_synopsis = "Do checks on the code signing certificates"
    snapshot = False
    enabled_by_default = False
    check_tools = False

    def _execute(self):
        known_expired_certs = ["ORCLS10SE", "ORCLS11SE",
                               "ORCLCodeSign_2014.pem",
                               "ORCLCodeSign_2017.pem",
                               "ORCLCodeSign_2018.pem"]
        vbcerts = "etc/certs/elfsign"
        env = self.bldenv.get_build_environment()
        days = int(env['CERT_EXPIRY_WARNING_DAYS'])
        self.logger.log(TINFO_LEVEL, "Certificate expiry warning threshold "
                        "%d days" % days)

        expiry_seconds = days * 24 * 60 * 60
        expired_count = 0
        valid_count = 0
        cert_count = 0
        for cert in os.listdir(os.path.join(self.env["ROOT"], vbcerts)):
            if cert in known_expired_certs:
                self.logger.log(TINFO_LEVEL,
                                "Known expired certificate '%s'." % cert)
                continue

            cpath = os.path.join(self.env["ROOT"], vbcerts, cert)
            cmd = ("openssl x509 -enddate -noout "
                   "-checkend %(expiry_seconds)d -in %(cert)s" % {
                       "expiry_seconds": expiry_seconds, "cert": cpath})
            ret, out, err = util.cmdline_run(cmd)
            if ret > 0:
                self.logger.log(TINFO_LEVEL,
                                "Certificate '%s' has expired"
                                " or will expire within %s days: %s." %
                                (cert, days, out))
                expired_count += 1
            else:
                self.logger.log(TINFO_LEVEL,
                                "Certificate '%s' still valid." % cert)
                valid_count += 1
            cert_count += 1

        if expired_count == cert_count:
            """ All certificates have or are about to expire """
            self.logger.log(TWARNING_LEVEL,
                            "All (%d) certificates have "
                            "expired, or will expire within %s days." %
                            (expired_count, days))
        elif expired_count > 0:
            self.logger.log(TWARNING_LEVEL,
                            "%d certificate(s) expired, or will expire "
                            "within %s days." % (expired_count, days))

        self.logger.log(TINFO_LEVEL,
                        "%s of %s certificate(s) still valid" %
                        (valid_count, cert_count))
        self.exit_status = expired_count


class ParfaitRemindDb(Task):
    """Generate a flat file containing all known Parfait issues which can
    be used by pbchk to warn gatelings about problems in their code."""

    task_name = "parfait_remind_db"
    task_synopsis = "Generate a database needed by the parfait_remind pbchk"
    snapshot = False

    def _execute(self):

        parfait_host = urllib.parse.urlparse(self.env[
            "PARFAIT_SERVER_URL"]).netloc

        # Determine the baselineReportId, needed to lookup the parfait issues
        # for each build flavor.
        baseline_id_uri_template = "https://%(parfait_host)s/" + \
            "parfait-server/projects/1/reports?" + \
            "filter=task=%%22on-%(branch)s-%(flavor)s%%22&" + \
            "orderBy=executionTimestamp&" + \
            "length=1&direction=descending"

        # A URI to retrieve all parfait issues for a given
        # baseline id.
        baseline_uri_template = "https://%(parfait_host)s/" + \
            "parfait-server/projects/1/reports/" + \
            "%(baseline_id)s/differential/%(baseline_id)s/defects" + \
            "?filter=flag=%%22verified%%22,%%22unverified%%22" + \
            "&startIndex=%(start_index)s&length=%(length)s"

        if not self.env.get("PARFAIT_SERVER_PASSWORD"):
            self.logger.log(
                TWARNING_LEVEL,
                "A bldrc setting is required for 'parfait_server_password'")
            self.exit_status = 1
            return

        self.logger.log(TINFO_LEVEL, "Finding Parfait baseline ids...")

        try:
            # it doesn't matter which flavor we use to create the pw_mgr
            pw_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
            pw_mgr.add_password(
                None, baseline_id_uri_template % {
                    "flavor": self.mach, "branch": self.branch,
                    "parfait_host": parfait_host},
                "solaris", self.env["PARFAIT_SERVER_PASSWORD"])

            auth_handler = urllib.request.HTTPBasicAuthHandler(pw_mgr)
            opener = urllib.request.build_opener(auth_handler)
            # pylint bug: reports too-many-function-args when using the
            # import six wrapper.
            # pylint: disable=too-many-function-args
            urllib.request.install_opener(opener)
            # pylint: enable=too-many-function-args

            # Obtain the baseline report ids for each build flavor.
            baseline_ids = {}
            self.logger.log(TINFO_LEVEL, "FLAVOR     BASELINE ID")
            for flavor in ["sparc-nd", "sparc", "i386-nd", "i386"]:

                baseline_id_uri = baseline_id_uri_template % {
                    "flavor": flavor,
                    "branch": self.branch,
                    "parfait_host": parfait_host}

                url_obj = urllib.request.urlopen(baseline_id_uri)
                report = simplejson.loads(url_obj.read())
                if report:
                    baseline_ids[flavor] = report[0]["id"]
                    self.logger.log(
                        TINFO_LEVEL, "%-10s %s" %
                        (flavor, baseline_ids[flavor]))
                else:
                    baseline_ids[flavor] = "none"
                    self.logger.log(
                        TINFO_LEVEL, "%-10s %s" %
                        (flavor, baseline_ids[flavor]))

                # Due to a Python bug, we need to reset the auth_handler
                # http://bugs.python.org/issue9639
                auth_handler.retried = 0
        except (urllib.error.URLError, urllib.error.HTTPError) as e:
            self.logger.log(
                TWARNING_LEVEL, "Unable to lookup parfait baseline ids: %s" %
                e)
            self.exit_status = 1
            return

        self.logger.log(TINFO_LEVEL, "Downloading Parfait results...")

        # A dictionary mapping build flavor to a list of tuples describing
        # the open issues for that build flavor.
        open_issues = {}

        self.logger.log(TINFO_LEVEL, "FLAVOR     PROGRESS")

        for flavor, baseline_id in list(baseline_ids.items()):
            if baseline_id == "none":
                self.logger.log(
                    TINFO_LEVEL, "%-10s Baseline does not exist. Skipping"
                    % flavor)
                continue
            desired_record_length = 50000
            total_records_retrieved = 0
            total_records = None
            open_issues[flavor] = []

            # Page through Parfait results (desired_record_length
            # records at a time, the server has default limit
            # to 10000 but configured to 50000 - new in 3.0.2)
            while not total_records or \
                    total_records_retrieved != total_records:
                try:
                    report_uri = baseline_uri_template % {
                        "parfait_host": parfait_host,
                        "baseline_id": baseline_id,
                        "start_index": total_records_retrieved,
                        "length": desired_record_length}

                    url_obj = urllib.request.urlopen(report_uri)
                    json_obj = simplejson.loads(url_obj.read())
                    # Due to a Python bug, we need to reset the auth_handler
                    # http://bugs.python.org/issue9639
                    auth_handler.retried = 0
                    total_records_retrieved += len(json_obj)

                    if not total_records:
                        total_records = int(
                            url_obj.headers['matched-count'])

                    for item in json_obj:
                        open_issues[flavor].append(
                            (item["flag"],
                                item["location"],
                                item["hash"]))

                    self.logger.log(
                        TINFO_LEVEL, "%-10s %s of %s" %
                        (flavor, total_records_retrieved, total_records))
                except (ValueError, urllib.error.URLError,
                        urllib.error.HTTPError) as e:
                    self.logger.log(TWARNING_LEVEL, e)
                    self.exit_status = 1
                    return

        # Combine the totals for each build flavor.
        # file_totals is a dictionary mapping a source file path to a
        # dictionary containing the number of verified and unverified issues
        # for that path. Parfait uniquely identifies issues with a hash, so we
        # include that to avoid duplicates that might occur across build
        # flavors.
        file_totals = {}
        for flavor in open_issues:
            for issue_type, location, phash in \
                    open_issues[flavor]:

                if location not in file_totals:
                    file_totals[location] = {
                        "verified": 0,
                        "unverified": 0,
                        "hashes": set()}
                    file_totals[location][issue_type] = 1
                    file_totals[location]["hashes"].add(phash)
                else:
                    if phash not in file_totals[location]["hashes"]:
                        file_totals[location][issue_type] += 1
                        file_totals[location]["hashes"].add(phash)

        public_dir = "%s/public/%s" % (self.ws, self.branch)
        if not os.path.exists(public_dir):
            os.makedirs(public_dir)

        report_filename = "%s/parfait-remind-db.txt" % public_dir
        self.logger.log(TINFO_LEVEL, "Writing output to %s" % report_filename)

        with open(report_filename, "w") as report_file:
            report_file.write(
                "#\n"
                "# This is a list of files in the gate that have\n"
                "# verified or unverified issues in the Parfait database.\n"
                "# The format of the file is:\n"
                "# <path in workspace> <num verified issues> "
                "<num unverified issues>\n"
                "#\n")

            for location in sorted(file_totals.keys()):
                verified = file_totals[location]["verified"]
                unverified = file_totals[location]["unverified"]
                report_file.write(
                    "%s %s %s\n" % (location, verified, unverified))


class FortifyReminderDB(Task):
    """Generate a flat file containing all known Fortify issues which can
    be used by pbchk to warn gatelings about problems in their code."""

    task_name = "fortify_remind_db"
    task_synopsis = "Generate database file needed by the fortify_remind pbchk"
    snapshot = False

    def _execute(self):
        def _get_relbasepath(fvdl):
            with open(fvdl) as fl:
                xmlstring = fl.read()
            fl.close()
            # Remove the default namespace definition
            xmlstring = re.sub('\\sxmlns="[^"]+"', '', xmlstring, count=1)
            ET = xml.etree.ElementTree.fromstring(xmlstring)
            # find base dir
            sourcebasepath = ET.find('Build/SourceBasePath').text
            return sourcebasepath[sourcebasepath.find('/usr'):][1:]

        def _parse_local(fpr):
            """Extract static code analysis results from the local build"""
            # unzip the local fpr file to a temporary directory
            local_tempdir = tempfile.mkdtemp(prefix='fpr_unzip_',
                                             dir=self.env["WS"] +
                                             "/data/fortify")
            zip_ref = zipfile.ZipFile(fpr, 'r')
            zip_ref.extractall(local_tempdir)
            zip_ref.close()

            issues = {}
            f_issues = []

            fvdl = local_tempdir + '/audit.fvdl'
            rel_path = _get_relbasepath(fvdl)

            with open(fvdl) as fl:
                xmlstring = fl.read()
            fl.close()
            # Remove the default namespace definition
            xmlstring = re.sub('\\sxmlns="[^"]+"', '', xmlstring, count=1)
            ET = xml.etree.ElementTree.fromstring(xmlstring)
            vulns = ET.find('Vulnerabilities')
            for v in vulns.findall('Vulnerability'):
                id = v.find('InstanceInfo/InstanceID')

                for e in v.findall('AnalysisInfo/Unified/Trace/Primary/Entry'):
                    for n in e.findall('Node'):
                        if n.get('isDefault') == 'true':
                            s = n.find('SourceLocation')
                            issues[id.text] = rel_path + '/' + s.get('path')

            sorted_issues = sorted(list(issues.items()),
                                   key=operator.itemgetter(1))

            old_issue = sorted_issues[0][1]
            count = 0
            for issue in sorted_issues:
                count = count + 1
                if issue[1] != old_issue:
                    f_issues.append([old_issue, 0, count])
                    old_issue = issue[1]
                    count = 0

            shutil.rmtree(local_tempdir)
            return(issues, f_issues)

        def _subtract_server(fpr, issues, f_issues):
            """Extract static code analysis results from the server"""
            # unzip the server fpr file to a temporary directory
            server_tempdir = tempfile.mkdtemp(prefix='server_fpr_unzip_',
                                              dir=self.env["WS"] +
                                              "/data/fortify")
            zip_ref = zipfile.ZipFile(server_file, 'r')
            zip_ref.extractall(server_tempdir)
            zip_ref.close()

            rel_basepath = _get_relbasepath(server_tempdir + '/audit.fvdl')

            # adjust number of issues with already processed ones
            ET = xml.etree.ElementTree.parse(server_tempdir + '/audit.xml')
            root = ET.getroot()
            ns = {'audit': 'xmlns://www.fortify.com/schema/audit'}
            il = root.find('audit:IssueList', ns)
            for i in il.findall('Issue'):
                t = i.find('Tag')
                v = t.find('Value')
                id = i.get('instanceId')
                path = issues[id]
                index = [x[0] for x in f_issues].index(path)
                if v.text == "Not an Issue":
                    print("Deleting issue {} in {}".format(id, path))
                    f_issues[index] = (
                        rel_basepath + '/' + f_issues[index][0],
                        f_issues[index][1],
                        f_issues[index][2] - 1)
                else:
                    f_issues[index] = (
                        rel_basepath + '/' + f_issues[index][0],
                        f_issues[index][1] + 1,
                        f_issues[index][2] - 1)

            shutil.rmtree(server_tempdir)

        local_file = self.env["WS"] + "/data/fortify/solaris.fpr"
        server_file = self.env["WS"] + "/data/fortify/server_fpr.zip"

        r_token_file = open(self.env["FORTIFY_READ_TOKEN"], "r")
        r_token = r_token_file.readline().strip()
        r_token_file.close()

        # Download an fpr file with analysis results and triage info
        cmd = "{0}/bin/fortifyclient -url {1} " \
            "-user onreaduser --authtoken {2} " \
            "downloadFPR -file {3} -project {4} -version {5} " \
            .format(self.env["FORTIFY_HOME"],
                    self.env["FORTIFY_SERVER"],
                    r_token,
                    server_file,
                    self.env["FORTIFY_PROJECT"],
                    self.env["FORTIFY_PROJECT_VERSION"])

        self.logger.log(TINFO_LEVEL, "Executing {}".format(cmd))
        ret, result, err = util.cmdline_run(cmd, env_arg=self.env)

        if ret != 0:
            self.exit_status = 1
            self.logger.log(TWARNING_LEVEL, "fortifyclient run failed!")
            return

        issues, f_issues = _parse_local(local_file)
        _subtract_server(server_file, issues, f_issues)

        public_dir = "{0}/public/{1}".format(self.ws, self.branch)
        if not os.path.exists(public_dir):
            os.makedirs(public_dir)

        report_filename = "{0}/fortify-remind-db.txt".format(public_dir)
        self.logger.log(TINFO_LEVEL,
                        "Writing output to {0}".format(report_filename))

        with open(report_filename, "w") as report_file:
            report_file.write(
                "#\n"
                "# This is a list of files in the gate that have\n"
                "# verified or unverified issues in the Fortify database.\n"
                "# The format of the file is:\n"
                "# <path in workspace> <num verified issues> "
                "<num unverified issues>\n"
                "#\n")

            for location, verified, unverified in f_issues:
                report_file.write(
                    "{0} {1} {2}\n".format(location, verified, unverified))


class GenerateTPLTask(Task):
    """As part of a Tonic build, we generate a THIRDPARTYLICENSE.ON-BINARIES
    file after we build."""

    task_name = "generate_tpl"
    task_synopsis = "Generate THIRDPARTYLICENSE files"
    snapshot = False

    def _execute(self):
        if "TONICBUILD" not in self.env:
            self.logger.log(
                TINFO_LEVEL, "No need for THIRDPARTYLICENSE file: "
                "'tonicbuild' not set in bldrc.")
            return

        licenses_dir = "%s/licenses" % self.env["ROOT"]
        if os.path.exists(licenses_dir):
            if self.env.get("BIP"):
                license_list_loc = "%s" % self.src
            else:
                license_list_loc = \
                    "%s/usr/src/pkg/license-list" % self.env["BUILD"]
            cmd = "mktpl %s" % license_list_loc
            ret = util.cmdline_run(
                cmd, env_arg=self.env, logger=self.logger, cwd=licenses_dir)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Couldn't create THIRDPARTYLICENSE file")
                self.exit_status = 1
        else:
            self.logger.log(
                TWARNING_LEVEL, "No licenses found under %s" % licenses_dir)


class ClosedBinsTask(Task):
    """Creates the closed bins tarballs, which can be used by the
    'on_closed_bins' bldrc key to allow users who don't have access to
    usr/closed to build the source.

    This step grovels through the package manifests, so it needs to be run
    after findunref.

    We assume no DEBUG vs non-DEBUG package content variation here; if that
    changes, then the "make all" in $SRC/pkg will need to be moved into the
    conditionals and repeated for each different build.

    This task doesn't use separate debug/nondebug logs because the amount of
    log data generated is small.
    """

    task_name = "closed_tarball"
    task_synopsis = "Generates tarballs containing closed binaries"

    def _execute(self):

        if "TONICBUILD" not in self.env:
            self.logger.log(
                TWARNING_LEVEL, "Unable to generate closed bins without "
                "'tonic' set in bldrc file.")
            self.exit_status = 1
            return

        def _do_closed_bins(debug=False):
            closed_bins_basename = "on-closed-bins"
            suffix = "-nd"
            label = "non-DEBUG"
            n_flag = "-n"
            if debug:
                suffix = ""
                label = "DEBUG"
                n_flag = ""
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            cmd = "bindrop %s %s%s" % (n_flag, closed_bins_basename, suffix)
            ret = util.cmdline_run(
                cmd, env_arg=env, logger=self.logger, cwd=self.ws)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL, "Couldn't create %s closed bins - "
                    "bindrop(1) exited with a failure status." % label)
                self.exit_status = 1

        if self.bldenv.debug:
            _do_closed_bins(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                TINFO_LEVEL, "Debug closed bins created finished, starting "
                "non-debug closed bins")

        if self.bldenv.nondebug:
            _do_closed_bins()

        env = self.bldenv.get_build_environment()
        self.logger.log(TINFO_LEVEL, "Generating README.opensolaris")
        try:
            shutil.copy(
                "%s/tools/opensolaris/README.opensolaris" %
                self.src, env["ONBLD_DATA"])
        except Exception as e:
            self.logger.log(
                TWARNING_LEVEL, "Couldn't copy README.opensolaris from "
                "usr/src/tools/opensolaris.")
            self.exit_status = 1
            return


class UpdateParentTask(Task):
    """Copy the proto list, the proto area, the stubroot proto area and the
    unreferenced files from this workspace into the $PARENT_ROOT(s) and
    $PARENT_WS directories."""
    task_name = "update_parent"
    task_synopsis = \
        "Update a parent ws with data/proto from this workspace"
    snapshot = False

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, do_filelists=True):
        super(UpdateParentTask, self).__init__(bldenv, args=args, level=level)
        self.do_filelists = do_filelists

    def _execute(self):
        """Uses rsync to update a parent workspace."""

        # A flag we update when we attempt to update the parent.
        tried_update = False

        def lock_parent(dir):
            """Attempt to acquire a lock on the directory pointed to by 'dir'
            returning True if we have been able to acquire the lock, False
            otherwise.  We make 10 attempts to acquire the lock, waiting 20
            seconds between each attempt."""

            interval = 20
            max_attempts = 10

            locked_parent = False
            lock_attempts = 0
            while not locked_parent and lock_attempts < max_attempts:
                try:
                    self._acquire_lock(dir)
                    locked_parent = True
                except (TaskLockedException, OSError) as e:
                    self.logger.log(TINFO_LEVEL, e)
                    self.logger.log(
                        TINFO_LEVEL, "Will try again in %s seconds" % interval)
                    time.sleep(interval)
                    lock_attempts += 1
            return locked_parent

        def update_dirtree(src, dest, env):
            """Invoke rsync to copy the source files to the destination with
            the given environment."""
            user = env.get("RSYNC_IDENT_USER")
            ident_key = env.get("RSYNC_IDENT_KEY")

            if not user:
                user = env.get("USER")
            args = "/usr/bin/ssh -l %s" % user
            if ident_key:
                args += " -i %s" % ident_key
            cmd = (
                "/usr/bin/rsync -e \"%(args)s\" -aHz --delete "
                "%(src)s %(dest)s" %
                {"args": args, "src": src, "dest": dest})
            return util.cmdline_run(
                cmd, logger=self.logger, env_arg=env, shell=True)

        def do_update(parent_proto, debug=False):
            if debug:
                label = "DEBUG"
            else:
                label = "non-DEBUG"
            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)

            update_dirs = [
                ("proto area", env["ROOT"]),
                ("stub proto", env["STUBROOT"])]

            if env["FISH_IS_PRESENT"] == "yes":
                update_dirs.append(
                    ("fish proto", env["FISHROOT"]))

            for label, src_dir in update_dirs:
                self.logger.log(
                    TINFO_LEVEL, "copying %s %s to %s" %
                    (label, src_dir, parent_proto))

                ret = update_dirtree(src_dir, parent_proto, env)
                if ret != 0:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to copy %s %s to %s" %
                        (label, src_dir, parent_proto))
                    self.exit_status = 1

        def update_unref_files():
            """Copy local unref files to the parent, and produce a master list
            of unreferenced files. Ideally, we'd generate the master just once
            after all of the nightlies that are contributing their proto areas
            to the parent have finished, but there's no simple way to know when
            that will be. Instead, we assume that we're the last nightly to
            finish and merge all of the ${BRANCH_PREFIX}unref-${MACH}.out
            files in $PARENT_WS/keep.

            If we are in fact the final ${MACH} to finish, then this file
            will be the authoritative master list. Otherwise, another
            ${MACH}'s nightly will eventually overwrite ours with its own
            master, but in the meantime our temporary "master" will be no
            worse than any older master which was already on the parent."""

            passed_unref_update = True
            parent_ws = self.env["PARENT_WS"]
            branch_prefix = self.env["ONBLD_BRANCH_PREFIX"]
            # Copy our unref file into the parent.
            try:
                unref_path = "%s/%sunref-%s.out" % \
                    (self.keep, branch_prefix, self.mach)
                shutil.copy(unref_path, "%s/keep" % parent_ws)
            except IOError as e:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to copy %s to %s: %s" %
                    (unref_path, parent_ws, e))
                self.logger.log(
                    TWARNING_LEVEL, "Did check.findunref complete?")
                passed_unref_update = False

            # Merge any existing unref files in the parent.
            parent_unref_files = glob.glob(
                "%s/keep/%sunref-*.out" % (parent_ws, branch_prefix))
            if parent_unref_files:
                merge_path = os.path.join(
                    self.env["TMPDIR"],
                    "%sunref.merge" % branch_prefix)
                try:
                    shutil.copy(parent_unref_files[0], merge_path)
                except IOError as e:
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Unable to copy unref files for merging: %s" % e)
                    passed_unref_update = False
                # having copied the first match, we can merge remaining files
                parent_unref_files.pop(0)
                for unreffile in parent_unref_files:
                    unref_temp = tempfile.mkstemp(".unref")[1]
                    cmd = "comm -12 %s %s > %s" % \
                        (merge_path, unreffile, unref_temp)
                    ret = util.cmdline_run(
                        cmd, logger=self.logger, env_arg=self.env, shell=True)
                    if ret != 0:
                        self.logger.log(
                            TWARNING_LEVEL, "Unable to merge unref file!")
                        passed_unref_update = False
                    os.rename(unref_temp, merge_path)
                try:
                    shutil.copy(
                        merge_path, "%s/keep/%sunrefmaster.out" %
                        (parent_ws, branch_prefix))
                except IOError as e:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to copy unrefmaster: %s" % e)
                    passed_unref_update = False
            return passed_unref_update

        parent_proto = os.path.dirname(self.env["PARENT_ROOT"])
        locked_parent = lock_parent(parent_proto)
        if not locked_parent:
            self.logger.log(TWARNING_LEVEL, "Unable to lock parent_root")
            self.exit_status = 1
            return

        # We only need to copy the native and tool proto area once, since we
        # don't build them separately for debug vs. non-debug.
        for proto_src in ["NATIVEROOT", "TOOLROOT"]:
            if os.path.exists(self.env[proto_src]):
                tried_update = True
                self.logger.log(
                    TINFO_LEVEL, "copying $%s to %s" %
                    (proto_src, parent_proto))

                self.exit_status += update_dirtree(
                    self.env[proto_src], parent_proto, self.env)

        if self.bldenv.debug:
            do_update(parent_proto, debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            # We won't, write a separate log file for non-debug proto updating,
            # so no need to call switch_logs(..)
            self.logger.log(
                TINFO_LEVEL, "Debug parent root update finished, starting "
                "non-debug update")
        if self.bldenv.nondebug:
            do_update(parent_proto)

        self._release_lock(parent_proto)

        # Now update files produced by some of our check.* tasks
        branch_prefix = self.env["ONBLD_BRANCH_PREFIX"]

        update_parent_paths = []
        if self.bldenv.overrides.get("check.protocmp", True) and \
                self.bldenv.check and self.do_filelists:
            proto_list_path = "%s/%sproto_list_%s" % (
                self.env["LOGDIR"],
                self.env["ONBLD_BRANCH_PREFIX"],
                self.mach)
            proto_list_parent = os.path.join(
                self.env["PARENT_WS"],
                "public/%sproto_list_%s" %
                (self.env["ONBLD_BRANCH_PREFIX"], self.mach))
            update_parent_paths.append((proto_list_path, proto_list_parent))

        if self.bldenv.overrides.get("check.findunref", True) and \
                self.bldenv.check and self.do_filelists:
            unref_path = "%s/%sunref-%s.out" % \
                (self.keep, branch_prefix, self.mach)
            unref_path_parent = os.path.join(
                self.env["PARENT_WS"], "keep/%sunref-%s.out" %
                (branch_prefix, self.mach))
            update_parent_paths.append((unref_path, unref_path_parent))

        for (src_path, dst_path) in update_parent_paths:
            parent_dir = os.path.dirname(dst_path)
            try:
                os.makedirs(parent_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Unable to mkdir %s: %s" % (parent_dir, e))
                    self.exit_status = 1
                    continue
            self.logger.log(
                TINFO_LEVEL, "copying %s list to %s" % (src_path, dst_path))
            if not os.path.exists(src_path):
                self.logger.log(
                    TWARNING_LEVEL,
                    "%s does not exist: not copying to parent." %
                    src_path)
                continue
            try:
                shutil.copy(src_path, dst_path)
            except IOError as e:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Unable to copy %s: %s" % (src_path, e))
                self.exit_status = 1

        if self.bldenv.overrides.get("check.findunref", True) and \
                self.bldenv.check and self.do_filelists:
            if not update_unref_files():
                self.logger.log(
                    TWARNING_LEVEL, "Unable to update parent unref files")
                self.exit_status = 1

        if not tried_update:
            self.logger.log(
                TWARNING_LEVEL, "No updates were attempted! Check that "
                "'parent_nativeroot' and/or 'parent_root' is specified in "
                "the bldrc file.")


class XrefTask(Task):
    """Builds xref files in the workspace using the build_cscope.conf file in
    the workspace to determine the databases to construct. We build cscope,
    ctags, and etags cross-references."""

    task_name = "xref"
    task_synopsis = "Build xref databases for the workspace"
    check_tools = False
    snapshot = False
    require_workspace_lock = False

    def _execute(self):
        # Since xref is often run on workspaces before 'setup' has produced
        # tools in $NATIVEROOT, we might want to use tools on the build
        # machine instead.
        cscope_fast = "bin/%s/cscope-fast" % self.mach
        ws_tools_cscope_path = os.path.join(
            self.env["TOOLROOT"], "opt/onbld", cscope_fast)

        if os.path.exists(ws_tools_cscope_path):
            self.logger.log(
                TINFO_LEVEL, "using cscope-fast from %s" %
                self.env["TOOLROOT"])
            env = copy.deepcopy(self.env)
            # set CSCOPE so that the definition in xref.mk gets overridden
            env["CSCOPE"] = ws_tools_cscope_path

        elif not os.path.exists(
                os.path.join(self.env["BUILD_TOOLS"], cscope_fast)):
            self.logger.log(
                TINFO_LEVEL, "cscope-fast not found in $BUILD_TOOLS - falling "
                "back to tools installed on build machine in /opt/onbld.")
            env = copy.deepcopy(self.env)
            env["PATH"] = "/opt/onbld/bin:%s" % env["PATH"]
            env["BUILD_TOOLS"] = "/opt"

        else:
            env = self.env

        cscope_conf_path = env["CSCOPE_CONF_PATH"]
        try:
            open(cscope_conf_path).close()
            self.logger.log(TINFO_LEVEL, "Using %s" % cscope_conf_path)
        except IOError as e:
            self.logger.log(
                TWARNING_LEVEL, "Unable to open cscope conf file: %s" % e)
            self.exit_status = 1
            return

        # Get the cscope filter string based on the current branch
        cmd = "/usr/bin/hg -R %s branch" % self.ws
        ret, out, err = util.cmdline_run(cmd)
        if ret != 0:
            csfilter = "default"
        else:
            csfilter = out.strip()

        with open(cscope_conf_path, "r") as cscope_conf:
            for line in cscope_conf:
                # ignore comments and blank lines
                if line.startswith("#") or line == "\n" or line.strip() == "":
                    continue
                items = line.rstrip().split()
                if len(items) < 4:
                    self.logger.log(
                        TWARNING_LEVEL, "Invalid build_cscope.conf entry '%s' "
                        ". Expecting lines of the form \n"
                        "'name	flags	 dir [ dir ... ]'" % line)
                    self.exit_status = 1
                    return
                if items[0] != "*" and items[0] != csfilter:
                    continue

                name = items[1]
                flags = items[2]
                dirs = items[3:]

                if flags == "\"\"":
                    flags = ""

                self.logger.log(
                    TINFO_LEVEL, "Building xref databases for '%s'" % name)
                ret = util.cmdline_run(
                    "xref -c %s" % (" ".join(dirs)),
                    logger=self.logger, env_arg=env, cwd=self.src)
                if ret != 0:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to clobber xref databases for "
                        "'%s'" % name)
                    self.exit_status = 1
                    break
                ret = util.cmdline_run(
                    "xref %s %s" % (flags, " ".join(dirs)),
                    logger=self.logger, env_arg=env, cwd=self.src)
                if ret != 0:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to create xref databases for "
                        "'%s'" % name)
                    self.exit_status = 1
                    break


class CscopeTask(XrefTask):
    """This class exists simply to provide a subcommand synonym for XrefTask.
    """
    task_name = "cscope"
    task_synopsis = "(a synonym for 'xref')"


class HereTask(Task):
    """Runs a 'dmake install' in the current directory. Note that this runs
    either a debug or a non-debug command, not both."""
    task_name = "here"

    task_synopsis = "Runs a 'dmake <target> ...' in the current directory"
    snapshot = False

    def _execute(self):
        cwd = os.getcwd()
        if self.bldenv.debug:
            label = "DEBUG"
        elif self.bldenv.nondebug and not self.bldenv.debug:
            label = "NONDEBUG"

        # Replace the environment with a usr/fish environment if we're in
        # a usr/fish repo
        cmd = "hg log --template '{node}' -r 0"
        ret, out, err = util.cmdline_run(cmd)
        if out.strip() == "442de104a6ad8d968bf5b49b372b0b9ff8a5e406":
            self.env = self.bldenv.get_build_environment(
                debug=self.bldenv.debug, nondebug=not self.bldenv.debug,
                fish=True)

        # the fact that <bldrc> is a positional argument, not an
        # option is a pain. Do gruesome things to get around that.
        if not self.args:
            target = "install"
        elif len(self.args) == 1:
            if os.path.isfile(self.args[0]):
                target = "install"
            else:
                target = self.args[0]
        else:
            target = " ".join(self.args[1:])

        self.logger.log(
            TINFO_LEVEL, "Running %s dmake %s in %s" % (label, target, cwd))

        # For 'build here', when running on workspaces with cold filesystem
        # caches, using 'hg identify' can take some time, relative to the
        # length of building a short Makefile target.
        hg_info = get_hg_info(self.ws, self.env, hg_identify=False)
        for repo in hg_info:
            self.logger.log(
                TINFO_LEVEL, "%s changeset = %s" % (repo, hg_info[repo][0]))

        ret = util.cmdline_run(
            "dmake %s" % target, logger=self.logger, cwd=cwd,
            env_arg=self.env)
        if ret != 0:
            self.exit_status = 1
            self.logger.log(
                TWARNING_LEVEL, "dmake %s failed in %s" % (target, cwd))


class DummyTask(Task):
    """This task does nothing and exists so that we can obtain build
    environments without needing to do work and without checking
    tools versions."""
    task_name = "dummy"
    check_tools = False


class KernelTarTask(Task):
    """A task to create a tarball of the minimum-required kernel
    bits. This is then finagled into place using Install.py."""

    task_name = "kerneltar"
    task_synopsis = "Create a tarball of the kernel from a proto area."
    produce_log_extract = False

    def _execute(self):
        bldenv = self.bldenv

        def do_kerneltar(debug=False):
            bufsize = 20 * 1024 * 1024
            nondebug = not debug
            env = bldenv.get_build_environment(debug=debug, nondebug=nondebug)
            user = self.env.get("USER")
            # We're using /tmp because the $TMPDIR gets deleted when build
            # exits, and for compatibility with the old version of Install(1).
            basedir = "/tmp/Install.%s" % (user)

            if not os.path.exists(basedir):
                try:
                    os.mkdir(basedir)
                except OSError:
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Unable to create %s" % (basedir))
                    return 1

            try:
                os.access(basedir, os.R_OK | os.W_OK | os.F_OK)
            except OSError:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Unable to access %s" % (basedir))
                return 1

            root = env.get("ROOT")
            if nondebug:
                ktarsuffix = "-nondebug"
            else:
                ktarsuffix = "-debug"

            ktarname = "%s/%s%s.tar" % (basedir, user, ktarsuffix)
            # We're NOT using any compression.
            ktar = tarfile.open(
                name=os.path.join(basedir, ktarname),
                mode="w", bufsize=bufsize)

            kerntar = "kernel.tar"
            plattar = "platform.tar"
            usrkerntar = "usrkernel.tar"
            usrplattar = "usrplatform.tar"
            mdbtar = "mdb.tar"
            tfn = []
            tfn.append(kerntar)
            tfn.append(plattar)
            tfn.append(usrkerntar)
            tfn.append(usrplattar)
            tfn.append(mdbtar)

            os.chdir("%s/kernel" % (root))
            ktf = tarfile.open(
                name=os.path.join(basedir, kerntar),
                mode="w", bufsize=bufsize)
            ktf.add(".", recursive=True)
            ktf.close()

            os.chdir("%s/platform" % (root))
            ptf = tarfile.open(
                name=os.path.join(basedir, plattar),
                mode="w", bufsize=bufsize)
            ptf.add(".", recursive=True)
            ptf.close()

            os.chdir("%s/usr/kernel" % (root))
            uktf = tarfile.open(
                name=os.path.join(basedir, usrkerntar),
                mode="w", bufsize=bufsize)
            uktf.add(".", recursive=True)
            uktf.close()

            os.chdir("%s/usr/platform" % (root))
            uptf = tarfile.open(
                name=os.path.join(basedir, usrplattar),
                mode="w", bufsize=bufsize)
            for i in os.listdir("."):
                p = os.path.join(i, "lib/mdb/kvm")
                if os.path.exists(p):
                    uptf.add(p, recursive=True)
            uptf.close()

            os.chdir("%s/usr/lib/mdb/kvm" % (root))
            mdbtf = tarfile.open(
                name=os.path.join(basedir, mdbtar),
                mode="w", bufsize=bufsize)
            mdbtf.add(".", recursive=True)
            mdbtf.close()

            os.chdir(basedir)
            for tf in tfn:
                ktar.add(tf)

            ktar.close()
            # tidy up a little
            for tf in tfn:
                os.remove(tf)
            self.logger.info(
                "Install (%s) tarballs created in %s" %
                (ktarsuffix[1:], basedir))

            return 0

        if self.bldenv.debug:
            self.exit_status += do_kerneltar(debug=True)

        if self.bldenv.nondebug:
            self.exit_status += do_kerneltar()

        self.logger.info(
            "Use the Install(1) command to install tarballs")


class PkgServeTask(Task):
    """Use pkg.depot-config to create an Apache config file to serve
    the package repositories in this workspace pointed to by PKGARCHIVE,
    then start an Apache instance.

    Note that we're tied to using the same version of Apache as
    pkg.depot-config, since that generates our httpd.conf. When that
    version for pkg.depot-config changes, this task must change also.
    """

    task_name = "pkgserve"
    task_synopsis = "Serve packages built from this workspace over HTTP"
    snapshot = False
    require_workspace_lock = False
    check_tools = False

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(PkgServeTask, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs)

    def _execute(self):
        repos = []
        fish_dirs = []

        def add_repos(debug=True, pkgmerge=False):
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug)
            if pkgmerge:
                label = ""
                pkg_label = ""
            elif debug:
                label = "debug-"
                pkg_label = "-debug"
            else:
                label = ""
                pkg_label = "-nondebug"

            if "FISH_STANDALONE" in env:
                on_pkgarchive = os.path.join(
                    env["FW_HOOKS"],
                    "data/pkg/{0}{1}".format(self.mach, pkg_label))
                repo_path = on_pkgarchive
            else:
                if pkgmerge:
                    pkgmerge_type, pkgmerge_dest, \
                        debug_pkgs_local, debug_pkgs_altws, \
                        nondebug_pkgs_local, nondebug_pkgs_altws = \
                        _compute_pkgmerge_params(self.bldenv)
                    repo_path = pkgmerge_dest
                    self.logger.info("Serving merged repos")
                else:
                    repo_path = env["PKGARCHIVE"]

            nightly = os.path.join(repo_path, "repo.osnet")
            extra = os.path.join(repo_path, "repo.osnet-internal")

            if os.path.exists(nightly):
                repos.append("%snightly=%s" % (label, nightly))
            else:
                self.logger.warning(
                    "nightly repo does not exist at %s" % nightly)
            if os.path.exists(extra):
                repos.append("%sextra=%s" % (label, extra))
            else:
                self.logger.warning("extra repo does not exist at %s" % extra)

            # see if we have fish content that we should also serve
            if env["FISH_IS_PRESENT"] == "yes":
                fishenv = self.bldenv.get_build_environment(
                    debug=debug, nondebug=not debug, fish=True)
                fish_dirs.append(
                    (fishenv["PKGARCHIVE"], "{0}fish".format(label)))

        if self.bldenv.pkgmerge:
            add_repos(pkgmerge=True)
        else:
            if self.bldenv.debug:
                add_repos(debug=True)
            if self.bldenv.nondebug:
                add_repos(debug=False)

        try:
            port = int(self.env.get("ONBLD_PKGSERVE_PORT", 8000))
        except ValueError as e:
            self.logger.error(
                "$ONBLD_PKGSERVE_PORT is set to an invalid port number %s" %
                self.env["ONBLD_PKGSERVE_PORT"])
            self.exit_status = 1
            return

        found_port = False
        while not found_port and port < 65535:
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(("", port))
                found_port = True
                s.close()
            except socket.error:
                if self.env.get("ONBLD_PKGSERVE_PORT"):
                    self.logger.error(
                        "$ONBLD_PKGSERVE_PORT is already in use")
                    self.exit_status = 1
                    return
                else:
                    port += 1

        if port >= 65535:
            self.logger.error("Unable to find free port - giving up.")
            self.exit_status = 1
            return

        runtime = os.path.join(self.env["TMPDIR"], "apache", "htdocs")
        logs = os.path.join(self.env["TMPDIR"], "apache", "logs")
        cache = os.path.join(self.env["TMPDIR"], "apache", "cache")

        for dir in runtime, logs, cache:
            try:
                os.mkdir(dir)
            except OSError:
                pass

        repostr = "-d " + " -d ".join(repos)

        depot_config_path = "{0}/usr/lib/pkg.depot-config".format(
            self.env["PKG7_ROOT"])
        depot_template_path = "{0}/etc/pkg/depot".format(
            self.env["PKG7_ROOT"])
        cmdline = "{depot_config_path} {repostr} -r {runtime} " \
            "-T {templates} -l {logs} -p {port} -c {cache}".format(
                depot_config_path=depot_config_path,
                repostr=repostr,
                port=port,
                runtime=runtime,
                templates=depot_template_path,
                logs=logs,
                cache=cache)
        # we need this to prevent pkg.depot-config trying to chown to pkg5srv
        self.env["PKG5_TEST_ENV"] = "true"
        ret = util.cmdline_run(cmdline, logger=self.logger, env_arg=self.env)
        if ret != 0:
            self.logger.error("Unable to create Apache configuration")
            if not os.path.exists(depot_config_path):
                self.logger.error(
                    "pkg:/package/pkg/depot must be installed "
                    "to use 'pkgserve'.")

            self.exit_status = 1
            return

        # Create a custom apache conf fragment to serve fish upgrade pkgs
        for path, alias in fish_dirs:
            fragment = """
<Directory "{0}">
    AllowOverride None
    Options None
    Require all granted
    Options Indexes FollowSymLinks
</Directory>
Alias /{1} {0}
""".format(path, alias)
            with open("{0}/depot_httpd.conf".format(runtime), "a") as conf:
                conf.write(fragment)

        httpd_path = "/usr/apache2/2.4/bin/httpd"
        if not os.path.exists(httpd_path):
            self.logger.error(
                "Apache is not installed at {httpd}".format(
                    httpd=httpd_path))
            self.exit_status = 1
            return
        try:
            cmdline = (
                "{httpd} -f {runtime}/depot_httpd.conf "
                "-k start").format(httpd=httpd_path, runtime=runtime)
            ret = util.cmdline_run(
                cmdline, logger=self.logger, env_arg=self.env)
            if ret != 0:
                self.logger.error("Unable to start Apache.")
                error_path = "{runtime}/logs/error_log".format(runtime=runtime)
                if os.path.exists(error_path):
                    with open(error_path) as error_log:
                        for line in error_log.readlines():
                            self.logger.error(line)
                self.exit_status = 1
                return
            self.logger.info("Apache started on:")

            for host in util.get_hostnames():
                self.logger.info("    http://%s:%s" % (host, port))

            if fish_dirs:
                for desc, ext in [
                    ("AK upgrade packages:", "pkg"),
                    ("AK iso images:", "iso")
                ]:
                    self.logger.info("")
                    self.logger.info(desc)
                    for host in util.get_hostnames():
                        for path, alias in fish_dirs:
                            if not os.path.exists(path):
                                continue
                            entries = os.listdir(path)
                            for entry in entries:
                                if not entry.endswith(".%s" % ext):
                                    continue
                                self.logger.info(
                                    "    http://%s:%s/%s/%s" %
                                    (host, port, alias, entry))

            self.logger.info("Use ^C to exit")
            while True:
                sys.stdin.read()
        except KeyboardInterrupt:
            cmdline = (
                "{httpd} -f {runtime}/depot_httpd.conf "
                "-k stop").format(httpd=httpd_path, runtime=runtime)
            ret = util.cmdline_run(
                cmdline, logger=self.logger, env_arg=self.env)
            if ret != 0:
                self.logger.error("Failed to stop apache!")
                self.exit_status = 1
                return


class SavePackagesTask(Task):
    """If PKGARCHIVE and PKGSURFARCHIVE are stored within the confines of
    this workspace, this task attempts to save the most recent set of
    packages built in this workspace to $ONBLD_KEEP, so that we can
    use those as our PKGSURFARCHIVE repositories. OTherwise, the
    'clobber' task will delete the packages from $ONBLD_DATA before
    we get a chance to use them.

    If PKGSURFARCHIVE directory is outside this workspace (not the default),
    or if it does NOT reside within $ONBLD_KEEP/pkgsurf-ref.pkg,
    we assume that the user has specified this, and we do not reuse old
    PKGARCHIVE contents for the pkgsurf reference repository.

    Any previous PKGSURFARCHIVE contents in $ONBLD_KEEP are removed.

    If 'pkgmerge' is set, we rescue the merged packages directory instead.

    This task never fails - if we can't save previous packages, then pkgsurf
    will attempt to resurface against whatever packages remain in
    PKGSURFARCHIVE, if any. If no packages exist, pkgsurf does nothing.
    """

    task_name = "save_packages"
    task_synopsis = "Move packages to $PKGSURFARCHIVE as a pkgsurf reference"
    snapshot = False
    require_workspace_lock = False
    check_tools = False

    def _execute(self):

        def do_save_pkgs(debug=False, pkgmerge=False):
            """Attempt to save packages, return a boolean indicating
            whether we were successful."""

            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug)

            psurfref = os.path.join(
                self.keep, "{0}pkgsurf-ref.pkg".format(
                    env["ONBLD_BRANCH_PREFIX"]))
            pkgsurfarchive = env.get("PKGSURFARCHIVE", "")
            if not pkgsurfarchive.startswith(psurfref):
                self.logger.log(
                    TINFO_LEVEL,
                    "pkgsurfarchive is not within the default pkgsurf "
                    "reference directory: not salvaging old pkgarchive "
                    "repositories.")
                return False

            if pkgmerge:
                # we want to save the merged repositories if we have
                # been merging in the past.  Saving the separate
                # debug/non-debug package repositories as well would
                # be a waste of IO.
                pkgmerge_type, pkgmerge_dest, \
                    debug_pkgs_local, debug_pkgs_altws, \
                    nondebug_pkgs_local, nondebug_pkgs_altws = \
                    _compute_pkgmerge_params(self.bldenv)
                pkgarchive = pkgmerge_dest
                pkgsurfarchive = os.path.normpath(os.path.join(
                    pkgsurfarchive, "..", os.path.basename(pkgmerge_dest)))
            else:
                pkgarchive = env.get("PKGARCHIVE")

            if not os.path.exists(pkgarchive):
                self.logger.log(
                    TINFO_LEVEL,
                    "No existing packages found at {0}".format(pkgarchive))
                return False

            # If a 'packages' task failed from the previous build, we can
            # have an empty repository. We won't clobber the current
            # pkgsurfarchive contents with those empty repositories.
            repo_cmd = "pkgrepo -s {0}/repo.osnet list -HF tsv".format(
                pkgarchive)
            ret, out, err = util.cmdline_run(repo_cmd, env_arg=env)
            if out == "":
                self.logger.log(
                    TINFO_LEVEL,
                    "Repository at {0} does not contain any packages: not "
                    "saving old pkgarchive repositories.".format(pkgarchive))
                return False

            if os.path.exists(pkgsurfarchive):
                self.logger.log(
                    TINFO_LEVEL,
                    "Removing old saved repositories from {0}".format(
                        pkgsurfarchive))
                shutil.rmtree(pkgsurfarchive)

            try:
                self.logger.log(
                    TINFO_LEVEL,
                    "Moving existing pkgarchive repositories from "
                    "{0} to {1}".format(pkgarchive, pkgsurfarchive))
                try:
                    os.makedirs(os.path.dirname(pkgsurfarchive))
                except OSError:
                    pass
                os.rename(pkgarchive, pkgsurfarchive)
            except OSError as e:
                # see if we attempted to do a cross-device rename
                if e.errno == errno.EXDEV:
                    self.logger.log(
                        TINFO_LEVEL,
                        "Unable to rename {0}, cross-device link. "
                        "Copying tree instead.".format(pkgarchive))
                    shutil.copytree(pkgarchive, pkgsurfarchive)
                    shutil.rmtree(pkgarchive)
                else:
                    self.logger.log(TWARNING_LEVEL, e)
                    return False
            return True

        if not self.bldenv.pkgsurf:
            self.logger.log(
                TINFO_LEVEL,
                "pkgsurf(1) disabled, not saving packages.")
        elif self.bldenv.generate_splice:
            self.logger.log(
                TINFO_LEVEL,
                "generate_splice set, not saving packages.")
        elif self.bldenv.pkgmerge:
            do_save_pkgs(debug=True, pkgmerge=True)
        else:
            saved = True
            if self.bldenv.debug:
                saved = do_save_pkgs(debug=True)

            # don't try to save nondebug if debug saving was attempted
            # but failed
            if self.bldenv.nondebug and saved:
                do_save_pkgs()


PKGMERGE_MONO_ARCH_BI_FLAVOR, \
    PKGMERGE_BI_ARCH_MONO_FLAVOR, \
    PKGMERGE_BI_ARCH_BI_FLAVOR, \
    PKGMERGE_NONE = list(range(4))


def _compute_pkgmerge_params(bldenv):
    """Given the current build environment, return a tuple describing the
    type of pkgmerge we could perform, the destination repository for that
    merge, and the paths for our:
     - local debug packages
     - remote debug packages
     - local non-debug packages
     - remote non-debug packages.
    """

    debug_pkgs_local = None
    debug_pkgs_altws = None
    nondebug_pkgs_local = None
    nondebug_pkgs_altws = None

    env = None

    if bldenv.debug:
        env = bldenv.get_build_environment(
            debug=True, nondebug=False)
        debug_pkgs_local = env.get("PKGARCHIVE")
        debug_pkgs_altws = env.get("ALTWS_PKGARCHIVE")

    if bldenv.nondebug:
        env = bldenv.get_build_environment(
            debug=False, nondebug=True)
        nondebug_pkgs_local = env.get("PKGARCHIVE")
        nondebug_pkgs_altws = env.get("ALTWS_PKGARCHIVE")

    # this value is consistent across debug/nondebug flavors
    branch = env["ONBLD_BRANCH_PREFIX"]

    if debug_pkgs_local and nondebug_pkgs_local and \
            nondebug_pkgs_local and nondebug_pkgs_altws:
        pkgmerge_type = PKGMERGE_BI_ARCH_BI_FLAVOR
        pkgmerge_dest = os.path.abspath(
            "{0}/../{1}merged".format(env["PKGARCHIVE"], branch))

    elif debug_pkgs_local and nondebug_pkgs_local:
        pkgmerge_type = PKGMERGE_MONO_ARCH_BI_FLAVOR
        pkgmerge_dest = os.path.abspath(
            "{0}/../{1}merged-{2}".format(
                env["PKGARCHIVE"], branch, env["MACH"]))

    elif debug_pkgs_local and debug_pkgs_altws:
        pkgmerge_type = PKGMERGE_BI_ARCH_MONO_FLAVOR
        pkgmerge_dest = os.path.abspath(
            "{0}/../{1}merged-debug".format(env["PKGARCHIVE"], branch))

    elif nondebug_pkgs_local and nondebug_pkgs_altws:
        pkgmerge_type = PKGMERGE_BI_ARCH_MONO_FLAVOR
        pkgmerge_dest = os.path.abspath(
            "{0}/../{1}merged-nondebug".format(env["PKGARCHIVE"], branch))

    else:
        pkgmerge_type = PKGMERGE_NONE
        pkgmerge_dest = None

    return (
        pkgmerge_type, pkgmerge_dest,
        debug_pkgs_local, debug_pkgs_altws,
        nondebug_pkgs_local, nondebug_pkgs_altws)


class PkgmergeTask(Task):
    task_name = "pkgmerge"
    task_synopsis = "Merge packages from one or more repositories"
    snapshot = True
    require_workspace_lock = True
    check_tools = False

    def _execute(self):

        if not self.env.get("PKGARCHIVE"):
            self.logger.log(
                TWARNING_LEVEL,
                "Unable to pkgmerge - 'pkgarchive' set to '<None>'")
            self.exit_status = 1
            return

        def add_pkgmerge_args(arg_list, pkg_name, arg, repo_path):
            """Add a (pkgmerge_arg, pkg_ver) tuple to the arg_list
            using pkg_name to determine pkg version which eventually
            gets used as a sanity-check to see if all package repositories
            contain the same package versions."""
            cmd = "pkgrepo -s {0} list -H {1}@latest".format(
                repo_path, pkg_name)
            ret, out, err = util.cmdline_run(cmd, env_arg=self.env)
            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Unable to find package version for {0} in {1}".format(
                        pkg_name, repo_path))
                self.logger.log(TWARNING_LEVEL, out)
                self.logger.log(TWARNING_LEVEL, err)
                arg_list.append(("{0}{1}".format(arg, repo_path), None))
            else:
                ver = out.split()[-1].split(":")[0]
                arg_list.append(("{0}{1}".format(arg, repo_path), ver))
            return

        def run_pkgmerge(
                arg_list, dest_repo, description, canary_pkg, default_pub,
                build_flavor="unknown"):
            """Run a pkgmerge using the list of arguments, writing to the
            destination repository.

            We verify that the package versions of the 'canary_pkg' stored
            in the arg_list are the same, reporting an error otherwise. Note
            that pkgmerge *doesn't* have this restriction, but it's a useful
            safety mechanism for ON to enforce so that users are always
            merging repositories from complementary builds.
            """
            self.logger.log(TINFO_LEVEL, description)
            versions = set()

            merge_cmd = "pkgmerge"
            bad_ver = False
            for arg, ver in arg_list:
                if ver is None:
                    bad_ver = True
                merge_cmd += " {0}".format(arg)
                versions.add(ver)

            # This is a sanity check, pkgmerge may still fail because
            # packages other than the ones we checked have different,
            # versions.
            if len(versions) > 1 or bad_ver:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Repositories were missing or invalid, or we found "
                    "different package versions for pkg:/{0} in repositories "
                    "for this merge!".format(canary_pkg))
                self.logger.log(TWARNING_LEVEL, "{:<30} {}".format(
                    "PKG VERSION", "PKGMERGE ARGS"))
                for arg, ver in arg_list:
                    self.logger.log(
                        TWARNING_LEVEL, "{:<30} {}".format(ver, arg))
                return 1

            if os.path.exists(dest_repo):
                try:
                    self.logger.log(
                        TINFO_LEVEL,
                        "Removing existing repository {0}".format(dest_repo))
                    shutil.rmtree(dest_repo)
                except OSError as e:
                    self.logger.log(
                        TWARNING_LEVEL, "Unable to remove existing "
                        "repository {0}: {1}".format(dest_repo, e),
                        exc_info=True)
                    return 1

            repo_cmd = "pkgrepo create {0}".format(dest_repo)
            ret = util.cmdline_run(
                repo_cmd, logger=self.logger, env_arg=self.env)
            if ret != 0:
                self.logger.log(TWARNING_LEVEL, "pkgrepo create failed")
                return ret

            prepo_set = [
                "publisher/prefix={0}".format(default_pub),
                "osnet/merge-description=\"{0}\"".format(description),
                "osnet/merged=true"
            ]
            repo_cmd = "pkgrepo -s {0} set {1}".format(
                dest_repo, " ".join(prepo_set))
            util.cmdline_run(repo_cmd, logger=self.logger, env_arg=self.env)

            hg_info = get_hg_info(self.ws, self.env)

            merge_cmd += " -d {0}".format(dest_repo)
            ret = util.cmdline_run(
                merge_cmd, logger=self.logger, env_arg=self.env)
            if ret != 0:
                self.logger.log(TWARNING_LEVEL, "pkgmerge failed")

            cmd = "pkgrepo -s {0} refresh".format(dest_repo)
            ret += util.cmdline_run(cmd, logger=self.logger, env_arg=self.env)

            _add_repository_metadata(
                None, self.ws, hg_info, self.env, self.logger,
                build_flavor=build_flavor, repo=dest_repo)

            return ret

        pkgmerge_type, pkgmerge_dest, \
            debug_pkgs_local, debug_pkgs_altws, \
            nondebug_pkgs_local, nondebug_pkgs_altws = \
            _compute_pkgmerge_params(self.bldenv)

        pkgmerge_args = []
        alt_mach = None

        noop_msg = "Not doing a pkgmerge for this build since the {0} " \
            "pkgarchive path is identical to the altws_pkgarchive path"
        noop_pkgmerge = False

        if self.bldenv.debug:
            if debug_pkgs_local == debug_pkgs_altws:
                self.logger.log(
                    TINFO_LEVEL, noop_msg.format("debug"))
                noop_pkgmerge = True

        if self.bldenv.nondebug:
            if nondebug_pkgs_local == nondebug_pkgs_altws:
                self.logger.log(
                    TINFO_LEVEL, noop_msg.format("non-debug"))
                noop_pkgmerge = True

        if noop_pkgmerge:
            self.exit_status = 0
            return

        if debug_pkgs_altws or nondebug_pkgs_altws:
            if self.mach == "sparc":
                alt_mach = "i386"
            else:
                alt_mach = "sparc"

        # If we have set ALTWS_KEEP, then we need to wait till the relevant
        # parts of that build have completed.
        if self.env.get("ALTWS_KEEP"):
            # the environment has already validated $ALTWS_KEEP
            mach, path = self.env.get("ALTWS_KEEP").split(":", 1)
            # We allow the alternative architecture workspace to also be
            # the current workspace (e.g. the same path, nfs-shared to two
            # build machines) but require that one build machine is the
            # 'alternative'. If the alternative workspace is the same as
            # this one, there's no pkgmerging to do during this build.
            if mach == self.mach:
                self.logger.log(
                    TINFO_LEVEL,
                    "'altws_keep' {0} is the same architecture as the current "
                    "build, {1}. Not doing pkgmerge as part of this "
                    "build.".format(path, mach))
                self.exit_status = 0
                return
            elif mach != alt_mach:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Unexepected 'altws_keep' architecture {0}, was expecting "
                    "{1}".format(mach, alt_mach))
                self.exit_status = 1
                return

            try:
                branch = self.env["ONBLD_BRANCH_PREFIX"]
                if debug_pkgs_altws:
                    self.wait_task(
                        "{0}packages".format(branch), path, mach,
                        timeout=int(
                            self.env.get("PKGMERGE_ALTWS_TIMEOUT", "3600")))
                if nondebug_pkgs_altws:
                    self.wait_task(
                        "{0}packages-nd".format(branch), path, mach,
                        timeout=int(
                            self.env.get("PKGMERGE_ALTWS_TIMEOUT", "3600")))
            except TaskException as e:
                self.logger.log(
                    TWARNING_LEVEL, "Unable to wait for task: {0}".format(e))
                self.exit_status = 1
                return

        for flavor, pkg_path in [
                ("debug", debug_pkgs_altws),
                ("non-debug", nondebug_pkgs_altws)]:
            if pkg_path and not os.path.isdir(pkg_path):
                self.logger.log(
                    TWARNING_LEVEL,
                    "'altws_pkgarchive' was set, but {0} packages "
                    "were not found at {1}".format(
                        flavor, pkg_path))
                self.exit_status = 1
                return

        # We want to check that our repositories are nominally related to ON
        # and contain matching package versions. To do that, we need to
        # define packages we expect to find in each repository.
        # tuples of the form
        # (<repository name>, <package to version-check>, <default publisher>)
        repo_sanity_data = [("repo.osnet", "system/core-os", "nightly")]
        if self.env["CLOSED_IS_PRESENT"] == "yes":
            repo_sanity_data.append(
                ("repo.osnet-internal", "test/gk-test", "extra"))

        #
        # We now have information for the 4 possible outcomes:
        #
        # * 4-way debug/nondebug/sparc/i386 pkgmerge
        # * 2-way sparc/i386 single-flavor pkgmerge
        # * 2-way debug/nondebug single-arch pkgmerge
        # * Anything else - an error
        #
        if pkgmerge_type == PKGMERGE_BI_ARCH_BI_FLAVOR:

            for repo_name, canary_pkg, default_pub in repo_sanity_data:
                pkgmerge_args = []
                dest_repo = os.path.join(pkgmerge_dest, repo_name)

                description = "4-way debug/nondebug/sparc/i386 pkgmerge"
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},debug.osnet=true,".format(self.mach),
                    os.path.join(debug_pkgs_local, repo_name))
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},debug.osnet=true,".format(alt_mach),
                    os.path.join(debug_pkgs_altws, repo_name))
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},debug.osnet=false,".format(self.mach),
                    os.path.join(nondebug_pkgs_local, repo_name))
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},debug.osnet=false,".format(alt_mach),
                    os.path.join(nondebug_pkgs_altws, repo_name))
                self.exit_status += run_pkgmerge(
                    pkgmerge_args, dest_repo, description, canary_pkg,
                    default_pub, build_flavor="both")

        elif pkgmerge_type == PKGMERGE_BI_ARCH_MONO_FLAVOR:

            description = "2-way sparc/i386 single-flavor pkgmerge"
            if debug_pkgs_local:
                local_pkgs = debug_pkgs_local
                altws_pkgs = debug_pkgs_altws
                build_flavor = "debug"
            else:
                local_pkgs = nondebug_pkgs_local
                altws_pkgs = nondebug_pkgs_altws
                build_flavor = "nondebug"
            for repo_name, canary_pkg, default_pub in repo_sanity_data:
                pkgmerge_args = []
                dest_repo = os.path.join(pkgmerge_dest, repo_name)
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},".format(self.mach),
                    os.path.join(local_pkgs, repo_name))
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s arch={0},".format(alt_mach),
                    os.path.join(altws_pkgs, repo_name))
                self.exit_status += run_pkgmerge(
                    pkgmerge_args, dest_repo, description, canary_pkg,
                    default_pub, build_flavor=build_flavor)

        elif pkgmerge_type == PKGMERGE_MONO_ARCH_BI_FLAVOR:

            description = "2-way debug/nondebug single-arch pkgmerge"
            for repo_name, canary_pkg, default_pub in repo_sanity_data:
                pkgmerge_args = []
                dest_repo = os.path.join(pkgmerge_dest, repo_name)
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s debug.osnet=true,",
                    os.path.join(debug_pkgs_local, repo_name))
                add_pkgmerge_args(
                    pkgmerge_args, canary_pkg,
                    "-s debug.osnet=false,",
                    os.path.join(nondebug_pkgs_local, repo_name))
                self.exit_status += run_pkgmerge(
                    pkgmerge_args, dest_repo, description, canary_pkg,
                    default_pub, build_flavor="both")

        else:
            self.logger.log(
                TWARNING_LEVEL, "Unable to pkgmerge when only a single "
                "build flavor (debug or nondebug) was specified.")
            self.exit_status = 1
            return

        if self.exit_status > 0:
            self.exit_status = 1


class ArchiveTask(Task):
    task_name = "archive"
    task_synopsis = "Archive build products"
    snapshot = False
    require_workspace_lock = True
    check_tools = False
    dir_pairs = []

    def _execute(self):

        def do_archive(src, dst, ignore=None):
            self.logger.log(TINFO_LEVEL,
                            "Archiving {0} to {1}".format(src, dst))
            if os.path.exists(dst):
                shutil.rmtree(dst)
            shutil.copytree(src, dst, symlinks=True, ignore=ignore)

        envs = []
        if self.bldenv.debug is True:
            envs.append(self.bldenv.get_build_environment(
                debug=True, nondebug=False))
        if self.bldenv.nondebug is True:
            envs.append(self.bldenv.get_build_environment(
                debug=False, nondebug=True))

        keep = envs[0].get("ONBLD_KEEP")
        mach = envs[0].get("MACH")
        archive_to = envs[0].get("ARCHIVE_TO_DIR")
        kid = envs[0].get("BUILDID_THISBUILD")

        archive_log_dir = os.path.join(archive_to,
                                       os.path.basename(self.log_dir))
        if not os.path.exists(archive_log_dir):
            self.logger.log(TINFO_LEVEL,
                            "Initializing {0}".format(archive_log_dir))
            os.makedirs(archive_log_dir)

        do_archive(
            os.path.join(keep, mach, "signing-material"),
            os.path.join(archive_to, "signing-material-{0}".format(mach)))

        if self.bldenv.pkgmerge:
            pkgmerge_type, pkgmerge_dest, \
                debug_pkgs_local, debug_pkgs_altws, \
                nondebug_pkgs_local, nondebug_pkgs_altws = \
                _compute_pkgmerge_params(self.bldenv)
            do_archive(
                pkgmerge_dest,
                os.path.normpath(os.path.join(
                    envs[0].get("KSPLICE_SPLICE_PKGARCHIVE"),
                    "..", os.path.basename(pkgmerge_dest))))

        for env in envs:
            kbr = env.get("KSPLICE_BASELINE_ROOT")
            ksr = env.get("KSPLICE_SPLICE_ROOT")
            ksr_limited = "{0}-limited".format(ksr)
            ksbl = "{0}-baseline".format(ksr)
            ksbl_full = "{0}-baseline-full".format(ksr)
            ksksr = env.get("KSPLICE_SPLICE_KSPLICEROOT")
            build = env.get("BUILD")
            pdir_rel = os.path.join("usr", "src", "pkg",
                                    "packages.{0}".format(mach))
            pdir = os.path.join(build, pdir_rel)

            do_archive(
                build, env.get("KSPLICE_SPLICE_BUILD"),
                ignore=shutil.ignore_patterns(
                    os.path.join(pdir_rel, "limited-root")))
            do_archive(env.get("PKGARCHIVE"),
                       env.get("KSPLICE_SPLICE_PKGARCHIVE"))
            do_archive(env.get("ROOT"), ksr)

            if os.path.exists(ksbl):
                os.unlink(ksbl)
            if os.path.exists(ksbl_full):
                os.unlink(ksbl_full)
            if not self.bldenv.generate_splice:
                self.logger.log(
                    TINFO_LEVEL,
                    "Creating {0} and {1} links in baseline archive".format(
                        ksbl, ksbl_full))
                os.symlink(os.path.basename(ksr_limited), ksbl)
                os.symlink(os.path.basename(ksr), ksbl_full)
                self.logger.log(
                    TINFO_LEVEL,
                    "Creating empty {0} in baseline archive".format(ksksr))
                if os.path.exists(ksksr):
                    shutil.rmtree(ksksr)
                os.mkdir(ksksr)
                do_archive(os.path.join(pdir, "limited-root"), ksr_limited)
            else:
                kde = "ksplice-{0}-delivery-exceptions".format(kid)
                self.logger.log(
                    TINFO_LEVEL,
                    "Creating {0} and {1} links in splice archive".format(
                        ksbl, ksbl_full))
                os.symlink("{0}-baseline".format(kbr), ksbl)
                os.symlink("{0}-baseline-full".format(kbr), ksbl_full)
                do_archive(os.path.join(pdir, kde),
                           os.path.join(archive_to, kde))
                do_archive(os.path.join(pdir, "combined-kspliceroot"), ksksr)


class PkgsurfTask(Task):
    task_name = "pkgsurf"
    task_synopsis = "Resurface package repositories"
    snapshot = True
    require_workspace_lock = True
    check_tools = False

    def _execute(self):

        def do_pkgsurf(debug=False, pkgmerge=False):
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug)
            pkgarchive = env.get("PKGARCHIVE")
            if not self.bldenv.generate_splice:
                pkgsurfarchive = env.get("PKGSURFARCHIVE")
            else:
                pkgsurfarchive = env.get("KSPLICE_BASELINE_PKGARCHIVE")

            if pkgarchive is None or pkgsurfarchive is None:
                self.logger.log(
                    TWARNING_LEVEL,
                    "pkgsurf(1) enabled for this build but no 'pkgarchive' or "
                    "'pkgsurfarchive' was set!")
                return 1

            if pkgmerge:
                # work out the name of merged archive we should resurface.
                # We only care about the pkgmerge_dest return value.
                pkgmerge_type, pkgmerge_dest, \
                    debug_pkgs_local, debug_pkgs_altws, \
                    nondebug_pkgs_local, nondebug_pkgs_altws = \
                    _compute_pkgmerge_params(self.bldenv)
                pkgarchive = pkgmerge_dest
                # If pkgmerge is set, then the repositories to be
                # resurfaced are merged. *Always* use the
                # corresponding merged reference repositories.
                pkgsurfarchive = os.path.normpath(os.path.join(
                    pkgsurfarchive, "..", os.path.basename(pkgmerge_dest)))

            self.logger.log(
                TINFO_LEVEL, "Target repositories: {0}".format(pkgarchive))
            self.logger.log(
                TINFO_LEVEL, "Reference repositories: {0}".format(
                    pkgsurfarchive))

            repos = [("repo.osnet", ["nightly", "localizable"])]
            exclude_pkgs = {
                "nightly": ["system/osnet-splice"],
                "localizable": [],
                "extra": [
                    "consolidation/osnet/osnet-internal-incorporation"]}
            if self.env["CLOSED_IS_PRESENT"] == "yes":
                repos.append(("repo.osnet-internal", ["extra"]))

            ret = 0
            for repo_name, pubs in repos:
                pkgrepo = os.path.join(pkgarchive, repo_name)
                pkgrepo_ref = os.path.join(pkgsurfarchive, repo_name)

                if not os.path.exists(pkgrepo):
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Unable to resurface {0}: "
                        "no such repository".format(pkgrepo))
                    return 1

                # In freshly checked out workspaces, we won't have a reference
                # repository. At the same time, having pkgsurf enabled by
                # default is worthwhile, so we can't make this fatal.
                if not os.path.exists(pkgrepo_ref):
                    self.logger.log(
                        TINFO_LEVEL,
                        "Skipping resurfacing of {0}: "
                        "no such reference repository".format(pkgrepo_ref))
                    continue

                # Make a copy of the raw (unresurfaced) repository if it
                # hasn't already been resurfaced.
                cmd = "pkgrepo -s {0} get -H osnet/resurfaced ".format(pkgrepo)
                pkgrepo_ret, out, err = util.cmdline_run(cmd, env_arg=env)

                if not out.strip().endswith("true"):
                    pkgrepo_raw = "{0}.raw".format(pkgrepo)
                    raw_basename = os.path.basename(pkgrepo_raw)

                    # XXX incremental builds always destroy the packages
                    # repository at present, so we will never have a
                    # $PKGARCHIVE that has already been resurfaced from a
                    # previous build. If that changes, we might want to
                    # look into rotating old ".raw" repositories.
                    if os.path.exists(pkgrepo_raw):
                        self.logger.log(
                            TINFO_LEVEL,
                            "Removing old .raw repositories at {0}".format(
                                raw_basename))
                        shutil.rmtree(pkgrepo_raw)

                    self.logger.log(
                        TINFO_LEVEL, "Copying {0} to {1}".format(
                            repo_name, raw_basename))
                    shutil.copytree(pkgrepo, pkgrepo_raw)
                else:
                    self.logger.log(
                        TINFO_LEVEL, "Repository {0} has already been "
                        "resurfaced. Not saving raw version".format(repo_name))

                # XXX run pkgsurf for each publisher until 22086013 is fixed.
                for pub in pubs:
                    if exclude_pkgs[pub]:
                        exclude_opts = "-c " + " -c ".join(exclude_pkgs[pub])
                    else:
                        exclude_opts = ""

                    cmd = "pkgsurf -u -s {0} -r {1} {2} -p {3}".format(
                        pkgrepo, pkgrepo_ref, exclude_opts, pub)
                    ret += util.cmdline_run(
                        cmd, logger=self.logger,
                        env_arg=env)

                if ret == 0:
                    for key, value in [
                            ("osnet/resurfaced", "true"),
                            ("osnet/resurface-reference",
                                os.path.realpath(pkgrepo_ref))]:
                        cmd = "pkgrepo -s {0} set {1}={2}".format(
                            pkgrepo, key, value)
                        util.cmdline_run(cmd, logger=self.logger, env_arg=env)

                    hg_info = get_hg_info(self.ws, self.env)
                    _add_repository_metadata(
                        None, self.ws, hg_info, self.env, self.logger,
                        repo=pkgrepo,
                        build_flavor="debug" if debug else "nondebug")

            return ret

        if self.bldenv.pkgmerge:
            # resurface our already-merged repositories
            # no need to do this separately for debug/nondebug
            self.exit_status = do_pkgsurf(pkgmerge=True)
        else:
            if self.bldenv.debug and self.bldenv.nondebug:
                self.logger.log(
                    TINFO_LEVEL, "This build does not have 'pkgmerge' set.")
                self.logger.log(
                    TINFO_LEVEL, "While this will not cause pkgsurf to fail, "
                    "it does mean that the debug and non-debug repositories "
                    "produced by this build could be separately reversioned "
                    "to different package versions. Subsequent attempts to "
                    "pkgmerge those repositories would then fail. Proceed "
                    "with caution.")

            if self.bldenv.debug:
                self.exit_status += do_pkgsurf(debug=True)

            if self.bldenv.debug and self.bldenv.nondebug:
                self.logger.log(
                    TINFO_LEVEL,
                    "Debug pkgsurf finished, starting non-debug pkgsurf")
                log.switch_logs(
                    self.logger, logsuffix="-nd", old_logsuffix="",
                    level=TASK_LEVEL)

            if self.bldenv.nondebug:
                self.exit_status += do_pkgsurf()

        if self.exit_status > 0:
            self.exit_status = 1

    def summary(self, max_lines=100):
        """Use a custom summary for pkgsurf, because the log is short enough
        to just dump what's likely all of it in case of error."""

        if self.exit_status == 0:
            return ""

        logfile_path = os.path.join(self.log_dir, self.log_name)
        if not os.path.exists(logfile_path):
            return ""

        summary = []
        cols = log.log_prefix_cols[self.bldenv.log_format] - 1
        with open(logfile_path, "r") as logfile:
            for full_line in open(logfile_path, "r").readlines():
                line = full_line[cols:]
                summary.append(line)
                if len(summary) >= max_lines:
                    summary.append(".\n.\n(additional lines truncated)")
                    break

        return "".join(summary[:max_lines])


class PkgdiffTask(Task):
    task_name = "pkgdiff"
    task_synopsis = "Compare reference and resurfaced package repositories"
    snapshot = False
    require_workspace_lock = True
    check_tools = False

    def _execute(self):
        def do_pkgdiff(debug=False, pkgmerge=False):
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=not debug)
            pkgarchive = env.get("PKGARCHIVE")
            if not self.bldenv.generate_splice:
                pkgsurfarchive = env.get("PKGSURFARCHIVE")
            else:
                pkgsurfarchive = env.get("KSPLICE_BASELINE_PKGARCHIVE")

            if pkgmerge:
                # Work out the name of merged archive.
                # We only care about the pkgmerge_dest return value.
                pkgmerge_type, pkgmerge_dest, \
                    debug_pkgs_local, debug_pkgs_altws, \
                    nondebug_pkgs_local, nondebug_pkgs_altws = \
                    _compute_pkgmerge_params(self.bldenv)
                pkgarchive = pkgmerge_dest
                # If pkgmerge is set, then the repositories to be
                # resurfaced are merged. *Always* use the
                # corresponding merged reference repositories.
                file_suffix = os.path.basename(pkgmerge_dest)
                pkgsurfarchive = os.path.normpath(os.path.join(
                    pkgsurfarchive, "..", file_suffix))
            elif debug:
                file_suffix = "debug"
            else:
                file_suffix = "nondebug"

            self.logger.log(
                TINFO_LEVEL, "Resurfaced repositories: {0}".format(pkgarchive))
            self.logger.log(
                TINFO_LEVEL, "Reference repositories: {0}".format(
                    pkgsurfarchive))

            # In freshly checked out workspaces, we won't have a
            # reference repository. At the same time, having pkgsurf
            # enabled by default is worthwhile, so we can't make this
            # fatal.
            ret = 0
            if not os.path.exists(os.path.join(pkgsurfarchive, "repo.osnet")):
                self.logger.log(
                    TINFO_LEVEL,
                    "Skipping pkgdiff of {0}: "
                    "no reference repository".format(pkgsurfarchive))
            else:
                ret = util.cmdline_run(
                    "dmake diff_repos CMP_REF={0} CMP_TGT={1} CMP_SFX={2}".
                    format(pkgsurfarchive, pkgarchive, file_suffix),
                    env_arg=env, logger=self.logger,
                    cwd=os.path.join(self.src, "pkg"))

            return ret

        if self.bldenv.pkgmerge:
            # compare our already-merged repositories
            # no need to do this separately for debug/nondebug
            self.exit_status = do_pkgdiff(pkgmerge=True)
        else:
            if self.bldenv.debug:
                self.exit_status += do_pkgdiff(debug=True)

            if self.bldenv.debug and self.bldenv.nondebug:
                self.logger.log(
                    TINFO_LEVEL,
                    "Debug pkgdiff finished, starting non-debug pkgdiff")
                log.switch_logs(
                    self.logger, logsuffix="-nd", old_logsuffix="",
                    level=TASK_LEVEL)

            if self.bldenv.nondebug:
                self.exit_status += do_pkgdiff()

        if self.exit_status > 0:
            self.exit_status = 1


class UpdateDiagnosticTask(Task):
    """Download a new copy of the stackdb diagnostic database
    stackdb"""
    task_name = "update_diag_db"
    task_synopsis = \
        "Download a new copy of the stackdb diagnostic database"
    snapshot = False

    def _execute(self):
        """Run the task"""
        path_to_perfpit_db = "usr/src/cmd/stackdiag/stackdb"
        self.logger.log(TINFO_LEVEL, "Downloading updated diagnostic database")
        ret = util.cmdline_run(
            "dmake download",
            logger=self.logger, cwd=os.path.join(self.ws, path_to_perfpit_db),
            env_arg=self.env)
        if ret != 0:
            self.exit_status = 1
            self.logger.log(
                TWARNING_LEVEL, "dmake download failed in %s"
                % os.path.join(self.ws, path_to_perfpit_db))


class TestScript(object):
    """An object that describes a given build test script."""

    def __init__(self):
        self.source_regexps = []
        self.args = ""
        self.name = "No name"
        self.cfg_path = None
        self.cwd = None
        self.synopsis = "[ no synopsis available ]"
        self.privileges = "basic"
        self.arch = "all"
        self.passed = None
        self.script = None
        self.notice = None
        self.requires_install = False

    def __repr__(self):
        return repr((self.name, self.script, self.args))

    def report_status(self, logger):
        # Emit a standard output line for DIY to find.
        if self.passed == "skip":
            status = "UNSUPPORTED"
        elif self.passed:
            status = "PASS"
        else:
            status = "FAIL"

        logger.log(TINFO_LEVEL, "ON_unit_tests | RESULT | %s | %s | %s" %
                   (self.cfg_path, self.name, status))

    def run(self, logger, env):
        # Wrap the do_run() function so we always report the status.
        self.do_run(logger, env)
        self.report_status(logger)

    def do_run(self, logger, env):
        logger.log(TINFO_LEVEL, "Starting test %s" % self.name)
        logger.log(TINFO_LEVEL, "Test synopsis: %s" % self.synopsis)

        if not self.script:
            logger.log(
                TINFO_LEVEL, "No test script associated with %s" % self.name)
            self.passed = True
            return
        if self.arch != "all" and self.arch != env["MACH"]:
            logger.log(
                TINFO_LEVEL, "Test %s not appropriate for this architecture" %
                self.name)
            self.passed = "skip"
            return
        if self.requires_install and env['ROOT'] != '/':
            logger.log(
                TINFO_LEVEL, "Test %s must be installed and run on a system "
                "booted with matching bits" % self.name)
            self.passed = "skip"
            return
        try:
            st = os.stat(self.script)
            if not st.st_mode & (stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH):
                raise OSError("mode is %s" % oct(st.st_mode & 0o777))
        except OSError as e:
            logger.log(
                TWARNING_LEVEL, "Unable to execute %s: %s" % (self.script, e))
            logger.log(
                TWARNING_LEVEL, "Test %s from %s failed" %
                (self.name, self.cfg_path))
            self.passed = False
            return

        if self.privileges != "basic":
            cmdline = "/usr/bin/ppriv -q %s" % self.privileges
            ret = util.cmdline_run(
                cmdline, logger=logger, cwd=self.cwd, env_arg=env)
            if ret != 0:
                logger.log(
                    TWARNING_LEVEL, "Unable to execute %s from %s: "
                    "insufficient privileges - required %s" %
                    (self.name, self.cfg_path, self.privileges))
                # log an info-level message, which will be appear at the
                # command line
                logger.info("Skipped test %s" % self.name)
                self.passed = "skip"
                return

        if not self.cwd:
            self.cwd = env["TMPDIR"]

        cmdline = "%s %s" % (self.script, self.args)
        ret = util.cmdline_run(
            cmdline, logger=logger, cwd=self.cwd, env_arg=env)
        if ret == 0:
            self.passed = True
        else:
            self.passed = False
            logger.log(
                TWARNING_LEVEL, "Test %s from %s failed" %
                (self.name, self.cfg_path))
        return


def load_tests(ws, env, logger):
    """Loads all tests from a workspace test.d directory, returning a list of
    TestScript objects describing the tests. 'env' points to the os.environ
    that should be used when replacing a limited set of '@MACRO@' strings
    present in config file values for 'script' and 'args' values.

    A test descriptor file can consist of one or more sections. Each
    section should be of the format:

    [<test name>]
    synopsis = <description of test>
    script = <path to test script>
    args = <CLI args to run script with>
    arch = <optional architecture, either sparc or i386>
    requires_install = <optional, true if it cannot be run from workspace>
    cwd = <directory from which to run the test>
    privileges = <optional list of privileges>
    pattern.<desc1> = <regexp matching files that should trigger this test>
    .                   .
    .                   .
    pattern.<descN> = <regexp matching files that should trigger this test>
    notice = [optional notice to be used for pbchk]
    """

    tests = []
    test_sects = {}
    allowed_macros = [
        "WS", "ROOT", "SRC", "BUILD", "TMPDIR", "NATIVEROOT", "TOOLROOT"]

    # To allow use by other consolidations, define a list of test.d paths
    # where we expect to find cfg files. The first one that exists is used,
    # and we intentionally fall back to the tests installed on the test
    # machine.
    test_dirs = [
        "%s/usr/src/tools/test.d" % ws,
        "%s/tools" % ws,
        "/opt/onbld/test/test.d"
    ]
    test_dir = None
    for tdir in test_dirs:
        if os.path.exists(tdir):
            test_dir = tdir
            logger.log(TINFO_LEVEL, "Loading tests from %s" % tdir)
            break
    else:
        return []

    def replace_macros(value):
        val = value
        for macro in allowed_macros:
            if macro not in env:
                continue
            # Special case to handle ROOT=/
            if macro == "ROOT" and env["ROOT"] == "/":
                val = val.replace("TOOLROOT", "ROOT")
            val = val.replace("@%s@" % macro, env[macro])
        return val

    for entry in os.listdir(test_dir):
        if entry.endswith(".cfg"):
            cfg_path = "%s/%s" % (test_dir, entry)
            if not os.path.isfile(cfg_path):
                continue
            try:
                cfg = configparser.RawConfigParser()
                cfg.read(cfg_path)
                sections = cfg.sections()
                if not sections:
                    raise configparser.Error("no sections found")

                for sect in sections:
                    # keep track of the section names, and the number
                    # of times they occur to guarantee uniqueness.
                    test_sects.setdefault(sect, 0)
                    test_sects[sect] += 1
                    test = TestScript()
                    test.name = sect
                    try:
                        test.script = cfg.get(sect, "script")
                    except configparser.NoOptionError:
                        pass
                    test.cfg_path = cfg_path

                    if test.script and not test.script.startswith("/"):
                        # we don't have an absolute path, macros are fine
                        test.script = replace_macros(test.script)

                    # some optional settings in the config file
                    for key in [
                            "args", "cwd", "privileges", "synopsis",
                            "notice", "arch", "requires_install"]:
                        try:
                            val = replace_macros(cfg.get(sect, key))
                            test.__setattr__(key, val)
                        except configparser.NoOptionError:
                            pass

                    for key in sorted(cfg.options(sect)):
                        if key.startswith("pattern."):
                            try:
                                regexp = cfg.get(sect, key)
                                test.source_regexps.append(
                                    (re.compile(regexp), regexp))
                            except re.error:
                                logger.log(
                                    TWARNING_LEVEL,
                                    "Skipping bad RE in %s: %s" % (
                                        cfg_path, regexp))

                    # Always add the .cfg file to each section so changes
                    # to it will also trigger the tests.
                    test.source_regexps.append(
                        (re.compile(cfg_path), cfg_path))

                    tests.append(test)
            except configparser.Error as e:
                raise ValueError("Unable to parse test config: %s" % e)

    dup_sects = [(name, num)
                 for name, num in list(test_sects.items()) if num > 1]
    if dup_sects:
        raise ValueError(
            "Section names must be unique but duplicates were "
            "discovered: %s" % "\n".join(
                item[0] for item in dup_sects))

    def tests_keycmp(item):
        return (item.name, item.cfg_path)

    return sorted(tests, key=tests_keycmp)


def get_skipped_tests(tests):
    return [test.name for test in tests if test.passed == "skip"]


def get_failed_tests(tests):
    return [test.name for test in tests if test.passed is False]


class ONTestTask(Task):
    """This runs a one or more tests known to the build. We name tests with
    any combination of the test.cfg basename that describes the tests, or
    the individual section name within any one test.cfg file."""
    task_name = "test"
    task_synopsis = "Runs tests matching test.d/*.cfg file or section names"
    check_tools = False

    def _execute(self):
        try:
            all_tests = load_tests(self.ws, self.env, self.logger)
        except ValueError as e:
            self.logger.log(TWARNING_LEVEL, e)
            self.exit_status = 1
            return

        if not all_tests:
            self.logger.error("No tests found!")
            self.exit_status = 1
            return

        # list the available tests. Note that we explicitly use
        # print here, rather than logging.
        if not self.args:
            log.destroy_logfiles(self.logger)
            print("\nUsage: build test <file name or section name>\n")

            print("\n{0:20} {1:30} {2}".format("FILE", "SECTION NAME",
                  "SYNOPSIS"))
            for test in all_tests:
                test_base = os.path.basename(test.cfg_path)
                print("{0:20} {1:22} {2}".format(test_base, test.name,
                      test.synopsis))
            print()
            return

        if os.path.exists("/etc/versions/osnet"):
            self.logger.log(TINFO_LEVEL, "/etc/versions/osnet contents:")
            with open("/etc/versions/osnet") as versions:
                self.logger.log(TINFO_LEVEL, versions.read().strip())

        known_tests = set([test.name for test in all_tests])
        known_tests.update(
            [os.path.basename(test.cfg_path) for test in all_tests])

        unknown_tests = set(self.args) - known_tests
        if unknown_tests:
            self.logger.error(
                "Unknown tests: %s" % " ".join(sorted(unknown_tests)))
            self.exit_status = 1
            return

        for test in all_tests:
            test_base = os.path.basename(test.cfg_path)
            if test.name in self.args or test_base in self.args:
                self.logger.info("Running test %s" % test.name)
                test.run(self.logger, self.env)
                self.logger.log(TINFO_LEVEL, "\n")
                if not test.passed:
                    self.exit_status = 1

        failed_tests = get_failed_tests(all_tests)
        skipped_tests = get_skipped_tests(all_tests)
        if skipped_tests:
            self.logger.info(
                "The following tests were skipped:\n\t%s" %
                "\n\t".join(skipped_tests))
        if failed_tests:
            self.logger.info(
                "The following tests failed:\n\t%s" %
                "\n\t".join(failed_tests))


class ExceptionListCheckTask(Task):
    """This task validates exception lists which contain names of source files
    and warns if any entries listed don't exist. This will then keep the lists
    from becoming stale.
    """

    task_name = "check.exceptions"
    task_synopsis = "Validate the exception lists against the source tree"
    snapshot = False
    enabled_by_default = True
    check_tools = False

    def _execute(self):
        lists_to_check = [
            "cddlchk",
            "copyright",
            "cstyle",
            "hdrchk",
            "jsstyle",
            "keywords",
            "makefilechk",
            "mapfilechk",
            "permchk"
        ]
        failcnt = 0

        self.logger.log(TINFO_LEVEL, "Validating exception lists:")
        ws = self.ws
        for repo in [
                ws,
                "%s/usr/closed" % ws,
                "%s/usr/man" % ws,
                "%s/usr/fish" % ws]:

            for elist in lists_to_check:
                linecnt = 0
                try:
                    excpt_list = "%s/exception_lists/%s" % (repo, elist)
                    with open(excpt_list) as excpt:
                        self.logger.log(TINFO_LEVEL, "\t%s" %
                                        (excpt_list.replace(ws + "/", "")))
                        for line in excpt:
                            linecnt += 1
                            line = line.strip()
                            if line.startswith("usr/"):
                                if len(glob.glob(
                                       os.path.join(repo, line),
                                       recursive=True)) == 0:
                                    self.logger.log(TWARNING_LEVEL,
                                                    "%s: %d: Unused entry: %s",
                                                    elist, linecnt, line)
                                    failcnt += 1
                except FileNotFoundError:
                    pass
        self.logger.log(TINFO_LEVEL, "Complete - %d failures" % failcnt)
        self.exit_status = failcnt


class CheckBuildTest(Task):
    """This task runs all tests known to the build. Tests are executed
    serially, in alphabetical order of the test names defined in each *.cfg
    test section."""
    task_name = "check.all_tests"
    task_synopsis = "Run all tests known to the workspace"
    enabled_by_default = False
    check_tools = False

    def _execute(self):

        if os.path.exists("/etc/versions/osnet"):
            self.logger.log(TINFO_LEVEL, "/etc/versions/osnet contents:")
            with open("/etc/versions/osnet") as versions:
                self.logger.log(TINFO_LEVEL, versions.read().strip())
        try:
            all_tests = load_tests(self.ws, self.env, self.logger)
        except ValueError as e:
            self.logger.log(TWARNING_LEVEL, e)
            self.exit_status = 1
            return

        for test in all_tests:
            test.run(self.logger, self.env)
            self.logger.log(TINFO_LEVEL, "\n")
            if not test.passed:
                self.exit_status = 1

        failed_tests = get_failed_tests(all_tests)
        skipped_tests = get_skipped_tests(all_tests)
        if skipped_tests:
            self.logger.log(
                TWARNING_LEVEL,
                "The following tests were skipped:\n\t%s" %
                "\n\t".join(skipped_tests))
        if failed_tests:
            self.logger.log(
                TWARNING_LEVEL,
                "The following tests failed:\n\t%s" %
                "\n\t".join(failed_tests))


class AllTests(CheckBuildTest):
    """This class exists simply to provide a subcommand synonym for
    CheckBuildTest."""
    task_name = "all_tests"
    task_synopsis = "(a synonym for 'check.all_tests')"


class CheckHgBuildTest(Task):
    """This task runs unit tests associated with the source files that have
    changed since the last build. Tests are executed serially, in alphabetical
    order of the test names defined in each *.cfg test section that contains
    a pattern.* entry that matches any of the changed files."""

    task_name = "check.tests"
    task_synopsis = "Run tests for sources changed since the last build"
    enabled_by_default = False
    check_tools = False

    def _get_changed_files(self):
        """If we're running this task as part of 'nightly' or 'qnightly'
        then .build.$MACH.tstamp refers to the build being performed now
        and .build.$MACH.tstamp.prev is the previous build.

        If we're running this task as part of "check.tests" then
        .build.$MACH.tstamp is the previous build."""

        if "check.tests" in sys.argv:
            tstamp_path = "%s/.build.%s.tstamp" % (self.keep, self.mach)
        else:
            tstamp_path = "%s/.build.%s.tstamp.prev" % (self.keep, self.mach)

        self.logger.log(
            TINFO_LEVEL, "Determining changed files in %s since %s" %
            (self.ws, tstamp_path))

        hg_info = get_hg_info(self.ws, self.env, hg_identify=True)
        hg_changed_files = set()
        changed_files = set()
        old_hg_info = {}

        if os.path.exists(tstamp_path):
            with open(tstamp_path, "r") as tstamp:
                for line in tstamp.readlines():
                    toks = line.strip().split()
                    if len(toks) >= 2:
                        # we only care about the workspace name and path
                        # and don't care about the number of heads, so set
                        # that to 1.
                        repo = toks[0].replace(self.ws, "")
                        old_hg_info[toks[0]] = (toks[1], 1)

        if old_hg_info:
            for repo in hg_info:
                if repo not in old_hg_info:
                    self.logger.log(
                        TWARNING_LEVEL, "No entry for %s in prev timestamp" %
                        repo)
                    continue

                # find the root of the repo, "", usr/closed or usr/man
                repo_root = repo.replace(self.ws, "").lstrip("/")

                # hg_info is a dictionary mapping the repo name to a tuple of
                # the form (<node info>, <num heads>). <node info> can look
                # like "h6a4cfb2ce79d+ tip", so we ignore the tag and remove
                # the '+' if necessary.
                old_cs = old_hg_info[repo][0].split()[0].rstrip("+")
                new_cs = hg_info[repo][0].split()[0].rstrip("+")

                # Determine the files that have changed using the
                # hg log from the old revision to the new revision, pruning
                # the old revision itself.
                cmd = "hg log -R %s -r %s:%s -P %s --template '{files} '" % \
                    (repo, old_cs, new_cs, old_cs)
                ret, out, err = util.cmdline_run(cmd)
                if ret != 0:
                    self.logger.log(
                        TWARNING_LEVEL,
                        "Unable to find differences between %s and %s for %s" %
                        (old_cs, new_cs, repo))
                if not out:
                    self.logger.log(
                        TINFO_LEVEL,
                        "No new hg changesets found for %s" % repo)
                    continue
                else:
                    for path in out.split():
                        path = os.path.join(repo_root, path)
                        hg_changed_files.add(path)

        for repo in hg_info:
            repo_root = repo.replace(self.ws, "").lstrip("/")
            # get changes to files that are not yet committed
            cmd = "hg status -R %s -mard" % repo
            ret, out, err = util.cmdline_run(cmd)

            if ret != 0:
                self.logger.log(
                    TWARNING_LEVEL,
                    "Unable to determine uncommited changed files in %s" %
                    repo)
            if not out:
                self.logger.log(
                    TINFO_LEVEL, "No uncommited changes in %s" % repo)
            else:
                try:
                    tstamp = os.stat(tstamp_path).st_mtime
                except OSError:
                    # the time stamp file might not exist
                    tstamp = None
                for line in out.split("\n"):
                    if not line:
                        continue

                    path = line[line.index(" ") + 1:]
                    abs_changed_path = os.path.join(self.ws, repo_root, path)
                    changed_path = os.path.join(repo_root, path)

                    try:
                        if not tstamp:
                            # no previous build, so treat all changes as new
                            changed_files.add(changed_path)
                        if not os.path.exists(abs_changed_path):
                            # a deleted file is something that could
                            # trigger a new test run. We can't tell how long
                            # ago it was deleted, unfortunately.
                            changed_files.add(changed_path)
                        elif os.stat(abs_changed_path).st_mtime > tstamp:
                            changed_files.add(changed_path)
                    except OSError:
                        # if we can't stat the file, we just ignore it
                        pass

        return hg_changed_files, changed_files

    def _execute(self):
        try:
            all_tests = load_tests(self.ws, self.env, self.logger)
        except ValueError as e:
            self.logger.log(TWARNING_LEVEL, e)
            self.exit_status = 1
            return

        hg_changed_files, changed_files = self._get_changed_files()
        if hg_changed_files:
            self.logger.log(
                TINFO_LEVEL,
                "Files modified in SCM since last build:")
        for change in sorted(hg_changed_files):
            self.logger.log(TINFO_LEVEL, change)

        if changed_files and hg_changed_files:
            self.logger.log(TINFO_LEVEL, "")

        if changed_files:
            self.logger.log(
                TINFO_LEVEL,
                "Files modified in the workspace since last build:")
        for change in sorted(changed_files):
            self.logger.log(TINFO_LEVEL, change)

        all_changed_files = set()
        all_changed_files.update(changed_files)
        all_changed_files.update(hg_changed_files)

        for test in all_tests:
            run_test = False
            for path in sorted(all_changed_files):
                for cpattern, pattern in test.source_regexps:
                    if cpattern.match(path):
                        self.logger.log(
                            TINFO_LEVEL,
                            "%s change triggers %s due to matching regexp %s" %
                            (path, test.name, pattern))
                        run_test = True
                        break
                # if we've already decided to run this test, no need to test
                # any more paths.
                if run_test:
                    break

            if run_test:
                test.run(self.logger, self.env)
                self.logger.log(TINFO_LEVEL, "")
                if not test.passed:
                    self.exit_status = 1
            else:
                self.logger.log(
                    TINFO_LEVEL, "Skipping %s: no patterns matched" %
                    test.name)

        failed_tests = get_failed_tests(all_tests)
        skipped_tests = get_skipped_tests(all_tests)
        if skipped_tests:
            self.logger.log(
                TWARNING_LEVEL,
                "The following tests were skipped:\n%s" %
                "\n".join(skipped_tests))
        if failed_tests:
            self.logger.log(
                TWARNING_LEVEL,
                "The following tests failed:\n%s" % "\n".join(failed_tests))


#
# A series of tasks that are only used for testing. These could be broken
# out into their own module in the future.
#

class TestTask(Task):
    """Test tasks should not write log files, so we override the default
    constructor to make this happen."""

    task_name = "buildtest.test_unused"
    snapshot = False
    check_tools = False

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(TestTask, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs)


class TestStacktrace(TestTask):
    """Blow up with a stack trace."""

    task_name = "buildtest.stacktrace"

    def _execute(self):
        """Blow up immediately."""
        raise Exception("Not the comfy chair!")


class TestLongRunning(TestTask):
    """A long running task that can be used to check things like
    KeyboardInterrupt exception handling and SIGTERM."""

    task_name = "buildtest.longrun"

    def _execute(self):
        """do nothing."""
        self.logger.log(TINFO_LEVEL, "Waiting for 30 seconds.")
        time.sleep(30)

    def _execute_finish(self):
        """All done"""
        super(TestLongRunning, self)._execute_finish()
        self.logger.log(TWARNING_LEVEL, "_execute_finish for longrun")


class TestTaskWait(TestTask):
    """Tests that we can wait for a buildtest.longrun task correctly"""

    task_name = "buildtest.wait"

    def _execute(self):
        """do nothing."""
        if not self.env.get("ALTWS_KEEP"):
            self.logger.info("No ALTWS_KEEP setting!")
            self.exit_status = 1
            return

        mach, path = self.env.get("ALTWS_KEEP").split(":", 1)
        self.logger.log(TINFO_LEVEL, "Waiting for buildtest.longrun.")
        self.wait_task("buildtest.longrun", path, mach, poll_interval=2)
        self.logger.log(TINFO_LEVEL, "buildtest.longrun completed")


class TestFail(TestTask):
    """A task that does nothing but fail."""

    task_name = "buildtest.fail"

    def _execute(self):
        self.logger.log(TWARNING_LEVEL, "failing now.")
        self.exit_status = 1


class TestCheckFail1(TestTask):
    """A check task that does nothing but fail and must always pass."""

    task_name = "check.buildtest.fail1"
    must_pass = True

    def _execute(self):
        self.logger.log(TWARNING_LEVEL, "failing now.")
        self.exit_status = 1


class TestCheckFail2(TestTask):
    """A check task that does nothing but fail."""

    task_name = "check.buildtest.fail2"
    must_pass = False

    def _execute(self):
        self.logger.log(TWARNING_LEVEL, "failing now.")
        self.exit_status = 1


class TestPass(TestTask):
    """A task that does nothing but pass."""

    task_name = "buildtest.pass"

    def _execute(self):
        known_tasks = self.get_known_tasks()
        self.logger.log(TINFO_LEVEL, "Known tasks:")
        for task in known_tasks:
            self.logger.log(TINFO_LEVEL, "    %s\n" % str(task))
        self.logger.log(TINFO_LEVEL, "all done - passing.")
        self.exit_status = 0


class TestPass1(TestPass):
    task_name = "buildtest.pass1"
    pass


class TestPass2(TestPass):
    task_name = "buildtest.pass2"
    pass


class TestComposite(CompositeTask):
    task_name = "buildtest.composite"

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        super(TestComposite, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)
        self.add_task(TestPass1(bldenv))
        self.add_task(TestPass2(bldenv))


class TestMultiPass(CompositeTask):
    task_name = "buildtest.multipass"

    def __init__(self, bldenv, args=None, fail_fast=False, level=TASK_LEVEL,
                 propagate_logs=True):
        super(TestMultiPass, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast)
        self.add_task(TestComposite(bldenv))
        self.add_task(TestComposite(bldenv))


class TestPrintBldEnv(TestTask):
    """A task that just prints our environment. This is also possible with
    'build -i bldrc /usr/bin/env' but this could be useful for the
    testsuite."""

    task_name = "buildtest.environment"

    def _execute(self):
        def check_cmp(item):
            """A 'comparator' function, actually a key mapping function."""
            if "must_pass" in item:
                return (1, item)
            return (0, item)

        for key in sorted(self.env):
            self.logger.log(STATE_LEVEL, "%s=%s" % (key, self.env[key]))
        for key in sorted(self.bldenv.overrides, key=check_cmp):
            self.logger.log(
                STATE_LEVEL, "%s:%s" %
                (key, str(self.bldenv.overrides[key]).lower()))


class TestDualFlavor(TestTask):
    """A task that runs in two flavors, allowing us to check that the build
    environment correctly switches flavor."""

    task_name = "buildtest.dual_flavor"

    def _execute(self):
        self.exit_status = 0

        def do_run(debug=False):
            """Simulate a run where we switch environments, emitting all
            environment variables."""

            nondebug = not debug
            env = self.bldenv.get_build_environment(
                debug=debug, nondebug=nondebug)
            for key in sorted(env):
                self.logger.log(TINFO_LEVEL, "%20s %-60s" % (key, env[key]))

        if self.bldenv.debug:
            do_run(debug=True)

        if self.bldenv.debug and self.bldenv.nondebug:
            self.logger.log(
                STATE_LEVEL, "\n\nDebug run finished, starting non-debug "
                "run\n\n")

        if self.bldenv.nondebug:
            do_run()


class TestHgInfo(TestTask):
    """A task that prints mercurial information for the workspace."""

    task_name = "buildtest.hginfo"

    def _execute(self):
        self.logger.log(STATE_LEVEL, "=== hg tip ===")
        info = get_hg_info(self.ws, self.env, force_cache_refresh=True)
        for key in info:
            self.logger.log(STATE_LEVEL, "%20s %-10s" % (key, info[key]))

        self.logger.log(STATE_LEVEL, "=== hg identify ===")
        info = get_hg_info(
            self.ws, self.env, hg_identify=True, force_cache_refresh=True)
        for key in info:
            self.logger.log(STATE_LEVEL, "%20s %-10s" % (key, info[key]))


class TestNightlyTask(NightlyTask):
    """TestNightly* allows us to test mail_msg generation, without having to
    run a full nightly build, so subclasses NightlyTask rather than TestTask
    so that NightlyTask's generate_mail_msg() method is called during
    self._execute_finish ."""

    snapshot = False
    task_name = "buildtest.nightly_unused"
    core_os_ver = "pkg:/system/core-os@5.6-5.6"
    entire_ver = "pkg:/entire@2.5.1"
    osnet_ver = "pkg:/developer/opensolaris/osnet@4.0"
    onbld_ver = "pkg:/developer/build/onbld@65533"
    onbld_int_ver = "pkg:/developer/build/onbld-internal@49149"
    pkgvers = "5.12.0.0.0.44.0"
    build_machine = socket.gethostname().split(".")[0]
    linkers_ver = ""

    def __init__(self, bldenv, args=None, fail_fast=True, level=TASK_LEVEL,
                 propagate_logs=True, no_logfiles=True):
        # specify avoid_task_creation so we don't inherit all of the Tasks
        # normally added by 'nightly'
        super(TestNightlyTask, self).__init__(
            bldenv, args=args, level=level, propagate_logs=propagate_logs,
            fail_fast=fail_fast, avoid_task_creation=True)


class TestNightlyFail(TestNightlyTask):
    """Simulate a failing nightly build."""
    task_name = "buildtest.nightly_fail"

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(TestNightlyFail, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs)
        self.add_task("This is a failing nightly build.")
        self.add_task(TestPass(bldenv))
        self.add_task(TestPass(bldenv))
        self.add_task(TestFail(bldenv))


class TestNightlyPass(TestNightlyTask):
    """Simulate a passing nightly build."""
    task_name = "buildtest.nightly_pass"

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(TestNightlyPass, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs)
        self.add_task("This is a passing nightly build.")
        self.add_task(TestPass(bldenv))
        self.add_task(TestPass(bldenv))
        self.add_task(TestComposite(bldenv))


class TestNightlyPassConditional(TestNightlyTask):
    """Simulate a nightly build that passes, but for which a check task
    fails."""
    task_name = "buildtest.nightly_conditional"

    def __init__(
            self, bldenv, args=None, level=TASK_LEVEL, propagate_logs=True,
            no_logfiles=True, name=None):
        super(TestNightlyPassConditional, self).__init__(
            bldenv, args=args, level=level, no_logfiles=no_logfiles,
            propagate_logs=propagate_logs, fail_fast=False)
        self.add_task("This is a conditional pass nightly build.")
        self.add_task(TestPass(bldenv))
        self.add_task(TestPass(bldenv))
        if not bldenv.check:
            self.add_task("Ignoring all checks")
            return
        if bldenv.overrides.get("check.buildtest.fail1", True):
            self.add_task(TestCheckFail1(bldenv))
        else:
            self.add_task("Skipping check.buildtest.fail1")
        if bldenv.overrides.get("check.buildtest.fail2", True):
            self.add_task(TestCheckFail2(bldenv))
        else:
            self.add_task("Skipping check.buildtest.fail2")
